/*
 * arith.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * arithmetic routine
 *      $Header: /Users/kako/asuka/src/RCS/arith.c,v 2.25 2016/09/06 05:50:02 kako Exp $
 */
static char const rcsid[] =
    "$Id: arith.c,v 2.25 2016/09/06 05:50:02 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"
#include "mathop.h"

#define RATIONAL_GCD
#define MILLER_RABIN 1
#define USE_MT

#ifdef USE_MT
#include "mt19937.h"
#endif

long            ratio_log2(Sex a);
Sex             reciprocal(Sex a);
Sex             gcd_y(Sex a, Sex b);
Sex             gcdjj(Sex a, Sex b);
Sex             gcdji(Sex a, Sex b);
Sex             Lbfloat_man(Sex a);
Sex             Lbfloat_exp(Sex a);
Sex             Lprimep(Sex a);
Sex             Lgen_prime(Sex a, Sex b);
Sex             Lnext_prime(Sex a);
Sex             Lprime_before(Sex a);

Sex             zfactor(Sex a);
Sex             zfactor1(Sex a, Sex f);

unsigned long   power_mod(unsigned long a, unsigned long b,
			  unsigned long n);
bool            miller_rabin(unsigned long n, unsigned long b);
bool            is_prime(long n);
Sex             modular_expt(Sex a, Sex b, Sex m);
Sex             modular_times(Sex a, Sex b, Sex m);
long            int_gcd(long a, long b);

extern long     mult2mm(long a, long b, long m);
extern long     pow2mm(long a, long b, long m);

/***************************************************************/
/***************************************************************/

// #define DEBUG_ARITH
#define DEBUG_SHOW 1

#ifdef DEBUG_ARITH
Sex             xplus2(Sex, Sex);
Sex             xdiff2(Sex, Sex);
Sex             xtimes2(Sex, Sex);
Sex             xquot2(Sex, Sex);
Sex             xremaind2(Sex, Sex);
extern void     debug_printf(char *fmt0, ...);
#endif


Sex
math_er(Sex a, Sex b)
{
    return aerror2("arithmetic operation", a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

two_args       *plus_table_1[] =
    { plus2il, plus2ib, plus2im, plus2ir, plus2if,
    plus2ie, plus2ic, plus2iv, plus2id, plus2iz, plus2iE, plus2iZ,
    plus2iV, plus2iM, plus2iq, plus2iY
};


two_args       *plus_table[TYPEMAX][TYPEMAX] = {
    {plus2ll, plus2lb, plus2lm, plus2lr, plus2lf,
     plus2le, plus2lc, plus2lv, plus2ld, plus2lz, plus2lE, plus2lZ,
     plus2lV, plus2lM, plus2lq, plus2lY}
    ,
    {math_er, plus2bb, plus2bm, plus2br, plus2bf,
     plus2be, plus2bc, plus2bv, plus2bd, plus2bz, plus2bE, plus2bZ,
     plus2bV, plus2bM, plus2bq, plus2bY}
    ,
    {math_er, math_er, plus2mm, plus2mr, plus2mf,
     plus2me, plus2mc, plus2mv, plus2md, plus2mz, plus2mE, plus2mZ,
     plus2mV, math_er, math_er, plus2mY}
    ,
    {math_er, math_er, math_er, plus2rr, plus2rf,
     plus2re, plus2rc, plus2rv, plus2rd, plus2rz, plus2rE, plus2rZ,
     plus2rV, plus2rM, plus2rq, plus2rY}
    ,
    {math_er, math_er, math_er, math_er, plus2ff,
     plus2fe, plus2fc, plus2fv, plus2fd, plus2fz, plus2fE, plus2fZ,
     plus2fV, math_er, math_er, plus2fY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     plus2ee, plus2ec, plus2ev, plus2ed, plus2ez, plus2eE, plus2eZ,
     plus2eV, math_er, math_er, plus2eY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, plus2cc, plus2cv, plus2cd, plus2cz, plus2cE, plus2cZ,
     plus2cV, plus2cM, plus2cq, plus2cY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, plus2vv, plus2vd, plus2vz, plus2vE, plus2vZ,
     plus2vV, math_er, math_er, plus2vY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, plus2dd, plus2dz, plus2dE, plus2dZ,
     plus2dV, math_er, math_er, plus2dY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, plus2zz, plus2zE, plus2zZ,
     plus2zV, math_er, math_er, plus2zY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, plus2EE, plus2EZ,
     plus2EV, math_er, math_er, plus2EY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, math_er, plus2ZZ,
     plus2ZV, math_er, math_er, plus2ZY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, math_er, math_er,
     plus2VV, math_er, math_er, plus2VY}
    ,
    {plus2Ml, plus2Mb, math_er, plus2Mr, math_er,
     math_er, plus2Mc, math_er, math_er, math_er, math_er, math_er,
     math_er, plus2MM, plus2Mq, plus2MY}
    ,
    {plus2ql, plus2qb, math_er, plus2qr, math_er,
     math_er, plus2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, plus2qM, plus2qq, plus2qY}
    ,
    {plus2Yl, plus2Yb, math_er, plus2Yr, math_er,
     math_er, plus2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, plus2YM, plus2Yq, plus2YY}
};

#ifdef DEBUG_ARITH

#if DEBUG_SHOW
Sex
plus2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    if (is_bigint(a) || is_bigint(b)) {
	stackcheck();
	push2(a, b);
	debug_printf("(equal (plus2 %p %p)", a, b);
	pop2(b, a);
	a = xplus2(a, b);
	push(a);
	debug_printf(" %p)\n", a);
	pop(a);
    } else {
	a = xplus2(a, b);
    }
    return a;
}
#else

Sex
plus2(a, b)
     Sex             a,
                     b;
{
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]
    c = xplus2(a, b);
    d = xdiff2(c, b);
    if (eql(d, a) == NO) {
	debug_printf("%p + %p != p  bat\n", a, b, c, d);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    a = stack[-2];
    popv(5);
    return a;
}
#endif


Sex
xplus2(a, b)
#else
Sex
plus2(a, b)
#endif
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n = mkint(a) + mkint(b);
	    long            t = n & fix_mask;
	    if (t == 0 || t == fix_mask)
		return mkfxat(n);
	    return mkbigat1(n);
	} else if (is_numhdr(b))
	    return plus_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return plus_table_1[numtype_of(a)] (b, a);
	else if (is_numhdr(b)) {
	    long            i = numtype_of(a);
	    long            j = numtype_of(b);
	    if (i > j)
		return plus_table[j][i] (b, a);
	    else
		return plus_table[i][j] (a, b);
	}
    }
    return aerror2("plus2", a, b);
}


Sex
plus2ii(Sex a, Sex b)
{
    long            n = mkint(a) + mkint(b);
    long            t = n & fix_mask;
    if (t == 0 || t == fix_mask)
	return mkfxat(n);
    return mkbigat1(n);
}




/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *diff_table_1[] =
    { diff2il, diff2ib, diff2im, diff2ir, diff2if,
    diff2ie, diff2ic, diff2iv, diff2id, diff2iz, diff2iE, diff2iZ,
    diff2iV, diff2iM, diff2iq, diff2iY
};

two_args       *diff_table_2[] =
    { diff2li, diff2bi, diff2mi, diff2ri, diff2fi,
    diff2ei, diff2ci, diff2vi, diff2di, diff2zi, diff2Ei, diff2Zi,
    diff2Vi, diff2Mi, diff2qi, diff2Yi
};

two_args       *diff_table[TYPEMAX][TYPEMAX] = {
    {diff2ll, diff2lb, diff2lm, diff2lr, diff2lf,
     diff2le, diff2lc, diff2lv, diff2ld, diff2lz, diff2lE, diff2lZ,
     diff2lV, diff2lM, diff2lq, diff2lY}
    ,
    {diff2bl, diff2bb, diff2bm, diff2br, diff2bf,
     diff2be, diff2bc, diff2bv, diff2bd, diff2bz, diff2bE, diff2bZ,
     diff2bV, diff2bM, diff2bq, diff2bY}
    ,
    {diff2ml, diff2mb, diff2mm, diff2mr, diff2mf,
     diff2me, diff2mc, diff2mv, diff2md, diff2mz, diff2mE, diff2mZ,
     diff2mV, math_er, math_er, diff2mY}
    ,
    {diff2rl, diff2rb, diff2rm, diff2rr, diff2rf,
     diff2re, diff2rc, diff2rv, diff2rd, diff2rz, diff2rE, diff2rZ,
     diff2rV, diff2rM, diff2rq, diff2rY}
    ,
    {diff2fl, diff2fb, diff2fm, diff2fr, diff2ff,
     diff2fe, diff2fc, diff2fv, diff2fd, diff2fz, diff2fE, diff2fZ,
     diff2fV, math_er, math_er, diff2fY}
    ,
    {diff2el, diff2eb, diff2em, diff2er, diff2ef,
     diff2ee, diff2ec, diff2ev, diff2ed, diff2ez, diff2eE, diff2eZ,
     diff2eV, math_er, math_er, diff2eY}
    ,
    {diff2cl, diff2cb, diff2cm, diff2cr, diff2cf,
     diff2ce, diff2cc, diff2cv, diff2cd, diff2cz, diff2cE, diff2cZ,
     diff2cV, diff2cM, diff2cq, diff2cY}
    ,
    {diff2vl, diff2vb, diff2vm, diff2vr, diff2vf,
     diff2ve, diff2vc, diff2vv, diff2vd, diff2vz, diff2vE, diff2vZ,
     diff2vV, math_er, math_er, diff2vY}
    ,
    {diff2dl, diff2db, diff2dm, diff2dr, diff2df,
     diff2de, diff2dc, diff2dv, diff2dd, diff2dz, diff2dE, diff2dZ,
     diff2dV, math_er, math_er, diff2dY}
    ,
    {diff2zl, diff2zb, diff2zm, diff2zr, diff2zf,
     diff2ze, diff2zc, diff2zv, diff2zd, diff2zz, diff2zE, diff2zZ,
     diff2zV, math_er, math_er, diff2zY}
    ,
    {diff2El, diff2Eb, diff2Em, diff2Er, diff2Ef,
     diff2Ee, diff2Ec, diff2Ev, diff2Ed, diff2Ez, diff2EE, diff2EZ,
     diff2EV, math_er, math_er, diff2EY}
    ,
    {diff2Zl, diff2Zb, diff2Zm, diff2Zr, diff2Zf,
     diff2Ze, diff2Zc, diff2Zv, diff2Zd, diff2Zz, diff2ZE, diff2ZZ,
     diff2ZV, math_er, math_er, diff2ZY}
    ,
    {diff2Vl, diff2Vb, diff2Vm, diff2Vr, diff2Vf,
     diff2Ve, diff2Vc, diff2Vv, diff2Vd, diff2Vz, diff2VE, diff2VZ,
     diff2VV, math_er, math_er, diff2VY}
    ,
    {diff2Ml, diff2Mb, math_er, diff2Mr, math_er,
     math_er, diff2Mc, math_er, math_er, math_er, math_er, math_er,
     math_er, diff2MM, diff2Mq, diff2MY}
    ,
    {diff2ql, diff2qb, math_er, diff2qr, math_er,
     math_er, diff2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, diff2qM, diff2qq, diff2qY}
    ,
    {diff2Yl, diff2Yb, math_er, diff2Yr, math_er,
     math_er, diff2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, diff2YM, diff2Yq, diff2YY}
};


#ifdef DEBUG_ARITH

#if DEBUG_SHOW
Sex
diff2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    if (is_bigint(a) || is_bigint(b)) {
	stackcheck();
	push2(a, b);
	debug_printf("(equal (difference %p %p)", a, b);
	pop2(b, a);
	a = xdiff2(a, b);
	push(a);
	debug_printf(" %p)\n", a);
	pop(a);
    } else {
	a = xdiff2(a, b);
    }
    return a;
}
#else

Sex
diff2(a, b)
     Sex             a,
                     b;
{
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]
    c = xdiff2(a, b);
    d = xplus2(c, b);
    if (eql(d, a) == NO) {
	debug_printf("%p - %p != %p ... %p\n", a, b, c, d);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    a = stack[-2];
    popv(5);
    return a;
}

#endif


Sex
xdiff2(a, b)
#else
Sex
diff2(a, b)
#endif
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n = mkint(a) - mkint(b);
	    long            t = n & fix_mask;
	    if (t == 0 || t == fix_mask)
		return mkfxat(n);
	    return mkbigat1(n);
	} else if (is_numhdr(b))
	    return diff_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return diff_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return diff_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("difference", a, b);
}



Sex
diff2ii(Sex a, Sex b)
{
    long            n = mkint(a) - mkint(b);
    long            t = n & fix_mask;
    if (t == 0 || t == fix_mask)
	return mkfxat(n);
    return mkbigat1(n);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2ii(Sex a, Sex b)
{
    long            aa = mkint(a),
	bb = mkint(b);
    long            sign = 1;
    unsigned long
                    low,
                    high;

    if (aa < 0) {
	sign = -1;
	aa = -aa;
    }
    if (bb < 0) {
	sign = -sign;
	bb = -bb;
    }
    high = umul(&low, aa, bb);
    if (high == 0) {
	if (sign > 0) {
	    if (low <= (unsigned long) fix_max)
		return mkfxat(low);
	} else {
	    if (low <= (unsigned long) big_min)
		return mkfxat(-(long) (low));
	}
	a = new_binteger(sign > 0 ? T_BIGNUM : T_BIGNEG, bnum_len(1));
	bignum_value(a)[0] = low;
	return a;
    }
    a = new_binteger(sign > 0 ? T_BIGNUM : T_BIGNEG, bnum_len(2));
    bignum_value(a)[0] = low;
    bignum_value(a)[1] = high;
    return a;
}




two_args       *times_table_1[] =
    { times2il, times2ib, times2im, times2ir, times2if,
    times2ie, times2ic, times2iv, times2id, times2iz, times2iE, times2iZ,
    times2iV, times2iM, times2iq, times2iY
};


two_args       *times_table[TYPEMAX][TYPEMAX] = {
    {times2ll, times2lb, times2lm, times2lr, times2lf,
     times2le, times2lc, times2lv, times2ld, times2lz, times2lE, times2lZ,
     times2lV, times2lM, times2lq, times2lY}
    ,
    {math_er, times2bb, times2bm, times2br, times2bf,
     times2be, times2bc, times2bv, times2bd, times2bz, times2bE, times2bZ,
     times2bV, times2bM, times2bq, times2bY}
    ,
    {math_er, math_er, times2mm, times2mr, times2mf,
     times2me, times2mc, times2mv, times2md, times2mz, times2mE, times2mZ,
     times2mV, math_er, math_er, times2mY}
    ,
    {math_er, math_er, math_er, times2rr, times2rf,
     times2re, times2rc, times2rv, times2rd, times2rz, times2rE, times2rZ,
     times2rV, times2rM, times2rq, times2rY}
    ,
    {math_er, math_er, math_er, math_er, times2ff,
     times2fe, times2fc, times2fv, times2fd, times2fz, times2fE, times2fZ,
     times2fV, math_er, math_er, times2fY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     times2ee, times2ec, times2ev, times2ed, times2ez, times2eE, times2eZ,
     times2eV, math_er, math_er, times2eY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, times2cc, times2cv, times2cd, times2cz, times2cE, times2cZ,
     times2cV, times2cM, times2cq, times2cY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, times2vv, times2vd, times2vz, times2vE, times2vZ,
     times2vV, math_er, math_er, times2vY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, times2dd, times2dz, times2dE, times2dZ,
     times2dV, math_er, math_er, times2dY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, times2zz, times2zE, times2zZ,
     times2zV, math_er, math_er, times2zY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, times2EE, times2EZ,
     times2EV, math_er, math_er, times2EY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, math_er, times2ZZ,
     times2ZV, math_er, math_er, times2ZY}
    ,
    {math_er, math_er, math_er, math_er, math_er,
     math_er, math_er, math_er, math_er, math_er, math_er, math_er,
     times2VV, math_er, math_er, times2VY}
    ,
    {times2Ml, times2Mb, math_er, times2Mr, math_er,
     math_er, times2Mc, math_er, math_er, math_er, math_er, math_er,
     math_er, times2MM, times2Mq, times2MY}
    ,
    {times2ql, times2qb, math_er, times2qr, math_er,
     math_er, times2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, times2qM, times2qq, times2qY}
    ,
    {times2Yl, times2Yb, math_er, times2Yr, math_er,
     math_er, times2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, times2YM, times2Yq, times2YY}
};

#ifdef DEBUG_ARITH

#if DEBUG_SHOW
Sex
times2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    if (is_bigint(a) || is_bigint(b)) {
	stackcheck();
	push2(a, b);
	debug_printf("(equal (times2 %p %p)", a, b);
	pop2(b, a);
	a = xtimes2(a, b);
	push(a);
	debug_printf(" %p)\n", a);
	pop(a);
    } else {
	a = xtimes2(a, b);
    }
    return a;
}
#else
Sex
times2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    stackcheck();
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[0]
    c = xtimes2(a, b);
    if (b == mkfxat(0)) {
	if (c == mkfxat(0)) {
	    r = c;
	    popv(4);
	    return r;
	}
	debug_printf("%p * %p != %p\n", a, b, c);
	r = c;
	popv(4);
	return r;
    }
    if (!is_rational(a) && !is_rational(b) && !zerop(d = xremaind2(c, b))) {
	debug_printf("%p % %p / %p != %p but ... %p\n", a, b, b, a, d);
    }
    if (is_rational(a) || is_rational(b))
	d = ratio2(c, b);
    else
	d = xquot2(c, b);
    if (eql(d, a) == NO) {
	debug_printf("%p * %p = %p / %p != %p but %p\n", a, b, c, b, a, d);
    }
#undef a
#undef b
#undef c
#undef d
    a = stack[-1];
    popv(4);
    return a;
}
#endif
#endif

Sex
#ifdef DEBUG_ARITH
xtimes2(a, b)
#else
times2(a, b)
#endif
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return times2ii(a, b);
	} else if (is_numhdr(b))
	    return times_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return times_table_1[numtype_of(a)] (b, a);
	else if (is_numhdr(b)) {
	    int             i = numtype_of(a);
	    int             j = numtype_of(b);
	    if (i > j)
		return times_table[j][i] (b, a);
	    return times_table[i][j] (a, b);
	}
    }
    return aerror2("times2", a, b);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


two_args       *quot_table_1[] =
    { quot2il, quot2ib, quot2im, quot2ir, quot2if,
    quot2ie, quot2ic, quot2iv, quot2id, quot2iz, quot2iE, quot2iZ,
    quot2iV, quot2iM, quot2iq, quot2iY
};

two_args       *quot_table_2[] =
    { quot2li, quot2bi, quot2mi, quot2ri, quot2fi,
    quot2ei, quot2ci, quot2vi, quot2di, quot2zi, quot2Ei, quot2Zi,
    quot2Vi, quot2Mi, quot2qi, quot2Yi
};

two_args       *quot_table[TYPEMAX][TYPEMAX] = {
    {quot2ll, quot2lb, quot2lm, quot2lr, quot2lf,
     quot2le, quot2lc, quot2lv, quot2ld, quot2lz, quot2lE, quot2lZ,
     quot2lV, quot2lM, quot2lq, quot2lY}
    ,
    {quot2bl, quot2bb, quot2bm, quot2br, quot2bf,
     quot2be, quot2bc, quot2bv, quot2bd, quot2bz, quot2bE, quot2bZ,
     quot2bV, quot2bM, quot2bq, quot2bY}
    ,
    {quot2ml, quot2mb, quot2mm, quot2mr, quot2mf,
     quot2me, quot2mc, quot2mv, quot2md, quot2mz, quot2mE, quot2mZ,
     quot2mV, math_er, math_er, quot2mY}
    ,
    {quot2rl, quot2rb, quot2rm, quot2rr, quot2rf,
     quot2re, quot2rc, quot2rv, quot2rd, quot2rz, quot2rE, quot2rZ,
     quot2rV, quot2rM, quot2rq, quot2rY}
    ,
    {quot2fl, quot2fb, quot2fm, quot2fr, quot2ff,
     quot2fe, quot2fc, quot2fv, quot2fd, quot2fz, quot2fE, quot2fZ,
     quot2fV, math_er, math_er, quot2fY}
    ,
    {quot2el, quot2eb, quot2em, quot2er, quot2ef,
     quot2ee, quot2ec, quot2ev, quot2ed, quot2ez, quot2eE, quot2eZ,
     quot2eV, math_er, math_er, quot2eY}
    ,
    {quot2cl, quot2cb, quot2cm, quot2cr, quot2cf,
     quot2ce, quot2cc, quot2cv, quot2cd, quot2cz, quot2cE, quot2cZ,
     quot2cV, quot2cM, quot2cq, quot2cY}
    ,
    {quot2vl, quot2vb, quot2vm, quot2vr, quot2vf,
     quot2ve, quot2vc, quot2vv, quot2vd, quot2vz, quot2vE, quot2vZ,
     quot2vV, math_er, math_er, quot2vY}
    ,
    {quot2dl, quot2db, quot2dm, quot2dr, quot2df,
     quot2de, quot2dc, quot2dv, quot2dd, quot2dz, quot2dE, quot2dZ,
     quot2dV, math_er, math_er, quot2dY}
    ,
    {quot2zl, quot2zb, quot2zm, quot2zr, quot2zf,
     quot2ze, quot2zc, quot2zv, quot2zd, quot2zz, quot2zE, quot2zZ,
     quot2zV, math_er, math_er, quot2zY}
    ,
    {quot2El, quot2Eb, quot2Em, quot2Er, quot2Ef,
     quot2Ee, quot2Ec, quot2Ev, quot2Ed, quot2Ez, quot2EE, quot2EZ,
     quot2EV, math_er, math_er, quot2EY}
    ,
    {quot2Zl, quot2Zb, quot2Zm, quot2Zr, quot2Zf,
     quot2Ze, quot2Zc, quot2Zv, quot2Zd, quot2Zz, quot2ZE, quot2ZZ,
     quot2ZV, math_er, math_er, quot2ZY}
    ,
    {quot2Vl, quot2Vb, quot2Vm, quot2Vr, quot2Vf,
     quot2Ve, quot2Vc, quot2Vv, quot2Vd, quot2Vz, quot2VE, quot2VZ,
     quot2VV, math_er, math_er, quot2VY}
    ,
    {quot2Ml, quot2Mb, math_er, quot2Mr, math_er,
     math_er, quot2Mc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2MM, quot2Mq, quot2MY}
    ,
    {quot2ql, quot2qb, math_er, quot2qr, math_er,
     math_er, quot2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2qM, quot2qq, quot2qY}
    ,
    {quot2Yl, quot2Yb, math_er, quot2Yr, math_er,
     math_er, quot2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2YM, quot2Yq, quot2YY}
};


#ifdef DEBUG_ARITH
#if DEBUG_SHOW
Sex
quot2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    if (is_bigint(a) || is_bigint(b)) {
	stackcheck();
	push2(a, b);
	debug_printf("(equal (quotient %p %p)", a, b);
	pop2(b, a);
	a = xquot2(a, b);
	push(a);
	debug_printf(" %p)\n", a);
	pop(a);
    } else {
	a = xquot2(a, b);
    }
    return a;
}
#else
Sex
quot2(a, b)
     Sex             a,
                     b;
{
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]
    c = xquot2(a, b);
    d = xremaind2(a, b);
    e = xtimes2(c, b);
    e = plus2(e, d);
    if (eql(e, a) == NO) {
	debug_printf("%p / %p != %p ... %p\n", a, b, c, d);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    a = stack[-2];
    popv(5);
    return a;
}
#endif
#endif

Sex
#ifdef DEBUG_ARITH
xquot2(a, b)
#else
quot2(a, b)
#endif
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n = mkint(a) / mkint(b);
	    long            t = n & fix_mask;
	    if (t == 0 || t == fix_mask)
		return mkfxat(n);
	    return mkbigat1(n);
	} else if (is_numhdr(b))
	    return quot_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return quot_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return quot_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("quotient", a, b);
}



Sex
quot2ii(Sex a, Sex b)
{
    long            n = mkint(a) / mkint(b);
    long            m = n & fix_mask;
    if (m == 0 || m == fix_mask)
	return mkfxat(n);
    return mkbigat1(n);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/




two_args       *ratio_table_1[] =
    { ratio2il, ratio2ib, quot2im, quot2ir, quot2if,
    quot2ie, quot2ic, quot2iv, quot2id, quot2iz, quot2iE, quot2iZ,
    quot2iV, quot2iM, quot2iq, quot2iY
};

two_args       *ratio_table_2[] =
    { ratio2li, ratio2bi, quot2mi, quot2ri, quot2fi,
    quot2ei, quot2ci, quot2vi, quot2di, quot2zi, quot2Ei, quot2Zi,
    quot2Vi, quot2Mi, quot2qi, quot2Yi
};

two_args       *ratio_table[TYPEMAX][TYPEMAX] = {
    {ratio2ll, ratio2lb, quot2lm, quot2lr, quot2lf,
     quot2le, quot2lc, quot2lv, quot2ld, quot2lz, quot2lE, quot2lZ,
     quot2lV, quot2lM, quot2lq, quot2lY}
    ,
    {ratio2bl, ratio2bb, quot2bm, quot2br, quot2bf,
     quot2be, quot2bc, quot2bv, quot2bd, quot2bz, quot2bE, quot2bZ,
     quot2bV, quot2bM, quot2bq, quot2bY}
    ,
    {quot2ml, quot2mb, quot2mm, quot2mr, quot2mf,
     quot2me, quot2mc, quot2mv, quot2md, quot2mz, quot2mE, quot2mZ,
     quot2mV, math_er, math_er, quot2mY}
    ,
    {quot2rl, quot2rb, quot2rm, quot2rr, quot2rf,
     quot2re, quot2rc, quot2rv, quot2rd, quot2rz, quot2rE, quot2rZ,
     quot2rV, quot2rM, quot2rq, quot2rY}
    ,
    {quot2fl, quot2fb, quot2fm, quot2fr, quot2ff,
     quot2fe, quot2fc, quot2fv, quot2fd, quot2fz, quot2fE, quot2fZ,
     quot2fV, math_er, math_er, quot2fY}
    ,
    {quot2el, quot2eb, quot2em, quot2er, quot2ef,
     quot2ee, quot2ec, quot2ev, quot2ed, quot2ez, quot2eE, quot2eZ,
     quot2eV, math_er, math_er, quot2eY}
    ,
    {quot2cl, quot2cb, quot2cm, quot2cr, quot2cf,
     quot2ce, quot2cc, quot2cv, quot2cd, quot2cz, quot2cE, quot2cZ,
     quot2cV, quot2cM, quot2cq, quot2cY}
    ,
    {quot2vl, quot2vb, quot2vm, quot2vr, quot2vf,
     quot2ve, quot2vc, quot2vv, quot2vd, quot2vz, quot2vE, quot2vZ,
     quot2vV, math_er, math_er, quot2vY}
    ,
    {quot2dl, quot2db, quot2dm, quot2dr, quot2df,
     quot2de, quot2dc, quot2dv, quot2dd, quot2dz, quot2dE, quot2dZ,
     quot2dV, math_er, math_er, quot2dY}
    ,
    {quot2zl, quot2zb, quot2zm, quot2zr, quot2zf,
     quot2ze, quot2zc, quot2zv, quot2zd, quot2zz, quot2zE, quot2zZ,
     quot2zV, math_er, math_er, quot2zY}
    ,
    {quot2El, quot2Eb, quot2Em, quot2Er, quot2Ef,
     quot2Ee, quot2Ec, quot2Ev, quot2Ed, quot2Ez, quot2EE, quot2EZ,
     quot2EV, math_er, math_er, quot2EY}
    ,
    {quot2Zl, quot2Zb, quot2Zm, quot2Zr, quot2Zf,
     quot2Ze, quot2Zc, quot2Zv, quot2Zd, quot2Zz, quot2ZE, quot2ZZ,
     quot2ZV, math_er, math_er, quot2ZY}
    ,
    {quot2Vl, quot2Vb, quot2Vm, quot2Vr, quot2Vf,
     quot2Ve, quot2Vc, quot2Vv, quot2Vd, quot2Vz, quot2VE, quot2VZ,
     quot2VV, math_er, math_er, quot2VY}
    ,
    {quot2Ml, quot2Mb, math_er, quot2Mr, math_er,
     math_er, quot2Mc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2MM, quot2Mq, quot2MY}
    ,
    {quot2ql, quot2qb, math_er, quot2qr, math_er,
     math_er, quot2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2qM, quot2qq, quot2qY}
    ,
    {quot2Yl, quot2Yb, math_er, quot2Yr, math_er,
     math_er, quot2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, quot2YM, quot2Yq, quot2YY}
};



Sex
ratio2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return ratio2ii(a, b);
	} else if (is_numhdr(b))
	    return ratio_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return ratio_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return ratio_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("ratio", a, b);
}



/*****************************************************************************/
/***              Lisp-callable versions of arithmetic functions           ***/
/*****************************************************************************/



Sex
Ladd1(Sex a)
{
    if (a == fxat_max)
	return mkbigat1(1 + mkint(a));
    if (is_fixnum(a)) {
	return (Sex) ((unsigned long) a + 0x10);
    } else if (is_numhdr(a)) {
	return plus_table_1[numtype_of(a)] (mkfxat(1), a);
    } else
	return aerror2("add1", a, mkfxat(1));
}

Sex
Lsub1(Sex a)
{
    if (a == fxat_min)
	return mkbigat1(mkint(a) - 1);
    if (is_fixnum(a)) {
	return (Sex) ((unsigned long) a - 0x10);
    } else if (is_numhdr(a)) {
	return plus_table_1[numtype_of(a)] (mkfxat(-1), a);
    } else
	return aerror2("sub1", a, mkfxat(-1));
}


Sex
Lash(Sex a, Sex b)
{
    a = ash(a, b);
    return a;
}

Sex
Lash1(Sex a, Sex b)
{
    bool            negative = NO;
    if (!is_fixnum(b))
	return aerror("ash1");
    if (minusp(a)) {
	negative = YES;
	a = negate(a);
    }
    a = ash(a, b);
    if (negative) {
	a = negate(a);
    }
    return a;
}

static char     msd_table[256] = {
    0, 1, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8,
    8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8
};

long
msd(unsigned long top)
{
    long            r = 0;
#ifdef _LP64
    if (top >= 0x100000000)
	r += 32, top >>= 32;
#endif
    if (top >= 0x10000)
	r += 16, top >>= 16;
    if (top >= 0x100)
	r += 8, top >>= 8;
    return (r + msd_table[top]);
}

Sex
Lmsd(Sex a)
{
    long            top;
    long            r = 0;
    if (is_fixnum(a)) {
	top = mkint(a);
	if (top < 0)
	    return aerror1("negative arg for msd", a);
    } else if (is_numhdr(a) &&
	       (numhdr(a) == T_BIGNUM || numhdr(a) == T_LOGICAL)) {
	r = bnum_n(bignum_length(a)) - 1;
	top = bignum_value(a)[r];
	r = LPSIZE * r;
    } else
	return aerror1("bad arg for msd", a);
#ifdef _LP64
    if (top >= 0x100000000)
	r += 32, top >>= 32;
#endif
    if (top >= 0x10000)
	r += 16, top >>= 16;
    if (top >= 0x100)
	r += 8, top >>= 8;
    return mkfxat(r + msd_table[top]);
}

static char     lsd_table[256] = {
    8, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    7, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    6, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    5, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0,
    4, 0, 1, 0, 2, 0, 1, 0, 3, 0, 1, 0, 2, 0, 1, 0
};

long
lsd(unsigned long top)
{
    long            r = 0;
#ifdef _LP64
    if ((top & 0xffffffff) == 0)
	r += 32, top >>= 32;
#endif
    if ((top & 0xffff) == 0)
	r += 16, top >>= 16;
    if ((top & 0xff) == 0)
	r += 8, top >>= 8;
    return (r + lsd_table[top & 0xff]);
}

Sex
Llsd(Sex a)
{
    long            top;
    long            r = 0;
    if (is_fixnum(a)) {
	top = mkint(a);
	if (top == 0)
	    return a;
	if (top < 0)
	    return aerror1("negative arg for lsd", a);
    } else if (is_bigint(a)) {
	while ((top = bignum_value(a)[r]) == 0)
	    r++;
	r = LPSIZE * r;
    } else
	return aerror1("bad arg for lsd", a);
#ifdef _LP64
    if ((top & 0xffffffff) == 0)
	r += 32, top >>= 32;
#endif
    if ((top & 0xffff) == 0)
	r += 16, top >>= 16;
    if ((top & 0xff) == 0)
	r += 8, top >>= 8;
    return mkfxat(r + lsd_table[top & 0xff]);
}

static char     log2_table[256] = {
    -1, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7,
    7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 7
};


long
integer_log2(long a)
{
    long            l = 0;

    if (a < 0)
	a = -a;
#ifdef _LP64
    if (a >= 0x100000000) {
	l = 32;
	a = a >> 32;
    }
#endif
    if (a >= 0x10000) {
	l += 16;
	a = a >> 16;
    }
    if (a >= 0x100) {
	l += 8;
	a = a >> 8;
    }
    return (l + log2_table[a]);
}


long
bignum_log2(Sex a)
{
    long            l;
    unsigned long   n;

    l = bnum_n(bignum_length(a)) - 1;
    n = bignum_value(a)[l];
    l = l * SEX * 8;
#ifdef _LP64
    if (n >= 0x100000000) {
	l = l + 32;
	n = n >> 32;
    }
#endif
    if (n >= 0x10000) {
	l = l + 16;
	n = n >> 16;
    }
    if (n >= 0x100) {
	l = l + 8;
	n = n >> 8;
    }
    return (l + log2_table[n]);
}

long
ratio_log2(Sex a)
{
    long            n,
                    m;

    Sex             b;
    b = rational_den(a);
    a = rational_num(a);
    if (a == mkfxat(0))
	return fix_min;
    n = is_fixnum(a) ? integer_log2(a) : bignum_log2(a);
    m = is_fixnum(b) ? integer_log2(b) : bignum_log2(b);
    n = n - m;
    if (n < fix_min)
	return fix_min;
    return n;
}


long
float_log2(double a)
{
    DoubleFloat     u;
    if (a == 0.0)
	return fix_min;
    u.d = a;
    return ieee_exponent(u);
}

long
bfloat_log2(Sex a)
{
    long            n,
                    m;
    long            l;
    m = bfloat_exp(a);
    a = bfloat_man(a);

    l = bnum_n(bignum_length(a)) - 1;
    n = bignum_value(a)[l];
    if (n == 0)
	return fix_min;
    n = m + l * SEX * 8 + msd(n) - 1;
    return n;
}


Sex
Lnumber_log2(a)
     Sex             a;
{
    return mkfxat(number_log2(a));
}

long
number_log2(a)
     Sex             a;
{
    long            l,
                    n;

    if (is_fixnum(a)) {
	if (a == mkfxat(0))
	    return fix_min;
	return integer_log2(mkint(a));
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    n = bignum_log2(a);
	    if (n == -1)
		return fix_min;
	    return n;
	case N_RATIONAL:
	    return ratio_log2(a);
	case N_FLOAT:
	    return float_log2(float_val(a));
	case N_COMPLEX:
	    n = number_log2(complex_real(a));
	    l = number_log2(complex_imag(a));
	    if (n < l)
		n = l;
	    return n;
	case N_EFLOAT:
	    return float_log2(efloat_value(a));
	case N_ECOMPLEX:
	    n = float_log2(ecomplex_real(a));
	    l = float_log2(ecomplex_imag(a));
	    if (n < l)
		n = l;
	    return n;
	case N_INTERVAL:
	    n = float_log2(interval_lower(a) + interval_upper(a));
	    if (n <= fix_min)
		return fix_min;
	    return n - 1;
	case N_BFLOAT:
	    return bfloat_log2(a);
	case N_BEFLOAT:
	    return bfloat_log2(befloat_value(a));
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    n = bfloat_log2(a);
	    if (n <= fix_min)
		return fix_min;
	    return n - 1;
	case N_BECOMPLEX:
	    n = bfloat_log2(becomplex_real(a));
	    l = bfloat_log2(becomplex_imag(a));
	    if (n < l)
		n = l;
	    return n;
	}
    }
    aerror2("number-log2", a, nil);
    return 0;
}

Sex
divide2ii(Sex a, Sex b)
{
    long            n = mkint(a) / mkint(b);
    long            m = mkint(a) % mkint(b);
    long            r = n & fix_mask;
    if (r == 0 || r == fix_mask)
	a = mkfxat(n);
    else
	a = mkbigat1(n);
    b = mkfxat(m);
    return cons(a, b);
}



Sex
divide2ib(Sex a, Sex b)
{
    if ((long) a == fix_mask && bignum_length(b) == bnum_len(1) &&
	!is_bigneg(a) && bignum_value(b)[0] == big_min) {
	return cons(mkfxat(-1), mkfxat(0));
    } else {
	return cons(mkfxat(0), a);
    }
}


Sex
Ldivide(Sex a, Sex b)
{
    Sex             q,
                    r;
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return divide2ii(a, b);
	else if (is_bigint(b))
	    return divide2ib(a, b);
    } else if (is_bigint(a)) {
	if (is_fixnum(b))
	    return divide2bi(a, b);
	else if (is_bigint(b))
	    return divide2bb(a, b);
    }
    stackcheck();
    push2(a, b);
    q = quot2(a, b);
    pop2(b, a);
    push(q);
    r = remaind2(a, b);
    pop(q);
    q = cons(q, r);
    return q;
}



/*
 * generic negation
 */

Sex
Lminus(Sex a)
{
    return negate(a);
}


Sex
negate(Sex a)
{
    if (is_fixnum(a)) {
	long            aa = -mkint(a);
	if (aa != big_min)
	    return mkfxat(aa);
	else
	    return mkbigat1(aa);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return negatel(a);
	case N_BIGNUM:
	    return negateb(a);
	case N_MODULAR:
	    return negatem(a);
	case N_RATIONAL:
	    {
		Sex             n = rational_num(a),
		    d = rational_den(a);
		stackcheck();
		push(d);
		n = negate(n);
		pop(d);
		return make_ratio(n, d);
	    }
	case N_COMPLEX:
	    {
		Sex             r = complex_real(a),
		    i = complex_imag(a);
		stackcheck();
		push(i);
		r = negate(r);
		pop(i);
		push(r);
		i = negate(i);
		pop(r);
		return Lmkcomplex(r, i);
	    }
	case N_EFLOAT:
	    {
		double          v = efloat_value(a),
		    e = efloat_error(a);
		return make_efloat(-v, e);
	    }
	case N_FLOAT:
	    {
		double          d = floatit(a);
		return make_float(-d);
	    }
	case N_INTERVAL:
	    return make_interval(-interval_upper(a), -interval_lower(a));
	case N_ECOMPLEX:
	    return make_ecomplex(-ecomplex_real(a), -ecomplex_imag(a),
				 ecomplex_error(a));
	case N_BFLOAT:
	    return negated(a);
	case N_BEFLOAT:
	    return negateE(a);
	case N_BECOMPLEX:
	    return negateZ(a);
	case N_BINTERVAL:
	    return negateV(a);
	case N_BMODULAR:
	    return negateM(a);
	case N_FMODULAR:
	    return negateY(a);
	default:
	    break;
	}
    }
    return aerror1("bad arg for minus", a);
}



#ifdef USE_MT
#ifdef _LP64
#define random genrand64_int63
#define random2 genrand64_int64
#define srandom init_genrand64
#else
#define random genrand_int31
#define random2 genrand_int32
#define srandom init_genrand
#endif

#else

#if GCCTYPE == GNU_WIN32
#define random rand
#define random2 rand
#define srandom srand

#else

#ifdef VisualC
#define random rand
#define random2 rand
#define srandom srand
#endif

#endif

#endif


long
my_random(long p)
{
    unsigned long   r;
    if (p == 0)
	return 0;
    r = random();
    r = r % p;
    return r;
}

Sex
Lrandom(Sex a)
{
    long            r;
#ifdef _LP64
    int            *p;
    long            s;
#endif
    if (is_fixnum(a)) {
	long            n;
	r = random();
	n = mkint(a);
	if (n <= 0)
	    return mkfxat(0);
#if defined(_LP64) && !defined(USE_MT)
	s = random();
	s = ((s << 32) | (r & 0x00000000ffffffffUL)) &
	    0x8000000000000000UL;
#endif
	r = r % n;
	return mkfxat(r);
    } else if (is_bigint(a)) {
	Sex             c;
	long            len,
	                i;

	push(a);
	len = numlen(a);
	c = new_bignum(len + 1);
	pop(a);
	len = bnum_n(len);
#if defined(_LP64) && !defined(USE_MT)
	p = (unsigned int *) &bignum_value(c)[0];

	for (i = 0; i <= 2 * len; i++) {
	    *p++ = random2();
	}
#else
	for (i = 0; i <= len; i++) {
	    bignum_value(c)[i] = random2();
	}
#endif
	for (i = len; i > 0; i--) {
	    if (bignum_value(c)[i] != 0)
		break;
	}
	if (i == 0) {
	    r = bignum_value(c)[0];
	    if (r <= fix_max)
		return mkfxat(r);
	}
	if (i != len) {
	    numlen(c) = bnum_len(i + 1);
	}
	a = rem2bb(c, a);
	return a;
    }
    return error1("random", a);
}

Sex
Lsrandom(Sex a)
{
    if (!is_fixnum(a))
	return aerror1("srandom", a);
    srandom(mkint(a));
    return nil;
}

#ifdef USE_MT
Sex
Lrandom_int1()
{
    Sex             a;
    a = mkfxat(random());
    a = a & int_max;
    return a;
}

Sex
Lrandom_int2()
{
    Sex             a;
    a = mkfxat(random());
    return a;
}
#else
#ifdef _LP64
Sex
Lrandom_int1()
{
    long            n;
    Sex             a;

    n = (random() << 31) | random();;
    a = mkfxat(n);
    a = a & int_max;
    return a;
}

Sex
Lrandom_int2()
{
    long            n;
    Sex             a;

    n = (random() << 31) | random();;
    a = mkfxat(n);
    return a;
}
#else
Sex
Lrandom_int1()
{
    Sex             a;
    a = mkfxat(random());
    a = a & int_max;
    return a;
}

Sex
Lrandom_int2()
{
    Sex             a;
    a = mkfxat(random());
    return a;
}
#endif
#endif


#ifdef USE_MT
#ifdef _LP64
Sex
Lrandom_real1()
{
    unsigned long   n = random2();
    double          r = (n >> 11) * (1.0 / 9007199254740991.0);
    return make_float(r);
}

Sex
Lrandom_real2()
{
    unsigned long   n = random2();
    double          r = (n >> 11) * (1.0 / 9007199254740992.0);
    return make_float(r);
}

Sex
Lrandom_real3()
{
    unsigned long   n = random2();
    double          r =
	((long) (n >> 12) + 0.5) * (1.0 / 4503599627370496.0);
    return make_float(r);
}
#else
Sex
Lrandom_real1()
{
    unsigned long   n = random2();
    double          r = ((double) n) * (1.0 / 4294967295.0);
    // divide by 2^32-1
    return make_float(r);
}

Sex
Lrandom_real2()
{
    unsigned long   n = random2();
    double          r = ((double) n) * (1.0 / 4294967296.0);
    // divide by 2^32
    return make_float(r);
}

Sex
Lrandom_real3()
{
    unsigned long   n = random2();
    double          r = ((double) n + 0.5) * (1.0 / 4294967296.0);
    // divide by 2^32
    return make_float(r);
}
#endif
#else
Sex
Lrandom_real1()
{
    unsigned long   n = random2();
    double          r = ((double) n) * (1.0 / 2147483647.0);
    // divide by 2^31-1
    return make_float(r);
}

Sex
Lrandom_real2()
{
    unsigned long   n = random2();
    double          r = ((double) n) * (1.0 / 2147483648.0);
    // divide by 2^31
    return make_float(r);
}

Sex
Lrandom_real3()
{
    unsigned long   n = random2();
    double          r = ((double) n + 0.5) * (1.0 / 2147483648.0);
    // divide by 2^31
    return make_float(r);
}
#endif



/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 * Arithmetic functions. remainder, =, minusp, plusp
 */





two_args       *rem_table_1[] = { rem2il, rem2ib, rem2im, rem2ir, rem2if,
    rem2ie, rem2ic, rem2iv, rem2id, rem2iz, rem2iE, rem2iZ,
    rem2iV, rem2iM, rem2iq, rem2iY
};

two_args       *rem_table_2[] = { rem2li, rem2bi, rem2mi, rem2ri, rem2fi,
    rem2ei, rem2ci, rem2vi, rem2di, rem2zi, rem2Ei, rem2Zi,
    rem2Vi, rem2Mi, rem2qi, rem2Yi
};

two_args       *rem_table[TYPEMAX][TYPEMAX] = {
    {rem2ll, rem2lb, rem2lm, rem2lr, rem2lf,
     rem2le, rem2lc, rem2lv, rem2ld, rem2lz, rem2lE, rem2lZ,
     rem2lV, rem2lM, rem2lq, rem2lY}
    ,
    {rem2bl, rem2bb, rem2bm, rem2br, rem2bf,
     rem2be, rem2bc, rem2bv, rem2bd, rem2bz, rem2bE, rem2bZ,
     rem2bV, rem2bM, rem2bq, rem2bY}
    ,
    {rem2ml, rem2mb, rem2mm, rem2mr, rem2mf,
     rem2me, rem2mc, rem2mv, rem2md, rem2mz, rem2mE, rem2mZ,
     rem2mV, rem2mM, rem2mq, rem2mY}
    ,
    {rem2rl, rem2rb, rem2rm, rem2rr, rem2rf,
     rem2re, rem2rc, rem2rv, rem2rd, rem2rz, rem2rE, rem2rZ,
     rem2rV, rem2rM, rem2rq, rem2rY}
    ,
    {rem2fl, rem2fb, rem2fm, rem2fr, rem2ff,
     rem2fe, rem2fc, rem2fv, rem2fd, rem2fz, rem2fE, rem2fZ,
     rem2fV, rem2fM, math_er, rem2fY}
    ,
    {rem2el, rem2eb, rem2em, rem2er, rem2ef,
     rem2ee, rem2ec, rem2ev, rem2ed, rem2ez, rem2eE, rem2eZ,
     rem2eV, rem2eM, math_er, rem2eY}
    ,
    {rem2cl, rem2cb, rem2cm, rem2cr, rem2cf,
     rem2ce, rem2cc, rem2cv, rem2cd, rem2cz, rem2cE, rem2cZ,
     rem2cV, rem2cM, rem2cq, rem2cY}
    ,
    {rem2vl, rem2vb, rem2vm, rem2vr, rem2vf,
     rem2ve, rem2vc, rem2vv, rem2vd, rem2vz, rem2vE, rem2vZ,
     rem2vV, rem2vM, math_er, rem2vY}
    ,
    {rem2dl, rem2db, rem2dm, rem2dr, rem2df,
     rem2de, rem2dc, rem2dv, rem2dd, rem2dz, rem2dE, rem2dZ,
     rem2dV, rem2dM, math_er, rem2dY}
    ,
    {rem2zl, rem2zb, rem2zm, rem2zr, rem2zf,
     rem2ze, rem2zc, rem2zv, rem2zd, rem2zz, rem2zE, rem2zZ,
     rem2zV, rem2zM, math_er, rem2zY}
    ,
    {rem2El, rem2Eb, rem2Em, rem2Er, rem2Ef,
     rem2Ee, rem2Ec, rem2Ev, rem2Ed, rem2Ez, rem2EE, rem2EZ,
     rem2EV, rem2EM, math_er, rem2EY}
    ,
    {rem2Zl, rem2Zb, rem2Zm, rem2Zr, rem2Zf,
     rem2Ze, rem2Zc, rem2Zv, rem2Zd, rem2Zz, rem2ZE, rem2ZZ,
     rem2ZV, rem2ZM, math_er, rem2ZY}
    ,
    {rem2Vl, rem2Vb, rem2Vm, rem2Vr, rem2Vf,
     rem2Ve, rem2Vc, rem2Vv, rem2Vd, rem2Vz, rem2VE, rem2VZ,
     rem2VV, rem2VM, math_er, rem2VY}
    ,
    {rem2Ml, rem2Mb, rem2Mm, rem2Mr, rem2Mf,
     rem2Me, rem2Mc, rem2Mv, rem2Md, rem2Mz, rem2ME, rem2MZ,
     rem2MV, rem2MM, rem2Mq, rem2MY}
    ,
    {rem2ql, rem2qb, rem2qm, rem2qr, rem2qf,
     rem2qe, rem2qc, rem2qv, rem2qd, rem2qz, rem2qE, rem2qZ,
     rem2qV, rem2qM, rem2qq, rem2qY}
    ,
    {rem2Yl, rem2Yb, rem2Ym, rem2Yr, rem2Yf,
     rem2Ye, rem2Yc, rem2Yv, rem2Yd, rem2Yz, rem2YE, rem2YZ,
     rem2YV, rem2YM, rem2Yq, rem2YY}
};


#ifdef DEBUG_ARITH
#if DEBUG_SHOW
Sex
remaind2(a, b)
     Sex             a,
                     b;
{
    Sex             r;
    if (is_bigint(a) || is_bigint(b)) {
	stackcheck();
	push2(a, b);
	debug_printf("(equal (remainder %p %p)", a, b);
	pop2(b, a);
	a = xremaind2(a, b);
	push(a);
	debug_printf(" %p)\n", a);
	pop(a);
    } else {
	a = xremaind2(a, b);
    }
    return a;
}
#else
Sex
remaind2(a, b)
     Sex             a,
                     b;
{
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define e stack[0]
    /*
     * c = xquot2(a, b); d = xremaind2(a, b); e = xtimes2(c, b); 
     */
    c = quot2(a, b);
    d = xremaind2(a, b);
    e = times2(c, b);
    e = plus2(e, d);
    if (eql(e, a) == NO) {
	debug_printf("%p / %p != %p ... %p\n", a, b, c, d);
    }
#undef a
#undef b
#undef c
#undef d
#undef e
    a = stack[-1];
    popv(5);
    return a;
}
#endif
#endif

Sex
#ifdef DEBUG_ARITH
xremaind2(a, b)
#else
remaind2(a, b)
#endif
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n = mkint(a) % mkint(b);
	    return mkfxat(n);
	} else if (is_numhdr(b))
	    return rem_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return rem_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return rem_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("remainder", a, b);
}


Sex
rem2ii(Sex a, Sex b)
{
    long            n = mkint(a) % mkint(b);
    return mkfxat(n);
}

Sex
mkzero2(Sex a, Sex b)
{
    return mkfxat(0);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



two_args       *mod_table_1[] = { mod2il, mod2ib, mod2im, mod2ir, mod2if,
    mod2ie, mod2ic, mod2iv, mod2id, mod2iz, mod2iE, mod2iZ,
    mod2iV, mod2iM, mod2iq, mod2iY
};

two_args       *mod_table_2[] = { mod2li, mod2bi, mod2mi, mod2ri, mod2fi,
    mod2ei, mod2ci, mod2vi, mod2di, mod2zi, mod2Ei, mod2Zi,
    mod2Vi, mod2Mi, mod2qi, mod2iY
};

two_args       *mod_table[TYPEMAX][TYPEMAX] = {
    {mod2ll, mod2lb, mod2lm, mod2lr, mod2lf,
     mod2le, mod2lc, mod2lv, mod2ld, mod2lz, mod2lE, mod2lZ,
     mod2lV, mod2lM, mod2lq, mod2lY}
    ,
    {mod2bl, mod2bb, mod2bm, mod2br, mod2bf,
     mod2be, mod2bc, mod2bv, mod2bd, mod2bz, mod2bE, mod2bZ,
     mod2bV, mod2bM, mod2bq, mod2bY}
    ,
    {mod2ml, mod2mb, mod2mm, mod2mr, mod2mf,
     mod2me, mod2mc, mod2mv, mod2md, mod2mz, mod2mE, mod2mZ,
     mod2mV, mod2mM, mod2mq, mod2mY}
    ,
    {mod2rl, mod2rb, mod2rm, mod2rr, mod2rf,
     mod2re, mod2rc, mod2rv, mod2rd, mod2rz, mod2rE, mod2rZ,
     mod2rV, mod2rM, mod2rq, mod2rY}
    ,
    {mod2fl, mod2fb, mod2fm, mod2fr, mod2ff,
     mod2fe, mod2fc, mod2fv, mod2fd, mod2fz, mod2fE, mod2fZ,
     mod2fV, mod2fM, math_er, mod2fY}
    ,
    {mod2el, mod2eb, mod2em, mod2er, mod2ef,
     mod2ee, mod2ec, mod2ev, mod2ed, mod2ez, mod2eE, mod2eZ,
     mod2eV, mod2eM, math_er, mod2eY}
    ,
    {mod2cl, mod2cb, mod2cm, mod2cr, mod2cf,
     mod2ce, mod2cc, mod2cv, mod2cd, mod2cz, mod2cE, mod2cZ,
     mod2cV, mod2cM, mod2cq, mod2cY}
    ,
    {mod2vl, mod2vb, mod2vm, mod2vr, mod2vf,
     mod2ve, mod2vc, mod2vv, mod2vd, mod2vz, mod2vE, mod2vZ,
     mod2vV, mod2vM, math_er, mod2vY}
    ,
    {mod2dl, mod2db, mod2dm, mod2dr, mod2df,
     mod2de, mod2dc, mod2dv, mod2dd, mod2dz, mod2dE, mod2dZ,
     mod2dV, mod2dM, math_er, mod2dY}
    ,
    {mod2zl, mod2zb, mod2zm, mod2zr, mod2zf,
     mod2ze, mod2zc, mod2zv, mod2zd, mod2zz, mod2zE, mod2zZ,
     mod2zV, mod2zM, math_er, mod2zY}
    ,
    {mod2El, mod2Eb, mod2Em, mod2Er, mod2Ef,
     mod2Ee, mod2Ec, mod2Ev, mod2Ed, mod2Ez, mod2EE, mod2EZ,
     mod2EV, mod2EM, math_er, mod2EY}
    ,
    {mod2Zl, mod2Zb, mod2Zm, mod2Zr, mod2Zf,
     mod2Ze, mod2Zc, mod2Zv, mod2Zd, mod2Zz, mod2ZE, mod2ZZ,
     mod2ZV, mod2ZM, math_er, mod2ZY}
    ,
    {mod2Vl, mod2Vb, mod2Vm, mod2Vr, mod2Vf,
     mod2Ve, mod2Vc, mod2Vv, mod2Vd, mod2Vz, mod2VE, mod2VZ,
     mod2VV, mod2VM, math_er, mod2VY}
    ,
    {mod2Ml, mod2Mb, mod2Mm, mod2Mr, mod2Mf,
     mod2Me, mod2Mc, mod2Mv, mod2Md, mod2Mz, mod2ME, mod2MZ,
     mod2MV, mod2MM, mod2Mq, mod2MY}
    ,
    {mod2ql, mod2qb, mod2qm, mod2qr, math_er,
     math_er, mod2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, mod2qM, mod2qq, mod2qY}
    ,
    {mod2Yl, mod2Yb, mod2Ym, mod2Yr, math_er,
     math_er, mod2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, mod2YM, mod2Yq, mod2YY}
};


Sex
mod2ii(Sex a, Sex b)
{
    long            n,
                    m;
    n = mkint(a);
    m = mkint(b);
    if (m == 0)
	return aerror2("modulus", a, b);
    n = n % m;
    if (m < 0) {
	if (n > 0)
	    n = n + m;
    } else if (n < 0)
	n = n + m;
    return mkfxat(n);
}

Sex
mod2xx(Sex a, Sex b)
{
    bool            f = minusp(b);
    a = remaind2(a, b);
    if (f) {
	if (plusp(a)) {
	    a = plus2(a, b);
	}
    } else if (minusp(a)) {
	a = plus2(a, b);
    }
    return a;
}





two_args       *smod_table_1[] =
    { smod2il, smod2ib, smod2im, smod2ir, smod2if,
    smod2ie, smod2ic, smod2iv, smod2id, smod2iz, smod2iE, smod2iZ,
    smod2iV, smod2iM, smod2iq, smod2iY
};

two_args       *smod_table_2[] =
    { smod2li, smod2bi, smod2mi, smod2ri, smod2fi,
    smod2ei, smod2ci, smod2vi, smod2di, smod2zi, smod2Ei, smod2Zi,
    smod2Vi, smod2Mi, smod2qi, smod2Yi
};

two_args       *smod_table[TYPEMAX][TYPEMAX] = {
    {smod2ll, smod2lb, smod2lm, smod2lr, smod2lf,
     smod2le, smod2lc, smod2lv, smod2ld, smod2lz, smod2lE, smod2lZ,
     smod2lV, smod2lM, smod2lq, smod2lY}
    ,
    {smod2bl, smod2bb, smod2bm, smod2br, smod2bf,
     smod2be, smod2bc, smod2bv, smod2bd, smod2bz, smod2bE, smod2bZ,
     smod2bV, smod2bM, smod2bq, smod2bY}
    ,
    {smod2ml, smod2mb, smod2mm, smod2mr, smod2mf,
     smod2me, smod2mc, smod2mv, smod2md, smod2mz, smod2mE, smod2mZ,
     smod2mV, smod2mM, smod2mq, smod2mY}
    ,
    {smod2rl, smod2rb, smod2rm, smod2rr, smod2rf,
     smod2re, smod2rc, smod2rv, smod2rd, smod2rz, smod2rE, smod2rZ,
     smod2rV, smod2rM, smod2rq, smod2rY}
    ,
    {smod2fl, smod2fb, smod2fm, smod2fr, smod2ff,
     smod2fe, smod2fc, smod2fv, smod2fd, smod2fz, smod2fE, smod2fZ,
     smod2fV, smod2fM, math_er, smod2fY}
    ,
    {smod2el, smod2eb, smod2em, smod2er, smod2ef,
     smod2ee, smod2ec, smod2ev, smod2ed, smod2ez, smod2eE, smod2eZ,
     smod2eV, smod2eM, math_er, smod2eY}
    ,
    {smod2cl, smod2cb, smod2cm, smod2cr, smod2cf,
     smod2ce, smod2cc, smod2cv, smod2cd, smod2cz, smod2cE, smod2cZ,
     smod2cV, smod2cM, smod2cq, smod2cY}
    ,
    {smod2vl, smod2vb, smod2vm, smod2vr, smod2vf,
     smod2ve, smod2vc, smod2vv, smod2vd, smod2vz, smod2vE, smod2vZ,
     smod2vV, smod2vM, math_er, smod2vY}
    ,
    {smod2dl, smod2db, smod2dm, smod2dr, smod2df,
     smod2de, smod2dc, smod2dv, smod2dd, smod2dz, smod2dE, smod2dZ,
     smod2dV, smod2dM, math_er, smod2dY}
    ,
    {smod2zl, smod2zb, smod2zm, smod2zr, smod2zf,
     smod2ze, smod2zc, smod2zv, smod2zd, smod2zz, smod2zE, smod2zZ,
     smod2zV, smod2zM, math_er, smod2zY}
    ,
    {smod2El, smod2Eb, smod2Em, smod2Er, smod2Ef,
     smod2Ee, smod2Ec, smod2Ev, smod2Ed, smod2Ez, smod2EE, smod2EZ,
     smod2EV, smod2EM, math_er, smod2EY}
    ,
    {smod2Zl, smod2Zb, smod2Zm, smod2Zr, smod2Zf,
     smod2Ze, smod2Zc, smod2Zv, smod2Zd, smod2Zz, smod2ZE, smod2ZZ,
     smod2ZV, smod2ZM, math_er, smod2ZY}
    ,
    {smod2Vl, smod2Vb, smod2Vm, smod2Vr, smod2Vf,
     smod2Ve, smod2Vc, smod2Vv, smod2Vd, smod2Vz, smod2VE, smod2VZ,
     smod2VV, smod2VM, math_er, smod2VY}
    ,
    {smod2Ml, smod2Mb, smod2Mm, smod2Mr, smod2Mf,
     smod2Me, smod2Mc, smod2Mv, smod2Md, smod2Mz, smod2ME, smod2MZ,
     smod2MV, smod2MM, smod2Mq, smod2MY}
    ,
    {smod2ql, smod2qb, smod2qm, smod2qr, math_er,
     math_er, smod2qc, math_er, math_er, math_er, math_er, math_er,
     math_er, smod2qM, smod2qq, smod2qY}
    ,
    {smod2Yl, smod2Yb, smod2Ym, smod2Yr, math_er,
     math_er, smod2Yc, math_er, math_er, math_er, math_er, math_er,
     math_er, smod2YM, smod2Yq, smod2YY}
};


Sex
smod2ii(Sex a, Sex b)
{
    long            n,
                    m;
    long            m2;
    n = mkint(a);
    m = mkint(b);
    if (m == 0)
	return aerror2("smodulus", a, b);
    if (m < 0)
	m = -m;
    n = n % m;
    m2 = m / 2;
    if (n > m2)
	n = n - m;
    else if (n <= -m2)
	n = n + m;
    return mkfxat(n);
}

Sex
smod2xx(Sex a, Sex b)
{
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define m stack[0]
    if (minusp(b))
	b = negate(b);
    m = quot2(b, mkfxat(2));
    a = remaind2(a, b);
    if (lessp2(m, a)) {
	a = diff2(a, b);
    } else {
	m = negate(m);
	if (lessp2(a, m))
	    a = plus2(a, b);
    }
#undef a
#undef b
#undef m
    a = stack[-2];
    popv(3);
    return a;
}

#if 0
Sex             modulusx(Sex a, Sex b);
Sex             smodulusx(Sex a, Sex b);

Sex
modulus(Sex a, Sex b)
{
    Sex             c;
    push2(a, b);
    c = modulusx(a, b);
    pop2(b, a);
    push(c);
    lisp_printf("%p mod %p = %p\n", a, b, c);
    pop(c);
    return c;
}
Sex
smodulus(Sex a, Sex b)
{
    Sex             c;
    push2(a, b);
    c = smodulusx(a, b);
    pop2(b, a);
    push(c);
    lisp_printf("%p smod %p = %p\n", a, b, c);
    pop(c);
    return c;
}
#endif

Sex
modulus(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n,
	                    m;
	    n = mkint(a);
	    m = mkint(b);
	    if (m == 0)
		return aerror2("modulus", a, b);
	    n = n % m;
	    if (m < 0) {
		if (n > 0)
		    n = n + m;
	    } else if (n < 0)
		n = n + m;
	    return mkfxat(n);
	} else if (is_numhdr(b))
	    return mod_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b))
	    return mod_table_2[numtype_of(a)] (a, b);
	else if (is_numhdr(b))
	    return mod_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("modulus", a, b);
}

Sex
smodulus(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    long            n,
	                    m;
	    long            m2;
	    n = mkint(a);
	    m = mkint(b);
	    if (m == 0)
		return aerror2("smodulus", a, b);
	    m2 = m / 2;
	    n = n % m;
	    if (n > m2)
		n = n - m;
	    else if (n <= -m2)
		n = n + m;
	    return mkfxat(n);
	} else if (is_numhdr(b))
	    return smod_table_1[numtype_of(b)] (a, b);
    } else if (is_numhdr(a)) {
	if (is_fixnum(b)) {
	    if (is_bigint(a) || is_logical(a)) {
		long            n,
		                m,
		                m2;
		m = mkint(b);
		a = mod2bi(a, b);
		n = mkint(a);
		m2 = m / 2;
		if (n > m2)
		    n = n - m;
		else if (n <= -m2)
		    n = n + m;
		return mkfxat(n);
	    }
	    return smod_table_2[numtype_of(a)] (a, b);
	} else if (is_numhdr(b))
	    return smod_table[numtype_of(a)][numtype_of(b)] (a, b);
    }
    return aerror2("smodulus", a, b);
}


Sex
reciprocal(Sex a)
{
    if (is_fixnum(a)) {
	return ratio2ii(mkfxat(1), a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return ratio2il(mkfxat(1), a);
	case N_BIGNUM:
	    return ratio2ib(mkfxat(1), a);
	case N_MODULAR:
	    return quot2im(mkfxat(1), a);
	case N_RATIONAL:
	    return quot2ir(mkfxat(1), a);
	case N_FLOAT:
	    return quot2if(mkfxat(1), a);
	case N_COMPLEX:
	    return quot2ic(mkfxat(1), a);
	case N_EFLOAT:
	    return quot2ie(mkfxat(1), a);
	case N_INTERVAL:
	    return quot2iv(mkfxat(1), a);
	case N_ECOMPLEX:
	    return quot2iz(mkfxat(1), a);
	case N_BFLOAT:
	    return quot2id(mkfxat(1), a);
	case N_BEFLOAT:
	    return quot2iE(mkfxat(1), a);
	case N_BINTERVAL:
	    return quot2iV(mkfxat(1), a);
	case N_BECOMPLEX:
	    return quot2iZ(mkfxat(1), a);
	case N_BMODULAR:
	    return quot2iM(mkfxat(1), a);
	}
    }
    return aerror2("recip", a, nil);
}

Sex
gcd_y(Sex a, Sex b)
{
    Sex             g;
    push2(a, b);
    g = gcd(a, b);
    pop2(b, a);
    push(g);
    if (minusp(g)) {
	// lisp_printf("gcd(%p, %p) --> %p\n", a, b, g);
	// lisp_printf("result is big ? %d \n", is_big(g) ? 1 : 0);
    }
    pop(g);
    return g;
}


Sex
gcd(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return gcdii(a, b);
	} else if (is_numhdr(b)) {
	    switch (numtype_of(b)) {
	    case N_LOGICAL:
	    case N_BIGNUM:
		return gcdbi(b, a);
	    case N_RATIONAL:
#ifdef RATIONAL_GCD
		push(b);
		a = gcdji(rational_num(b), a);
		pop(b);
		b = rational_den(b);
		return make_ratio(a, b);
#else
		return gcdji(rational_num(b), a);
#endif
	    case N_FLOAT:
	    case N_BFLOAT:
	    case N_EFLOAT:
	    case N_BEFLOAT:
	    case N_INTERVAL:
	    case N_BINTERVAL:
	    case N_MODULAR:
	    case N_BMODULAR:
	    case N_FMODULAR:
		return mkfxat(1);
	    }
	}
	return aerror2("gcd", a, b);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    if (is_fixnum(b)) {
		return gcdbi(a, b);
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
		    return gcdbb(a, b);
		case N_RATIONAL:
#ifdef RATIONAL_GCD
		    push(b);
		    a = gcdjj(a, rational_num(b));
		    pop(b);
		    b = rational_den(b);
		    return make_ratio(a, b);
#else
		    return gcdjj(a, rational_num(b));
#endif
		case N_FLOAT:
		case N_BFLOAT:
		case N_EFLOAT:
		case N_BEFLOAT:
		case N_INTERVAL:
		case N_BINTERVAL:
		case N_MODULAR:
		case N_BMODULAR:
		case N_FMODULAR:
		    return mkfxat(1);
		}
	    }
	    return aerror2("gcd", a, b);
	case N_RATIONAL:
	    if (is_fixnum(b)) {
#ifdef RATIONAL_GCD
		push(a);
		b = gcdji(rational_num(a), b);
		pop(a);
		a = rational_den(a);
		return make_ratio(b, a);
#else
		return gcdji(rational_num(a), b);
#endif
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
#ifdef RATIONAL_GCD
		    push(a);
		    b = gcdjj(rational_num(a), b);
		    pop(a);
		    return make_ratio(b, rational_den(a));
#else
		    return gcdjj(rational_num(a), b);
#endif
		case N_RATIONAL:
		    return gcdrr(a, b);
		case N_FLOAT:
		case N_BFLOAT:
		case N_EFLOAT:
		case N_BEFLOAT:
		case N_INTERVAL:
		case N_BINTERVAL:
		case N_MODULAR:
		case N_BMODULAR:
		    return mkfxat(1);
		}
	    }
	    return aerror2("gcd", a, b);
	case N_FLOAT:
	case N_BFLOAT:
	case N_EFLOAT:
	case N_BEFLOAT:
	case N_INTERVAL:
	case N_BINTERVAL:
	case N_MODULAR:
	case N_BMODULAR:
	case N_FMODULAR:
	    return mkfxat(1);
	}
    }
    return aerror2("gcd", a, b);
}

Sex
gcdrr(Sex a, Sex b)
{
    Sex             c,
                    d;
#ifdef RATIONAL_GCD
    push2(a, b);
    c = gcdjj(rational_num(a), rational_num(b));
    pop2(b, a);
    push(c);
    d = Llcm(rational_den(a), rational_den(b));
    pop(c);
    return make_ratio(c, d);
#else
    return gcdjj(rational_num(a), rational_num(b));
#endif
}

Sex
gcdjj(Sex a, Sex b)
{
    if (is_fixnum(b))
	return gcdji(a, b);
    if (is_fixnum(a))
	return gcdji(b, a);
    return gcdbb(a, b);
}

Sex
gcdji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return gcdii(a, b);
    if (b == mkfxat(0)) {
	if (is_bigneg(a))
	    return negateb(a);
	else if (is_big(a))
	    return a;
	if ((long) a < 0) {
	    long            aa = -mkint(a);
	    if (aa != big_min)
		return mkfxat(aa);
	    return mkbigat1(aa);
	}
	return a;
    }
    a = rem2bi(a, b);
    return gcdii(b, a);
}


Sex
gcdbi(Sex a, Sex b)
{
    if (b == mkfxat(0)) {
	if (is_bigneg(a))
	    return negateb(a);
	else if (is_big(a))
	    return a;
	if ((long) a < 0) {
	    long            aa = -mkint(a);
	    if (aa != big_min)
		return mkfxat(aa);
	    return mkbigat1(aa);
	}
	return a;
    }
    a = rem2bi(a, b);
    return gcdii(b, a);
}




	/*
	 * moved to bignum.c 2007.8.31 by f.kako 
	 */
#if 0
Sex
gcdbb(Sex a, Sex b)
{
    stackcheck();
    while (!is_fixnum(b)) {
	push(b);
	b = rem2bb(a, b);
	pop(a);
    }
    return gcdbi(a, b);
}
#endif


Sex
gcdii(Sex a, Sex b)
{
    long            p,
                    q,
                    r;

    p = mkint(a);
    if (p < 0)
	p = -p;
    q = mkint(b);
    if (q < 0)
	q = -q;
    if (p == 0)
	return mkfxat(q);
    while (q != 0) {
	r = p % q;
	p = q;
	q = r;
    }
    if (p == big_min)
	return mkbigat1(p);
    else
	return mkfxat(p);
}


Sex
Llcm(Sex a, Sex b)
{
    Sex             g;
    stackcheck();
    push2(b, a);
    g = gcd(a, b);
    pop(a);
    a = quot2(a, g);
    pop(b);
    a = times2(a, b);
    return a;
}

Sex
Lbfloat_man(Sex a)
{
    return bfloat_man(a);
}

Sex
Lbfloat_exp(Sex a)
{
    return mkfxat(bfloat_exp(a));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

#define PRIMEP_CONFIDENCE	40	/* 
					 * * * tolerance is 2^(-2*40)   
					 */


unsigned long
power_mod(unsigned long a, unsigned long b, unsigned long n)
{
    unsigned long   q,
                    p;
    unsigned long   h,
                    l;
    p = 1;
    while (b > 1) {
	if (b % 2 == 1) {
	    // p = umuldiv(&q, a, p, n);
	    h = umul(&l, a, p);
	    q = udiv(&p, h, l, n);

	}
	// a = umuldiv(&q, a, a, n);
	h = umul(&l, a, a);
	q = udiv(&a, h, l, n);
	b = b / 2;
    }
    // umuldiv(&q, a, p, n);
    h = umul(&l, a, p);
    q = udiv(&a, h, l, n);
    return a;
}


bool
miller_rabin(unsigned long n, unsigned long b)
{
    unsigned long   i,
                    q,
                    r,
                    s;
    unsigned long   p;
    unsigned long   h,
                    l;

    if (n == b)
	return YES;
    if (n % b == 0)
	return NO;
    q = n - 1;
    s = 0;
    while (q % 2 == 0) {
	q = q / 2;
	s++;
    }

    r = pow2mm(b, q, n);
    if (r == 1)
	return YES;

    for (i = 0; i < s; i++) {
	if (r == n - 1)
	    return YES;
	r = mult2mm(r, r, n);
    }
    return NO;
}


long
int_gcd(long a, long b)
{
    long            r;
    while ((r = a % b) != 0) {
	a = b;
	b = r;
    }
    return b;
}

bool
is_prime(long n)
{
    long            i,
                    r;
    unsigned long   x,
                    y,
                    m,
                    h,
                    l,
                    k;
    unsigned long   s,
                    w;


    if (n < 2)
	return NO;
    if (n == 2)
	return YES;
    if ((n & 1) == 0)
	return NO;
    if (n < 9080191) {
	if (!miller_rabin(n, 31))
	    return NO;
	if (!miller_rabin(n, 73))
	    return NO;
	return YES;
    }
#if _LP64
    if (n < 4759123141) {
	if (!miller_rabin(n, 2))
	    return NO;
	if (!miller_rabin(n, 7))
	    return NO;
	if (!miller_rabin(n, 61))
	    return NO;
	return YES;
    }
    if (n < 341550071728321) {
	if (!miller_rabin(n, 2))
	    return NO;
	if (!miller_rabin(n, 3))
	    return NO;
	if (!miller_rabin(n, 5))
	    return NO;
	if (!miller_rabin(n, 7))
	    return NO;
	if (!miller_rabin(n, 11))
	    return NO;
	if (!miller_rabin(n, 13))
	    return NO;
	if (!miller_rabin(n, 17))
	    return NO;
	return YES;
    }

    m = n - 1;
    s = 0;
    while (m % 2 == 0) {
	m = m / 2;
	s++;
    }
    for (i = 1; i < 60; i++) {
	w = random();
	w = (w % (n - 2)) + 1;
	x = power_mod(w, m, n);
	if (x == 1)
	    continue;
	for (k = 1; k <= s; k++) {
	    h = umul(&l, x, x);
	    (void) udiv(&y, h, l, n);
	    if (y == 1 && x != n - 1 && x != 1)
		return NO;
	    x = y;
	}
	if (x != 1)
	    return NO;
    }
    return YES;
#else
    if (!miller_rabin(n, 2))
	return NO;
    if (!miller_rabin(n, 7))
	return NO;
    if (!miller_rabin(n, 61))
	return NO;
    return YES;
#endif
}

Sex
modular_times(Sex a, Sex b, Sex m)
{
    if (is_fixnum(m)) {
	unsigned long   aa,
	                bb,
	                mm;
	unsigned long   h,
	                l;
	a = remaind2(a, m);
	b = remaind2(b, m);
	aa = mkint(a);
	bb = mkint(b);
	mm = mkint(m);
	h = umul(&l, aa, bb);
	bb = udiv(&aa, h, l, mm);
	return mkfxat(bb);
    }
    push(m);
    a = times2(a, b);
    pop(m);
    a = remaind2(a, m);
    return a;
}

Sex
modular_expt(Sex a, Sex b, Sex m)
{
    Sex             p;
    if (is_fixnum(m)) {
	long            aa,
	                bb,
	                mm;
	a = remaind2(a, m);
	b = remaind2(b, m);
	aa = mkint(a);
	bb = mkint(b);
	mm = mkint(m);
	return power_mod(aa, bb, mm);
    }
    p = mkfxat(1);
    push4(a, b, p, m);
#define a stack[-3]
#define b stack[-2]
#define p stack[-1]
#define m stack[-0]
    while (Lgreaterp(b, mkfxat(1)) != nil) {
	if (Levenp(b) != nil) {
	    p = times2(a, p);
	    p = remaind2(p, m);
	}
	a = times2(a, a);
	a = remaind2(a, m);
	b = leftshift(b, 1);
    }
    a = times2(a, p);
    a = remaind2(a, m);
#undef a
#undef b
#undef p
#undef m
    a = stack[-3];
    popv(4);
    return a;
}

Sex
Lprimep(Sex a)
{
    if (is_fixnum(a)) {
#if MILLER_RABIN == 1
	if (is_prime(mkint(a)))
	    return true;
	return nil;
#else
	long            n,
	                i;
	n = mkint(a);
	if (n < 0)
	    n = -n;
	if (n <= 1)
	    return nil;
	if (n < 4)
	    return true;
	if (n % 2 == 0)
	    return nil;
	for (i = 3; i * i <= n; i += 2)
	    if (n % i == 0)
		return nil;
	return true;
#endif
    } else if (is_bigint(a)) {
	long            i,
	                l;
	/*
	 * w; 
	 */
	/*
	 * Sex save; 
	 */
	long            k;
	Sex             result;
	/*
	 * Based on an algorithm of M.Rabin published in the Journal of Number
	 * Theory Vol 12, pp 128-138 (1980).
	 */

	if (Levenp(a) != nil || rem2bi(a, mkfxat(3)) == mkfxat(0))
	    return nil;

	push4(a, nil, nil, nil);
	push2(nil, nil);
#define a stack[-5]
#define m stack[-4]
#define x stack[-3]
#define y stack[-2]
#define w stack[-1]
#define a1 stack[0]
	m = Lsub1(a);
	a1 = m;
	l = 0;
	while (Levenp(m) != nil) {
	    m = rightshift(m, 1);
	    l++;
	}
	i = 1;
	result = true;
	while (result != nil && i <= PRIMEP_CONFIDENCE) {
	    w = Ladd1(Lrandom(Lsub1(a1)));
	    x = modular_expt(w, m, a);
	    if (x != mkfxat(1)) {
		for (k = 1; k <= l; k++) {
		    y = times2(x, x);
		    y = remaind2(y, a);
		    if (y == mkfxat(1) && !numeq2(x, a1) && x != mkfxat(1))
			result = nil;
		    else
			x = y;
		}
	    }
	    if (x != mkfxat(1))
		result = nil;
	    i++;
	}
#undef a
#undef m
#undef x
#undef y
#undef w
#undef a1
	popv(6);
	return result;
    } else
	return error("primep : not a integer");
}

Sex
Lgen_prime(Sex a, Sex b)
{
    long            l,
                    u;
    long            i;

    if (!is_fixnum(a) || !is_fixnum(b))
	error2("gen-prime: invalid argument", a, b);
    l = mkint(a);
    u = mkint(b);
    stackcheck();
    push(nil);
#define c stack[0]
    if (l < 3) {
	c = cons(mkfxat(2), nil);
	l = 3;
    } else if (l % 2 == 0)
	l++;
    while (l <= u) {
	for (i = 3; i * i <= l; i += 2)
	    if (l % i == 0)
		goto again;
	c = cons(mkfxat(l), c);
      again:
	l += 2;
    }
#undef c
    pop(a);
    return Lnreverse(a);
}

Sex
Lnext_prime(Sex a)
{
    long            l;
    long            i;

    if (!is_fixnum(a))
	error1("next-prime: invalid argument", a);
    l = mkint(a) + 1;
    if (l <= 2)
	return mkfxat(2);
    if (l % 2 == 0)
	l++;
    while (1) {
	for (i = 3; i * i <= l; i += 2)
	    if (l % i == 0)
		goto again;
	return mkfxat(l);
      again:
	l = l + 2;
    }
}


Sex
Lprime_before(Sex a)
{
    long            l;
    long            i;

    if (!is_fixnum(a))
	error1("prime-before: invalid argument", a);
    l = mkint(a) - 1;

    if (l == 2)
	return mkfxat(2);
    if (l < 2)
	l = mkint(fxat_max);
    if (l % 2 == 0)
	l--;
    while (1) {
	for (i = 3; i * i <= l; i += 2)
	    if (l % i == 0)
		goto again;
	return mkfxat(l);
      again:
	l = l - 2;
    }
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
Sex
zfactor(Sex n)
{
    return zfactor1(n, true);
}

Sex
zfactor1(Sex a, bool f)
{
    return nil;
}

Sex
Lillegal_bignum_zero(Sex a)
{
    unsigned long  *pa;
    long            l,
                    i;

    if (is_bigint(a)) {
	l = bignum_length(a);
	l = bnum_n(l);
	pa = &bignum_value(a)[l - 1];
	for (i = 0; i < l; i++) {
	    if (*pa-- != 0)
		break;
	    return true;
	}

    }
    return nil;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


extern Sex      Lgeqn(Sex a, Sex b);


setup_type const arith_setup[] = {
    {"frexp", (lisp_proc *) Lfrexp, 1},
    {"ceiling", (lisp_proc *) Lceiling, 1},
    {"floor", (lisp_proc *) Lfloor, 1},
    {"lcm", (lisp_proc *) Llcm, 2},
    {"round", (lisp_proc *) Lround, 1},
    {"truncate", (lisp_proc *) Ltruncate, 1},
    {"gcdn", (lisp_proc *) gcd, 2},
#if 1
    {"gcd", (lisp_proc *) gcd, 2},
#else
#ifdef ASUKA
    {"gcd", (lisp_proc *) ggcd2, 2},
#else
    {"gcd", (lisp_proc *) gcd, 2},
#endif
#endif
    {"denominator", (lisp_proc *) Ldenominator, 1},
    {"numerator", (lisp_proc *) Lnumerator, 1},
    {"ash", (lisp_proc *) Lash, 2},
    {"ash1", (lisp_proc *) Lash1, 2},
    {"divide", (lisp_proc *) Ldivide, 2},
    {"evenp", (lisp_proc *) Levenp, 1},
    {"logand2", (lisp_proc *) logand2, 2},
    {"logeqv2", (lisp_proc *) logeqv2, 2},
    {"lognot", (lisp_proc *) lognot, 1},
    {"logxor2", (lisp_proc *) logxor2, 2},
    {"logor2", (lisp_proc *) logor2, 2},
    {"lsd", (lisp_proc *) Llsd, 1},
    {"srandom", (lisp_proc *) Lsrandom, 1},
    {"random", (lisp_proc *) Lrandom, 1},
    {"max2", (lisp_proc *) Lmax2, 2},
    {"min2", (lisp_proc *) Lmin2, 2},
#ifdef ASUKA
    {"minus", (lisp_proc *) gnegate, 1},
#else
    {"minus", (lisp_proc *) negate, 1},
#endif
    {"minusp", (lisp_proc *) Lminusp, 1},
#ifdef ASUKA
    {"mod", (lisp_proc *) gmodulus, 2},
    {"smod", (lisp_proc *) gsmodulus, 2},
#else
    {"mod", (lisp_proc *) modulus, 2},
#endif
    {"msd", (lisp_proc *) Lmsd, 1},
    {"oddp", (lisp_proc *) Loddp, 1},
    {"onep", (lisp_proc *) Lonep, 1},
    {"plus2", (lisp_proc *) plus2, 2},
    {"plusp", (lisp_proc *) Lplusp, 1},
    {"times2", (lisp_proc *) times2, 2},
    {"zerop", (lisp_proc *) Lzerop, 1},
    {"add1", (lisp_proc *) Ladd1, 1},
    {"difference", (lisp_proc *) diff2, 2},
    {"eqn", (lisp_proc *) Leqn, 2},
    {"appeqn", (lisp_proc *) Lappeq, 2},
    {"geq", (lisp_proc *) Lgeq, 2},
    {"greaterp", (lisp_proc *) Lgreaterp, 2},
    {"leq", (lisp_proc *) Lleq, 2},
    {"lep", (lisp_proc *) Lleq, 2},
    {"gep", (lisp_proc *) Lgeq, 2},
    {"lessp", (lisp_proc *) Llessp, 2},
    {"cgeq", (lisp_proc *) Lcgeq, 2},
    {"cgreaterp", (lisp_proc *) Lcgreaterp, 2},
    {"cleq", (lisp_proc *) Lcleq, 2},
    {"clep", (lisp_proc *) Lcleq, 2},
    {"cgep", (lisp_proc *) Lcgeq, 2},
    {"clessp", (lisp_proc *) Lclessp, 2},
    {"ratio", (lisp_proc *) ratio2, 2},
    {"quotient", (lisp_proc *) quot2, 2},
    {"remainder", (lisp_proc *) remaind2, 2},
    {"reciprocal", (lisp_proc *) reciprocal, 1},
    {"sub1", (lisp_proc *) Lsub1, 1},
    {"precision", (lisp_proc *) Lprecision, 1},
    /*
     * type conversion routines 
     */
    {"fix", (lisp_proc *) Lfix, 1},
    {"mklogical", (lisp_proc *) Llogical, 1},
    {"logical", (lisp_proc *) Llogical, 1},
    {"rational", (lisp_proc *) Lrational, 1},
    {"mkrational", (lisp_proc *) Lmkrational, 2},
    {"float", (lisp_proc *) Lfloat, 1},
    {"mkfloat", (lisp_proc *) Lfloat, 1},
    {"efloat", (lisp_proc *) Lefloat, 1},
    {"mkefloat", (lisp_proc *) Lmkefloat, 2},
    {"mkefloat1", (lisp_proc *) Lmkefloat1, 1},
    {"interval", (lisp_proc *) Linterval, 1},
    {"mkinterval", (lisp_proc *) Lmkinterval, 2},
    {"complex", (lisp_proc *) Lcomplex, 1},
    {"mkcomplex", (lisp_proc *) Lmkcomplex, 2},
    {"ecomplex", (lisp_proc *) Lecomplex, 1},
    {"mkecomplex", (lisp_proc *) Lmkecomplex, 3},
    {"bfloat", (lisp_proc *) Lbfloat, 1},
    {"mkbfloat", (lisp_proc *) Lbfloat, 1},
    {"befloat", (lisp_proc *) Lbefloat, 1},
    {"mkbefloat", (lisp_proc *) Lmkbefloat, 2},
    {"binterval", (lisp_proc *) Lbinterval, 1},
    {"mkbinterval", (lisp_proc *) Lmkbinterval, 2},
    {"becomplex", (lisp_proc *) Lbecomplex, 1},
    {"mkbecomplex", (lisp_proc *) Lmkbecomplex, 3},
    {"#c", (lisp_proc *) Lmkcomplex, 2},
    {"#ef", (lisp_proc *) Lmkefloat, 2},
    {"#ec", (lisp_proc *) Lmkecomplex, 3},
    {"#cc", (lisp_proc *) Lcomplex_conjugate, 1},
    {"#bef", (lisp_proc *) Lmkbefloat, 2},
    {"mkmodular", (lisp_proc *) Lmkmodular, 1},
    {"unmodular", (lisp_proc *) Lunmodular, 1},
    /*
     * interval 
     */
    {"#itvlow", (lisp_proc *) Linterval_lower, 1},
    {"#itvupp", (lisp_proc *) Linterval_upper, 1},
    {"#itvmid", (lisp_proc *) Linterval_mid, 1},
    {"#itvwid", (lisp_proc *) Linterval_width, 1},
    {"#itv", (lisp_proc *) Lmkinterval, 2},
    /*
     * efloat
     */
    {"#epart", (lisp_proc *) Lefloat_error, 1},
    {"#vpart", (lisp_proc *) Lefloat_value, 1},
    {"#cvpart", (lisp_proc *) Lecomplex_value, 1},
    {"#cepart", (lisp_proc *) Lecomplex_error, 1},
    {"#rvpart", (lisp_proc *) Lecomplex_real, 1},
    {"#ivpart", (lisp_proc *) Lecomplex_imag, 1},
    {"#rpart", (lisp_proc *) Lcomplex_real, 1},
    {"#ipart", (lisp_proc *) Lcomplex_imag, 1},
    {"#bepart", (lisp_proc *) Lbefloat_error, 1},
    {"#bvpart", (lisp_proc *) Lbefloat_value, 1},
    /*
     */
#ifdef ASUKA
    {"+", (lisp_proc *) gplus2, 2},
    {"-", (lisp_proc *) gdiff2, 2},
    {"*", (lisp_proc *) gtimes2, 2},
    {"/", (lisp_proc *) gratio2, 2},
    {"//", (lisp_proc *) gremaind2, 2},
    {"^", (lisp_proc *) gexpt, 2},
#else
    {"+", (lisp_proc *) plus2, 2},
    {"-", (lisp_proc *) diff2, 2},
    {"*", (lisp_proc *) times2, 2},
    {"/", (lisp_proc *) quot2, 2},
    {"//", (lisp_proc *) remaind2, 2},
    {"^", (lisp_proc *) Lexpt, 2},
#endif
#ifdef ASUKA
    {"=", (lisp_proc *) Lgeqn, 2},
#else
    {"=", (lisp_proc *) Leqn, 2},
#endif
    {"=:=", (lisp_proc *) Lappeq, 2},
    {"<", (lisp_proc *) Llessp, 2},
    {">", (lisp_proc *) Lgreaterp, 2},
    {">=", (lisp_proc *) Lgeq, 2},
    {"<=", (lisp_proc *) Lleq, 2},
    /*
     */
    {"bfloat_man", (lisp_proc *) Lbfloat_man, 1},
    {"bfloat_exp", (lisp_proc *) Lbfloat_exp, 1},
    {"leftshift", (lisp_proc *) Lleftshift, 2},
    {"gen-prime", (lisp_proc *) Lgen_prime, 2},
    {"gen_prime", (lisp_proc *) Lgen_prime, 2},
    {"next-prime", (lisp_proc *) Lnext_prime, 1},
    {"next_prime", (lisp_proc *) Lnext_prime, 1},

    {"primep", (lisp_proc *) Lprimep, 1},

    {"number-log2", (lisp_proc *) Lnumber_log2, 1},
    {"safe-fp-plus", (lisp_proc *) Lsafe_fp_plus, 2},
    {"safe-fp-times", (lisp_proc *) Lsafe_fp_times, 2},
    {"safe-fp-quot", (lisp_proc *) Lsafe_fp_quot, 2},

    {"illegal-bignum-zero", (lisp_proc *) Lillegal_bignum_zero, 1},
    {"illegal_bignum_zero", (lisp_proc *) Lillegal_bignum_zero, 1},

    {NULL, 0, 0}
};
