/*
 * becomplex.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * big effective complex number
 *      $Header: /Users/kako/asuka/src/RCS/becomplex.c,v 2.4 2015/12/09 03:44:28 kako Exp $
 */

static char const rcsid[] =
    "$Id: becomplex.c,v 2.4 2015/12/09 03:44:28 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"



/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex             becomplexit(Sex a);
Sex             becomplexiti(Sex a);
Sex             becomplexitf(Sex a);

Sex             Lbecomplex_conjugate(Sex a);
Sex             Lbecomplex_real(Sex a);
Sex             Lbecomplex_imag(Sex a);

Sex             plus2Ec(Sex a, Sex b);
Sex             plus2Zc(Sex a, Sex b);
Sex             times2VZ(Sex a, Sex b);
double          absZ(Sex a);

bool            lessp_f_d(double x, Sex a);

bool            appeq2Zc(Sex a, Sex b);


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
fixitZ(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = fixitd(becomplex_real(a));
    pop(a);
    push(c);
    a = fixitd(becomplex_imag(a));
    pop(c);
    return Lmkcomplex(c, a);
}

Sex
becomplexiti(Sex a)
{
    Sex             c;
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatiti(a);
    ci = bfloatiti(mkfxat(0));
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = 0.0;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitb(Sex a)
{
    Sex             c;
    double          x;
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatitb(a);
    x = floatitb(a);
    x = fabs(x) * current_precision;
    ci = bfloatiti(mkfxat(0));
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = x;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitf(Sex a)
{
    Sex             c;
    double          x;
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatitf(float_val(a));
    /*
     * x = float_val(a) * flt_min;
     */
    x = float_val(a) * current_precision;
    x = fabs(x);
    ci = bfloatiti(mkfxat(0));
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = x;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitd(Sex a)
{
    Sex             c;
    double          x;
    stackcheck();
    push2(a, nil);
#define cr stack[-1]
#define ci stack[0]
    ci = bfloatiti(mkfxat(0));
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    x = floatitd(cr);
    x = fabs(x) * current_precision;
    becomplex_error(c) = x;
    popv(2);
#undef cr
#undef ci
    return c;
}


Sex
becomplexitc(Sex a)
{
    Sex             c;
    double          xr,
                    xi;
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatit(complex_real(a));
    ci = bfloatit(complex_imag(a));
    xr = fabs(floatitd(cr));
    xi = fabs(floatitd(ci));
    if (xr < xi)
	xr = xi;
    xr = xr * current_precision;
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = xr;
    popv(3);
#undef a
#undef cr
#undef ci
    return c;
}

Sex
becomplexite(Sex a)
{
    Sex             c;
    double          e;
    e = efloat_error(a);
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatitf(efloat_value(a));
    ci = bfloatitf(0.0);
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = e;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitv(Sex a)
{
    Sex             c;
    double          v,
                    e;

    v = interval_upper(a) + interval_lower(a);
    e = interval_upper(a) - interval_lower(a);
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatitf(v * 0.5);
    ci = bfloatitf(0.0);
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = e * 0.5;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitz(Sex a)
{
    Sex             c;
    double          e;
    double          zr,
                    zi;

    zr = ecomplex_real(a);
    zi = ecomplex_imag(a);
    e = ecomplex_error(a);
    stackcheck();
    push2(nil, nil);
#define cr stack[-1]
#define ci stack[0]
    cr = bfloatitf(zr);
    ci = bfloatitf(zi);
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = e;
    popv(2);
#undef cr
#undef ci
    return c;
}

Sex
becomplexitE(Sex a)
{
    Sex             c;
    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define ci stack[0]
    ci = bfloatitf(0.0);
    c = new_becomplex();
    becomplex_real(c) = befloat_value(a);
    becomplex_imag(c) = ci;
    becomplex_error(c) = befloat_error(a);
    popv(2);
#undef a
#undef ci
    return c;
}

Sex
becomplexitV(Sex a)
{
    Sex             c;
    double          e;
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define cr stack[-1]
#define ci stack[0]
    cr = plus2dd(binterval_upper(a), binterval_lower(a));
    ci = diff2dd(binterval_upper(a), binterval_lower(a));
    e = floatitd(ci) * 0.5;
    cr = leftshiftd(cr, -1);
    ci = bfloatitf(0.0);
    c = new_becomplex();
    becomplex_real(c) = cr;
    becomplex_imag(c) = ci;
    becomplex_error(c) = e;
    popv(3);
#undef a
#undef cr
#undef ci
    return c;
}

Sex
Lbecomplex(Sex a)
{
    if (is_fixnum(a)) {
	return becomplexiti(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return becomplexitb(a);
	case N_FLOAT:
	    return becomplexitf(a);
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_BFLOAT:
	    return becomplexitd(a);
	case N_EFLOAT:
	    return becomplexite(a);
	case N_INTERVAL:
	    return becomplexitv(a);
	case N_BEFLOAT:
	    return becomplexitE(a);
	case N_BINTERVAL:
	    return becomplexitV(a);
	case N_COMPLEX:
	    return becomplexitc(a);
	case N_ECOMPLEX:
	    return becomplexitz(a);
	case N_BECOMPLEX:
	    return a;
	}
    }
    return error1("bad argument to becomplex", a);
}



Sex
Lmkbecomplex(Sex r, Sex i, Sex e)
{
    Sex             a;
    double          er;

    er = floatit(e);
    er = fabs(er);
    stackcheck();
    push2(r, i);
    stack[-1] = bfloatit(r);
    stack[0] = bfloatit(stack[0]);
    a = new_becomplex();
    pop2(i, r);
    becomplex_real(a) = r;
    becomplex_imag(a) = i;
    becomplex_error(a) = er;
    return a;
}



Sex
make_becomplex(Sex r, Sex i, double e)
{
    Sex             a;
    ASSERT(is_bfloat(r) && is_bfloat(i));
    stackcheck();
    push2(r, i);
    a = new_becomplex();
    pop2(i, r);
    becomplex_real(a) = r;
    becomplex_imag(a) = i;
    becomplex_error(a) = e;
    return a;
}



Sex
Lbecomplex_conjugate(Sex u)
{
    Sex             v;
    if (is_fixnum(u))
	return u;
    if (!is_numhdr(u))
	aerror("complelx_conjugate: not number");
    switch (numtype_of(u)) {
    case N_COMPLEX:
	stackcheck();
	push(u);
	v = new_complex();
	u = stack[0];
	stack[0] = v;
	complex_real(v) = complex_real(u);
	complex_imag(stack[0]) = negate(complex_imag(u));
	pop(v);
	return v;
    case N_ECOMPLEX:
	return make_ecomplex(ecomplex_real(u), -ecomplex_imag(u),
			     ecomplex_error(u));
    case N_BECOMPLEX:
	{
	    Sex             ur,
	                    ui;
	    long            ue;
	    ue = becomplex_error(u);
	    ur = becomplex_real(u);
	    ui = becomplex_imag(u);
	    stackcheck();
	    push(ur);
	    ui = negate(ui);
	    pop(ur);
	    return make_becomplex(ur, ui, ue);
	}
    default:
	return u;
    }
}

Sex
Lbecomplex_real(Sex a)
{
    if (is_fixnum(a))
	return a;
    if (!is_numhdr(a))
	aerror("complelx real: not number");
    switch (numtype_of(a)) {
    case N_COMPLEX:
	return complex_real(a);
    case N_ECOMPLEX:
	return make_efloat(ecomplex_real(a), ecomplex_error(a));
    case N_BECOMPLEX:
	return make_befloat(becomplex_real(a), becomplex_error(a));
    default:
	return a;
    }
}

Sex
Lbecomplex_imag(Sex a)
{
    if (is_fixnum(a))
	return mkfxat(0);
    if (!is_numhdr(a))
	aerror("complelx real: not number");
    switch (numtype_of(a)) {
    case N_COMPLEX:
	return complex_imag(a);
    case N_ECOMPLEX:
	return make_efloat(ecomplex_imag(a), ecomplex_error(a));
    case N_BECOMPLEX:
	return make_befloat(becomplex_imag(a), becomplex_error(a));
    default:
	return a;
    }
}

void
print_becomplex(Sex u, putchar_fn * fn)
{
    double          er;
    long            n,
                    m;
    Sex             c;
    double          d;
    extern void     print_error(double d, putchar_fn * fn);
    er = becomplex_error(u);
    n = bfloat_log2(becomplex_real(u));
    m = bfloat_log2(becomplex_imag(u));
    if (n < m) {
	n = m;
    }
    stackcheck();
    push2(becomplex_real(u), becomplex_imag(u));
    (*fn) ('#');
    (*fn) ('E');
    (*fn) ('C');
    (*fn) ('(');
    print_bfloat(stack[-1], fn);
    (*fn) (',');
    print_bfloat(stack[0], fn);
    (*fn) (',');
    popv(2);
    /*
     * print_expt10(er + n, fn); (*fn) (','); print_int(er, fn); (*fn)
     * (')'); 
     */
    print_error(er, fn);
    (*fn) (')');
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2fZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
#if TOWARD_LOWER_PRECISION
    double          xr,
                    xi;
    e = becomplex_error(b);
    xr = floatitd(becomplex_real(b));
    xi = floatitd(becomplex_imag(b));
    xr = xr + float_val(a);
    return make_ecomplex(xr, xi, e);

#else
    e = becomplex_error(b);
    stackcheck();
    push(b);
    c = plus2(a, becomplex_real(b));
    pop(b);
    d = becomplex_imag(b);
    return make_becomplex(c, d, e);
#endif
}

Sex
plus2xZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
    e = becomplex_error(b);
    stackcheck();
    push(b);
    c = plus2(a, becomplex_real(b));
    pop(b);
    d = becomplex_imag(b);
    return make_becomplex(c, d, e);
}

Sex
plus2cZ(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	b = stack[0];
	stack[0] = a;
	b = ecomplexitZ(b);
	pop(a);
	return plus2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return plus2ZZ(a, b);
    }
#else
    Sex             c,
                    d;
    double          e = becomplex_error(b);
    if (is_shortfloat(a)) {
	double          x,
	                y,
	                u,
	                v;
	x = floatit(complex_real(a));
	y = floatit(complex_imag(a));
	u = floatitd(becomplex_real(b));
	v = floatitd(becomplex_imag(b));
	x = x + u;
	y = y + v;
	return make_ecomplex(x, y, e);
    }
    stackcheck();
    push2(a, b);
    c = plus2(complex_real(a), becomplex_real(b));
    push(c);
    b = stack[-1];
    a = stack[-2];
    d = plus2(complex_imag(a), becomplex_imag(b));
    b = stack[-1];
    c = stack[0];
    popv(3);
    return make_becomplex(c, d, e);
#endif
}

Sex
plus2eZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return plus2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    return plus2EZ(a, b);
#endif
}


Sex
plus2vZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    push(b);
    a = efloatitv(a);
    pop(b);
    return plus2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitv(a);
    pop(b);
    return plus2EZ(a, b);
#endif
}

Sex
plus2zZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return plus2zz(a, b);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return plus2ZZ(a, b);
#endif
}


Sex
plus2EZ(Sex a, Sex b)
{
    double          e;

    e = befloat_error(a);
    if (e < becomplex_error(b))
	e = becomplex_error(b);
    stackcheck();
    push(b);
    a = plus2dd(befloat_value(a), becomplex_real(b));
    pop(b);
    b = becomplex_imag(b);
    return make_becomplex(a, b, e);
}


Sex
plus2ZZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
    e = becomplex_error(a);
    if (e < becomplex_error(b))
	e = becomplex_error(b);
    stackcheck();
    push2(a, b);
    c = plus2(becomplex_real(a), becomplex_real(b));
    a = stack[-1];
    b = stack[0];
    push(c);
    d = plus2(becomplex_imag(a), becomplex_imag(b));
    pop(c);
    a = stack[-1];
    b = stack[0];
    popv(2);
    return make_becomplex(c, d, e);
}

Sex
plus2ZV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return plus2EZ(b, a);
}

#if 0
Sex
plus2Ec(Sex a, Sex b)
{
    double          n;
    Sex             br,
                    bi;

    n = befloat_error(a);
    stackcheck();
    push2(b, a);
    br = bfloatit(complex_real(b));
    pop(a);
    br = plus2dd(befloat_value(a), br);
    pop(b);
    push(br);
    bi = bfloatit(complex_imag(b));
    pop(br);
    return make_becomplex(br, bi, n);
}
#endif


#if 0
Sex
plus2cV(Sex a, Sex b)
{
    double          e;

    stackcheck();
    push5(a, b, nil, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define cr stack[-3]
#define ci stack[-2]
#define vv stack[-1]
#define ve stack[0]
    vv = plus2dd(binterval_upper(b), binterval_lower(b));
    ve = diff2dd(binterval_upper(b), binterval_lower(b));
    e = floatit(ve) * 0.5;
    vv = leftshiftd(vv, -1);
    cr = bfloatit(complex_real(a));
    vv = plus2dd(vv, cr);
    ci = bfloatit(complex_imag(a));
#undef a
#undef b
    a = cr;
    b = ci;
#undef cr
#undef ci
#undef vv
#undef ve
    popv(6);
    return make_becomplex(a, b, e);
}
#endif

Sex
plus2zV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return plus2ez(b, a);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return plus2ZV(a, b);
#endif
}



Sex
plus2Zc(Sex a, Sex b)
{
    stackcheck();
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	pop(a);
	return plus2zZ(b, a);
    } else {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return plus2ZZ(a, b);
    }
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
negateZ(Sex a)
{
    double          e;
    Sex             r,
                    i;

    r = becomplex_real(a);
    i = becomplex_imag(a);
    e = becomplex_error(a);
    stackcheck();
    push(i);
    r = negated(r);
    pop(i);
    push(r);
    i = negated(i);
    pop(r);
    return make_becomplex(r, i, e);
}

Sex
diff2fZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return diff2fz(a, b);
#else
    Sex             c,
                    d;
    double          e;
    stackcheck();
    push(b);
    c = diff2(a, becomplex_real(b));
    pop(b);
    d = becomplex_imag(b);
    e = becomplex_error(b);
    return make_becomplex(c, d, e);
#endif
}

Sex
diff2xZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
    stackcheck();
    push(b);
    c = diff2(a, becomplex_real(b));
    pop(b);
    d = becomplex_imag(b);
    e = becomplex_error(b);
    return make_becomplex(c, d, e);
}

Sex
diff2Zf(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    // lisp_printf("diff2Zf: a = %p\n", a);
    pop(b);
    return diff2zf(a, b);
#else
    Sex             c,
                    d;
    double          e;
    stackcheck();
    push(a);
    c = diff2(becomplex_real(a), b);
    pop(a);
    d = becomplex_imag(a);
    e = becomplex_error(a);
    return make_becomplex(c, d, e);
#endif
}

Sex
diff2Zx(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
    stackcheck();
    push(a);
    c = diff2(becomplex_real(a), b);
    pop(a);
    d = becomplex_imag(a);
    e = becomplex_error(a);
    return make_becomplex(c, d, e);
}

Sex
diff2cZ(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	b = stack[0];
	stack[0] = a;
	b = ecomplexitZ(b);
	pop(a);
	return diff2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return diff2ZZ(a, b);
    }
#else
    Sex             c,
                    d;
    double          e;
    if (is_shortfloat(a)) {
	double          x,
	                y,
	                u,
	                v;
	x = floatit(complex_real(a));
	y = floatit(complex_imag(a));
	u = floatitd(becomplex_real(b));
	v = floatitd(becomplex_imag(b));
	x = x - u;
	y = y - v;
	e = becomplex_error(b);
	return make_ecomplex(x, y, e);
    }
    stackcheck();
    push2(a, b);
    c = diff2(complex_real(a), becomplex_real(b));
    a = stack[-1];
    b = stack[0];
    push(c);
    d = diff2(complex_imag(a), becomplex_imag(b));
    pop(c);
    b = stack[0];
    e = becomplex_error(b);
    popv(2);
    return make_becomplex(c, d, e);
#endif
}


Sex
diff2Zc(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	a = stack[0];
	stack[0] = b;
	a = ecomplexitZ(a);
	pop(b);
	return diff2zz(a, b);
    } else {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return diff2ZZ(a, b);
    }
#else
    Sex             c,
                    d;
    double          e;

    stackcheck();
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	pop(a);
	return diff2Zz(a, b);
    }
    push2(a, b);
    c = diff2(becomplex_real(a), complex_real(b));
    a = stack[-1];
    b = stack[0];
    push(c);
    d = diff2(becomplex_imag(a), complex_imag(b));
    pop(c);
    a = stack[-1];
    e = becomplex_error(a);
    popv(2);
    return make_becomplex(c, d, e);
#endif
}


Sex
diff2eZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return diff2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    return diff2EZ(a, b);
#endif
}

Sex
diff2Ze(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return diff2ze(a, b);
#else
    stackcheck();
    push(a);
    b = befloatite(b);
    pop(a);
    return diff2ZE(a, b);
#endif
}


Sex
diff2vZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    push(b);
    a = efloatitv(a);
    pop(b);
    return diff2ez(a, b);
#else
    double          vv,
                    ve;
    stackcheck();
    push(b);
    vv = (interval_upper(a) + interval_lower(a)) / 2;
    ve = (interval_upper(a) - interval_lower(a)) / 2;
    if (ve < becomplex_error(b))
	ve = becomplex_error(b);
    a = bfloatitf(vv);
    b = stack[0];
    a = diff2dd(a, becomplex_real(b));
    pop(b);
    return make_becomplex(a, becomplex_imag(b), ve);
#endif
}

Sex
diff2Zv(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitv(b);
    pop(a);
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return diff2ze(a, b);

#else
    double          vv,
                    ve;
    stackcheck();
    push(a);
    vv = (interval_upper(b) + interval_lower(b)) / 2;
    ve = (interval_upper(b) - interval_lower(b)) / 2;
    if (ve < becomplex_error(a))
	ve = becomplex_error(a);
    b = bfloatitf(vv);
    a = stack[0];
    b = diff2dd(becomplex_real(a), b);
    pop(a);
    return make_becomplex(b, becomplex_imag(a), ve);
#endif
}

Sex
diff2zZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return diff2zz(a, b);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return diff2ZZ(a, b);
#endif
}

Sex
diff2Zz(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return diff2zz(a, b);
#else
    stackcheck();
    push(a);
    b = becomplexitz(b);
    pop(b);
    return diff2ZZ(a, b);
#endif
}


Sex
diff2EZ(Sex a, Sex b)
{
    double          e;

    e = befloat_error(a);
    if (e < becomplex_error(b))
	e = becomplex_error(b);
    stackcheck();
    push(becomplex_imag(b));
    a = diff2dd(befloat_value(a), becomplex_real(b));
    pop(b);
    push(a);
    b = negated(b);
    pop(a);
    return make_becomplex(a, b, e);
}


Sex
diff2ZE(Sex a, Sex b)
{
    double          e;

    e = becomplex_error(a);
    if (e < befloat_error(b))
	e = befloat_error(b);
    stackcheck();
    push(becomplex_imag(a));
    a = diff2(becomplex_real(a), befloat_value(b));
    pop(b);
    return make_becomplex(a, b, e);
}

Sex
diff2VZ(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    return diff2EZ(a, b);
}


Sex
diff2ZV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return diff2ZE(a, b);
}


Sex
diff2ZZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;

    e = becomplex_error(a);
    if (e < becomplex_error(b))
	e = becomplex_error(b);
    stackcheck();
    push2(a, b);
    c = diff2dd(becomplex_real(a), becomplex_real(b));
    pop2(b, a);
    push(c);
    d = diff2dd(becomplex_imag(a), becomplex_imag(b));
    pop(c);
    return make_becomplex(c, d, e);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2fZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return times2fz(a, b);
#else
    Sex             c,
                    d;
    double          e;
    double          f;

    e = becomplex_error(b);
    f = float_val(a);
    e = fabs(a * f);
    stackcheck();
    push2(a, b);
    c = times2(a, becomplex_real(b));
    a = stack[-1];
    b = stack[0];
    stack[0] = c;
    d = times2(a, becomplex_imag(b));
    c = stack[0];
    popv(2);
    return make_becomplex(c, d, e);
#endif
}

Sex
times2xZ(Sex a, Sex b)
{
    Sex             c,
                    d;
    double          e;
    double          f;

    e = becomplex_error(b);
    f = floatit(a);
    e = fabs(e * f);
    stackcheck();
    push2(a, b);
    c = times2(a, becomplex_real(b));
    a = stack[-1];
    b = stack[0];
    stack[0] = c;
    d = times2(a, becomplex_imag(b));
    c = stack[0];
    popv(2);
    return make_becomplex(c, d, e);
}


Sex
times2cZ(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	b = stack[0];
	stack[0] = a;
	b = ecomplexitZ(b);
	pop(a);
	return times2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return times2ZZ(a, b);
    }
#else
    Sex             w;
    double          e;
    double          dr,
                    di;
    stackcheck();

    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return times2zZ(a, b);
    }

    e = becomplex_error(b);
    dr = fabs(floatit(complex_real(a)));
    di = fabs(floatit(complex_imag(a)));
    if (dr < di)
	dr = di;
    e = e * dr;

    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[0]
    c = times2(complex_real(a), becomplex_real(b));
    w = times2(complex_imag(a), becomplex_imag(b));
    c = diff2(c, w);
    d = times2(complex_real(a), becomplex_imag(b));
    w = times2(complex_imag(a), becomplex_real(b));
    d = plus2(d, w);
#undef a
#undef b
#undef c
#undef d
    a = stack[-1];
    b = stack[0];
    popv(4);
    return make_becomplex(a, b, e);
#endif
}

Sex
times2eZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return times2ez(a, b);
#else
    double          dr,
                    di;
    double          ea,
                    eb;

    dr = fabs(floatit(becomplex_real(a)));
    di = fabs(floatit(becomplex_imag(a)));
    if (dr < di)
	dr = di;
    ea = efloat_error(a);
    eb = becomplex_error(b);
    ea = ea * dr;
    eb = eb * fabs(efloat_value(a));
    if (ea < eb)
	ea = eb;

    stackcheck();
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define c stack[0]
    c = bfloatitf(efloat_value(a));
    a = times2(c, becomplex_real(b));
    b = times2(c, becomplex_imag(b));
#undef a
#undef b
#undef c
    popv(1);
    pop2(b, a);
    return make_becomplex(a, b, ea);
#endif
}

Sex
times2vZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = efloatitv(a);
    pop(b);
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return times2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitv(a);
    pop(b);
    return times2EZ(a, b);
#endif
}

/*
 * ??????? kokomade 
 */

Sex
times2zZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return times2zz(a, b);
#else
    double          ea,
                    eb;

    ea = fabs(ecomplex_real(a));
    eb = fabs(ecomplex_imag(a));
    if (ea > eb)
	eb = ea;
    eb = eb * becomplex_error(b);
    ea = ecomplex_error(a) * absZ(b);

    if (eb > ea)
	ea = eb;
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define f stack[0]
    f = bfloatitf(ecomplex_real(a));
    c = times2dd(f, becomplex_real(b));
    d = times2dd(f, becomplex_imag(b));
    f = bfloatitf(ecomplex_imag(a));
    a = times2dd(f, becomplex_imag(b));
    c = diff2dd(c, a);
    b = times2dd(f, becomplex_real(b));
    d = plus2dd(d, b);
#undef a
#undef b
#undef c
#undef d
#undef f
    a = stack[-2];
    b = stack[-1];
    popv(5);
    return make_becomplex(a, b, ea);
#endif
}



Sex
times2EZ(Sex a, Sex b)
{
    Sex             c;
    double          ea,
                    eb;

    ea = befloat_error(a);
    eb = becomplex_error(b);
    ea = ea * absZ(b);
    eb = eb * fabs(floatit(befloat_value(a)));
    if (eb > ea)
	ea = eb;

    a = befloat_value(a);
    stackcheck();
    push2(a, b);
    c = times2dd(a, becomplex_real(b));
    b = stack[0];
    a = stack[-1];
    stack[0] = c;
    b = times2dd(a, becomplex_imag(b));
    c = stack[0];
    popv(2);
    return make_becomplex(c, b, ea);
}

Sex
times2VZ(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    return times2EZ(a, b);
}

Sex
times2ZZ(Sex a, Sex b)
{
    double          ea,
                    eb;

    ea = becomplex_error(a);
    eb = becomplex_error(b);
    ea = ea * absZ(b);
    eb = eb * absZ(a);
    if (eb > ea)
	ea = eb;
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define d stack[-1]
#define f stack[ 0]
    c = times2dd(becomplex_real(a), becomplex_real(b));
    f = times2dd(becomplex_imag(a), becomplex_imag(b));
    c = diff2dd(c, f);
    d = times2dd(becomplex_real(a), becomplex_imag(b));
    f = times2dd(becomplex_imag(a), becomplex_real(b));
    d = plus2dd(d, f);
#undef a
#undef b
#undef c
#undef d
#undef f
    a = stack[-2];
    b = stack[-1];
    popv(5);
    return make_becomplex(a, b, ea);
}

Sex
times2ZV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return times2EZ(b, a);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

double
absZ(Sex a)
{
    double          f,
                    g;
    f = fabs(floatitd(becomplex_real(a)));
    g = fabs(floatitd(becomplex_imag(a)));
    /*
     * return fmax(f, g);
     */
    return f > g ? f : g;
}

Sex
quot2fZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return quot2fz(a, b);
#else
    double          e;
    e = becomplex_error(b) / absZ(b);
    stackcheck();

    push4(a, b, nil, nil);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define zr stack[-1]
#define zi stack[ 0]
    a = bfloatit(a);
    e = e * floatitd(a);
    zr = times2dd(becomplex_real(b), becomplex_real(b));
    zi = times2dd(becomplex_imag(b), becomplex_imag(b));
    c = plus2dd(zr, zi);

    zr = quot2dd(becomplex_real(b), c);
    zi = quot2dd(becomplex_imag(b), c);

    zr = times2dd(a, zr);
    zi = negated(times2dd(a, zi));
    a = new_becomplex();
    becomplex_real(a) = zr;
    becomplex_imag(a) = zi;
    becomplex_error(a) = e * absZ(a);
#undef a
#undef b
#undef c
#undef zr
#undef zi
    a = stack[-4];
    popv(5);
    return a;
#endif
}


Sex
quot2xZ(Sex a, Sex b)
{
    double          e;
    e = becomplex_error(b) / absZ(b);
    stackcheck();

    push4(a, b, nil, nil);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define zr stack[-1]
#define zi stack[ 0]
    a = bfloatit(a);
    e = e * floatitd(a);
    zr = times2dd(becomplex_real(b), becomplex_real(b));
    zi = times2dd(becomplex_imag(b), becomplex_imag(b));
    c = plus2dd(zr, zi);

    zr = quot2dd(becomplex_real(b), c);
    zi = quot2dd(becomplex_imag(b), c);

    zr = times2dd(a, zr);
    zi = negated(times2dd(a, zi));
    a = new_becomplex();
    becomplex_real(a) = zr;
    becomplex_imag(a) = zi;
    becomplex_error(a) = e * absZ(a);
#undef a
#undef b
#undef c
#undef zr
#undef zi
    a = stack[-4];
    popv(5);
    return a;
}


Sex
quot2Zf(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return quot2zf(a, b);
#else
    Sex             c;
    double          e;
    e = becomplex_error(a);
    e = fabs(e / floatit(b));
    stackcheck();
    push2(a, b);
    c = quot2(becomplex_real(a), b);
    pop2(b, a);
    push(c);
    a = quot2(becomplex_imag(a), b);
    pop(c);
    return make_becomplex(c, a, e);
#endif
}

Sex
quot2Zx(Sex a, Sex b)
{
    Sex             c;
    double          e;
    e = becomplex_error(a);
    e = fabs(e / floatit(b));
    stackcheck();
    push2(a, b);
    c = quot2(becomplex_real(a), b);
    pop2(b, a);
    push(c);
    a = quot2(becomplex_imag(a), b);
    pop(c);
    return make_becomplex(c, a, e);
}

Sex
quot2cZ(Sex a, Sex b)
{
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return quot2zZ(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return quot2ZZ(a, b);
    }
}

Sex
quot2eZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return quot2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatite(a);
    pop(b);
    return quot2EZ(a, b);
#endif
}

Sex
quot2vZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    push(b);
    a = efloatitv(a);
    pop(b);
    return quot2ez(a, b);
#else
    stackcheck();
    push(b);
    a = befloatitv(a);
    pop(b);
    return quot2EZ(a, b);
#endif
}

Sex
quot2zZ(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = ecomplexitZ(b);
    pop(a);
    return quot2zz(a, b);
#else
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return quot2ZZ(a, b);
#endif
}

Sex
quot2EZ(Sex a, Sex b)
{
    double          f,
                    g;
    double          ea,
                    eb;
    ea = befloat_error(a);
    eb = becomplex_error(b);

    f = fabs(floatitd(befloat_value(a)));
    g = 1.0 / absZ(b);
    eb = eb * g * f;
    if (ea < eb)
	ea = eb;
    ea = ea * g;

    stackcheck();

    push4(a, b, nil, nil);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define zr stack[-1]
#define zi stack[ 0]
    zr = times2dd(becomplex_real(b), becomplex_real(b));
    zi = times2dd(becomplex_imag(b), becomplex_imag(b));
    c = plus2dd(zr, zi);

    zr = quot2dd(becomplex_real(b), c);
    zi = quot2dd(becomplex_imag(b), c);

    zr = times2dd(befloat_value(a), zr);
    zi = negated(times2dd(befloat_value(a), zi));
    a = new_becomplex();
    becomplex_real(a) = zr;
    becomplex_imag(a) = zi;
    becomplex_error(a) = ea;
#undef a
#undef b
#undef c
#undef zr
#undef zi
    a = stack[-4];
    popv(5);
    return a;
}


Sex
quot2VZ(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = befloatitV(a);
    pop(b);
    return quot2EZ(a, b);
}



Sex
quot2ZZ(Sex a, Sex b)
{
    double          f,
                    g;
    double          ea,
                    eb;
    ea = becomplex_error(a);
    eb = becomplex_error(b);
    f = absZ(a);
    g = 1.0 / absZ(b);
    eb = eb * g * f;
    if (ea < eb)
	ea = eb;
    ea = ea * g;
    stackcheck();

    push4(a, b, nil, nil);
    push2(nil, nil);
#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define zr stack[-1]
#define zi stack[ 0]
    c = times2dd(becomplex_real(b), becomplex_real(b));
    d = times2dd(becomplex_imag(b), becomplex_imag(b));
    c = plus2dd(c, d);
    zr = quot2dd(becomplex_real(b), c);
    zi = quot2dd(becomplex_imag(b), c);
    c = times2dd(becomplex_real(a), zr);
    d = times2dd(becomplex_imag(a), zi);
    c = plus2dd(c, d);
    b = times2dd(becomplex_real(a), zi);
    d = times2dd(becomplex_imag(a), zr);
    b = diff2dd(d, b);
    a = new_becomplex();
    becomplex_real(a) = c;
    becomplex_imag(a) = b;
    becomplex_error(a) = ea;
#undef a
#undef b
#undef c
#undef d
#undef zr
#undef zi
    a = stack[-5];
    popv(6);
    return a;
}


Sex
quot2Zc(Sex a, Sex b)
{
    stackcheck();
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	pop(a);
	return quot2Zz(a, b);
    } else {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return quot2ZZ(a, b);
    }
}


Sex
quot2Ze(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return quot2ze(a, b);
#else
    stackcheck();
    push(a);
    b = befloatite(b);
    pop(a);
    return quot2ZE(a, b);
#endif
}


Sex
quot2Zz(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return quot2zz(a, b);
#else
    stackcheck();
    push(a);
    b = becomplexitz(b);
    pop(a);
    return quot2ZZ(a, b);
#endif
}


Sex
quot2Zv(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitv(b);
    pop(a);
    push(b);
    a = ecomplexitZ(a);
    pop(b);
    return quot2ze(a, b);
#else
    stackcheck();
    push(a);
    b = befloatitv(b);
    pop(a);
    return quot2ZE(a, b);
#endif
}


Sex
quot2ZE(Sex a, Sex b)
{
    Sex             c;
    double          f,
                    g;
    double          ea,
                    eb;

    ea = becomplex_error(a);
    eb = befloat_error(b);
    f = absZ(a);
    g = 1.0 / floatitd(befloat_value(b));
    g = fabs(g);

    eb = eb * f * g;
    if (eb < ea)
	ea = eb;

    stackcheck();
    push2(a, b);

    c = quot2dd(becomplex_real(a), befloat_value(b));
    pop2(b, a);
    push(c);
    b = quot2dd(becomplex_imag(a), befloat_value(b));
    pop(c);
    return make_becomplex(c, b, ea);
}

Sex
quot2ZV(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    return quot2ZE(a, b);
}

Sex
quot2zV(Sex a, Sex b)
{
#if TOWARD_LOWER_PRECISION
    stackcheck();
    push(a);
    b = efloatitV(b);
    pop(a);
    return quot2ze(a, b);
#else
    stackcheck();
    push(a);
    b = befloatitV(b);
    pop(a);
    push(b);
    a = becomplexitz(a);
    pop(b);
    return quot2ZE(a, b);
#endif
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 *    compare routine
 */



bool
lessp2xZ(Sex a, Sex b)
{
    return (zeropd(becomplex_imag(b)) ? lessp2(a, becomplex_real(b))
	    : bool_er(a, b));
}

bool
lessp2Zx(Sex a, Sex b)
{
    return (zeropd(becomplex_imag(a)) ? lessp2(becomplex_real(a), b)
	    : bool_er(a, b));
}

bool
lessp2ZZ(Sex a, Sex b)
{
    if (bfloat_log2(becomplex_imag(a)) > becomplex_error(a))
	return NO;
    if (bfloat_log2(becomplex_imag(b)) > becomplex_error(b))
	return NO;
    if (lessp2dd(becomplex_real(a), becomplex_real(b)))
	return YES;
    return NO;
}

bool
lessp2Zc(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(b)) {
	push(a);
	b = ecomplexitc(b);
	a = stack[0];
	stack[0] = b;
	a = ecomplexitZ(a);
	pop(b);
	return lessp2zz(a, b);
    } else {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return lessp2ZZ(a, b);
    }
#else
    Sex             br;
    long            na;

    na = number_log2(becomplex_imag(a));
    if (na > becomplex_error(a))
	return NO;
    if (!zerop(complex_imag(b)))
	return NO;
    stackcheck();
    push(a);
    br = bfloatit(complex_real(b));
    pop(a);
    br = diff2dd(becomplex_real(a), br);
    if (minusp(br))
	return YES;
    return NO;
#endif
}

bool
lessp2Zz(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = becomplexitz(b);
    pop(a);
    return lessp2ZZ(a, b);
}

bool
lessp2cZ(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	b = stack[0];
	stack[0] = a;
	b = ecomplexitZ(b);
	pop(a);
	return lessp2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return lessp2ZZ(a, b);
    }
#else
    Sex             ar;
    long            nb;

    nb = number_log2(becomplex_imag(b));
    if (nb > ecomplex_error(b))
	return NO;
    if (!zerop(complex_imag(a)))
	return NO;
    stackcheck();
    push(b);
    ar = bfloatit(complex_real(a));
    pop(b);
    ar = diff2dd(ar, becomplex_real(b));
    if (minusp(ar))
	return YES;
    return NO;
#endif
}

bool
lessp2zZ(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = becomplexitz(a);
    pop(b);
    return lessp2ZZ(a, b);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2EZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2dd(befloat_value(a), becomplex_real(b));
}

bool
numeq2ZZ(Sex a, Sex b)
{
    if (!numeq2dd(becomplex_real(a), becomplex_real(b)))
	return NO;
    return numeq2dd(becomplex_imag(a), becomplex_imag(b));
}


bool
numeq2bZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2bd(a, becomplex_real(b));
}

bool
numeq2cZ(Sex a, Sex b)
{
#if 1
    stackcheck();
    if (is_shortfloat(a)) {
	push(b);
	a = ecomplexitc(a);
	b = stack[0];
	stack[0] = a;
	b = ecomplexitZ(b);
	pop(a);
	return numeq2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return numeq2ZZ(a, b);
    }
#else
    if (!numeq2(becomplex_real(b), complex_real(a)))
	return NO;
    return numeq2(becomplex_imag(b), complex_imag(a));
#endif
}

bool
numeq2dZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2dd(a, becomplex_real(b));
}

bool
numeq2eZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq_f_d(efloat_value(a), becomplex_real(b));
}

bool
numeq_f_d(double x, Sex a)
{
    Sex             c;
    push(a);
    c = bfloatitf(x);
    pop(a);
    return numeq2dd(c, a);
}
bool
lessp_f_d(double x, Sex a)
{
    Sex             c;
    push(a);
    c = bfloatitf(x);
    pop(a);
    return lessp2dd(c, a);
}


bool
numeq2fZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2fd(a, becomplex_real(b));
}

bool
numeq2iZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2id(a, becomplex_real(b));
}

bool
numeq2rZ(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(b)))
	return NO;
    return numeq2rd(a, becomplex_real(b));
}

bool
numeq2vZ(Sex a, Sex b)
{
    double          v;

    if (!zeropd(becomplex_imag(b)))
	return NO;
    v = interval_upper(a) + interval_lower(a);
    return numeq_f_d(v * 0.5, becomplex_real(b));
}

bool
numeq2zZ(Sex a, Sex b)
{
    if (!numeq_f_d(ecomplex_real(a), becomplex_real(b)))
	return NO;
    return numeq_f_d(ecomplex_imag(a), becomplex_imag(b));
}

bool
numeq2zE(Sex a, Sex b)
{
    if (ecomplex_imag(a) != 0.0)
	return NO;
    return numeq_f_d(ecomplex_real(a), befloat_value(b));
}

bool
numeq2zV(Sex a, Sex b)
{
    double          x;
    if (ecomplex_imag(a) != 0)
	return NO;
    x = ecomplex_real(b);
    b = plus2dd(binterval_lower(b), binterval_upper(b));
    return numeq_f_d(x + x, b);
}

bool
numeq2ZV(Sex a, Sex b)
{
    if (!zeropd(becomplex_imag(a)))
	return NO;
    stackcheck();
    push(a);
    b = plus2dd(binterval_lower(b), binterval_upper(b));
    b = leftshiftd(b, -1);
    pop(a);
    return numeq2dd(becomplex_real(a), b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
appeq2EZ(Sex a, Sex b)
{
    return zerop(diff2EZ(a, b));
}

bool
appeq2ZZ(Sex a, Sex b)
{
    return zerop(diff2ZZ(a, b));
}

bool
appeq2Zc(Sex a, Sex b)
{
    return zerop(diff2Zc(a, b));
}

bool
appeq2bZ(Sex a, Sex b)
{
    return zerop(diff2bZ(a, b));
}

bool
appeq2cZ(Sex a, Sex b)
{
    return zerop(diff2cZ(a, b));
}

bool
appeq2dZ(Sex a, Sex b)
{
    return zerop(diff2dZ(a, b));
}

bool
appeq2eZ(Sex a, Sex b)
{
    return zerop(diff2eZ(a, b));
}

bool
appeq2fZ(Sex a, Sex b)
{
    return zerop(diff2fZ(a, b));
}

bool
appeq2iZ(Sex a, Sex b)
{
    return zerop(diff2iZ(a, b));
}

bool
appeq2rZ(Sex a, Sex b)
{
    return zerop(diff2rZ(a, b));
}

bool
appeq2vZ(Sex a, Sex b)
{
    return zerop(diff2vZ(a, b));
}

bool
appeq2zZ(Sex a, Sex b)
{
    return zerop(diff2zZ(a, b));
}

bool
appeq2zE(Sex a, Sex b)
{
    return zerop(diff2zE(a, b));
}

bool
appeq2zV(Sex a, Sex b)
{
    return zerop(diff2zV(a, b));
}

bool
appeq2ZV(Sex a, Sex b)
{
    return zerop(diff2ZV(a, b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
