/*
 * bignum.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/bignum.c,v 2.18 2016/04/03 23:43:37 kako Exp $
 */

static char const rcsid[] =
    "$Id: bignum.c,v 2.18 2016/04/03 23:43:37 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"

#ifdef USE_GMP
#ifdef VisualC
#include "gmp.h"
#else
#include <gmp.h>
#endif
#endif


#define SAVE_SPACE

#ifndef KARATSUBA
#define KARATSUBA	16
#endif


static Sex      plus2_bn(Sex a, long n, long t);
static Sex      diff2_bn(Sex a, long n, long t);
Sex             fixit(Sex a);
Sex             mkbigat1s(long s, long n);
Sex             mkbigats(long s, long a1, unsigned long a0);
long            bfadd(unsigned long a[], long la, long b,
		      unsigned long c[]);
long            bfsub(unsigned long a[], long la, long b,
		      unsigned long c[]);
void            print_phex(Sex u, long k, putchar_fn * fn);


long            bbmul_karatsuba(unsigned long a[], long la,
				unsigned long b[], long lb,
				unsigned long c[], long lc,
				unsigned long d[]);
long            bbsqr_karatsuba(unsigned long a[], long la,
				unsigned long c[], long lc,
				unsigned long d[]);
Sex             times2bb_karatsuba(Sex a, Sex b);
#ifdef HAVE_GMP
Sex             times2bb_gmp(Sex a, Sex b);
#endif
Sex             times2bb_classic(Sex a, Sex b);

Sex
fixitl(Sex a)
{
    Sex             c;
    long            i;
    long            len = numlen(a);

    if (len == bnum_len(1)) {
	i = logical_value(a)[0];
	if ((i & fix_mask) == 0)
	    return mkfxat(i);
    }
    stackcheck();
    push(a);
    c = new_bignum(len);
    pop(a);
    len = bnum_n(len);
    for (i = 0; i < len; i++)
	bignum_value(c)[i] = logical_value(a)[i];
    return c;
}



Sex
fixitc(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = Lfix(complex_real(a));
    pop(a);
    push(c);
    a = Lfix(complex_imag(a));
    pop(c);
    return Lmkcomplex(c, a);
}

Sex
fixitz(Sex a)
{
    double          zi,
                    zr;
    Sex             c;

    zr = ecomplex_real(a);
    zi = ecomplex_imag(a);
    a = fixitf(zr);
    stackcheck();
    push(a);
    c = fixitf(zi);
    pop(a);
    return Lmkcomplex(a, c);
}

Sex
fixit(Sex a)
{
    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_EFLOAT:
	    return fixitf(efloat_value(a));
	case N_INTERVAL:
	    return fixitf((interval_lower(a) + interval_upper(a)) * 0.5);
	case N_RATIONAL:
	    return fixitf(floatit(rational_num(a)) /
			  floatit(rational_den(a)));
	case N_FLOAT:
	    return fixitf(float_val(a));
	case N_BFLOAT:
	    return fixitd(a);
	case N_BEFLOAT:
	    return fixitd(befloat_value(a));
	case N_BINTERVAL:
	    return fixitV(a);
	}
    }
    return aerror1("fixit", a);
}






Sex
Lfix(Sex a)
{
    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_EFLOAT:
	    return fixitf(efloat_value(a));
	case N_INTERVAL:
	    return fixitf((interval_lower(a) + interval_upper(a)) * 0.5);
	case N_RATIONAL:
	    return fixitf(floatit(rational_num(a)) /
			  floatit(rational_den(a)));
	case N_FLOAT:
	    return fixitf(float_val(a));
	case N_COMPLEX:
	    return fixitc(a);
	case N_BFLOAT:
	    return fixitd(a);
	case N_ECOMPLEX:
	    return fixitz(a);
	case N_BEFLOAT:
	    return fixitd(befloat_value(a));
	case N_BINTERVAL:
	    return fixitV(a);
	case N_BECOMPLEX:
	    return fixitZ(a);
	}
    }
    return aerror1("fix", a);
}






Sex
mkbigat1(long n)
{				/* 
				 * make a one word size bignum 
				 */
    Sex             w;

    if (n >= 0) {
	w = new_bignum(bnum_len(1));
	bignum_value(w)[0] = n;
    } else {
	w = new_bigneg(bnum_len(1));
	bignum_value(w)[0] = -n;
    }
    return w;
}

Sex
mkbigat1s(long s, long n)
{				/* 
				 * make a one word size bignum 
				 */
    Sex             w;

    if (s >= 0) {
	w = new_bignum(bnum_len(1));
	bignum_value(w)[0] = n;
    } else {
	w = new_bigneg(bnum_len(1));
	bignum_value(w)[0] = n;
    }
    return w;
}

Sex
mkbigat(long a1, unsigned long a0)
{				/* 
				 * make a 2-word bignum 
				 */
    Sex             w;
    if (a1 >= 0) {
	w = new_bignum(bnum_len(2));
	bignum_value(w)[0] = a0;
	bignum_value(w)[1] = a1;
    } else {
	w = new_bigneg(bnum_len(2));
	bignum_value(w)[0] = a0;
	bignum_value(w)[1] = -a1;
    }
    return w;
}


Sex
mkbigats(long s, long a1, unsigned long a0)
{				/* 
				 * make a 2-word bignum 
				 */
    Sex             w;
    if (s >= 0) {
	w = new_bignum(bnum_len(2));
	bignum_value(w)[0] = a0;
	bignum_value(w)[1] = a1;
    } else {
	w = new_bigneg(bnum_len(2));
	bignum_value(w)[0] = a0;
	bignum_value(w)[1] = a1;
    }
    return w;
}




long
bfadd(a, la, b, c)		/* 
				 * * * * assume b < 0x80000000     
				 */
     unsigned long   a[],
                     c[];
     long            la;
     long            b;
{
    long            i;
    unsigned long   x = b;

    for (i = 0; i < la && x != 0; i++) {
	c[i] = a[i] + x;
#if 0
	x = (a[i] & ~c[i]) >> 63;
#else
	x = c[i] < x;
#endif
    }
    for (; i < la; i++)
	c[i] = a[i];

    return (long) x;
}

long
bfsub(a, la, b, c)
     unsigned long   a[],
                     c[];
     long            la;
     long            b;
{
    long            i;
    unsigned long   x = b;

    for (i = 0; i < la && x != 0; i++) {
	c[i] = a[i] - x;
	x = a[i] < x;
    }
    for (; i < la; i++)
	c[i] = a[i];
    return (long) x;
}

long
bbadd(a, la, b, lb, c)		/* 
				 * * * * la >= lb              
				 */
     unsigned long   a[],
                     b[],
                     c[];
     long            la,
                     lb;
{
    long            i;
    unsigned long   carry;
    unsigned long   x;

    carry = 0;
    for (i = 0; i < lb; i++) {

	x = a[i] + carry;
#if 0
	if (~a[i] < carry)
	    carry = 1;
	else
	    carry = 0;
#else
	carry = x < carry;
#endif
	c[i] = x + b[i];
#if 0
	if (~x < b[i])
	    carry = 1;
#else
	carry += c[i] < x;
#endif
    }
    for (; carry != 0 && i < la; i++) {
	c[i] = a[i] + 1;
	if (a[i] != -1)
	    carry = 0;
    }
    for (; i < la; i++)
	c[i] = a[i];
    return (long) carry;
}

long
bbsub(a, la, b, lb, c)		/* 
				 * * * * la >= lb              
				 */
     unsigned long   a[],
                     b[],
                     c[];
     long            la,
                     lb;
{
    long            i;
    unsigned long   x,
                    y;
    unsigned long   carry;

    carry = 0;
    for (i = 0; i < lb; i++) {
	x = a[i] - b[i];
	y = a[i] >= b[i] ? 0 : 1;
	c[i] = x - carry;
	if (x < carry)
	    y = 1;
	carry = y;
    }
    for (; carry != 0 && i < la; i++) {
	c[i] = a[i] - 1;
	carry = (a[i] >= 1) ? 0 : 1;
    }
    if (carry) {
	for (i = 0; i < la; i++)
	    c[i] = ~c[i];
	y = 1;
	for (i = 0; i < la && y != 0; i++) {
	    x = c[i] + 1;
	    y = c[i] == -1 ? 1 : 0;
	    c[i] = x;
	}
	for (i = la - 1; i >= 0 && c[i] == 0; i--);
	return -(i + 1);
    } else if (i < la) {
	for (; i < la; i++)
	    c[i] = a[i];
	return la;
    } else {			/* 
				 * here i == la 
				 */
	for (i--; i >= 0 && c[i] == 0; i--);
	return i + 1;
    }
}


#ifndef USE_ASMCODE

#if !defined(_LP64) && defined(USE_LONG_LONG)

unsigned long
umul(l, a, b)
     unsigned long
                    *l;
     unsigned long   a,
                     b;
{
    unsigned long long r;
    r = (unsigned long long) a *(unsigned long long) b;
    *l = (unsigned long) r;
    return (unsigned long) (r >> 32);
}

unsigned long
umuladd(l, a, b, c)
     unsigned long
                    *l;
     unsigned long   a,
                     b,
                     c;
{
    unsigned long long r;
    r = (unsigned long long) a *(unsigned long long) b
	+ (unsigned long long) c;
    *l = (unsigned long) r;
    return (unsigned long) (r >> 32);
}

unsigned long
udiv(r, a, b, c)
     unsigned long
                    *r;
     unsigned long   a,
                     b,
                     c;
{
    unsigned long long p;
    p = ((unsigned long long) a << 32) | (unsigned long long) b;
    *r = p % c;
    return p / c;
}

unsigned long
umulrem(a, b, c)
     unsigned long
                     a;
     unsigned long   b,
                     c;
{
    unsigned long long r;
    r = (unsigned long long) a *(unsigned long long) b;
    r = r % (unsigned long long) c;
    return (unsigned long) r;
}

#else

unsigned long
umul(l, a, b)
     unsigned long
                    *l;
     unsigned long   a,
                     b;
{
    unsigned long   ah = a >> N_SHIFT,
	bh = b >> N_SHIFT;
    unsigned long   w0,
                    w1,
                    w2,
                    w3;
    a = a & L_HALF;
    b = b & L_HALF;
    w0 = a * b;			/* 
				 * low * low 
				 */
    w1 = w0 >> N_SHIFT;
    w0 = (w0 & L_HALF);
    w2 = ah * bh;

    w3 = a * bh;
    w2 += w3 >> N_SHIFT;
    w1 += (w3 & L_HALF);

    w3 = ah * b;
    w2 += (w3 >> N_SHIFT);
    w1 += (w3 & L_HALF);

    w1 += (w0 >> N_SHIFT);
    w2 += (w1 >> N_SHIFT);
    w0 = ((w1 & L_HALF) << N_SHIFT) + (w0 & L_HALF);
    *l = w0;
    return w2;
}

unsigned long
umuladd(l, a, b, c)
     unsigned long
                    *l;
     unsigned long   a,
                     b,
                     c;
{
    unsigned long   ah = a >> N_SHIFT,
	bh = b >> N_SHIFT,
	ch = c >> N_SHIFT;
    unsigned long   w0,
                    w1,
                    w2,
                    w3;
    a = a & L_HALF;
    b = b & L_HALF;
    c = c & L_HALF;
    w0 = a * b;			/* 
				 * low * low 
				 */
    w1 = w0 >> N_SHIFT;
    w0 = (w0 & L_HALF) + c;
    w1 += ch;
    w2 = ah * bh;

    w3 = a * bh;
    w2 += w3 >> N_SHIFT;
    w1 += (w3 & L_HALF);

    w3 = ah * b;
    w2 += (w3 >> N_SHIFT);
    w1 += (w3 & L_HALF);

    w1 += (w0 >> N_SHIFT);
    w2 += (w1 >> N_SHIFT);
    w0 = ((w1 & L_HALF) << N_SHIFT) + (w0 & L_HALF);
    *l = w0;
    return w2;
}


unsigned long
udiv(r, a, b, c)
     unsigned long
                    *r;
     unsigned long   a,
                     b,
                     c;
{
    unsigned long   i;
    unsigned long   p;

    if (a == 0) {
	p = b / c;
	*r = b - p * c;
	return p;
    } else if ((c & U_HALF) == 0) {
	a = (a << N_SHIFT) | (b >> N_SHIFT);
	p = a / c;
	a = a - p * c;
	a = (a << N_SHIFT) | (b & L_HALF);
	b = a / c;
	*r = a - b * c;
	return (p << N_SHIFT) + b;
    }
    p = 0;
    for (i = 0; i < LPSIZE; i++) {
	p = p << 1;
	if ((long) a < 0) {
	    a = a << 1;
	    if ((long) b < 0)
		a = a + 1;
	    b = b << 1;
	    a = a - c;
	    p = p + 1;
	} else {
	    a = a << 1;
	    if ((long) b < 0)
		a = a + 1;
	    b = b << 1;
	    if (a >= c) {
		a = a - c;
		p = p + 1;
	    }
	}
    }
    *r = a;
    return p;
}

unsigned long
umulrem(a, b, c)
     unsigned long
                     a;
     unsigned long   b,
                     c;
{
    unsigned long   h,
                    l;
    unsigned long   r;

    h = umul(&l, a, b);
    (void) udiv(&r, h, l, c);
    return r;
}
#endif
#endif



long
bfmul(a, la, b, c)		/* 
				 * * * * b > 0              
				 */
     unsigned long   a[],
                     c[];
     long            la;
     unsigned long   b;
{
    long            i;

    unsigned long   carry;

    carry = 0;
    for (i = 0; i < la; i++) {
	carry = umuladd(&c[i], a[i], b, carry);
    }
    return (long) carry;
}




long
bbmul_karatsuba(a, la, b, lb, c, lc, d)
     unsigned long   a[],
                     b[],
                     c[],
                     d[];
     long            la,
                     lb,
                     lc;
{
    long            n;
    long            i;
    long            la1,
                    lb1;
    unsigned long   carry1,
                    carry2;
    unsigned long   carry;
    unsigned long   x,
                    y;
    if (la < lb) {
	unsigned long  *e;
	e = a;
	la1 = la;
	a = b;
	la = lb;
	b = e;
	lb = la1;
    }
    if (lb <= KARATSUBA) {
	n = bbmul(a, la, b, lb, c);
	for (i = la + lb; i < lc; i++)
	    c[i] = 0;
	return n;
    }
    n = lc / 4;
    la1 = la - n;
    lb1 = lb - n;
    if (lb1 <= 0) {
	if (la1 <= 0) {
	    bbmul_karatsuba(a, la, b, lb, c, 2 * n, d);
	    for (i = 0; i < 2 * n; i++)
		c[2 * n + i] = 0;
	    return c[la + lb - 1];
	}

	bbmul_karatsuba(a + n, la1, b, lb, c + n, 2 * n, d);
	for (i = 0; i < n; i++)
	    c[3 * n + i] = 0;
	bbmul_karatsuba(a, n, b, lb, d, 2 * n, c);
	for (i = 0; i < n; i++)
	    c[i] = d[i];

	carry = 0;
	for (; i < 2 * n; i++) {
	    x = d[i] + carry;
	    if (~d[i] < carry)
		carry = 1;
	    else
		carry = 0;
	    y = x + c[i];
	    if (~x < c[i])
		carry = 1;
	    c[i] = y;
	}
	for (; carry != 0; i++) {
	    if (c[i] != -1)
		carry = 0;
	    c[i] = c[i] + 1;
	}
	return c[la + lb - 1];
    }
    /*
     * calculate (a1+a2) and (b1+b2) 
     */

    /*
     * add a[i] + a[n+i] 
     */
    /*
     * bbadd(a  , n, a+n, la1, d);
     */
    carry1 = 0;
    for (i = 0; i < la1; i++) {
	x = a[i] + carry1;
	if (~a[i] < carry1)
	    carry1 = 1;
	else
	    carry1 = 0;
	d[i] = x + a[n + i];
	if (~x < a[n + i])
	    carry1 = 1;
    }
    for (; carry1 != 0 && i < n; i++) {
	if (a[i] != -1)
	    carry1 = 0;
	d[i] = a[i] + 1;
    }
    for (; i < n; i++)
	d[i] = a[i];

    /*
     * add b[i] + b[n+i] 
     */
    /*
     * bbadd(b  , n, b+n, lb1, d+n);
     */

    carry2 = 0;
    for (i = 0; i < lb1; i++) {
	x = b[i] + carry2;
	if (~b[i] < carry2)
	    carry2 = 1;
	else
	    carry2 = 0;
	d[n + i] = x + b[n + i];
	if (~x < b[n + i])
	    carry2 = 1;
    }
    for (; carry2 != 0 && i < n; i++) {
	if (b[i] != -1)
	    carry2 = 0;
	d[n + i] = b[i] + 1;
    }
    for (; i < n; i++)
	d[n + i] = b[i];
    bbmul_karatsuba(d, n, d + n, n, c + n, 2 * n, c);
    c[3 * n] = carry1 & carry2;
    if (carry1) {
	carry1 = 0;
	for (i = 0; i < n; i++) {
	    x = d[n + i] + carry1;
	    if (~d[n + i] < carry1)
		carry1 = 1;
	    else
		carry1 = 0;
	    y = x + c[2 * n + i];
	    if (~x < c[2 * n + i])
		carry1 = 1;
	    c[2 * n + i] = y;
	}
    }
    if (carry2) {
	carry2 = 0;
	for (i = 0; i < n; i++) {
	    x = d[i] + carry2;
	    if (~d[i] < carry2)
		carry2 = 1;
	    else
		carry2 = 0;
	    y = x + c[2 * n + i];
	    if (~x < c[2 * n + i])
		carry2 = 1;
	    c[2 * n + i] = y;
	}
    }
    c[3 * n] += carry1 + carry2;
    for (i = 1; i < n; i++)
	c[3 * n + i] = 0;

    /*
     * calculate a1 * b1 
     */
    bbmul_karatsuba(a + n, la1, b + n, lb1, d, 2 * n, c);
    /*
     * bbadd(c+2*n, 2*n, d, 2*n, c+2*n);
     */
    carry = 0;
    for (i = 0; i < 2 * n; i++) {
	x = d[i] + carry;
	if (~d[i] < carry)
	    carry = 1;
	else
	    carry = 0;
	y = x + c[2 * n + i];
	if (~x < c[2 * n + i])
	    carry = 1;
	c[2 * n + i] = y;
    }
    /*
     * bbsub(c+n  , 2*n, d, 2*n, c+n);
     */
    carry = 0;
    for (i = 0; i < 2 * n; i++) {
	x = c[n + i] - d[i];
	y = c[n + i] < d[i];
	c[n + i] = x - carry;
	if (x < carry)
	    y = 1;
	carry = y;
    }

    for (; carry != 0 && i < 3 * n; i++) {
	if (c[n + i] != 0)
	    carry = 0;
	c[n + i] = c[n + i] - 1;
    }
    /*
     * calculate a2 * b2 
     */
    bbmul_karatsuba(a, n, b, n, d, 2 * n, c);
    for (i = 0; i < n; i++)
	c[i] = d[i];
    /*
     * bbadd(c  , 2*n, d, 2*n, c);
     */
    carry = 0;
    for (; i < 2 * n; i++) {
	x = d[i] + carry;
	carry = x < carry;
	y = x + c[i];
	carry += y < x;
	c[i] = y;
    }
    for (; carry != 0 && i < 4 * n; i++) {
	carry = c[i] == -1;
	c[i] = c[i] + 1;
    }

    /*
     * bbsub(c+n, 2*n, d, 2*n, c+n);
     */
    carry = 0;
    for (i = 0; i < 2 * n; i++) {
	x = c[n + i] - d[i];
	y = c[n + i] >= d[i] ? 0 : 1;
	c[n + i] = x - carry;
	if (x < carry)
	    y = 1;
	carry = y;
    }

    for (; carry != 0 && i < 3 * n; i++) {
	x = c[n + i] - 1;
	if (c[n + i] != 0)
	    carry = 0;
	c[n + i] = x;
    }
    return c[la + lb - 1];
}

long
bbsqr_karatsuba(a, la, c, lc, d)
     unsigned long   a[],
                     c[],
                     d[];
     long            la,
                     lc;
{
    long            n;
    long            i;
    long            la1;
    unsigned long   carry;
    unsigned long   x,
                    y;

    if (la <= KARATSUBA) {
	n = bbsqr(a, la, c);
	for (i = la + la; i < lc; i++)
	    c[i] = 0;
	return n;
    }
    n = lc / 4;
    la1 = la - n;

    if (la1 <= 0) {
	bbsqr_karatsuba(a, la, c, 2 * n, d);
	for (i = la + la; i < lc; i++)
	    c[i] = 0;
	return c[la + la - 1];
    }

    /*
     * calculate (a1*a2) 
     */


    bbmul_karatsuba(a, n, a + n, la1, c + n, 2 * n, c);
    /*
     * shift left 1 bit 
     */
    x = 0;
    for (i = n; i < 3 * n; i++) {
	y = c[i];
	y = y << 1;
	carry = y < x;
	c[i] = y | x;
	x = carry;
    }
    c[3 * n] = x;
    for (i = 1; i < n; i++)
	c[3 * n + i] = 0;

    /*
     * calculate a1 * a1 
     */
    bbsqr_karatsuba(a + n, la1, d, 2 * n, c);
    /*
     * bbadd(c+2*n, 2*n, d, 2*n, c+2*n);
     */
    carry = 0;
    for (i = 0; i < 2 * n; i++) {
	x = d[i] + carry;
	carry = x < carry;
	y = x + c[2 * n + i];
	carry += y < x;
	c[2 * n + i] = y;
    }
    /*
     * calculate a2 * a2 
     */
    bbsqr_karatsuba(a, n, d, 2 * n, c);
    for (i = 0; i < n; i++)
	c[i] = d[i];
    /*
     * bbadd(c  , 2*n, d, 2*n, c);
     */
    carry = 0;
    for (; i < 2 * n; i++) {
	x = d[i] + carry;
	carry = x < carry;
	y = x + c[i];
	carry += y < x;
	c[i] = y;
    }
    for (; carry != 0 && i < 4 * n; i++) {
	carry = c[i] == -1;
	c[i] = c[i] + 1;
    }
    return c[la + la - 1];
}



long
bbmul(a, la, b, lb, c)		/* 
				 * * * * here la >= lb      
				 */
     unsigned long   a[],
                     b[],
                     c[];
     long            la,
                     lb;
{
    long            i,
                    j;
    unsigned long   h,
                    l,
                    carry;
    long            lc = la + lb;

    for (i = 0; i < lc; i++)
	c[i] = 0;

    carry = 0;
    for (j = 0; j < lb; j++) {
	carry = 0;
	for (i = 0; i < la; i++) {
	    h = umuladd(&l, a[i], b[j], c[i + j]);
	    if (~l < carry)
		h = h + 1;
	    c[i + j] = l + carry;
	    carry = h;
	}
	c[la + j] = carry;
    }
    return (long) carry;
}

long
bbsqr(a, la, c)			/* 
				 */
     unsigned long   a[],
                     c[];
     long            la;
{
    long            i,
                    j;
    unsigned long   h,
                    l,
                    carry;
    long            lc = la << 1;

    for (i = 0; i < lc; i++)
	c[i] = 0;

    for (j = 0; j < la; j++) {
	carry = 0;
	for (i = 0; i < la; i++) {
	    h = umuladd(&l, a[i], a[j], c[i + j]);
	    if (~l < carry)
		h = h + 1;
	    c[i + j] = l + carry;
	    carry = h;
	}
	c[la + j] = carry;
    }
    return (long) carry;
}

long
bfdiv(a, la, b, c)		/* 
				 * * * * b > 0              
				 */
     unsigned long   a[],
                     c[];
     long            la;
     unsigned long   b;
{
    long            i;
    unsigned long   carry;

    carry = 0;
    for (i = la - 1; i >= 0; i--) {
	c[i] = udiv(&carry, carry, a[i], b);
    }
    return (long) carry;
}

long
bfrem(a, la, b)			/* 
				 * * * * b > 0              
				 */
     unsigned long   a[];
     long            la;
     unsigned long   b;
{
    long            i;
    unsigned long   q,
                    carry;

    carry = 0;
    for (i = la - 1; i >= 0; i--) {
	q = udiv(&carry, carry, a[i], b);
    }
    return (long) carry;
}



#ifdef SAVE_SPACE
#else
long
bbquo(a, la, b, lb, c, d, e)
     unsigned long   a[],
                     b[],
                     c[],
                     d[],
                     e[];
     long            la,
                     lb;

{
    unsigned long  *numr,
                   *divr,
                   *quot;
    long            scale;
    long            i,
                    j,
                    n;
    unsigned long   qh;
    unsigned long   carry;

    if (lb == 1) {
	carry = 0;
	for (i = la - 1; i >= 0; i--) {
	    e[i] = udiv(&carry, carry, a[i], b[0]);
	}
	return e[la - 1] == 0 ? la - 1 : la;
    }
    numr = c;
    divr = d;
    quot = e;

    if (b[lb - 1] & 0x80000000) {
	scale = 0;
	for (i = 0; i < la; i++)
	    c[i] = a[i];
	c[la] = 0;
	for (i = 0; i < lb; i++)
	    d[i] = b[i];
    } else {			/* 
				 * need normalization 
				 */
	i = 0;
	n = b[lb - 1];
	if ((n & 0xffff0000) == 0) {
	    n <<= 16;
	    i += 16;
	}
	if ((n & 0xff000000) == 0) {
	    n <<= 8;
	    i += 8;
	}
	if ((n & 0xf0000000) == 0) {
	    n <<= 4;
	    i += 4;
	}
	while ((long) n > 0) {
	    i++;
	    n <<= 1;
	}
	scale = i;
	for (i = lb - 1; i > 0; i--)
	    d[i] = (b[i] << scale) | (b[i - 1] >> (32 - scale));
	d[0] = b[0] << scale;
	c[la] = a[la - 1] >> (32 - scale);
	for (i = la - 1; i > 0; i--)
	    c[i] = (a[i] << scale) | (a[i - 1] >> (32 - scale));
	c[0] = a[0] << scale;
    }
    for (j = la - lb; j >= 0; j--) {
	unsigned long   xh,
	                xl,
	                yh,
	                yl;
	unsigned long   r;

	if (numr[lb + j] == divr[lb - 1]) {
	    qh = 0xffffffff;
	    if (~numr[lb + j] < numr[lb + j - 1])
		goto ok;
	    r = numr[lb + j] + numr[lb + j - 1];
	} else {
	    qh = udiv(&r, numr[lb + j], numr[lb + j - 1], divr[lb - 1]);
	}
	xh = r;
	xl = numr[lb + j - 2];
	yh = umul(&yl, qh, divr[lb - 2]);
	if (yh > xh || (yh == xh && yl > xl)) {
	    qh = qh - 1;
	    if (yl < xl)
		yh = yh - 1;
	    yl = yl - xl;
	    yh = yh - xh;
	    xh = divr[lb - 1];
	    xl = divr[lb - 2];
	    if (yh > xh || (yh == xh && yl > xl))
		qh = qh - 1;
	}
      ok:
	if (qh == 0) {
	    quot[j] = 0;
	    continue;
	}
	yl = yh = 0;
	carry = yh;
	for (i = 0; i < lb; i++) {
	    xl = numr[i + j];
	    yh = umuladd(&yl, divr[i], qh, carry);
	    if (xl < yl)
		yh = yh + 1;
	    numr[i + j] = xl - yl;
	    carry = yh;
	}
	yl = yh;
	xl = numr[lb + j];
	if (yl > xl) {
	    unsigned long   m;
	    unsigned long   c;

	    qh = qh - 1;
	    c = 0;
	    for (i = 0; i < lb; i++) {
		m = numr[i + j] + c;
		if (~numr[i + j] < c)
		    c = 1;
		else
		    c = 0;
		numr[i + j] = m + divr[i];
		if (~m < divr[i])
		    c = 1;
	    }
	}
	numr[lb + j] = 0;
	quot[j] = qh;
    }
    return quot[la - lb] == 0 ? la - lb : la - lb + 1;
}

long
bbrem(a, la, b, lb, c, d, e)
     unsigned long   a[],
                     b[],
                     c[],
                     d[],
                     e[];
     long            la,
                     lb;

{
    unsigned long  *numr,
                   *divr,
                   *quot;
    long            scale;
    long            i,
                    j,
                    n;
    unsigned long   qh;
    long            carry;

    if (lb == 1) {
	carry = 0;
	for (i = la - 1; i >= 0; i--) {
	    e[i] = udiv(&carry, carry, a[i], b[0]);
	}
	c[0] = carry;
	return carry ? 1 : 0;
    }
    numr = c;
    divr = d;
    quot = e;
    if (b[lb - 1] & 0x80000000) {
	scale = 0;
	for (i = 0; i < la; i++)
	    c[i] = a[i];
	c[la] = 0;
	for (i = 0; i < lb; i++)
	    d[i] = b[i];
    } else {			/* 
				 * need normalization 
				 */
	i = 0;
	n = b[lb - 1];
	if ((n & 0xffff0000) == 0) {
	    n <<= 16;
	    i += 16;
	}
	if ((n & 0xff000000) == 0) {
	    n <<= 8;
	    i += 8;
	}
	if ((n & 0xf0000000) == 0) {
	    n <<= 4;
	    i += 4;
	}
	while ((long) n > 0) {
	    i++;
	    n <<= 1;
	}
	scale = i;
	for (i = lb - 1; i > 0; i--)
	    d[i] = (b[i] << scale) | (b[i - 1] >> (LPSIZE - scale));
	d[0] = b[0] << scale;
	c[la] = a[la - 1] >> (LPSIZE - scale);
	for (i = la - 1; i > 0; i--)
	    c[i] = (a[i] << scale) | (a[i - 1] >> (32 - scale));
	c[0] = a[0] << scale;
    }
    for (j = la - lb; j >= 0; j--) {
	unsigned long   xh,
	                xl,
	                yh,
	                yl;
	unsigned long   r;

	if (numr[lb + j] == divr[lb - 1]) {
	    qh = 0xffffffff;
	    if (~numr[lb + j] < numr[lb + j - 1])
		goto ok;
	    r = numr[lb + j] + numr[lb + j - 1];
	} else {
	    qh = udiv(&r, numr[lb + j], numr[lb + j - 1], divr[lb - 1]);
	}
	xh = r;
	xl = numr[lb + j - 2];
	yh = umul(&yl, qh, divr[lb - 2]);
	if (yh > xh || (yh == xh && yl > xl)) {
	    qh = qh - 1;
	    if (yl < xl)
		yh = yh - 1;
	    yl = yl - xl;
	    yh = yh - xh;
	    xh = divr[lb - 1];
	    xl = divr[lb - 2];
	    if (yh > xh || (yh == xh && yl > xl))
		qh = qh - 1;
	}
      ok:
	if (qh == 0) {
	    quot[j] = 0;
	    continue;
	}
	carry = 0;
	for (i = 0; i < lb; i++) {
	    xl = numr[i + j];
	    yh = umul(&yl, divr[i], qh, carry);
	    if (xl < yl)
		yh = yh + 1;
	    numr[i + j] = xl - yl;
	    carry = yh;
	}
	yl = yh;
	xl = numr[lb + j];
	if (yl > xl) {
	    unsigned long   m;
	    unsigned long   c;

	    qh = qh - 1;
	    c = 0;
	    for (i = 0; i < lb; i++) {
		m = numr[i + j] + c;
		if (~numr[i + j] < c)
		    c = 1;
		else
		    c = 0;
		numr[i + j] = m + divr[i];
		if (~m < divr[i])
		    c = 1;
	    }
	}
	numr[lb + j] = 0;
	quot[j] = qh;
    }
    for (i = lb - 1; i >= 0 && numr[i] == 0; i--);
    lb = i + 1;
    if (scale) {
	for (i = 0; i < lb; i++) {
	    c[i] = (c[i] >> scale) | (c[i + 1] << (LPSIZE - scale));
	}
    }
    return (c[lb - 1] == 0) ? lb - 1 : lb;
}
#endif
long
bbdiv(a, la, b, lb, c, d, e, r)
     unsigned long   a[],
                     b[],
                     c[],
                     d[],
                     e[];
     long            la,
                     lb;
     long           *r;

{
    unsigned long  *numr,
                   *divr,
                   *quot;
    long            scale;
    long            lc;
    long            i,
                    j,
                    n;
    unsigned long   qh;
    unsigned long   carry;

    if (lb == 1) {
	carry = 0;
	for (i = la - 1; i >= 0; i--) {
	    e[i] = udiv(&carry, carry, a[i], b[0]);
	}
	c[0] = carry;
	*r = carry ? 1 : 0;
	return e[la - 1] == 0 ? la - 1 : la;
    }
    numr = c;
    divr = d;
    quot = e;
#ifdef _LP64
#define FLAG	0x8000000000000000L
#else
#define FLAG	0x80000000
#endif
    if (b[lb - 1] & FLAG) {
	scale = 0;
	for (i = 0; i < la; i++)
	    c[i] = a[i];
	c[la] = 0;
	for (i = 0; i < lb; i++)
	    d[i] = b[i];
    } else {			/* 
				 * need normalization 
				 */
	i = 0;
	n = b[lb - 1];
	if ((n & 0xffff0000) == 0) {
	    n <<= 16;
	    i += 16;
	}
	if ((n & 0xff000000) == 0) {
	    n <<= 8;
	    i += 8;
	}
	if ((n & 0xf0000000) == 0) {
	    n <<= 4;
	    i += 4;
	}
	while ((long) n > 0) {
	    i++;
	    n <<= 1;
	}
	scale = i;
	for (i = lb - 1; i > 0; i--)
	    d[i] = (b[i] << scale) | (b[i - 1] >> (LPSIZE - scale));
	d[0] = b[0] << scale;
	c[la] = a[la - 1] >> (LPSIZE - scale);
	for (i = la - 1; i > 0; i--)
	    c[i] = (a[i] << scale) | (a[i - 1] >> (LPSIZE - scale));
	c[0] = a[0] << scale;
    }
    for (j = la - lb; j >= 0; j--) {
	unsigned long   xh,
	                xl,
	                yh,
	                yl;
	unsigned long   r;

	if (numr[lb + j] == divr[lb - 1]) {
	    qh = -1;
	    if (~numr[lb + j] < numr[lb + j - 1])
		goto ok;
	    r = numr[lb + j] + numr[lb + j - 1];
	} else {
	    qh = udiv(&r, numr[lb + j], numr[lb + j - 1], divr[lb - 1]);
	}
	xh = r;
	xl = numr[lb + j - 2];
	yh = umul(&yl, qh, divr[lb - 2]);
	if (yh > xh || (yh == xh && yl > xl)) {
	    qh = qh - 1;
	    if (yl < xl)
		yh = yh - 1;
	    yl = yl - xl;
	    yh = yh - xh;
	    xh = divr[lb - 1];
	    xl = divr[lb - 2];
	    if (yh > xh || (yh == xh && yl > xl))
		qh = qh - 1;
	}
      ok:
	if (qh == 0) {
	    quot[j] = 0;
	    continue;
	}
	carry = 0;
	for (i = 0; i < lb; i++) {
	    xl = numr[i + j];
	    yh = umuladd(&yl, divr[i], qh, carry);
	    if (xl < yl)
		yh = yh + 1;
	    numr[i + j] = xl - yl;
	    carry = yh;
	}
	yl = yh;
	xl = numr[lb + j];
	if (yl > xl) {
	    unsigned long   m;
	    unsigned long   c;

	    qh = qh - 1;
	    c = 0;
	    for (i = 0; i < lb; i++) {
		m = numr[i + j] + c;
		c = m < c;
		numr[i + j] = m + divr[i];
		if (~m < divr[i])
		    c = 1;
	    }
	}
	numr[lb + j] = 0;
	quot[j] = qh;
    }
    lc = la - lb;
    if (quot[lc])
	lc++;
    for (i = lb - 1; i >= 0 && numr[i] == 0; i--);
    lb = i + 1;
    if (scale) {
	for (i = 0; i < lb; i++) {
	    c[i] = (c[i] >> scale) | (c[i + 1] << (LPSIZE - scale));
	}
    }
    *r = (c[lb - 1] == 0) ? lb - 1 : lb;
    return lc;
}




/*
 */


Sex
absi(Sex a)
{
    if ((long) a < 0) {
	long            n = mkint(a);
	return mkfxat(-n);
    }
    return a;
}

Sex
absb(Sex a)
{
    if (is_bigneg(a))
	return negateb(a);
    return a;
}

Sex
absj(Sex a)
{
    if (is_fixnum(a)) {
	if ((long) a < 0) {
	    long            n = mkint(a);
	    return mkfxat(-n);
	}
    } else if (is_bigneg(a))
	return negateb(a);
    return a;
}


Sex
plus2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return plus2ii(a, b);
    else
	return plus2ib(b, a);
}

Sex
plus2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return plus2ii(a, b);
    else
	return plus2ib(a, b);
}

Sex
plus2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return plus2ii(a, b);
	else
	    return plus2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return plus2ib(b, a);
	else
	    return plus2bb(a, b);
    }
}


Sex
plus2ib(Sex a, Sex b)
{
    long            n = mkint(a);
    if (n == 0)
	return b;
    if (n > 0) {
	if (is_bigneg(b))
	    return diff2_bn(b, n, T_BIGNEG);
	else
	    return plus2_bn(b, n, T_BIGNUM);
    } else {
	n = -n;
	if (is_bigneg(b))
	    return plus2_bn(b, n, T_BIGNEG);
	else
	    return diff2_bn(b, n, T_BIGNUM);
    }
}

Sex
diff2bi(Sex a, Sex b)
{
    long            n = mkint(b);
    if (n == 0)
	return a;
    if (n > 0) {
	if (is_bigneg(a))
	    return plus2_bn(a, n, T_BIGNEG);
	else
	    return diff2_bn(a, n, T_BIGNUM);
    } else {
	n = -n;
	if (is_bigneg(a))
	    return diff2_bn(a, n, T_BIGNEG);
	else
	    return plus2_bn(a, n, T_BIGNUM);
    }
}

Sex
diff2ib(Sex a, Sex b)
{
    long            n = mkint(a);
    if (n == 0)
	return negateb(b);
    if (n > 0) {
	if (is_bigneg(b))
	    return plus2_bn(b, n, T_BIGNUM);
	else
	    return diff2_bn(b, n, T_BIGNEG);
    } else {
	n = -n;
	if (is_bigneg(b))
	    return diff2_bn(b, n, T_BIGNUM);
	else
	    return plus2_bn(b, n, T_BIGNEG);
    }
}



static          Sex
plus2_bn(Sex a, long n, long t)
{
    long            len = bignum_length(a),
	i,
	s;
    Sex             c;


    len = bnum_n(len);
    if (len == 1) {
	s = (long) bignum_value(a)[0] + n;
	if (~bignum_value(a)[0] < (unsigned long) n) {
	    c = new_binteger(t, bnum_len(2));
	    bignum_value(c)[0] = s;
	    bignum_value(c)[1] = 1;
	    return c;
	}
	c = new_binteger(t, bnum_len(1));
	bignum_value(c)[0] = s;
	return c;
    }
    stackcheck();
    push(a);
    c = new_binteger(t, bnum_len(len));
    pop(a);
    for (i = 0; i < len; i++) {
	s = bignum_value(a)[i] + n;
	if (~(bignum_value(a)[i]) < (unsigned long) n)
	    n = 1;
	else
	    n = 0;
	bignum_value(c)[i] = s;
    }
    if (n) {
	if (len & 1) {
	    bignum_value(c)[len] = 1;
	    bignum_length(c) = bnum_len(len + 1);
	    return c;
	} else {
	    push(c);
	    a = new_binteger(t, bnum_len(len + 1));
	    pop(c);
	    for (i = 0; i < len; i++)
		bignum_value(a)[i] = bignum_value(c)[i];
	    bignum_value(a)[len] = 1;
	    return a;
	}
    }
    return c;
}

static          Sex
diff2_bn(Sex a, long n, long t)
{
    long            len = bignum_length(a),
	i,
	s;
    Sex             c;


    len = bnum_n(len);
    if (len == 1) {
	s = (long) bignum_value(a)[0] - n;
	if (t == T_BIGNUM) {
	    if ((s & fix_mask) == 0)
		return mkfxat(s);
	} else {
	    if ((s & fix_mask) == 0 || s == -fix_mask)
		return mkfxat(-s);
	}
	c = new_binteger(t, bnum_len(1));
	bignum_value(c)[0] = s;
	return c;
    }
    stackcheck();
    push(a);
    c = new_binteger(t, bnum_len(len));
    pop(a);
    for (i = 0; i < len; i++) {
	s = bignum_value(a)[i] - n;
	if (bignum_value(a)[i] < (unsigned long) n)
	    n = 1;
	else
	    n = 0;
	bignum_value(c)[i] = s;
    }
    if (n) {
	return aerror2("diff2_bn:this should not happen", a, mkfxat(n));
    }
    if (s == 0) {
	push(c);
	a = new_binteger(t, bnum_len(len - 1));
	pop(c);
	len--;
	for (i = 0; i < len; i++)
	    bignum_value(a)[i] = bignum_value(c)[i];
	return a;
    }
    return c;
}


Sex
diff2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return diff2ii(a, b);
    else
	return diff2bi(a, b);
}

Sex
diff2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return diff2ii(a, b);
    else
	return diff2ib(a, b);
}

Sex
diff2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return diff2ii(a, b);
	else
	    return diff2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return diff2bi(a, b);
	else
	    return diff2bb(a, b);
    }
}






Sex
plus2bb(Sex a, Sex b)
{

    long            ta = numhdr(a),
	tb = numhdr(b);
    long            la = bignum_length(a),
	lb = bignum_length(b),
	i,
	n;
    Sex             c;
    long            m;
    long            sign;

    sign = ta & T_NEG;
    if (la < lb) {
	Sex             t = a;
	long            t1 = la;
	a = b;
	la = lb;
	b = t;
	lb = t1;
	sign = tb & T_NEG;
    }
    if (ta == tb) {		/* 
				 * a + b or -a - b 
				 */
	if (la == bnum_len(1)) {	/* 
					 * both are single word length 
					 */
	    unsigned long   a0 = bignum_value(a)[0];
	    unsigned long   b0 = bignum_value(b)[0];
	    unsigned long   x = a0 + b0;
	    if (~a0 < b0) {
		c = new_binteger(sign | T_BIGNUM, bnum_len(2));
		bignum_value(c)[0] = x;
		bignum_value(c)[1] = 1;
		return c;
	    }
	    c = new_binteger(sign | T_BIGNUM, bnum_len(1));
	    bignum_value(c)[0] = x;
	    return c;
	}
	stackcheck();
	push2(a, b);
	c = new_binteger(sign | T_BIGNUM, la);
	pop2(b, a);
	la = bnum_n(la);
	lb = bnum_n(lb);
	n = bbadd(bignum_value(a), la, bignum_value(b), lb,
		  bignum_value(c));
	if (n == 0)
	    return c;
	if (la & 1) {
	    bignum_length(c) = bnum_len(la + 1);
	    bignum_value(c)[la] = 1;
	    return c;
	}
	push(c);
	b = new_binteger(sign | T_BIGNUM, bnum_len(la + 1));
	pop(c);
	for (i = 0; i < la; i++)
	    bignum_value(b)[i] = bignum_value(c)[i];
	bignum_value(b)[la] = n;
	return b;

    } else {			/* 
				 * a - b or -a + b 
				 */
	stackcheck();
	push2(a, b);
	c = new_binteger(sign | T_BIGNUM, la);
	pop2(b, a);
	la = bnum_n(la);
	lb = bnum_n(lb);
	n = bbsub(bignum_value(a), la, bignum_value(b), lb,
		  bignum_value(c));


	if (n < 0) {
	    n = -n;
	    sign = sign ^ T_NEG;
	    numhdr(c) = (TAG_HDRS + T_BIGNUM) | sign;
	} else if (n == 0) {
	    return mkfxat(0);
	}
	if (n == 1) {
	    m = bignum_value(c)[0];
	    if (sign == 0 && (m & fix_mask) == 0)
		return mkfxat(m);
	    else if (sign && ((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
	if (n == la)
	    return c;

	bignum_length(c) = bnum_len(n);
	if ((n & 1) != 0) {
	    if (la == n + 1)
		return c;
	    n++;
	}
	bignum_value(c)[n] = TAG_HDRS + T_STRING;
	bignum_value(c)[n + 1] = bnum_len(la - n - 2);
	return c;
    }
}

Sex
diff2bb(Sex a, Sex b)
{

    long            ta = numhdr(a),
	tb = numhdr(b);
    long            la = bignum_length(a),
	lb = bignum_length(b),
	i,
	n;
    Sex             c;
    long            m;
    long            sign;

    sign = ta & T_NEG;
    if (la < lb) {
	Sex             t = a;
	long            t1 = la;
	a = b;
	la = lb;
	b = t;
	lb = t1;
	sign = (tb & T_NEG) ^ T_NEG;
    }
    stackcheck();
    if (ta != tb) {		/* 
				 * a + b or -a - b 
				 */
	push2(a, b);
	c = new_binteger(sign | T_BIGNUM, la);
	pop2(b, a);
	la = bnum_n(la);
	lb = bnum_n(lb);
	n = bbadd(bignum_value(a), la, bignum_value(b), lb,
		  bignum_value(c));
	if (n == 0)
	    return c;
	if (la & 1) {
	    bignum_length(c) = bnum_len(la + 1);
	    bignum_value(c)[la] = 1;
	    return c;
	}
	push(c);
	b = new_binteger(sign | T_BIGNUM, bnum_len(la + 1));
	pop(c);
	for (i = 0; i < la; i++)
	    bignum_value(b)[i] = bignum_value(c)[i];
	bignum_value(b)[la] = n;
	return b;

    } else {			/* 
				 * a - b or -a + b 
				 */
	push2(a, b);
	c = new_binteger(sign | T_BIGNUM, la);
	pop2(b, a);
	la = bnum_n(la);
	lb = bnum_n(lb);
	n = bbsub(bignum_value(a), la, bignum_value(b), lb,
		  bignum_value(c));


	if (n < 0) {
	    n = -n;
	    sign = sign ^ T_NEG;
	    numhdr(c) = (TAG_HDRS + T_BIGNUM) | sign;
	} else if (n == 0) {
	    return mkfxat(0);
	}
	if (n == 1) {
	    m = bignum_value(c)[0];
	    if (sign == 0 && (m & fix_mask) == 0)
		return mkfxat(m);
	    else if (sign && ((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
	if (n == la)
	    return c;

	bignum_length(c) = bnum_len(n);
	if ((n & 1) != 0) {
	    if (la == n + 1)
		return c;
	    n++;
	}
	bignum_value(c)[n] = TAG_HDRS + T_STRING;
	bignum_value(c)[n + 1] = bnum_len(la - n - 2);
	return c;
    }
}



Sex
times2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return times2ii(a, b);
    else
	return times2ib(b, a);
}

Sex
times2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return times2ii(a, b);
    else
	return times2ib(a, b);
}

Sex
times2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return times2ii(a, b);
	else
	    return times2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return times2ib(b, a);
	else
	    return times2bb(a, b);
    }
}



Sex
times2ib(Sex a, Sex b)
{
    long            aa = mkint(a),
	lb,
	i;
    long            n,
                    sign;


    if (aa == 0)
	return mkfxat(0);
    else if (aa == 1)
	return b;
    else if (aa == -1)
	return negateb(b);
    sign = numhdr(b) & T_NEG;
    if (aa < 0) {
	aa = -aa;
	sign = sign ^ T_NEG;
    }
    lb = bignum_length(b);
    stackcheck();
    push(b);
    a = new_binteger(sign | T_BIGNUM, lb);
    pop(b);
    lb = bnum_n(lb);
    n = bfmul(bignum_value(b), lb, aa, bignum_value(a));
    if (n) {
	if (lb & 1) {
	    bignum_length(a) = bnum_len(lb + 1);
	    bignum_value(a)[lb] = n;
	    return a;
	}
	push(a);
	b = new_binteger(sign | T_BIGNUM, bnum_len(lb + 1));
	pop(a);
	for (i = 0; i < lb; i++)
	    bignum_value(b)[i] = bignum_value(a)[i];
	bignum_value(b)[lb] = n;
	return b;
    } else {
	return a;
    }
}



static long     karatsuba_buff[(128 + 2 + 4) * SEX / 4];


Sex
times2bb_karatsuba(Sex a, Sex b)
{
    Sex             c,
                    d;
    long            lena,
                    lenb,
                    lenc,
                    lend,
                    n;
    long            sign;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lenb > lena) {
	n = lena;
	lena = lenb;
	lenb = n;
	c = a;
	a = b;
	b = c;
    }
    lenc = lena + lenb;
    // printf("times2bb: lena = %d, lenb = %d\n", lena, lenb);

    if (lena == 1) {
	unsigned long   a0 = bignum_value(a)[0];
	unsigned long   b0 = bignum_value(b)[0];
	unsigned long   h,
	                l;
	h = umul(&l, a0, b0);
	c = new_binteger(T_BIGNUM | sign, bnum_len(2));
	bignum_value(c)[0] = l;
	bignum_value(c)[1] = h;
	return c;
    }
    stackcheck();
    if (lenb > KARATSUBA) {
	lend = (lena + 1) / 2;
	n = 0;
	while (lend > KARATSUBA) {
	    lend = (lend + 1) / 2;
	    n++;
	}
	while (n != 0) {
	    lend = 2 * lend;
	    n--;
	}
	lend = 2 * lend;
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lend * 2));
	if (lend <= 128) {
	    d = (Sex) (alignit(&karatsuba_buff[0]) | TAG_NUMBERS);
	} else {
	    push(c);
	    d = new_bignum(bnum_len(lend));
	    pop(c);
	}
	lend = 2 * lend;
	pop2(b, a);
	if (a == b)
	    n = bbsqr_karatsuba(bignum_value(a), lena,
				bignum_value(c), lend, bignum_value(d));
	else
	    n = bbmul_karatsuba(bignum_value(a), lena, bignum_value(b),
				lenb, bignum_value(c), lend,
				bignum_value(d));
	/*
	 * karatsuba routine has bugs!! pfactorial(27000, 29500) returns
	 * bad result 
	 */
	if (n == 0)		/* 
				 * * * * result is lenc-1 length *      
				 */
	    lenc--;
	bignum_length(c) = bnum_len(lenc);
	if (lenc & 1)
	    lenc++;
	if (lend > lenc) {	/* 
				 * here lenc and lend both are even 
				 */
	    bignum_value(c)[lenc] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc + 1] = bnum_len(lend - lenc - 2);
	}
	return c;
    } else {
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lenc));
	pop2(b, a);
	if (lenb == 1) {
	    n = bfmul(bignum_value(a), lena, bignum_value(b)[0],
		      bignum_value(c));
	    if (n) {
		bignum_value(c)[lena] = n;
		return c;
	    }
	} else {
	    n = bbmul(bignum_value(a), lena, bignum_value(b), lenb,
		      bignum_value(c));
	    if (n)		/* 
				 * * * * result is lenc length
				 *  
				 */
		return c;
	}
	bignum_length(c) = bnum_len(lenc - 1);
	if (lenc & 1) {
	    bignum_value(c)[lenc - 1] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc] = bnum_len(0);
	}
	return c;
    }
}

#ifdef HAVE_GMP
Sex
times2bb_gmp(Sex a, Sex b)
{
    Sex             c,
                    d;
    long            lena,
                    lenb,
                    lenc,
                    lend,
                    n;
    long            sign;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lenb > lena) {
	n = lena;
	lena = lenb;
	lenb = n;
	c = a;
	a = b;
	b = c;
    }
    lenc = lena + lenb;
    // printf("times2bb: lena = %d, lenb = %d\n", lena, lenb);

    if (lena == 1) {
	unsigned long   a0 = bignum_value(a)[0];
	unsigned long   b0 = bignum_value(b)[0];
	unsigned long   h,
	                l;
	h = umul(&l, a0, b0);
	c = new_binteger(T_BIGNUM | sign, bnum_len(2));
	bignum_value(c)[0] = l;
	bignum_value(c)[1] = h;
	return c;
    }
    stackcheck();
    {
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lenc));
	pop2(b, a);
	/*
	 * if (a == b) { #if GMPVERSION >= 432 __gmpn_sqr(bignum_value(c), 
	 * bignum_value(a), lena); #else __gmpn_sqr_n(bignum_value(c),
	 * bignum_value(a), lena); #endif n = bignum_value(c)[lenc - 1]; } 
	 * else 
	 */
	n = __gmpn_mul(bignum_value(c), bignum_value(a), lena,
		       bignum_value(b), lenb);
	/*
	 * Caution!! these use of gmp library is not safe. GMP data format 
	 * and Lisp data format may differ!. 2007.8.30 by f.kako 
	 */
	if (n)
	    return c;
	bignum_length(c) = bnum_len(lenc - 1);
	if (lenc & 1) {
	    bignum_value(c)[lenc - 1] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc] = bnum_len(0);
	}
	return c;
    }
}
#endif

Sex
times2bb_classic(Sex a, Sex b)
{
    Sex             c,
                    d;
    long            lena,
                    lenb,
                    lenc,
                    lend,
                    n;
    long            sign;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lenb > lena) {
	n = lena;
	lena = lenb;
	lenb = n;
	c = a;
	a = b;
	b = c;
    }
    lenc = lena + lenb;
    // printf("times2bb: lena = %d, lenb = %d\n", lena, lenb);

    if (lena == 1) {
	unsigned long   a0 = bignum_value(a)[0];
	unsigned long   b0 = bignum_value(b)[0];
	unsigned long   h,
	                l;
	h = umul(&l, a0, b0);
	c = new_binteger(T_BIGNUM | sign, bnum_len(2));
	bignum_value(c)[0] = l;
	bignum_value(c)[1] = h;
	return c;
    }
    stackcheck();
    {
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lenc));
	pop2(b, a);
	if (lenb == 1) {
	    n = bfmul(bignum_value(a), lena, bignum_value(b)[0],
		      bignum_value(c));
	    if (n) {
		bignum_value(c)[lena] = n;
		return c;
	    }
	} else {
	    n = bbmul(bignum_value(a), lena, bignum_value(b), lenb,
		      bignum_value(c));
	    if (n)		/* 
				 * * * * result is lenc length
				 *  
				 */
		return c;
	}
	bignum_length(c) = bnum_len(lenc - 1);
	if (lenc & 1) {
	    bignum_value(c)[lenc - 1] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc] = bnum_len(0);
	}
	return c;
    }
}


#ifdef TEST_KARATUBA
Sex
times2bb(Sex a, Sex b)
{
    Sex             c,
                    d;

    ASSERT(is_bigint(a) && is_bigint(b));

    push2(a, b);
    c = times2bb_karatsuba(a, b);
    ASSERT(is_bigint(c));
    push(c);
    a = stack[-2];
    b = stack[-1];
    d = times2bb_gmp(a, b);
    ASSERT(is_bigint(d));
    c = stack[0];
    if (!numeq2bb(c, d)) {
	long            la,
	                lb;
	la = bignum_length(c);
	lb = bignum_length(d);
	printf
	    ("times2bb: lc = %ld, ld = %ld, numhdr(c)= 0x%16lx, d = 0x%16lx\n",
	     la, lb, numhdr(c), numhdr(d));
	la = bnum_n(la) - 1;
	while (la >= 0) {
	    printf("%4ld: 0x%16lx -- 0x%16lx: %c\n", la,
		   bignum_value(c)[la], bignum_value(d)[la],
		   bignum_value(c)[la] == bignum_value(d)[la] ? 'Y' : 'N');
	    la--;
	}

	a = stack[-2];
	stack[-2] = d;
	lisp_printf("times2bb( %p", a);
	lisp_printf(", %p)", stack[-1]);
	lisp_printf(" = %p", d);
	lisp_printf(" != %p\n", stack[0]);
	c = stack[-2];
    }
    popv(3);
    return c;
}

#else

Sex
times2bb(Sex a, Sex b)
{
    Sex             c,
                    d;
    long            lena,
                    lenb,
                    lenc,
                    lend,
                    n;
    long            sign;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lenb > lena) {
	n = lena;
	lena = lenb;
	lenb = n;
	c = a;
	a = b;
	b = c;
    }
    lenc = lena + lenb;
    // printf("times2bb: lena = %d, lenb = %d\n", lena, lenb);

    if (lena == 1) {
	unsigned long   a0 = bignum_value(a)[0];
	unsigned long   b0 = bignum_value(b)[0];
	unsigned long   h,
	                l;
	h = umul(&l, a0, b0);
	c = new_binteger(T_BIGNUM | sign, bnum_len(2));
	bignum_value(c)[0] = l;
	bignum_value(c)[1] = h;
	return c;
    }
    stackcheck();
#ifdef USE_GMP
    {
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lenc));
	pop2(b, a);
	/*
	 * if (a == b) { #if GMPVERSION >= 432 __gmpn_sqr(bignum_value(c), 
	 * bignum_value(a), lena); #else __gmpn_sqr_n(bignum_value(c),
	 * bignum_value(a), lena); #endif n = bignum_value(c)[lenc - 1]; } 
	 * else 
	 */
	n = __gmpn_mul(bignum_value(c), bignum_value(a), lena,
		       bignum_value(b), lenb);
	/*
	 * Caution!! these use of gmp library is not safe. GMP data format 
	 * and Lisp data format may differ!. 2007.8.30 by f.kako 
	 */
	if (n)
	    return c;
	bignum_length(c) = bnum_len(lenc - 1);
	if (lenc & 1) {
	    bignum_value(c)[lenc - 1] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc] = bnum_len(0);
	}
	return c;
    }
#else
#ifdef USE_KARATSUBA
    if (lenb > KARATSUBA) {
	lend = (lena + 1) / 2;
	n = 0;
	while (lend > KARATSUBA) {
	    lend = (lend + 1) / 2;
	    n++;
	}
	while (n != 0) {
	    lend = 2 * lend;
	    n--;
	}
	lend = 2 * lend;
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lend * 2));
	if (lend <= 128) {
	    d = (Sex) (alignit(&karatsuba_buff[0]) | TAG_NUMBERS);
	} else {
	    push(c);
	    d = new_bignum(bnum_len(lend));
	    pop(c);
	}
	lend = 2 * lend;
	pop2(b, a);
	if (a == b)
	    n = bbsqr_karatsuba(bignum_value(a), lena,
				bignum_value(c), lend, bignum_value(d));
	else
	    n = bbmul_karatsuba(bignum_value(a), lena, bignum_value(b),
				lenb, bignum_value(c), lend,
				bignum_value(d));
	/*
	 * karatsuba routine has bugs!! pfactorial(27000, 29500) returns
	 * bad result 
	 */
	if (n == 0)		/* 
				 * * * * result is lenc-1 length *      
				 */
	    lenc--;
	bignum_length(c) = bnum_len(lenc);
	if (lenc & 1)
	    lenc++;
	if (lend > lenc) {	/* 
				 * here lenc and lend both are even 
				 */
	    bignum_value(c)[lenc] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc + 1] = bnum_len(lend - lenc - 2);
	}
	return c;
    } else
#endif
    {
	push2(a, b);
	c = new_binteger(T_BIGNUM | sign, bnum_len(lenc));
	pop2(b, a);
	if (lenb == 1) {
	    n = bfmul(bignum_value(a), lena, bignum_value(b)[0],
		      bignum_value(c));
	    if (n) {
		bignum_value(c)[lena] = n;
		return c;
	    }
	} else {
	    n = bbmul(bignum_value(a), lena, bignum_value(b), lenb,
		      bignum_value(c));
	    if (n)		/* 
				 * * * * result is lenc length
				 *  
				 */
		return c;
	}
	bignum_length(c) = bnum_len(lenc - 1);
	if (lenc & 1) {
	    bignum_value(c)[lenc - 1] = T_STRING + TAG_HDRS;
	    bignum_value(c)[lenc] = bnum_len(0);
	}
	return c;
    }
#endif
}
#endif


/*
 * Division
 */

Sex
quot2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return quot2ii(a, b);
    else
	return quot2bi(a, b);
}

Sex
quot2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return quot2ii(a, b);
    else
	return quot2ib(a, b);
}

Sex
quot2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return quot2ii(a, b);
	else
	    return quot2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return quot2bi(a, b);
	else
	    return quot2bb(a, b);
    }
}



Sex
quot2ib(Sex a, Sex b)
{
    if (a == mkfxat(fix_mask) && bignum_length(b) == bnum_len(1) &&
	!is_bigneg(b) && bignum_value(b)[0] == big_min)
	return mkfxat(-1);
    else
	return mkfxat(0);
}


Sex
quot2bi(Sex a, Sex b)
{
    long            n,
                    m;
    long            len,
                    sign;

    if (b == mkfxat(1))
	return a;
    else if (b == mkfxat(-1))
	return negateb(a);
    else if (b == mkfxat(0))
	return aerror2("bad arg for quotient", a, b);
    sign = numhdr(a) & T_NEG;
    n = mkint(b);
    if (n < 0) {
	sign = sign ^ T_NEG;
	n = -n;
    }
    stackcheck();
    push(a);
    len = bignum_length(a);
    b = new_binteger(sign | T_BIGNUM, len);
    pop(a);
    len = bnum_n(len);
    bfdiv(bignum_value(a), len, n, bignum_value(b));
    if (bignum_value(b)[len - 1] == 0) {	/* 
						 * shrink bignum 
						 */
	bignum_length(b) = bnum_len(len - 1);
	if (len & 1) {
	    bignum_value(b)[len - 1] = T_STRING + TAG_HDRS;
	    bignum_value(b)[len] = bnum_len(0);
	}
	len--;
    }
    if (len == 1) {
	m = bignum_value(b)[0];
	if ((m & fix_mask) == 0) {
	    if (sign)
		return mkfxat(-m);
	    else
		return mkfxat(m);
	} else if (m == big_min && sign) {
	    return mkfxat(-big_min);
	}
    }
    return b;
}


Sex
quot2bb(Sex a, Sex b)
{
    Sex             c;
    long            la,
                    lb,
                    lc;
    long            i,
                    m;
    long            sign;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    la = bignum_length(a);
    lb = bignum_length(b);
    if (la < lb)
	return mkfxat(0);
    stackcheck();
#ifdef USE_GMP
    push2(a, b);
    c = new_binteger(sign | T_BIGNUM, la + 3 * SEX);
    pop2(b, a);
    la = bnum_n(la);
    lb = bnum_n(lb);
    lc = la - lb + 1;
    __gmpn_tdiv_qr(bignum_value(c), &bignum_value(c)[lc], 0L,
		   bignum_value(a), la, bignum_value(b), lb);
    m = bignum_value(c)[lc - 1];
    if (m == 0)
	lc--;
    if (lc == 0)
	return mkfxat(0);
    if (lc == 1) {
	m = bignum_value(c)[lc - 1];
	if (sign == 0) {
	    if ((m & fix_mask) == 0)
		return mkfxat(m);
	} else {
	    if (((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
    }
    bignum_length(c) = bnum_len(lc);
    if (lc & 1)
	lc++;
    bignum_value(c)[lc] = TAG_HDRS | T_STRING;
    bignum_value(c)[lc + 1] =
	((lb + 2 + la - lb + 1 + 2) - (lc + 2)) * SEX;
    return c;
#else
    push2(a, b);
    c = new_binteger(sign | T_BIGNUM, 2 * la);
    la = bnum_n(la);
    lb = bnum_n(lb);
    pop2(b, a);
#ifdef SAVE_SPACE
    lc = bbdiv(bignum_value(a), la, bignum_value(b), lb,
	       bignum_value(c), &bignum_value(c)[la + 1],
	       &bignum_value(c)[la + 1 + lb], &lc);
#else
    lc = bbquo(bignum_value(a), la, bignum_value(b), lb,
	       bignum_value(c), &bignum_value(c)[la + 1],
	       &bignum_value(c)[la + 1 + lb]);
#endif
    if (lc == 0)
	return mkfxat(0);
    m = bignum_value(c)[la + 1 + lb];
    if (lc == 1) {
	if (sign == 0) {
	    if ((m & fix_mask) == 0)
		return mkfxat(m);
	} else {
	    if (((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
    }
    bignum_length(c) = bnum_len(lc);
    for (i = 0; i < lc; i++)
	bignum_value(c)[i] = bignum_value(c)[la + 1 + lb + i];
    if (lc & 1)
	lc++;
    bignum_value(c)[lc] = TAG_HDRS | T_STRING;
    bignum_value(c)[lc + 1] =
	((la + 2 + 1 + lb + 2 + la - lb + 1 - 2) - (lc + 2)) * SEX;
    return c;
#endif
}

Sex
rem2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return rem2ii(a, b);
    else
	return rem2bi(a, b);
}

Sex
rem2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return rem2ii(a, b);
    else
	return rem2ib(a, b);
}

Sex
rem2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return rem2ii(a, b);
	else
	    return rem2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return rem2bi(a, b);
	else
	    return rem2bb(a, b);
    }
}



Sex
rem2ib(Sex a, Sex b)
{
    if (a == mkfxat(fix_mask) && bignum_length(b) == bnum_len(1) &&
	!is_bigneg(b) && bignum_value(b)[0] == big_min)
	return mkfxat(0);
    else
	return a;
}


Sex
rem2bi(Sex a, Sex b)
{
    long            n;

    if (b == mkfxat(0))
	return aerror2("bad arg for remainder", a, b);
    else if (b == mkfxat(1) || b == mkfxat(-1))
	return mkfxat(0);
    n = mkint(b);
    n = bfrem(bignum_value(a), bnum_n(bignum_length(a)), n < 0 ? -n : n);
    if (is_bigneg(a))
	return mkfxat(-n);
    else
	return mkfxat(n);
}



Sex
rem2bb(Sex a, Sex b)
{
    Sex             c;
    long            la,
                    lb,
                    lc;
    long
                    m,
                    sign;

    sign = numhdr(a) & T_NEG;
    la = bignum_length(a);
    lb = bignum_length(b);
    if (la < lb)
	return a;
    stackcheck();
#ifdef USE_GMP
    push2(a, b);
    c = new_binteger(sign | T_BIGNUM, la + 4 * SEX);
    la = bnum_n(la);
    lb = bnum_n(lb);
    pop2(b, a);
    __gmpn_tdiv_qr(&bignum_value(c)[lb + 1], bignum_value(c), 0L,
		   bignum_value(a), la, bignum_value(b), lb);

    for (lc = lb - 1; lc >= 0 && bignum_value(c)[lc] == 0; lc--);

    if (lc < 0)
	return mkfxat(0);
    if (lc == 0) {
	m = bignum_value(c)[0];
	if (sign == 0) {
	    if ((m & fix_mask) == 0)
		return mkfxat(m);
	} else {
	    if (((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
    }
    bignum_length(c) = bnum_len(lc + 1);
    if ((lc & 1) == 0)
	lc++;
    bignum_value(c)[lc + 1] = TAG_HDRS | T_STRING;
    bignum_value(c)[lc + 2] = ((lb + 2 + 1 + la - lb + 1 + 2) -
			       (lc + 1 + 2)) * SEX;
    return c;
#else
    push2(a, b);
    c = new_binteger(sign | T_BIGNUM, 2 * la);
    la = bnum_n(la);
    lb = bnum_n(lb);
    pop2(b, a);
#ifdef SAVE_SPACE
    bbdiv(bignum_value(a), la, bignum_value(b), lb,
	  bignum_value(c), &bignum_value(c)[la + 1],
	  &bignum_value(c)[la + 1 + lb], &lc);
#else
    lc = bbrem(bignum_value(a), la, bignum_value(b), lb,
	       bignum_value(c), &bignum_value(c)[la + 1],
	       &bignum_value(c)[la + 1 + lb]);
#endif
    if (lc == 0)
	return mkfxat(0);

    m = bignum_value(c)[0];
    if (lc == 1) {
	if (sign == 0) {
	    if ((m & fix_mask) == 0)
		return mkfxat(m);
	} else {
	    if (((m & fix_mask) == 0 || m == big_min))
		return mkfxat(-m);
	}
    }
    bignum_length(c) = bnum_len(lc);
    if (lc & 1)
	lc++;
    bignum_value(c)[lc] = TAG_HDRS | T_STRING;
    bignum_value(c)[lc + 1] = ((la + 2 + 1 + lb + 2 + la - lb + 1 - 2) -
			       (lc + 2)) * SEX;
    return c;
#endif
}

Sex
divide2bb(Sex a, Sex b)
{
    Sex             c;
    Sex             d;
    long            la,
                    lb,
                    lc,
                    ld,
                    l;
    long            i,
                    m;
    long            sign,
                    signr;

    sign = ((numhdr(a) ^ numhdr(b))) & T_NEG;
    signr = numhdr(a) & T_NEG;
    la = bignum_length(a);
    lb = bignum_length(b);
    if (la < lb)
	return cons(mkfxat(0), a);
    stackcheck();
    push2(a, b);
#ifdef USE_GMP
    c = new_binteger(sign | T_BIGNUM, la - lb + HDR + SEX);
    push(c);
    d = new_binteger(signr | T_BIGNUM, lb);
    pop(c);
    la = bnum_n(la);
    lb = bnum_n(lb);
    pop2(b, a);
    __gmpn_tdiv_qr(bignum_value(c), bignum_value(d), 0L,
		   bignum_value(a), la, bignum_value(b), lb);
    lc = la - lb + 1;
    if (bignum_value(c)[lc - 1] == 0)
	lc--;
    for (ld = lb - 1; ld >= 0 && bignum_value(d)[ld] == 0; ld--);
    ld++;

    b = nil;
    if (ld == 0) {
	b = mkfxat(0);
    } else {
	if (ld == 1) {
	    m = bignum_value(d)[0];
	    if (signr == 0) {
		if ((m & fix_mask) == 0)
		    b = mkfxat(m);
	    } else {
		if (((m & fix_mask) == 0 || m == big_min))
		    b = mkfxat(-m);
	    }
	}
    }
    if (b == nil) {
	bignum_length(d) = bnum_len(ld);
	b = d;
	if (lb & 1)
	    lb++;
	if (ld & 1)
	    ld++;
	if (lb - ld > 0) {
	    bignum_value(d)[ld] = TAG_HDRS | T_STRING;
	    bignum_value(d)[ld + 1] = bnum_len(lb - ld - 2);
	}
    }
    a = nil;
    if (lc == 0) {
	a = mkfxat(0);
    } else {
	if (lc == 1) {
	    m = bignum_value(c)[0];
	    if (sign == 0) {
		if ((m & fix_mask) == 0)
		    a = mkfxat(m);
	    } else {
		if (((m & fix_mask) == 0 || m == big_min))
		    a = mkfxat(-m);
	    }
	}
    }
    if (a == nil) {
	if (lc == la - lb) {
	    bignum_length(c) = bnum_len(lc);
	    if ((lc & 1) == 0) {
		bignum_value(c)[lc] = TAG_HDRS | T_STRING;
		bignum_value(c)[lc + 1] = bnum_len(0);
	    }
	}
	a = c;
    }
    return cons(a, b);

#else
    c = new_binteger(sign | T_BIGNUM, 2 * la + HDR);
    la = bnum_n(la);
    lb = bnum_n(lb);
    pop2(b, a);
    lc = bbdiv(bignum_value(a), la, bignum_value(b), lb,
	       bignum_value(c), &bignum_value(c)[la + 1],
	       &bignum_value(c)[la + 1 + lb + 2], &ld);
    b = nil;
    l = -2;
    if (ld == 0) {
	b = mkfxat(0);
    } else {
	if (ld == 1) {
	    m = bignum_value(c)[0];
	    if (signr == 0) {
		if ((m & fix_mask) == 0)
		    b = mkfxat(m);
	    } else {
		if (((m & fix_mask) == 0 || m == big_min))
		    b = mkfxat(-m);
	    }
	}
    }
    if (b == nil) {
	bignum_length(c) = bnum_len(ld);
	if (ld & 1)
	    ld++;
	l = ld;
	b = c;
    }
    d = ((Sex) & (bignum_value(c)[l])) | TAG_NUMBERS;
    numhdr(d) = numhdr(c);
    numhdr(c) = numhdr(a);
    bignum_length(d) = (2 * la + 3) * SEX;
    a = nil;
    if (lc == 0) {
	a = mkfxat(0);
    } else {
	if (lc == 1) {
	    m = bignum_value(c)[la + 1 + lb + 2];
	    if (sign == 0) {
		if ((m & fix_mask) == 0)
		    a = mkfxat(m);
	    } else {
		if (((m & fix_mask) == 0 || m == big_min))
		    a = mkfxat(-m);
	    }
	}
    }
    if (a == nil) {
	bignum_length(d) = bnum_len(lc);
	for (i = 0; i < lc; i++)
	    bignum_value(d)[i] = bignum_value(c)[la + 1 + lb + i + 2];
	if (lc & 1)
	    lc++;

	ld = ((la + 2 + 1 + lb + 2 + la - lb + 1 - 2 + 2) -
	      (l + 2 + lc + 2)) * SEX;
	if (ld > 0) {
	    bignum_value(d)[lc] = TAG_HDRS | T_STRING;
	    bignum_value(d)[lc + 1] = ld;
	}
	a = d;
    }
    return cons(a, b);
#endif
}

Sex
divide2bi(Sex a, Sex b)
{
    long            n,
                    m;
    long            len,
                    sign;

    if (b == mkfxat(1))
	return cons(a, mkfxat(0));
    else if (b == mkfxat(-1))
	return cons(negateb(a), mkfxat(0));
    else if (b == mkfxat(0))
	return aerror2("bad arg for quotient", a, b);
    sign = numhdr(a) & T_NEG;
    n = mkint(b);
    if (n < 0) {
	sign = sign ^ T_NEG;
	n = -n;
    }
    stackcheck();
    push(a);
    len = bignum_length(a);
    b = new_binteger(sign | T_BIGNUM, len);
    pop(a);
    len = bnum_n(len);
    n = bfdiv(bignum_value(a), len, n, bignum_value(b));
    if (bignum_value(b)[len - 1] == 0) {	/* 
						 * shrink bignum 
						 */
	bignum_length(b) = bnum_len(len - 1);
	if (len & 1) {
	    bignum_value(b)[len - 1] = T_STRING + TAG_HDRS;
	    bignum_value(b)[len] = bnum_len(0);
	}
	len--;
    }
    a = mkfxat(is_bigneg(a) ? -n : n);
    if (len == 1) {
	m = bignum_value(b)[0];
	if ((m & fix_mask) == 0) {
	    b = mkfxat(sign ? -m : m);
	    return cons(b, a);
	} else if (m == big_min && sign) {
	    return cons(mkfxat(-big_min), a);
	}
    }
    return cons(b, a);
}

Sex
gcdbb(Sex a, Sex b)
{
    stackcheck();
    while (!is_fixnum(b)) {
	push(b);
	b = rem2bb(a, b);
	pop(a);
    }
    return gcdbi(a, b);
}




Sex
mod2ji(Sex a, Sex b)
{
    if (is_fixnum(a))
	return mod2ii(a, b);
    else
	return mod2bi(a, b);
}

Sex
mod2ij(Sex a, Sex b)
{
    if (is_fixnum(b))
	return mod2ii(a, b);
    else
	return mod2ib(a, b);
}

Sex
mod2jj(Sex a, Sex b)
{
    if (is_fixnum(a)) {
	if (is_fixnum(b))
	    return mod2ii(a, b);
	else
	    return mod2ib(a, b);
    } else {
	if (is_fixnum(b))
	    return mod2bi(a, b);
	else
	    return mod2bb(a, b);
    }
}



Sex
mod2bi(Sex a, Sex b)
{
    bool            f = numhdr(a) & T_NEG;
    long            n;
    a = rem2bi(a, b);
    n = mkint(a);
    if (f) {
	if (n > 0)
	    a = plus2ii(a, b);
    } else if (n < 0)
	a = plus2ii(a, b);
    return a;
}

Sex
mod2ib(Sex a, Sex b)
{
    long            n;
    n = mkint(a);
    if (numhdr(b) & T_NEG) {
	if (n > 0)
	    a = plus2ib(a, b);
    } else if (n < 0)
	a = plus2ib(a, b);
    return a;
}

Sex
mod2bb(Sex a, Sex b)
{
    bool            f = numhdr(b) & T_NEG;
    a = rem2bb(a, b);
    if (f) {
	if (plusp(a))
	    a = plus2(a, b);
    } else if (minusp(a))
	a = plus2(a, b);
    return a;
}



Sex
smod2bi(Sex a, Sex b)
{
    long            n;
    long            m = mkint(b);
    long            m2;

    if (m < 0)
	m = -m;
    m2 = m / 2;

    a = rem2bi(a, b);
    n = mkint(a);

    if (n > m2)
	n = n - m;
    else if (n < -m2)
	n = n + m;
    return mkfxat(n);

}

#if 0
Sex
smod2ib(Sex a, Sex b)
{
    if (numhdr(b) & T_NEG) {
	if (a > 0)
	    a = plus2ib(a, b);
    } else if (a < 0)
	a = plus2ib(a, b);
    return a;
}
#endif

#if 0
Sex
smod2bb(Sex a, Sex b)
{
    Sex             b2;
    bool            f = numhdr(b) & T_NEG;
    push2(a, b);
  b2:= quot2bi(b, mkfxat(2));
    pop2(b, a);
    push(b);
    a = rem2bb(a, b);
    pop(b);
    if (f) {
	if (plusp(a)) {
	    a = plus2(a, b);
	}
    } else if (minusp(a))
	a = plus2(a, b);
    return a;
}
#endif


Sex
smod2ib(Sex a, Sex b)
{
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define m stack[0]
    if (minusp(b))
	b = negate(b);
    m = quot2(b, mkfxat(2));
    a = remaind2(a, b);
    if (lessp2(m, a)) {
	a = diff2(a, b);
    } else {
	m = negate(m);
	if (lessp2(a, m))
	    a = plus2(a, b);
    }
#undef a
#undef b
#undef m
    a = stack[-2];
    popv(3);
    return a;
}

Sex
smod2bb(Sex a, Sex b)
{
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define m stack[0]
    if (minusp(b))
	b = negate(b);
    m = quot2(b, mkfxat(2));
    a = remaind2(a, b);
    if (lessp2(m, a)) {
	a = diff2(a, b);
    } else {
	m = negate(m);
	if (lessp2(a, m))
	    a = plus2(a, b);
    }
#undef a
#undef b
#undef m
    a = stack[-2];
    popv(3);
    return a;
}





/*
 * Arithmetic comparison: lessp2
 */

bool
lessp2ib(Sex a, Sex b)
{
    if (is_bigneg(b))
	return NO;
    return YES;
}

bool
lessp2bi(Sex a, Sex b)
{
    if (is_bigneg(a))
	return YES;
    return NO;
}


bool
lessp2bb(Sex a, Sex b)
{
    long            ta = numhdr(a),
	tb = numhdr(b);
    long            lena = bignum_length(a),
	lenb = bignum_length(b);
    if (ta < tb)
	return NO;
    else if (ta > tb)
	return YES;
    ta = (ta & T_NEG) >> S_NEG;
    tb = 1 - ta;
    if (lena < lenb)
	return tb;
    else if (lena > lenb)
	return ta;
    lena = bnum_n(lena);
    while (--lena >= 0) {
	unsigned long   da = bignum_value(a)[lena],
	    db = bignum_value(b)[lena];
	if (da < db)
	    return tb;
	if (da > db)
	    return ta;
    }
    return NO;
}



bool
numeq2bb(Sex a, Sex b)
{
    long            la = bignum_length(a);
    if (numhdr(a) != numhdr(b))
	return NO;
    if (la != (long) bignum_length(b))
	return NO;
    la = bnum_n(la) - 1;
    while (la >= 0) {
	if (bignum_value(a)[la] != bignum_value(b)[la])
	    return NO;
	else
	    la--;
    }
    return YES;
}


void
print_bignum(Sex u, putchar_fn * fn)
{
    long            len = numlen(u);
    long            i,
                    len1,
                    len2;
    Sex             w;
    long            sign;
    char            my_buff[32];
    unsigned long   dig;


    sign = numhdr(u) & T_NEG;

    if (len == bnum_len(1)) {
	dig = bignum_value(u)[0];
	for (i = 0; dig != 0; i++) {
	    my_buff[i] = (long) (dig % 10) + '0';
	    dig = dig / 10;
	}
	if (sign)
	    fn('-');
	while (i > 0)
	    fn(my_buff[--i]);
	return;
    }
    stackcheck();
    push(u);
    if (sign)
	fn('-');
    len1 = HDR + (11 * len) / 10;
    len1 = (long) alignit(len1);
    w = new_bignum(len1);
    pop(u);

    len = bnum_n(len);
    len1 = bnum_n(len1);
    for (i = 0; i < len; i++)
	bignum_value(w)[i] = bignum_value(u)[i];
    len2 = len1;

    while (len > 0) {
	unsigned long   carry;
	long            k;
	carry = 0;
	for (k = len - 1; k >= 0; k--) {
	    bignum_value(w)[k] = udiv(&carry, carry,
				      bignum_value(w)[k], M10_PRINT_EXP);
	}
	if (bignum_value(w)[len - 1] == 0)
	    len--;
	bignum_value(w)[--len2] = carry;
    }
    push(w);
    if (bignum_value(w)[0] == 0)
	dig = bignum_value(w)[len2++];
    else
	dig = bignum_value(w)[0];
    i = 0;
    do {
	long            nxt = dig % 10;
	dig = dig / 10;
	my_buff[i++] = (long) nxt + '0';
    } while (dig != 0);
    while (i > 0)
	fn(my_buff[--i]);
    pop(w);
    while (len2 < len1) {
	dig = bignum_value(w)[len2++];
	push(w);
	for (i = M10_PRINT_DIG - 1; i >= 0; i--) {
	    long            nxt = dig % 10;
	    dig = dig / 10;
	    my_buff[i] = (long) nxt + '0';
	}
	for (i = 0; i < M10_PRINT_DIG; i++)
	    fn(my_buff[i]);
	pop(w);
    }
}

void
print_phex(Sex u, long k, putchar_fn * fn)
{
    static char     hexdig[] = "0123456789abcdef";
    long            n,
                    m;
    unsigned long   a;

    fn('0');
    fn('x');
    if (is_fixnum(u)) {
	if (k > SEX * 2)
	    k = SEX * 2;
	a = mkint(u);
	for (m = k * 4 - 4; m >= 0; m -= 4)
	    fn(hexdig[(a >> m) & 0xf]);
    } else {
	n = bnum_n(bignum_length(u));
	if (n > k / (2 * SEX)) {
	    n = k / (2 * SEX);
	    k = k - n * (2 * SEX);
	} else {
	    n--;
	    k = 2 * SEX;
	}
	a = bignum_value(u)[n];
	for (m = 4 * k - 4; m >= 0; m -= 4)
	    fn(hexdig[(a >> m) & 0xf]);
	for (n--; n >= 0; n--) {
	    a = bignum_value(u)[n];
	    for (m = LPSIZE - 4; m >= 0; m -= 4)
		fn(hexdig[(a >> m) & 0xf]);
	}
    }
}


void
print_bighex(Sex u, putchar_fn * fn)
{
    static char     hexdig[] = "0123456789abcdef";
    long            n,
                    m;
    unsigned long   a;

    fn('0');
    fn('x');
    for (n = bnum_n(bignum_length(u)) - 1; n >= 0; n--) {
	a = bignum_value(u)[n];
	/*
	 * fn(' ');      
	 */
	for (m = LPSIZE - 4; m >= 0; m -= 4)
	    fn(hexdig[(a >> m) & 0xf]);
    }
}




Sex
copyb(Sex a)
/*
 * copy a bignum.
 */
{
    Sex             b;
    long            len = bignum_length(a),
	i;
    stackcheck();
    push(a);
    b = new_bignum(len);
    pop(a);
    len = bnum_n(len);
    for (i = 0; i < len; i++)
	bignum_value(b)[i] = bignum_value(a)[i];
    return b;
}

Sex
negateb(Sex a)
{
    Sex             b;
    long            len = bignum_length(a);
    long            i,
                    type;

    if (is_bigneg(a)) {
	if (len == bnum_len(1)) {
	    i = -(long) bignum_value(a)[0];
	    if (i == fix_mask)
		return mkfxat(i);
	}
	type = T_BIGNUM;
    } else {
	if (len == bnum_len(1) && bignum_value(a)[0] == big_min)
	    return mkfxat(big_min);
	type = T_BIGNEG;
    }
    stackcheck();
    push(a);
    b = new_binteger(type, len);
    pop(a);
    len = bnum_n(len);
    for (i = 0; i < len; i++)
	bignum_value(b)[i] = bignum_value(a)[i];
    return b;
}
