/*
 * bmodular.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * short int and modular arithmetics
 *      $Header: /Users/kako/asuka/src/RCS/bmodular.c,v 2.9 2016/02/03 01:08:44 kako Exp $
 */

static char const rcsid[] =
    "$Id: bmodular.c,v 2.9 2016/02/03 01:08:44 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "mathop.h"
#include "compare.h"


#ifdef _LP64
#define M_SQROOT        3037000499
#else
#define M_SQROOT        46341
#endif

void            print_bmodular(Sex u, putchar_fn * fn);
Sex             Lmkbmodular(Sex a);
Sex             Lbmodular(Sex a, Sex m);
Sex             times2Mi(Sex a, Sex b);
Sex             recipiM(Sex n, Sex m);
Sex             recipM(Sex n);
bool            numeq2Mr(Sex a, Sex b);
bool            numeq2Mc(Sex a, Sex b);
Sex             Lbmodular_difference(Sex a, Sex b);
Sex             Lbmodular_plus(Sex a, Sex b);
Sex             Lbmodular_times(Sex a, Sex b);
Sex             Lbmodular_quotient(Sex a, Sex b);
Sex             Lbmodular_expt(Sex a, Sex b);
Sex             Lbmodular_minus(Sex a);
Sex             Lbmodular_number(Sex a);
Sex             Lbmodular_reciprocal(Sex n);
Sex             Lset_modulus(Sex a);
Sex             Lset_big_modulus(Sex a);

/***************************************************************/
/***************************************************************/
/***************************************************************/


void
print_bmodular(Sex u, putchar_fn * fn)
{
    print_sex(bmodular_value(u), fn);
}


Sex
Lbmodular(Sex a, Sex m)
{

    Sex             c;

    stackcheck();
    if (is_fixnum(a)) {
	push(m);
	c = new_bmodular();
	pop(m);
	bmodular_value(c) = a;
	bmodular_modulus(c) = m;
	return c;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    push2(a, m);
	    a = remaind2(a, m);
	    if (minusp(a)) {
		m = stack[0];
		a = plus2(a, m);
	    }
	    stack[-1] = a;
	    c = new_bmodular();
	    pop2(m, a);
	    bmodular_value(c) = a;
	    bmodular_modulus(c) = m;
	    return c;
	case N_MODULAR:
	    return a;
	case N_RATIONAL:
	    {
		push2(m, rational_den(a));
		a = Lbmodular(rational_num(a), m);
		push(a);
		a = Lbmodular(stack[-1], stack[-2]);
		c = stack[0];
		popv(3);
		return quot2MM(c, a);
	    }
	case N_COMPLEX:
	    {
		push2(m, complex_imag(a));
		a = Lbmodular(complex_real(a), m);
		push(a);
		a = Lbmodular(stack[-1], stack[-2]);
		push(a);
		c = new_complex();
		complex_imag(c) = stack[0];
		complex_real(c) = stack[-1];
		popv(4);
		return c;
	    }
	case N_BMODULAR:
	    return a;
	}
    }
    return error1("bad argument to mkbmodular ", a);
}

Sex
Lmkbmodular(Sex a)
{

    Sex             c;

    stackcheck();
    if (is_fixnum(a)) {
	c = new_bmodular();
	bmodular_value(c) = a;
	bmodular_modulus(c) = current_big_modulus;
	return c;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    a = remaind2(a, current_big_modulus);
	    if (minusp(a)) {
		a = plus2(a, current_big_modulus);
	    }
	    push(a);
	    c = new_bmodular();
	    pop(a);
	    bmodular_value(c) = a;
	    bmodular_modulus(c) = current_big_modulus;
	    return c;
	case N_MODULAR:
	    return a;
	case N_RATIONAL:
	    {
		push(rational_den(a));
		a = Lbmodular(rational_num(a), current_big_modulus);
		push(a);
		a = Lbmodular(stack[0], current_big_modulus);
		pop(c);
		return quot2MM(c, a);
	    }
	case N_COMPLEX:
	    {
		push(complex_imag(a));
		a = Lbmodular(complex_real(a), current_big_modulus);
		push(a);
		a = Lbmodular(stack[0], current_big_modulus);
		push(a);
		c = new_complex();
		complex_imag(c) = stack[0];
		complex_real(c) = stack[-1];
		popv(2);
		return c;
	    }
	case N_BMODULAR:
	    return a;
	}
    }
    return error1("bad argument to mkbmodular ", a);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2iM(Sex a, Sex b)
{
    Sex             m;
    m = bmodular_modulus(b);
    push2(a, m);
    a = plus2(a, bmodular_value(b));
    a = remaind2(a, stack[0]);
    if (minusp(a)) {
	a = plus2(a, stack[0]);
    }
    stack[-1] = a;
    b = new_bmodular();
    bmodular_value(b) = stack[-1];
    bmodular_modulus(b) = stack[0];
    popv(2);
    return b;
}


Sex
plus2bM(Sex a, Sex b)
{
    Sex             m;
    m = bmodular_modulus(b);
    push2(a, m);
    a = plus2(a, bmodular_value(b));
    a = remaind2(a, stack[0]);
    if (minusp(a)) {
	a = plus2(a, stack[0]);
    }
    stack[-1] = a;
    b = new_bmodular();
    bmodular_value(b) = stack[-1];
    bmodular_modulus(b) = stack[0];
    popv(2);
    return b;
}



Sex
plus2Mi(Sex a, Sex b)
{
    return plus2iM(b, a);
}


Sex
plus2Mb(Sex a, Sex b)
{
    return plus2bM(b, a);
}

Sex
plus2Mr(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return plus2MM(a, b);
}

Sex
plus2rM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return plus2MM(a, b);
}

Sex
plus2MM(Sex a, Sex b)
{
    Sex             m;

    m = bmodular_modulus(a);
    push(m);
    if (!numeq2(m, bmodular_modulus(b))) {
	popv(1);
	return math_er(a, b);
    }
    a = plus2jj(bmodular_value(a), bmodular_value(b));
    m = stack[0];
    if (lessp2(m, a))
	a = diff2(a, stack[0]);
    push(a);
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
diff2iM(Sex a, Sex b)
{
    Sex             m;
    m = bmodular_modulus(b);
    push2(a, m);
    a = diff2(a, bmodular_value(b));
    a = remaind2(a, stack[0]);
    if (minusp(a)) {
	a = plus2(a, stack[0]);
    }
    stack[-1] = a;
    b = new_bmodular();
    bmodular_value(b) = stack[-1];
    bmodular_modulus(b) = stack[0];
    popv(2);
    return b;
}



Sex
diff2bM(Sex a, Sex b)
{
    return diff2iM(a, b);
}



Sex
diff2Mi(Sex a, Sex b)
{
    Sex             m;
    m = bmodular_modulus(a);
    push2(b, m);
    a = diff2(bmodular_value(a), b);
    a = remaind2(a, stack[0]);
    if (minusp(a)) {
	a = plus2(a, stack[0]);
    }
    stack[-1] = a;
    b = new_bmodular();
    bmodular_value(b) = stack[-1];
    bmodular_modulus(b) = stack[0];
    popv(2);
    return b;
}



Sex
diff2Mb(Sex a, Sex b)
{
    return diff2Mi(a, b);
}

Sex
diff2Mr(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return diff2MM(a, b);
}

Sex
diff2rM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return diff2MM(a, b);
}

Sex
diff2MM(Sex a, Sex b)
{
    Sex             m;

    m = bmodular_modulus(a);
    push(m);
    if (!numeq2(m, bmodular_modulus(b))) {
	popv(1);
	return math_er(a, b);
    }
    a = diff2jj(bmodular_value(a), bmodular_value(b));
    m = stack[0];
    if (minusp(a))
	a = plus2(a, stack[0]);
    push(a);
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
times2iM(Sex a, Sex b)
{
    push(bmodular_modulus(b));
    a = times2(a, bmodular_value(b));
    a = remaind2(a, stack[0]);
    if (minusp(a))
	a = plus2(a, stack[0]);
    push(a);
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}


Sex
times2bM(Sex a, Sex b)
{
    return times2iM(a, b);
}


Sex
times2Mi(Sex a, Sex b)
{
    return times2iM(b, a);
}


Sex
times2Mb(Sex a, Sex b)
{
    return times2iM(b, a);
}

Sex
times2Mr(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return times2MM(a, b);
}

Sex
times2rM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return times2MM(a, b);
}


Sex
times2MM(Sex a, Sex b)
{
    Sex             m;

    m = bmodular_modulus(a);
    if (!numeq2(m, bmodular_modulus(b))) {
	return math_er(a, b);
    }
    push(m);
    a = times2jj(bmodular_value(a), bmodular_value(b));
    a = remaind2(a, stack[0]);
    push(a);
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
quot2iM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return quot2MM(a, b);
}


Sex
quot2bM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return quot2MM(a, b);
}


Sex
quot2Mi(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return quot2MM(a, b);
}


Sex
quot2Mb(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return quot2MM(a, b);
}

Sex
quot2Mr(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return quot2MM(a, b);
}

Sex
quot2rM(Sex a, Sex b)
{
    push(b);
    a = Lbmodular(a, bmodular_modulus(b));
    pop(b);
    return quot2MM(a, b);
}

Sex
recipiM(Sex n, Sex m)
{
    Sex             c;

    push4(m, n, m, n);
    push4(mkfxat(0), mkfxat(1), nil, nil);
#define a stack[-7]
#define b stack[-6]
#define m stack[-5]
#define n stack[-4]
#define x stack[-3]
#define y stack[-2]
#define t stack[-1]
#define w stack[-0]
    while (!onep(b)) {
	if (zerop(b)) {
	    popv(7);
	    return aerror1("recipiM ", mkfxat(n));
	}
	w = Ldivide(a, b);
	t = b;
	b = cdr(w);
	a = t;
	t = y;
	w = times2(y, car(w));
	y = diff2(x, w);
	/*
	 * y = diff2(x, times2(y, car(w))); 
	 */
	x = t;
    }
    if (minusp(y))
	y = plus2(y, m);
    c = y;
#undef a
#undef b
#undef m
#undef n
#undef x
#undef y
#undef t
#undef w
    popv(8);
    return c;
}


Sex
recipM(Sex n)
{
    Sex             a = value(star_bmodulus);
    if (!is_number(a))
	return aerror1("recipM", a);
    push4(value(star_bmodulus), n, mkfxat(0), mkfxat(1));
    push2(nil, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define y stack[-2]
#define w stack[-1]
#define t stack[0]
    while (b != mkfxat(1)) {
	if (b == mkfxat(0))
	    return aerror1("modular-reciprocal", n);
	w = Ldivide(a, b);
	a = b;
	b = cdr(w);		/* 
				 * remainder 
				 */
	t = y;
	w = times2jj(y, car(w));
	y = diff2jj(x, w);
	x = t;
    }
    if (minusp(y))
	y = plus2jj(y, value(star_bmodulus));
    n = y;
#undef a
#undef b
#undef x
#undef y
#undef w
#undef t
    popv(6);
    return n;
}


Sex
quot2MM(Sex a, Sex b)
{
    Sex             m;

    m = bmodular_modulus(a);
    if (!numeq2jj(m, bmodular_modulus(b))) {
	return math_er(a, b);
    }
    push2(m, a);
    b = recipiM(bmodular_value(b), m);
    a = stack[0];
    a = times2jj(bmodular_value(a), b);
    a = rem2jj(a, stack[-1]);
    stack[0] = a;
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}


/***************************************************************/
/***************************************************************/


Sex
negateM(Sex a)
{
    push(bmodular_modulus(a));
    a = negate(bmodular_value(a));
    push(a);
    a = new_bmodular();
    bmodular_value(a) = stack[0];
    bmodular_modulus(a) = stack[-1];
    popv(2);
    return a;
}
/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2iM(Sex a, Sex b)
{
    return numeq2jj(a, bmodular_value(b));
}

bool
numeq2bM(Sex a, Sex b)
{
    return numeq2jj(a, bmodular_value(b));
}

bool
numeq2MM(Sex a, Sex b)
{
    return numeq2jj(bmodular_value(a), bmodular_value(b));
}

bool
numeq2Mr(Sex a, Sex b)
{
    push(a);
    b = Lbmodular(b, bmodular_modulus(a));
    pop(a);
    return numeq2jj(bmodular_value(a), bmodular_value(b));
}

bool
numeq2Mc(Sex a, Sex b)
{
    if (!zerop(complex_imag(b)))
	return NO;
    return numeq2(a, complex_real(b));
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
Lbmodular_difference(Sex a, Sex b)
{
    a = diff2jj(a, b);
    if (minusp(a))
	a = plus2jj(a, current_big_modulus);
    return a;
}

Sex
Lbmodular_minus(Sex a)
{
    a = diff2jj(current_big_modulus, a);
    return a;
}

Sex
Lbmodular_number(Sex a)
{
    a = rem2jj(a, current_big_modulus);
    if (minusp(a))
	a = plus2jj(a, current_big_modulus);
    return a;
}

Sex
Lbmodular_plus(Sex a, Sex b)
{
    a = plus2jj(a, b);
    a = rem2jj(a, current_big_modulus);
    return a;
}

Sex
Lbmodular_reciprocal(Sex n)
{

    n = recipiM(mkint(n), current_big_modulus);
    return n;
}



Sex
Lbmodular_times(Sex a, Sex b)
{
    a = times2jj(a, b);
    a = rem2jj(a, current_big_modulus);
    return a;
}

Sex
Lbmodular_quotient(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = Lbmodular_reciprocal(b);
    pop(a);
    return Lbmodular_times(a, b);
}

Sex
Lbmodular_expt(Sex a, Sex b)
{
    Sex             p;
    p = mkfxat(1);
    push3(a, b, p);
#define a stack[-2]
#define b stack[-1]
#define p stack[-0]
    while (Lgreaterp(b, mkfxat(1)) != nil) {
	if (Levenp(b) != nil) {
	    p = times2jj(a, p);
	    p = rem2jj(p, current_big_modulus);
	}
	a = times2jj(a, a);
	a = rem2jj(a, current_big_modulus);
	b = leftshift(b, 1);
    }
    a = times2jj(a, p);
    a = rem2jj(a, current_big_modulus);
#undef a
#undef b
#undef p
    a = stack[-2];
    popv(3);
    return a;
}



Sex
Lset_modulus(Sex a)
{
    long            r,
                    old = current_modulus;
    if (!is_fixnum(a))
	return aerror1("set-small-modulus", a);
    r = mkint(a);
    current_modulus = r;
    return mkfxat(old);
}

Sex
Lset_big_modulus(Sex a)
{
    Sex             old = current_big_modulus;
    current_big_modulus = a;
    return old;
}

/***************************************************************/
/***************************************************************/

setup_type const bmodular_setup[] = {
    {"bmodular-difference", (lisp_proc *) Lbmodular_difference, 2},
    {"bmodular-minus", (lisp_proc *) Lbmodular_minus, 1},
    {"bmodular-number", (lisp_proc *) Lbmodular_number, 1},
    {"bmodular-plus", (lisp_proc *) Lbmodular_plus, 2},
    {"bmodular-quotient", (lisp_proc *) Lbmodular_quotient, 2},
    {"bmodular-reciprocal", (lisp_proc *) Lbmodular_reciprocal, 1},
    {"bmodular-times", (lisp_proc *) Lbmodular_times, 2},
    {"bmodular-expt", (lisp_proc *) Lbmodular_expt, 2},
    {"set-modulus", (lisp_proc *) Lset_modulus, 1},
    {"set-big-modulus", (lisp_proc *) Lset_big_modulus, 1},
    {NULL, 0, 0}
};
