/*
 * complex.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/complex.c,v 2.7 2014/10/27 02:28:08 kako Exp $
 */

static char const rcsid[] =
    "$Id: complex.c,v 2.7 2014/10/27 02:28:08 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "complex.h"


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
is_shortfloat(Sex a)
{
    Sex             ar,
                    ai;
    ar = complex_real(a);
    ai = complex_imag(a);
    if (is_numhdr(ar)) {
	if ((numhdr(ar) & (T_APP | T_BIG)) == T_APP)
	    return YES;
    }
    if (is_numhdr(ai)) {
	if ((numhdr(ai) & (T_APP | T_BIG)) == T_APP)
	    return YES;
    }
    return NO;
}

bool
is_longfloat(Sex a)
{
    Sex             ar,
                    ai;
    ar = complex_real(a);
    ai = complex_imag(a);
    if (is_numhdr(ar)) {
	if (!(numhdr(ar) & T_APP) || (numhdr(ar) & T_BIG))
	    return YES;
    }
    if (is_numhdr(ai)) {
	if (!(numhdr(ai) & T_APP) || (numhdr(ai) & T_BIG))
	    return YES;
    }
    return NO;
}


Sex
Lcomplex_conjugate(Sex u)
{
    Sex             v;
    if (is_fixnum(u))
	return u;
    if (!is_numhdr(u))
	aerror("complelx_conjugate: not number");
    switch (numtype_of(u)) {
    case N_COMPLEX:
	stackcheck();
	push(u);
	v = new_complex();
	u = stack[0];
	stack[0] = v;
	complex_real(v) = complex_real(u);
	complex_imag(stack[0]) = negate(complex_imag(u));
	pop(v);
	return v;
    case N_ECOMPLEX:
	return make_ecomplex(ecomplex_real(u), -ecomplex_imag(u),
			     ecomplex_error(u));
    case N_BECOMPLEX:
	{
	    Sex             ur,
	                    ui;
	    bf_error_type   ue;
	    ue = becomplex_error(u);
	    ur = becomplex_real(u);
	    ui = becomplex_imag(u);
	    stackcheck();
	    push(ur);
	    ui = negate(ui);
	    pop(ur);
	    return make_becomplex(ur, ui, ue);
	}
    default:
	return u;
    }
}

Sex
Lcomplex_real(Sex a)
{
    if (is_fixnum(a))
	return a;
    if (!is_numhdr(a))
	aerror("complelx real: not number");
    switch (numtype_of(a)) {
    case N_COMPLEX:
	return complex_real(a);
    case N_ECOMPLEX:
	return make_efloat(ecomplex_real(a), ecomplex_error(a));
    case N_BECOMPLEX:
	return make_befloat(becomplex_real(a), becomplex_error(a));
    default:
	return a;
    }
}

Sex
Lcomplex_imag(Sex a)
{
    if (is_fixnum(a))
	return mkfxat(0);
    if (!is_numhdr(a))
	aerror("complelx imag: not number");
    switch (numtype_of(a)) {
    case N_COMPLEX:
	return complex_imag(a);
    case N_ECOMPLEX:
	return make_efloat(ecomplex_imag(a), ecomplex_error(a));
    case N_BECOMPLEX:
	return make_befloat(becomplex_imag(a), becomplex_error(a));
    default:
	return mkfxat(0);
    }
}


Sex
Lcomplex(Sex a)
{
    Sex             c;
    if (is_fixnum(a)) {
	c = new_complex();
	complex_real(c) = a;
	complex_imag(c) = mkfxat(0);
	return c;
    } else if (is_numhdr(a)) {
	stackcheck();
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_EFLOAT:
	case N_INTERVAL:
	case N_RATIONAL:
	case N_FLOAT:
	case N_BFLOAT:
	case N_BINTERVAL:
	case N_BEFLOAT:
	    stackcheck();
	    push(a);
	    c = new_complex();
	    pop(a);
	    complex_real(c) = a;
	    complex_imag(c) = mkfxat(0);
	    return c;
	case N_COMPLEX:
	    return a;
	case N_ECOMPLEX:
	    {
		double          r,
		                i,
		                e;
		Sex             b;
		r = ecomplex_real(a);
		i = ecomplex_imag(a);
		e = ecomplex_error(a);
		a = make_efloat(r, e);
		push(a);
		b = make_efloat(i, e);
		push(b);
		c = new_complex();
		pop2(b, a);
		complex_real(c) = a;
		complex_imag(c) = b;
		return c;
	    }
#if 0
	    stackcheck();
	    push2(a, nil);
	    c = make_float(ecomplex_real(a));
	    stack[-0] = c;
	    a = stack[-1];
	    c = make_float(ecomplex_imag(a));
	    stack[-1] = c;
	    c = new_complex();
	    complex_real(c) = stack[-0];
	    complex_imag(c) = stack[-1];
	    popv(2);
	    return c;
#endif
	case N_BECOMPLEX:
	    {
		Sex             r,
		                i;
		double          e;
		r = becomplex_real(a);
		i = becomplex_imag(a);
		e = becomplex_error(a);
		r = make_befloat(r, e);
		push(r);
		i = make_befloat(i, e);
		push(i);
		c = new_complex();
		pop2(i, r);
		complex_real(c) = r;
		complex_imag(c) = i;
		return c;
	    }
#if 0
	    stackcheck();
	    push(a);
	    c = new_complex();
	    pop(a);
	    complex_real(c) = becomplex_real(a);
	    complex_imag(c) = becomplex_imag(a);
	    return c;
#endif
	}
    }
    return aerror("complelx: not number");
}

Sex
Lmkcomplex(Sex r, Sex i)
{
    Sex             v;
    if (i == mkfxat(0))
	return r;
    stackcheck();
    push2(r, i);
    v = new_complex();
    pop2(i, r);
    complex_real(v) = r;
    complex_imag(v) = i;
    return v;
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2xc(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = plus2(a, complex_real(b));
    pop(b);
    b = complex_imag(b);
    return Lmkcomplex(a, b);
}

Sex
plus2cx(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = plus2(b, complex_real(a));
    pop(a);
    a = complex_imag(a);
    return Lmkcomplex(b, a);
}

Sex
plus2mc(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = plus2(a, complex_real(b));
    b = stack[0];
    stack[0] = a;
    b = Lmkmodular(complex_imag(b));
    pop(a);
    return Lmkcomplex(a, b);
}


Sex
plus2cc(Sex a, Sex b)
{
    Sex             c,
                    d;
    stackcheck();
    push2(a, b);
    c = plus2(complex_real(a), complex_real(b));
    b = stack[0];
    a = stack[-1];
    stack[0] = c;
    d = plus2(complex_imag(a), complex_imag(b));
    c = stack[0];
    popv(2);
    return Lmkcomplex(c, d);
}

Sex
plus2Mc(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = Lcomplex(b);
    pop(a);
    return plus2cc(a, b);
}

Sex
plus2cM(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = Lcomplex(a);
    pop(b);
    return plus2cc(a, b);
}




/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
diff2xc(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = diff2(a, complex_real(b));
    pop(b);
    b = complex_imag(b);
    return Lmkcomplex(a, b);
}

Sex
diff2cx(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = diff2(complex_real(a), b);
    pop(a);
    a = complex_imag(a);
    return Lmkcomplex(b, a);
}
Sex
diff2mc(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = diff2(a, complex_real(b));
    b = stack[0];
    stack[0] = a;
    b = Lmkmodular(negate(complex_imag(b)));
    pop(a);
    return Lmkcomplex(a, b);
}

Sex
diff2cm(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = diff2(complex_real(a), b);
    a = stack[0];
    stack[0] = b;
    a = Lmkmodular(complex_imag(a));
    pop(b);
    return Lmkcomplex(a, b);
}

Sex
diff2cc(Sex a, Sex b)
{
    Sex             c,
                    d;
    stackcheck();
    push2(a, b);
    c = diff2(complex_real(a), complex_real(b));
    b = stack[0];
    a = stack[-1];
    stack[0] = c;
    d = diff2(complex_imag(a), complex_imag(b));
    c = stack[0];
    popv(2);
    return Lmkcomplex(c, d);
}

Sex
diff2Mc(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = Lcomplex(b);
    pop(a);
    return diff2cc(a, b);
}

Sex
diff2cM(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = Lcomplex(a);
    pop(b);
    return diff2cc(a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
times2cx(Sex a, Sex b)
{
    return times2xc(b, a);
}

Sex
times2xc(Sex a, Sex b)
{
    Sex             r = complex_real(b),
	i = complex_imag(b);
    stackcheck();
    push2(a, r);
    i = times2(a, i);
    r = stack[0];
    a = stack[-1];
    stack[0] = i;
    r = times2(a, r);
    i = stack[0];
    popv(2);
    return Lmkcomplex(r, i);
}

Sex
times2mc(Sex a, Sex b)
{
    stackcheck();
    push3(a, b, nil);
    stack[0] = times2(a, complex_real(b));
    b = times2(stack[-2], complex_imag(stack[-1]));
    a = stack[0];
    popv(3);
    return Lmkcomplex(a, b);
}

Sex
times2cc(Sex a, Sex b)
{
    Sex             w = nil;
    stackcheck();
    push4(complex_real(a), complex_imag(a),
	  complex_real(b), complex_imag(b));
    push2(nil, nil);
#define u   stack[0]
#define v   stack[-1]
#define ib  stack[-2]
#define rb  stack[-3]
#define ia  stack[-4]
#define ra  stack[-5]
    u = times2(ra, rb);
    v = times2(ia, ib);
    v = negate(v);
    u = plus2(u, v);		/* 
				 * real part of result 
				 */
    v = times2(ra, ib);
    ib = times2(rb, ia);
    v = plus2(v, ib);		/* 
				 * imaginary part 
				 */
    w = Lmkcomplex(u, v);
    popv(6);
    return w;
#undef u
#undef v
#undef ib
#undef rb
#undef ia
#undef ra
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
quot2xc(Sex a, Sex b)
{
    stackcheck();
    push5(a, b, nil, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define u stack[-1]
#define v stack[0]
    /*
     * a / (p + iq) is : (a * (p - iq)) / (p^2 + q^2)
     */
    u = complex_real(b);
    u = times2(u, u);
    v = complex_imag(b);
    v = times2(v, v);
    c = plus2(u, v);
    c = quot2(a, c);
    u = times2(complex_real(b), c);
    v = times2(complex_imag(b), c);
    v = negate(v);
#undef u
#undef v
#undef c
#undef a
#undef b
    a = stack[-1];
    b = stack[0];
    popv(5);
    return Lmkcomplex(a, b);
}



Sex
quot2cx(Sex a, Sex b)
{
    Sex             r = complex_real(a),
	i = complex_imag(a);
    /*
     * (r + i) / a is : (r/a) + (i/a)
     */
    if (zerop(b))
	return aerror2("bad arg for quotient", a, b);
    stackcheck();
    push2(b, r);
    i = quot2(i, b);
    pop2(r, b);
    push(i);
    r = quot2(r, b);
    pop(i);
    return Lmkcomplex(r, i);
}

Sex
quot2cc(Sex a, Sex b)
{
    Sex             w = nil;
    stackcheck();
    push4(complex_real(a), complex_imag(a),
	  complex_real(b), complex_imag(b));
    push3(nil, nil, nil);
#define r   stack[0]
#define u   stack[-1]
#define v   stack[-2]
#define ib  stack[-3]
#define rb  stack[-4]
#define ia  stack[-5]
#define ra  stack[-6]
    r = times2(rb, rb);
    u = times2(ib, ib);
    r = plus2(r, u);		/* r = |b|^2 */
    u = times2(ra, rb);
    v = times2(ia, ib);
    u = plus2(u, v);		/* 
				 * real part of result 
				 * ra*rb+ia*ib 
				 */
    u = quot2(u, r);		/* (ra*rb+ia*ib)/r */
    v = times2(ra, ib);
    ib = times2(rb, ia);
    v = diff2(ib, v);		/* 
				 * imaginary part 
				 * -ra*ib + rb*ia
				 */
    v = quot2(v, r);
    w = Lmkcomplex(u, v);
    popv(7);
    return w;
#undef r
#undef u
#undef v
#undef ib
#undef rb
#undef ia
#undef ra
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
rem2cc(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = quot2cc(a, b);
    pop(b);
    c = times2(b, c);
    pop(a);
    c = diff2(a, c);
    return c;
}

Sex
rem2xc(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = quot2xc(a, b);
    pop(b);
    c = times2(b, c);
    pop(a);
    c = diff2(a, c);
    return c;
}


Sex
rem2cx(Sex a, Sex b)
{
    Sex             r = complex_real(a),
	i = complex_imag(a);
    /*
     * (r + i) / a is : (r/a) + (i/a)
     */
    if (zerop(b))
	return aerror2("bad arg for remaind", a, b);
    stackcheck();
    push(r);
    i = remaind2(i, b);
    pop(r);
    if (zerop(i))
	return remaind2(r, b);
    push(i);
    r = remaind2(r, b);
    pop(i);
    return Lmkcomplex(r, i);
}


#if 0
Sex
rem2cc(Sex a, Sex b)
{
    Sex             w = nil;
    stackcheck();
    push4(complex_real(a), complex_imag(a),
	  complex_real(b), complex_imag(b));
    push3(nil, nil, nil);
#define r   stack[0]
#define u   stack[-1]
#define v   stack[-2]
#define ib  stack[-3]
#define rb  stack[-4]
#define ia  stack[-5]
#define ra  stack[-6]
    r = times2(rb, rb);
    u = times2(ib, ib);
    r = plus2(r, u);
    u = times2(ra, rb);
    v = times2(ia, ib);
    u = plus2(u, v);		/* 
				 * real part of result 
				 */
    u = remaind2(u, r);
    v = times2(ra, ib);
    v = negate(v);
    ib = times2(rb, ia);
    v = plus2(v, ib);		/* 
				 * imaginary part 
				 */
    v = remaind2(v, r);
    w = Lmkcomplex(u, v);
    popv(7);
    return w;
#undef r
#undef u
#undef v
#undef ib
#undef rb
#undef ia
#undef ra
}


Sex
rem2cx(Sex a, Sex b)
{
    Sex             r = complex_real(a),
	i = complex_imag(a);
    /*
     * (r + i) / a is : (r/a) + (i/a)
     */
    if (zerop(b))
	return aerror2("bad arg for remaind", a, b);
    stackcheck();
    push(r);
    i = remaind2(i, b);
    pop(r);
    if (zerop(i))
	return remaind2(r, b);
    push(i);
    r = remaind2(r, b);
    pop(i);
    return Lmkcomplex(r, i);
}

Sex
rem2xc(Sex a, Sex b)
{
    Sex             w;
    stackcheck();
    push3(a, complex_real(b), complex_imag(b));
    push3(nil, nil, nil);
#define r   stack[0]
#define u   stack[-1]
#define v   stack[-2]
#define ib  stack[-3]
#define rb  stack[-4]
#define ra  stack[-5]
    r = times2(rb, rb);
    u = times2(ib, ib);
    r = plus2(r, u);
    u = times2(ra, rb);		/* 
				 * real part of result 
				 */
    u = remaind2(u, r);
    v = times2(ra, ib);
    v = negate(v);		/* 
				 * imaginary part 
				 */
    v = remaind2(v, r);
    w = Lmkcomplex(u, v);
    popv(6);
    return w;
#undef r
#undef u
#undef v
#undef ib
#undef rb
#undef ra
}
#endif

/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
lessp2cx(Sex a, Sex b)
{
    bool            r;
    stackcheck();
    push2(a, b);
    r = lessp2(complex_real(a), b);
    if (r) {
	popv(2);
	return YES;
    }
    pop(b);
    a = stack[0];
    r = numeq2(complex_real(a), b);
    pop(a);
    if (r)
	return minusp(complex_imag(a));
    return NO;
}

bool
lessp2xc(Sex a, Sex b)
{
    bool            r;
    stackcheck();
    push2(b, a);
    r = lessp2(a, complex_real(b));
    if (r) {
	popv(2);
	return YES;
    }
    pop(a);
    b = stack[0];
    r = numeq2(a, complex_real(b));
    pop(b);
    if (r)
	return minusp(complex_imag(b));
    return NO;
}

bool
lessp2cc(Sex a, Sex b)
{
    bool            r;
    stackcheck();
    push2(a, b);
    r = lessp2(complex_real(a), complex_real(b));
    if (r) {
	popv(2);
	return YES;
    }
    a = stack[-1];
    b = stack[0];
    r = numeq2(complex_real(a), complex_real(b));
    pop2(b, a);
    if (r)
	return lessp2(complex_imag(a), complex_imag(b));
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
numeq2xc(Sex a, Sex b)
{
    if (!zerop(complex_imag(b)))
	return NO;
    else
	return numeq2(a, complex_real(b));
}

bool
numeq2cx(Sex a, Sex b)
{
    if (!zerop(complex_imag(a)))
	return NO;
    else
	return numeq2(b, complex_real(a));
}


bool
numeq2cc(Sex a, Sex b)
{
    bool            r;
    stackcheck();
    push2(a, b);
    r = numeq2(complex_real(a), complex_real(b));
    if (!r) {
	popv(2);
	return NO;
    }
    pop2(b, a);
    return numeq2(complex_imag(a), complex_imag(b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
appeq2xc(Sex a, Sex b)
{
    if (!zerop(complex_imag(b)))
	return NO;
    else
	return appeq2(a, complex_real(b));
}

bool
appeq2cx(Sex a, Sex b)
{
    if (!zerop(complex_imag(a)))
	return NO;
    else
	return appeq2(b, complex_real(a));
}


bool
appeq2cc(Sex a, Sex b)
{
    bool            r;
    stackcheck();
    push2(a, b);
    r = appeq2(complex_real(a), complex_real(b));
    if (!r) {
	popv(2);
	return NO;
    }
    pop2(b, a);
    return appeq2(complex_imag(a), complex_imag(b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
