/*
 * ecomplex.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * effective complex number
 *      $Header: /Users/kako/asuka/src/RCS/ecomplex.c,v 2.10 2014/10/27 04:45:50 kako Exp $
 */

static char const rcsid[] =
    "$Id: ecomplex.c,v 2.10 2014/10/27 04:45:50 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>


#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"

bool            numeq2zc(Sex b, Sex a);
bool            appeq2zc(Sex b, Sex a);

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
Lecomplex_real(Sex a)
{
    if (is_fixnum(a))
	return a;
    if (!is_numhdr(a))
	aerror("complelx real: not number");
    if (is_complex(a))
	return complex_real(a);
    if (is_ecomplex(a))
	return make_float(ecomplex_real(a));
    if (is_becomplex(a))
	return becomplex_real(a);
    return a;
}

Sex
Lecomplex_imag(Sex a)
{
    if (is_fixnum(a))
	return mkfxat(0);
    if (!is_numhdr(a))
	aerror("complelx imag: not number");
    if (is_complex(a))
	return complex_imag(a);
    if (is_ecomplex(a))
	return make_float(ecomplex_imag(a));
    if (is_becomplex(a))
	return becomplex_imag(a);
    return mkfxat(0);
}

Sex
Lecomplex_value(Sex a)
{
    Sex             c,
                    b;
    double          r,
                    i;

    if (is_fixnum(a))
	return a;
    if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_INTERVAL:
	case N_RATIONAL:
	case N_FLOAT:
	case N_BFLOAT:
	case N_BINTERVAL:
	case N_COMPLEX:
	    return a;
	case N_EFLOAT:
	    return make_float(efloat_value(a));
	case N_BEFLOAT:
	    return befloat_value(a);
	case N_ECOMPLEX:
	    r = ecomplex_real(a);
	    i = ecomplex_imag(a);
	    a = make_float(r);
	    stackcheck();
	    push(a);
	    b = make_float(i);
	    push(b);
	    c = new_complex();
	    pop2(b, a);
	    complex_imag(c) = b;
	    complex_real(c) = a;
	    return c;
	case N_BECOMPLEX:
	    stackcheck();
	    push(a);
	    c = new_complex();
	    pop(a);
	    complex_real(c) = becomplex_real(a);
	    complex_imag(c) = becomplex_imag(a);
	    return c;
	}
    }
    return aerror("ecomplex_value: not number");
}

Sex
Lecomplex_error(Sex a)
{
    double          d,
                    r;

    if (is_fixnum(a))
	return mkfxat(0);
    if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_RATIONAL:
	case N_INTERVAL:
	case N_BINTERVAL:
	case N_COMPLEX:
	    return mkfxat(0);
	case N_FLOAT:
	case N_BFLOAT:
	    return make_float(0.0);
	case N_EFLOAT:
	    return make_float(efloat_error(a));
	case N_BEFLOAT:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    return make_float(befloat_error(a));
#else
	    d = floatit(befloat_value(a));
	    d = fabs(d);
	    d = ldexp(d, befloat_error(a));
	    return make_float(d);
#endif
	case N_ECOMPLEX:
	    return make_float(ecomplex_error(a));
	case N_BECOMPLEX:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    return make_float(becomplex_error(a));
#else
	    r = fabs(floatit(becomplex_real(a)));
	    d = fabs(floatit(becomplex_imag(a)));
	    if (d < r)
		d = r;
	    d = ldexp(d, becomplex_error(a));

	    return make_float(d);
#endif
	}
    }
    return aerror("ecomplex_error: not number");
}

Sex
Lecomplex(Sex a)
{
    double          r,
                    i;
    double          e = EFLOAT_ERROR;
    if (is_fixnum(a)) {
	r = (double) mkint(a);
	return make_ecomplex(r, 0.0, fabs(r) * e);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    r = floatitb(a);
	    return make_ecomplex(r, 0.0, fabs(r) * e);
	case N_BIGNUM:
	    r = floatitb(a);
	    return make_ecomplex(r, 0.0, fabs(r) * e);
	case N_EFLOAT:
	    return make_ecomplex(efloat_value(a), 0.0, efloat_error(a));
	case N_INTERVAL:
	    return make_ecomplex((interval_lower(a) + interval_upper(a)) *
				 0.5, 0.0,
				 (interval_upper(a) -
				  interval_lower(a)) * 0.5);
	case N_RATIONAL:
	    r = floatit(rational_num(a)) / floatit(rational_den(a));
	    return make_ecomplex(r, 0.0, fabs(r) * e);
	case N_FLOAT:
	    r = float_val(a);
	    return make_ecomplex(r, 0.0, fabs(r) * e);
	case N_COMPLEX:
	    return ecomplexitc(a);
	case N_BFLOAT:
	    r = floatitd(a);
	    return make_ecomplex(r, 0.0, fabs(r) * e);
	case N_ECOMPLEX:
	    return a;
	case N_BINTERVAL:
	    return ecomplexitV(a);
	case N_BEFLOAT:
	    return ecomplexitE(a);
	case N_BECOMPLEX:
	    return ecomplexitZ(a);
	default:
	    return make_ecomplex(0.0, 0.0, 0.0);
	}
    } else {
	return make_ecomplex(0.0, 0.0, 0.0);
    }
}

Sex
Lmkecomplex(Sex r, Sex i, Sex e)
{
    Sex             a;
    double          dr,
                    di,
                    de;
    dr = floatit(r);
    di = floatit(i);
    de = floatit(e);
    a = new_ecomplex();
    ecomplex_real(a) = dr;
    ecomplex_imag(a) = di;
    ecomplex_error(a) = fabs(de);
    return a;
}


Sex
ecomplexitc(Sex a)
{
    Sex             w;
    double          r,
                    i;
    double          c,
                    e;

    r = floatit(complex_real(a));
    i = floatit(complex_imag(a));
    if (is_fixnum(complex_real(a)))
	c = 0;
    else
	c = fabs(r);
    if (is_fixnum(complex_imag(a)))
	e = 0;
    else
	e = fabs(i);
    if (e < c)
	e = c;
    e = e * EFLOAT_ERROR;
    w = new_ecomplex();
    ecomplex_real(w) = r;
    ecomplex_imag(w) = i;
    ecomplex_error(w) = e;
    return w;

}

Sex
ecomplexitE(Sex a)
{
    Sex             w;
    double          v;
    bf_error_type   e;

    v = floatit(befloat_value(a));
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    e = befloat_error(a);
#else
    e = fabs(v);
    e = ldexp(e, befloat_error(a));
#endif
    w = new_ecomplex();
    ecomplex_real(w) = v;
    ecomplex_imag(w) = 0.0;
    ecomplex_error(w) = e;
    return w;

}

Sex
ecomplexitV(Sex a)
{
    Sex             w;
    double          u,
                    l;

    u = floatitd(binterval_upper(a));
    l = floatitd(binterval_lower(a));
    w = new_ecomplex();
    ecomplex_real(w) = (u + l) * 0.5;
    ecomplex_imag(w) = 0.0;
    ecomplex_error(w) = (u - l) * 0.5;
    return w;

}

Sex
ecomplexitZ(Sex a)
{
    Sex             w;
    double          r,
                    i;
    bf_error_type   e;

    r = floatitd(becomplex_real(a));
    i = floatitd(becomplex_imag(a));
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    e = becomplex_error(a);
#else
    e = fabs(r);
    if (e < fabs(i))
	e = fabs(i);
    e = ldexp(e, becomplex_error(a));
#endif
    w = new_ecomplex();
    ecomplex_real(w) = r;
    ecomplex_imag(w) = i;
    ecomplex_error(w) = e;
    return w;

}


Sex
make_ecomplex(double r, double i, double e)
{
    Sex             a;
    a = new_ecomplex();
    ecomplex_real(a) = r;
    ecomplex_imag(a) = i;
    ecomplex_error(a) = e;
    return a;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
plus2iz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = floatit(a) + ecomplex_real(b);
    fi = ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}


Sex
plus2zx(Sex b, Sex a)
{
    double          fr,
                    fi,
                    fe;
    fr = floatit(a) + ecomplex_real(b);
    fi = ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}

Sex
plus2xz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe,
                    f;
    fr = ecomplex_real(b);
    fi = ecomplex_imag(b);
    fe = ecomplex_error(b);
    f = floatit(a);
    return make_ecomplex(fr + f, fi, fe);
}


Sex
plus2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return plus2zz(a, b);
#else
    Sex             ar,
                    ai;
    ar = complex_real(a);
    ai = complex_imag(a);
#if TOWARD_LOWER_PRECISION
    if (!is_big(ar) || !is_big(ai)) {
	a = ecomplexitc(a);
	return plus2zz(a, b);
    } else {
	a = becomplexitc(a);
	return plus2zZ(b, a);
    }
#else
    if (is_big(ar) || is_big(ai)) {
	a = becomplexitc(a);
	return plus2zZ(b, a);
    } else {
	a = ecompexitc(a);
	return plus2zz(a, b);
    }
#endif
#endif
}


Sex
plus2fz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = float_val(a) + ecomplex_real(b);
    fi = ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}


Sex
plus2ez(Sex a, Sex b)
{
    double          fr;
    fr = efloat_value(a) + ecomplex_real(b);
    if (efloat_error(a) < ecomplex_error(b))
	return make_ecomplex(fr, ecomplex_imag(b), ecomplex_error(b));
    else
	return make_ecomplex(fr, ecomplex_imag(b), efloat_error(a));
}


Sex
plus2vz(Sex a, Sex b)
{
    double          fr,
                    wv,
                    we;

    wv = (interval_lower(a) + interval_upper(a)) / 2;
    we = (interval_lower(a) - interval_upper(a)) / 2;

    fr = wv + ecomplex_real(b);
    if (we < ecomplex_error(b))
	return make_ecomplex(fr, ecomplex_imag(b), ecomplex_error(b));
    else
	return make_ecomplex(fr, ecomplex_imag(b), we);
}


Sex
plus2zz(Sex a, Sex b)
{
    double          fr,
                    fi;
    fr = ecomplex_real(a) + ecomplex_real(b);
    fi = ecomplex_imag(a) + ecomplex_imag(b);
    if (ecomplex_error(a) < ecomplex_error(b))
	return make_ecomplex(fr, fi, ecomplex_error(b));
    else
	return make_ecomplex(fr, fi, ecomplex_error(a));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
diff2iz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = mkint(a) - ecomplex_real(b);
    fi = -ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}


Sex
diff2zi(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = ecomplex_real(a) - mkint(b);
    fi = ecomplex_imag(a);
    fe = ecomplex_error(a);
    return make_ecomplex(fr, fi, fe);
}


Sex
diff2fz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = float_val(a) - ecomplex_real(b);
    fi = -ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}


Sex
diff2zf(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = ecomplex_real(a) - float_val(b);
    fi = ecomplex_imag(a);
    fe = ecomplex_error(a);
    return make_ecomplex(fr, fi, fe);
}

Sex
diff2xz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = floatit(a) - ecomplex_real(b);
    fi = -ecomplex_imag(b);
    fe = ecomplex_error(b);
    return make_ecomplex(fr, fi, fe);
}


Sex
diff2zx(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe;
    fr = ecomplex_real(a) - floatit(b);
    fi = ecomplex_imag(a);
    fe = ecomplex_error(a);
    return make_ecomplex(fr, fi, fe);
}


Sex
diff2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return diff2zz(a, b);
#else
    Sex             ar,
                    ai;
    ar = complex_real(a);
    ai = complex_imag(a);
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (!is_big(ar) || !is_big(ai)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return diff2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return diff2Zz(a, b);
    }
#else
    if (is_big(ar) || is_big(ai)) {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return diff2Zz(a, b);
    } else {
	push(b);
	a = ecompexitc(a);
	pop(b);
	return diff2zz(a, b);
    }
#endif
#endif
}

Sex
diff2zc(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(a);
    b = ecomplexitc(b);
    pop(a);
    return diff2zz(a, b);
#else
    Sex             br,
                    bi;
    br = complex_real(b);
    bi = complex_imag(b);
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (!is_big(br) || !is_big(bi)) {
	push(a);
	b = ecomplexitc(b);
	pop(a);
	return diff2zz(a, b);
    } else {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return diff2zZ(a, b);
    }
#else
    if (is_big(br) || is_big(bi)) {
	push(a);
	b = becomplexitc(b);
	pop(a);
	return diff2zZ(a, b);
    } else {
	push(a);
	b = ecompexitc(b);
	pop(a);
	return diff2zz(a, b);
    }
#endif
#endif
}



Sex
diff2ez(Sex a, Sex b)
{
    double          fr;
    double          e;

    fr = efloat_value(a) - ecomplex_real(b);
    e = efloat_error(a);
    if (e < ecomplex_error(b))
	e = ecomplex_error(b);
    return make_ecomplex(fr, -ecomplex_imag(b), e);
}

Sex
diff2ze(Sex a, Sex b)
{
    double          fr;
    double          e;

    fr = ecomplex_real(a) - efloat_value(b);
    e = efloat_error(b);
    if (e < ecomplex_error(a))
	e = ecomplex_error(a);
    return make_ecomplex(fr, ecomplex_imag(a), e);
}


Sex
diff2vz(Sex a, Sex b)
{
    double          fr,
                    wv,
                    we;

    wv = (interval_lower(a) + interval_upper(a)) / 2;
    we = (interval_lower(a) - interval_upper(a)) / 2;

    fr = wv - ecomplex_real(b);
    if (we < ecomplex_error(b))
	we = ecomplex_error(b);
    return make_ecomplex(fr, -ecomplex_imag(b), we);
}

Sex
diff2zv(Sex a, Sex b)
{
    double          fr,
                    wv,
                    we;

    wv = (interval_lower(b) + interval_upper(b)) / 2;
    we = (interval_lower(b) - interval_upper(b)) / 2;

    fr = ecomplex_real(a) - wv;
    if (we < ecomplex_error(a))
	we = ecomplex_error(a);
    return make_ecomplex(fr, -ecomplex_imag(b), we);
}


Sex
diff2zz(Sex a, Sex b)
{
    double          fr,
                    fi;
    double          e;
    fr = ecomplex_real(a) - ecomplex_real(b);
    fi = ecomplex_imag(a) - ecomplex_imag(b);
    e = ecomplex_error(a);
    if (e < ecomplex_error(b))
	e = ecomplex_error(b);
    return make_ecomplex(fr, fi, e);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
times2xz(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe;
    w = floatit(a);
    fr = w * ecomplex_real(b);
    fi = w * ecomplex_imag(b);
    fe = fabs(w * ecomplex_error(b));
    return make_ecomplex(fr, fi, fe);
}


Sex
times2fz(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe;
    w = float_val(a);
    fr = w * ecomplex_real(b);
    fi = w * ecomplex_imag(b);
    fe = fabs(w * ecomplex_error(b));
    return make_ecomplex(fr, fi, fe);
}




Sex
times2ez(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe,
                    fee;

    fr = efloat_value(a) * ecomplex_real(b);
    fi = efloat_value(a) * ecomplex_imag(b);
    fe = fabs(efloat_value(a) * ecomplex_error(b));
    fee = fabs(efloat_error(a) * ecomplex_real(b));
    if (fe < fee)
	fe = fee;
    fee = fabs(efloat_error(a) * ecomplex_imag(b));
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}

Sex
times2vz(Sex a, Sex b)
{
    double          fr,
                    fi,
                    fe,
                    fee,
                    wv,
                    we;

    wv = (interval_lower(a) + interval_upper(a)) / 2;
    we = (interval_lower(a) - interval_upper(a)) / 2;

    fr = wv * ecomplex_real(b);
    fi = wv * ecomplex_imag(b);
    fe = fabs(wv * ecomplex_error(b));
    fee = fabs(we * ecomplex_real(b));
    if (fe < fee)
	fe = fee;
    fee = fabs(we * ecomplex_imag(b));
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}


Sex
times2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return times2zz(a, b);
#else
    Sex             ar,
                    ai;
    ar = complex_real(a);
    ai = complex_imag(a);
    stackcheck();
#if TOWARD_LOWER_PRECISION
    if (!is_big(ar) || !is_big(ai)) {
	push(b);
	a = ecomplexitc(a);
	pop(b);
	return times2zz(a, b);
    } else {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return times2zZ(b, a);
    }
#else
    if (is_big(ar) || is_big(ai)) {
	push(b);
	a = becomplexitc(a);
	pop(b);
	return times2zZ(b, a);
    } else {
	push(b);
	a = ecompexitc(a);
	pop(b);
	return times2zz(a, b);
    }
#endif
#endif
}


Sex
times2zz(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe,
                    fee;

    fr = ecomplex_real(a) * ecomplex_real(b);
    fi = ecomplex_imag(a) * ecomplex_imag(b);
    fr = fr - fi;
    fi = ecomplex_imag(a) * ecomplex_real(b);
    w = ecomplex_real(a) * ecomplex_imag(b);
    fi = fi + w;
    fe = fabs(ecomplex_real(a) * ecomplex_error(b));
    fee = fabs(ecomplex_imag(a) * ecomplex_error(b));
    if (fe < fee)
	fe = fee;
    fee = fabs(ecomplex_real(b) * ecomplex_error(a));
    if (fe < fee)
	fe = fee;
    fee = fabs(ecomplex_imag(b) * ecomplex_error(a));
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
quot2zx(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe;
    w = floatit(b);
    fr = ecomplex_real(a) / w;
    fi = ecomplex_imag(a) / w;
    fe = fabs(ecomplex_error(a) / w);
    return make_ecomplex(fr, fi, fe);
}


Sex
quot2xz(Sex a, Sex b)
{
    double          w,
                    zr,
                    zi,
                    ze,
                    fr,
                    fi,
                    fe;

    zr = ecomplex_real(b) * ecomplex_real(b);
    zi = ecomplex_imag(b) * ecomplex_imag(b);
    w = zr + zi;
    zr = ecomplex_real(b) / w;
    zi = ecomplex_imag(b) / w;
    ze = ecomplex_error(b) / w;

    w = floatit(a);
    fr = w * zr;
    fi = -w * zi;
    fe = fabs(w * ze);
    return make_ecomplex(fr, fi, fe);
}


Sex
quot2zf(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe;
    w = float_val(b);
    fr = ecomplex_real(a) / w;
    fi = ecomplex_imag(a) / w;
    fe = fabs(ecomplex_error(a) / w);
    return make_ecomplex(fr, fi, fe);
}


Sex
quot2fz(Sex a, Sex b)
{
    double          w,
                    zr,
                    zi,
                    ze,
                    fr,
                    fi,
                    fe;

    zr = ecomplex_real(b) * ecomplex_real(b);
    zi = ecomplex_imag(b) * ecomplex_imag(b);
    w = zr + zi;
    zr = ecomplex_real(b) / w;
    zi = ecomplex_imag(b) / w;
    ze = ecomplex_error(b) / w;

    w = float_val(a);
    fr = w * zr;
    fi = -w * zi;
    fe = fabs(w * ze);
    return make_ecomplex(fr, fi, fe);
}






Sex
quot2ze(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe,
                    fee;

    fr = ecomplex_real(a) / efloat_value(b);
    fi = ecomplex_imag(a) / efloat_value(b);
    w = efloat_error(b) / (efloat_value(b) * efloat_value(b));
    fe = fabs(efloat_value(b) * ecomplex_error(a));
    fee = fabs(w * ecomplex_real(a));
    if (fe < fee)
	fe = fee;
    fee = fabs(w * ecomplex_imag(a));
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}

Sex
quot2ez(Sex a, Sex b)
{
    double          w,
                    zr,
                    zi,
                    ze,
                    fr,
                    fi,
                    fe,
                    fee;

    zr = ecomplex_real(b) * ecomplex_real(b);
    zi = ecomplex_imag(b) * ecomplex_imag(b);
    w = zr + zi;
    zr = ecomplex_real(b) / w;
    zi = ecomplex_imag(b) / w;
    ze = ecomplex_error(b) / w;

    fr = efloat_value(a) * zr;
    fi = efloat_value(a) * zi;
    fe = fabs(efloat_value(a) * ze);
    fee = fabs(efloat_error(a) * zr);
    if (fe < fee)
	fe = fee;
    fee = fabs(efloat_error(a) * zi);
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}

Sex
quot2zv(Sex a, Sex b)
{
    double          w,
                    fr,
                    fi,
                    fe,
                    fee,
                    wv,
                    we;

    wv = (interval_lower(b) + interval_upper(b)) / 2;
    we = (interval_lower(b) - interval_upper(b)) / 2;

    fr = ecomplex_real(a) / wv;
    fi = ecomplex_imag(a) / wv;
    w = we / (wv * wv);
    fe = fabs(ecomplex_error(a) / wv);
    fee = fabs(ecomplex_real(a) * w);
    if (fe < fee)
	fe = fee;
    fee = fabs(ecomplex_imag(a) * w);
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}

Sex
quot2vz(Sex a, Sex b)
{
    double          w,
                    zr,
                    zi,
                    ze,
                    fr,
                    fi,
                    fe,
                    fee,
                    wv,
                    we;

    zr = ecomplex_real(b) * ecomplex_real(b);
    zi = ecomplex_imag(b) * ecomplex_imag(b);
    w = zr + zi;
    zr = ecomplex_real(b) / w;
    zi = ecomplex_imag(b) / w;
    ze = ecomplex_error(b) / w;

    wv = (interval_lower(b) + interval_upper(b)) / 2;
    we = (interval_lower(b) - interval_upper(b)) / 2;

    fr = wv * zr;
    fi = wv * zi;
    fe = fabs(wv * ze);
    fee = fabs(we * zr);
    if (fe < fee)
	fe = fee;
    fee = fabs(we * zi);
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}

Sex
quot2cz(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return quot2zz(a, b);
}

Sex
quot2zc(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = ecomplexitc(b);
    pop(a);
    return quot2zz(a, b);
}

Sex
quot2ce(Sex a, Sex b)
{
    double          fr,
                    fe;
    double          gr,
                    gi;

    fr = efloat_value(b);
    fe = efloat_error(b);
    gr = floatit(complex_real(a));
    gi = floatit(complex_imag(a));

    gr = gr / fr;
    gi = gi / fr;
    return make_ecomplex(gr, gi, fe);
}

Sex
quot2ec(Sex a, Sex b)
{
    double          fr,
                    fe;
    double          gr,
                    gi,
                    ge1,
                    ge2,
                    w;

    fr = efloat_value(a);
    fe = efloat_error(a);
    gr = floatit(complex_real(b));
    gi = floatit(complex_imag(b));

    w = gr * gr + gi * gi;
    gr = fr * gr / w;
    gi = -fr * gi / w;
    fe = fe / w;
    ge1 = fabs(fe * gr);
    ge2 = fabs(fe * gi);
    if (ge1 < ge2)
	ge1 = ge2;

    return make_ecomplex(gr, gi, ge1);
}

Sex
quot2cv(Sex a, Sex b)
{
    double          fr,
                    fe;
    double          gr,
                    gi,
                    ge1,
                    ge2;

    fr = (interval_lower(b) + interval_upper(b)) * 0.5;
    fe = (interval_lower(b) - interval_upper(b)) * 0.5;
    gr = floatit(complex_real(a));
    gi = floatit(complex_imag(a));

    gr = gr / fr;
    gi = gi / fr;
    ge1 = fabs(fe * gr);
    ge2 = fabs(fe * gi);
    if (ge2 > ge1)
	ge1 = ge2;
    return make_ecomplex(gr, gi, ge1);
}

Sex
quot2vc(Sex a, Sex b)
{
    double          fr,
                    fe;
    double          gr,
                    gi,
                    ge1,
                    ge2,
                    w;

    fr = (interval_lower(a) + interval_upper(a)) * 0.5;
    fe = (interval_lower(a) - interval_upper(a)) * 0.5;
    gr = floatit(complex_real(b));
    gi = floatit(complex_imag(b));

    w = gr * gr + gi * gi;
    gr = fr * gr / w;
    gi = -fr * gi / w;
    fe = fe / w;
    ge1 = fabs(fe * gr);
    ge2 = fabs(fe * gi);
    if (ge1 < ge2)
	ge1 = ge2;

    return make_ecomplex(gr, gi, ge1);
}

Sex
quot2zz(Sex a, Sex b)
{
    double          w,
                    wr,
                    wi,
                    we,
                    fr,
                    fi,
                    fe,
                    fee;

    wr = ecomplex_real(b) * ecomplex_real(b);
    wi = ecomplex_imag(b) * ecomplex_imag(b);
    w = wr + wi;
    wr = ecomplex_real(b) / w;
    wi = ecomplex_imag(b) / w;
    we = ecomplex_error(b) / w;

    fr = ecomplex_real(a) * wr;
    fi = ecomplex_imag(a) * wi;
    fr = fr + fi;
    fi = ecomplex_imag(a) * wr;
    w = ecomplex_real(a) * wi;
    fi = fi - w;
    fe = fabs(ecomplex_real(a) * we);
    fee = fabs(ecomplex_imag(a) * we);
    if (fe < fee)
	fe = fee;
    fee = fabs(wr * ecomplex_error(a));
    if (fe < fee)
	fe = fee;
    fee = fabs(wi * ecomplex_error(a));
    if (fe < fee)
	fe = fee;
    return make_ecomplex(fr, fi, fe);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 *    compare routines
 */

bool
numeq2bz(Sex a, Sex b)
{
    double          d;

    if (fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    d = floatitb(a);
    if (fabs(ecomplex_real(b) - d) > ecomplex_error(b))
	return NO;
    return YES;
}


bool
numeq2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return numeq2zz(a, b);
#else
    double          ar,
                    ai;
    ar = floatit(complex_real(a)) - ecomplex_real(b);
    ai = floatit(complex_imag(a)) - ecomplex_imag(b);
    if (fabs(ar) > ecomplex_error(b) || fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
#endif
}

bool
numeq2dz(Sex a, Sex b)
{
    double          d;
    d = floatitd(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2ez(Sex a, Sex b)
{
    double          ar,
                    ai,
                    ae;
    ar = efloat_value(a) - ecomplex_real(b);
    ai = ecomplex_imag(b);
    ae = efloat_error(a);
    if (ae < ecomplex_error(b))
	ae = ecomplex_error(b);
    if (fabs(ar) > ae)
	return NO;
    if (fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2fz(Sex a, Sex b)
{
    double          d;
    d = float_val(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2iz(Sex a, Sex b)
{
    double          d;
    d = mkint(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2rz(Sex a, Sex b)
{
    double          d;
    d = floatit(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2vz(Sex a, Sex b)
{
    double          ar,
                    ai,
                    ae;

    ar = (interval_upper(a) + interval_lower(a)) * 0.5;
    ae = (interval_upper(a) - interval_lower(a)) * 0.5;
    ar = ar - ecomplex_real(b);
    ai = ecomplex_imag(b);

    if (ae < ecomplex_error(b))
	ae = ecomplex_error(b);
    if (fabs(ar) > ae)
	return NO;
    if (fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2zc(Sex b, Sex a)
{
    double          ar,
                    ai;
    ar = floatit(complex_real(a)) - ecomplex_real(b);
    ai = floatit(complex_imag(a)) - ecomplex_imag(b);
    if (fabs(ar) > ecomplex_error(b) || fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
numeq2zz(Sex a, Sex b)
{
    double          e;
    e = ecomplex_error(a);
    if (e < ecomplex_error(b))
	e = ecomplex_error(b);
    if (fabs(ecomplex_real(a) - ecomplex_real(b)) > e)
	return NO;
    if (fabs(ecomplex_imag(a) - ecomplex_imag(b)) > e)
	return NO;
    return YES;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
lessp2xz(Sex a, Sex b)
{
    Sex             c;

    if (ecomplex_imag(b) != 0.0)
	return bool_er(a, b);
    stackcheck();
    push2(a, b);
    c = new_efloat();
    pop2(b, a);
    efloat_value(c) = ecomplex_real(b);
    efloat_error(c) = ecomplex_error(b);
    return lessp2(a, c);
}

bool
lessp2zx(Sex a, Sex b)
{
    Sex             c;

    if (ecomplex_imag(a) != 0.0)
	return bool_er(a, b);
    stackcheck();
    push2(a, b);
    c = new_efloat();
    pop2(b, a);
    efloat_value(c) = ecomplex_real(a);
    efloat_error(c) = ecomplex_error(a);
    return lessp2(c, b);
}



bool
lessp2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return lessp2zz(a, b);
#else
    double          ar,
                    ai;
    ar = floatit(complex_real(a));
    ai = floatit(complex_imag(a));
    if (fabs(ecomplex_imag(b)) < ecomplex_error(b) &&
	ai == 0.0 && ar < ecomplex_real(b) - ecomplex_error(b))
	return YES;
    return NO;
#endif
}

bool
lessp2zc(Sex a, Sex b)
{
    double          br,
                    bi;
    br = floatit(complex_real(b));
    bi = floatit(complex_imag(b));
    if (fabs(ecomplex_imag(a)) < ecomplex_error(a) &&
	bi == 0.0 && ecomplex_real(a) + ecomplex_error(a) < br)
	return YES;
    return NO;
}

bool
lessp2zz(Sex a, Sex b)
{
    if (fabs(ecomplex_imag(a)) < ecomplex_error(a) &&
	fabs(ecomplex_imag(b)) < ecomplex_error(b) &&
	ecomplex_real(a) + ecomplex_error(a) <
	ecomplex_real(b) - ecomplex_error(b))
	return YES;
    return NO;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
appeq2bz(Sex a, Sex b)
{
    double          d;

    if (fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    d = floatitb(a);
    if (fabs(ecomplex_real(b) - d) > ecomplex_error(b))
	return NO;
    return YES;
}


bool
appeq2cz(Sex a, Sex b)
{
#if 1
    stackcheck();
    push(b);
    a = ecomplexitc(a);
    pop(b);
    return appeq2zz(a, b);
#else
    double          ar,
                    ai;
    ar = floatit(complex_real(a)) - ecomplex_real(b);
    ai = floatit(complex_imag(a)) - ecomplex_imag(b);
    if (fabs(ar) > ecomplex_error(b) || fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
#endif
}

bool
appeq2dz(Sex a, Sex b)
{
    double          d;
    d = floatitd(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2ez(Sex a, Sex b)
{
    double          ar,
                    ai,
                    ae;
    ar = efloat_value(a) - ecomplex_real(b);
    ai = ecomplex_imag(b);
    ae = efloat_error(a);
    if (ae < ecomplex_error(b))
	ae = ecomplex_error(b);
    if (fabs(ar) > ae)
	return NO;
    if (fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2fz(Sex a, Sex b)
{
    double          d;
    d = float_val(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2iz(Sex a, Sex b)
{
    double          d;
    d = mkint(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2rz(Sex a, Sex b)
{
    double          d;
    d = floatit(a);
    if (fabs(d - ecomplex_real(b)) > ecomplex_error(b) ||
	fabs(ecomplex_imag(b)) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2vz(Sex a, Sex b)
{
    double          ar,
                    ai,
                    ae;

    ar = (interval_upper(a) + interval_lower(a)) * 0.5;
    ae = (interval_upper(a) - interval_lower(a)) * 0.5;
    ar = ar - ecomplex_real(b);
    ai = ecomplex_imag(b);

    if (ae < ecomplex_error(b))
	ae = ecomplex_error(b);
    if (fabs(ar) > ae)
	return NO;
    if (fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2zc(Sex b, Sex a)
{
    double          ar,
                    ai;
    ar = floatit(complex_real(a)) - ecomplex_real(b);
    ai = floatit(complex_imag(a)) - ecomplex_imag(b);
    if (fabs(ar) > ecomplex_error(b) || fabs(ai) > ecomplex_error(b))
	return NO;
    return YES;
}

bool
appeq2zz(Sex a, Sex b)
{
    double          e;
    e = ecomplex_error(a);
    if (e < ecomplex_error(b))
	e = ecomplex_error(b);
    if (fabs(ecomplex_real(a) - ecomplex_real(b)) > e)
	return NO;
    if (fabs(ecomplex_imag(a) - ecomplex_imag(b)) > e)
	return NO;
    return YES;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/
