/*
 * efloat.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * effective float routine
 *      $Header: /Users/kako/asuka/src/RCS/efloat.c,v 2.10 2014/10/27 02:28:08 kako Exp $
 */

static char const rcsid[] =
    "$Id: efloat.c,v 2.10 2014/10/27 02:28:08 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "efloat.h"



/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Lefloat_value(Sex a)
{
    if (is_fixnum(a))
	return a;
    if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_INTERVAL:
	case N_RATIONAL:
	case N_FLOAT:
	case N_BFLOAT:
	case N_BINTERVAL:
	case N_COMPLEX:
	    return a;
	case N_EFLOAT:
	    return make_float(efloat_value(a));
	case N_BEFLOAT:
	    return befloat_value(a);
	case N_ECOMPLEX:
	case N_BECOMPLEX:
	    return Lcomplex(a);
	}
    }
    return aerror("efloat_value: not number");
}

Sex
Lefloat_error(Sex a)
{
    double          d,
                    r;

    if (is_fixnum(a))
	return mkfxat(0);
    if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_RATIONAL:
	case N_INTERVAL:
	case N_BINTERVAL:
	case N_COMPLEX:
	    return mkfxat(0);
	case N_FLOAT:
	    return make_float(0.0);
	case N_BFLOAT:
	    return bfloatiti(mkfxat(0));
	case N_EFLOAT:
	    return make_float(efloat_error(a));
	case N_BEFLOAT:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    return make_float(befloat_error(a));
#else
	    d = floatit(befloat_value(a));
	    d = fabs(d);
	    d = ldexp(d, befloat_error(a));
	    return make_float(d);
#endif
	case N_ECOMPLEX:
	    return make_float(ecomplex_error(a));
	case N_BECOMPLEX:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    return make_float(becomplex_error(a));
#else
	    r = fabs(floatit(becomplex_real(a)));
	    d = fabs(floatit(becomplex_imag(a)));
	    if (d < r)
		d = r;
	    d = ldexp(d, becomplex_error(a));

	    return make_float(d);
#endif
	}
    }
    return aerror("efloat_value: not number");
}


Sex
Lmkefloat(Sex a, Sex b)
{
    double          e;
    e = floatit(b);
    a = Lefloat(a);
    if (is_ecomplex(a))
	ecomplex_error(a) = e;
    else
	efloat_error(a) = e;
    return a;
}

Sex
Lmkefloat_by_ratio(Sex a, Sex b)
{
    double          e;
    double          r;
    double          i;
    e = floatit(b);
    a = Lefloat(a);
    if (is_ecomplex(a)) {
	r = ecomplex_real(a);
	r = fabs(r);
	i = ecomplex_imag(a);
	i = fabs(i);
	if (r < i)
	    r = i;
	ecomplex_error(a) = r * e;
    } else {
	r = efloat_value(a);
	r = fabs(r);
	efloat_error(a) = r * e;
    }
    return a;
}

Sex
efloatitc(Sex a)
{
    double          r,
                    i,
                    e;
    double          c;
    r = floatit(complex_real(a));
    i = floatit(complex_imag(a));
    if (is_fixnum(complex_real(a)))
	c = 0;
    else
	c = fabs(r);
    if (is_fixnum(complex_imag(a)))
	e = 0;
    else
	e = fabs(i);
    if (e < c)
	e = c;
    e = c * EFLOAT_ERROR;
    if (i == 0.0) {
	a = new_efloat();
	efloat_value(a) = r;
	efloat_error(a) = e;
	return a;
    } else {
	a = new_ecomplex();
	ecomplex_real(a) = r;
	ecomplex_imag(a) = i;
	ecomplex_error(a) = e;
	return a;
    }
}

Sex
efloatitE(Sex a)
{
    DoubleFloat     u;
    long            n;
    double          v;
    bf_error_type   e;

    v = floatitd(befloat_value(a));
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    e = befloat_error(a);
#else
    e = fabs(v);
    e = ldexp(e, befloat_error(a));
#endif
    a = new_efloat();
    efloat_value(a) = v;
    efloat_error(a) = e;
    return a;
}

Sex
efloatitZ(Sex a)
{
    double          r,
                    i;
    double          ia;
    bf_error_type   e;

    r = floatitd(becomplex_real(a));
    i = floatitd(becomplex_imag(a));
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    e = becomplex_error(a);
#else
    e = fabs(r);
    ia = fabs(i);
    if (e < ia)
	e = ia;
    e = ldexp(e, becomplex_error(a));
#endif

    if (i == 0.0) {
	a = new_efloat();
	efloat_value(a) = r;
	efloat_error(a) = e;
	return a;
    } else {
	a = new_ecomplex();
	ecomplex_real(a) = r;
	ecomplex_imag(a) = i;
	ecomplex_error(a) = e;
	return a;
    }
}

Sex
efloatitv(Sex a)
{
    double          d,
                    e;
    d = interval_upper(a) + interval_lower(a);
    e = interval_upper(a) - interval_lower(a);
    d = d * 0.5;
    e = e * 0.5;
    return make_efloat(d, e);
}

Sex
efloatitV(Sex a)
{
    double          u,
                    l;
    u = floatitd(binterval_upper(a));
    l = floatitd(binterval_lower(a));
    a = new_efloat();
    efloat_value(a) = (u + l) * 0.5;
    efloat_error(a) = (u - l) * 0.5;
    return a;
}


Sex
efloatit(Sex a)
{
    double          v,
                    e;

    e = 0.0;
    if (is_fixnum(a)) {
	v = mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    v = floatitl(a);
	    break;
	case N_BIGNUM:
	    v = floatitb(a);
	    break;
	case N_FLOAT:
	    v = float_val(a);
	    break;
	case N_RATIONAL:
	    v = floatit(rational_num(a));
	    v = v / floatit(rational_den(a));
	    break;
	case N_BFLOAT:
	    v = floatitd(a);
	    break;
	case N_INTERVAL:
	    v = (interval_upper(a) + interval_lower(a)) * 0.5;
	    e = (interval_upper(a) - interval_lower(a)) * 0.5;
	    break;
	case N_BINTERVAL:
	    return efloatitV(a);
	case N_EFLOAT:
	    return a;
	case N_BEFLOAT:
	    return efloatitE(a);
	default:
	    return error1("bad argument to efloat", a);
	}
    } else
	return error1("bad argument to efloat", a);

    a = new_efloat();
    efloat_value(a) = v;
    efloat_error(a) = e;
    return a;
}

Sex
Lefloat(Sex a)
{
    double          v,
                    e;

    e = EFLOAT_ERROR;
    if (is_fixnum(a)) {
	v = mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    v = floatitl(a);
	    break;
	case N_BIGNUM:
	    v = floatitb(a);
	    break;
	case N_FLOAT:
	    v = float_val(a);
	    break;
	case N_RATIONAL:
	    v = floatit(rational_num(a));
	    v = v / floatit(rational_den(a));
	    break;
	case N_COMPLEX:
	    return efloatitc(a);
	case N_BFLOAT:
	    v = floatitd(a);
	    break;
	case N_INTERVAL:
	    v = (interval_upper(a) + interval_lower(a)) * 0.5;
	    e = (interval_upper(a) - interval_lower(a)) * 0.5;
	    a = new_efloat();
	    efloat_value(a) = v;
	    efloat_error(a) = e;
	    return a;
	case N_BINTERVAL:
	    return efloatitV(a);
	case N_EFLOAT:
	case N_ECOMPLEX:
	    return a;
	case N_BEFLOAT:
	    return efloatitE(a);
	case N_BECOMPLEX:
	    return efloatitZ(a);
	default:
	    return error1("bad argument to efloat", a);
	}
    } else
	return error1("bad argument to efloat", a);

    a = new_efloat();
    efloat_value(a) = v;
    efloat_error(a) = e * fabs(v);
    return a;
}

Sex
Lmkefloat1(Sex a)
{
    double          v,
                    e;

    e = 0.0;
    if (is_fixnum(a)) {
	v = mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    v = floatitl(a);
	    e = fabs(v) * EFLOAT_ERROR;
	    break;
	case N_BIGNUM:
	    v = floatitb(a);
	    e = fabs(v) * EFLOAT_ERROR;
	    break;
	case N_FLOAT:
	    v = float_val(a);
	    e = fabs(v) * EFLOAT_ERROR;
	    break;
	case N_RATIONAL:
	    v = floatit(rational_num(a));
	    v = v / floatit(rational_den(a));
	    e = fabs(v) * EFLOAT_ERROR;
	    break;
	case N_COMPLEX:
	    return efloatitc(a);
	case N_BFLOAT:
	    v = floatitd(a);
	    e = fabs(v) * EFLOAT_ERROR;
	    break;
	case N_INTERVAL:
	    v = (interval_upper(a) + interval_lower(a)) * 0.5;
	    e = (interval_upper(a) - interval_lower(a)) * 0.5;
	    break;
	case N_BINTERVAL:
	    return efloatitV(a);
	case N_EFLOAT:
	case N_ECOMPLEX:
	    return a;
	case N_BEFLOAT:
	    return efloatitE(a);
	case N_BECOMPLEX:
	    return efloatitZ(a);
	default:
	    return error1("bad argument to efloat", a);
	}
    } else
	return error1("bad argument to efloat", a);

    a = new_efloat();
    efloat_value(a) = v;
    efloat_error(a) = e;
    return a;
}

Sex
make_efloat(double v, double e)
{
    Sex             r;
    r = new_efloat();
    efloat_value(r) = v;
    efloat_error(r) = e;
    return r;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2ie(Sex a, Sex b)
{
    Sex             c;
    double          wa = mkint(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa + wb;
    efloat_error(c) = we;
    return c;
}


Sex
plus2ei(Sex b, Sex a)
{
    Sex             c;
    double          wa = mkint(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa + wb;
    efloat_error(c) = we;
    return c;
}

Sex
plus2de(Sex a, Sex b)
{
    return plus2ed(b, a);
}


Sex
plus2xe(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa + wb;
    efloat_error(c) = we;
    return c;
}

Sex
plus2ex(Sex b, Sex a)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa + wb;
    efloat_error(c) = we;
    return c;
}

Sex
plus2fe(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a) + efloat_value(b);
    double          wb = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}

Sex
plus2ef(Sex b, Sex a)
{
    Sex             c;
    double          wa = float_val(a) + efloat_value(b);
    double          wb = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}

Sex
plus2ee(Sex a, Sex b)
{
    Sex             c;
    double          wc = efloat_value(a) + efloat_value(b);
    double          ea = efloat_error(a);
    double          eb = efloat_error(b);
    DoubleFloat    *u,
                   *v;
    long            ia,
                    ib,
                    ja,
                    jb;


    if (ea == 0.0)
	ea = eb;
    else if (eb == 0.0);
    else {
	u = (DoubleFloat *) & efloat_error(a);
	ja = ieee_exponent(u[0]);
	u = (DoubleFloat *) & efloat_error(b);
	jb = ieee_exponent(u[0]);
	u = (DoubleFloat *) & efloat_value(a);
	ia = ieee_exponent(u[0]);
	v = (DoubleFloat *) & efloat_value(b);
	ib = ieee_exponent(v[0]);
	if (ia - ja == ib - jb &&
	    ((ieee_signum(u[0]) ^ ieee_signum(v[0])) & 0x80000000) == 0)
	    ea = ea + eb;
	else if (ea < eb)
	    ea = eb;
    }

    c = new_efloat();
    efloat_value(c) = wc;
    efloat_error(c) = ea;

    return c;
}

Sex
plus2ec(Sex a, Sex b)
{
    double          r,
                    i,
                    e;
    double          cr,
                    ci;
    r = efloat_value(a);
    e = efloat_error(a);
#if TOWARD_LOWER_PRECISION
    cr = floatit(complex_real(b));
    ci = floatit(complex_imag(b));
    return make_ecomplex(r + cr, ci, e);
#else
    if (!is_longfloat(b)) {
	cr = floatit(complex_real(b));
	ci = floatit(complex_imag(b));
	return make_ecomplex(r + cr, ci, e);
    } else {
	Sex             sr,
	                si;
	stackcheck();
	push(b);
	sr = bfloatit(complex_real(b));
	push(sr);
	b = stack[-1];
	si = bfloatit(complex_imag(b));
	push(si);
	a = bfloatitf(r);
	sr = stack[-1];
	a = plus2dd(a, sr);
	si = stack[0];
	popv(3);
	return make_becomplex(a, si, e);
    }
#endif
}


Sex
plus2ve(Sex a, Sex b)
{
    return plus2ev(b, a);
}

Sex
plus2ev(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a);
    double          wb = (interval_upper(b) + interval_lower(b)) / 2;
    double          ea = efloat_error(a);
    double          eb = (interval_upper(b) - interval_lower(b)) / 2;

    if (eb > ea)
	ea = eb;
    c = new_efloat();
    efloat_value(c) = wa + wb;
    efloat_error(c) = ea;

    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
diff2ie(Sex a, Sex b)
{
    Sex             c;
    double          wa = mkint(a) - efloat_value(b);
    double          wb = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}


Sex
diff2ei(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a) - mkint(b);
    double          wb = efloat_error(a);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}

Sex
diff2xe(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a) - efloat_value(b);
    double          wb = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}


Sex
diff2ex(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a) - floatit(b);
    double          wb = efloat_error(a);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}

Sex
diff2fe(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a) - efloat_value(b);
    double          wb = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}



Sex
diff2ef(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a) - float_val(b);
    double          wb = efloat_error(a);
    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = wb;
    return c;
}

Sex
diff2ee(Sex a, Sex b)
{
    double          wc = efloat_value(a) - efloat_value(b);
    double          ea = efloat_error(a);
    double          eb = efloat_error(b);
    DoubleFloat    *u,
                   *v;
    long            ia,
                    ib,
                    ja,
                    jb;

    if (ea == 0.0)
	ea = eb;
    else if (eb == 0.0);
    else {
	u = (DoubleFloat *) & efloat_error(a);
	ja = ieee_exponent(u[0]);
	u = (DoubleFloat *) & efloat_error(b);
	jb = ieee_exponent(u[0]);
	u = (DoubleFloat *) & efloat_value(a);
	ia = ieee_exponent(u[0]);
	v = (DoubleFloat *) & efloat_value(b);
	ib = ieee_exponent(v[0]);
	if (ia - ja == ib - jb &&
	    ((ieee_signum(u[0]) ^ ieee_signum(v[0])) & 0x80000000) != 0)
	    ea = ea + eb;
	else if (ea < eb)
	    ea = eb;
    }

    a = new_efloat();
    efloat_value(a) = wc;
    efloat_error(a) = ea;

    return a;
}

Sex
diff2ec(Sex a, Sex b)
{
    double          r,
                    i,
                    e;
    double          cr,
                    ci;
    r = efloat_value(a);
    e = efloat_error(a);
#if TOWARD_LOWER_PRECISION
    cr = floatit(complex_real(b));
    ci = floatit(complex_imag(b));
    return make_ecomplex(r - cr, -ci, e);
#else
    if (!is_longfloat(b)) {
	cr = floatit(complex_real(b));
	ci = floatit(complex_imag(b));
	return make_ecomplex(r - cr, -ci, e);
    } else {
	Sex             sr,
	                si;
	stackcheck();
	push(b);
	sr = bfloatit(complex_real(b));
	push(sr);
	b = stack[-1];
	si = bfloatit(complex_imag(b));
	si = negated(si);
	push(si);
	a = bfloatitf(r);
	sr = stack[-1];
	a = diff2dd(a, sr);
	si = stack[0];
	popv(3);
	return make_becomplex(a, si, e);
    }
#endif
}

Sex
diff2ce(Sex a, Sex b)
{
    double          r,
                    i,
                    e;
    double          cr,
                    ci;
    r = efloat_value(b);
    e = efloat_error(b);
#if TOWARD_LOWER_PRECISION
    cr = floatit(complex_real(a));
    ci = floatit(complex_imag(a));
    return make_ecomplex(cr - r, ci, e);
#else
    if (!is_longfloat(a)) {
	cr = floatit(complex_real(a));
	ci = floatit(complex_imag(a));
	return make_ecomplex(cr - r, ci, e);
    } else {
	Sex             sr,
	                si;
	stackcheck();
	push(b);
	sr = bfloatit(complex_real(b));
	push(sr);
	b = stack[-1];
	si = bfloatit(complex_imag(b));
	push(si);
	a = bfloatitf(r);
	sr = stack[-1];
	sr = diff2dd(sr, a);
	si = stack[0];
	popv(3);
	return make_becomplex(sr, si, e);
    }
#endif
}


Sex
diff2ev(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a);
    double          wb = (interval_upper(b) + interval_lower(b)) / 2;
    double          eb = (interval_upper(b) - interval_lower(b)) / 2;
    double          ea = efloat_error(a);

    if (ea < eb)
	ea = eb;

    c = new_efloat();
    efloat_value(c) = wa - wb;
    efloat_error(c) = ea;

    return c;
}


Sex
diff2ve(Sex a, Sex b)
{
    Sex             c;
    double          wa = (interval_upper(a) + interval_lower(a)) / 2;
    double          wb = efloat_value(b);
    double          ea = (interval_upper(a) - interval_lower(a)) / 2;
    double          eb = efloat_error(a);

    if (eb > ea)
	ea = eb;
    c = new_efloat();
    efloat_value(c) = wa - wb;
    efloat_error(c) = ea;
    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
times2ie(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}


Sex
times2ei(Sex b, Sex a)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}



Sex
times2fe(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}

Sex
times2ef(Sex b, Sex a)
{
    Sex             c;
    double          wa = float_val(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}

Sex
times2xe(Sex a, Sex b)
{
    Sex             c;
    double          wa;
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    wa = floatit(a);
    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}

Sex
times2ex(Sex b, Sex a)
{
    Sex             c;
    double          wa;
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    wa = floatit(a);
    c = new_efloat();
    efloat_value(c) = wa * wb;
    efloat_error(c) = fabs(wa * we);
    return c;
}


Sex
times2ee(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a) * efloat_value(b);
    double          e1 = fabs(efloat_error(a) * efloat_value(b));
    double          e2 = fabs(efloat_error(b) * efloat_value(a));
    if (e1 < e2)
	e1 = e2;
    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = e1;

    return c;
}

Sex
times2ec(Sex a, Sex b)
{
    double          r,
                    e;
    double          cr,
                    ci;
    double          ar,
                    ai;
    r = efloat_value(a);
    e = efloat_error(a);
#if TOWARD_LOWER_PRECISION
    cr = floatit(complex_real(b));
    ci = floatit(complex_imag(b));
    ar = fabs(cr);
    ai = fabs(ci);
    if (ai > ar)
	ar = ai;
    return make_ecomplex(r * cr, r * ci, e * ar);
#else
    if (!is_longfloat(b)) {
	cr = floatit(complex_real(b));
	ci = floatit(complex_imag(b));
	ar = fabs(cr);
	ai = fabs(ci);
	if (ai > ar)
	    ar = ai;
	return make_ecomplex(r * cr, r * ci, e * ai);
    } else {
	Sex             sr,
	                si;
	stackcheck();
	cr = floatit(complex_real(b));
	ci = floatit(complex_imag(b));
	cr = fabs(cr);
	ci = fabs(ci);
	if (ci > cr)
	    cr = ci;
	push(b);
	sr = bfloatit(complex_real(b));
	push(sr);
	b = stack[-1];
	si = bfloatit(complex_imag(b));
	push(si);
	a = bfloatitf(r);
	sr = stack[-1];
	a = plus2dd(a, sr);
	si = stack[0];
	popv(3);
	return make_becomplex(a, si, e * ci);
    }
#endif
}

Sex
times2ve(Sex a, Sex b)
{
    return times2ev(b, a);
}

Sex
times2ev(Sex a, Sex b)
{
    Sex             c;

    double          vv = (interval_upper(b) + interval_lower(b)) / 2;
    double          ve = (interval_upper(b) - interval_lower(b)) / 2;
    double          wa = efloat_value(a) * vv;
    double          e1 = fabs(efloat_error(a) * vv);
    double          e2 = fabs(efloat_value(a) * ve);

    if (e1 < e2)
	e1 = e2;
    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = e1;
    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
quot2ie(Sex a, Sex b)
{
    Sex             c;
    double          wa = mkint(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa / wb;
    efloat_error(c) = fabs(efloat_value(c) * we / wb);
    return c;
}


Sex
quot2ei(Sex a, Sex b)
{
    Sex             c;
    double          wa = mkint(b);
    double          wb = efloat_value(a);
    double          we = efloat_error(a);

    c = new_efloat();
    efloat_value(c) = wb / wa;
    efloat_error(c) = fabs(we / wa);
    return c;
}

Sex
quot2xe(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa / wb;
    efloat_error(c) = fabs(efloat_value(c) * we / wb);
    return c;
}


Sex
quot2ex(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(b);
    double          wb = efloat_value(a);
    double          we = efloat_error(a);

    c = new_efloat();
    efloat_value(c) = wb / wa;
    efloat_error(c) = fabs(we / wa);
    return c;
}

Sex
quot2fe(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a);
    double          wb = efloat_value(b);
    double          we = efloat_error(b);

    c = new_efloat();
    efloat_value(c) = wa / wb;
    efloat_error(c) = fabs(efloat_value(c) * we / wb);
    return c;
}


Sex
quot2ef(Sex a, Sex b)
{
    Sex             c;
    double          wb = float_val(b);
    double          wa = efloat_value(a);
    double          we = efloat_error(a);

    c = new_efloat();
    efloat_value(c) = wa / wb;
    efloat_error(c) = fabs(we / wb);
    return c;
}

Sex
quot2ee(Sex a, Sex b)
{
    Sex             c;
    double          wa = efloat_value(a) / efloat_value(b);
    double          e1 = fabs(efloat_error(a) / efloat_value(b));
    double          e2 = fabs(wa * efloat_error(b) / efloat_value(b));

    if (e1 < e2)
	e1 = e2;
    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = e1;
    return c;
}


Sex
quot2ev(Sex a, Sex b)
{
    Sex             c;
    double          vv = (interval_upper(b) + interval_lower(b)) / 2;
    double          ve = (interval_upper(b) - interval_lower(b)) / 2;
    double          wa = efloat_value(a) / vv;
    double          e1 = fabs(efloat_error(a) / vv);
    double          e2 = fabs(wa * ve / vv);

    if (e1 < e2)
	e1 = e2;

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = e1;

    return c;
}


Sex
quot2ve(Sex a, Sex b)
{
    Sex             c;
    double          vv = (interval_upper(a) + interval_lower(a)) / 2;
    double          ve = (interval_upper(a) - interval_lower(a)) / 2;
    double          wa = vv / efloat_value(b);
    double          e1 = fabs(ve / efloat_value(b));
    double          e2 = fabs(wa * efloat_error(b) / efloat_value(b));

    if (e1 < e2)
	e1 = e2;

    c = new_efloat();
    efloat_value(c) = wa;
    efloat_error(c) = e1;

    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



bool
numeq2ie(Sex a, Sex b)
{
    if (efloat_error(b) != 0.0)
	return NO;
    else
	return (mkint(a) == efloat_value(b));
}

bool
numeq2be(Sex a, Sex b)
{
    if (efloat_error(b) != 0.0)
	return NO;
    else
	return (floatit(a) == efloat_value(b));
}

bool
numeq2re(Sex a, Sex b)
{
    if (efloat_error(b) != 0.0)
	return NO;
    else
	return (floatit(a) == efloat_value(b));
}

Sex
numeq2fe(Sex a, Sex b)
{
    if (efloat_error(b) != 0.0)
	return NO;
    else
	return float_val(a) == efloat_value(b);
}

bool
numeq2ed(Sex a, Sex b)
{
    push(b);
    a = bfloatitf(efloat_value(a));
    pop(b);
    return numeq2dd(a, b);
}

bool
numeq2ee(Sex a, Sex b)
{
    return (efloat_value(a) == efloat_value(b)) &&
	(efloat_error(a) == efloat_error(b));
}

bool
numeq2ev(Sex a, Sex b)
{
    double          vv = (interval_upper(b) + interval_lower(b)) / 2;
    double          ve = (interval_upper(b) - interval_lower(b)) / 2;

    return (vv == efloat_value(a) && ve == efloat_error(a));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



bool
lessp2ie(Sex a, Sex b)
{
    double          f = mkint(a);
    double          ev = efloat_value(b);
    double          ee = efloat_error(b);

    if (f < ev - ee)
	return YES;
    return NO;
}

bool
lessp2ei(Sex a, Sex b)
{
    double          f = mkint(b);
    double          ev = efloat_value(a);
    double          ee = efloat_error(a);

    if (ev + ee < f)
	return YES;
    return NO;
}


bool
lessp2ex(Sex a, Sex b)
{
    double          f;
    double          ev = efloat_value(a);
    double          ee = efloat_error(a);

    f = floatit(b);
    if (ev + ee < f)
	return YES;
    return NO;
}

bool
lessp2xe(Sex a, Sex b)
{
    double          f;
    double          ev = efloat_value(b);
    double          ee = efloat_error(b);

    f = floatit(a);
    if (f < ev - ee)
	return YES;
    return NO;
}

bool
lessp2ef(Sex a, Sex b)
{
    double          f = float_val(b);
    double          ev = efloat_value(a);
    double          ee = efloat_error(a);

    if (ev + ee < f)
	return YES;
    return NO;
}

bool
lessp2fe(Sex a, Sex b)
{
    double          f = float_val(a);
    double          ev = efloat_value(b);
    double          ee = efloat_error(b);

    if (f < ev - ee)
	return YES;
    return NO;
}


bool
lessp2ee(Sex a, Sex b)
{

    if (efloat_value(a) + efloat_error(a) <
	efloat_value(b) - efloat_error(b))
	return YES;
    return NO;
}

bool
lessp2ev(Sex a, Sex b)
{

    if (efloat_value(a) + efloat_error(a) < interval_lower(b))
	return YES;
    return NO;
}

bool
lessp2ve(Sex a, Sex b)
{

    if (interval_upper(a) < efloat_value(b) - efloat_error(b))
	return YES;
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
appeq2ie(Sex a, Sex b)
{
    double          f = mkint(a);
    if (f >= efloat_value(b) - efloat_error(b) &&
	f <= efloat_value(b) + efloat_error(b))
	return YES;
    return NO;
}

bool
appeq2be(Sex a, Sex b)
{
    double          f = floatitb(a);
    if (f >= efloat_value(b) - efloat_error(b) &&
	f <= efloat_value(b) + efloat_error(b))
	return YES;
    return NO;
}

bool
appeq2re(Sex a, Sex b)
{
    double          f;

    f = floatit(rational_num(a)) / floatit(rational_den(a));
    if (f >= efloat_value(b) - efloat_error(b) &&
	f <= efloat_value(b) + efloat_error(b))
	return YES;
    return NO;
}

Sex
appeq2fe(Sex a, Sex b)
{
    double          f = float_val(a);
    if (f >= efloat_value(b) - efloat_error(b) &&
	f <= efloat_value(b) + efloat_error(b))
	return YES;
    return NO;
}

bool
appeq2ed(Sex a, Sex b)
{
    double          f = floatitd(b);
    if (f >= efloat_value(a) - efloat_error(a) &&
	f <= efloat_value(a) + efloat_error(a))
	return YES;
    return NO;
}

bool
appeq2ee(Sex a, Sex b)
{
    return (efloat_value(a) == efloat_value(b)) &&
	(efloat_error(a) == efloat_error(b));
}

bool
appeq2ev(Sex a, Sex b)
{
    double          vv = (interval_upper(b) + interval_lower(b)) / 2;
    double          ve = (interval_upper(b) - interval_lower(b)) / 2;

    return (vv == efloat_value(a) && ve == efloat_error(a));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
