/*
 * float.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/float.c,v 2.3 2013/03/28 07:15:05 kako Exp $
 */

static char const rcsid[] =
    "$Id: float.c,v 2.3 2013/03/28 07:15:05 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "float.h"

#define SAFE_FP

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
make_float(double a)
{
    Sex             r;
    r = new_float();
    float_val(r) = a;
    return r;
}

Sex
Lfrexp(Sex a)
{
    double          d;
    int             x;
    d = floatit(a);
    d = frexp(d, &x);
    a = make_float(d);
    return cons(mkfxat((long) x), a);
}

#define TWO_32    4294967296.0	/* * 2^32 */
#define TWO_31    2147483648.0	/* * 2^31 */
#define TWO_64    18446744073709551616.0	/* 2^64 */
#define TWO_63    9223372036854775808.0	/* 2^63 */
#ifdef _LP64
#define TWO_LP 		TWO_64
#define TWO_LP_1	TWO_63
#else
#define TWO_LP 		TWO_32
#define TWO_LP_1	TWO_31
#endif


double
floatitl(Sex v)
{
    DoubleFloat     u;
    long            len = numlen(v);

    len = bnum_n(len);
    if (len > 0)
	u.i[0] = logical_value(v)[0];
    u.i[1] = 0;
    if (len > 1)
	u.i[1] = logical_value(v)[1];
    return u.d;
}

double
floatitb(Sex v)
{
    long            n = bnum_n(numlen(v)) - 1;
    long            x = LPSIZE * (long) n;
    double          r = (double) bignum_value(v)[n];
    double          s;

    if (r < 0)
	r = r + TWO_LP_1;

    switch (n) {
    default:
	r = TWO_LP * r;
	s = (double) bignum_value(v)[--n];
	if (s < 0)
	    s = s + TWO_LP_1;
	r = r + s;
	x -= LPSIZE;
    case 1:
	r = TWO_LP * r;
	s = (double) bignum_value(v)[--n];
	if (s < 0)
	    s = s + TWO_LP_1;
	r = r + s;
	x -= LPSIZE;
    case 0:
	break;
    }
    if (is_bigneg(v))
	return -ldexp(r, x);
    else
	return ldexp(r, x);
}

double
floatitr(Sex a)
{
    double          r,
                    s;
    Sex             b;
    b = rational_den(a);
    a = rational_num(a);
    if (is_fixnum(a))
	r = mkint(a);
    else
	r = floatitb(a);
    if (is_fixnum(b))
	s = mkint(b);
    else
	s = floatitb(b);
    return r / s;
}


double
floatitd(Sex v)
{
    long            n;
    long            x;
    double          r;
    double          s;
    long            exp;

    exp = bfloat_exp(v);
    v = bfloat_man(v);
    n = bnum_n(numlen(v)) - 1;
    x = LPSIZE * (long) n + exp;
    r = (double) bignum_value(v)[n];

    if (r < 0)
	r = r + TWO_LP_1;

    switch (n) {
    default:
	r = TWO_LP * r;
	s = (double) bignum_value(v)[--n];
	if (s < 0)
	    s = s + TWO_LP_1;
	r = r + s;
	x -= LPSIZE;
    case 1:
	r = TWO_LP * r;
	s = (double) bignum_value(v)[--n];
	if (s < 0)
	    s = s + TWO_LP_1;
	r = r + s;
	x -= LPSIZE;
    case 0:
	break;
    }
    if (is_bigneg(v))
	return -ldexp(r, x);
    else
	return ldexp(r, x);
}

double
floatitE(Sex a)
{
    return floatitd(befloat_value(a));
}


double
floatitV(Sex a)
{
    double          u,
                    l;

    u = floatitd(binterval_upper(a));
    l = floatitd(binterval_lower(a));
    return (u + l) * 0.5;
}

Sex
floatitz(Sex a)
{
    Sex             cr,
                    ci;
    double          r,
                    i;
    r = ecomplex_real(a);
    i = ecomplex_imag(a);
    if (i == 0.0)
	return make_float(r);
#if !CONVERT_FROM_COMPLEX
    return aerror1("convert ecomplex to float", a);
#else
    cr = new_float();
    stackcheck();
    push(cr);
    ci = new_float();
    pop(cr);
    float_val(cr) = r;
    float_val(ci) = i;
    return Lmkcomplex(cr, ci);
#endif
}

Sex
floatitZ(Sex a)
{
    Sex             cr,
                    ci;
    double          r,
                    i;
    r = floatitd(becomplex_real(a));
    i = floatitd(becomplex_imag(a));
    if (i == 0.0)
	return make_float(r);
#if !CONVERT_FROM_COMPLEX
    return aerror1("convert becomplex to float", a);
#else
    cr = new_float();
    stackcheck();
    push(cr);
    ci = new_float();
    pop(cr);
    float_val(cr) = r;
    float_val(ci) = i;
    return Lmkcomplex(cr, ci);
#endif
}

Sex
floatitc(Sex a)
{
    Sex             cr,
                    ci;
    double          r,
                    i;
    r = floatit(complex_real(a));
    i = floatit(complex_imag(a));
    if (i == 0.0)
	return make_float(r);
#if !CONVERT_FROM_COMPLEX
    return aerror1("convert complex to float", a);
#else
    cr = new_float();
    stackcheck();
    push(cr);
    ci = new_float();
    pop(cr);
    float_val(cr) = r;
    float_val(ci) = i;
    return Lmkcomplex(cr, ci);
#endif
}

double
floatit(Sex a)
{
    if (is_fixnum(a)) {
	return (double) mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return floatitl(a);
	case N_BIGNUM:
	    return floatitb(a);
	case N_EFLOAT:
	    return efloat_value(a);
	case N_BFLOAT:
	    return floatitd(a);
	case N_INTERVAL:
	    return (interval_lower(a) + interval_upper(a)) * 0.5;
	case N_RATIONAL:
	    return floatitr(a);
	case N_FLOAT:
	    return float_val(a);
	case N_BEFLOAT:
	    return floatitE(a);
	case N_BINTERVAL:
	    return floatitV(a);
	case N_COMPLEX:
	    return floatitc(a);
	case N_ECOMPLEX:
	    return floatitz(a);
	case N_BECOMPLEX:
	    return floatitZ(a);
	}
    }
    return aerror1("floatit", a);
}

Sex
Lfloat(Sex a)
{
    double          f;

    if (is_fixnum(a)) {
	f = mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    f = floatitl(a);
	    break;
	case N_BIGNUM:
	    f = floatitb(a);
	    break;
	case N_EFLOAT:
	    f = efloat_value(a);
	    break;
	case N_BFLOAT:
	    f = floatitd(a);
	    break;
	case N_INTERVAL:
	    f = (interval_lower(a) + interval_upper(a)) * 0.5;
	    break;
	case N_RATIONAL:
	    f = floatitr(a);
	    break;
	case N_FLOAT:
	    return a;
	case N_COMPLEX:
	    return floatitc(a);
	case N_ECOMPLEX:
	    return floatitz(a);
	case N_BECOMPLEX:
	    return floatitZ(a);
	case N_BEFLOAT:
	    f = floatitE(a);
	    break;
	case N_BINTERVAL:
	    f = floatitV(a);
	    break;
	default:
	    return aerror1("floatit", a);
	}
    } else {
	return aerror1("floatit", a);
    }
    a = new_float();
    float_val(a) = f;
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2if(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = float_val(b);
    c = new_float();
    float_val(c) = wa + wb;
    return c;
}

Sex
plus2fi(Sex b, Sex a)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = float_val(b);
    c = new_float();
    float_val(c) = wa + wb;
    return c;
}


Sex
plus2ff(Sex a, Sex b)
{
    Sex             c;
    double          wc = float_val(a) + float_val(b);
    c = new_float();
    float_val(c) = wc;
    return c;
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
diff2if(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a) - float_val(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}



Sex
diff2fi(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a) - floatit(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}


Sex
diff2ff(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a) - float_val(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}



/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2fi(Sex b, Sex a)
{
    Sex             c;
    double          wa = floatit(a) * float_val(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}

Sex
times2if(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a) * float_val(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}

Sex
times2rf(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a) * float_val(b);
    c = new_float();
    float_val(c) = wa;
    return c;
}



Sex
times2ff(Sex a, Sex b)
{
    Sex             c;
    double          wc = float_val(a) * float_val(b);
    c = new_float();
    float_val(c) = wc;
    return c;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
quot2if(Sex a, Sex b)
{
    Sex             c;
    double          wa = floatit(a);
    double          wb = float_val(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    c = new_float();
    float_val(c) = wa / wb;
    return c;
}


Sex
quot2fi(Sex a, Sex b)
{
    Sex             c;
    double          wa = float_val(a);
    double          wb = floatit(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    c = new_float();
    float_val(c) = wa / wb;
    return c;
}


Sex
quot2ff(Sex a, Sex b)
{
    Sex             c;
    double          wb,
                    wc;
    wb = float_val(b);
    if (wb == 0.0)
	return aerror2("bad arg for quotient", a, b);
    wc = float_val(a) / wb;
    c = new_float();
    float_val(c) = wc;
    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
lessp2if(Sex a, Sex b)
{
    return ((double) mkint(a) < float_val(b));
}



bool
lessp2bf(Sex a, Sex b)
{
    double          x,
                    y;

    y = float_val(b);
    x = floatit(a);
    return (x < y);
}

bool
lessp2rf(Sex a, Sex b)
{
    return (floatit(a) < float_val(b));
}


bool
lessp2fi(Sex a, Sex b)
{
    return (float_val(a) < (double) mkint(b));
}

bool
lessp2fb(Sex a, Sex b)
{
    double          x,
                    y;

    x = float_val(a);
    y = floatit(b);
    return (x < y);
}


bool
lessp2fr(Sex a, Sex b)
{
    return (float_val(a) < floatit(b));
}

bool
lessp2ff(Sex a, Sex b)
{
    return (float_val(a) < float_val(b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/





bool
numeq2if(Sex a, Sex b)
{
    return ((double) mkint(a) == float_val(b));
}

bool
numeq2bf(Sex a, Sex b)
{
    return (floatitb(a) == float_val(b));
}

bool
numeq2rf(Sex a, Sex b)
{
    return (floatitr(a) == float_val(b));
}



bool
numeq2ff(Sex a, Sex b)
{
    return (float_val(a) == float_val(b));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
#define FLOAT_2_52	4503599627370496.0e0

Sex
Lroundd(Sex a)
{
    Sex             b;
    stackcheck();
    push(a);
    b = bfloatitf(0.5);
    pop(a);
    if (is_bigneg(bfloat_man(a)))
	a = diff2dd(a, b);
    else
	a = plus2dd(a, b);
    return fixitd(a);
}

Sex
Lceilingd(Sex a)
{
    if (!is_bigneg(bfloat_man(a)) && is_decimal_part(a)) {
	a = fixitd(a);
	return Ladd1(a);
    }
    return fixitd(a);
}

Sex
Lfloord(Sex a)
{
    if (is_bigneg(bfloat_man(a)) && is_decimal_part(a)) {
	a = fixitd(a);
	return Lsub1(a);
    }
    return fixitd(a);
}


long
is_decimal_part(Sex a)
{
    long            i,
                    n,
                    m;
    long            l;

    l = bfloat_exp(a);
    a = bfloat_man(a);
    if (l >= 0)
	return 0;

    l = -l;
    n = l >> LP_LOG2;
    m = l & (LPSIZE - 1);
    l = bnum_n(numlen(a)) - 1;
    for (i = 0; i < n; i++)
	if (bignum_value(a)[i])
	    return 1;
    if (n > l)
	return 0;
    if (m == 0)
	return 0;
    if (bignum_value(a)[n] << (LPSIZE - m))
	return 1;
    return 0;
}



Sex
Lroundf(double v)
{
#if defined(Linux) || defined(VisualC)
    if (v > 0)
	return fixitf(v + 0.5);
    else
	return fixitf(v - 0.5);
#else
    return fixitf(round(v));
#endif
}


#ifdef _LP64
Sex
Lfloorf_sub(s, a1, exp)
     long            s,
                     exp;
     unsigned long
                     a1;
{
    long            f;
    long            len,
                    i;
    unsigned long   a2;
    Sex             c;


    f = exp >> LP_LOG2;		/* 
				 * f should  be greater or equal to 32 
				 */
    exp = exp & (LPSIZE - 1);
    if (exp == 0) {
	a2 = 0;
    } else {
	a2 = a1 >> (LPSIZE - exp);
	a1 = (a1 << exp);
    }
    len = bnum_len(f + 1);
    if (a2)
	len += SEX;
    c = new_binteger(s < 0 ? T_BIGNEG : T_BIGNUM, len);
    for (i = 0; i < f; i++)
	bignum_value(c)[i] = 0;
    bignum_value(c)[f] = a1;
    if (a2)
	bignum_value(c)[f + 1] = a2;
    return c;
}

Sex
Lceilingf(double v)
{				/* 
				 * ceiling the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a1 = ieee_word(u);
    f = ieee_signum(u);
    if (exp < 0) {
	if (v > 0.0)
	    return mkfxat(1);
	return mkfxat(0);
    }
    a0 = 0;
    if (exp <= 52) {
	if (exp < 52) {
	    a0 = a1 << (12 + exp);
	    a1 = a1 >> (52 - exp);
	}
	if (f < 0)
	    a1 = -a1;
	else if (a0)
	    a1 = a1 + 1;
	return mkfxat(a1);
    }
    exp = exp - 52;
    if (exp < 12) {
	a1 = a1 << exp;
	if (f < 0)
	    return mkbigat1s(-1L, a1);
	else if (a0) {
	    a1 = a1 + 1;
	    if (a1 == 0)
		return mkbigat(1L, 0L);
	}
	return mkbigat1s(1L, a1);
    }
    return Lfloorf_sub(f, a1, exp);
}


Sex
fixitf(double v)
{				/* 
				 * truncate the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a1 = ieee_word(u);
    f = ieee_signum(u);
    if (exp < 0) {
	return mkfxat(0);
    }
    a0 = 0;
    if (exp <= 52) {
	if (exp < 52) {
	    a0 = a1 << (12 + exp);
	    a1 = a1 >> (52 - exp);
	}
	if (f < 0)
	    a1 = -a1;
	return mkfxat(a1);
    }
    exp = exp - 52;
    if (exp < 12) {
	a1 = a1 << exp;
	if (f < 0)
	    return mkbigat1s(-1L, a1);
	return mkbigat1s(1L, a1);
    }
    return Lfloorf_sub(f, a1, exp);
}

Sex
Lfloorf(double v)
{				/* 
				 * floor the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a1 = ieee_word(u);
    f = ieee_signum(u);
    if (exp < 0) {
	if (v < 0.0)
	    return mkfxat(-1);
	return mkfxat(0);
    }
    a0 = 0;
    if (exp <= 52) {
	if (exp < 52) {
	    a0 = a1 << (12 + exp);
	    a1 = a1 >> (52 - exp);
	}
	if (f < 0) {
	    a1 = -a1;
	    if (a0)
		a1 = a1 - 1;
	}
	return mkfxat(a1);
    }
    exp = exp - 52;
    if (exp < 12) {
	a1 = a1 << exp;
	if (f > 0)
	    return mkbigat1s(1L, a1);
	else if (a0) {
	    a1 = a1 + 1;
	    if (a1 == 0)
		return mkbigat(-1L, 0L);
	}
	return mkbigat1s(-1L, -a1);
    }
    return Lfloorf_sub(f, a1, exp);
}

#else

Sex
Lfloorf_sub(s, a0, a1, exp)
     long            s,
                     exp;
     unsigned long   a0,
                     a1;
{
    long            f;
    long            len,
                    i;
    unsigned long   a2;
    Sex             c;


    f = exp >> LP_LOG2;		/* 
				 * f should  be greater or equal to 32 
				 */
    exp = exp & (LPSIZE - 1);
    if (exp == 0) {
	a2 = 0;
    } else {
	a2 = a1 >> (LPSIZE - exp);
	a1 = (a1 << exp) | (a0 >> (LPSIZE - exp));
	a0 = a0 << exp;
    }
    len = bnum_len(f + 1);
    if (a2)
	len += SEX;
    c = new_binteger(s < 0 ? T_BIGNEG : T_BIGNUM, len);
    for (i = 0; i < f; i++)
	bignum_value(c)[i] = 0;
    if (f > 0)
	bignum_value(c)[f - 1] = a0;
    bignum_value(c)[f] = a1;
    if (a2)
	bignum_value(c)[f + 1] = a2;
    return c;
}

Sex
Lceilingf(double v)
{				/* 
				 * ceiling the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a0 = ieee_loword(u);
    a1 = ieee_hiword(u);
    f = ieee_signum(u);
    if (exp < 0) {
	if (v > 0.0)
	    return mkfxat(1);
	return mkfxat(0);
    }
    if (exp == 20) {
	if (f < 0)
	    a1 = -a1;
	else if (a0)
	    a1 = a1 + 1;
	return mkfxat(a1);
    }
    if (exp < 20) {
	a0 = a1 << (12 + exp);
	a1 = a1 >> (20 - exp);
	if (f < 0)
	    a1 = -a1;
	else if (a0)
	    a1 = a1 + 1;
	return mkfxat(a1);
    }
    exp = exp - 20;
    if (exp < 12) {
	a1 = (a1 << exp) | (a0 >> (32 - exp));
	a0 = a0 << exp;
	if (f < 0)
	    return mkbigat1s(-1, a1);
	else if (a0) {
	    a1 = a1 + 1;
	    if (a1 == 0)
		return mkbigat(1, 0);
	}
	return mkbigat1s(1, a1);
    }
    return Lfloorf_sub(f, a0, a1, exp);
}


Sex
fixitf(double v)
{				/* 
				 * truncate the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a0 = ieee_loword(u);
    a1 = ieee_hiword(u);
    f = ieee_signum(u);
    if (exp < 0)
	return mkfxat(0);
    if (exp == 20) {
	if (f < 0)
	    a1 = -a1;
	return mkfxat(a1);
    }
    if (exp < 20) {
	a0 = a1 << (12 + exp);
	a1 = a1 >> (20 - exp);
	if (f < 0)
	    a1 = -a1;
	return mkfxat(a1);
    }
    exp = exp - 20;
    if (exp < 12) {
	a1 = (a1 << exp) | (a0 >> (32 - exp));
	a0 = a0 << exp;
	if (f < 0)
	    return mkbigat1s(-1, a1);
	return mkbigat1s(1, a1);
    }
    return Lfloorf_sub(f, a0, a1, exp);
}

Sex
Lfloorf(double v)
{				/* 
				 * floor the floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1;

    u.d = v;
    exp = ieee_exponent(u);
    a0 = ieee_loword(u);
    a1 = ieee_hiword(u);
    f = ieee_signum(u);
    if (exp < 0) {
	if (v < 0.0)
	    return mkfxat(-1);
	return mkfxat(0);
    }
    if (exp == 20) {
	if (f < 0) {
	    a1 = -a1;
	    if (a0)
		a1 = a1 - 1;
	}
	return mkfxat(a1);
    }
    if (exp < 20) {
	a0 = a1 << (12 + exp);
	a1 = a1 >> (20 - exp);
	if (f < 0) {
	    a1 = -a1;
	    if (a0)
		a1 = a1 - 1;
	}
	return mkfxat(a1);
    }
    exp = exp - 20;
    if (exp < 12) {
	a1 = (a1 << exp) | (a0 >> (32 - exp));
	a0 = a0 << exp;
	if (f < 0) {
	    a1 = -a1;
	    if (a0) {
		a1 = a1 - 1;
		if (a1 == 0)
		    return mkbigat(-1, 0);
	    }
	    return mkbigat1s(-1, -a1);
	}
	return mkbigat1s(1, a1);
    }
    return Lfloorf_sub(f, a0, a1, exp);
}
#endif

#if 0
Sex
Lceilingf(double x)
{
    double          y;
    if (x < 0) {
	x = -x;
	if (x < FLOAT_2_52) {
	    y = FLOAT_2_52 + x;
	    y -= FLOAT_2_52;
	    x = x < y ? y - 1 : y;
	}
	x = -x;
    } else {
	if (x < FLOAT_2_52) {
	    y = FLOAT_2_52 + x;
	    y -= FLOAT_2_52;
	    x = x > y ? y + 1 : y;
	}
    }
    return fixitf(x);
}

Sex
Lfloorf(double x)
{
    double          y;
    if (x < 0) {
	x = -x;
	if (x < FLOAT_2_52) {
	    y = FLOAT_2_52 + x;
	    y -= FLOAT_2_52;
	    x = x > y ? y + 1 : y;
	}
	x = -x;
    } else {
	if (x < FLOAT_2_52) {
	    y = FLOAT_2_52 + x;
	    y -= FLOAT_2_52;
	    x = x < y ? y - 1 : y;
	}
    }
    return fixitf(x);
}


Sex
fixitf(double v)
{				/* 
				 * truncate floating number 
				 */
    DoubleFloat     u;

    long            exp,
                    f;
    unsigned long   a0,
                    a1,
                    a2;
    long            i,
                    len;
    Sex             c;

    u.d = v;
    exp = ieee_exponent(u);
    a0 = ieee_loword(u);
    a1 = ieee_hiword(u);
    if (exp < 0)
	return mkfxat(0);
    if (exp == 20) {
	if (ieee_signum(u) < 0)
	    a1 = -a1;
	return mkfxat(a1);
    }
    if (exp < 20) {
	a0 = a1 << (12 + exp);
	a1 = a1 >> (20 - exp);
	if (ieee_signum(u) < 0)
	    a1 = -a1;
	return mkfxat(a1);
    }
    exp = exp - 20;
    if (exp < 12) {
	a1 = (a1 << exp) | (a0 >> (32 - exp));
	a0 = a0 << exp;
	if (ieee_signum(u) < 0)
	    return mkbigat1s(-1, a1);
	return mkbigat1s(1, a1);
    }
    f = exp >> 5;		/* 
				 * f should  be greater or equal to 32 
				 */
    exp = exp & 31;
    if (exp == 0) {
	a2 = 0;
    } else {
	a2 = a1 >> (32 - exp);
	a1 = (a1 << exp) | (a0 >> (32 - exp));
	a0 = a0 << exp;
    }
    len = (f << 2) + 8 + 4;
    if (a2)
	len += 4;
    c = new_binteger(ieee_signum(u) < 0 ? T_BIGNEG : T_BIGNUM, len);
    for (i = 0; i < f; i++)
	bignum_value(c)[i] = 0;
    if (f > 0)
	bignum_value(c)[f - 1] = a0;
    bignum_value(c)[f] = a1;
    if (a2)
	bignum_value(c)[f + 1] = a2;
    return c;
}

#endif

Sex
Lceiling(Sex a)
{
    Sex             b;
    double          cr,
                    ci;

    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	stackcheck();
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_FLOAT:
	    return Lceilingf(float_val(a));
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_BFLOAT:
	    return Lceilingd(a);
	case N_EFLOAT:
	    return Lceilingf(efloat_value(a));
	case N_INTERVAL:
	    return Lceilingf((interval_lower(a) + interval_upper(a)) *
			     0.5);
	case N_BEFLOAT:
	    return Lceilingd(befloat_value(a));
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    a = leftshiftd(a, -1);
	    return Lceilingd(a);
	case N_COMPLEX:
	    b = complex_imag(a);
	    a = complex_real(a);
	    push(b);
	    a = Lceiling(a);
	    pop(b);
	    push(a);
	    b = Lceiling(b);
	    pop(a);
	    return Lmkcomplex(a, b);
	case N_ECOMPLEX:
	    cr = ecomplex_real(a);
	    ci = ecomplex_imag(a);
	    a = Lceilingf(cr);
	    push(a);
	    b = Lceilingf(ci);
	    pop(a);
	    return Lmkcomplex(b, a);
	case N_BECOMPLEX:
	    b = becomplex_imag(a);
	    a = becomplex_real(a);
	    push(b);
	    a = Lceilingd(a);
	    pop(b);
	    push(a);
	    b = Lceilingd(b);
	    pop(a);
	    return Lmkcomplex(a, b);
	}
    }
    return error1("bad argument to Lceiling", a);
}



Sex
Lround(Sex a)
{
    Sex             b;
    double          cr,
                    ci;

    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	stackcheck();
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_FLOAT:
	    return Lroundf(float_val(a));
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_BFLOAT:
	    return Lroundd(a);
	case N_EFLOAT:
	    return Lroundf(efloat_value(a));
	case N_INTERVAL:
	    return Lroundf((interval_lower(a) + interval_upper(a)) * 0.5);
	case N_BEFLOAT:
	    return Lroundd(befloat_value(a));
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    a = leftshiftd(a, -1);
	    return Lroundd(a);
	case N_COMPLEX:
	    push(a);
	    b = Lround(complex_real(a));
	    pop(a);
	    push(b);
	    a = Lround(complex_imag(a));
	    pop(b);
	    return Lmkcomplex(b, a);
	case N_ECOMPLEX:
	    cr = ecomplex_real(a);
	    ci = ecomplex_imag(a);
	    a = Lroundf(cr);
	    push(a);
	    b = Lroundf(ci);
	    pop(a);
	    return Lmkcomplex(b, a);
	case N_BECOMPLEX:
	    b = becomplex_imag(a);
	    a = becomplex_real(a);
	    push(b);
	    a = Lroundd(a);
	    pop(b);
	    push(a);
	    b = Lroundd(b);
	    pop(a);
	    return Lmkcomplex(a, b);
	}
    }
    return error1("bad argument to Lround", a);
}

Sex
Lfloor(Sex a)
{
    Sex             b;
    double          cr,
                    ci;

    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	stackcheck();
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_FLOAT:
	    return Lfloorf(float_val(a));
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_BFLOAT:
	    return Lfloord(a);
	case N_EFLOAT:
	    return Lfloorf(efloat_value(a));
	case N_INTERVAL:
	    return Lfloorf((interval_lower(a) + interval_upper(a)) * 0.5);
	case N_BEFLOAT:
	    return Lfloord(befloat_value(a));
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    a = leftshiftd(a, -1);
	    return Lfloord(a);
	case N_COMPLEX:
	    b = complex_imag(a);
	    a = complex_real(a);
	    push(b);
	    a = Lfloor(a);
	    pop(b);
	    push(a);
	    b = Lfloor(b);
	    pop(a);
	    return Lmkcomplex(a, b);
	case N_ECOMPLEX:
	    cr = ecomplex_real(a);
	    ci = ecomplex_imag(a);
	    a = Lfloorf(cr);
	    push(a);
	    b = Lfloorf(ci);
	    pop(a);
	    return Lmkcomplex(b, a);
	case N_BECOMPLEX:
	    b = becomplex_imag(a);
	    a = becomplex_real(a);
	    push(b);
	    a = Lfloord(a);
	    pop(b);
	    push(a);
	    b = Lfloord(b);
	    pop(a);
	    return Lmkcomplex(a, b);
	}
    }
    return error1("bad argument to Lfloor", a);
}


Sex
Ltruncate(Sex a)
{
    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return fixitl(a);
	case N_BIGNUM:
	    return a;
	case N_FLOAT:
	    return fixitf(float_val(a));
	case N_RATIONAL:
	    a = bfloatit(a);
	case N_BFLOAT:
	    return fixitd(a);
	case N_EFLOAT:
	    return fixitf(efloat_value(a));
	case N_INTERVAL:
	    return fixitf((interval_lower(a) + interval_upper(a)) * 0.5);
	case N_BEFLOAT:
	    return fixitd(befloat_value(a));
	case N_BINTERVAL:
	    return fixitV(a);
	case N_COMPLEX:
	    return fixitc(a);
	case N_ECOMPLEX:
	    return fixitz(a);
	case N_BECOMPLEX:
	    return fixitZ(a);
	}
    }
    return error1("bad argument to Ltruncate", a);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 * needed for reduce 
 */

Sex
Lsafe_fp_plus(Sex a, Sex b)
{
    DoubleFloat     ff;
    double          fa,
                    fb;

#ifdef SAFE_FP
    if (!is_float(a) || !is_float(b))
	error2("invalid argument to safe_fp_plus", a, b);
#endif
    fa = float_val(a);
    fb = float_val(b);
    if (fa == 0.0)
	return b;
    if (fb == 0.0)
	return a;
    ff.d = fa + fb;
    if (ieee_exponent(ff) == 0x400)
	return nil;
    return make_float(ff.d);
}


Sex
Lsafe_fp_quot(Sex a, Sex b)
{
    double          fa,
                    fb;
    DoubleFloat     ff;

#ifdef SAFE_FP
    if (!is_float(a) || !is_float(b))
	error2("invalid argument to safe_fp_quot", a, b);
    fa = float_val(a);
    fb = float_val(b);
#else
    fa = floatit(a);
    fb = floatit(b);
#endif
    if (fa == 0.0 || fb == 1.0)
	return a;
    if (fb == 0.0)
	error2("invalid argument to safe_fp_quot", a, b);
    ff.d = fa / fb;
    if (ieee_exponent(ff) == 0x400)
	return nil;
    return make_float(ff.d);
}


Sex
Lsafe_fp_times(Sex a, Sex b)
{
    double          fa,
                    fb;
    DoubleFloat     ff;

#ifdef SAFE_FP
    if (!is_float(a) || !is_float(b))
	error2("invalid argument to safe_fp_times", a, b);
    fa = float_val(a);
    fb = float_val(b);
#else
    fa = floatit(a);
    fb = floatit(b);
#endif
    fa = float_val(a);
    fb = float_val(b);
    if (fa == 0.0)
	return a;
    if (fb == 0.0)
	return b;
    if (fa == 1.0)
	return b;
    if (fb == 1.0)
	return a;
    ff.d = fa * fb;
    if (ieee_exponent(ff) == 0x400)
	return nil;
    return make_float(ff.d);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/
