/*
 * gc.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * Garbage  Collections
 *      $Header: /Users/kako/asuka/src/RCS/gc.c,v 2.33 2017/11/27 00:45:12 kako Exp $
 */

static char const rcsid[] =
    "$Id: gc.c,v 2.33 2017/11/27 00:45:12 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "code.h"
#include "gc.h"

#ifdef ASUKA
#if 0
#define yyvs  yyv
#define yyvsp yypv
#endif
#endif

static void     mark_sysvars(void);
static void     relocate_sysvars(void);
static void     mark_bps(void);
static void     relocate_bps(void);
static void     mark_x(Sex *);
static void     relocate_x(Sex *);

#ifdef ASUKA
extern Sex      F5_Rules;
extern Sex      star_evaluation_points;
#endif

Sex            *literal_vectors[100];
long            number_of_literal_vectors = 0;
long            number_of_gc_called = 0;

/***************************************************************/

void            mark(Sex * pp);
void            relocate(Sex * cp);
static void     relocate_conscell(void);
void            nullptr(Sex * p);
extern Sex      throw_fn(Sex u, Sex e);

/***************************************************************/
/***************************************************************/

#ifndef _LP64

#if MEMORY_MODEL == MEMORY_MIXED

extern int      memory_low;

int
mixed_set_p_mark(int w)
{
    w = w & 0x7fffffff;
    if (w < memory_low)
	w = w | 0x80000000;
    return w;
}

int
mixed_reset_p_mark(int w)
{
    w = w & 0x7fffffff;
    if (w >= memory_low)
	w = w | 0x80000000;
    return w;
}

int
mixed_is_p_marked(int w)
{
    int             u = w & 0x7fffffff;
    if (u < memory_low) {
	if (w & 0x80000000)
	    return 1;
    } else if ((w & 0x80000000) == 0)
	return 1;
    return 0;
}

#endif

#if MEMORY_MODEL == MEMORY_VARIABLE

extern int      memory_model;
extern int      memory_low;

int
variable_set_p_mark(int w)
{
    switch (memory_model) {
    case MEMORY_LOW:
	return (int) w | GC_BIT_P;
    case MEMORY_HIGH:
	return (int) w & GC_BIT_P;
    case MEMORY_MIXED:
	w = w & 0x7fffffff;
	if (w < memory_low)
	    w = w | 0x80000000;
	return w;
    default:
	return error("gc: invalid memory mode\n");
    }
}

int
variable_reset_p_mark(int w)
{
    switch (memory_model) {
    case MEMORY_LOW:
	return (int) w & ~GC_BIT_P;
    case MEMORY_HIGH:
	return (int) w | ~GC_BIT_P;
    case MEMORY_MIXED:
	w = w & 0x7fffffff;
	if (w >= memory_low)
	    w = w | 0x80000000;
	return w;
    default:
	return error("gc: invalid memory mode\n");
    }
}

int
variable_is_p_marked(int w)
{
    int             u;
    switch (memory_model) {
    case MEMORY_LOW:
	return ((int) (w) < 0);
    case MEMORY_HIGH:
	return ((int) (w) > 0);
    case MEMORY_MIXED:
	u = w & 0x7fffffff;
	if (u < memory_low) {
	    if (w & 0x80000000)
		return 1;
	} else if ((w & 0x80000000) == 0)
	    return 1;
	return 0;
    default:
	return error("gc: invalid memory mode\n");
    }
}

#endif
#endif

/***************************************************************/
/***************************************************************/


void
mark(Sex * pp)
{
    Sex             p,
                    q;
    Sex            *sp = stack,
	*sl = stacklimit;
    Tag             h;
    long            i;
    Sex            *spsave;

    *++sp = (Sex) NULL;
    if (sp >= sl)
	fatal_error("stack over while gbc...");
  top:
    if (pp == NULL)
	return;
  top2:
    p = *pp;
    switch ((long) p & TAG_BITS) {
    case TAG_CONS:
	p = reset_p_mark(p);
	q = car(p);
	if (is_immed(q)) {
	    if (is_i_marked(q)) {
		pp = (Sex *) (*sp--);
		goto top;
	    }
	    car(p) = set_i_mark(q);
	    pp = &cdr(p);
	    goto top2;
	} else if (is_p_marked(q)) {
	    pp = (Sex *) (*sp--);
	    goto top;
	} else {
	    car(p) = set_p_mark(q);
	    q = cdr(p);
	    if (!is_immed(q) && q != nil) {
		if (sp >= sl)
		    fatal_error("stack over while gbc...");
		*++sp = (Sex) & cdr(p);
	    }
	    pp = &car(p);
	    goto top2;
	}
    case TAG_RELOC:
    case TAG_FIXNUM:
    case TAG_HDRS:
	pp = (Sex *) (*sp--);
	goto top;
    case TAG_SYMBOL:
	p = reset_p_mark(p);
	if (p == nil) {
	    pp = (Sex *) (*sp--);
	    goto top;
	}
	h = qtag(p);
	if (!is_hdrs(h)) {	/* 
				 * already visited 
				 */
	    *pp = set_p_mark(h);
	    qtag(p) = (Tag) pp;
	    pp = (Sex *) (*sp--);
	    goto top;
	}
	*pp = set_i_mark(h);
	qtag(p) = (Tag) pp;
	if (sp >= sl)
	    fatal_error("stack over while gbc...");
	q = value(p);
	if (!is_immed(q) && q != nil)
	    *++sp = (Sex) & value(p);
	if ((SYM_PROC & h) != 0 && ((PROC_CCODE | PROC_BCODE) & h) == 0)
	    *++sp = (Sex) & proc(p);
#ifdef ALGEBRAIC_PROPERTY
	q = gvalue(p);
	if (!is_immed(q) && q != nil)
	    *++sp = (Sex) & gvalue(p);
	if ((SYM_APROC & h) != 0 && (GPROC_CCODE & h) == 0)
	    *++sp = (Sex) & gproc(p);
#endif
	pp = &plist(p);
	goto top2;

    case TAG_NUMBERS:
	p = reset_p_mark(p);
	h = numhdr(p);
	if (!is_hdrs(h)) {	/* 
				 * already visited 
				 */
	    *pp = set_p_mark(h);
	    numhdr(p) = (Tag) pp;
	    pp = (Sex *) (*sp--);
	    goto top;
	}
	*pp = set_i_mark(h);
	numhdr(p) = (Tag) pp;
	switch (numhdr_of(h)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_FLOAT:
	case N_EFLOAT:
	case N_ECOMPLEX:
	case N_INTERVAL:
	case N_MODULAR:
	case N_QMODULAR:
	case N_FMODULAR:
	    pp = (Sex *) (*sp--);
	    goto top;
	case N_BEFLOAT:
	    pp = (Sex *) ((long) p + 2 * SEX - TAG_NUMBERS);
	    goto top2;
	case N_BMODULAR:
	case N_RATIONAL:
	case N_COMPLEX:
	case N_BINTERVAL:
	    if (sp >= sl)
		fatal_error("stack over while gbc...");
	    *++sp = (Sex) ((long) p + 2 * SEX - TAG_NUMBERS);
	case N_BFLOAT:
	    pp = (Sex *) ((long) p + 3 * SEX - TAG_NUMBERS);
	    goto top2;
	case N_BECOMPLEX:
	    if (sp >= sl)
		fatal_error("stack over while gbc...");
	    *++sp = (Sex) ((long) p + 2 * SEX - TAG_NUMBERS);
	    pp = (Sex *) ((long) p + 3 * SEX - TAG_NUMBERS);
	    goto top2;
	default:
	    lisp_printf("Bad numeric type found %08x\n", (long) h);
	    abort();
	    return;
	}

    case TAG_BINARY:
	p = reset_p_mark(p);
	h = binhdr(p);
	if (!is_hdrs(h)) {
	    *pp = set_p_mark(h);
	} else {
	    *pp = set_i_mark(h);
	}
	binhdr(p) = (Tag) pp;
	pp = (Sex *) (*sp--);
	goto top;

    case TAG_VECTOR:
	p = reset_p_mark(p);
	h = vechdr(p);
	if (!is_hdrs(h)) {
	    *pp = set_p_mark(h);
	    vechdr(p) = (Tag) pp;
	    pp = (Sex *) (*sp--);
	    goto top;
	}
	*pp = set_i_mark(h);
	vechdr(p) = (Tag) pp;
	i = (long) alignits(veclen(p));
	spsave = stack;
	stack = sp;
	switch (type_of_header(h)) {
	case T_V_ARRAY:
	    while (i >= 4 * SEX) {
		i -= SEX;
		pp = (Sex *) ((char *) p - TAG_VECTOR + i);
		if (!is_immed(*pp))
		    mark(pp);
	    }
	    break;
	default:
	    while (i >= 3 * SEX) {
		i -= SEX;
		pp = (Sex *) ((char *) p - TAG_VECTOR + i);
		if (!is_immed(*pp))
		    mark(pp);
	    }
	    break;

	}
	stack = spsave;
	pp = (Sex *) (*sp--);
	goto top;
    }
}


Sex
Lgc()
{
    return reclaim(nil, GC_USER, 0);
}

Sex
Lverbos(Sex a)
{
    long            code,
                    old_code = verbos_flag;
    if (a == nil)
	code = 0;
    else if (is_fixnum(a))
	code = mkint(a);
    else
	code = 1;
    miscflags = (miscflags & ~(GC_MSG_BITS | LOAD_ECHO_FLAG)) |
	(code & (GC_MSG_BITS | LOAD_ECHO_FLAG));
    return mkfxat(old_code);
}


bool            already_in_gc;
bool            interrupt_pending;


static void
compress_cell(void)
{
    Cons_Cell      *bottom = cellbase;
    Cons_Cell      *top = cellpointer - 1;
    Sex             p;
#ifdef DEBUG
    printf("compress cell: 0x%08lx -- 0x%08lx\n", (long) bottom,
	   (long) top);
#endif
    while (YES) {
	top++;
	while (YES) {
	    bottom--;
	    if (bottom < top) {
		cellpointer = top;
		return;
	    }
	    p = car(bottom);
	    if (is_immed(p)) {
		if (is_i_marked(p))
		    car(bottom) = reset_i_mark(p);
		else
		    break;
	    } else if (is_p_marked(p))
		car(bottom) = reset_p_mark(p);
	    else
		break;

	}
	top--;
	while (YES) {
	    top++;
	    if (bottom < top) {
		cellpointer = top;
		return;
	    }
	    p = car(top);
	    if (is_immed(p)) {
		if (is_i_marked(p)) {
		    car(top) = reset_i_mark(p);
		    break;
		}
	    } else if (is_p_marked(p)) {
		car(top) = reset_p_mark(p);
		break;
	    }
	}
	car(bottom) = car(top);
	cdr(bottom) = cdr(top);
	car(top) = set_p_mark((Sex) bottom + TAG_RELOC);
	cdr(top) = set_p_mark((Sex) bottom + SEX + TAG_RELOC);
    }
}

static void
adjust_vec_heap(void)
{
    char           *bottom = (char *) fwdbase;
    char           *top = (char *) fwdpointer;
    char           *p = bottom;
    Tag             h;
    Sex             h1,
                    h2;
    char           *p1;
    long            l;
    long            t_type[4] = { TAG_BINARY, TAG_SYMBOL,
	TAG_NUMBERS, TAG_VECTOR
    };

    while (bottom < top) {
	if (is_hdrs(h = *(Tag *) bottom)) {
	    bottom += alignit(*(Sex *) (bottom + SEX));
	    continue;
	}
	h1 = h;
	l = alignit(*(Sex *) (bottom + SEX));

	while (!is_hdrs(h1)) {
	    h1 = reset_p_mark(h1);
	    h2 = *(Sex *) h1;
	    /*
	     * the following code can be h1 = (Sex)h2 & ~3 
	     */
	    if (is_relhdr(h2))	/* 
				 * * * * * * relocated cons cell      
				 */
		h1 = (Sex) ((char *) h2 - TAG_RELOC);
	    else
		h1 = h2;
	}

	/*
	 * 
	 * if (is_symbol_header(h1)) p1 = p + TAG_SYMBOL; else if
	 * (is_numhdr_header(h1)) p1 = p + TAG_NUMBERS; else if
	 * (is_binhdr_header(h1)) p1 = p + TAG_BINARY; else p1 = p + TAG_VECTOR; 
	 */


	p1 = p + t_type[(h1 >> 4) & 3];


	while (!is_hdrs(h)) {
	    h = reset_p_mark(h);
	    h1 = *(Sex *) h;
	    if (is_relhdr(h1))	/* 
				 * * * * * * relocated cons cell      
				 */
		h = (Sex) ((char *) h1 - TAG_RELOC);
	    else {
		*(Sex *) h = (Sex) p1;
		h = h1;
	    }
	}
	*(Sex *) bottom = set_i_mark(h);
	bottom += l;
	p += l;
    }
}

static void
move_vec_heap(void)
{
    Sex            *top = (Sex *) fwdpointer;
    Sex            *bottom = (Sex *) fwdbase;
    Sex            *p = bottom;
    Tag             h;
    long            l;

    while (bottom < top) {
	if (!is_i_marked(h = (Tag) bottom[0])) {
	    bottom += alignit(bottom[1]) / SEX;
	    continue;
	}
	l = alignit(bottom[1]) / SEX;
	*p = reset_i_mark(h);
	if (p != bottom) {
	    p++;
	    bottom++;
	    while (--l != 0) {
		*p++ = *bottom++;
	    }
	} else {
	    p += l;
	    bottom += l;
	}
    }
    fwdpointer = p;
#ifdef CLEAR_FREE_AREA
    while (p < top)
	*p++ = 0;
#endif
}


void
relocate(Sex * cp)
{
    Sex             p = (*cp);
    if (p == nil)
	return;
    else if (is_cons(p)) {
	p = car(p);
	if (is_relhdr(p) && is_p_marked(p))
	    *cp = reset_p_mark(p - TAG_RELOC + TAG_CONS);
    }
}

static void
relocate_conscell(void)
{
    Cons_Cell      *bottom = cellbase;
    Cons_Cell      *top = cellpointer;

#ifdef DEBUG
    printf("relocate: cell = 0x%08lx -- 0x%08lx\n", (long) bottom,
	   (long) top);
#endif
    for (; top < bottom; top++) {
	relocate(&car(top));
	relocate(&cdr(top));
    }
}

static void
relocate_vecheap(void)
{
    Sex            *top = fwdpointer;
    Sex            *bottom = fwdbase;
    long            i,
                    l,
                    ll;
    Tag             h;
    Symbol_Head    *s;

    while (bottom < top) {
	h = *(Tag *) & bottom[0];
	l = *(Sex *) & bottom[1];
	if (is_symbol_header(h)) {
	    s = (Symbol_Head *) bottom;
	    if ((h & SYM_PROC) != 0
		&& (h & (PROC_CCODE | PROC_BCODE)) == 0)
		relocate((Sex *) & (s->proc));
	    relocate(&(s->value));
#ifdef ALGEBRAIC_PROPERTY
	    if ((h & SYM_APROC) != 0 && (h & GPROC_CCODE) == 0)
		relocate((Sex *) & (s->gproc));
	    relocate(&(s->gvalue));
#endif
	    relocate(&(s->plist));
	} else {
	    switch (type_of_header(h)) {
	    case T_BFLOAT:
		relocate(&bottom[3]);
		break;
	    case T_BEFLOAT:
		relocate(&bottom[2]);
		break;
	    case T_BMODULAR:
	    case T_RATIONAL:
	    case T_COMPLEX:
	    case T_BINTERVAL:
		relocate(&bottom[2]);
		relocate(&bottom[3]);
		break;
	    case T_BECOMPLEX:
		relocate(&bottom[2]);
		relocate(&bottom[3]);
		break;
	    case T_SIMPLE:
	    case T_ARRAY:
	    case T_VAR:
	    case T_ALGNUM:
	    case T_TRNNUM:
	    case T_POLY:
	    case T_DIPOLY:
	    case T_ALGFUN:
	    case T_RPOLY:
	    case T_VECT:
	    case T_MATRIX:
	    case T_TENSOR:
	    case T_PREFIX:
	    case T_RELEXP:
	    case T_SETEXP:
	    case T_LOGEXP:
	    case T_TROPICAL:
	    case T_HASH:
	    case T_PACKAGE:
		ll = (l + (SEX - 1)) / SEX;
		for (i = 2; i < ll; i++)
		    relocate(&bottom[i]);
		break;
	    case T_V_ARRAY:
		l = *(Sex *) & bottom[2];
		ll = (l + (SEX - 1)) / SEX;
		for (i = 3; i < ll; i++)
		    relocate(&bottom[i]);
		break;
	    default:
		break;
	    }
	}
	bottom += (alignit(l)) / SEX;
    }
}




Sex
reclaim(Sex p, long stg_class, long size)
{
    Sex            *sp;
    long            fwd_needed = 0;
    long            freespace;

    double          tv0;

#ifdef DEBUG
    printf("reclaim: cell = 0x%08lx -- 0x%08lx, 0x%08lx\n",
	   (long) cellbase, (long) celllimit, (long) cellpointer);
    printf("reclaim: fwd = 0x%08lx -- 0x%08lx, 0x%08lx\n", (long) fwdbase,
	   (long) fwdlimit, (long) fwdpointer);
#endif

    // miscflags = miscflags | GC_MSG_BITS;

    if (stg_class == GC_FWD)
	fwd_needed = size;
    already_in_gc = YES;

    stackcheck();
    push(p);

    tv0 = get_time();
    number_of_gc_called++;


    if (verbos_flag > 0) {
	static char     c[3] = { 'u', 'c', 'f' };
	lisp_printf("\n   Now collecting(%c) ---", c[stg_class]);
    }
    /*
     * marking phase 
     */
    qtag(nil) = set_i_mark(qtag(nil));
    mark(&plist(nil));
    mark_sysvars();
    // mark_bps();
    for (sp = stack; sp > (Sex *) stackbase; sp--) {
	mark(sp);
    }
    for (sp = bstack; sp > (Sex *) bindbase; sp--) {
	mark(sp);
    }
    /*
     * salvage garbages 
     */

    qtag(nil) = reset_i_mark(qtag(nil));

    compress_cell();


    /*
     * relocation of full word area 
     */

    adjust_vec_heap();
    move_vec_heap();


    /*
     * relocation for moved cons cell 
     */


    relocate(&(plist(nil)));
    relocate_sysvars();
    // relocate_bps();

    for (sp = stack; sp > (Sex *) stackbase; sp--)
	relocate(sp);
    for (sp = bstack; sp > (Sex *) bindbase; sp--) {
	relocate(sp);
    }

    relocate_conscell();
    relocate_vecheap();

    /*
     * end of garbage collection 
     */

    gc_time += get_time() - tv0;

    pop(p);

    freespace = (long) cellpointer - (long) fwdpointer - SPARE;
    if (freespace < 0)
	fatal_error("in gbc. No space left at all");
    else if (freespace < fwd_needed)
	fwd_needed = -1;
    else {
	freespace = ((freespace - fwd_needed) / 2) & (-CELL);
	fwdlimit = (Sex *) ((long) fwdpointer + freespace);
	celllimit = (Cons_Cell *) ((long) cellpointer - freespace);
	/*
	 * printf("cell limit is decreased by %d ( 0x%08x)\n",
	 * (long)freespace, (long)freespace); printf("fwd limit =
	 * 0x%08lx\n", (long)fwdlimit); 
	 */
    }
    if (verbos_flag > 0) {
	float           x;
	x = (float) ((long) cellpointer - (long) fwdpointer);
	x = 1 - x / ((long) cellbase - (long) fwdbase);
	lisp_printf(" %dKcells + %dKw (%.2f%%) used, %d stack left.\n",
		    ((long) cellbase -
		     (long) cellpointer) / (1024 * CELL),
		    ((long) fwdpointer -
		     (long) fwdbase) / (1024 * SEX), x * 100,
		    ((long) stacklimit - (long) stack) / (2 * SEX));
    }
    poll_for_attn();		/* 
				 * Should cause ctrl-C checking 
				 */
    if (interrupt_pending) {
	interrupt_pending = NO;
	already_in_gc = NO;
	return interrupted(p);
    }
    already_in_gc = NO;

    if (fwd_needed < 0)
	error("full word space too short for allocation");

#if 0
    lisp_printf("trap_time = %f, tv = %f\n", trap_time, tv0 - base_time);
#endif
    // if (trap_time != 0 && tv0 - base_time - io_time - gc_time >
    // trap_time) {
    if (trap_time > 0 && tv0 - base_time > trap_time) {
	Sex             a;
	trap_time = 0;
	a = cons(time_out_symbol, nil);
	a = cons(time_out_symbol, a);
	throw_fn(a, nil);
    }
    return p;
}


static Sex     *system_vars[] = {
    &env,
    &true,
    &lambda,
    &funarg,
    &undefined_var,
    &evalhook,
    &applyhook,
    &exit_tag,
    &exit_val,
    &catch_tags,
    &current_package,
    &current_charvec,

    &sugar_charvec,
    &asuka_charvec,
    &current_sugar_package,
    &current_asuka_package,

    &uniq_cons_table,
    &uniq_string_table,
    &fake_cons_cell,
    &gensym_base,
    &io_workspace,
    &io_workspace2,
#ifdef SAVE_WORK_SAPCE
    &work_space,
#endif
    &supervisor,
    &standard_input,
    &standard_output,
    &standard_errout,
    &lisp_input,
    &lisp_output,
    &lisp_errout,
    &file_input,
    &file_output,
    &fsl_input,
    &fsl_output,
    &prompt_message,
    &lisp_prompt_message,
    &emsg_star,
    &current_big_modulus,
    &star_modulus,
    &star_bmodulus,
    &star_var_order,
    &pi_value,
    &e_value,
    &log2_value,
    &log10_value,
    &sqrt2_value,
    &sqrt5_value,
    &prime_list,
    &kernel_order,
    &dipoly_kernel_order,

    &asuka_mode,
    &plot_device,
    &plot_devices,

    &true_symbol,
    &false_symbol,
    &time_out_symbol,

#ifdef ASUKA
    &asuka_true_symbol,
    &asuka_false_symbol,
    &asuka_nil_symbol,
    &catch_symbol,
    &throw_symbol,
    &any_symbol,
    &switch_symbol,
    &case_symbol,
    &F5_Rules,
    &star_evaluation_points,
#endif


    &append_symbol,
    &input_symbol,
    &output_symbol,
    &aref_symbol,
    &bappend_symbol,
    &binput_symbol,
    &boutput_symbol,
    &car_symbol,
    &cdr_symbol,
    &caar_symbol,
    &cadr_symbol,
    &cdar_symbol,
    &cddr_symbol,
    &comma_at_symbol,
    &comma_symbol,
    &echo_symbol,
    &eecho_symbol,
    &alist_symbol,
    &eval_symbol,

    &subr_symbol,
    &fsubr_symbol,
    &expr_symbol,
    &fexpr_symbol,
    &macro_symbol,
    &smacro_symbol,
    &cmacro_symbol,
    &csmacro_symbol,
    &code_symbol,
    &fcode_symbol,
    &mcode_symbol,
    &scode_symbol,

    &nomess_symbol,

    &function_symbol,
    &progn_symbol,
    &prog2_symbol,
    &quote_symbol,
    &cons_symbol,
    &nconc_symbol,
    &bye_symbol,

    &raise_symbol,
    &lower_symbol,
    &hankaku_symbol,



#ifdef ASUKA
    &expreope_symbol,
    &name_symbol,
    &for_symbol,
    &foreach_symbol,
    &if_symbol,
    &setq_symbol,

    &arrowope_symbol,
    &onbeepre_symbol,

    &lambda_symbol,

    &itrsyth_symbol,
    &noncomu_symbol,
    &omit_symbol,
    &lesser_equal_symbol,
    &greater_equal_symbol,
    &not_equal_symbol,
    &colon_equal_symbol,
    &semicolon_semicolon_symbol,
    &twbeepre_symbol,
    &thbeepre_symbol,
    &list_symbol,
    &set_symbol,
    &prog_symbol,
    &return_symbol,
    &reversip_symbol,
    &go_symbol,
    &defproc_symbol,
    &not_symbol,
    &set_diff_symbol,
    &ge_symbol,
    &le_symbol,
    &gt_symbol,
    &lt_symbol,
    &add_symbol,
    &sub_symbol,
    &times_symbol,

    &error_symbol,
    &debug_symbol,
    &debugn_symbol,
    &write_symbol,
    &writen_symbol,
    &printf_symbol,
    &read_symbol,
    &import_symbol,
    &save_symbol,
    &load_symbol,
    &clear_symbol,
    &algebraic_symbol,

    &sin_symbol,
    &cos_symbol,
    &tan_symbol,
    &asin_symbol,
    &acos_symbol,
    &atan_symbol,
    &sinh_symbol,
    &cosh_symbol,
    &tanh_symbol,
    &asinh_symbol,
    &acosh_symbol,
    &atanh_symbol,
    &exp_symbol,
    &log_symbol,
    &log2_symbol,
    &log10_symbol,
    &sqrt_symbol,
    &cbrt_symbol,


    &array_symbol,
    &mkarray1_symbol,
    &mkarray2_symbol,
    &mkarray3_symbol,
    &getarray1_symbol,
    &getarray2_symbol,
    &getarray3_symbol,
    &setarray1_symbol,
    &setarray2_symbol,
    &setarray3_symbol,
    &matrix_symbol,
    &mkmatrix_symbol,
    &vector_symbol,
    &mkvector_symbol,


#endif



    &proc_frame,
    &proc_vars,
    &proc_gvars,
    &loop_frame,
    NULL
};



static void
relocate_sysvars()
{
    Sex           **p;
    Sex            *q;
    long            i;
#ifdef  ASUKA
    extern Sex     *yyvs,
                   *yyvsp;
#endif

    p = system_vars;
    while (*p != NULL) {
	relocate(*p++);
    }


    for (i = 0; i < number_of_literal_vectors; i++) {
	for (q = literal_vectors[i]; *q != (Sex) NULL; q++)
	    relocate(q);
    }
#ifdef ASUKA
    if (yyvs) {
	for (q = yyvs; q <= yyvsp; q++)
	    if (*q)
		relocate(q);
    }
#endif
}

void
nullptr(Sex * p)
{
    printf("null pointer %lx\n", (unsigned long) p);
}

static void
mark_sysvars()
{
    long            i;
    Sex           **p;
    Sex            *q;
	int n;
#ifdef  ASUKA
    extern Sex     *yyvs,
                   *yyvsp;
#endif

    p = system_vars;
    while (*p != NULL) {
	mark(*p++);
    }


    for (i = 0; i < number_of_literal_vectors; i++) {
	for (q = literal_vectors[i]; *q != (Sex) NULL; q++) {
	    mark(q);
	}
    }
#ifdef ASUKA
    if (yyvs) {
	for (q = yyvs; q <= yyvsp; q++)
	    if (*q)
		mark(q);
    }
#endif
}


/***************************************************************/
/***************************************************************/


static void
mark_x(Sex * p)
{
    union {
	long            u;
	short           s[4];
    } u;

    if ((long) p & (SEX - 1)) {
	u.s[0] = ((short *) p)[0];
	u.s[1] = ((short *) p)[1];
#ifdef _LP64
	u.s[2] = ((short *) p)[2];
	u.s[3] = ((short *) p)[3];
#endif
	mark((Sex *) & u);
	((short *) p)[0] = u.s[0];
	((short *) p)[1] = u.s[1];
#ifdef _LP64
	((short *) p)[2] = u.s[2];
	((short *) p)[3] = u.s[3];
#endif
    } else
	mark(p);
}

static void
relocate_x(Sex * p)
{
    union {
	long            u;
	short           s[4];
    } u;

    if ((long) p & (SEX - 1)) {
	u.s[0] = ((short *) p)[0];
	u.s[1] = ((short *) p)[1];
#ifdef _LP64
	u.s[2] = ((short *) p)[2];
	u.s[3] = ((short *) p)[3];
#endif
	relocate((Sex *) & u);
	((short *) p)[0] = u.s[0];
	((short *) p)[1] = u.s[1];
#ifdef _LP64
	((short *) p)[2] = u.s[2];
	((short *) p)[3] = u.s[3];
#endif
    } else
	relocate(p);
}



static void
relocate_bps()
{

    long            i;


    for (i = 0; i < bytecode_nreg; i++) {
	relocate(&bytecode_regs[i]);
    }

    for (i = 0; i < atomorg; i++)
	relocate(&atoms[i]);

}

static void
mark_bps()
{

    long            i;


    for (i = 0; i < bytecode_nreg; i++) {
	mark(&bytecode_regs[i]);
    }
    for (i = 0; i < atomorg; i++)
	mark(&atoms[i]);

}

/***************************************************************/
/***************************************************************/
