/*
 * gcompare.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * Arithmetic functions. lessp, greaterp, eqn etc. 
 *      $Header: /Users/kako/asuka/src/RCS/gcompare.c,v 2.6 2015/06/26 08:18:05 kako Exp $
 */

static char const rcsid[] =
    "$Id: gcompare.c,v 2.6 2015/06/26 08:18:05 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "compare.h"

#include "garith.h"
#include "gcompare.h"


bool            gnumeq2(Sex a, Sex b);
bool            geq2(Sex a, Sex b);
bool            geq_list(Sex a, Sex b);
Sex             Lgeqn(Sex a, Sex b);
bool            glessp2(Sex a, Sex b);
bool            gzerop(Sex a);
bool            gzeropp(Sex a);
bool            gonep(Sex a);
bool            gonepp(Sex a);
bool            gonepD(Sex a);

typedef bool    bool_one(Sex);
typedef bool    bool_two(Sex, Sex);

#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)


/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 * Numeric equality - note that comparisons involving non-numbers are errors
 * here (unlike the position in eql, equal, equalp).  Also this must be coded
 * so that it never provokes garbage collection.
 */
extern bool     numeq2ji(Sex a, Sex b);
extern bool     numeq2ij(Sex a, Sex b);
extern bool     numeq2jj(Sex a, Sex b);
extern bool     numeq2il(Sex a, Sex b);
extern bool     bool_yes(Sex a, Sex b);
extern bool     bool_no(Sex a, Sex b);

/***************************************************************/
/***************************************************************/
/***************************************************************/

static          bool
vecequal(Sex a, Sex b)
{
    Tag             ha = vechdr(a),
	hb = vechdr(b);
    int             l;

    if (ha != hb || (l = veclen(a)) != veclen(b))
	return NO;
    l = bnum_n(l + SEX - 1);
    while ((l--) != 0) {
	Sex             ea = vector_value(a, l);
	Sex             eb = vector_value(b, l);
	if (ea == eb)
	    continue;
	if (equal(ea, eb) == nil)
	    return NO;
    }
    return YES;
}

bool
gnumeq2(Sex a, Sex b)
{
    if (a == b)
	return YES;
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return numeq2(a, b);
	else if (is_vechdr(b))
	    return NO;
	return NO;
    } else if (is_poly(a) && is_poly(b)) {
	a = gdiff2pp(a, b);
	if (is_number(a))
	    return zerop(a);
	return NO;
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return NO;
	else if (is_vechdr(b)) {
	    if (gnumtype_of(a) != gnumtype_of(b))
		return NO;
	    return vecequal(a, b);
	}
	return NO;
    }
    return NO;
    /*
     * aerror2("gnumeq2", a, b);
     */
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

two_args       *geq_table_1[] =
    { bool_no, geq2nv, geq2na, geq2nt, geq2np, geq2nD,
    geq2nA, geq2nr, geq2nV, geq2nM, geq2nT, geq2nP,
    geq2nR, geq2nS, geq2nL, geq2nX
};

two_args       *geq_table_2[] =
    { bool_no, geq2vn, geq2an, geq2tn, geq2pn, geq2Dn,
    geq2An, geq2rn, geq2Vn, geq2Mn, geq2Tn, geq2Pn,
    geq2Rn, geq2Sn, geq2Ln, geq2Xn
};

two_args       *geq_table[GTYPEMAX][GTYPEMAX] = {
    {bool_no, bool_no, bool_no, bool_no, bool_no, bool_no,
     bool_no, bool_no, bool_no, bool_no, bool_no, bool_no,
     bool_no, bool_no, bool_no, bool_no}
    ,
    {bool_no, geq2vv, geq2va, geq2vt, geq2vp, geq2vD,
     geq2vA, geq2vr, geq2vV, geq2vM, geq2vT, geq2vP,
     geq2vR, geq2vS, geq2vL, geq2vX}
    ,
    {bool_no, geq2av, geq2aa, geq2at, geq2ap, geq2aD,
     geq2aA, geq2ar, geq2aV, geq2aM, geq2aT, geq2aP,
     geq2aR, geq2aS, geq2aL, geq2aX}
    ,
    {bool_no, geq2tv, geq2ta, geq2tt, geq2tp, geq2tD,
     geq2tA, geq2tr, geq2tV, geq2tM, geq2tT, geq2tP,
     geq2tR, geq2tS, geq2tL, geq2tX}
    ,
    {bool_no, geq2pv, geq2pa, geq2pt, geq2pp, geq2pD,
     geq2pA, geq2pr, geq2pV, geq2pM, geq2pT, geq2pP,
     geq2pR, geq2pS, geq2pL, geq2pX}
    ,
    {bool_no, geq2Dv, geq2Da, geq2Dt, geq2Dp, geq2DD,
     geq2DA, geq2Dr, geq2DV, geq2DM, geq2DT, geq2DP,
     geq2DR, geq2DS, geq2DL, geq2DX}
    ,
    {bool_no, geq2Av, geq2Aa, geq2At, geq2Ap, geq2AD,
     geq2AA, geq2Ar, geq2AV, geq2AM, geq2AT, geq2AP,
     geq2AR, geq2AS, geq2AL, geq2AX}
    ,
    {bool_no, geq2rv, geq2ra, geq2rt, geq2rp, geq2rD,
     geq2rA, geq2rr, geq2rV, geq2rM, geq2rT, geq2rP,
     geq2rR, geq2rS, geq2rL, geq2rX}
    ,
    {bool_no, geq2Vv, geq2Va, geq2Vt, geq2Vp, geq2VD,
     geq2VA, geq2Vr, geq2VV, geq2VM, geq2VT, geq2VP,
     geq2VR, geq2VS, geq2VL, geq2VX}
    ,
    {bool_no, geq2Mv, geq2Ma, geq2Mt, geq2Mp, geq2MD,
     geq2MA, geq2Mr, geq2MV, geq2MM, geq2MT, geq2MP,
     geq2MR, geq2MS, geq2ML, geq2MX}
    ,
    {bool_no, geq2Tv, geq2Ta, geq2Tt, geq2Tp, geq2TD,
     geq2TA, geq2Tr, geq2TV, geq2TM, geq2TT, geq2TP,
     geq2TR, geq2TS, geq2TL, geq2TX}
    ,
    {bool_no, geq2Pv, geq2Pa, geq2Pt, geq2Pp, geq2PD,
     geq2PA, geq2Pr, geq2PV, geq2PM, geq2PT, geq2PP,
     geq2PR, geq2PS, geq2PL, geq2PX}
    ,
    {bool_no, geq2Rv, geq2Ra, geq2Rt, geq2Rp, geq2RD,
     geq2RA, geq2Rr, geq2RV, geq2RM, geq2RT, geq2RP,
     geq2RR, geq2RS, geq2RL, geq2RX}
    ,
    {bool_no, geq2Sv, geq2Sa, geq2St, geq2Sp, geq2SD,
     geq2SA, geq2Sr, geq2SV, geq2SM, geq2ST, geq2SP,
     geq2SR, geq2SS, geq2SL, geq2SX}
    ,
    {bool_no, geq2Lv, geq2La, geq2Lt, geq2Lp, geq2LD,
     geq2LA, geq2Lr, geq2LV, geq2LM, geq2LT, geq2LP,
     geq2LR, geq2LS, geq2LL, geq2LX}
    ,
    {bool_no, geq2Xv, geq2Xa, geq2Xt, geq2Xp, geq2XD,
     geq2XA, geq2Xr, geq2XV, geq2XM, geq2XT, geq2XP,
     geq2XR, geq2XS, geq2XL, geq2XX}
};




bool
geq2(a, b)
     Sex             a,
                     b;
{
    if (a == b)
	return YES;

#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return numeq2(a, b);
	else if (is_vechdr(b))
	    return geq_table_1[gnumtype_of(b)] (a, b);
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return geq_table_2[gnumtype_of(a)] (a, b);
	else if (is_vechdr(b))
	    return geq_table[gnumtype_of(a)][gnumtype_of(b)] (a, b);
    } else if (is_cons(a)) {
	if (is_cons(b))
	    return geq_list(a, b);
	return NO;
    }
    return equal(a, b) == nil ? NO : YES;
}

bool
geq_list(Sex a, Sex b)
{
    bool            r = YES;
    push2(a, b);
#define a stack[-1]
#define b stack[0]
    while (a != nil) {
	if (b == nil) {
	    r = NO;
	    break;
	}
	r = geq2(car(a), car(b));
	if (r == NO)
	    break;
	a = cdr(a);
	b = cdr(b);
    }
    if (b != nil)
	r = NO;
#undef a
#undef b
    popv(2);
    return r;
}

bool
geq2Dn(Sex a, Sex b)
{
    Sex             v;
    Sex            *p;
    int             i,
                    n;
    a = dipoly_pol(a);
    if (a == nil)
	return zerop(b);
    if (cdr(a) != nil)
	return NO;
    a = car(a);
#if DIPOLY_COEFF == 0
    v = cdr(a);
    n = mkint(vect_dimx(v));
    p = &vect_value(v, 0);
    for (i = 0; i < n; i++) {
	if (*p++ != mkfxat(0))
	    return NO;
    }
    return geq2(car(a), b);
#else
    v = a;
    n = mkint(vect_dimx(v));
    p = &vect_value(v, 1);
    for (i = 1; i < n; i++) {
	if (*p++ != mkfxat(0))
	    return NO;
    }
    a = vect_value(a, 0);
    return geq2(a, b);
#endif
}

bool
geq2nD(Sex a, Sex b)
{
    return geq2Dn(b, a);
}

bool
geq2DD(Sex a, Sex b)
{
    if (a == b)
	return YES;
    if (dipoly_order(a) != dipoly_order(b))
	return NO;
    if (equal(a, b) == nil)
	return NO;
    return YES;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
Lgeqn(Sex a, Sex b)
{
    bool            r;
    // r = gnumeq2(a, b);
    r = geq2(a, b);
    return Bool_to_Sex(r);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
glessp2(Sex a, Sex b)
{
#ifdef USE_SYMBOL_AS_POLYNOMIAL
    if (is_symbol(a)) {
	push(b);
	a = symbol_to_polynomial(a);
	pop(b);
    }
    if (is_symbol(b)) {
	push(a);
	b = symbol_to_polynomial(b);
	pop(a);
    }
#endif
    if (is_number(a)) {
	if (is_number(b))
	    return lessp2(a, b);
	else if (is_vechdr(b))
	    return NO;
    } else if (is_vechdr(a)) {
	if (is_number(b))
	    return NO;
	else if (is_vechdr(b)) {
	    if (gnumtype_of(a) != gnumtype_of(b))
		return NO;
	    return vecequal(a, b);
	}
    }
    return aerror2("glessp2", a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/
Sex
Lgzerop(Sex a)
{
    bool            r = NO;
    if (is_number(a)) {
	if (is_fixnum(a))
	    r = (a == mkfxat(0));
	else
	    r = zerop(a);
    } else if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_DIPOLY:
	    r = gzeropD(a);
	    break;
	case N_VECT:
	    r = gzeropV(a);
	    break;
	case N_MATRIX:
	    r = gzeropM(a);
	    break;
	}
    }
    return Bool_to_Sex(r);
}


bool
gzerop(Sex a)
{
    if (is_number(a))
	return zerop(a);
    else if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_POLY:
	    if (poly_pol(a) == nil)
		return YES;
	    return NO;
	case N_DIPOLY:
	    if (dipoly_pol(a) == nil)
		return YES;
	    return NO;
	case N_VECT:
	    return gzeropV(a);
	case N_MATRIX:
	    return gzeropM(a);
	}
    }
    return NO;
}


bool
gzeropp(Sex a)
{
    if (poly_pol(a) == nil)
	return YES;
    return NO;
}

bool
gzeropD(Sex a)
{
    if (dipoly_pol(a) == nil)
	return YES;
    return NO;
}

bool
gzeropV(Sex a)
{
    long            i,
                    n;
    Sex            *p;

    n = mkint(vect_dimx(a));
    p = &vect_value(a, 0);
    for (i = 0; i < n; i++) {
	if (!gzerop(*p++))
	    return NO;
    }
    return YES;
}

bool
gzeropM(Sex a)
{
    long            i,
                    sx,
                    sy;
    Sex            *p;

    sx = mkint(matrix_dimx(a));
    sy = mkint(matrix_dimy(a));
    sx = sx * sy;
    p = &matrix_value(a, 0);
    for (i = 0; i < sx; i++) {
	if (!gzerop(*p++))
	    return NO;
    }
    return YES;
}

bool
gonep(Sex a)
{
    if (is_number(a))
	return onep(a);
    else if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_POLY:
	    return gonepp(a);
	case N_DIPOLY:
	    return gonepD(a);
	}
    }
    return NO;
}


bool
gonepp(Sex a)
{
    a = poly_pol(a);
    if (cdr(a) != nil)
	return NO;
    a = lterm(a);
    if (power(a) != mkfxat(0))
	return NO;
    a = coeff(a);
    return gonep(a);
}

bool
gonepD(Sex a)
{
    Sex             f;
    int             order;
    long            i,
                    n;

    f = dipoly_order(a);
    order = mkint(car(f));

    f = dipoly_pol(a);
    if (f == nil)
	return NO;
    a = car(f);
    f = cdr(f);
    if (f != nil)
	return NO;
#if DIPOLY_COEFF == 0
    f = car(a);
    a = cdr(a);
    if (!onep(f))
	return NO;
    i = order / 2;
    for (; i < n + order / 2; i++) {
	f = vect_value(a, i);
	if (f != mkfxat(0))
	    return NO;
    }
#else
    f = vect_value(a, 0);
    if (!onep(f))
	return NO;
    i = order / 2 + 1;
    for (; i <= n + order / 2; i++) {
	f = vect_value(a, i);
	if (f != mkfxat(0))
	    return NO;
    }
#endif
    return YES;
}
