/*
 * hash.c               (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 *  Hash functions.
 * $Header: /Users/kako/asuka/src/RCS/hash.c,v 2.3 2017/11/27 00:45:12 kako Exp $
 */

static char const rcsid[] =
    "$Id: hash.c,v 2.3 2017/11/27 00:45:12 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "kctype.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"

unsigned long   hash_equal(Sex x, long depth);
Sex             Lsethash(Sex a, Sex t, Sex c);
Sex             Lremhash(Sex a, Sex t);
Sex             Lgethash(Sex key, Sex t);
Sex             Lhash_equal(Sex p);
Sex             Lhcons(Sex a, Sex b);
Sex             Lhmapcar(Sex f, Sex x);
Sex             Lhappend(Sex r, Sex s);
Sex             hlist_fn(Sex a, Sex env);
Sex             Lmaphcons(Sex fn);
Sex             Lmaphash(Sex fn, Sex t);

unsigned long   hash_eql(Sex x);
HashEntry      *gethash(Sex key, Sex hashtable);
void            extend_hashtable(Sex hashtable);
Sex             sethash(Sex key, Sex hashtable, Sex value);
Sex             Lclrhash(Sex t);
Sex             Lhash_table_count(Sex t);
Sex             Lnext_hash_table_entry(Sex t, Sex ind);
Sex             Lhash_table_p(Sex t);
Sex             Lhash_table_size(Sex t);
Sex             Lmake_hash_table(void);
void            extend_uniq_table(void);
Sex             Lsxhash(Sex a);
Sex             Lmkhash(Sex a, Sex b, Sex c);
Sex             Lmake_string_unique(Sex a);

/*
 */

unsigned long
hash_eql(Sex x)
{
    unsigned long   h = 0;
    unsigned long  *p;
    unsigned char  *q;
    long            l;


    switch (TAG_BITS & x) {
    case TAG_CONS:
	h = hash_eql(car(x));
	h <<= 1;
	return h ^ hash_eql(cdr(x));
    case TAG_FIXNUM:
	return mkint(x);
    case TAG_SYMBOL:
	q = (unsigned char *) &pname(x);
	p = (unsigned long *) q;
	l = symbol_length(x);
	while (l >= 0) {
	    h = (h << 1) ^ (*p++);
	    l -= SEX;
	}
	return h;
    case TAG_BINARY:
	q = (unsigned char *) &string_value(x, 0);
	p = (unsigned long *) q;
	l = string_length(x);
	while (l >= 0) {
	    h = (h << 1) ^ (*p++);
	    l -= SEX;
	}
	return h;
    case TAG_VECTOR:
	p = (unsigned long *) &vector_value(x, 0);
	l = bnum_n(veclen(x));
	while (l-- > 0) {
	    h ^= hash_eql(*p++);
	}
	return h;

    case TAG_NUMBERS:
	switch (numtype_of(x)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    l = bnum_n(bignum_length(x));
	    p = (unsigned long *) &bignum_value(x)[0];
	    while (l-- > 0) {
		h ^= *p++;
	    }
	    return h;
	case N_MODULAR:
	    return hash_eql(modular_value(x));
	case N_RATIONAL:
	    return hash_eql(rational_num(x)) + hash_eql(rational_den(x));
	case N_FLOAT:
	    p = (unsigned long *) &float_val(x);
#ifdef _LP64
	    return p[0];
#else
	    return p[0] ^ p[1];
#endif
	case N_EFLOAT:
	    p = (unsigned long *) &efloat_value(x);
#ifdef _LP64
	    return p[0] ^ p[1];
#else
	    return p[0] ^ p[1] ^ p[2] ^ p[3];
#endif
	case N_COMPLEX:
	    return hash_eql(complex_real(x))
		^ hash_eql(complex_imag(x));
	case N_INTERVAL:
	    p = (unsigned long *) &interval_lower(x);
#ifdef _LP64
	    h = p[0];
#else
	    h = p[0] ^ p[1];
#endif
	    p = (unsigned long *) &interval_upper(x);
#ifdef _LP64
	    return h ^ p[0];
#else
	    return h ^ p[0] ^ p[1];
#endif
	case N_BFLOAT:
	    return hash_eql(bfloat_man(x)) ^ bfloat_exp(x);
	case N_ECOMPLEX:
	    p = (unsigned long *) &ecomplex_real(x);
#ifdef _LP64
	    return p[0] ^ p[1] ^ p[2];
#else
	    return p[0] ^ p[1] ^ p[2] ^ p[3] ^ p[4] ^ p[5];
#endif
	case N_BEFLOAT:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    p = (unsigned long *) &befloat_error(x);
#ifdef _LP64
	    return hash_eql(befloat_value(x)) ^ p[0];
#else
	    return hash_eql(befloat_value(x)) ^ p[0] ^ p[1];
#endif
#else
	    return hash_eql(befloat_value(x)) ^ befloat_error(x);
#endif
	case N_BECOMPLEX:
#ifdef BEFLOAT_ERROR_IS_DOUBLE
	    p = (unsigned long *) &becomplex_error(x);
	    return hash_eql(becomplex_real(x))
		^ hash_eql(becomplex_imag(x))
		^ p[0] ^ p[1];
#else
	    return hash_eql(becomplex_real(x))
		^ hash_eql(becomplex_imag(x))
		^ becomplex_error(x);
#endif
	case N_BINTERVAL:
	    return hash_eql(binterval_upper(x))
		^ hash_eql(binterval_lower(x));
	}

    }
    return x / 4;
}

unsigned long
hash_equal(Sex x, long depth)
{
    unsigned long   h = 0;
    long            i;
    char           *s;
    Sex            *p;


    while (1) {
	if (depth++ > 3)
	    return h;
	switch ((long) x & TAG_BITS) {
	case TAG_FIXNUM:
	    return mkint(x);
	case TAG_CONS:
	    h = h << 1;
	    h ^= hash_equal(car(x), depth);
	    x = cdr(x);
	    continue;

	case TAG_BINARY:
	    i = string_length(x);
	    s = &string_value(x, 0);
	    while (i >= 0) {
		h = h << 1;
		h ^= (*s++ & 0377) * 12345 + 1;
		i--;
	    }
	    return (h);
	case TAG_SYMBOL:
	    i = symbol_length(x);
	    s = (char *) &pname(x);
	    while (i >= 0) {
		h = h << 1;
		h ^= (*s++ & 0377) * 12345 + 1;
		i--;
	    }
	    return (h);
	case TAG_VECTOR:
	    p = &vector_value(x, 0);
	    i = bnum_n(veclen(x));
	    while (i > 0) {
		h = h << 1;
		h ^= hash_equal(*p++, depth);
		i--;
	    }
	    return h;
	default:
	    h = h << 1;
	    return h ^ hash_eql(x);
	}
    }
}


HashEntry      *
gethash(Sex key, Sex hashtable)
{
    long            hsize;
    HashEntry      *e;
    Sex             hkey;
    long            i,
                    j;

    if (!is_hashtable(hashtable))
	error("not a hash table");
    hsize = mkint(hash_size(hashtable));
    i = hash_equal(key, 0);
    e = (HashEntry *) & vector_value(hash_entry(hashtable), 0);
    i &= int_max;
    i %= hsize;
    j = i;
    while (1) {
	hkey = e[i].key;
	if (hkey == nil) {
	    return (&e[i]);
	}
	if (equal(key, hkey))
	    return (&e[i]);
	i++;
	if (i >= hsize)
	    i = 0;
    }
    return NULL;
}

void
extend_hashtable(Sex hashtable)
{
    long            hsize,
                    new_hsize;
    HashEntry      *e,
                   *e_new;
    Sex             hkey;
    long            i,
                    n;
    Sex             new_table;

    hsize = hash_size(hashtable);
    new_hsize = hsize * 3 / 2;
    push(hashtable);
    new_table = getvector_init(new_hsize * sizeof(HashEntry) + HDR, nil);
    pop(hashtable);
    e = (HashEntry *) & vector_value(hash_entry(hashtable), 0);
    e_new = (HashEntry *) & vector_value(new_table, 0);
    for (i = 0; i < hsize; i++) {
	if (e[i].key == nil)
	    continue;
	n = hash_equal(e[i].key, 0);
	n = (n & int_max) % (new_hsize);
	while (1) {
	    hkey = e_new[n].key;
	    if (hkey == nil) {
		e_new[n].key = e[i].key;
		e_new[n].value = e[i].value;
		break;
	    }
	    n++;
	    if (n >= new_hsize)
		n = 0;
	}
    }
    hash_entry(hashtable) = new_table;
    hash_size(hashtable) = new_hsize;
}


Sex
sethash(Sex key, Sex hashtable, Sex value)
{
    HashEntry      *e;
    long            count;
    long            size;

    size = mkint(hash_size(hashtable));
    count = mkint(hash_count(hashtable));

    if (size * 9 < count * 10)
	extend_hashtable(hashtable);
    e = gethash(key, hashtable);
    if (e == NULL)
	error("sethash: this should not happen");
    if (e->key == nil)
	hash_count(hashtable) = mkfxat(count + 1);
    e->key = key;
    e->value = value;
    return value;
}



Sex
Lgethash(Sex key, Sex t)
{
    HashEntry      *e;

    if (!is_hashtable(t))
	error("not a hash table");
    e = gethash(key, t);
    if (e == NULL || e->key == nil)
	return nil;
    return e->value;
}


Sex
Lsethash(Sex a, Sex t, Sex c)
{
    if (!is_hashtable(t))
	error("not a hash table");
    return sethash(a, t, c);
}

Sex
Lremhash(Sex a, Sex t)
{
    /*
     * HashEntry *e; 
     */

    if (!is_hashtable(t))
	error("not a hash table");
    /*
     * not implemented yet
     */



    return nil;
}

Sex
Lclrhash(Sex t)
{
    long            i,
                    n;
    Sex             v;
    HashEntry      *e;

    if (!is_hashtable(t))
	error("not a hash table");
    v = hash_entry(t);
    e = (HashEntry *) & vector_value(v, 0);
    n = mkint(hash_size(t));
    for (i = 0; i < n; i++) {
	e->key = nil;
	e->value = nil;
	e++;
    }
    hash_count(t) = mkfxat(0);
    return true;
}

Sex
Lhash_table_count(Sex t)
{
    if (!is_hashtable(t))
	error("not a hash table");
    return hash_count(t);
}


Sex
Lsxhash(Sex a)
{
    return mkfxat(hash_equal(a, 0));
}

Sex
Lmaphash(Sex fn, Sex t)
{
    long            i;
    long            n;
    HashEntry      *e;
    if (!is_hashtable(t))
	error("not a hash table");

    stackcheck();
    push3(fn, t, nil);
#define fn stack[-2]
#define t stack[-1]
#define vv stack[0]

    vv = hash_entry(t);
    n = mkint(hash_size(t));
    for (i = 0; i < n; i++) {
	e = (HashEntry *) & vector_value(vv, 0);
	if (e[i].key != nil)
	    Lfuncall2(fn, e[i].key, e[i].value);
    }
    popv(3);
#undef fn
#undef t
#undef v
    return nil;
}


Sex
Lnext_hash_table_entry(Sex t, Sex ind)
{
    long            i;
    long            n;
    HashEntry      *e;

    if (!is_fixnum(ind))
	error("not a fix num");
    i = mkint(ind);
    if (!is_hashtable(t))
	error("not a hash table");
    if (i < 0) {
	error("needs non negative index");
    }
    e = (HashEntry *) & vector_value(hash_entry(t), 0);
    n = mkint(hash_size(t));
    while (i < n) {
	if (e[i].key != nil) {
	    return Llist3(mkfxat(i + 1), e[i].key, e[i].value);
	}
	i++;
    }
    return Llist3(mkfxat(-1), nil, nil);
}

Sex
Lhash_table_p(Sex t)
{
    return Bool_to_Sex(is_hashtable(t));
}

Sex
Lhash_table_size(Sex t)
{
    if (!is_hashtable(t))
	error("not a hash table");
    return hash_size(t);
}

Sex
Lhash_equal(Sex p)
{
    return mkfxat(hash_equal(p, 0));
}


Sex
Lmake_hash_table(void)
{
    long            n = INIT_HASH_SIZE;
    Sex             h,
                    v;

    v = getvector_init(n * sizeof(HashEntry) + HDR, nil);
    stackcheck();
    push(v);
    h = getvector(TAG_VECTOR, T_HASH, sizeof(HashTable));
    pop(v);
    hash_count(h) = mkfxat(0);
    hash_size(h) = mkfxat(INIT_HASH_SIZE);
    hash_entry(h) = v;
    return h;
}

Sex
Lmkhash(Sex a, Sex b, Sex c)
{
        return Lmake_hash_table();
}

/*
 * hashed cons routine 
 */

void
extend_uniq_table(void)
{
    long            hsize,
                    new_hsize;
    Sex            *e,
                   *e_new;
    Sex             hkey;
    long            i,
                    n;
    Sex             new_table;

    hsize = uniq_cons_size;
    new_hsize = hsize * 3 / 2;
    new_table = getvector_init(new_hsize * sizeof(HashEntry) + HDR, nil);
    e = &vector_value(uniq_cons_table, 0);
    e_new = &vector_value(new_table, 0);
    for (i = 0; i < hsize; i++) {
	if (e[i] == nil)
	    continue;
	n = hash_equal(e[i], 0);
	n = (n & int_max) % (new_hsize);
	while (1) {
	    hkey = e_new[n];
	    if (hkey == nil) {
		e_new[n] = e[i];
		break;
	    }
	    n++;
	    if (n >= new_hsize)
		n = 0;
	}
    }
    uniq_cons_table = new_table;
    uniq_cons_size = new_hsize;
}


Sex
getuniq(Sex key)
{
    long            hsize;
    Sex            *e;
    Sex             hkey;
    long            i,
                    j;

    hsize = uniq_cons_size;
    if (9 * hsize < 10 * uniq_cons_count) {
	push(key);
	extend_uniq_table();
	pop(key);
	hsize = uniq_cons_size;
    }
    i = hash_equal(key, 0);
    e = &vector_value(uniq_cons_table, 0);
    i &= int_max;
    i %= hsize;
    j = i;
    while (1) {
	hkey = e[i];
	if (hkey == nil) {
	    e[i] = key;
	    uniq_cons_count++;
	    return key;
	}
	if (equal(key, hkey))
	    return hkey;
	if (++i >= hsize)
	    i = 0;
	if (i == j) {
	    error("getuniq: this should not happen");
	}
    }
}

Sex
Lhcons(Sex a, Sex b)
{
    Sex             c;
    car(fake_cons_cell) = a;
    cdr(fake_cons_cell) = b;
    c = getuniq(fake_cons_cell);
    if (c == fake_cons_cell) {
	push(c);
	fake_cons_cell = cons(car_symbol, cdr_symbol);
	pop(c);
    }
    return c;
}


Sex
Lhmapcar(Sex f, Sex x)
{
    Sex             a,
                    b;
    if (x == nil)
	return nil;
    stackcheck();
    push2(f, x);
    a = Lfuncall1(f, car(x));
    pop2(x, f);
    push(a);
    b = Lhmapcar(f, cdr(x));
    pop(a);
    return Lhcons(a, b);
}



Sex
Lhappend(Sex r, Sex s)
{
    Sex             a;
    if (r == nil)
	return s;
    stackcheck();
    push(car(r));
    a = Lhappend(cdr(r), s);
    pop(r);
    return Lhcons(r, a);
}

Sex
hlist_fn(Sex a, Sex env)
{
    Sex             x;

    if (a == nil)
	return nil;
    x = car(a);
    stackcheck();
    push2(a, env);
    x = eval(x, env);
    pop2(env, a);
    push(x);
    a = hlist_fn(cdr(a), env);
    pop(x);
    return Lhcons(x, a);
}

Sex
Lmaphcons(Sex fn)
{
    long            i;
    Sex             e;

    stackcheck();
    push(fn);
#define fn stack[0]

    for (i = 0; i < uniq_cons_size; i++) {
	e = vector_value(uniq_cons_table, i);
	if (e != nil)
	    Lfuncall1(fn, e);
    }
#undef fn
    popv(1);
    return true;
}

/***************************************************************/

Sex
Lmake_string_unique(Sex a)
{
        Sex w;
        stackcheck();
        push(a);
        w = Lgethash(a, uniq_string_table);
        if(w == nil) {
                a = stack[0];
                Lsethash(a, uniq_string_table, a);
                a = stack[0];
        } else
                a = w;
        popv(1);
        return a;
}

/***************************************************************/

setup_type const hash_setup[] = {
    {"make-hash-table", (lisp_proc *) Lmake_hash_table, 0},
    {"make_hash_table", (lisp_proc *) Lmake_hash_table, 0},
    {"hash-table-p", (lisp_proc *) Lhash_table_p, 1},
    {"hash_table_p", (lisp_proc *) Lhash_table_p, 1},
    {"mkhash", (lisp_proc *) Lmkhash, 3},
    {"gethash", (lisp_proc *) Lgethash, 2},
    {"remhash", (lisp_proc *) Lremhash, 2},
    {"maphash", (lisp_proc *) Lmaphash, 2},
    {"clrhash", (lisp_proc *) Lclrhash, 1},
    {"hash-table-count", (lisp_proc *) Lhash_table_count, 1},
    {"hash_table_count", (lisp_proc *) Lhash_table_count, 1},
    {"hash-table-size", (lisp_proc *) Lhash_table_size, 1},
    {"hash_table_size", (lisp_proc *) Lhash_table_size, 1},
    {"sxhash", (lisp_proc *) Lsxhash, 1},
    {"hash-equal", (lisp_proc *) Lhash_equal, 1},
    {"hash_equal", (lisp_proc *) Lhash_equal, 1},
    {"sethash", (lisp_proc *) Lsethash, 3},
    {"next-hash-table-entry", (lisp_proc *) Lnext_hash_table_entry, 2},
    {"next_hash_table_entry", (lisp_proc *) Lnext_hash_table_entry, 2},
    {"hcons", (lisp_proc *) Lhcons, 2},
    {"maphcons", (lisp_proc *) Lmaphcons, 1},
    {"hmapcar", (lisp_proc *) Lhmapcar, 2},
    {"happend", (lisp_proc *) Lhappend, 2},
    {"hlist", (lisp_proc *) hlist_fn, -1},
    {"make-string-unique", (lisp_proc *) Lmake_string_unique, 1},
    {NULL, NULL, 0}
};
