/*
 * interval.c           (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * interval number
 * $Header: /Users/kako/asuka/src/RCS/interval.c,v 2.5 2014/10/20 00:23:49 kako Exp $
 */

static char const rcsid[] =
    "$Id: interval.c,v 2.5 2014/10/20 00:23:49 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "interval.h"



/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 * interval arithmetic
 */
Sex
Linterval_lower(Sex a)
{
    if (is_interval(a))
	return make_float(interval_lower(a));
    else if (is_binterval(a))
	return binterval_lower(a);
    return nil;
}

Sex
Linterval_upper(Sex a)
{
    if (is_interval(a))
	return make_float(interval_upper(a));
    else if (is_binterval(a))
	return binterval_upper(a);
    return nil;
}


Sex
Linterval_mid(Sex a)
{
    if (is_interval(a))
	return make_float((interval_upper(a) + interval_lower(a)) * 0.5);
    else if (is_binterval(a)) {
	return quot2di(plus2dd(binterval_upper(a), binterval_lower(a)),
		       mkfxat(2));
    }
    return nil;
}


Sex
Linterval_width(Sex a)
{
    if (is_interval(a))
	return make_float(interval_upper(a) - interval_lower(a));
    else if (is_binterval(a))
	return diff2dd(binterval_upper(a), binterval_lower(a));
    return nil;
}

Sex
make_interval(double a, double b)
{
    Sex             c;

    c = new_interval();
    interval_lower(c) = a;
    interval_upper(c) = b;
    return c;
}

Sex
Lmkinterval(Sex a, Sex b)
{
    double          x,
                    y;
    stackcheck();
    push(b);
    a = Linterval(a);
    pop(b);
    x = interval_lower(a);
    y = interval_upper(a);
    b = Linterval(b);
    if (x > interval_lower(b))
	x = interval_lower(b);
    if (y < interval_upper(b))
	y = interval_upper(b);
    return make_interval(x, y);
}

Sex
Linterval_by_ratio(Sex a, Sex f)
{
    double          x,
                    y;
    x = floatit(a);
    y = floatit(f);
    y = fabs(y);
    if (x > 0)
	return make_interval((1 - y) * x, (1 + y) * x);
    else
	return make_interval((1 + y) * x, (1 - y) * x);
}

Sex
intervalitE(Sex a)
{
#ifdef BEFLOAT_ERROR_IS_DOUBLE
    double          e;
    double          x;

    e = befloat_error(a);
    x = floatitd(befloat_value(a));
    return make_interval(x - e, x + e);
#else
    long            e;
    double          x,
                    y;

    e = befloat_error(a);
    a = befloat_value(a);
    x = floatitd(a);
    y = fabs(ldexp(x, e));
    return make_interval(x - y, x + y);
#endif
}

Sex
intervalitV(Sex a)
{
    double          x,
                    y;
    x = floatitd(binterval_lower(a));
    y = floatitd(binterval_upper(a));
    return make_interval(x, y);
}

Sex
Linterval(Sex a)
{
    double          x,
                    y;

    if (is_fixnum(a)) {
	x = mkint(a);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    x = floatitb(a);
	    break;
	case N_FLOAT:
	    x = float_val(a);
	    break;
	case N_RATIONAL:
	    x = floatit(rational_num(a)) / floatit(rational_den(a));
	    break;
	case N_BFLOAT:
	    x = floatitd(a);
	    break;
	case N_EFLOAT:
	    x = efloat_value(a) + efloat_error(a);
	    y = efloat_value(a) - efloat_error(a);
	    return make_interval(y, x);
	case N_INTERVAL:
	    return a;
	case N_BEFLOAT:
	    return intervalitE(a);
	case N_BINTERVAL:
	    return intervalitV(a);
	default:
	    return error1("bad argument to interval", a);
	}
    } else
	return error1("bad argument to interval", a);
    return make_interval(x, x);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


static          Sex
plus2_f_v(double f, Sex a)
{
    double          wl = interval_lower(a);
    double          wu = interval_upper(a);
    return make_interval(wl + f, wu + f);
}

Sex
plus2iv(Sex a, Sex b)
{
    return plus2_f_v((double) mkint(a), b);
}

Sex
plus2xv(Sex a, Sex b)
{
    return plus2_f_v(floatit(a), b);
}


Sex
plus2fv(Sex a, Sex b)
{
    return plus2_f_v(float_val(a), b);
}


Sex
plus2vv(Sex a, Sex b)
{
    double          wl,
                    wu;
    wl = interval_lower(a) + interval_lower(b);
    wu = interval_upper(a) + interval_upper(b);
    return make_interval(wl, wu);
}



Sex
plus2vd(Sex a, Sex b)
{
    return plus2_f_v(floatitd(b), a);
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

static          Sex
diff2_f_v(double f, Sex a)
{
    return make_interval(f - interval_upper(a), f - interval_lower(a));
}

static          Sex
diff2_v_f(Sex a, double f)
{
    return make_interval(interval_lower(a) - f, interval_upper(a) - f);
}


Sex
diff2iv(Sex a, Sex b)
{
    return diff2_f_v((double) mkint(a), b);
}

Sex
diff2xv(Sex a, Sex b)
{
    return diff2_f_v(floatit(a), b);
}



Sex
diff2vi(Sex a, Sex b)
{
    return diff2_v_f(a, (double) mkint(b));
}

Sex
diff2vx(Sex a, Sex b)
{
    return diff2_v_f(a, floatit(b));
}

Sex
diff2fv(Sex a, Sex b)
{
    return diff2_f_v(float_val(a), b);
}


Sex
diff2vf(Sex a, Sex b)
{
    return diff2_v_f(a, float_val(b));
}


Sex
diff2vv(Sex a, Sex b)
{
    double          wl,
                    wu;
    wl = interval_lower(a) - interval_upper(b);
    wu = interval_upper(a) - interval_lower(b);
    return make_interval(wl, wu);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


static          Sex
times2_f_v(double f, Sex a)
{
    double          wl = interval_lower(a) * f;
    double          wu = interval_upper(a) * f;
    if (f > 0.0)
	return make_interval(wl, wu);
    else
	return make_interval(wu, wl);
}

Sex
times2iv(Sex a, Sex b)
{
    return times2_f_v((double) mkint(a), b);
}

Sex
times2xv(Sex a, Sex b)
{
    return times2_f_v(floatit(a), b);
}


Sex
times2fv(Sex a, Sex b)
{
    return times2_f_v(float_val(a), b);
}



Sex
times2vd(Sex a, Sex b)
{
    return times2_f_v(floatit(b), a);
}


Sex
times2vv(Sex a, Sex b)
{
    double          vl,
                    vu;

    if (interval_lower(a) > 0.0) {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a);
	    vu = interval_upper(a);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a);
	    vu = interval_lower(a);
	} else {
	    vl = interval_upper(a);
	    vu = interval_upper(a);
	}
	vl = vl * interval_lower(b);
	vu = vu * interval_upper(b);
    } else if (interval_upper(a) < 0.0) {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a);
	    vu = interval_upper(a);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a);
	    vu = interval_lower(a);
	} else {
	    vl = interval_lower(a);
	    vu = interval_lower(a);
	}
	vl = vl * interval_upper(b);
	vu = vu * interval_lower(b);
    } else {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a) * interval_upper(b);
	    vu = interval_upper(a) * interval_upper(b);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a) * interval_lower(b);
	    vu = interval_lower(a) * interval_lower(b);
	} else {
	    double          x;
	    vl = interval_upper(a) * interval_lower(b);	/* 
							 * -
							 */
	    x = interval_lower(a) * interval_upper(b);	/* 
							 * -
							 */
	    if (x < vl)
		vl = x;
	    vu = interval_upper(a) * interval_upper(b);	/* 
							 * +
							 */
	    x = interval_lower(a) * interval_lower(b);	/* 
							 * +
							 */
	    if (vu < x)
		vu = x;
	}
    }
    return make_interval(vl, vu);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

static          Sex
quot2_f_v(double f, Sex a)
{
    double          wl;
    double          wu;
    if (interval_lower(a) <= 0.0 && interval_upper(a) >= 0.0)
	return aerror("quot2xv: zero divide");
    wl = f / interval_lower(a);
    wu = f / interval_upper(a);
    if (wu > wl)
	return make_interval(wl, wu);
    else
	return make_interval(wu, wl);
}

static          Sex
quot2_v_f(Sex a, double f)
{
    double          wl = interval_lower(a) / f;
    double          wu = interval_upper(a) / f;
    if (wu > wl)
	return make_interval(wl, wu);
    else
	return make_interval(wu, wl);
}

Sex
quot2iv(Sex a, Sex b)
{
    return quot2_f_v((double) mkint(a), b);
}

Sex
quot2xv(Sex a, Sex b)
{
    return quot2_f_v(floatit(a), b);
}

Sex
quot2vi(Sex a, Sex b)
{
    return quot2_v_f(a, (double) mkint(b));
}

Sex
quot2vx(Sex a, Sex b)
{
    return quot2_v_f(a, floatit(b));
}

Sex
quot2fv(Sex a, Sex b)
{
    return quot2_f_v(float_val(a), b);
}


Sex
quot2vf(Sex a, Sex b)
{
    return quot2_v_f(a, float_val(b));
}

Sex
quot2vv(Sex a, Sex b)
{
    double          vl,
                    vu;

    if (interval_lower(a) > 0.0) {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a);
	    vu = interval_upper(a);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a);
	    vu = interval_lower(a);
	} else {
	    return aerror("quot2vv: zero divide");
	}
	vl = vl / interval_upper(b);
	vu = vu / interval_lower(b);
    } else if (interval_upper(a) < 0.0) {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a);
	    vu = interval_upper(a);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a);
	    vu = interval_lower(a);
	} else {
	    return aerror("quot2vv: zero divide");
	}
	vl = vl / interval_lower(b);
	vu = vu / interval_upper(b);
    } else {
	if (interval_lower(b) > 0.0) {
	    vl = interval_lower(a) / interval_lower(b);
	    vu = interval_upper(a) / interval_lower(b);
	} else if (interval_upper(b) < 0.0) {
	    vl = interval_upper(a) / interval_upper(b);
	    vu = interval_lower(a) / interval_upper(b);
	} else {
	    return aerror("quot2vv: zero divide");
	}
    }
    return make_interval(vl, vu);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
numeq2iv(Sex a, Sex b)
{
    double          f,
                    v;
    f = mkint(a) * 2;
    v = interval_upper(b) + interval_lower(b);
    if (f == v)
	return YES;
    return NO;
}


Sex
numeq2xv(Sex a, Sex b)
{
    double          f,
                    v;
    f = floatit(a) * 2;
    v = interval_upper(b) + interval_lower(b);
    if (f == v)
	return YES;
    return NO;
}

Sex
numeq2vx(Sex a, Sex b)
{
    double          f,
                    v;
    f = floatit(b) * 2;
    v = interval_upper(a) + interval_lower(a);
    if (f == v)
	return YES;
    return NO;
}

Sex
numeq2fv(Sex a, Sex b)
{
    double          f,
                    v;
    f = float_val(a) * 2;
    v = interval_upper(b) + interval_lower(b);
    if (f == v)
	return YES;
    return NO;
}


Sex
numeq2vv(Sex a, Sex b)
{
    double          au = interval_upper(a);
    double          al = interval_lower(a);
    double          bu = interval_upper(b);
    double          bl = interval_lower(b);

    if (al == bl && au == bu)
	return YES;
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
lessp2iv(Sex a, Sex b)
{
    double          f = mkint(a);
    double          vl = interval_lower(b);

    if (f < vl)
	return YES;
    return NO;
}


Sex
lessp2vi(Sex a, Sex b)
{
    double          f = mkint(b);
    double          vu = interval_upper(a);

    if (vu < f)
	return YES;
    return NO;
}


Sex
lessp2xv(Sex a, Sex b)
{
    double          f;
    double          vl = interval_lower(b);

    f = floatit(a);
    if (f < vl)
	return YES;
    return NO;
}


Sex
lessp2vx(Sex a, Sex b)
{
    double          f;
    double          vu = interval_upper(a);

    f = floatit(b);
    if (vu < f)
	return YES;
    return NO;
}


Sex
lessp2fv(Sex a, Sex b)
{
    double          f = float_val(a);
    double          vl = interval_lower(b);

    if (f < vl)
	return YES;
    return NO;
}


Sex
lessp2vf(Sex a, Sex b)
{
    double          f = float_val(b);
    double          vu = interval_upper(a);

    if (vu < f)
	return YES;
    return NO;
}


Sex
lessp2vv(Sex a, Sex b)
{
    double          au = interval_upper(a);
    double          bl = interval_lower(b);

    if (au < bl)
	return YES;
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
appeq2iv(Sex a, Sex b)
{
    double          f = mkint(a);
    double          vu = interval_upper(b);
    double          vl = interval_lower(b);

    if (f >= vl && f <= vu)
	return YES;
    return NO;
}


Sex
appeq2xv(Sex a, Sex b)
{
    double          f;
    double          vu = interval_upper(b);
    double          vl = interval_lower(b);

    f = floatit(a);
    if (f >= vl && f <= vu)
	return YES;
    return NO;
}

Sex
appeq2vx(Sex a, Sex b)
{
    double          f;
    double          vu = interval_upper(a);
    double          vl = interval_lower(a);

    f = floatit(b);
    if (f >= vl && f <= vu)
	return YES;
    return NO;
}


Sex
appeq2fv(Sex a, Sex b)
{
    double          f = float_val(a);
    double          vu = interval_upper(b);
    double          vl = interval_lower(b);

    if (f >= vl && f <= vu)
	return YES;
    return NO;
}


Sex
appeq2vv(Sex a, Sex b)
{
    double          au = interval_upper(a);
    double          al = interval_lower(a);
    double          bu = interval_upper(b);
    double          bl = interval_lower(b);

    if (al <= bl && au >= bl)
	return YES;
    if (bl <= al && bu >= al)
	return YES;
    return NO;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/
