/*
 * list.c               (C) Copyright 2007 by Fujio Kako, Nara Women's University.
 *  Basic functions.
 * $Header: /Users/kako/asuka/src/RCS/list.c,v 2.24 2016/05/23 01:26:26 kako Exp $
 */

static char const rcsid[] =
    "$Id: list.c,v 2.24 2016/05/23 01:26:26 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "kctype.h"
#include "errors.h"
#include "externs.h"
#include "entries.h"
#include "arith.h"

#define INCLUDE_FEND

#include "list.h"

extern long     number_of_gc_called;

/************************************************************************** 
 *
 * Some basic functions                                                 
 * 
 *************************************************************************/



static          Sex
Acar(Sex A)
{
    if (!is_cons(A))
	return aerror1("car", A);
    return car(A);
}

static          Sex
Acdr(Sex A)
{
    if (!is_cons(A))
	return aerror1("cdr", A);
    return cdr(A);
}

static          Sex
Arplaca(Sex A, Sex Q)
{
    if (!is_cons(A))
	return aerror1("rplaca", A);
    car(A) = Q;
    return A;
}

static          Sex
Arplacd(Sex A, Sex Q)
{
    if (!is_cons(A))
	return aerror1("rplacd", A);
    cdr(A) = Q;
    return A;
}



static          Sex
Aatom(Sex p)
{
    if (is_number(p))
	return true;
    else if (is_symbol(p))
	return true;
    else if (is_string(p))
	return true;
    return nil;
    // return Bool_to_Sex(!is_cons(p));
}

static          Sex
Aeq(Sex a, Sex b)
{
    return Bool_to_Sex(a == b);
}

static          Sex
Anthcdr(Sex u, Sex n)
{
    long            i;
    if (u == nil || !is_fixnum(n))
	error2("nth: invalid argument", u, n);
    i = mkint(n);
    while (i-- > 0) {
	if (!is_cons(u))
	    error2("nth: invalid argument", u, n);
	u = cdr(u);
    }
    return u;
}

static          Sex
Areplace(Sex u, Sex f, Sex n)
{
    long            i;
    Sex             v;
    if (u == nil || !is_fixnum(n))
	error2("replace: invalid argument", u, n);
    v = u;
    i = mkint(n);
    while (i-- > 1) {
	if (!is_cons(u))
	    error2("replace: invalid argument", u, n);
	u = cdr(u);
    }
    if (u == nil)
	error2("replace: invalid argument", u, n);
    car(u) = f;
    return v;
}


static          Sex
Apnth(Sex u, Sex n)
{
    long            i;
    if (u == nil || !is_fixnum(n))
	error2("nth: invalid argument", u, n);
    i = mkint(n);
    while (i-- > 1) {
	if (!is_cons(u))
	    error2("nth: invalid argument", u, n);
	u = cdr(u);
    }
    return u;
}

static          Sex
Anth(Sex u, Sex n)
{
    long            i;
    if (u == nil || !is_fixnum(n))
	error2("nth: invalid argument", u, n);
    i = mkint(n);
    while (i-- > 1) {
	if (!is_cons(u))
	    error2("nth: invalid argument", u, n);
	u = cdr(u);
    }
    return car(u);
}

static          Sex
Alistp(Sex p)
{
    while (is_cons(p))
	p = cdr(p);
    if (p == nil)
	return true;
    return nil;
}

static          Sex
Anumberp(Sex p)
{
    return Bool_to_Sex(is_number(p));
}



static          Sex
Afloatp(Sex p)
{
    return is_float(p) ? true : nil;
}


static          Sex
Aget(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (!is_symbol(a))
	return nil;
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return cdr(w);
	pl = cdr(pl);
    }
    return nil;
}

static          Sex
Aput(Sex a, Sex b, Sex c)
{
    Sex             pl;
    if (!is_symbol(a))
	return c;
    pl = plist(a);
    while (pl != nil) {
	Sex             w = car(pl);
	if (car(w) == b) {
	    cdr(w) = c;
	    return c;
	} else
	    pl = cdr(pl);
    }
    stackcheck();
    push2(a, c);
    b = Lacons(b, c, plist(a));
    pop2(c, a);
    plist(a) = b;
    return c;
}

static          Sex
Aremprop(Sex a, Sex b)
{
    Sex             pl,
                    prevp;
    if (!is_symbol(a))
	return nil;
    prevp = nil;
    pl = plist(a);
    while (pl != nil) {
	Sex             w = car(pl);
	if (car(w) == b) {
	    pl = cdr(pl);
	    if (prevp == nil)
		plist(a) = pl;
	    else
		cdr(prevp) = pl;
	    return cdr(w);
	}
	prevp = pl;
	pl = cdr(prevp);
    }
    return nil;
}


/*****************************************************************************/
/*
 * Storage allocation.                                                  
 */
/*****************************************************************************/


static          Sex
Acons(Sex a, Sex b)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    car(r) = a;
    cdr(r) = b;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Arcons(Sex a, Sex b)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    cdr(a) = (Sex) r;
    car(r) = b;
    cdr(r) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Axcons(Sex b, Sex a)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    car(r) = a;
    cdr(r) = b;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Ancons(Sex a)
{
    Cons_Cell      *r;
    r = cellpointer - 1;
    car(r) = a;
    cdr(r) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Alist2(Sex a, Sex b)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Alist2star(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = c;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Aacons(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 2;
    car(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    cdr(r) = c;
    car(r + 1) = a;
    cdr(r + 1) = b;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Alist3(Sex a, Sex b, Sex c)
{
    Cons_Cell      *r;
    r = cellpointer - 3;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}


static          Sex
Alist4(Sex a, Sex b, Sex c, Sex d)
{
    Cons_Cell      *r;
    r = cellpointer - 4;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Alist5(Sex a, Sex b, Sex c, Sex d, Sex e)
{
    Cons_Cell      *r;
    r = cellpointer - 5;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}


#ifdef ASUKA
static          Sex
Alist6(Sex a, Sex b, Sex c, Sex d, Sex e, Sex f)
{
    /*
     * Sex s; 
     */
    Cons_Cell      *r;
    r = cellpointer - 6;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = (Sex) ((char *) (r + 5) + TAG_CONS);
    car(r + 5) = f;
    cdr(r + 5) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}

static          Sex
Alist7(Sex a, Sex b, Sex c, Sex d, Sex e, Sex f, Sex g)
{
    /*
     * Sex s; 
     */
    Cons_Cell      *r;
    r = cellpointer - 7;
    car(r) = a;
    cdr(r) = (Sex) ((char *) (r + 1) + TAG_CONS);
    car(r + 1) = b;
    cdr(r + 1) = (Sex) ((char *) (r + 2) + TAG_CONS);
    car(r + 2) = c;
    cdr(r + 2) = (Sex) ((char *) (r + 3) + TAG_CONS);
    car(r + 3) = d;
    cdr(r + 3) = (Sex) ((char *) (r + 4) + TAG_CONS);
    car(r + 4) = e;
    cdr(r + 4) = (Sex) ((char *) (r + 5) + TAG_CONS);
    car(r + 5) = f;
    cdr(r + 5) = (Sex) ((char *) (r + 6) + TAG_CONS);
    car(r + 6) = g;
    cdr(r + 6) = nil;
    cellpointer = r;
    if (cellpointer <= celllimit)
	return reclaim((Sex) ((char *) r + TAG_CONS), GC_CONS, 0);
    else
	return (Sex) ((char *) r + TAG_CONS);
}
#endif


static          Sex
Acaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caar", a);
}
static          Sex
Acdar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdar", a);
}
static          Sex
Acadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("cadr", a);
}
static          Sex
Acddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cddr", a);
}


static          Sex
Acaaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caaar", a);
}
static          Sex
Acdaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdaar", a);
}
static          Sex
Acadar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("cadar", a);
}
static          Sex
Acddar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cddar", a);
}
static          Sex
Acaadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caadr", a);
}
static          Sex
Acdadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdadr", a);
}
static          Sex
Acaddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("caddr", a);
}
static          Sex
Acdddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cdddr", a);
}

static          Sex
Acaaaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = car(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caaaar", a);
}
static          Sex
Acdaaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = car(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdaaar", a);
}
static          Sex
Acadaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("cadaar", a);
}
static          Sex
Acddaar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cddaar", a);
}
static          Sex
Acaadar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caadar", a);
}
static          Sex
Acdadar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdadar", a);
}
static          Sex
Acaddar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("caddar", a);
}
static          Sex
Acdddar(Sex a)
{
    if (is_cons(a) && is_cons(a = car(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cdddar", a);
}
static          Sex
Acaaadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = car(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caaadr", a);
}
static          Sex
Acdaadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = car(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdaadr", a);
}
static          Sex
Acadadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("cadadr", a);
}
static          Sex
Acddadr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = car(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cddadr", a);
}
static          Sex
Acaaddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return car(a);
    return aerror1("caaddr", a);
}
static          Sex
Acdaddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = car(a)))
	return cdr(a);
    return aerror1("cdaddr", a);
}
static          Sex
Acadddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return car(a);
    return aerror1("cadddr", a);
}
static          Sex
Acddddr(Sex a)
{
    if (is_cons(a) && is_cons(a = cdr(a)) &&
	is_cons(a = cdr(a)) && is_cons(a = cdr(a)))
	return cdr(a);
    return aerror1("cddddr", a);
}



static          Sex
Aidp(Sex a)
{
    return Bool_to_Sex(is_symbol(a));
}


static          Sex
Apairp(Sex a)
{
    return Bool_to_Sex(is_cons(a));
}

static          Sex
Aconstantp(Sex a)
{
    return Bool_to_Sex(is_number(a) || is_vechdr(a) || is_binhdr(a));
}



static          Sex
Abigp(Sex p)
{
    return is_bigint(p) ? true : nil;
}

static          Sex
Afixp(Sex p)
{
    switch (((long) p) & TAG_BITS) {
    case TAG_FIXNUM:
	return true;
    case TAG_NUMBERS:
	switch (numtype_of(p)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    return true;
	}
    }
    return nil;
}

static          Sex
Alogicalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_LOGICAL)
	return true;
    return nil;
}


static          Sex
Arationalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_RATIONAL)
	return true;
    return nil;
}

static          Sex
Abfloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BFLOAT)
	return true;
    return nil;
}


static          Sex
Aefloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_EFLOAT)
	return true;
    return nil;
}


static          Sex
Abefloatp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BEFLOAT)
	return true;
    return nil;
}

static          Sex
Acomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_COMPLEX)
	return true;
    return nil;
}

static          Sex
Aecomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_ECOMPLEX)
	return true;
    return nil;
}


static          Sex
Abecomplexp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BECOMPLEX)
	return true;
    return nil;
}

static          Sex
Aintervalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_INTERVAL)
	return true;
    return nil;
}

static          Sex
Abintervalp(Sex p)
{
    if (is_numhdr(p) && type_of_header(numhdr(p)) == T_BINTERVAL)
	return true;
    return nil;
}



bool
AAstringp(Sex a)
{
    if (is_string(a))
	return YES;
    return NO;
}

static          Sex
Astringp(Sex a)
{
    if (!is_string(a))
	return nil;
    else
	return true;
}


static          Sex
Avectorp(Sex a)
{
    if (!is_simple_vector(a))
	return nil;
    return true;
}


static          Sex
Apackagep(Sex a)
{
    if (!is_package(a))
	return nil;
    return true;
}

static          Sex
Apackage_cnt(Sex a)
{
    if (!is_package(a))
	return nil;
    return package_cnt(a);
}

static          Sex
Ahashp(Sex a)
{
    if (!is_hashtable(a))
	return nil;
    return true;
}


static          Sex
Aflagp(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (!is_symbol(a))
	return nil;
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return true;
	pl = cdr(pl);
    }
    return nil;
}

static          Sex
Aflagpcar(Sex a, Sex b)
{
    Sex             pl,
                    w;
    if (a != nil) {
	if (!is_cons(a))
	    return nil;
	a = car(a);
	if (!is_symbol(a))
	    return nil;
    }
    pl = plist(a);
    while (pl != nil) {
	w = car(pl);
	if (car(w) == b)
	    return true;
	pl = cdr(pl);
    }
    return nil;
}

static          Sex
Aflag(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             v = car(a),
	    pl;
	a = cdr(a);
	if (!is_symbol(v))
	    continue;
	push2(a, b);
	pl = plist(v);
	while (pl != nil) {
	    Sex             w = car(pl);
	    if (car(w) == b) {
		cdr(w) = true;
		goto already_flagged;
	    } else
		pl = cdr(pl);
	}
	push(v);
	b = Lacons(b, true, plist(v));
	pop(v);
	plist(v) = b;
      already_flagged:
	pop2(b, a);
    }
    return nil;
}

static          Sex
Aremflag(Sex a, Sex b)
{
    while (is_cons(a)) {
	Sex             pl,
	                prevp,
	                v = car(a);
	a = cdr(a);
	if (!is_symbol(v))
	    continue;
	prevp = nil;
	pl = plist(v);
	while (pl != nil) {
	    Sex             w = car(pl);
	    if (car(w) == b) {
		pl = cdr(pl);
		if (prevp == nil)
		    plist(v) = pl;
		else
		    cdr(prevp) = pl;
		break;
	    }
	    prevp = pl;
	    pl = cdr(prevp);
	}
    }
    return nil;
}


static          Sex
Amkquote(Sex a)
{
    a = Llist2(quote_symbol, a);
    return a;
}

static          Sex
Apair(Sex a, Sex b)
{
    Sex             r = nil;
    stackcheck();
    while (is_cons(a) && is_cons(b)) {
	push2(a, b);
	r = Lacons(car(a), car(b), r);
	pop2(b, a);
	a = cdr(a);
	b = cdr(b);
    }
    a = nil;
    while (r != nil) {
	b = cdr(r);
	cdr(r) = a;
	a = r;
	r = b;
    }
    return a;
}

static long
memqcount(Sex a, Sex b)
/*
 * Counts how many times a is a member of the list b
 */
{
    long            r = 0;
    if (is_symbol(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		r++;
	    b = cdr(b);
	}
	return r;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb)
	    r++;
	b = cdr(b);
    }
    return r;
}

static long
membercount(Sex a, Sex b)
/*
 * Counts how many times a is a member of the list b
 */
{
    long            r = 0;
    if (is_symbol(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		r++;
	    b = cdr(b);
	}
	return r;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb || equal(a, cb) != nil)
	    r++;
	b = cdr(b);
    }
    return r;
}

static          Sex
Aintersection(Sex a, Sex b)
{
    Sex             r = nil,
	w;
    long            n1,
                    n2;
    stackcheck();
    push(b);
    while (is_cons(a)) {
	push2(a, r);
	w = Lmember(car(a), stack[-2]);
	if (w != nil) {
	    n1 = membercount(car(stack[-1]), stack[0]);
	    if (n1 != 0) {
		n2 = membercount(car(stack[-1]), stack[-2]);
		if (n2 > n1)
		    n1 = 0;
	    }
	    if (n1 == 0) {
		pop(r);
		a = stack[0];
		r = cons(car(a), r);
		pop(a);
	    } else
		pop2(r, a);
	} else
	    pop2(r, a);
	a = cdr(a);
    }
    popv(1);
    a = nil;
    while (is_cons(r)) {
	b = r;
	r = cdr(r);
	cdr(b) = a;
	a = b;
    }
    return a;
}

static          Sex
Aintersectionq(Sex a, Sex b)
{
    Sex             r = nil,
	w;
    long            n1,
                    n2;
    stackcheck();
    push(b);
    while (is_cons(a)) {
	push2(a, r);
	w = Lmemq(car(a), stack[-2]);
	if (w != nil) {
	    n1 = memqcount(car(stack[-1]), stack[0]);
	    if (n1 != 0) {
		n2 = memqcount(car(stack[-1]), stack[-2]);
		if (n2 > n1)
		    n1 = 0;
	    }
	    if (n1 == 0) {
		pop(r);
		a = stack[0];
		r = cons(car(a), r);
		pop(a);
	    } else
		pop2(r, a);
	} else
	    pop2(r, a);
	a = cdr(a);
    }
    popv(1);
    a = nil;
    while (is_cons(r)) {
	b = r;
	r = cdr(r);
	cdr(b) = a;
	a = b;
    }
    return a;
}

Sex
Amerge(Sex a, Sex b)
{
    Sex             c,
                    d;
    stackcheck();
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[0]
    while (is_cons(a) && is_cons(b)) {
	c = car(a);
	d = car(b);
	if (c == d || equal(c, d) != nil) {
	    r = cons(c, r);
	    a = cdr(a);
	    b = cdr(b);
	} else if (orderp(c, d) > 0) {
	    r = cons(c, r);
	    a = cdr(a);
	} else {
	    r = cons(d, r);
	    b = cdr(b);
	}
    }
    while (is_cons(a)) {
	r = cons(car(a), r);
	a = cdr(a);
    }
    while (is_cons(b)) {
	r = cons(car(b), r);
	b = cdr(b);
    }
    c = Lnreverse(r);
#undef a
#undef b
#undef r
    popv(3);
    return c;
}


static          Sex
Aunion(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             c;
	push2(a, b);
	c = Lmember(car(a), b);
	pop(b);
	if (c == nil) {
	    b = cons(car(stack[0]), b);
	}
	pop(a);
	a = cdr(a);
    }
    return b;
}

static          Sex
Aunionq(Sex a, Sex b)
{
    stackcheck();
    while (is_cons(a)) {
	Sex             c;
	push2(a, b);
	c = Lmemq(car(a), b);
	pop(b);
	if (c == nil) {
	    b = cons(car(stack[0]), b);
	}
	pop(a);
	a = cdr(a);
    }
    return b;
}

Sex
Asetdiff(Sex a, Sex b)
{
    Sex             r = nil;
    Sex             c;
    stackcheck();
    push2(a, b);
    push(r);
#define a stack[-2]
#define b stack[-1]
#define r stack[-0]
    while (is_cons(a)) {
	c = Lmember(car(a), b);
	if (c == nil) {
	    r = cons(car(a), r);
	}
	a = cdr(a);
    }
#undef a
#undef b
#undef r
    pop(r);
    popv(2);
    return Lnreverse(r);
}

static          Sex
Asetdiffq(Sex a, Sex b)
{
    Sex             r = nil;
    Sex             c;
    stackcheck();
    push2(a, b);
    push(r);
    while (is_cons(stack[-2])) {
	c = Lmemq(car(stack[-2]), stack[-1]);
	if (c == nil) {
	    stack[0] = cons(car(stack[-2]), stack[0]);
	}
	stack[-2] = cdr(stack[-2]);
    }
    pop(r);
    popv(2);
    return Lnreverse(r);
}

static          Sex
Asort(Sex a)
{
    int             i;
    long            n;
    Sex             b;
    Sex            *p;

    b = a;
    n = 0;
    while (is_cons(b)) {
	n++;
	b = cdr(b);
    }
    push(a);
    b = make_vector(n);
    pop(a);
    p = &vect_value(b, 0);
    for (i = 0; i < n; i++) {
	*p++ = car(a);
	a = cdr(a);
    }
    p = &vect_value(b, 0);
    quick_sort(p, 0, n - 1);
    a = nil;
    for (i = 0; i < n; i++)
	a = cons(*p++, a);
    return Lnreverse(a);
}

static void
quick_sort(Sex * a, long left, long right)
{
    long            l_save,
                    r_save;
    long            ll;
    Sex             pivot;

    l_save = left;
    r_save = right;
    pivot = a[left];

    while (left < right) {
	while (orderp(a[right], pivot) <= 0 && left < right)
	    right--;
	if (left != right) {
	    a[left] = a[right];
	    left++;
	}
	while (orderp(a[left], pivot) >= 0 && left < right)
	    left++;
	if (left != right) {
	    a[right] = a[left];
	    right--;
	}
    }
    a[left] = pivot;
    ll = left;
    left = l_save;
    right = r_save;
    if (left < ll)
	quick_sort(a, left, ll - 1);
    if (right > ll)
	quick_sort(a, ll + 1, right);
}



Sex
Aerror(Sex msg, Sex a)
{
    push2(msg, a);
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: %p, ", stack[-1]);
	lisp_printf("%p\n", stack[0]);
    }
    popv(2);
    longjmp(*errorset_buffer, 2);
    return nil;
}

static          Sex
Aerror0()
{
    miscflags &= ~(MESSAGES_FLAG);
    return nil;
}

static          Sex
Aerror1()
{
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: \n");
    }
    longjmp(*errorset_buffer, 2);
    return nil;
}

static          Sex
Aerror2(Sex msg)
{
    if (miscflags & MESSAGES_FLAG) {
	lisp_printf("\n *** error: %P\n", msg);
    }
    longjmp(*errorset_buffer, 2);
    return nil;
}

static          Sex
Astop(Sex code)
{
    if (!is_fixnum(code))
	return aerror("stop");
    return nil;
}

static          Sex
Afluid(Sex a)
{
    Sex             q,
                    b;

    b = a;
    stackcheck();
    while (is_cons(b)) {
	q = car(b);
	if (!is_symbol(q)) {
	    push2(a, b);
	    warnings("non symbol", q, nil);
	    pop2(b, a);
	    /*
	     * return aerror("fluid"); 
	     */
	} else {
	    qtag(q) |= SYM_FLUID;
	    if (value(q) == undefined_var)
		value(q) = nil;
	}
	b = cdr(b);
    }
    return a;
}

static          Sex
Aglobal(Sex a)
{
    Sex             q,
                    b;
    b = a;
    while (is_cons(b)) {
	q = car(b);
	if (!is_symbol(q))
	    return aerror("global");
	qtag(q) |= (SYM_FLUID | SYM_GLOBAL);
	if (value(q) == undefined_var)
	    value(q) = nil;
	b = cdr(b);
    }
    return a;
}

static          Sex
Aunfluid(Sex a)
{
    if (!is_symbol(a))
	return nil;
    qtag(a) &= ~(SYM_FLUID | SYM_GLOBAL);
    value(a) = nil;
    return a;
}

static          Sex
Aunglobal(Sex a)
{
    long            t;

    if (!is_symbol(a))
	return nil;
    t = qtag(a) & ~(SYM_GLOBAL);
    if ((t & SYM_FLUID) == 0)
	value(a) = nil;
    qtag(a) = t;
    return a;
}

static          Sex
Afluidp(Sex a)
{
    if (!is_symbol(a))
	return nil;
    /*
     * else if ((qtag(a) & (SYM_FLUID | SYM_GLOBAL)) == SYM_FLUID) 
     */
    else if ((qtag(a) & SYM_FLUID) != 0)
	return true;
    else
	return nil;
}

static          Sex
Aglobalp(Sex a)
{
    if (!is_symbol(a))
	return nil;
    else if ((qtag(a) & SYM_GLOBAL) != 0)
	return true;
    else
	return nil;
}

static          Sex
Aboundp(Sex a)
{
    if (!is_symbol(a))
	return nil;
    else if ((qtag(a) & SYM_FLUID) == 0)
	return nil;
    else
	return true;
}

static          Sex
Asymbol_value(Sex a)
{
    if (!is_symbol(a))
	return a;
    else
	return value(a);
}


static          Sex
Aset(Sex a, Sex b)
{
    if (!is_symbol(a) || a == nil || a == true) {
	return aerror("set");
    }
    value(a) = b;
    return b;
}




static          Sex
Acodep(Sex a)
{
    if (!is_symbol(a))
	return nil;
    if ((qtag(a) & PROC_CCODE) == 0)
	return nil;
    else
	return true;
}

static          Sex
Agetvector(long tag, long type, long size)
{
    while (YES) {
	char           *r = (char *) fwdpointer;
	long            alloc_size = (long) alignit(size);

	if (alloc_size > (long) fwdlimit - (long) r) {
	    reclaim(nil, GC_FWD, alloc_size);
	    continue;
	}
	fwdpointer = (Sex *) (r + alloc_size);
	*((Tag *) r) = type | TAG_HDRS;
	*((long *) (r + SEX)) = size;

	return (Sex) (r + tag);
    }
}

static          Sex
Agetvector_init(long n, Sex k)
{
    Sex             p;
    push(k);
    p = Agetvector(TAG_VECTOR, T_SIMPLE, n);
    pop(k);
    if ((n & SEX) != 0)
	n += SEX;
    n = bnum_n(n);
    while (n > 0) {
	n--;
	vector_value(p, n) = k;
    }
    return p;
}

double          base_time,
                io_time,
                gc_time;

static          Sex
Afloat_time()
{
    double          clock;

    clock = get_time();
    clock = clock - base_time - io_time - gc_time;
    return make_float(clock);
}


static          Sex
Atime()
{
    double          clock;

    clock = get_time();
    clock = clock - base_time - io_time - gc_time;
    return make_float((clock * 1000));
}

static          Sex
Agctime()
{
    return make_float((gc_time * 1000));
}

static          Sex
Agccalled()
{
    return mkfxat(number_of_gc_called);
}

static          Sex
Adate()
{
    Sex             w;
    time_t          t = time(NULL);
    char            today[32];
#ifdef Android
    strcpy(today, (const time_t *) ctime(&t));
#else
    strcpy(today, ctime(&t));
#endif
    w = make_string2(today, 24);
    return w;
}


static          Sex
deleqip(Sex a, Sex l)
{
    Sex             w,
                    r;
    if (l == nil)
	return nil;
    if (car(l) == a)
	return cdr(l);
    r = l;
    while (w = l, (l = cdr(l)) != nil) {
	if (car(l) == a) {
	    cdr(w) = cdr(l);
	    return r;
	}
    }
    return r;
}


static          Sex
Aputd(Sex a, Sex b, Sex c)
{
    long            ptype;

    if (!is_symbol(a))
	return aerror1("putd", a);
    if (b == expr_symbol) {
	ptype = PROC_EXPR;
    } else if (b == fexpr_symbol) {
	ptype = PROC_FEXPR;
    } else if (b == macro_symbol) {
	ptype = PROC_MACRO;
    } else if (b == smacro_symbol) {
	ptype = PROC_SMACRO;
    } else {
	return aerror1("putc", b);
    }
    if (is_symbol(c)) {
	if ((qtag(c) & SYM_PROC) == 0)
	    return aerror1("putd", c);
	qtag(a) = (qtag(a) & ~PROC_MASK) | (qtag(c) & PROC_MASK);
	proc(a) = proc(c);
	qnargs(a) = qnargs(c);
	return a;
    } else if (!is_cons(c)) {
	return aerror1("putd", c);
    } else if (is_fixnum(car(c))) {
	Sex             d = cdr(c);
	if (!is_logical(d))
	    aerror1("putd", c);
	ptype |= PROC_CCODE;
	proc(a) = (lisp_proc *) logical_value(d)[0];
	qnargs(a) = mkint(car(c));
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(c) == lambda) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else if (car(b) == funarg) {
	Sex             bvl;
	long            nargs;
	c = cdr(c);
	bvl = car(c);
	nargs = 0;
	while (is_cons(bvl)) {
	    nargs++;
	    bvl = cdr(bvl);
	}
	proc(a) = (lisp_proc *) c;
	qnargs(a) = nargs;
	qtag(a) = (qtag(a) & ~PROC_MASK) | ptype;
	return a;
    } else
	return aerror1("putd", b);
}



static          Sex
Agetd(Sex a)
{
    Tag             h;
    Sex             type;
    long            narg;
    if (!is_symbol(a))
	return aerror1("getd", a);
    h = qtag(a);
    switch (h & PROC_MASK) {
    case PROC_CSMACRO:
	a = cons(lambda, Lget(a, smacro_symbol));
	type = smacro_symbol;
	break;
    case PROC_SMACRO:
	a = cons(lambda, (Sex) proc(a));
	type = smacro_symbol;
	break;
    case PROC_CMACRO:
	a = cons(lambda, Lget(a, macro_symbol));
	type = macro_symbol;
	break;
    case PROC_MACRO:
	a = cons(lambda, (Sex) proc(a));
	type = macro_symbol;
	break;
    case PROC_EXPR:
	a = cons(lambda, (Sex) proc(a));
	type = expr_symbol;
	break;
    case PROC_FEXPR:
	a = cons(lambda, (Sex) proc(a));
	type = fexpr_symbol;
	break;
    case PROC_SUBR:
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = expr_symbol;
	break;
    case PROC_FSUBR:
	narg = qnargs(a);
	a = mklogical((long) proc(a));
	a = cons(mkfxat(narg), a);
	type = fexpr_symbol;
	break;
    default:
	return nil;
    }
    return cons(type, a);
}


static          Sex
Amovd(Sex a, Sex b)
{
    b = Lgetd(b);
    if (is_cons(b)) {
	Lputd(a, car(b), cdr(b));
    }
    return b;
}


static          Sex
Aremd(Sex a)
{
    Sex             res;
    if (!is_symbol(a))
	return aerror1("remd", a);
    res = Lgetd(a);
    if (res == nil)
	return nil;
    qtag(a) = qtag(a) & ~(SYM_TRACED | PROC_MASK);
    proc(a) = (lisp_proc *) undefined_proc;
    qnargs(a) = 0;
    return res;
}


static          Sex
Atraceall(Sex a)
{
    if (a == nil)
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    else
	trace_flag |= (SYM_TRACED | SYM_PROC);
    return a;
}

static          Sex
Atrace(Sex a)
{
    Sex             q,
                    m;

    if (is_cons(a)) {
	q = a;
	while (is_cons(q)) {
	    m = car(q);
	    if (is_symbol(m) && (qtag(m) & SYM_PROC) != 0)
		qtag(m) |= SYM_TRACED;
	    q = cdr(q);
	}
	trace_flag |= SYM_TRACED;
    } else if (is_fixnum(a)) {
	trace_level = mkint(a);
    } else if (a == nil)
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    else
	trace_flag |= SYM_TRACED;
    return a;
}

static          Sex
Auntrace(Sex a)
{
    Sex             q,
                    m;

    if (is_cons(a)) {
	q = a;
	while (is_cons(q)) {
	    m = car(q);
	    if (is_symbol(m) && (qtag(m) & SYM_PROC) != 0)
		qtag(m) &= ~SYM_TRACED;
	    q = cdr(q);
	}
    } else if (a == nil)
	trace_flag |= SYM_TRACED;
    else
	trace_flag &= ~(SYM_TRACED | SYM_PROC);
    return a;
}

static          Sex
Amacro_function(Sex a)
{
    if (!is_symbol(a))
	return nil;
    else if ((qtag(a) & PROC_MASK) != PROC_MACRO)
	return nil;
    else
	return cons(lambda, (Sex) proc(a));
}



static          Sex
Asymbol_length(Sex a)
{
    if (!is_symbol(a))
	return aerror1("symbol-name", a);
    return mkfxat(symbol_length(a));
}

static          Sex
Asymbol_name(Sex a)
{
    if (!is_symbol(a))
	return aerror1("symbol-name", a);
    return make_string2(pname(a), symbol_length(a));
}

static          Sex
Aplist(Sex a)
{
    if (!is_symbol(a))
	return aerror1("plist", a);
    return plist(a);
}


static          Sex
Aatom_tag(Sex a)
{
    if (!is_symbol(a))
	return aerror1("plist", a);
    return mklogical(qtag(a));
}

static          Sex
Anumber_tag(Sex a)
{
    if (!is_numhdr(a))
	return aerror1("number-tag", a);
    return mklogical(numhdr(a));
}

static          Sex
Anumber_length(Sex a)
{
    if (!is_numhdr(a))
	return aerror1("number-tag", a);
    return mkfxat(bignum_length(a));
}


/************************************************************************** 
 *
 * array and matrix functions                                                 
 * 
 *************************************************************************/

static          Sex
Amkvect(Sex n)
{
    long            n1;
    if (!is_fixnum(n) || (n1 = mkint(n)) < 0)
	return aerror1("mkvect", n);
    return Agetvector_init(bnum_len(n1 + 1), nil);
}

long
Acheckarray1(Sex n1)
{
    long            d1;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0)
	return aerror1("checkarray1", n1);
    return d1;
}

long
Acheckarray2(Sex n1, Sex n2)
{
    long            d1,
                    d2;
    long            n;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0
	|| !is_fixnum(n2) || (d2 = mkint(n2)) <= 0)
	return aerror2("checkarray2", n1, n2);
    n = d1 * d2;
    if (n <= 0)
	return aerror2("checkarray2", n1, n2);
    return n;
}


long
Acheckarray3(Sex n1, Sex n2, Sex n3)
{
    long            n;
    long            d1,
                    d2,
                    d3;
    if (!is_fixnum(n1) || (d1 = mkint(n1)) <= 0
	|| !is_fixnum(n2) || (d2 = mkint(n2)) <= 0
	|| !is_fixnum(n3) || (d3 = mkint(n3)) <= 0)
	return aerror3("checkarray3", n1, n2, n3);
    n = d1 * d2 * d3;
    if (n <= 0)
	return aerror3("checkarray3", n1, n2, n3);
    return n;
}

static          Sex
Amkarray1(Sex n1)
{
    long            i,
                    d1;
    Sex             p;
    Sex            *m;
    d1 = Acheckarray1(n1);
    p = Agetvector(TAG_VECTOR, T_ARRAY, bnum_len(d1 + 1));
    m = &array_value(p, 0);
    for (i = 0; i < d1; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n1 = mkfxat(d1);
    n1 = Lncons(n1);
    pop(p);
    array_dim(p) = n1;
    /*
     * printf("mkarray: len = %d\n", veclen(p)); 
     */
    return p;
}

static          Sex
Amkarray2(Sex n1, Sex n2)
{
    long            i,
                    n;
    Sex             p;
    Sex            *m;
    n = Acheckarray2(n1, n2);
    p = Agetvector(TAG_VECTOR, T_ARRAY, bnum_len(n + 1));
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n2 = Lncons(n2);
    n1 = cons(n1, n2);
    pop(p);
    array_dim(p) = n1;
    return p;
}

static          Sex
Amkarray3(Sex n1, Sex n2, Sex n3)
{
    long            i,
                    n;
    Sex             p;
    Sex            *m;
    n = Acheckarray3(n1, n2, n3);
    p = Agetvector(TAG_VECTOR, T_ARRAY, bnum_len(n + 1));
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    push(p);
    n3 = Lncons(n3);
    n2 = cons(n2, n3);
    n1 = cons(n1, n2);
    pop(p);
    array_dim(p) = n1;
    return p;
}

static          Sex
mkarray_fn(Sex args, Sex env)
{
    long            i;
    long            n;
    long            dim[10];
    Sex             p,
                    r;
    Sex            *m;

    if (!is_cons(args))
	return aerror1("mkarray", args);
    stackcheck();
    n = 1;
    for (i = 0; i < 10; i++) {
	push2(args, env);
	r = car(args);
	r = eval(r, env);
	if (!is_fixnum(r))
	    return aerror1("mkarray", r);
	dim[i] = mkint(r);
	n = n * dim[i];
	if (n < 0 || n > 0x0fffffff)
	    return aerror1("mkarray", r);
	pop2(env, args);
	args = cdr(args);
	if (!is_cons(args))
	    break;
    }
    r = nil;
    for (; i >= 0; i--)
	r = cons(mkfxat(dim[i]), r);
    push(r);
    p = Agetvector(TAG_VECTOR, T_ARRAY, bnum_len(n + 1));
    pop(r);
    m = &array_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKARRAY_INIT_VALUE;
    }
    array_dim(p) = r;
    return p;
}


static          Sex
Amkiarray1(Sex n1)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = Acheckarray1(n1);
    p = Agetvector(TAG_BINARY, T_IARRAY, n * sizeof(int) + HDR);
    m = &iarray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amkiarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = Acheckarray2(n1, n2);
    p = Agetvector(TAG_BINARY, T_IARRAY2, n * sizeof(int) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &iarray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amkiarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    int            *m;
    Sex             p;

    n = Acheckarray3(n1, n2, n3);
    p = Agetvector(TAG_BINARY, T_IARRAY3, n * sizeof(int) + HDR + 2 * SEX);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &iarray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amklarray1(Sex n1)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = Acheckarray1(n1);
    p = Agetvector(TAG_BINARY, T_IARRAY, n * sizeof(long) + HDR);
    m = &larray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amklarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = Acheckarray2(n1, n2);
    p = Agetvector(TAG_BINARY, T_IARRAY2, n * sizeof(long) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &larray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amklarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    long           *m;
    Sex             p;

    n = Acheckarray3(n1, n2, n3);
    p = Agetvector(TAG_BINARY, T_IARRAY3,
		   n * sizeof(long) + HDR + 2 * SEX);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &larray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0;
    }
    return p;
}

static          Sex
Amkfarray1(Sex n1)
{
    long            n,
                    i;
    float          *m;
    Sex             p;

    n = Acheckarray1(n1);
    p = Agetvector(TAG_BINARY, T_SARRAY, n * sizeof(float) + HDR);
    m = &sarray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}
static          Sex
Amkfarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    float          *m;
    Sex             p;

    n = Acheckarray2(n1, n2);
    p = Agetvector(TAG_BINARY, T_SARRAY2, n * sizeof(float) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &sarray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

static          Sex
Amkfarray3(Sex n1, Sex n2, Sex n3)
{
    long
                    n,
                    i;
    float          *m;
    Sex             p;

    n = Acheckarray3(n1, n2, n3);
    p = Agetvector(TAG_BINARY, T_SARRAY3,
		   n * sizeof(float) + HDR + 2 * SEX);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &sarray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

static          Sex
Amkdarray1(Sex n1)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = Acheckarray1(n1);
    p = Agetvector(TAG_BINARY, T_DARRAY, n * sizeof(double) + HDR);
    m = &darray_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

static          Sex
Amkdarray2(Sex n1, Sex n2)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = Acheckarray2(n1, n2);
    p = Agetvector(TAG_BINARY, T_DARRAY2, n * sizeof(double) + HDR + SEX);
    barray_dimy(p) = mkint(n2);
    m = &darray2_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

static          Sex
Amkdarray3(Sex n1, Sex n2, Sex n3)
{
    long            n,
                    i;
    double         *m;
    Sex             p;

    n = Acheckarray3(n1, n2, n3);
    p = Agetvector(TAG_BINARY, T_DARRAY3,
		   n * sizeof(double) + HDR + 2 * SEX);
    barray_dimy(p) = mkint(n2);
    barray_dimz(p) = mkint(n3);
    m = &darray3_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = 0.0;
    }
    return p;
}

static          Sex
Agetarray1(Sex v, Sex n1)
{
    long            d1,
                    hl;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0)
	return aerror1("getarray1", n1);

    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 1)
	    return aerror1("getarray not 1d array ", p);
	hl = mkint(car(p));
	if (d1 >= hl)
	    return aerror1("getarray index range", n1);
	else
	    return array_value(v, d1);
    }
    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_1)
	    return aerror1("getarray not array ", v);
	hl = binlen(v) - HDR;
	switch (type_of_header(i)) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    i = iarray_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    i = larray_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    f = sarray_value(v, d1);
	    return make_float(f);
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (d1 >= hl) {
		return aerror1("getarray index range", n1);
	    }
	    f = darray_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray not array ", v);
	}
    }


    return aerror1("getarray", v);
}

static          Sex
Agetarray2(Sex v, Sex n1, Sex n2)
{
    long            d1,
                    d2,
                    hl;
    long            sy;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0)
	return aerror2("getarray2", n1, n2);
    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 2)
	    return aerror1("getarray2 not 2d array", p);
	hl = mkint(car(p));
	sy = mkint(cadr(p));
	if (d1 >= hl || d2 >= sy)
	    return aerror2("getarray2 index range", n1, n2);
	d1 = d1 * sy + d2;
	return array_value(v, d1);
    }

    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_2)
	    return aerror1("getarray2 not array ", v);
	sy = barray_dimy(v);
	if (d2 >= sy)
	    return aerror2("getarray2 index range", n1, n2);
	hl = binlen(v) - HDR - SEX;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY2:
	    hl = hl / sizeof(int);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    i = iarray2_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY2:
	    hl = hl / sizeof(long);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    i = larray2_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY2:
	    hl = hl / sizeof(float);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    f = sarray2_value(v, d1);
	    return make_float(f);
	case T_DARRAY2:
	    hl = hl / sizeof(double);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("getarray2 index range", n1, n2);
	    }
	    f = darray2_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray2 not array ", v);
	}
    }
    return aerror1("getarray2", v);
}

static          Sex
Agetarray3(Sex v, Sex n1, Sex n2, Sex n3)
{
    long            d1,
                    d2,
                    d3,
                    hl;
    long            sy,
                    sz;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0
	|| !is_fixnum(n3) || (d3 = mkint(n3) - A_OFF) < 0)
	return aerror3("mkdarray3", n1, n2, n3);
    if (is_array(v)) {
	p = array_dim(v);

	if (list_length(p) != 3)
	    return aerror1("getarray not a 3d array ", p);
	sy = mkint(cadr(p));
	sz = mkint(cadr(cdr(p)));
	hl = mkint(car(p));
	if (d1 >= hl || d2 >= sy || d3 >= sz)
	    return aerror3("getarray3 index range", n1, n2, n3);
	d1 = (d1 * sy + d2) * sz + d3;
	return array_value(v, d1);
    }
    if (is_binhdr(v)) {
	long            i;
	double          f;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_3)
	    return aerror1("getarray2 not array ", v);
	hl = binlen(v) - HDR - 2 * SEX;
	sy = barray_dimy(v);
	sz = barray_dimz(v);
	if (d2 >= sy || d3 >= sz)
	    return aerror3("getarray3 index range", n1, n2, n3);
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY3:
	    hl = hl / sizeof(int);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray index range", n1, n2);
	    }
	    i = iarray3_value(v, d1);
	    return mkfxat(i);
	case T_LARRAY3:
	    hl = hl / sizeof(long);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray index range", n1, n2);
	    }
	    i = larray3_value(v, d1);
	    return mkfxat(i);
	case T_SARRAY3:
	    hl = hl / sizeof(float);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray index range", n1, n2);
	    }
	    f = sarray3_value(v, d1);
	    return make_float(f);
	case T_DARRAY3:
	    hl = hl / sizeof(double);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror2("getarray index range", n1, n2);
	    }
	    f = darray3_value(v, d1);
	    return make_float(f);
	default:
	    return aerror1("getarray not array ", v);
	}
    }
    return aerror1("getarray", v);
}

static          Sex
getarray_fn(Sex arg, Sex env)
{
    long            n;
    Sex             a1,
                    a2,
                    a3;
    Sex             b;

    stackcheck();
    n = list_length(arg);
    push2(arg, env);
    push4(nil, nil, nil, nil);
#define arg stack[-5]
#define env stack[-4]
#define v stack[-3]
#define n1 stack[-2]
#define n2 stack[-1]
#define n3 stack[0]
    if (n >= 2 && n <= 4) {
	v = eval(car(arg), env);
	arg = cdr(arg);
	switch (n) {
	case 2:
	    a1 = eval(car(arg), env);
	    b = v;
	    popv(6);
	    return Lgetarray1(b, a1);
	case 3:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    a2 = eval(car(arg), env);
	    a1 = n1;
	    b = v;
	    popv(6);
	    return Lgetarray2(b, a1, a2);
	case 4:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    n2 = eval(car(arg), env);
	    arg = cdr(arg);
	    a3 = eval(car(arg), env);
	    a2 = n2;
	    a1 = n1;
	    b = v;
	    popv(6);
	    return Lgetarray3(b, a1, a2, a3);
	}
    }
#undef arg
#undef env
#undef v
#undef n1
#undef n2
#undef n3
    return nil;
}

static          Sex
Asetarray1(Sex v, Sex x, Sex n1)
{
    long            d1,
                    hl;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0)
	return aerror1("setarray1", n1);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 1)
	    return aerror1("setarray not a one dimensional array", v);
	hl = mkint(car(p));
	if (d1 >= hl) {
	    return aerror1("setarray index range", n1);
	}
	array_value(v, d1) = x;
	return x;
    }
    if (is_binhdr(v)) {
	long            i;

	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_1)
	    return aerror1("setarray1", n1);
	hl = binlen(v) - HDR;
	switch (type_of_header(i)) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    iarray_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    larray_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    sarray_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (d1 >= hl) {
		return aerror1("setarray index range", n1);
	    }
	    darray_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    return aerror1("setarray not vector ", v);
}


static          Sex
Asetarray2(Sex v, Sex x, Sex n1, Sex n2)
{
    long            d1,
                    d2,
                    hl;
    long            sy;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0)
	return aerror2("setarray2", n1, n2);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 2)
	    return aerror1("setarray not a two dimensional array", v);
	sy = mkint(cadr(p));
	hl = mkint(car(p));
	if (d1 >= hl || d2 >= sy)
	    return aerror2("setarray index range", n1, n2);
	d1 = d1 * sy + d2;
	array_value(v, d1) = x;
	return x;
    }
    if (is_binhdr(v)) {
	long            i;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_2)
	    return aerror1("setarray2", n1);
	sy = barray_dimy(v);
	hl = binlen(v) - HDR - SEX;
	if (d2 >= sy)
	    return aerror2("setarray2 index range", n1, n2);
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY2:
	    hl = hl / sizeof(int);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    iarray2_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY2:
	    hl = hl / sizeof(long);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    larray2_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY2:
	    hl = hl / sizeof(float);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    sarray2_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY2:
	    hl = hl / sizeof(double);
	    d1 = d1 * sy + d2;
	    if (d1 >= hl) {
		return aerror2("setarray index range", n1, n2);
	    }
	    darray2_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    return aerror1("setarray not vector ", v);
}


static          Sex
Asetarray3(Sex v, Sex x, Sex n1, Sex n2, Sex n3)
{
    long            d1,
                    d2,
                    d3,
                    hl;
    long            sy,
                    sz;
    Sex             p;

    if (!is_fixnum(n1) || (d1 = mkint(n1) - A_OFF) < 0
	|| !is_fixnum(n2) || (d2 = mkint(n2) - A_OFF) < 0
	|| !is_fixnum(n3) || (d3 = mkint(n3) - A_OFF) < 0)
	return aerror3("setarray3", n1, n2, n3);
    if (is_array(v)) {
	p = array_dim(v);
	if (list_length(p) != 3)
	    return aerror1("setarray not a three dimensional array", v);
	hl = mkint(car(p));
	sy = mkint(cadr(p));
	sz = mkint(cadr(cdr(p)));

	if (d1 >= hl || d2 >= sy || d3 >= sz)
	    return aerror3("setarray index range", n1, n2, n3);
	d1 = (d1 * sy + d2) * sz + d3;
	array_value(v, d1) = x;
	return x;
    }
    if (is_binhdr(v)) {
	long            i;
	i = binhdr(v);
	if ((i & T_ARRAY_MASK) != T_ARRAY_3)
	    return aerror1("setarray3", n1);
	sy = barray_dimy(v);
	sz = barray_dimz(v);
	if (d2 >= sy || d3 >= sz)
	    return aerror3("setarray3 index range", n1, n2, n3);
	hl = binlen(v) - HDR - 2 * SEX;
	switch (type_of_header(i)) {
	case T_IARRAY3:
	    hl = hl / sizeof(int);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    iarray3_value(v, d1) = mkint(x);
	    return x;
	case T_LARRAY3:
	    hl = hl / sizeof(long);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    larray3_value(v, d1) = mkint(x);
	    return x;
	case T_SARRAY3:
	    hl = hl / sizeof(float);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    sarray3_value(v, d1) = (float) floatit(x);
	    return x;
	case T_DARRAY3:
	    hl = hl / sizeof(double);
	    d1 = (d1 * sy + d2) * sz + d3;
	    if (d1 >= hl) {
		return aerror3("setarray index range", n1, n2, n3);
	    }
	    darray3_value(v, d1) = floatit(x);
	    return x;
	default:
	    return aerror1("setarray not vector ", v);
	}
    }
    return aerror1("setarray not vector ", v);
}



Sex
Alist2array(Sex a, Sex v)
{
    Sex             p;
    if (!is_array(a))
	return aerror2("list2array", a, v);
    p = array_dim(a);
    switch (list_length(p)) {
    case 1:
	return Alist2array1(a, v);
    case 2:
	return Alist2array2(a, v);
    case 3:
	return Alist2array3(a, v);
    default:
	return aerror1("list2array not a 1,2 or 3 dimensional array", a);
    }
}

static          Sex
Alist2array1(Sex a, Sex v)
{
    long            sx;
    long            j;
    Sex             p;
    Sex            *q;

    p = array_dim(a);
    sx = mkint(car(p));
    q = (Sex *) & array_value(a, 0);

    for (j = 0; j < sx; j++) {
	if (!is_cons(v))
	    break;
	q[j] = car(v);
	v = cdr(v);
    }
    return a;
}


static          Sex
Alist2array2(Sex a, Sex v)
{
    long            i,
                    j;
    long            sx,
                    sy;
    Sex             p;
    Sex            *q;

    p = array_dim(a);
    sy = mkint(cadr(p));
    sx = mkint(car(p));
    q = (Sex *) & array_value(a, 0);
    for (i = 0; i < sx; i++) {
	if (!is_cons(v))
	    break;
	p = car(v);
	v = cdr(v);
	for (j = 0; j < sy; j++) {
	    if (!is_cons(p))
		break;
	    q[j] = car(p);
	    p = cdr(p);
	}
	q += sy;
    }

    return a;
}


static          Sex
Alist2array3(Sex a, Sex v)
{
    long            i,
                    j,
                    k;
    long            sx,
                    sy,
                    sz;
    Sex             p,
                    r;
    Sex            *q;

    p = array_dim(a);
    sx = mkint(car(p));
    sy = mkint(cadr(p));
    sz = mkint(cadr(cdr(p)));
    q = (Sex *) & array_value(a, 0);
    for (i = 0; i < sx; i++) {
	if (!is_cons(v))
	    break;
	p = car(v);
	v = cdr(v);
	for (j = 0; j < sy; j++) {
	    if (!is_cons(p)) {
		q += (sy - j) * sz;
		break;
	    }
	    r = car(p);
	    p = cdr(p);
	    for (k = 0; k < sz; k++) {
		if (!is_cons(r))
		    break;
		q[k] = car(r);
		r = cdr(r);
	    }
	    q += sz;
	}
    }

    return a;
}

static          Sex
Aarray2list(Sex a)
{
    Sex             p;
    if (!is_array(a))
	return aerror1("array2list", a);
    p = array_dim(a);
    switch (list_length(p)) {
    case 1:
	return Aarray2list1(a, p);
    case 2:
	return Aarray2list2(a, p);
    case 3:
	return Aarray2list3(a, p);
    default:
	return aerror1("array2list not a 1,2 or 3 dimensional array", a);
    }
}


static          Sex
Aarray2list1(Sex a, Sex p)
{
    long            sx;
    long            j;
    Sex             q;

    sx = mkint(car(p));

    stackcheck();
    push(a);
    q = nil;
    for (j = sx - 1; j >= 0; j--) {
	q = cons(array_value(stack[0], j), q);
    }
    popv(1);
    return q;
}

static          Sex
Aarray2list2(Sex a, Sex p)
{
    long            sx,
                    sy;
    long            j,
                    k;
    Sex             q,
                    r,
                    l;

    sx = mkint(car(p));
    sy = mkint(cadr(p));

    stackcheck();
    push(a);
    q = nil;
    l = sx * sy - 1;
    for (j = 0; j < sx; j++) {
	push(q);
	r = nil;
	for (k = 0; k < sy; k++) {
	    r = cons(array_value(stack[-1], l), r);
	    l--;
	}
	pop(q);
	q = cons(r, q);
    }
    popv(1);
    return q;
}


static          Sex
Aarray2list3(Sex a, Sex p)
{
    long            sx,
                    sy,
                    sz;
    long            j,
                    k,
                    l,
                    m;
    Sex             q,
                    r,
                    s;

    sx = mkint(car(p));
    sy = mkint(cadr(p));
    sz = mkint(caddr(p));
    stackcheck();

    push(a);
    q = nil;
    m = sx * sy * sz - 1;
    for (j = 0; j < sx; j++) {
	push(q);
	r = nil;
	for (k = 0; j < sy; j++) {
	    push(r);
	    s = nil;
	    for (l = 0; j < sz; j++) {
		s = cons(array_value(stack[-2], m), s);
		m--;
	    }
	    pop(r);
	    r = cons(s, r);
	}
	pop(q);
	q = cons(r, q);
    }
    popv(1);
    return q;
}



static          Sex
Asetvector(Sex a, Sex v)
{
    long            sx;
    long            j;
    Sex            *q;

    if (!is_vect(a))
	return aerror2("setvector", a, v);

    sx = mkint(vect_dimx(a));
    q = (Sex *) & vect_value(a, 0);

    for (j = 0; j < sx; j++) {
	if (!is_cons(v))
	    break;
	q[j] = car(v);
	v = cdr(v);
    }
    return a;
}

static          Sex
Asetmatrix(Sex a, Sex v)
{
    long            i,
                    j;
    long            sx,
                    sy;
    Sex             p;
    Sex            *q;

    if (!is_matrix(a))
	return aerror2("setmatrix", a, v);
    sx = mkint(matrix_dimx(a));
    sy = mkint(matrix_dimy(a));
    q = (Sex *) & matrix_value(a, 0);
    for (i = 0; i < sx; i++) {
	if (!is_cons(v))
	    break;
	p = car(v);
	v = cdr(v);
	for (j = 0; j < sy; j++) {
	    if (!is_cons(p))
		break;
	    q[j] = car(p);
	    p = cdr(p);
	}
	q += sy;
    }

    return a;
}

static          Sex
Alist2vector(Sex a)
{
    long            n;
    long            j;
    Sex             p;
    Sex            *q;

    n = 0;
    p = a;
    while (is_cons(p)) {
	p = cdr(p);
	n++;
    }
    push(a);
    p = make_vector(n);
    pop(a);
    q = (Sex *) & vect_value(p, 0);

    for (j = 0; j < n; j++) {
	*q++ = car(a);
	a = cdr(a);
    }
    return p;
}


static          Sex
Avector2list(Sex a)
{
    long            sx;
    long            j;
    Sex             q;

    if (!is_vect(a))
	return aerror("vector2list");

    sx = mkint(vect_dimx(a));

    push(a);
    q = nil;
    for (j = 0; j < sx; j++) {
	q = cons(vect_value(stack[0], j), q);
    }
    popv(1);
    return q;
}



static          Sex
setarray_fn(Sex arg, Sex env)
{
    long            n;
    Sex             a1,
                    a2,
                    a3;
    Sex             b,
                    c;

    stackcheck();
    n = list_length(arg);
    push2(arg, env);
    push5(nil, nil, nil, nil, nil);
#define arg stack[-6]
#define env stack[-5]
#define v stack[-4]
#define x stack[-3]
#define n1 stack[-2]
#define n2 stack[-1]
#define n3 stack[0]
    if (n >= 3 && n <= 5) {
	v = eval(car(arg), env);
	arg = cdr(arg);
	x = eval(car(arg), env);
	arg = cdr(arg);
	switch (n) {
	case 3:
	    a1 = eval(car(arg), env);
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray1(b, c, a1);
	case 4:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    a2 = eval(car(arg), env);
	    a1 = n1;
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray2(b, c, a1, a2);
	case 5:
	    n1 = eval(car(arg), env);
	    arg = cdr(arg);
	    n2 = eval(car(arg), env);
	    arg = cdr(arg);
	    a3 = eval(car(arg), env);
	    a2 = n2;
	    a1 = n1;
	    b = v;
	    c = x;
	    popv(7);
	    return Lsetarray3(b, c, a1, a2, a3);
	}
    }
#undef arg
#undef env
#undef v
#undef x
#undef n1
#undef n2
#undef n3
    return nil;
}

static          Sex
Aputv(Sex v, Sex n, Sex x)
{
    long            n1,
                    hl;
    if (!is_fixnum(n) || (n1 = mkint(n)) < 0)
	return aerror1("putv offset not fixnum", n);
    if (is_simple_vector(v)) {
	hl = bnum_n(veclen(v));
	if (n1 >= hl) {
	    return aerror1("putv index range", n);
	}
	vector_value(v, n1) = x;
	return x;
    }
#if 0
    if (is_binhdr(v)) {
	hl = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    iarray_value(v, n1) = mkint(x);
	    return x;
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    larray_value(v, n1) = mkint(x);
	    return x;
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    sarray_value(v, n1) = (float) floatit(x);
	    return x;
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (n1 >= hl) {
		return aerror1("putv index range", n);
	    }
	    darray_value(v, n1) = floatit(x);
	    return x;
	default:
	    return aerror1("putv not vector ", v);
	}
    }
#endif
    return aerror1("putv not vector ", v);
}

static          Sex
Aaref(Sex v, Sex n)
{
    Sex             m;
    long            n1,
                    hl;
    while (is_cons(n)) {
	m = car(n);
	if (!is_simple_vector(v))
	    return aerror1("aref", v);
	else if (!is_fixnum(m))
	    return aerror1("aref offset not fixnum", m);
	hl = bnum_n(veclen(v));
	n1 = mkint(m);
	if (n1 < 0 || n1 >= hl)
	    return aerror1("aref index range", m);
	else
	    v = vector_value(v, n1);
	n = cdr(n);
    }
    return v;
}

static          Sex
Agetv(Sex v, Sex n)
{
    long            n1,
                    hl;
    long            i;
    double          f;

    if (!is_fixnum(n) || (n1 = mkint(n)) < 0)
	return aerror1("getv offset not fixnum", n);
    if (is_simple_vector(v)) {
	hl = bnum_n(veclen(v));
	if (n1 >= hl)
	    return aerror1("getv index range", n);
	else
	    return vector_value(v, n1);
    }
#if 0
    if (is_binhdr(v)) {
	hl = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    hl = hl / sizeof(int);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    i = iarray_value(v, n1);
	    return mkfxat(i);
	case T_LARRAY:
	    hl = hl / sizeof(long);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    i = larray_value(v, n1);
	    return mkfxat(i);
	case T_SARRAY:
	    hl = hl / sizeof(float);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    f = sarray_value(v, n1);
	    return make_float(f);
	case T_DARRAY:
	    hl = hl / sizeof(double);
	    if (n1 >= hl) {
		return aerror1("getv index range", n);
	    }
	    f = darray_value(v, n1);
	    return make_float(f);
	default:
	    return aerror1("getv not vector ", v);
	}
    }
#endif
    return aerror1("getv", v);
}


static          Sex
Aupbv(Sex v)
{
    Tag             h;
    long            n;
    if ((is_simple_vector(v))) {
	h = vechdr(v);
	n = bnum_n(veclen(v)) - 1;
	return mkfxat(n);
    }
#if 0
    if (is_binhdr(v)) {
	n = binlen(v) - HDR;
	switch (type_of_header(binhdr(v))) {
	case T_IARRAY:
	    n = n / sizeof(int);
	    return mkfxat(n);
	case T_LARRAY:
	    n = n / sizeof(long);
	    return mkfxat(n);
	case T_SARRAY:
	    n = n / sizeof(float);
	    return mkfxat(n);
	case T_DARRAY:
	    n = n / sizeof(double);
	    return mkfxat(n);
	default:
	    return nil;
	}
    }
#endif
    return nil;
}


/************************************************************************** 
 *
 * equal 
 * 
 *************************************************************************/


bool
AAeql(Sex a, Sex b)
{
    Tag             h;
    long            hh;

  again:
    if (a == b)
	return YES;
    if ((((long) a ^ (long) b) & TAG_BITS) != 0)
	return NO;
    switch ((long) a & TAG_BITS) {
    case TAG_NUMBERS:
	if ((h = numhdr(a)) != numhdr(b) || (hh = numlen(a)) != numlen(b))
	    return NO;
	switch (numhdr_of(h)) {
	case N_FLOAT:
	case N_BIGNUM:
	case N_EFLOAT:
	case N_INTERVAL:
	case N_ECOMPLEX:
	    hh = hh - TAG_NUMBERS;
	    while (hh > (2 * SEX - TAG_NUMBERS)) {
		hh -= SEX;
		if (*(Sex *) ((char *) a + hh) !=
		    *(Sex *) ((char *) b + hh))
		    return NO;
	    }
	    return YES;
	case N_BFLOAT:
	    if (numvec(a, 0) != numvec(b, 0))
		return NO;
	    a = numvec(a, 1);
	    b = numvec(b, 1);
	    goto again;
	case N_BEFLOAT:
	    if (numvec(a, 1) != numvec(b, 1))
		return NO;
	    a = numvec(a, 0);
	    b = numvec(b, 0);
	    goto again;
	case N_BECOMPLEX:
	    if (eql(numvec(a, 2), numvec(b, 2)) == NO)
		return NO;
	case N_RATIONAL:
	case N_COMPLEX:
	case N_BINTERVAL:
	    if (eql(numvec(a, 1), numvec(b, 1)) == NO)
		return NO;
	case N_MODULAR:
	    a = numvec(a, 0);
	    b = numvec(b, 0);
	    goto again;
	}
	return NO;
    default:
	return NO;
    }
}

static          Sex
numequal(Sex a, Sex b)
{
    Tag             ha;
    long            l;

  again:
    if (a == b)
	return true;
    if (is_fixnum(a)) {
	if (!is_fixnum(b))
	    return nil;
	return Bool_to_Sex(a == b);
    } else if (is_fixnum(b))
	return nil;


    if ((ha = numhdr(a)) != numhdr(b))
	return nil;
    switch (numhdr_of(ha)) {
    case N_BFLOAT:
	if (numvec(a, 0) != numvec(b, 0))
	    return nil;
	a = numvec(a, 1);
	b = numvec(b, 1);
	goto again;
    case N_BEFLOAT:
	if (numvec(a, 1) != numvec(b, 1))
	    return nil;
    case N_MODULAR:
	a = numvec(a, 0);
	b = numvec(b, 0);
	goto again;
    case N_BECOMPLEX:
	if (numequal(numvec(a, 2), numvec(b, 2)) == nil)
	    return nil;
    case N_RATIONAL:
    case N_COMPLEX:
    case N_BINTERVAL:
	if (numequal(numvec(a, 1), numvec(b, 1)) == nil)
	    return nil;
	a = numvec(a, 0);
	b = numvec(b, 0);
	goto again;
    default:
	if ((l = numlen(a)) != numlen(b))
	    return nil;
	l = bnum_n(l);
	while ((l--) != 0)
	    if (numvec(a, l) != numvec(b, l))
		return nil;
	return true;

    }
}

static          Sex
strequal(Sex a, Sex b)
{
    Tag             ha = binhdr(a),
	hb = binhdr(b);
    long            l;

    if (ha != hb || (l = binlen(a)) != binlen(b))
	return nil;
    l = bnum_n(l + (SEX - 1));
    while ((l--) != 0)
	if (object_value(a, l) != object_value(b, l))
	    return nil;
    return true;
}

static          Sex
vecequal(Sex a, Sex b)
{
    Tag             ha = vechdr(a),
	hb = vechdr(b);
    long            l;

    if (ha != hb || (l = veclen(a)) != veclen(b))
	return nil;
    l = bnum_n(l + (SEX - 1));
    while ((l--) != 0) {
	Sex             ea = vector_value(a, l);
	Sex             eb = vector_value(b, l);
	if (ea == eb)
	    continue;
	if (equal(ea, eb) == nil)
	    return nil;
    }
    return true;
}



static          Sex
AAequal(Sex a, Sex b)
{
    long            ta;

    while (YES) {
	if ((((long) a ^ (long) b) & TAG_BITS) != 0)
	    return nil;
	switch (ta = (long) a & TAG_BITS) {
	case TAG_CONS:
	    if (!is_cons(b))
		return nil;
	    else {
		Sex             ca = car(a),
		    cb = car(b);
		if (ca == cb) {
		    a = cdr(a);
		    b = cdr(b);
		    if (a == b)
			return true;
		    continue;
		}
		if (equal(ca, cb) != true)
		    return nil;
		a = cdr(a);
		b = cdr(b);
		if (a == b)
		    return true;
		continue;
	    }
	case TAG_FIXNUM:
	case TAG_HDRS:
	case TAG_SYMBOL:
	    return nil;
	case TAG_NUMBERS:
	    return numequal(a, b);
	case TAG_VECTOR:
	    return vecequal(a, b);
	case TAG_BINARY:
	    return strequal(a, b);
	}
    }
}


static          Sex
Aeql(Sex a, Sex b)
{
    if (a == b)
	return true;
    return Bool_to_Sex(eql(a, b));
}

static          Sex
Aeqcar(Sex a, Sex b)
{
    if (!is_cons(a) || car(a) != b)
	return nil;
    return true;
}

static          Sex
Aequal(Sex a, Sex b)
{
    if (a == b)
	return true;
    else
	return AAequal(a, b);
}

static          Sex
Aneq(Sex a, Sex b)
{
    if (a == b)
	return nil;
    if (equal(a, b) == nil)
	return true;
    return nil;
}

static          Sex
Anull(Sex a)
{
    return Bool_to_Sex(a == nil);
}

static          Sex
Aendp(Sex a)
{
    if (a == nil)
	return true;
    else if (is_cons(a))
	return nil;
    else
	return error1("endp: bad argument", a);
}


static          Sex
Areverse(Sex a)
{
    Sex             r;
    stackcheck();
    r = nil;
    push(nil);
    while (is_cons(a)) {
	stack[0] = cdr(a);
	r = cons(car(a), r);
	a = stack[0];
    }
    popv(1);
    return r;
}


static          Sex
Aassoc(Sex a, Sex b)
{
    if (is_symbol(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    Sex             c = car(b);
	    if (is_cons(c) && a == car(c))
		return c;
	    b = cdr(b);
	}
	return nil;
    }
    while (is_cons(b)) {
	Sex             c = car(b);
	if (is_cons(c)) {
	    Sex             cc = car(c);
	    if (a == cc || equal(a, cc) != nil)
		return c;
	}
	b = cdr(b);
    }
    return nil;
}

static          Sex
Arassoc(Sex a, Sex b)
{
    while (is_cons(b)) {
	Sex             c = car(b);
	if (is_cons(c)) {
	    Sex             cc = cdr(c);
	    if (a == cc || equal(a, cc) != nil)
		return c;
	}
	b = cdr(b);
    }
    return nil;
}

static          Sex
Aatsoc(Sex a, Sex b)
{
    Sex             c;

    while (is_cons(b)) {
	c = car(b);
	if (is_cons(c) && a == car(c))
	    return c;
	b = cdr(b);
    }
    return nil;
}

static          Sex
Amember(Sex a, Sex b)
{
    if (is_symbol(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    if (a == car(b))
		return b;
	    b = cdr(b);
	}
	return nil;
    }
    while (is_cons(b)) {
	Sex             cb = car(b);
	if (a == cb || equal(a, cb) != nil)
	    return b;
	b = cdr(b);
    }
    return nil;
}


static          Sex
Amemq(Sex a, Sex b)
{
    while (is_cons(b)) {
	if (a == car(b))
	    return b;
	b = cdr(b);
    }
    return nil;
}

static          bool
smemq(Sex a, Sex b)
{
    Sex             w;
    while (is_cons(b)) {
	w = car(b);
	if (w == quote_symbol)
	    return NO;
	else if (smemq(a, w))
	    return YES;
	else
	    b = cdr(b);
    }
    return (a == b);
}

static          Sex
Asmemq(Sex a, Sex b)
{
    return Bool_to_Sex(smemq(a, b));
}


long
AAlist_length(Sex a)
{
    long            n;
    n = 0;
    while (is_cons(a)) {
	a = cdr(a);
	n++;
    }
    return n;
}


static          Sex
Alength(Sex a)
{
    long            n = 0;

    switch (tag_type(a)) {
    case TAG_CONS:
	while (is_cons(a)) {
	    a = cdr(a);
	    n++;
	}
	return mkfxat(n);
    case TAG_SYMBOL:
	if (a == nil)
	    return mkfxat(0);
	break;
    case TAG_FIXNUM:
    case TAG_NUMBERS:
	break;
    case TAG_VECTOR:
	switch (type_of_header(vechdr(a))) {
	case T_VECT:
	    return vect_dimx(a);
	case T_MATRIX:
	    n = mkint(matrix_dimx(a));
	    n = n * mkint(matrix_dimy(a));
	    return mkfxat(n);
	case T_SIMPLE:
	    n = bnum_n(veclen(a)) - 1;
	    return mkfxat(n);
	case T_ARRAY:
	    a = array_dim(a);
	    n = 1;
	    while (is_cons(a)) {
		n = n * mkint(car(a));
		a = cdr(a);
	    }
	    return mkfxat(n);
	}
	break;
    case TAG_BINARY:
	n = binlen(a) - HDR;
	switch (type_of_header(binhdr(a))) {
	case T_IARRAY:
	case T_IARRAY2:
	case T_IARRAY3:
	    n = n / sizeof(int);
	    return mkfxat(n);

	case T_LARRAY:
	case T_LARRAY2:
	case T_LARRAY3:
	    n = n / sizeof(long);
	    return mkfxat(n);
	case T_SARRAY:
	case T_SARRAY2:
	case T_SARRAY3:
	    n = n / sizeof(float);
	    return mkfxat(n);
	case T_DARRAY:
	case T_DARRAY2:
	case T_DARRAY3:
	    n = n / sizeof(double);
	    return mkfxat(n);
	}
    }
    error("length: not a list");
    return nil;
}


static          Sex
Aappend(Sex a, Sex b)
{
    Sex             r = nil;

    stackcheck();
    push(b);
    push(nil);
    while (is_cons(a)) {
	stack[0] = cdr(a);
	r = cons(car(a), r);
	a = stack[0];
    }
    popv(1);
    pop(b);
    while (r != nil) {
	a = cdr(r);
	cdr(r) = b;
	b = r;
	r = a;
    }
    return b;
}

static          Sex
Aappende(Sex a, Sex b)
{
    Sex             r = nil;

    stackcheck();
    push(a);
    b = Lncons(b);
    push(b);
    while (is_cons(a)) {
	stack[-1] = cdr(a);
	r = cons(car(a), r);
	a = stack[-1];
    }
    pop(b);
    popv(1);
    while (r != nil) {
	a = cdr(r);
	cdr(r) = b;
	b = r;
	r = a;
    }
    return b;
}



static          Sex
Adelete(Sex a, Sex b)
{
    Sex             r;
    Sex             q;
    stackcheck();
    push2(a, b);
    r = nil;
    if (is_symbol(a) || is_fixnum(a)) {
	while (is_cons(b)) {
	    q = car(b);
	    if (q == stack[-1]) {
		b = cdr(b);
		break;
	    }
	    stack[0] = cdr(b);
	    r = cons(car(b), r);
	    b = stack[0];
	}
    } else {
	while (is_cons(b)) {
	    q = car(b);
	    if (q == a || equal(q, a) != nil) {
		b = cdr(b);
		break;
	    }
	    stack[0] = cdr(b);
	    r = cons(car(b), r);
	    b = stack[0];
	    a = stack[-1];
	}
    }
    popv(2);
    while (r != nil) {
	q = cdr(r);
	cdr(r) = b;
	b = r;
	r = q;
    }
    return b;
}


static          Sex
Adeleq(Sex a, Sex b)
{
    Sex             r;
    Sex             q;
    stackcheck();
    push2(a, b);
    r = nil;
    while (is_cons(b)) {
	q = car(b);
	if (q == stack[-1]) {
	    b = cdr(b);
	    break;
	}
	stack[0] = cdr(b);
	r = cons(car(b), r);
	b = stack[0];
    }
    popv(2);
    while (r != nil) {
	q = cdr(r);
	cdr(r) = b;
	b = r;
	r = q;
    }
    return b;
}

static          Sex
Aaconc(Sex a, Sex b)
{
    Sex             c;
    Sex             next;

    stackcheck();
    push(a);
    b = cons(b, nil);
    pop(a);
    if (!is_cons(a))
	return b;
    c = a;
    while (YES) {
	next = cdr(c);
	if (!is_cons(next)) {
	    cdr(c) = b;
	    return a;
	} else {
	    c = next;
	}
    }
}


static          Sex
Anconc(Sex a, Sex b)
{
    Sex             c;
    Sex             next;
    if (a == nil)
	return b;
    if (!is_cons(a))
	error("first arg. of nconc is not a list");
    c = a;
    while (YES) {
	next = cdr(c);
	if (next == nil) {
	    cdr(c) = b;
	    return a;
	} else if (!is_cons(next)) {
	    error("first arg. of nconc is not a list");
	} else {
	    c = next;
	}
    }
}


static          Sex
Asubstq(Sex a, Sex b, Sex c)
{
    Sex             w;
    if (c == b)
	return a;
    if (!is_cons(c)) {
	return c;
    }
    stackcheck();
    push3(a, b, c);
    w = Asubstq(a, b, car(c));
    pop2(c, b);
    a = stack[0];
    stack[0] = w;
    w = Asubstq(a, b, cdr(c));
    pop(a);
    return cons(a, w);
}

static          Sex
Asubst(Sex a, Sex b, Sex c)
{
    Sex             w;
    if (c == b)
	return a;
    if (is_symbol(b) || is_fixnum(b))
	return Asubstq(a, b, c);
    stackcheck();
    push3(a, b, c);
    if (c == b || equal(c, b) != nil) {
	popv(2);
	pop(a);
	return a;
    }
    if (!is_cons(stack[0])) {
	pop(c);
	popv(2);
	return c;
    }
    w = Lsubst(stack[-2], stack[-1], car(stack[0]));
    pop2(c, b);
    a = stack[0];
    stack[0] = w;
    w = Lsubst(a, b, cdr(c));
    pop(a);
    return cons(a, w);
}


static          Sex
Asublis(Sex al, Sex x)
{
    Sex             a1,
                    a2;
    stackcheck();
    push5(al, x, al, nil, nil);
#define carx stack[0]
#define cdrx stack[-1]
#define w    stack[-2]
#define x    stack[-3]
#define al   stack[-4]
    while (YES) {
	if (!is_cons(w)) {
	    if (!is_cons(x)) {
		a1 = x;
		popv(5);
		return a1;
	    }
	    carx = Asublis(al, car(x));
	    cdrx = Asublis(al, cdr(x));
	    if (carx == car(x) && cdrx == cdr(x)) {
		a1 = x;
		popv(5);
		return a1;
	    } else {
		a1 = cons(carx, cdrx);
		popv(5);
		return a1;
	    }
	}
	a1 = car(w);
	if (is_cons(a1)) {
	    a2 = car(a1);
	    if (a2 == x || equal(a2, x) != nil) {
		a1 = cdr(a1);
		popv(5);
		return a1;
	    }
	}
	w = cdr(w);
    }
}
#undef carx
#undef cdrx
#undef w
#undef x
#undef al

static          Sex
Asubla(Sex al, Sex x)
{
    Sex             a1;

    stackcheck();
    push5(al, x, al, nil, nil);
#define carx stack[0]
#define cdrx stack[-1]
#define w    stack[-2]
#define x    stack[-3]
#define al   stack[-4]
    while (YES) {
	if (!is_cons(w)) {
	    if (!is_cons(x)) {
		a1 = x;
		popv(5);
		return a1;
	    }
	    carx = Lsubla(al, car(x));
	    cdrx = Lsubla(al, cdr(x));
	    if (carx == car(x) && cdrx == cdr(x)) {
		a1 = x;
		popv(5);
		return a1;
	    } else {
		a1 = cons(carx, cdrx);
		popv(5);
		return a1;
	    }
	}
	a1 = car(w);
	if (is_cons(a1)) {
	    if (car(a1) == x) {
		a1 = cdr(a1);
		popv(5);
		return a1;
	    }
	}
	w = cdr(w);
    }
}
#undef carx
#undef cdrx
#undef w
#undef x
#undef al

static          Sex
Aquit()
{
    exit(0);
    return nil;
}

static          Sex
Abye()
{
    exit(0);
    return nil;
}


#ifdef INCLUDE_FEND

/*****************************************************************************/
/*
 * Standard lisp functions.
 */
/*****************************************************************************/



/*
 * Code for deflist 
 */

static          Sex
Adeflist(Sex u, Sex ind)
{
    Sex             x,
                    y;

    /*
     * if (compile_flag) return nil; 
     */
    if (u == nil)
	return nil;
    stackcheck();
    push3(u, ind, nil);
#define u stack[-2]
#define ind stack[-1]
#define v stack[0]
    v = car(u);
    Lput(car(v), ind, car(cdr(v)));
    x = Ldeflist(cdr(u), ind);
    y = car(v);
    popv(3);
    return cons(y, x);
#undef v
#undef u
#undef ind
}


static          Sex
Aexpand(Sex l, Sex fn)
{
    if (l == nil)
	return nil;
    else if (cdr(l) == nil)
	return car(l);
    else {
	Sex             r;
	stackcheck();
	push2(fn, l);
	r = Lexpand(cdr(l), fn);
	pop2(l, fn);
	return Llist3(fn, car(l), r);
    }
}


static          Sex
logxor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logxor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
logand_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logand2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
logor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = logor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogxor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogxor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogand_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogand2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
ilogor_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lilogor2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}


static          Sex
max_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lmax2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
min_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Lmin2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}


static          Sex
imax_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (max) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Limax2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}

static          Sex
imin_fn(Sex args, Sex env)
{
    Sex             r;
    if (!is_cons(args))
	return mkfxat(0);	/* 
				 * * (min) => 0 ?  
				 */
    stackcheck();
    push2(args, env);
    r = car(args);
    r = eval(r, env);
    pop2(env, args);
    args = cdr(args);
    while (is_cons(args)) {
	Sex             w;
	push3(env, args, r);
	w = car(args);
	w = eval(w, env);
	pop(r);
	r = Limin2(r, w);
	pop2(args, env);
	args = cdr(args);
    }
    return r;
}




/*
 * Code for copy 
 */

static          Sex
Acopy(Sex u)
{
    Sex             v;

    if (!is_cons(u))
	return u;
    stackcheck();
    push2(nil, nil);
    stack[-1] = cdr(u);
    stack[0] = Lcopy(car(u));
    u = Lcopy(stack[-1]);
    v = stack[0];
    popv(2);
    return cons(v, u);
}

/*
 * toplevel list or array copy 
 */

static          Sex
Aclone(Sex u)
{
    Sex             p;

    if (!is_cons(u))
	return Aarray_copy(u);
    stackcheck();
    push3(nil, nil, u);
#define p  stack[-2]
#define pp stack[-1]
#define u  stack[ 0]
    p = cons(nil, nil);
    pp = p;
    while (is_cons(u)) {
	p = Arcons(p, car(u));
	u = cdr(u);
    }
#undef p
    p = cdr(pp);
#undef pp
#undef u
    popv(3);
    return p;
}

static          Sex
Aarray_copy(Sex u)
{
    long            len;
    long            i;
    Sex             obj;

    switch ((long) (u) & TAG_BITS) {
    case TAG_VECTOR:
	len = veclen(u);
	push(u);
	obj = Agetvector(TAG_VECTOR, T_SIMPLE, len);
	pop(u);
	vechdr(obj) = vechdr(u);
	len = (len - HDR) / SEX;
	for (i = 0; i < len; i++)
	    vector_value(obj, i) = vector_value(u, i);
	return obj;
    case TAG_BINARY:
	len = binlen(u);
	push(u);
	obj = Agetvector(TAG_BINARY, T_IARRAY, len);
	pop(u);
	binhdr(obj) = binhdr(u);
	len = (len - HDR) / SEX;
	for (i = 0; i < len; i++)
	    iarray_value(obj, i) = iarray_value(u, i);
	return obj;
    default:
	return error1("array_copy", u);
    }
}



/*
 * Code for xn 
 */

static          Sex
Axn(Sex v, Sex u)
{
    Sex             w;
    stackcheck();
    push(v);
#define v stack[0]
    while (v != nil) {
	if (Lmember(car(v), u) == nil) {
	    v = cdr(v);
	    continue;
	}
	w = Ldelete(car(v), u);
	w = Lxn(cdr(v), w);
	u = car(v);
	popv(1);
	return cons(u, w);
    }
    popv(1);
    return nil;
#undef v
}





static          Sex
Alast(Sex u)
{
    while (is_cons(u)) {
	if (cdr(u) == nil)
	    return car(u);
	u = cdr(u);
    }
    return u;
}





static          Sex
Anum(Sex n)
{
    if (is_numhdr(n) && type_of_header(numhdr(n)) == T_RATIONAL)
	return rational_num(n);
    else if (is_rpoly(n))
	return rpoly_num(n);
    else
	return error("not a rational number");
}

static          Sex
Aden(Sex n)
{
    if (is_numhdr(n) && type_of_header(numhdr(n)) == T_RATIONAL)
	return rational_den(n);
    else if (is_rpoly(n))
	return rpoly_den(n);
    else
	return error("not a rational number");
}





static          Sex
Aprintprompt(Sex u)
{
    return Lsetpchar(u);
}

#endif

static          Sex
Aprintaddr(Sex x)
{
    return mkfxat(x);
}

#ifdef ASUKA
static          Sex
Amode_lisp()
{
    current_package = value(current_sugar_package);
    current_charvec = package_charvec(current_package);
    return true;
}
static          Sex
Amode_asuka()
{
    current_package = value(current_asuka_package);
    current_charvec = package_charvec(current_package);
    return true;
}
#endif



setup_type const list_setup[] = {
    {"acons", (lisp_proc *) Aacons, 2},
    {"atom", (lisp_proc *) Aatom, 1},
    {"boundp", (lisp_proc *) Aboundp, 1},
    {"car", (lisp_proc *) Acar, 1},
    {"cdr", (lisp_proc *) Acdr, 1},
    {"caar", (lisp_proc *) Acaar, 1},
    {"cadr", (lisp_proc *) Acadr, 1},
    {"cdar", (lisp_proc *) Acdar, 1},
    {"cddr", (lisp_proc *) Acddr, 1},
    {"caaar", (lisp_proc *) Acaaar, 1},
    {"caadr", (lisp_proc *) Acaadr, 1},
    {"cadar", (lisp_proc *) Acadar, 1},
    {"caddr", (lisp_proc *) Acaddr, 1},
    {"cdaar", (lisp_proc *) Acdaar, 1},
    {"cdadr", (lisp_proc *) Acdadr, 1},
    {"cddar", (lisp_proc *) Acddar, 1},
    {"cdddr", (lisp_proc *) Acdddr, 1},
    {"caaaar", (lisp_proc *) Acaaaar, 1},
    {"caaadr", (lisp_proc *) Acaaadr, 1},
    {"caadar", (lisp_proc *) Acaadar, 1},
    {"caaddr", (lisp_proc *) Acaaddr, 1},
    {"cadaar", (lisp_proc *) Acadaar, 1},
    {"cadadr", (lisp_proc *) Acadadr, 1},
    {"caddar", (lisp_proc *) Acaddar, 1},
    {"cadddr", (lisp_proc *) Acadddr, 1},
    {"cdaaar", (lisp_proc *) Acdaaar, 1},
    {"cdaadr", (lisp_proc *) Acdaadr, 1},
    {"cdadar", (lisp_proc *) Acdadar, 1},
    {"cdaddr", (lisp_proc *) Acdaddr, 1},
    {"cddaar", (lisp_proc *) Acddaar, 1},
    {"cddadr", (lisp_proc *) Acddadr, 1},
    {"cdddar", (lisp_proc *) Acdddar, 1},
    {"cddddr", (lisp_proc *) Acddddr, 1},
    {"first", (lisp_proc *) Acar, 1},
    {"second", (lisp_proc *) Acadr, 1},
    {"third", (lisp_proc *) Acaddr, 1},
    {"rest", (lisp_proc *) Acdr, 1},

    {"codep", (lisp_proc *) Acodep, 1},
    {"constantp", (lisp_proc *) Aconstantp, 1},
    {"date", (lisp_proc *) Adate, 0},
    {"error", (lisp_proc *) Aerror, 2},
    {"error1", (lisp_proc *) Aerror1, 0},
    {"error2", (lisp_proc *) Aerror2, 1},
    {"fixp", (lisp_proc *) Afixp, 1},
    {"bigp", (lisp_proc *) Abigp, 1},
    {"modularp", (lisp_proc *) Ais_modular, 1},
    {"logicalp", (lisp_proc *) Alogicalp, 1},
    {"floatp", (lisp_proc *) Afloatp, 1},
    {"bfloatp", (lisp_proc *) Abfloatp, 1},
    {"rationalp", (lisp_proc *) Arationalp, 1},
    {"complexp", (lisp_proc *) Acomplexp, 1},
    {"complp", (lisp_proc *) Acomplexp, 1},
    {"ecomplexp", (lisp_proc *) Aecomplexp, 1},
    {"becomplexp", (lisp_proc *) Abecomplexp, 1},
    {"efloatp", (lisp_proc *) Aefloatp, 1},
    {"befloatp", (lisp_proc *) Abefloatp, 1},
    {"intervalp", (lisp_proc *) Aintervalp, 1},
    {"bintervalp", (lisp_proc *) Abintervalp, 1},
    {"fluidp", (lisp_proc *) Afluidp, 1},
    {"gctime", (lisp_proc *) Agctime, 0},
    {"gccalled", (lisp_proc *) Agccalled, 0},
    {"get", (lisp_proc *) Aget, 2},
    {"globalp", (lisp_proc *) Aglobalp, 1},
    {"list2", (lisp_proc *) Alist2, 2},
    {"list3", (lisp_proc *) Alist3, 3},
    {"list4", (lisp_proc *) Alist4, 4},
    {"list5", (lisp_proc *) Alist5, 5},
    {"global", (lisp_proc *) Aglobal, 1},
    {"fluid", (lisp_proc *) Afluid, 1},
    {"mkquote", (lisp_proc *) Amkquote, 1},
    {"cons", (lisp_proc *) cons, 2},
    {"xcons", (lisp_proc *) Axcons, 2},
    {"ncons", (lisp_proc *) Ancons, 1},
    {"numberp", (lisp_proc *) Anumberp, 1},
    {"pair", (lisp_proc *) Apair, 2},
    {"put", (lisp_proc *) Aput, 3},
    {"remprop", (lisp_proc *) Aremprop, 2},
    {"rplaca", (lisp_proc *) Arplaca, 2},
    {"rplacd", (lisp_proc *) Arplacd, 2},
    {"stop", (lisp_proc *) Astop, 1},
    {"time", (lisp_proc *) Atime, 0},
    {"float-time", (lisp_proc *) Afloat_time, 0},
    {"union", (lisp_proc *) Aunion, 2},
    {"merge", (lisp_proc *) Amerge, 2},
    {"sort", (lisp_proc *) Asort, 1},
    {"unionq", (lisp_proc *) Aunionq, 2},
    {"intersection", (lisp_proc *) Aintersection, 2},
    {"intersectionq", (lisp_proc *) Aintersectionq, 2},
    {"setdiff", (lisp_proc *) Asetdiff, 2},
    {"setdiffq", (lisp_proc *) Asetdiffq, 2},
    {"unglobal", (lisp_proc *) Aunglobal, 1},
    {"unfluid", (lisp_proc *) Aunfluid, 1},
    {"flag", (lisp_proc *) Aflag, 2},
    {"flagp", (lisp_proc *) Aflagp, 2},
    {"flagpcar", (lisp_proc *) Aflagpcar, 2},
    {"flagp**", (lisp_proc *) Aflagp, 2},
    {"get*", (lisp_proc *) Aget, 2},
    {"idp", (lisp_proc *) Aidp, 1},
    {"pairp", (lisp_proc *) Apairp, 1},
    {"listp", (lisp_proc *) Alistp, 1},
    {"is_list", (lisp_proc *) Alistp, 1},
    {"remflag", (lisp_proc *) Aremflag, 2},
    {"set", (lisp_proc *) Aset, 2},
    {"stringp", (lisp_proc *) Astringp, 1},
    {"vectorp", (lisp_proc *) Avectorp, 1},
    {"append", (lisp_proc *) Aappend, 2},
    {"appende", (lisp_proc *) Aappende, 2},
    {"eq", (lisp_proc *) Aeq, 2},
    {"eqcar", (lisp_proc *) Aeqcar, 2},
    {"eql", (lisp_proc *) Aeql, 2},
    {"equal", (lisp_proc *) Aequal, 2},
    {"endp", (lisp_proc *) Aendp, 1},
    {"getd", (lisp_proc *) Agetd, 1},
    {"movd", (lisp_proc *) Amovd, 2},
    {"length", (lisp_proc *) Alength, 1},
    {"putd", (lisp_proc *) Aputd, 3},
    {"remd", (lisp_proc *) Aremd, 1},
    {"trace", (lisp_proc *) Atrace, 1},
    {"untrace", (lisp_proc *) Auntrace, 1},
    {"symbol-name", (lisp_proc *) Asymbol_name, 1},
    {"symbol-length", (lisp_proc *) Asymbol_length, 1},
    {"plist", (lisp_proc *) Aplist, 1},
    {"putv", (lisp_proc *) Aputv, 3},
    {"upbv", (lisp_proc *) Aupbv, 1},
    {"memq", (lisp_proc *) Amemq, 2},
    {"nconc", (lisp_proc *) Anconc, 2},
    {"aconc", (lisp_proc *) Aaconc, 2},
    {"map", (lisp_proc *) Lmap, 2},
    {"mapc", (lisp_proc *) Lmapc, 2},
    {"mapcar", (lisp_proc *) Lmapcar, 2},
    {"maplist", (lisp_proc *) Lmaplist, 2},
    {"mapcan", (lisp_proc *) Lmapcan, 2},
    {"mapcon", (lisp_proc *) Lmapcon, 2},
    {"mapv", (lisp_proc *) Lmapv, 2},
    {"map2", (lisp_proc *) Lmap2, 3},
    {"mapc2", (lisp_proc *) Lmapc2, 3},
    {"mapcar2", (lisp_proc *) Lmapcar2, 3},
    {"maplist2", (lisp_proc *) Lmaplist2, 3},
    {"mapcan2", (lisp_proc *) Lmapcan2, 3},
    {"mapcon2", (lisp_proc *) Lmapcon2, 3},
    {"mapv2", (lisp_proc *) Lmapv2, 3},
    {"neq", (lisp_proc *) Aneq, 2},
    {"not", (lisp_proc *) Anull, 1},
    {"null", (lisp_proc *) Anull, 1},
    {"reverse", (lisp_proc *) Areverse, 1},
    {"smemq", (lisp_proc *) Asmemq, 2},
    {"subla", (lisp_proc *) Asubla, 2},
    {"sublis", (lisp_proc *) Asublis, 2},
    {"delete", (lisp_proc *) Adelete, 2},
    {"deleq", (lisp_proc *) Adeleq, 2},
    {"rassoc", (lisp_proc *) Arassoc, 2},
    {"assoc", (lisp_proc *) Aassoc, 2},
    {"atsoc", (lisp_proc *) Aatsoc, 2},
    {"assq", (lisp_proc *) Aatsoc, 2},
    {"getv", (lisp_proc *) Agetv, 2},
    {"aref", (lisp_proc *) Aaref, 2},
    {"member", (lisp_proc *) Amember, 2},
    {"mkvect", (lisp_proc *) Amkvect, 1},
    {"mkarray", (lisp_proc *) mkarray_fn, -1},
    {"mkarray1", (lisp_proc *) Amkarray1, 1},
    {"mkarray2", (lisp_proc *) Amkarray2, 2},
    {"mkarray3", (lisp_proc *) Amkarray3, 3},
    {"mkiarray", (lisp_proc *) Amkiarray1, 1},
    {"mkivect", (lisp_proc *) Amkiarray1, 1},
    {"mkiarray2", (lisp_proc *) Amkiarray2, 2},
    {"mkiarray3", (lisp_proc *) Amkiarray3, 3},
    {"mkfvect", (lisp_proc *) Amkfarray1, 1},
    {"mkfarray2", (lisp_proc *) Amkfarray2, 2},
    {"mkfarray3", (lisp_proc *) Amkfarray3, 3},
    {"mkdvect", (lisp_proc *) Amkdarray1, 1},
    {"mkdarray2", (lisp_proc *) Amkdarray2, 2},
    {"mkdarray3", (lisp_proc *) Amkdarray3, 3},
    {"getarray", (lisp_proc *) getarray_fn, -1},
    {"getarray1", (lisp_proc *) Agetarray1, 2},
    {"getarray2", (lisp_proc *) Agetarray2, 3},
    {"getarray3", (lisp_proc *) Agetarray3, 4},
    {"set_array", (lisp_proc *) setarray_fn, -1},
    {"setarray1", (lisp_proc *) Asetarray1, 3},
    {"setarray2", (lisp_proc *) Asetarray2, 4},
    {"setarray3", (lisp_proc *) Asetarray3, 5},
    {"list2array", (lisp_proc *) Alist2array, 2},
    {"array2list", (lisp_proc *) Aarray2list, 1},
    {"list2vector", (lisp_proc *) Alist2vector, 1},
    {"vector2list", (lisp_proc *) Avector2list, 1},
    {"setarray", (lisp_proc *) Alist2array, 2},
    {"setmatrix", (lisp_proc *) Asetmatrix, 2},
    {"setvector", (lisp_proc *) Asetvector, 2},
    {"reversip", (lisp_proc *) Lnreverse, 1},
    {"subst", (lisp_proc *) Asubst, 3},
    {"quit", (lisp_proc *) Aquit, 0},
    {"bye", (lisp_proc *) Abye, 0},
#ifdef INCLUDE_FEND
    {"logxor", (lisp_proc *) logxor_fn, -1},
    {"logand", (lisp_proc *) logand_fn, -1},
    {"logor", (lisp_proc *) logor_fn, -1},
    {"ilogxor", (lisp_proc *) ilogxor_fn, -1},
    {"ilogand", (lisp_proc *) ilogand_fn, -1},
    {"ilogor", (lisp_proc *) ilogor_fn, -1},
    {"max", (lisp_proc *) max_fn, -1},
    {"min", (lisp_proc *) min_fn, -1},
    {"imax", (lisp_proc *) imax_fn, -1},
    {"imin", (lisp_proc *) imin_fn, -1},
    {"deflist", (lisp_proc *) Adeflist, 2},
    {"expand", (lisp_proc *) Aexpand, 2},
    {"copy", (lisp_proc *) Acopy, 1},
    {"clone", (lisp_proc *) Aclone, 1},
    {"xn", (lisp_proc *) Axn, 2},
    {"last", (lisp_proc *) Alast, 1},
    {"num", (lisp_proc *) Anum, 1},
    {"den", (lisp_proc *) Aden, 1},
    {"printprompt", (lisp_proc *) Aprintprompt, 1},
#endif
    {"printaddr", (lisp_proc *) Aprintaddr, 1},
    {"traceall", (lisp_proc *) Atraceall, 1},
#ifdef MS_DOS
    {"int10h", (lisp_proc *) Aint10h, 4},
    {"int10hr", (lisp_proc *) Aint10hr, 4},
#endif
    {"atom-tag", (lisp_proc *) Aatom_tag, 1},
    {"number-tag", (lisp_proc *) Anumber_tag, 1},
    {"number-length", (lisp_proc *) Anumber_length, 1},
    {"nth", (lisp_proc *) Anth, 2},
    {"pnth", (lisp_proc *) Apnth, 2},
    {"nthcdr", (lisp_proc *) Anthcdr, 2},
    {"replace", (lisp_proc *) Areplace, 3},
    {"package_cnt", (lisp_proc *) Apackage_cnt, 2},
    {"hashp", (lisp_proc *) Ahashp, 1},
    {"packagep", (lisp_proc *) Apackagep, 1},
#ifdef ASUKA
    {"lisp", (lisp_proc *) Amode_lisp, 0},
    {"asuka", (lisp_proc *) Amode_asuka, 0},
#endif
    {NULL, 0, 0}
};
