/*
 * logical.c            (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * Boolean operations
 * $Header: /Users/kako/asuka/src/RCS/logical.c,v 2.2 2018/01/05 02:35:21 kako Exp $
 */

static char const rcsid[] =
    "$Id: logical.c,v 2.2 2018/01/05 02:35:21 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"

Sex             logerror1(Sex a);
Sex             logerror2(Sex a, Sex b);

Sex
logerror2(Sex a, Sex b)
{
    return aerror2("bad args for logical operations", a, b);
}

Sex
logerror1(Sex a)
{
    return aerror1("bad arg for logical operations", a);
}

Sex
mklogical(long n)
{
    Sex             c;
    c = new_logical(bnum_len(1));
    logical_value(c)[0] = n;
    return c;
}

Sex
Llogical(Sex a)
{
    Sex             c;
    long            n;

    if (is_fixnum(a)) {
	n = mkint(a);
	c = new_logical(bnum_len(1));
	logical_value(c)[0] = n;
	return c;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	    return a;
	default:
	    n = numlen(a);
	    stackcheck();
	    push(a);
	    c = new_logical(n);
	    pop(a);
	    n = bnum_n(n);
	    while (--n >= 0)
		logical_value(c)[n] = numvec(a, n);
	    return c;
	}
    }
    return logerror1(a);
}



Sex
lognot(Sex a)
{
    if (is_fixnum(a))
	return (Sex) ((long) a ^ ((unsigned long)(-1) << 4));
    else if (is_bigint(a)) {
	a = plus2(a, mkfxat(1));
	return negate(a);
    } else
	return logerror1(a);
}

Sex
ash(Sex a, Sex b)
{
    long            n,
                    m,
                    al,
                    ah;
    long            n1,
                    n2,
                    type;
    Sex             c;

    if (!is_fixnum(b))
	return logerror2(a, b);
    n = mkint(b);
    if (n == 0)
	return a;
    if (is_fixnum(a)) {
	m = mkint(a);
	if (m == 0)
	    return a;
#ifdef _LP64
	if (n < -(28 + 32)) {
	    return mkfxat(m >> (28 + 32));	/* 
						 * ??? by f.k 
						 */
	}
#else
	if (n < -(28)) {
	    return mkfxat(m >> (28));	/* 
					 * ??? by f.k 
					 */
	}
#endif
	else if (n < 0) {
	    return mkfxat(m >> n);
	} else if (n < LPSIZE) {
	    if (m > 0) {
		ah = m >> (LPSIZE - n);
		al = m << n;
		if (ah == 0) {
		    if (al < big_min)
			return mkfxat(al);
		    c = new_bignum(bnum_len(1));
		    bignum_value(c)[0] = al;
		    return c;
		}
		c = new_bignum(bnum_len(2));
		bignum_value(c)[0] = al;
		bignum_value(c)[1] = ah;
		return c;
	    } else {
		m = -m;
		ah = m >> (LPSIZE - n);
		al = m << n;
		if (ah == 0) {
		    if (al <= big_min)
			return mkfxat(-al);
		    c = new_bigneg(bnum_len(1));
		    bignum_value(c)[0] = al;
		    return c;
		}
		c = new_bigneg(bnum_len(2));
		bignum_value(c)[0] = al;
		bignum_value(c)[1] = ah;
		return c;
	    }
	}
	n1 = n & (LPSIZE - 1);
	n2 = n / LPSIZE;
	if (m < 0) {
	    type = T_BIGNEG;
	    m = -m;
	} else {
	    type = T_BIGNUM;
	}

	ah = m >> (LPSIZE - n1);
	al = m << n1;
	if (ah == 0) {
	    c = new_binteger(type, bnum_len(n2 + 1));
	    bignum_value(c)[n2] = al;
	} else {
	    c = new_binteger(type, bnum_len(n2 + 2));
	    bignum_value(c)[n2] = al;
	    bignum_value(c)[n2 + 1] = ah;
	}
	for (n = 0; n < n2; n++)
	    bignum_value(c)[n] = 0;
	return c;
    }
    if (is_bigint(a))
	type = type_of_header(a);
    else
	return logerror2(a, b);

    if (n > 0) {
	long            la = bnum_n(bignum_length(a));
	long            n1 = n & (LPSIZE - 1);
	long            n2 = n >> LP_LOG2;
	long            lc;
	Sex             c;
	unsigned long   u;

	u = bignum_value(a)[--la];
	ah = u >> (LPSIZE - n1);
	al = u << n1;
	lc = la + n2;
	if (ah)
	    lc++;
	stackcheck();
	push(a);
	c = new_binteger(type, bnum_len(lc));
	pop(a);
	while (la > 0) {
	    if (ah)
		bignum_value(c)[lc--] = ah;
	    bignum_value(c)[lc--] = al | (u >> (LPSIZE - n1));
	    u = bignum_value(a)[--la];

	}
	return c;
    } else {
	long            la = bnum_n(bignum_length(a));
	long            n1;
	long            n2;
	long            lc;
	Sex             c;

	n = -n;
	n1 = n & (LPSIZE - 1);
	n2 = n >> LP_LOG2;
	lc = la - n2;
	if (lc < 0)
	    return mkfxat(0);
	if (lc == 0)
	    return mkfxat(1);
	stackcheck();
	push(a);
	c = new_bignum(bnum_len(lc));
	pop(a);
	return c;
    }
}


Sex
logior2(Sex a, Sex b)
{
    return (Sex) ((long) a | (long) b);
}



static          Sex
logeqvbb(Sex a, Sex b)
{
    long            lena,
                    lenb,
                    i;
    Sex             c;
    long            lenc;

    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lena > lenb) {
	c = a;
	lenc = lena;
	a = b;
	lena = lenb;
	b = c;
	lenb = lenc;
    }
    stackcheck();
    push2(a, b);
    c = new_logical(bnum_len(lena));
    pop2(b, a);
    for (i = 0; i < lena; i++)
	bignum_value(c)[i] = bignum_value(a)[i] ^ (~bignum_value(b)[i]);
    return c;
}



static          Sex
logeqvib(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = mkbigat1(mkint(a));
    pop(b);
    return logeqvbb(a, b);
}

Sex
logeqv2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return (Sex) ((long) a ^ (long) b ^ (long) mkfxat(-1));
	} else if (is_numhdr(b)) {
	    switch (numtype_of(b)) {
	    case N_LOGICAL:
	    case N_BIGNUM:
		return logeqvib(a, b);
	    }
	}
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    if (is_fixnum(b)) {
		return logeqvib(b, a);
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
		    return logeqvbb(a, b);
		}
	    }
	    break;
	}
    }
    return logerror2(a, b);
}


static          Sex
logxorbb(Sex a, Sex b)
{
    long            lena,
                    lenb,
                    i;
    Sex             c;
    long            lenc;

    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lena > lenb) {
	c = a;
	lenc = lena;
	a = b;
	lena = lenb;
	b = c;
	lenb = lenc;
    }
    stackcheck();
    push2(a, b);
    c = new_logical(bnum_len(lena));
    pop2(b, a);
    for (i = 0; i < lena; i++)
	bignum_value(c)[i] = bignum_value(a)[i] ^ bignum_value(b)[i];
    return c;
}



static          Sex
logxorib(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = mkbigat1(mkint(a));
    pop(b);
    return logxorbb(a, b);
}

Sex
logxor2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return (Sex) (((long) a ^ (long) b) | TAG_FIXNUM);
	} else if (is_numhdr(b)) {
	    switch (numtype_of(b)) {
	    case N_LOGICAL:
	    case N_BIGNUM:
		return logxorib(a, b);
	    }
	}
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    if (is_fixnum(b)) {
		return logxorib(b, a);
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
		    return logxorbb(a, b);
		}
	    }
	    break;
	}
    }
    return logerror2(a, b);
}


static          Sex
logandbb(Sex a, Sex b)
{
    long            lena,
                    lenb,
                    i;
    Sex             c;
    long            lenc;

    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lena > lenb) {
	c = a;
	lenc = lena;
	a = b;
	lena = lenb;
	b = c;
	lenb = lenc;
    }
    stackcheck();
    push2(a, b);
    c = new_logical(bnum_len(lena));
    pop2(b, a);
    for (i = 0; i < lena; i++)
	bignum_value(c)[i] = bignum_value(a)[i] & bignum_value(b)[i];
    return c;
}



static          Sex
logandib(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = mkbigat1(mkint(a));
    pop(b);
    return logandbb(a, b);
}

Sex
logand2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return (Sex) ((long) a & (long) b);
	} else if (is_numhdr(b)) {
	    switch (numtype_of(b)) {
	    case N_LOGICAL:
	    case N_BIGNUM:
		return logandib(a, b);
	    }
	}
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    if (is_fixnum(b)) {
		return logandib(b, a);
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
		    return logandbb(a, b);
		}
	    }
	    break;
	}
    }
    return logerror2(a, b);
}

static          Sex
logorbb(Sex a, Sex b)
{
    long            lena,
                    lenb,
                    i;
    Sex             c;
    long            lenc;

    lena = bnum_n(bignum_length(a));
    lenb = bnum_n(bignum_length(b));
    if (lena > lenb) {
	c = a;
	lenc = lena;
	a = b;
	lena = lenb;
	b = c;
	lenb = lenc;
    }
    stackcheck();
    push2(a, b);
    c = new_logical(bnum_len(lena));
    pop2(b, a);
    for (i = 0; i < lena; i++)
	bignum_value(c)[i] = bignum_value(a)[i] | bignum_value(b)[i];
    return c;
}



static          Sex
logorib(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = mkbigat1(mkint(a));
    pop(b);
    return logorbb(a, b);
}

Sex
logor2(a, b)
     Sex             a,
                     b;
{
    if (is_fixnum(a)) {
	if (is_fixnum(b)) {
	    return (Sex) ((long) a | (long) b);
	} else if (is_numhdr(b)) {
	    switch (numtype_of(b)) {
	    case N_LOGICAL:
	    case N_BIGNUM:
		return logorib(a, b);
	    }
	}
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    if (is_fixnum(b)) {
		return logorib(b, a);
	    } else if (is_numhdr(b)) {
		switch (numtype_of(b)) {
		case N_LOGICAL:
		case N_BIGNUM:
		    return logorbb(a, b);
		}
	    }
	    break;
	}
    }
    return logerror2(a, b);
}



Sex
negatel(Sex a)
{
    Sex             b;
    long            len = logical_length(a);
    long            i,
                    l,
                    carry;

    stackcheck();
    push(a);
    b = new_logical(len);
    pop(a);
    len = bnum_n(len);
    carry = 1;
    for (i = 0; i < len; i++) {
	l = (~logical_value(a)[i]) + carry;
	carry = (carry == 1 && l == 0) ? 1 : 0;
	logical_value(b)[i] = l;
    }
    return b;
}
