/*
 * matrix.c                (C) Copyright 2011 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/matrix.c,v 2.9 2015/03/12 03:09:17 kako Exp $
 */

static char const rcsid[] =
    "$Id: matrix.c,v 2.9 2015/03/12 03:09:17 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "print.h"
#include "entries.h"
#include "misc.h"

/***************************************************************/
/***************************************************************/

Sex             make_matrix(long a, long b);

Sex             new_matrix(long a, long b);
Sex             Ldeterminant(Sex a);
Sex             Ltranspose(Sex a);
Sex             Lident(Sex a);
Sex             Ldiag(Sex a);
Sex             Atrace(Sex a);
Sex             inverseMatrix(Sex a);
Sex             copyMatrix(Sex a);
Sex             addMatrix(Sex a, Sex b);
Sex             subMatrix(Sex a, Sex b);
Sex             scaleMatrix(Sex a, Sex b);
Sex             scaleaddMatrix(Sex a, Sex x, Sex b, Sex y);
Sex             copyMatrix(Sex a);

Sex             Arowvector(Sex a, Sex c);
Sex             Acolvector(Sex a, Sex c);
Sex             Asetrow(Sex a, Sex b, Sex c);
Sex             Asetcol(Sex a, Sex b, Sex c);
Sex             Aswap_row(Sex a, Sex b, Sex c);
Sex             Aswap_col(Sex a, Sex b, Sex c);
Sex             Ascale_row(Sex M, Sex vi, Sex a);
Sex             Ascale_col(Sex M, Sex vi, Sex a);
Sex             Amuladd_row(Sex M, Sex vi, Sex a, Sex vj);
Sex             Amuladd_col(Sex M, Sex vi, Sex a, Sex vj);
Sex             Ascale_muladd_row(Sex M, Sex vi, Sex a, Sex vj, Sex b);
Sex             Ascale_muladd_col(Sex M, Sex vi, Sex a, Sex vj, Sex b);
Sex             Ains_row(Sex a, Sex b, Sex c);
Sex             Ains_col(Sex a, Sex b, Sex c);
Sex             ins_rowvector(Sex a, Sex b, Sex c);
Sex             ins_colvector(Sex a, Sex b, Sex c);
Sex             Amerge_row(Sex a, Sex b);
Sex             Amerge_col(Sex a, Sex b);
Sex             Adel_row(Sex a, Sex b);
Sex             Adel_col(Sex a, Sex b);
Sex             Asubmatrix(Sex a, Sex b, Sex c);
Sex             Acofactor(Sex a, Sex b, Sex c);
Sex             Arank(Sex a);
Sex             matrix_subst(Sex a, Sex v, Sex b);
Sex             matrix_reval(Sex a);
Sex             matrix_diff(Sex a, Sex v);

Sex             Anewmatrix(Sex a);
Sex             Anewvector(Sex a);

extern Sex      make_vector(long n);
extern Sex      new_vector(long n);
extern Sex      addVector(Sex a, Sex b);
extern Sex      subVector(Sex a, Sex b);
extern Sex      scaleVector(Sex a, Sex b);
extern Sex      scaleaddVector(Sex a, Sex x, Sex b, Sex y);


/***************************************************************/
/***************************************************************/

void
print_matrix(Sex a, putchar_fn * fn)
{
    long            len,
                    i,
                    j,
                    k;
    long            dimx,
                    dimy;

    push(a);
    dimx = mkint(matrix_dimx(a));
    dimy = mkint(matrix_dimy(a));
    fn('[');
    k = 0;
    for (i = 0; i < dimx; i++) {
	if (i != 0)
	    fn(',');
	fn('[');
	for (j = 0; j < dimy; j++) {
	    if (j != 0)
		fn(',');
	    print_sex(matrix_value(stack[0], k), fn);
	    k++;
	}
	fn(']');
    }
    fn(']');
    popv(1);
}

/***********************************************************************/
/***********************************************************************/
Sex
make_matrix(long a, long b)
{
    long            i;
    long            n;
    long            dim[2];
    Sex             p;
    Sex            *m;

    n = a * b;
    if (n < 0 || n > 0x0fffffff)
	return aerror2("mkmatrix", a, b);

    p = getvector(TAG_VECTOR, T_MATRIX,
		  sizeof(Matrix) + (n - 1) * sizeof(Sex));
    matrix_dimx(p) = mkfxat(a);
    matrix_dimy(p) = mkfxat(b);
    m = &matrix_value(p, 0);
    for (i = 0; i < n; i++) {
	*m++ = MKMATRIX_INIT_VALUE;
    }
    return p;
}

Sex
new_matrix(long a, long b)
{
    Sex             p;
    long            n = a * b;
    if (n < 0 || n > 0x0fffffff)
	return aerror("matrix size");
    p = getvector(TAG_VECTOR, T_MATRIX, sizeof(Matrix) + (n - 1) * SEX);
    matrix_dimx(p) = mkfxat(a);
    matrix_dimy(p) = mkfxat(b);
    return p;
}
/***************************************************************/
/***************************************************************/

Sex
Anewvector(Sex v)
{
    Sex            *q;
    int             n;
    Sex             a;


    n = 0;
    a = v;
    while (is_cons(a)) {
	a = cdr(a);
	n++;
    }
    push(v);
    a = new_vector(n);
    pop(v);
    q = (Sex *) & vect_value(a, 0);

    while (is_cons(v)) {
	*q++ = car(v);
	v = cdr(v);
    }
    return a;
}

Sex
Anewmatrix(Sex v)
{
    long            i,
                    j;
    long            n;
    long            sx,
                    sy;
    Sex             a;
    Sex             p;
    Sex            *q;

    sx = 0;
    a = v;
    while (is_cons(a)) {
	sx++;
	a = cdr(a);
    }
    sy = 0;
    a = v;
    if (is_cons(a)) {
	a = car(a);
	while (is_cons(a)) {
	    sy++;
	    a = cdr(a);
	}
    }

    push(v);
    a = new_matrix(sx, sy);
    pop(v);
    q = (Sex *) & matrix_value(a, 0);
    for (i = 0; i < sx; i++) {
	if (!is_cons(v))
	    goto err;
	p = car(v);
	v = cdr(v);
	for (j = 0; j < sy; j++) {
	    if (!is_cons(p))
		goto err;
	    q[j] = car(p);
	    p = cdr(p);
	}
	q += sy;
    }

    return a;
  err:
    return aerror("list length");
}




/***************************************************************/
/***************************************************************/

Sex
gplus2MM(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;
    Sex             p;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    if (dx != matrix_dimx(b) || dy != matrix_dimy(b))
	aerror2("gplus2:", a, b);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	p = gplus2(matrix_value(a, i), matrix_value(b, i));
	matrix_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
gdiff2MM(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;
    Sex             p;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    if (dx != matrix_dimx(b) || dy != matrix_dimy(b))
	aerror2("gplus2:", a, b);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	p = gdiff2(matrix_value(a, i), matrix_value(b, i));
	matrix_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
gtimes2Mx(Sex a, Sex b)
{
    return gtimes2xM(b, a);
}


Sex
gtimes2xM(Sex b, Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;
    Sex             p;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	p = gtimes2(matrix_value(a, i), b);
	matrix_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
gtimes2MM(Sex a, Sex b)
{
    long            dx,
                    dy,
                    dz;
    long            i,
                    j,
                    k;

    Sex             c;
    Sex             p,
                    q;
    Sex             ax,
                    by;

    if (matrix_dimy(a) != matrix_dimx(b))
	aerror2("gplus2:", a, b);
    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    dz = mkint(matrix_dimy(b));
    push2(a, b);
    c = make_matrix(dx, dz);
    push2(c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define p stack[0]

    for (i = 0; i < dx; i++) {
	for (k = 0; k < dz; k++) {
	    p = mkfxat(0);
	    for (j = 0; j < dy; j++) {
		ax = matrix_value(a, i * dy + j);
		by = matrix_value(b, j * dz + k);
		q = gtimes2(ax, by);
		p = gplus2(p, q);
	    }
	    matrix_value(c, i * dz + k) = p;
	}
    }
#undef a
#undef b
#undef c
#undef p
    c = stack[-1];
    popv(4);
    return c;
}

Sex
gtimes2MV(Sex a, Sex b)
{
    long            dx,
                    dy,
                    dz;
    long            i,
                    j,
                    k;

    Sex             c;
    Sex             p,
                    q;
    Sex             ax,
                    by;

    if (matrix_dimy(a) != vect_dimx(b))
	aerror2("gplus2:", a, b);
    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    push2(a, b);
    c = make_vector(dx);
    push2(c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define p stack[0]

    for (i = 0; i < dx; i++) {
	p = mkfxat(0);
	for (j = 0; j < dy; j++) {
	    ax = matrix_value(a, i * dy + j);
	    by = vect_value(b, j);
	    q = gtimes2(ax, by);
	    p = gplus2(p, q);
	}
	vect_value(c, i) = p;
    }
#undef a
#undef b
#undef c
#undef p
    c = stack[-1];
    popv(4);
    return c;
}

Sex
gtimes2VM(Sex a, Sex b)
{
    long            dx,
                    dy,
                    dz;
    long            i,
                    j,
                    k;

    Sex             c;
    Sex             p,
                    q;
    Sex             ax,
                    by;

    if (vect_dimx(a) != matrix_dimx(b))
	aerror2("gplus2:", a, b);
    dx = mkint(matrix_dimx(b));
    dy = mkint(matrix_dimy(b));
    push2(a, b);
    c = make_vector(dy);
    push2(c, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define p stack[0]

    for (j = 0; j < dy; j++) {
	p = mkfxat(0);
	for (i = 0; i < dx; i++) {
	    ax = vect_value(a, i);
	    by = matrix_value(b, i * dy + j);
	    q = gtimes2(ax, by);
	    p = gplus2(p, q);
	}
	vect_value(c, j) = p;
    }
#undef a
#undef b
#undef c
#undef p
    c = stack[-1];
    popv(4);
    return c;
}

Sex
gratio2Mx(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;
    Sex             p;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	p = gratio2(matrix_value(a, i), b);
	matrix_value(c, i) = p;
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}


Sex
gratio2xM(Sex a, Sex b)
{
    push(a);
    b = inverseMatrix(b);
    pop(a);
    return gtimes2xM(a, b);
}

Sex
gratio2VM(Sex a, Sex b)
{
    push(a);
    b = inverseMatrix(b);
    pop(a);
    return gtimes2VM(a, b);
}




Sex
gratio2MM(Sex a, Sex b)
{
    push(a);
    b = inverseMatrix(b);
    pop(a);
    return gtimes2MM(a, b);
}


Sex
inverseMatrix(Sex a)
{
    long            n;
    long            i,
                    j,
                    k;

    Sex             c;
    Sex             r;
    Sex            *ax,
                   *ay;

    if (matrix_dimx(a) != matrix_dimy(a))
	aerror1("not square matrix", a);
    n = mkint(matrix_dimx(a));
    push(a);
    a = copyMatrix(a);
    push(a);
    c = Lident(mkfxat(n));
    push3(c, nil, nil);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define p stack[-1]
#define q stack[0]

    for (i = 0; i < n; i++) {
	for (j = i; j < n; j++) {
	    if (!zerop(matrix_value(b, i * n + j)))
		break;
	}
	if (j == n) {
	    r = b;
	    popv(5);
	    return aerror1("singular matrix", r);
	} else if (j != i) {
	    ax = &matrix_value(b, i * n);
	    ay = &matrix_value(b, j * n);
	    for (k = 0; k < n; k++) {
		q = *ax;
		*ax++ = *ay;
		*ay++ = q;
	    }
	    ax = &matrix_value(c, i * n);
	    ay = &matrix_value(c, j * n);
	    for (k = 0; k < n; k++) {
		q = *ax;
		*ax++ = *ay;
		*ay++ = q;
	    }
	}
	p = matrix_value(b, i * n + i);
	for (j = i + 1; j < n; j++) {
	    q = matrix_value(b, j * n + i);
	    if (zerop(q))
		continue;
	    q = gratio2(q, p);
	    for (k = i; k < n; k++) {
		r = gtimes2(q, matrix_value(b, i * n + k));
		r = gdiff2(matrix_value(b, j * n + k), r);
		matrix_value(b, j * n + k) = r;
	    }
	    for (k = 0; k < n; k++) {
		r = gtimes2(q, matrix_value(c, i * n + k));
		r = gdiff2(matrix_value(c, j * n + k), r);
		matrix_value(c, j * n + k) = r;
	    }

	}
    }
    for (i = n - 1; i >= 0; i--) {
	p = matrix_value(b, i * n + i);
	for (j = i - 1; j >= 0; j--) {
	    q = matrix_value(b, j * n + i);
	    q = gratio2(q, p);
	    for (k = 0; k < n; k++) {
		r = gtimes2(q, matrix_value(c, i * n + k));
		r = gdiff2(matrix_value(c, j * n + k), r);
		matrix_value(c, j * n + k) = r;
	    }
	}
	p = gratio2(mkfxat(1), p);
	if (!onep(p)) {
	    for (j = 0; j < n; j++) {
		r = gtimes2(p, matrix_value(c, i * n + j));
		matrix_value(c, i * n + j) = r;
	    }
	}
    }
    // lisp_printf("inverse: %p\n", b);
#undef a
#undef b
#undef c
#undef p
#undef q
    c = stack[-2];
    popv(5);
    return c;
}


Sex
gexpt2Mi(Sex a, Sex b)
{
    long            n = mkint(b);
    Sex             sx;
    Sex             c;

    if (matrix_dimx(a) != matrix_dimy(a))
	return aerror1("not a square matrix", a);
    sx = matrix_dimx(a);
    if (n == 0)
	return Lident(sx);
    if (n < 0) {
	a = inverseMatrix(a);
	n = -n;
    }
    if (n == 1)
	return a;
    push(a);
    c = Lident(sx);
    push(c);
#define a stack[-1]
#define c stack[0]
    while (n > 1) {
	if (n & 1)
	    c = gtimes2MM(c, a);
	a = gtimes2MM(a, a);
	n = n >> 1;
    }
    c = gtimes2MM(a, c);
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

/***************************************************************/
/*
 * inplace operation of matrices 
 */
/***************************************************************/

Sex
addMatrix(Sex a, Sex b)
{				/* a + b --> a */
    Sex             dx,
                    dy;
    long            i,
                    n;

    Sex             p;

    if (!is_matrix(a))
	return aerror("scaleMatrix");
    if (matrix_dimx(a) != matrix_dimx(b) ||
	matrix_dimy(a) != matrix_dimy(b))
	return aerror("scaleMatrix");
    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    n = mkint(dx) * mkint(dy);
#define a stack[-1]
#define b stack[-0]
    for (i = 0; i < n; i++) {
	p = gplus2(matrix_value(a, i), matrix_value(b, i));
	matrix_value(a, i) = p;
    }
#undef a
#undef b
    p = stack[-1];
    popv(2);
    return p;
}

Sex
subMatrix(Sex a, Sex b)
{				/* a - b --> a */
    Sex             dx,
                    dy;
    long            i,
                    n;

    Sex             p;

    if (!is_matrix(a))
	return aerror("scaleMatrix");
    if (matrix_dimx(a) != matrix_dimx(b) ||
	matrix_dimy(a) != matrix_dimy(b))
	return aerror("scaleMatrix");
    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    n = mkint(dx) * mkint(dy);
#define a stack[-1]
#define b stack[-0]
    for (i = 0; i < n; i++) {
	p = gdiff2(matrix_value(a, i), matrix_value(b, i));
	matrix_value(a, i) = p;
    }
#undef a
#undef b
    p = stack[-1];
    popv(2);
    return p;
}

Sex
scaleMatrix(Sex a, Sex b)
{				/* a * b --> a */
    Sex             dx,
                    dy;
    long            i,
                    n;

    Sex             p;

    if (!is_matrix(a))
	return aerror("scaleMatrix");
    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    n = mkint(dx) * mkint(dy);
#define a stack[-1]
#define b stack[-0]
    for (i = 0; i < n; i++) {
	p = gtimes2(matrix_value(a, i), b);
	matrix_value(a, i) = p;
    }
#undef a
#undef b
    p = stack[-1];
    popv(2);
    return p;
}


Sex
scaleaddMatrix(Sex a, Sex x, Sex b, Sex y)
{				/* a*x + b*y --> a */
    Sex             dx,
                    dy;
    long            i,
                    n;

    Sex             p,
                    q;

    if (!is_matrix(a))
	return aerror("scaleMatrix");
    if (matrix_dimx(a) != matrix_dimx(b) ||
	matrix_dimy(a) != matrix_dimy(b))
	return aerror("scaleMatrix");
    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push4(a, b, x, y);
    n = mkint(dx) * mkint(dy);
    push(nil);
#define a stack[-4]
#define b stack[-3]
#define x stack[-2]
#define y stack[-1]
#define p stack[0]
    for (i = 0; i < n; i++) {
	p = gtimes2(matrix_value(a, i), x);
	q = gtimes2(matrix_value(b, i), y);
	q = gplus2(p, q);
	matrix_value(a, i) = q;
    }
    q = a;
#undef a
#undef b
#undef x
#undef y
#undef p
    popv(5);
    return q;
}

/***************************************************************/
/***************************************************************/

Sex
matrix_converter(Sex a, one_args * cf)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = converter(matrix_value(a, i), cf);
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
matrix_converter2(Sex a, Sex b, two_args * cf)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = converter2(matrix_value(a, i), b, cf);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}



Sex
Aunmodular_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Aunmodular(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Amkmodular_mat(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Amkmodular(matrix_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Amodular_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Amodular(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Afmodular_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Afmodular(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Afloat_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Afloat(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Abfloat_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Abfloat(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Amkefloat_mat(Sex a, Sex f)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, f);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define f stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Amkefloat(matrix_value(a, i), f);
    }
#undef a
#undef f
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Amkbefloat_mat(Sex a, Sex f)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, f);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define f stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Amkbefloat(matrix_value(a, i), f);
    }
#undef a
#undef f
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
Aefloat_value_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Aefloat_value(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
Aefloat_error_mat(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Aefloat_error(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}


Sex
gnegateM(Sex a)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push(a);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = gnegate(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}

Sex
gmodulusM(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = gmodulus(matrix_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

Sex
gsmodulusM(Sex a, Sex b)
{
    Sex             dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = matrix_dimx(a);
    dy = matrix_dimy(a);
    push2(a, b);
    c = make_matrix(mkint(dx), mkint(dy));
    n = mkint(dx) * mkint(dy);
    push(c);
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = gsmodulus(matrix_value(a, i), b);
    }
#undef a
#undef b
#undef c
    c = stack[0];
    popv(3);
    return c;
}

/***************************************************************/
/***************************************************************/

Sex
matrix_diff(Sex a, Sex v)
{
    long            dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    push2(a, v);
    c = make_matrix((dx), (dy));
    n = (dx) * (dy);
    push(c);
#define a stack[-2]
#define x stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Adiff(matrix_value(a, i), x);
    }
#undef a
#undef x
#undef c
    c = stack[0];
    popv(3);
    return c;
}

/***************************************************************/
/***************************************************************/

Sex
matrix_subst(Sex a, Sex v, Sex b)
{
    long            dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    stackcheck();
    push3(a, v, b);
    c = make_matrix((dx), (dy));
    n = (dx) * (dy);
    push(c);
#define a stack[-3]
#define x stack[-2]
#define b stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Asub(matrix_value(a, i), x, b);
    }
#undef a
#undef x
#undef b
#undef c
    c = stack[0];
    popv(4);
    return c;
}

Sex
matrix_reval(Sex a)
{
    long            dx,
                    dy;
    int             i,
                    n;

    Sex             c;

    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    stackcheck();
    push(a);
    c = make_matrix((dx), (dy));
    n = (dx) * (dy);
    push(c);
#define a stack[-1]
#define c stack[-0]
    for (i = 0; i < n; i++) {
	matrix_value(c, i) = Areval(matrix_value(a, i));
    }
#undef a
#undef c
    c = stack[0];
    popv(2);
    return c;
}


/***************************************************************/
/***************************************************************/
Sex
copyMatrix(Sex a)
{
    long            dx,
                    dy;
    long            i,
                    n;

    Sex             c;

    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    stackcheck();
    push(a);
    c = make_matrix(dx, dy);
    pop(a);
    n = dx * dy;

    for (i = 0; i < n; i++) {
	matrix_value(c, i) = matrix_value(a, i);
    }
    return c;
}


Sex
Ldeterminant(Sex a)
{
    long            n;
    long            i,
                    j,
                    k;

    Sex             r;
    Sex            *ax,
                   *ay;

    if (matrix_dimx(a) != matrix_dimy(a))
	aerror1("not square matrix", a);
    n = mkint(matrix_dimx(a));
    a = copyMatrix(a);
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define p stack[-1]
#define q stack[0]

    for (i = 0; i < n; i++) {
	for (j = i; j < n; j++) {
	    if (!zerop(matrix_value(a, j * n + i)))
		break;
	}
	if (j == n) {
	    popv(3);
	    return mkfxat(0);
	} else if (j != i) {
	    ax = &matrix_value(a, i * n);
	    ay = &matrix_value(a, j * n);
	    for (k = 0; k < n; k++) {
		q = *ax;
		*ax++ = *ay;
		*ay++ = q;
	    }
	}
	p = matrix_value(a, i * n + i);
	for (j = i + 1; j < n; j++) {
	    q = matrix_value(a, j * n + i);
	    if (zerop(q))
		continue;
	    q = gratio2(q, p);
	    for (k = i; k < n; k++) {
		r = gtimes2(q, matrix_value(a, i * n + k));
		r = gdiff2(matrix_value(a, j * n + k), r);
		matrix_value(a, j * n + k) = r;
	    }
	}
    }
    p = matrix_value(a, 0);
    for (i = 1; i < n; i++) {
	p = gtimes2(p, matrix_value(a, i * n + i));
    }
    r = p;
#undef a
#undef p
#undef q
    popv(3);
    return r;
}

Sex
Ltranspose(Sex a)
{
    long            i,
                    j;
    long            dx,
                    dy;
    Sex             b;
    Sex            *av,
                   *bv;

    if (!is_matrix(a))
	aerror1("transpose: not a matrix %p\n", a);
    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    stackcheck();
    push(a);
    b = new_matrix(dy, dx);
    pop(a);
    av = &matrix_value(a, 0);
    bv = &matrix_value(b, 0);
    for (i = 0; i < dx; i++) {
	for (j = 0; j < dy; j++) {
	    bv[j * dx + i] = av[i * dy + j];
	}
    }
    return b;
}

Sex
Atrace(Sex a)
{
    long            i;
    long            dx,
                    dy;
    Sex             p;

    if (!is_matrix(a))
	return aerror1("transpose: not a matrix %p\n", a);
    dx = mkint(matrix_dimx(a));
    dy = mkint(matrix_dimy(a));
    if (dx != dy)
	return aerror("trace");
    stackcheck();
    push(a);
#define a stack[-0]
    p = matrix_value(a, 0);
    for (i = 1; i < dx; i++) {
	p = gplus2(p, matrix_value(a, i * dx + i));
    }
#undef a
    return p;
}

Sex
Lident(Sex a)
{
    long            n;
    long            i,
                    j;
    Sex             b;
    Sex            *bv;

    if (!is_fixnum(a) || ((n = mkint(a)) <= 0))
	aerror("ident: size");
    b = new_matrix(n, n);
    bv = &matrix_value(b, 0);
    for (i = 0; i < n; i++) {
	for (j = 0; j < n; j++) {
	    bv[j * n + i] = mkfxat(0);
	}
	bv[i * n + i] = mkfxat(1);
    }
    return b;
}


Sex
Ldiag(Sex a)
{
    long            n;
    long            i,
                    j;
    Sex             b;
    Sex             c;
    Sex            *bv;
    int             flg = 0;

    if (is_vect(a)) {
	n = mkint(vect_dimx(a));
	flg = 1;
    } else if (is_cons(a))
	n = list_length(a);
    else
	aerror("diag");
    push(a);
    b = new_matrix(n, n);
    pop(a);
    bv = &matrix_value(b, 0);
    for (i = 0; i < n; i++) {
	for (j = 0; j < n; j++) {
	    bv[j * n + i] = mkfxat(0);
	}
	if (flg) {
	    c = vect_value(a, i);
	} else {
	    c = car(a);
	    a = cdr(a);
	}
	bv[i * n + i] = c;
    }
    return b;
}


/***************************************************************/
/***************************************************************/

Sex
Arowvector(Sex a, Sex b)
{
    long            i,
                    k;
    long            n,
                    m;

    Sex            *ax,
                   *cx;
    Sex             c;

    if (!is_matrix(a))
	return aerror("rowvector: not matrix");
    k = 0;
    if (is_fixnum(b))
	k = mkint(b);
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (k < 1 || k > n)
	return aerror("rowvector: index out of range");
    stackcheck();
    push(a);
    c = new_vector(m);
    pop(a);
    ax = &matrix_value(a, (k - 1) * m);
    cx = &vect_value(c, 0);
    for (i = 0; i < m; i++) {
	*cx++ = *ax++;
    }
    return c;
}


Sex
Acolvector(Sex a, Sex b)
{
    long            i,
                    k;
    long            n,
                    m;

    Sex            *ax,
                   *cx;
    Sex             c;

    if (!is_matrix(a))
	return aerror("colvector: not matrix");
    k = 0;
    if (is_fixnum(b))
	k = mkint(b);
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (k < 1 || k > m)
	return aerror("colvector : index out of range");
    stackcheck();
    push(a);
    c = new_vector(n);
    pop(a);
    ax = &matrix_value(a, k - 1);
    cx = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	*cx++ = ax[i * m];
    }
    return c;
}

Sex
Asetrow(Sex a, Sex b, Sex c)
{
    long            i,
                    k;
    long            n,
                    m;

    Sex            *ax,
                   *cx;

    k = 0;
    if (!is_matrix(a) || !is_vect(c))
	return aerror("setrow");
    if (matrix_dimy(a) != vect_dimx(c))
	return aerror("setrow: size mismatch");
    if (is_fixnum(b))
	k = mkint(b);
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (k < 1 || k > n)
	return aerror("setrow: index out of range");
    ax = &matrix_value(a, (k - 1) * m);
    cx = &vect_value(c, 0);
    for (i = 0; i < m; i++) {
	*ax++ = *cx++;
    }
    return a;
}


Sex
Asetcol(Sex a, Sex b, Sex c)
{
    long            i,
                    k;
    long            n,
                    m;

    Sex            *ax,
                   *cx;

    k = 0;
    if (!is_matrix(a) || !is_vect(c))
	return aerror("setcol");
    if (matrix_dimx(a) != vect_dimx(c))
	return aerror("setcol: size mismatch");
    if (is_fixnum(b))
	k = mkint(b);
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (k < 1 || k > m)
	return aerror("setcol: index out of range");
    ax = &matrix_value(a, k - 1);
    cx = &vect_value(c, 0);
    for (i = 0; i < n; i++) {
	ax[i * m] = *cx++;
    }
    return a;
}

Sex
Aswap_row(Sex a, Sex b, Sex c)
{
    long            i,
                    j;
    long            n,
                    m;

    Sex            *ai,
                   *aj;
    Sex             tmp;

    if (!is_matrix(a) || !is_fixnum(b) || !is_fixnum(c))
	return aerror("swap_row");
    i = mkint(b);
    j = mkint(c);
    if (i == j)
	return a;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i <= 0 || i > n || j <= 0 || j > n)
	return aerror("swap_row: index out of range");
    ai = &matrix_value(a, (i - 1) * m);
    aj = &matrix_value(a, (j - 1) * m);
    for (i = 0; i < m; i++) {
	tmp = *ai;
	*ai++ = *aj;
	*aj++ = tmp;
    }
    return a;
}


Sex
Aswap_col(Sex a, Sex b, Sex c)
{
    long            i,
                    j;
    long            n,
                    m;

    Sex            *ai,
                   *aj;
    Sex             tmp;

    if (!is_matrix(a) || !is_fixnum(b) || !is_fixnum(c))
	return aerror("swap_col");
    i = mkint(b);
    j = mkint(c);
    if (i == j)
	return a;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i <= 0 || i > m || j <= 0 || j > m)
	return aerror("swap_col: index out of range");
    ai = &matrix_value(a, i - 1);
    aj = &matrix_value(a, j - 1);
    for (i = 0; i < n; i++) {
	tmp = *ai;
	*ai = *aj;
	*aj = tmp;
	ai += m;
	aj += m;
    }
    return a;
}


Sex
Ascale_row(Sex M, Sex vi, Sex a)
{
    long            i;
    long            n,
                    m;
    long            k;

    Sex             ai;

    if (!is_matrix(M) || !is_fixnum(vi))
	return aerror("scale_row");
    i = mkint(vi);
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > n)
	return aerror("scale_row: index out of range");
    stackcheck();
    push2(M, a);
#define M stack[-1]
#define a stack[0]
    i = (i - 1) * m;
    for (k = 0; k < m; k++) {
	ai = matrix_value(M, i);
	ai = gtimes2(ai, a);
	matrix_value(M, i) = ai;
	i++;
    }
#undef M
#undef a
    M = stack[-1];
    popv(2);
    return M;
}

Sex
Ascale_col(Sex M, Sex vi, Sex a)
{
    long            i;
    long            n,
                    m;
    long            k;

    Sex             ai;

    if (!is_matrix(M) || !is_fixnum(vi))
	return aerror("scale_col");
    i = mkint(vi);
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > m)
	return aerror("scale_col: index out of range");
    stackcheck();
    push2(M, a);
#define M stack[-1]
#define a stack[0]
    i = i - 1;
    for (k = 0; k < n; k++) {
	ai = matrix_value(M, i);
	ai = gtimes2(ai, a);
	matrix_value(M, i) = ai;
	i += m;
    }
#undef M
#undef a
    M = stack[-1];
    popv(2);
    return M;
}

Sex
Amuladd_row(Sex M, Sex vi, Sex a, Sex vj)
{
    long            i,
                    j;
    long            n,
                    m;
    long            k;

    Sex             ai,
                    aj;

    if (!is_matrix(M) || !is_fixnum(vi) || !is_fixnum(vj))
	return aerror("muladd_row");
    i = mkint(vi);
    j = mkint(vj);
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > n || j <= 0 || j > n)
	return aerror("muladd_row: index out of range");
    stackcheck();
    push2(M, a);
#define M stack[-1]
#define a stack[0]
    i = (i - 1) * m;
    j = (j - 1) * m;
    for (k = 0; k < m; k++) {
	aj = matrix_value(M, j);
	aj = gtimes2(aj, a);
	ai = matrix_value(M, i);
	aj = gplus2(aj, ai);
	matrix_value(M, i) = aj;
	i++;
	j++;
    }
#undef M
#undef a
    M = stack[-1];
    popv(2);
    return M;
}

Sex
Amuladd_col(Sex M, Sex vi, Sex a, Sex vj)
{
    long            i,
                    j;
    long            n,
                    m;
    long            k;

    Sex             ai,
                    aj;

    if (!is_matrix(M) || !is_fixnum(vi) || !is_fixnum(vj))
	return aerror("muladd_col");
    i = mkint(vi);
    j = mkint(vj);
    if (i == j)
	return M;
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > m || j <= 0 || j > m)
	return aerror("muladd_col: index out of range");
    stackcheck();
    push2(M, a);
#define M stack[-1]
#define a stack[0]
    i = i - 1;
    j = j - 1;
    for (k = 0; k < n; k++) {
	aj = matrix_value(M, j);
	aj = gtimes2(aj, a);
	ai = matrix_value(M, i);
	aj = gplus2(aj, ai);
	matrix_value(M, i) = aj;
	i += m;
	j += m;
    }
#undef M
#undef a
    M = stack[-1];
    popv(2);
    return M;
}


Sex
Ascale_muladd_row(Sex M, Sex vi, Sex a, Sex vj, Sex b)
{
    long            i,
                    j;
    long            n,
                    m;
    long            k;

    Sex             ai,
                    aj;

    if (!is_matrix(M) || !is_fixnum(vi) || !is_fixnum(vj))
	return aerror("scale_muladd_row");
    i = mkint(vi);
    j = mkint(vj);
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > n || j <= 0 || j > n)
	return aerror("scale_muladd_row: index out of range");
    stackcheck();
    push4(M, a, b, nil);
#define M stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    j = (j - 1) * m;
    i = (i - 1) * m;
    for (k = 0; k < m; k++) {
	aj = matrix_value(M, j);
	c = gtimes2(aj, a);
	ai = matrix_value(M, i);
	ai = gtimes2(ai, b);
	aj = gplus2(c, ai);
	matrix_value(M, i) = aj;
	i++;
	j++;
    }
#undef M
#undef a
#undef b
#undef c
    M = stack[-3];
    popv(4);
    return M;
}

Sex
Ascale_muladd_col(Sex M, Sex vi, Sex a, Sex vj, Sex b)
{
    long            i,
                    j;
    long            n,
                    m;
    long            k;

    Sex             ai,
                    aj;

    if (!is_matrix(M) || !is_fixnum(vi) || !is_fixnum(vj))
	return aerror("scale_muladd_col");
    i = mkint(vi);
    j = mkint(vj);
    n = mkint(matrix_dimx(M));
    m = mkint(matrix_dimy(M));
    if (i <= 0 || i > m || j <= 0 || j > m)
	return aerror("scale_muladd_col: index out of range");
    stackcheck();
    push4(M, a, b, nil);
#define M stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[-0]
    j = j - 1;
    i = i - 1;
    for (k = 0; k < n; k++) {
	aj = matrix_value(M, j);
	c = gtimes2(aj, a);
	ai = matrix_value(M, i);
	ai = gtimes2(ai, b);
	aj = gplus2(c, ai);
	matrix_value(M, i) = aj;
	i += m;
	j += m;
    }
#undef M
#undef a
#undef b
#undef c
    M = stack[-3];
    popv(4);
    return M;
}


/***************************************************************/

Sex
Ains_row(Sex a, Sex b, Sex c)
{
    long            i;
    long            j,
                    k;
    long            n,
                    m;

    Sex            *pa,
                   *pb;

    if (!is_matrix(a) || !is_fixnum(b))
	return aerror("ins_row");
    if (is_vect(c))
	return ins_rowvector(a, b, c);
    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i < 0 || i > n)
	return aerror("ins_row: index out of range");
    push2(a, c);
    b = new_matrix(n + 1, m);
    pop2(c, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    for (k = 0; k < n; k++) {
	if (k == i) {
	    for (j = 0; j < m; j++) {
		if (is_cons(c)) {
		    *pb++ = car(c);
		    c = cdr(c);
		} else {
		    *pb++ = mkfxat(0);
		}
	    }
	}
	for (j = 0; j < m; j++) {
	    *pb++ = *pa++;
	}
    }
    if (n == i) {
	for (j = 0; j < m; j++) {
	    if (is_cons(c)) {
		*pb++ = car(c);
		c = cdr(c);
	    } else {
		*pb++ = mkfxat(0);
	    }
	}
    }

    return b;
}

Sex
ins_rowvector(Sex a, Sex b, Sex c)
{
    long            i;
    long            j,
                    k;
    long            n,
                    m;

    Sex            *pa,
                   *pb,
                   *pc;

    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (m != mkint(vect_dimx(c)))
	return aerror("ins_rowvector");
    if (i < 0 || i > n)
	return aerror("ins_row: index out of range");
    push2(a, c);
    b = new_matrix(n + 1, m);
    pop2(c, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    pc = &vect_value(c, 0);
    for (k = 0; k < n; k++) {
	if (k == i) {
	    for (j = 0; j < m; j++) {
		*pb++ = *pc++;
	    }
	}
	for (j = 0; j < m; j++) {
	    *pb++ = *pa++;
	}
    }
    if (n == i) {
	for (j = 0; j < m; j++) {
	    *pb++ = *pc++;
	}
    }

    return b;
}


Sex
Ains_col(Sex a, Sex b, Sex c)
{
    long            i;
    long            j,
                    k;
    long            n,
                    m;

    Sex            *pa,
                   *pb;

    if (!is_matrix(a) || !is_fixnum(b))
	return aerror("ins_col");
    if (is_vect(c))
	return ins_colvector(a, b, c);
    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i < 0 || i > m)
	return aerror("ins_col: index out of range");
    push2(a, c);
    b = new_matrix(n, m + 1);
    pop2(c, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    for (k = 0; k < n; k++) {
	for (j = 0; j < m; j++) {
	    if (j == i) {
		*pb++ = is_cons(c) ? car(c) : mkfxat(0);
	    }
	    *pb++ = *pa++;
	}
	if (i == m) {
	    *pb++ = is_cons(c) ? car(c) : mkfxat(0);
	}
	if (is_cons(c))
	    c = cdr(c);
    }

    return b;
}

Sex
ins_colvector(Sex a, Sex b, Sex c)
{
    long            i;
    long            j,
                    k;
    long            n,
                    m;

    Sex            *pa,
                   *pb,
                   *pc;

    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (mkint(vect_dimx(c)) != n)
	return aerror("ins_colvector");
    if (i < 0 || i > m)
	return aerror("ins_col: index out of range");
    push2(a, c);
    b = new_matrix(n, m + 1);
    pop2(c, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    pc = &vect_value(c, 0);
    for (k = 0; k < n; k++) {
	for (j = 0; j < m; j++) {
	    if (j == i) {
		*pb++ = *pc++;
	    }
	    *pb++ = *pa++;
	}
	if (i == m) {
	    *pb++ = *pc++;
	}
    }

    return b;
}

Sex
Amerge_row(Sex a, Sex b)
{
    long            j,
                    k;
    long            n,
                    m,
                    l;
    Sex             c;

    Sex            *pa,
                   *pb,
                   *pc;

    if (!is_matrix(a) || !is_matrix(b))
	return aerror("merge_row");
    if (matrix_dimy(a) != matrix_dimy(b))
	return aerror("ins_row: matrix size mismatch");
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    l = mkint(matrix_dimx(b));
    push2(a, b);
    c = new_matrix(n + l, m);
    pop2(b, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    pc = &matrix_value(c, 0);
    for (k = 0; k < n; k++) {
	for (j = 0; j < m; j++)
	    *pc++ = *pa++;
    }
    for (k = 0; k < l; k++) {
	for (j = 0; j < m; j++)
	    *pc++ = *pb++;
    }
    return c;
}

Sex
Amerge_col(Sex a, Sex b)
{
    long            j,
                    k;
    long            n,
                    m,
                    l;
    Sex             c;

    Sex            *pa,
                   *pb,
                   *pc;

    if (!is_matrix(a) || !is_matrix(b))
	return aerror("merge_col");
    if (matrix_dimx(a) != matrix_dimx(b))
	return aerror("merge_col: matrix size mismatch");
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    l = mkint(matrix_dimy(a));
    push2(a, b);
    c = new_matrix(n, m + l);
    pop2(b, a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    pc = &matrix_value(c, 0);
    for (k = 0; k < n; k++) {
	for (j = 0; j < m; j++) {
	    *pc++ = *pa++;
	}
	for (j = 0; j < l; j++) {
	    *pc++ = *pb++;
	}
    }

    return c;
}


/***************************************************************/
/***************************************************************/
Sex
Adel_row(Sex a, Sex b)
{
    long            i;
    long            k,
                    j;
    long            n,
                    m;

    Sex            *pa,
                   *pb;

    if (!is_matrix(a) || !is_fixnum(b))
	return aerror("del_row");
    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i < 0 || i >= n)
	return aerror("del_row: index out of range");
    push(a);
    b = new_matrix(n - 1, m);
    pop(a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    for (k = 0; k < n; k++) {
	if (k == i) {
	    pa += m;
	    continue;
	}
	for (j = 0; j < m; j++) {
	    *pb++ = *pa++;
	}
    }
    return b;
}


Sex
Adel_col(Sex a, Sex b)
{
    long            i;
    long            k,
                    j;
    long            n,
                    m;

    Sex            *pa,
                   *pb;

    if (!is_matrix(a) || !is_fixnum(b))
	return aerror("del_col");
    i = mkint(b) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i < 0 || i >= m)
	return aerror("del_col: index out of range");
    push(a);
    b = new_matrix(n, m - 1);
    pop(a);
    pa = &matrix_value(a, 0);
    pb = &matrix_value(b, 0);
    for (k = 0; k < n; k++) {
	for (j = 0; j < m; j++) {
	    if (j == i) {
		pa++;
		continue;
	    }
	    *pb++ = *pa++;
	}
    }
    return b;
}

Sex
Asubmatrix(Sex a, Sex b, Sex c)
{
    long            i = 0,
	j = 0;
    long            k,
                    l;
    long            n,
                    m;

    Sex            *pa,
                   *pc;

    if (!is_matrix(a) || !is_fixnum(b) || !is_fixnum(c))
	return aerror("submatrix");
    i = mkint(b) - 1;
    j = mkint(c) - 1;
    n = mkint(matrix_dimx(a));
    m = mkint(matrix_dimy(a));
    if (i < 0 || i >= n || j < 0 || j >= m)
	return aerror("submatrix: index out of range");
    push(a);
    c = new_matrix(n - 1, m - 1);
    pop(a);
    pa = &matrix_value(a, 0);
    pc = &matrix_value(c, 0);
    for (k = 0; k < n; k++) {
	if (k == i) {
	    pa += m;
	    continue;
	}
	for (l = 0; l < m; l++) {
	    if (l == j) {
		pa++;
		continue;
	    }
	    *pc++ = *pa++;
	}
    }
    return c;
}


Sex
Acofactor(Sex a, Sex b, Sex c)
{
    long            n;


    if (!is_matrix(a) || !is_fixnum(b) || !is_fixnum(c))
	return aerror("cofactor");

    n = mkint(b) + mkint(c);
    a = Asubmatrix(a, c, b);
    a = Ldeterminant(a);
    if (n % 2 != 0)
	a = gnegate(a);
    return a;
}


Sex
Arank(Sex a)
{
    long            n,
                    m;
    long            i,
                    j,
                    k;
    long            r;

    Sex             s;
    Sex            *ax,
                   *ay;

    m = mkint(matrix_dimx(a));
    n = mkint(matrix_dimy(a));
    a = copyMatrix(a);
    push3(a, nil, nil);
#define a stack[-2]
#define p stack[-1]
#define q stack[0]

    r = m;
    for (i = 0; i < m; i++) {
	for (j = i - m + r; j < n; j++) {
	    if (!zerop(matrix_value(a, i * n + j)))
		break;
	}
	if (j >= n) {
	    r--;
	    continue;
	}
	if (j != i) {
	    ax = &matrix_value(a, i + i * n);
	    ay = &matrix_value(a, j + i * n);
	    for (k = 0; k < m; k++) {
		q = *ax;
		*ax = *ay;
		*ay = q;
		ax += n;
		ay += n;
	    }
	}
	p = matrix_value(a, i * n + i);
	for (j = i + 1; j < m; j++) {
	    q = matrix_value(a, j * n + i);
	    if (zerop(q))
		continue;
	    q = gratio2(q, p);
	    for (k = i; k < n; k++) {
		s = gtimes2(q, matrix_value(a, i * n + k));
		s = gdiff2(matrix_value(a, j * n + k), s);
		matrix_value(a, j * n + k) = s;
	    }
	}
    }
#undef a
#undef p
#undef q
    popv(3);
    return mkfxat(r);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


setup_type const matrix_setup[] = {
#ifdef ASUKA
    {"trace", (lisp_proc *) Atrace, 1},
    {"determinant", (lisp_proc *) Ldeterminant, 1},
    {"transpose", (lisp_proc *) Ltranspose, 1},
    {"ident", (lisp_proc *) Lident, 1},
    {"diag", (lisp_proc *) Ldiag, 1},
    {"inverseMatrix", (lisp_proc *) inverseMatrix, 1},
    {"newvector", (lisp_proc *) Anewvector, 1},
    {"newmatrix", (lisp_proc *) Anewmatrix, 1},

    {"addMatrix", (lisp_proc *) addMatrix, 2},
    {"subMatrix", (lisp_proc *) subMatrix, 2},
    {"scaleMatrix", (lisp_proc *) scaleMatrix, 2},
    {"scaleaddMatrix", (lisp_proc *) scaleaddMatrix, 4},

    {"addVector", (lisp_proc *) addVector, 2},
    {"subVector", (lisp_proc *) subVector, 2},
    {"scaleVector", (lisp_proc *) scaleVector, 2},
    {"scaleaddVector", (lisp_proc *) scaleaddVector, 4},

    {"rowvector", (lisp_proc *) Arowvector, 2},
    {"colvector", (lisp_proc *) Acolvector, 2},
    {"setrow", (lisp_proc *) Asetrow, 3},
    {"setcol", (lisp_proc *) Asetcol, 3},
    {"swap_row", (lisp_proc *) Aswap_row, 3},
    {"swap_col", (lisp_proc *) Aswap_col, 3},
    {"scale_row", (lisp_proc *) Ascale_row, 3},
    {"scale_col", (lisp_proc *) Ascale_col, 3},
    {"muladd_row", (lisp_proc *) Amuladd_row, 4},
    {"muladd_col", (lisp_proc *) Amuladd_col, 4},
    {"scale_muladd_row", (lisp_proc *) Ascale_muladd_row, 5},
    {"scale_muladd_col", (lisp_proc *) Ascale_muladd_col, 5},
    {"ins_row", (lisp_proc *) Ains_row, 3},
    {"ins_col", (lisp_proc *) Ains_col, 3},
    {"merge_row", (lisp_proc *) Amerge_row, 2},
    {"merge_col", (lisp_proc *) Amerge_col, 2},
    {"del_row", (lisp_proc *) Adel_row, 2},
    {"del_col", (lisp_proc *) Adel_col, 2},
    {"submatrix", (lisp_proc *) Asubmatrix, 3},
    {"cofactor", (lisp_proc *) Acofactor, 3},
    {"rank", (lisp_proc *) Arank, 1},
#endif
    {NULL, 0, 0}
};
