/*
 * modular.c            (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * short int and modular arithmetics
 * $Header: /Users/kako/asuka/src/RCS/modular.c,v 2.12 2016/02/23 08:50:34 kako Exp $
 */

static char const rcsid[] =
    "$Id: modular.c,v 2.12 2016/02/23 08:50:34 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "entries.h"
#include "mathop.h"
#include "compare.h"


#ifdef _LP64
#define M_SQROOT	3037000499
#else
#define M_SQROOT	46341
#endif

#define MODULAR_PRINT
bool            modular_print = YES;

/***************************************************************/

void            print_modular(Sex u, putchar_fn * fn);
Sex             Lmkmodular(Sex a);
Sex             plus2im(Sex a, Sex b);
Sex             plus2bm(Sex a, Sex b);
Sex             plus2mi(Sex a, Sex b);
Sex             plus2mb(Sex a, Sex b);
Sex             plus2mr(Sex a, Sex b);
Sex             plus2rm(Sex a, Sex b);
Sex             plus2mm(Sex a, Sex b);
Sex             diff2im(Sex a, Sex b);
Sex             diff2bm(Sex a, Sex b);
Sex             diff2mi(Sex a, Sex b);
Sex             diff2mb(Sex a, Sex b);
Sex             diff2mr(Sex a, Sex b);
Sex             diff2rm(Sex a, Sex b);
Sex             diff2mm(Sex a, Sex b);
Sex             times2im(Sex a, Sex b);
Sex             times2bm(Sex a, Sex b);
Sex             times2mi(Sex a, Sex b);
Sex             times2mb(Sex a, Sex b);
Sex             times2mr(Sex a, Sex b);
Sex             times2rm(Sex a, Sex b);
Sex             times2mm(Sex a, Sex b);
Sex             quot2im(Sex a, Sex b);
Sex             quot2bm(Sex a, Sex b);
Sex             quot2mi(Sex a, Sex b);
Sex             quot2mb(Sex a, Sex b);
Sex             quot2mr(Sex a, Sex b);
Sex             quot2rm(Sex a, Sex b);
long            recipim(long n, long m);
Sex             recipm(Sex n);
Sex             quot2mm(Sex a, Sex b);
// Sex quot2mmx(Sex a, Sex b);
Sex             negatem(Sex a);
bool            numeq2im(Sex a, Sex b);
bool            numeq2bm(Sex a, Sex b);
bool            numeq2mm(Sex a, Sex b);
bool            numeq2mr(Sex a, Sex b);
bool            numeq2mc(Sex a, Sex b);
Sex             Lmodular_difference(Sex a, Sex b);
Sex             Lmodular_minus(Sex a);
Sex             Lmodular_number(Sex a);
Sex             Lmodular_plus(Sex a, Sex b);
Sex             Lmodular_reciprocal(Sex n);
Sex             Lmodular_times(Sex a, Sex b);
Sex             Lmodular_quotient(Sex a, Sex b);
Sex             Lmodular_expt(Sex a, Sex b);
Sex             Lset_small_modulus(Sex a);
Sex             Liadd1(Sex a);
Sex             Lidifference(Sex a, Sex b);
Sex             Ligreaterp(Sex a, Sex b);
Sex             Lilessp(Sex a, Sex b);
Sex             Ligeq(Sex a, Sex b);
Sex             Lileq(Sex a, Sex b);
Sex             Lilognot(Sex a);
Sex             Lilogand2(Sex a, Sex b);
Sex             Lilogor2(Sex a, Sex b);
Sex             Lilogxor2(Sex a, Sex b);
Sex             Limin2(Sex a, Sex b);
Sex             Limax2(Sex a, Sex b);
Sex             Liminus(Sex a);
Sex             Liminusp(Sex a);
Sex             Liplus2(Sex a, Sex b);
Sex             Liquotient(Sex a, Sex b);
Sex             Liremainder(Sex a, Sex b);
Sex             Lirightshift(Sex a, Sex b);
Sex             Lisub1(Sex a);
Sex             Litimes2(Sex a, Sex b);
Sex             Lionep(Sex a);
Sex             Lizerop(Sex a);

long            mult2mm(long a, long b, long m);
long            div2mm(long a, long b, long m);
long            pow2mm(long a, long b, long m);

Sex             Lmodular_value(Sex a);
Sex             Lmodular_modulus(Sex a);

extern Sex      recipiM(Sex a, Sex b);

/***************************************************************/
/***************************************************************/


void
print_modular(Sex u, putchar_fn * fn)
{
#ifdef MODULAR_PRINT
    if (modular_print) {
	fn('(');
	print_sex(mkfxat(modular_value(u)), fn);
	fn('%');
	print_sex(mkfxat(modular_modulus(u)), fn);
	fn(')');
    } else {
	print_sex(mkfxat(modular_value(u)), fn);
    }
#else
    print_sex(mkfxat(modular_value(u)), fn);
#endif
}


Sex
Lmodular(Sex a, Sex b)
{
    if (is_fixnum(b))
	return Lmodular2(a, mkint(b));
    return Lbmodular(a, b);
}

Sex
Lmodular2(Sex a, long m)
{

    Sex             c;
    long            n;

    stackcheck();
    if (is_fixnum(a)) {
	n = mkint(a) % m;
	if (n < 0)
	    n = n + m;
	c = new_modular();
	modular_value(c) = n;
	modular_modulus(c) = m;
	return c;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    a = remaind2(a, mkfxat(m));
	    n = mkint(a);
	    if (n < 0)
		n = n + m;
	    c = new_modular();
	    modular_value(c) = n;
	    modular_modulus(c) = m;
	    return c;
	case N_MODULAR:
	    n = modular_value(a);
	    n = n % m;
	    a = new_modular();
	    modular_value(a) = n;
	    modular_modulus(a) = m;
	    return a;
	case N_RATIONAL:
	    {
		long            k;
		push(a);
		a = rational_num(a);
		a = remaind2(a, mkfxat(m));
		n = mkint(a);
		if (n < 0)
		    n = n + m;
		pop(a);
		a = rational_den(a);
		a = remaind2(a, mkfxat(m));
		k = mkint(a);
		k = recipim(k, m);
		k = mult2mm(k, n, m);
		a = new_modular();
		modular_value(a) = k;
		modular_modulus(a) = m;
		return a;
	    }
	case N_COMPLEX:
	    {
		push(a);
		a = Lmodular(complex_real(a), mkfxat(m));
		push(a);
		a = Lmodular(complex_imag(stack[-1]), mkfxat(m));
		push(a);
		c = new_complex();
		complex_imag(c) = stack[0];
		complex_real(c) = stack[-1];
		popv(3);
		return c;
	    }
	}
    }
    return error1("bad argument to modular ", a);
}



Sex
Lmkmodular(Sex a)
{

    Sex             c;
    long            n;
    long            m = current_modulus;;

    stackcheck();
    if (is_fixnum(a)) {
	n = mkint(a) % m;
	if (n < 0)
	    n = n + m;
	c = new_modular();
	modular_value(c) = n;
	modular_modulus(c) = m;
	return c;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	    a = remaind2(a, mkfxat(m));
	    n = mkint(a);
	    c = new_modular();
	    if (n < 0)
		n = n + m;
	    modular_value(c) = n;
	    modular_modulus(c) = m;
	    return c;
	case N_MODULAR:
	    n = modular_value(a);
	    n = n % m;
	    a = new_modular();
	    modular_value(a) = n;
	    modular_modulus(a) = m;
	    return a;
	case N_RATIONAL:
	    {
		long            k;
		push(a);
		a = rational_num(a);
		a = remaind2(a, mkfxat(m));
		n = mkint(a);
		if (n < 0)
		    n = n + m;
		pop(a);
		a = rational_den(a);
		a = remaind2(a, mkfxat(m));
		k = mkint(a);
		k = recipim(k, m);
		k = mult2mm(n, k, m);
		a = new_modular();
		modular_value(a) = k;
		modular_modulus(a) = m;
		return a;
	    }
	case N_COMPLEX:
	    {
		push(a);
		a = Lmkmodular(complex_real(a));
		push(a);
		a = Lmkmodular(complex_imag(stack[-1]));
		push(a);
		c = new_complex();
		complex_imag(c) = stack[0];
		complex_real(c) = stack[-1];
		popv(3);
		return c;
	    }
	}
    }
    return error1("bad argument to mkmodular ", a);
}

Sex
Lunmodular(Sex a)
{

    if (is_modular(a)) {
	long            m,
	                n,
	                mh;
	m = modular_modulus(a);
	mh = m / 2;
	n = modular_value(a);
	if (n > mh)
	    n -= m;
	return mkfxat(n);
    } else if (is_bmodular(a)) {
	Sex             m,
	                n,
	                mh;
	m = bmodular_modulus(a);
	n = bmodular_value(a);
	push2(m, n);
	mh = quot2(m, mkfxat(2));
	pop2(n, m);
	if (lessp2(mh, n))
	    diff2(n, m);
	return n;
    }
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2im(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(b);
    n = (mkint(a) + modular_value(b)) % m;
    if (n < 0)
	n = n + m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;
}


Sex
plus2bm(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(b);
    n = modular_value(b);
    a = remaind2(a, mkfxat(m));
    n = mkint(a) + n + m;
    n = n % m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;

}


Sex
plus2mi(Sex a, Sex b)
{
    return plus2im(b, a);
}


Sex
plus2mb(Sex a, Sex b)
{
    return plus2bm(b, a);
}

Sex
plus2mr(Sex a, Sex b)
{

    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return plus2mm(a, b);
}

Sex
plus2rm(Sex a, Sex b)
{
    return plus2mr(b, a);
}

Sex
plus2mm(Sex a, Sex b)
{
    long            m,
                    k;

    m = modular_modulus(a);
    if (m != modular_modulus(b))
	return math_er(a, b);
    k = modular_value(a) + modular_value(b);
    if (k >= m)
	k = k - m;
    a = new_modular();
    modular_value(a) = k;
    modular_modulus(a) = m;
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
diff2im(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(b);
    n = (mkint(a) - modular_value(b)) % m;
    if (n < 0)
	n = n + m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;
}


Sex
diff2bm(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(b);
    n = modular_value(b);
    a = remaind2(a, mkfxat(m));
    n = mkint(a) - n;
    if (n < 0)
	n = n + m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;
}


Sex
diff2mi(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(a);
    n = (-mkint(b) + modular_value(a)) % m;
    if (n < 0)
	n = n + m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;
}


Sex
diff2mb(Sex a, Sex b)
{
    long            n,
                    m;
    m = modular_modulus(a);
    n = modular_value(a);
    b = remaind2(b, mkfxat(m));
    n = n - mkint(b);
    if (n < 0)
	n = n + m;
    a = new_modular();
    modular_value(a) = n;
    modular_modulus(a) = m;
    return a;
}

Sex
diff2mr(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return diff2mm(a, b);
}

Sex
diff2rm(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return diff2mm(a, b);
}

Sex
diff2mm(Sex a, Sex b)
{
    long            m,
                    k;

    m = modular_modulus(a);
    if (m != modular_modulus(b))
	return math_er(a, b);
    k = modular_value(a) - modular_value(b);
    if (k < 0)
	k = k + m;
    a = new_modular();
    modular_value(a) = k;
    modular_modulus(a) = m;
    return a;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
times2im(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return times2mm(a, b);
}


Sex
times2bm(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return times2mm(a, b);
}


Sex
times2mi(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return times2mm(a, b);
}


Sex
times2mb(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return times2mm(a, b);
}

Sex
times2mr(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return times2mm(a, b);
}

Sex
times2rm(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return times2mm(a, b);
}


long
mult2mm(long a, long b, long m)
{
    unsigned long   r;

    if (m <= M_SQROOT)
	r = (a * b) % m;
    else {
	unsigned long   h,
	                l,
	                q;
	h = umul(&l, a, b);
	q = udiv(&r, h, l, m);
    }
    return (long) r;
}

long
div2mm(long a, long b, long m)
{
    b = recipim(b, m);
    return mult2mm(a, b, m);
}

long
pow2mm(long a, long b, long m)
{
    unsigned long   p;
    unsigned long   r;
    unsigned long   h,
                    l;


    p = a;
    if (b == 0)
	return 1;
    r = 1;
    if (m < M_SQROOT) {
	while (b > 1) {
	    if (b & 1)
		r = (r * a) % m;
	    a = (a * a) % m;
	    b >>= 1;
	}
	r = (r * a) % m;
	return r;
    }
    while (b > 1) {
	if (b & 1) {
	    h = umul(&l, r, p);
	    udiv(&r, h, l, m);
	}
	h = umul(&l, p, p);
	udiv(&p, h, l, m);
	b >>= 1;
    }
    h = umul(&l, r, p);
    udiv(&r, h, l, m);
    return r;
}

Sex
times2mm(Sex a, Sex b)
{
    unsigned long   r,
                    cm;
    long            aa,
                    bb;
    cm = modular_modulus(a);
    aa = modular_value(a);
    bb = modular_value(b);
    r = mult2mm(aa, bb, cm);
    a = new_modular();
    modular_value(a) = r;
    modular_modulus(a) = cm;
    return a;
}



/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
quot2im(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return quot2mm(a, b);
}


Sex
quot2bm(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return quot2mm(a, b);
}


Sex
quot2mi(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return quot2mm(a, b);
}


Sex
quot2mb(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return quot2mm(a, b);
}

Sex
quot2mr(Sex a, Sex b)
{
    push(a);
    b = Lmodular2(b, modular_modulus(a));
    pop(a);
    return quot2mm(a, b);
}

Sex
quot2rm(Sex a, Sex b)
{
    push(b);
    a = Lmodular2(a, modular_modulus(b));
    pop(b);
    return quot2mm(a, b);
}

long
recipim(long n, long m)
{
    long            a,
                    b,
                    x,
                    y;
    long            w;
    long            t;

    a = m;
    b = n;
    x = 0;
    y = 1;
    while (b != 1) {
	if (b == 0)
	    return aerror1("recipim ", mkfxat(n));
	w = a / b;
	t = b;
	b = a % b;
	a = t;
	t = y;
	y = x - y * w;
	x = t;
    }
    if (y < 0)
	y += m;
    return y;
}


Sex
recipm(Sex n)
{
#if 0
    if (!is_number(value(star_modulus)))
	return aerror1("modular-reciprocal", value(star_modulus));
#endif
    push4(value(star_modulus), n, mkfxat(0), mkfxat(1));
    push2(nil, nil);
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define y stack[-2]
#define w stack[-1]
#define t stack[0]
    a = mkfxat(current_modulus);
    while (b != mkfxat(1)) {
	if (b == mkfxat(0))
	    return aerror1("modular-reciprocal", n);
	w = Ldivide(a, b);
	a = b;
	b = cdr(w);		/* 
				 * remainder 
				 */
	t = y;
	w = times2jj(y, car(w));
	y = diff2jj(x, w);
	x = t;
    }
    if (minusp(y))
	// y = plus2jj(y, value(star_modulus));
	y = plus2jj(y, mkfxat(current_modulus));
    n = y;
#undef a
#undef b
#undef x
#undef y
#undef w
#undef t
    popv(6);
    return n;
}

#if 0
Sex
quot2mm(Sex a, Sex b)
{
    Sex             c;
    lisp_printf("quot2mm (%p, ", a);
    lisp_printf(" %p) = ", b);
    push2(a, b);
    c = quot2mmx(a, b);
    pop2(b, a);
    lisp_printf(" %p\n", c);
    return c;
}
#endif

Sex
quot2mm(Sex a, Sex b)
{
    long            n;
    long            aa;
    unsigned long   r;
    unsigned long   cm;

    cm = modular_modulus(a);

    aa = modular_value(a);
    n = modular_value(b);
    n = recipim(n, cm);


    if (cm <= M_SQROOT)
	r = (n * aa) % cm;
    else {
	unsigned long   h,
	                l,
	                q;
	h = umul(&l, n, aa);
	q = udiv(&r, h, l, cm);
    }
    a = new_modular();
    modular_value(a) = r;
    modular_modulus(a) = cm;
    return a;

}

Sex
negatem(Sex a)
{
    long            cm = modular_modulus(a);
    long            n = modular_value(a);

    a = new_modular();
    modular_value(a) = cm - n;
    modular_modulus(a) = cm;
    return a;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


bool
numeq2im(Sex a, Sex b)
{
    long            n = mkint(a);
    n = n % modular_modulus(b);
    if (n < 0)
	n = n + modular_modulus(b);
    return n == modular_value(b);
}

bool
numeq2bm(Sex a, Sex b)
{
    long            n = modular_value(b);
    long            cm = modular_modulus(b);
    long            m;
    a = remaind2(a, mkfxat(cm));
    m = mkint(a);
    if (m < 0)
	m = m + cm;
    return n == m;
}

bool
numeq2mm(Sex a, Sex b)
{
    return modular_value(a) == modular_value(b);
}

bool
numeq2mr(Sex a, Sex b)
{
    long            cm = modular_modulus(a);
    long            n = modular_value(a);
    b = Lmodular(b, mkfxat(cm));
    return modular_value(b) == n;
}

bool
numeq2mc(Sex a, Sex b)
{
    bool            r;
    push(a);
    r = zerop(complex_imag(b));
    pop(a);
    if (!r)
	return NO;
    return numeq2(a, complex_real(b));
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



/*
 *
 */

Sex
Lmodular_difference(Sex a, Sex b)
{
    long            r;
    r = mkint(a) - mkint(b);
    if (r < 0)
	r += current_modulus;
    return mkfxat(r);
}

Sex
Lmodular_minus(Sex a)
{
    if (a != mkfxat(0)) {
	long            r = current_modulus - mkint(a);
	a = mkfxat(r);
    }
    return a;
}

Sex
Lmodular_number(Sex a)
{
    long            r;
    a = remaind2(a, mkfxat(current_modulus));
    r = mkint(a);
    if (r < 0)
	r += current_modulus;
    return mkfxat(r);
}

Sex
Lmodular_plus(Sex a, Sex b)
{
    long            r;
    r = mkint(a) + mkint(b);
    if (r >= current_modulus)
	r -= current_modulus;
    return mkfxat(r);
}

Sex
Lmodular_reciprocal(Sex n)
{
    long            m;

    m = recipim(mkint(n), current_modulus);
    return mkfxat(m);
}



Sex
Lmodular_times(Sex a, Sex b)
{
    unsigned long   r,
                    cm;
    long            aa,
                    bb;
    cm = (unsigned long) current_modulus;
    aa = mkint(a);
    bb = mkint(b);
    if (cm <= M_SQROOT)
	r = (aa * bb) % cm;
    else {
	unsigned long   h,
	                l,
	                q;
	h = umul(&l, aa, bb);
	q = udiv(&r, h, l, cm);
    }
    return mkfxat(r);
}

Sex
Lmodular_quotient(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = Lmodular_reciprocal(b);
    pop(a);
    return Lmodular_times(a, b);
}


Sex
Lmodular_expt(Sex a, Sex b)
{
    unsigned long   x,
                    r,
                    p;
    unsigned long   h,
                    l;

    unsigned long   m = current_modulus;


    x = mkint(b);
    if (x == 0)
	return mkfxat(1);
    p = mkint(a);
    r = 1;
    if (m < M_SQROOT) {
	while (x > 1) {
	    if (x & 1)
		r = (r * p) % m;
	    p = (p * p) % m;
	    x >>= 1;
	}
	r = (r * p) % m;
	return mkfxat(r);
    }
    while (x > 1) {
	if (x & 1) {
	    h = umul(&l, r, p);
	    udiv(&r, h, l, current_modulus);
	}
	h = umul(&l, p, p);
	udiv(&p, h, l, current_modulus);
	x >>= 1;
    }
    h = umul(&l, r, p);
    udiv(&r, h, l, current_modulus);
    return mkfxat(r);
}


Sex
Lset_small_modulus(Sex a)
{
    long            r,
                    old = current_modulus;
    if (!is_fixnum(a))
	return aerror1("set-small-modulus", a);
    r = mkint(a);
    current_modulus = r;
    return mkfxat(old);
}

Sex
reciprocal2(Sex a, Sex b)
{
    long            m;
    if (is_fixnum(b)) {
	m = mkint(b);
	if (!is_fixnum(a))
	    a = remaind2(a, b);
	m = recipim(mkint(a), m);
	return mkfxat(m);
    }
    return recipiM(a, b);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Liadd1(Sex a)
{
    if (!is_fixnum(a))
	return aerror1("iadd1", a);
    return (Sex) ((long) a + 0x10);
}

Sex
Lidifference(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("idifference", a, b);
    return (Sex) ((long) a - (long) b + TAG_FIXNUM);
}

Sex
Ligreaterp(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("igreaterp", a, b);
    return Bool_to_Sex(a > b);
}

Sex
Lilessp(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("ilessp", a, b);
    return Bool_to_Sex(a < b);
}

Sex
Ligeq(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("igeq", a, b);
    return Bool_to_Sex(a >= b);
}

Sex
Lileq(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("ileq", a, b);
    return Bool_to_Sex(a <= b);
}


Sex
Lilognot(Sex a)
{
    if (!is_fixnum(a))
	return aerror1("ilognot", a);
    return (~(a - TAG_FIXNUM) + TAG_FIXNUM);
}

Sex
Lilogand2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("ilogand", a, b);
    return (a & b);
}

Sex
Lilogor2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("ilogor", a, b);
    return (a | b);
}

Sex
Lilogxor2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("ilogxor", a, b);
    return ((a ^ b) + TAG_FIXNUM);
}

Sex
Limin2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("imin", a, b);
    return (a < b ? a : b);
}

Sex
Limax2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("imax", a, b);
    return (a > b ? a : b);
}

Sex
Liminus(Sex a)
{
    if (!is_fixnum(a))
	return aerror1("iminus", a);
    return (Sex) (2 * TAG_FIXNUM - (long) a);
}

Sex
Liminusp(Sex a)
{
    return Bool_to_Sex((long) a < (long) mkfxat(0));
}


Sex
Liplus2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("iplus2", a, b);
    return (Sex) ((long) a + (long) b - TAG_FIXNUM);
}

Sex
Liquotient(Sex a, Sex b)
{
    long            aa,
                    bb,
                    c;
    if (!is_fixnum(a) || !is_fixnum(b) || b == mkfxat(0))
	return aerror2("iquotient", a, b);
    /*
     * C does not define the exact behaviour of /, % on -ve args 
     */
    aa = mkint(a);
    bb = mkint(b);
    c = aa % bb;
    if (aa < 0) {
	if (c > 0)
	    c -= bb;
    } else if (c < 0)
	c += bb;
    return mkfxat((aa - c) / bb);
}

Sex
Liremainder(Sex a, Sex b)
{
    long            aa,
                    bb,
                    c;
    if (!is_fixnum(a) || !is_fixnum(b) || b == mkfxat(0))
	return aerror2("iremainder", a, b);
    /*
     * C does not define the exact behaviour of /, % on -ve args 
     */
    aa = mkint(a);
    bb = mkint(b);
    c = aa % bb;
    if (aa < 0) {
	if (c > 0)
	    c -= bb;
    } else if (c < 0)
	c += bb;
    return mkfxat(c);
}

Sex
Lirightshift(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("irightshift", a, b);
    return mkfxat(mkint(a) >> mkint(b));
}

Sex
Lisub1(Sex a)
{
    if (!is_fixnum(a))
	return aerror1("isub1", a);
    return (Sex) ((long) a - 0x10);
}


Sex
Litimes2(Sex a, Sex b)
{
    if (!is_fixnum(a) || !is_fixnum(b))
	return aerror2("itimes2", a, b);
    return mkfxat(mkint(a) * mkint(b));
}

Sex
Lionep(Sex a)
{
    return Bool_to_Sex((long) a == (long) mkfxat(1));
}

Sex
Lizerop(Sex a)
{
    return Bool_to_Sex((long) a == (long) mkfxat(0));
}


/***************************************************************/
/***************************************************************/

Sex
Lmodular_value(Sex a)
{
    if (is_modular(a))
	return mkfxat(modular_value(a));
    if (is_bmodular(a))
	return bmodular_value(a);
    return nil;
}

Sex
Lmodular_modulus(Sex a)
{
    if (is_modular(a))
	return mkfxat(modular_modulus(a));
    if (is_bmodular(a))
	return bmodular_modulus(a);
    return nil;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

setup_type const modular_setup[] = {
    {"modular-difference", (lisp_proc *) Lmodular_difference, 2},
    {"modular-minus", (lisp_proc *) Lmodular_minus, 1},
    {"modular-number", (lisp_proc *) Lmodular_number, 1},
    {"modular-plus", (lisp_proc *) Lmodular_plus, 2},
    {"modular-quotient", (lisp_proc *) Lmodular_quotient, 2},
    {"modular-reciprocal", (lisp_proc *) Lmodular_reciprocal, 1},
    {"modular-times", (lisp_proc *) Lmodular_times, 2},
    {"modular-expt", (lisp_proc *) Lmodular_expt, 2},
    {"set-small-modulus", (lisp_proc *) Lset_small_modulus, 1},
    {"iadd1", (lisp_proc *) Liadd1, 1},
    {"idifference", (lisp_proc *) Lidifference, 2},
    {"igeq", (lisp_proc *) Ligeq, 2},
    {"igreaterp", (lisp_proc *) Ligreaterp, 2},
    {"ileq", (lisp_proc *) Lileq, 2},
    {"ilessp", (lisp_proc *) Lilessp, 2},
    {"ilogand2", (lisp_proc *) Lilogand2, 2},
    {"ilogor2", (lisp_proc *) Lilogor2, 2},
    {"ilogxor2", (lisp_proc *) Lilogxor2, 2},
    {"imax2", (lisp_proc *) Limax2, 2},
    {"imin2", (lisp_proc *) Limin2, 2},
    {"iminus", (lisp_proc *) Liminus, 1},
    {"iminusp", (lisp_proc *) Liminusp, 1},
    {"iplus2", (lisp_proc *) Liplus2, 2},
    {"iquotient", (lisp_proc *) Liquotient, 2},
    {"iremainder", (lisp_proc *) Liremainder, 2},
    {"irightshift", (lisp_proc *) Lirightshift, 2},
    {"isub1", (lisp_proc *) Lisub1, 1},
    {"itimes", (lisp_proc *) Litimes2, 2},
    {"itimes2", (lisp_proc *) Litimes2, 2},
    {"ionep", (lisp_proc *) Lionep, 1},
    {"izerop", (lisp_proc *) Lizerop, 1},
    {"modular_value", (lisp_proc *) Lmodular_value, 1},
    {"modular_modulus", (lisp_proc *) Lmodular_modulus, 1},
    {NULL, 0, 0}
};
