/*
 * polynomial.c                (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/src/RCS/polynomial.c,v 2.42 2017/04/20 08:14:14 kako Exp $
 */

static char const rcsid[] =
    "$Id: polynomial.c,v 2.42 2017/04/20 08:14:14 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "macros.h"
#include "garith.h"
#include "tri.h"
#include "polynomial.h"
#include "print.h"


// #define DEBUG

#define coeff(a)        car(a)
#define power(a)        cdr(a)
#define lterm(a)        car(a)
#define mkterm(c,p)        cons(c, p)
#define reduct(a)        cdr(a)


extern Sex      vect_subst(Sex p, Sex v, Sex b);
extern Sex      matrix_subst(Sex p, Sex v, Sex b);
extern Sex      list_subst(Sex p, Sex v, Sex b);
extern Sex      array_subst(Sex p, Sex v, Sex b);
extern Sex      simple_array_subst(Sex p, Sex v, Sex b);
extern Sex      vect_reval(Sex p);
extern Sex      matrix_reval(Sex p);
extern Sex      list_reval(Sex p);
extern Sex      array_reval(Sex p);
extern Sex      simple_array_reval(Sex p);

Sex             Adiff(Sex p, Sex v);
Sex             poly_diff(Sex p, Sex v);
Sex             dipoly_diff(Sex p, Sex v);
Sex             rpoly_diff(Sex p, Sex v);
extern Sex      vect_diff(Sex p, Sex v);
extern Sex      matrix_diff(Sex p, Sex v);
extern Sex      list_diff(Sex p, Sex v);
extern Sex      array_diff(Sex p, Sex v);
extern Sex      simple_array_diff(Sex p, Sex v);
extern Sex      prefix_diff(Sex p, Sex v);

extern long     term_order;
extern Sex      dipoly_kernel_order;

extern Sex      Lsetorder_lex(Sex a);
extern Sex      Lsetorder_tlex(Sex a);
extern Sex      Lsetorder_trevlex(Sex a);
extern Sex      Lsetorder_gradings(Sex a, Sex b);

extern Sex      Aexpti(Sex a, long n);


long            compare_term(Sex a, Sex b);
long            compare_term_lex(Sex a, Sex b);
long            compare_term_tlex(Sex a, Sex b);
long            compare_term_trevlex(Sex a, Sex b);
long            compare_term_gradings(Sex a, Sex b);

extern long     gradings_matrix_m;
extern long     gradings_matrix_n;
extern long     gradings_matrix[];


Sex             Lresultant(Sex a, Sex b, Sex x);
static Sex      Lresultant_det(Sex a, Sex b, Sex x);
static Sex      Lresultant_poly(Sex a, Sex b, Sex x);
static Sex      Lresultant_bezout1(Sex a, Sex b, Sex x);
static Sex      Lresultant_bezout2(Sex a, Sex b, Sex x);
static Sex      ext_add(Sex u, Sex v);
static Sex      ext_mult(Sex u, Sex v);
static Sex      sf2exb1(Sex u);
static Sex      sf2exb2(Sex u, Sex v);
static Sex      ext_ordexp(Sex u, Sex v);
static Sex      ext_ordexn(Sex u, Sex v);

extern Sex      Llist2(Sex a, Sex b);
extern Sex      Lappend(Sex a, Sex b);
extern Sex      Lreverse(Sex a);
extern Sex      equal(Sex a, Sex b);
extern Sex      Ainit(Sex a, Sex x);
extern Sex      Adeg(Sex a, Sex v);
extern Sex      Areduct(Sex a, Sex v);

extern Sex      Acoeff(Sex a, Sex v);
extern Sex      new_matrix(long n, long m);
extern Sex      Ldeterminant(Sex a);

/*
 * #define ONEP(x) onep(x) #define ZEROP(x) zerop(x) 
 */
#define ONEP(x) (x == mkfxat(1))
// #define ZEROP(x) (x == mkfxat(0))
#define ZEROP(x)        zerop(x)



/***************************************************************/
/***************************************************************/
/***************************************************************/


/*
 * typedef struct var_list { Sex *var; Sex *pow; struct var_list *next; }
 * var_list; 
 */

bool
minuspP(Sex a)
{

    if (is_fixnum(a)) {
	return ((long) a < 0);
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_BIGNUM:
	    return is_bigneg(a) ? YES : NO;
	case N_RATIONAL:
	    return minusp(rational_num(a));
	case N_FLOAT:
	    return (float_val(a) < 0.0);
	case N_BFLOAT:
	    return minusp(bfloat_man(a));
	}
    } else if (is_poly(a)) {
	a = poly_pol(a);
	if (reduct(a) != nil)
	    return NO;
	return minuspP(coeff(lterm(a)));
    }
    return NO;
}



void
print_poly_aux(Sex a, var_list * v, putchar_fn * fn, long first)
{
    Sex             x,
                    p;
    struct var_list vv,
                   *vp;

    x = poly_var(a);
    p = poly_pol(a);
    stackcheck();
    push4(x, p, nil, nil);
#define a stack[-1]
#define c stack[0]
#define p stack[-2]
#define x stack[-3]

    vv.var = &x;
    vv.next = v;
    while (p != nil) {		/* 
				 * ( (coef exp) ...)  
				 */
	a = lterm(p);
	p = reduct(p);
	c = power(a);
	vv.pow = &c;
	a = coeff(a);
	if (is_poly(a)) {
	    print_poly_aux(a, &vv, fn, first);
	} else {
	    if (!first && !minusp(a))
		fn('+');
	    if (a == mkfxat(1)) {
		for (vp = &vv; vp != NULL; vp = vp->next) {
		    if (*(vp->pow) != mkfxat(0))
			break;
		}
		if (vp == NULL)
		    fn('1');
	    } else
		print_sex(a, fn);
	    for (vp = &vv; vp != NULL; vp = vp->next) {
		if (*(vp->pow) != mkfxat(0)) {
		    print_sex(*(vp->var), fn);
		    if (*(vp->pow) != mkfxat(1)) {
			fn('^');
			print_sex(*(vp->pow), fn);
		    }
		}
	    }
	}
	first = NO;
    }
#undef a
#undef c
#undef p
#undef x
    popv(4);
}

void
print_polynomial(Sex a, putchar_fn * fn)
{
    Sex             x,
                    p;
    long            first = YES;

    x = poly_var(a);
    p = poly_pol(a);
    stackcheck();
    push4(x, p, nil, nil);
#define a stack[-1]
#define c stack[0]
#define p stack[-2]
#define x stack[-3]

    while (p != nil) {		/* 
				 * ( (coef exp) ...)  
				 */
      again:
	a = lterm(p);
	p = reduct(p);
	c = power(a);
	a = coeff(a);
	if (is_poly(a)) {
	    if (p == nil && c == mkfxat(0)) {
		x = poly_var(a);
		p = poly_pol(a);
		goto again;
	    }
	    if (!first && !minuspP(a))
		fn('+');
	    if (c == mkfxat(0)) {
		fn('(');
		print_sex(a, fn);
		fn(')');
	    } else {
		if (reduct(poly_pol(a)) == nil) {
		    print_sex(a, fn);
		} else {
		    fn('(');
		    print_sex(a, fn);
		    fn(')');
		}
		fn('*');
		print_sex(x, fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	} else {
	    if (!first && !minuspP(a))
		fn('+');
	    if (a == mkfxat(-1) && c != mkfxat(0)) {
		fn('-');
	    } else if (a != mkfxat(1) && c != mkfxat(0)) {
		print_sex(a, fn);
		fn('*');
	    } else if (c == mkfxat(0)) {
		print_sex(a, fn);
	    }
	    if (c != mkfxat(0)) {
		print_sex(x, fn);
		if (c != mkfxat(1)) {
		    fn('^');
		    print_sex(c, fn);
		}
	    }
	}
	first = NO;
    }
#undef a
#undef c
#undef p
#undef x
    popv(4);
}


void
print_rpoly(Sex a, putchar_fn * fn)
{
    push(a);
    fn('(');
    print_sex(rpoly_num(a), fn);
    fn(')');
    fn('/');
    fn('(');
    pop(a);
    print_sex(rpoly_den(a), fn);
    fn(')');
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
make_poly(Sex x, Sex p)
{
    Sex             d;

    if (p == nil || zerop(p))
	return mkfxat(0);
    d = lterm(p);
    if (reduct(p) == nil && power(d) == mkfxat(0))
	return coeff(d);

    stackcheck();
    push2(x, p);
    d = getvector(TAG_VECTOR, T_POLY, sizeof(Polynomial));
    pop2(p, x);
    poly_var(d) = x;
    poly_pol(d) = p;
    return d;
}

Sex
make_polynomial(Sex a, Sex b, Sex c)
{
    Sex             d;

    if (c == mkfxat(0))
	return b;
    stackcheck();
    push(a);
    b = mkterm(b, c);
    b = cons(b, nil);
    push(b);
    d = getvector(TAG_VECTOR, T_POLY, sizeof(Polynomial));
    pop2(b, a);
    poly_var(d) = a;
    poly_pol(d) = b;
    return d;
}

Sex
symbol_to_polynomial(Sex a)
{
    Sex             b,
                    d;

    if (is_poly(a))
	return a;
    ASSERT(is_symbol(a));
    stackcheck();
    push(a);
    b = mkterm(mkfxat(1), mkfxat(1));
    b = cons(b, nil);
    push(b);
    d = getvector(TAG_VECTOR, T_POLY, sizeof(Polynomial));
    pop2(b, a);
    poly_var(d) = a;
    poly_pol(d) = b;
    return d;
}

Sex
make_rpoly(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = getvector(TAG_VECTOR, T_RPOLY, sizeof(Rational_Polynomial));
    pop2(b, a);
    rpoly_num(c) = a;
    rpoly_den(c) = b;
    return c;
}

/***************************************************************/
/***************************************************************/

Sex
Apoly_spoly(Sex a, Sex b)
{
    long            o;
    Sex             as,
                    bs;
    Sex             s;
    int             na,
                    nb;
    int             nn;

    as = a;
    bs = b;

    stackcheck();
    push4(a, b, nil, nil);
    push3(as, bs, nil);
#define a stack[-6]
#define b stack[-5]
#define va stack[-4]
#define vb stack[-3]
#define as stack[-2]
#define bs stack[-1]
#define g stack[-0]

    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0) {
	    s = cons(poly_var(b), mkfxat(0));
	    va = cons(s, va);
	    s = cons(poly_var(b), power(lterm(poly_pol(b))));
	    vb = cons(s, vb);
	    b = coeff(lterm(poly_pol(b)));
	} else if (o < 0) {
	    s = cons(poly_var(a), power(lterm(poly_pol(a))));
	    va = cons(s, va);
	    s = cons(poly_var(a), mkfxat(0));
	    vb = cons(s, vb);
	    a = coeff(lterm(poly_pol(a)));
	} else {
	    s = cons(poly_var(a), power(lterm(poly_pol(a))));
	    va = cons(s, va);
	    s = cons(poly_var(b), power(lterm(poly_pol(b))));
	    vb = cons(s, vb);
	    a = coeff(lterm(poly_pol(a)));
	    b = coeff(lterm(poly_pol(b)));
	}
    }

    while (is_poly(a)) {
	s = cons(poly_var(a), power(lterm(poly_pol(a))));
	va = cons(s, va);
	s = cons(poly_var(a), mkfxat(0));
	vb = cons(s, vb);
	a = coeff(lterm(poly_pol(a)));
    }
    while (is_poly(b)) {
	s = cons(poly_var(b), mkfxat(0));
	va = cons(s, va);
	s = cons(poly_var(b), power(lterm(poly_pol(b))));
	vb = cons(s, vb);
	b = coeff(lterm(poly_pol(b)));
    }
    g = gcd(a, b);
    a = quot2(a, g);
    b = quot2(b, g);
    while (va != nil) {
	na = mkint(cdar(va));
	nb = mkint(cdar(vb));
	nn = na;
	if (nb > nn)
	    nn = nb;
	if (nn - nb != 0) {
	    a = mkterm(a, mkfxat(nn - nb));
	    a = cons(a, nil);
	    a = make_poly(caar(va), a);
	}
	if (nn - na != 0) {
	    b = mkterm(b, mkfxat(nn - na));
	    b = cons(b, nil);
	    b = make_poly(caar(vb), b);
	}
	va = cdr(va);
	vb = cdr(vb);
    }


    g = gtimes2(as, b);


    b = gtimes2(bs, a);
    a = gdiff2(g, b);
#undef a
#undef b
#undef va
#undef vb
#undef as
#undef bs
#undef g
    a = stack[-6];
    popv(7);
    return a;
}


Sex
Apoly_reduce(Sex a, Sex b)
{
    Sex             bs;
    bs = b;

    stackcheck();
    push4(a, b, bs, nil);
#define a stack[-3]
#define b stack[-2]
#define bs stack[-1]
#define c stack[-0]
    while (bs != nil) {
	c = car(bs);
	bs = cdr(bs);
#ifdef DEBUG
	lisp_printf("poly_reduce1: %p", a);
	lisp_printf(", %p", c);
#endif
	c = Apoly_reduce1(a, c);
#ifdef DEBUG
	lisp_printf(" ==> %p\n", c);
#endif
	if (a != c) {
	    a = c;
	    bs = b;
	}
    }
#undef a
    a = stack[-3];
#undef b
#undef bs
#undef c
    popv(4);
    return a;
}

Sex
Apoly_reduce1(Sex a, Sex b)
{
    long            o;
    Sex             as,
                    bs;
    Sex             c;
    Sex             s;
    int             na,
                    nb;
    int             nn;

    as = a;
    bs = b;

    stackcheck();
    push3(a, b, nil);
    push3(as, bs, nil);
#define a stack[-5]
#define b stack[-4]
#define va stack[-3]
#define as stack[-2]
#define bs stack[-1]
#define g stack[-0]

    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));

	if (o > 0) {
	    c = as;
	    popv(6);
	    return c;
	} else if (o < 0) {
	    s = cons(poly_var(a), power(lterm(poly_pol(a))));
	    va = cons(s, va);
	    a = coeff(lterm(poly_pol(a)));
	} else {
	    na = mkint(power(lterm(poly_pol(a))));
	    nb = mkint(power(lterm(poly_pol(b))));
	    if (na < nb) {
		c = as;
		popv(6);
		return c;
	    }
	    if (na != nb) {
		s = cons(poly_var(a), mkfxat(na - nb));
		va = cons(s, va);
	    }
	    a = coeff(lterm(poly_pol(a)));
	    b = coeff(lterm(poly_pol(b)));
	}
    }

    while (is_poly(a)) {
	s = cons(poly_var(a), power(lterm(poly_pol(a))));
	va = cons(s, va);
	a = coeff(lterm(poly_pol(a)));
    }
    if (is_poly(b)) {
	c = as;
	popv(6);
	return c;

    }
    g = gcd(a, b);
    a = quot2(a, g);
    b = quot2(b, g);
    a = ratio2(a, b);
    while (va != nil) {
	a = mkterm(a, cdar(va));
	a = cons(a, nil);
	a = make_poly(caar(va), a);
	va = cdr(va);
    }

    b = gtimes2(bs, a);
    a = gdiff2(as, b);
#undef a
#undef b
#undef va
#undef as
#undef bs
#undef g
    c = stack[-5];
    popv(6);
    return c;
}


Sex
Apoly_full_reduce(Sex a, Sex b)
{
    return Apoly_reduce(a, b);
}

Sex
Apoly_full_reduce1(Sex a, Sex b)
{
    return Apoly_reduce1(a, b);
}

Sex
Apoly_reduce_set(Sex a)
{

    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define bs stack[-1]
#define c stack[-0]
    while (bs != nil) {
	c = car(bs);
	bs = cdr(bs);
#ifdef DEBUG
	lisp_printf("poly_reduce1: %p", a);
	lisp_printf(", %p", c);
#endif
	c = Apoly_reduce1(a, c);
#ifdef DEBUG
	lisp_printf(" ==> %p\n", c);
#endif
	if (a != c) {
	    a = c;
	    bs = b;
	}
    }
#undef a
    a = stack[-3];
#undef b
#undef bs
#undef c
    popv(4);
    return a;
}

Sex
Apoly_have_nontrivial_gcd(Sex a, Sex b)
{
    long            o;
    Sex             na,
                    nb;

    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0) {
	    b = poly_pol(b);
	    b = coeff(lterm(b));
	} else if (o == 0) {
	    a = poly_pol(a);
	    b = poly_pol(b);
	    na = power(lterm(a));
	    nb = power(lterm(b));
	    if (na != mkfxat(0) && nb != mkfxat(0))
		return true;
	    a = coeff(lterm(a));
	    b = coeff(lterm(b));
	} else {
	    a = poly_pol(a);
	    a = coeff(lterm(a));
	}
    }
    return nil;
}

Sex
Apoly_lt_reducible(Sex a, Sex b)
{
    long            o;
    Sex             na,
                    nb;

    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0)
	    return nil;
	if (o == 0) {
	    a = poly_pol(a);
	    b = poly_pol(b);
	    na = power(lterm(a));
	    nb = power(lterm(b));
	    if (na < nb)
		return nil;
	    a = coeff(lterm(a));
	    b = coeff(lterm(b));
	} else {
	    a = poly_pol(a);
	    a = coeff(lterm(a));
	}
    }
    if (is_poly(b))
	return nil;
    return true;
}

Sex
Apoly_lt_divisible(Sex a, Sex b)
{
    long            o;
    Sex             na,
                    nb;

    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0)
	    return nil;
	if (o == 0) {
	    a = poly_pol(a);
	    b = poly_pol(b);
	    na = power(lterm(a));
	    nb = power(lterm(b));
	    if (na < nb)
		return nil;
	    a = coeff(lterm(a));
	    b = coeff(lterm(b));
	} else {
	    a = poly_pol(a);
	    a = coeff(lterm(a));
	}
    }
    if (is_poly(b))
	return nil;
    return true;
}

Sex
Apoly_LT_gcd(Sex a, Sex b)
{
    long            o;
    Sex             na,
                    nb;
    Sex             c,
                    x;


    stackcheck();
    c = mkfxat(1);
    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0) {
	    b = poly_pol(b);
	    b = coeff(lterm(b));
	} else if (o == 0) {
	    x = poly_var(a);
	    a = poly_pol(a);
	    b = poly_pol(b);
	    na = power(lterm(a));
	    nb = power(lterm(b));
	    if (na > nb)
		na = nb;
	    if (na != mkfxat(0)) {
		push2(a, b);
		push(c);
		c = make_polynomial(x, mkfxat(1), na);
		pop(b);
		c = gtimes2(c, b);
		pop2(b, a);
	    }
	    a = coeff(lterm(a));
	    b = coeff(lterm(b));
	} else {
	    a = poly_pol(a);
	    a = coeff(lterm(a));
	}
    }
    return c;
}


Sex
Apoly_LT_lcm(Sex a, Sex b)
{
    long            o;
    Sex             na,
                    nb;
    Sex             c,
                    x;


    stackcheck();
    c = mkfxat(1);
    while (is_poly(a) && is_poly(b)) {
	o = compare_var(poly_var(a), poly_var(b));
	if (o > 0) {
	    x = poly_var(b);
	    b = poly_pol(b);
	    nb = power(lterm(b));
	    if (nb != mkfxat(0)) {
		push2(a, b);
		push(c);
		c = make_polynomial(x, mkfxat(1), nb);
		pop(b);
		c = gtimes2(c, b);
		pop2(b, a);
	    }

	    b = coeff(lterm(b));
	} else if (o == 0) {
	    x = poly_var(a);
	    a = poly_pol(a);
	    b = poly_pol(b);
	    na = power(lterm(a));
	    nb = power(lterm(b));
	    if (na < nb)
		na = nb;
	    if (na != mkfxat(0)) {
		push2(a, b);
		push(c);
		c = make_polynomial(x, mkfxat(1), na);
		pop(b);
		c = gtimes2(c, b);
		pop2(b, a);
	    }
	    a = coeff(lterm(a));
	    b = coeff(lterm(b));
	} else {
	    x = poly_var(a);
	    a = poly_pol(a);
	    na = power(lterm(a));
	    if (na != mkfxat(0)) {
		push2(a, b);
		push(c);
		c = make_polynomial(x, mkfxat(1), na);
		pop(b);
		c = gtimes2(c, b);
		pop2(b, a);
	    }
	    a = coeff(lterm(a));
	}
    }
    return c;
}


/***************************************************************/
/***************************************************************/

#if 0

Sex
xquot2pp(Sex a, Sex b)
{
    long            o;

    if (!is_poly(a)) {
	if (!is_poly(b))
	    return ratio2(a, b);
	return mkfxat(0);
    } else if (!is_poly(b)) {
	return xxquot2pn(a, b);
    }

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return mkfxat(0);
    push(a);
    if (o < 0)
	c = xxquot2px(poly_pol(a), b);
    else
	c = xxquot2pp(poly_pol(a), poly_pol(b));
    pop(a);
    if (c == nil)
	return mkfxat(0);
    return make_poly(poly_var(a), c);
}

Sex
xquot2pn(Sex a, Sex b)
{
    if (!is_poly(a))
	return ratio2(a, b);
    x = poly_var(a);
    a = poly_pol(a);
    push(x);
    a = xxquot2pn(a, b);
    pop(x);
    if (a == nil)
	return mkfxat(0);
    return make_poly(x, a);
}

Sex
xxquot2pn(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xquot2pn(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}

Sex
xxquot2px(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xquot2pp(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}


Sex
xxquot2pp(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;

    while (a != nil) {
	d = diff2(power(lterm(a)), power(lterm(b)));
	if (minusp(d))
	    break;
	e = xquot2pp(coeff(lterm(a)), coeff(lterm(b)));
	if (ZEROP(e)) {
	    a = reduct(a);
	    continue;
	}
	e = mkterm(e, d);
	c = rcons(c, e);
	e = times2tp(e, reduct(b));
	a = diff2pp(reduct(a), e);
    }
    c = cdr(f);
    return c;
}


/***************************************************************/

Sex
xrem2pp(Sex a, Sex b)
{
    long            o;

    if (!is_poly(a)) {
	if (!is_poly(b))
	    return mkfxat(0);
	return a;
    } else if (!is_poly(b)) {
	return mkfxat(0);
    }

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return a;
    push(a);
    if (o < 0)
	c = xxrem2px(poly_pol(a), b);
    else
	c = xxrem2pp(poly_pol(a), poly_pol(b));
    pop(a);
    if (c == nil)
	return mkfxat(0);
    return make_poly(poly_var(a), c);
}

Sex
xrem2pn(Sex a, Sex b)
{
    if (!is_poly(a))
	return mkfxat(0);
    x = poly_var(a);
    a = poly_pol(a);
    push(x);
    a = xxrem2pn(a, b);
    pop(x);
    if (a == nil)
	return mkfxat(0);
    return make_poly(x, a);
}

Sex
xxrem2pn(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xrem2pn(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}

Sex
xxrem2px(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xrem2pp(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}


Sex
xxrem2pp(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;

    while (a != nil) {
	d = diff2(power(lterm(a)), power(lterm(b)));
	if (minusp(d))
	    break;
	e = xdiv2pp(coeff(lterm(a)), coeff(lterm(b)));
	g = cdr(e);
	e = car(e);
	e = mkterm(e, d);
	d = mkterm(g, d);
	c = rcons(c, d);
	e = times2tp(e, reduct(b));
	a = diff2pp(reduct(a), e);
    }
    c = cdr(f);
    return c;
}

/***************************************************************/


Sex
xdiv2pp(Sex a, Sex b)
{
    long            o;

    if (!is_poly(a)) {
	if (!is_poly(b)) {
	    a = ratio2(a, b);
	    return cons(a, mkfxat(0));
	}
	return cons(mkfxat(0), a);
    } else if (!is_poly(b)) {
	return xxdiv2pn(a, b);
    }

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return cons(mkfxat(0), a);
    push(a);
    if (o < 0)
	c = xxdiv2px(poly_pol(a), b);
    else
	c = xxdiv2pp(poly_pol(a), poly_pol(b));
    d = car(c);
    if (d != nil)
	d = make_poly(poly_var(a), d);
    else
	d = mkfxat(0);
    c = cdr(c);
    if (c != nil)
	c = make_poly(poly_var(a), c);
    else
	c = mkfxat(0);
    c = cons(d, c);
    pop(a);
    return c;
}



Sex
xdiv2pn(Sex a, Sex b)
{
    if (!is_poly(a)) {
	a = ratio2(a, b);
	return cons(a, mkfxat(0));
    }
    x = poly_var(a);
    a = poly_pol(a);
    push(x);
    a = xxdiv2pn(a, b);
    pop(x);
    if (a == nil)
	return mkfxat(0);
    return make_poly(x, a);
}

Sex
xxdiv2pn(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xdiv2pn(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}

Sex
xxdiv2px(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = xdiv2pp(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);
    return c;
}


Sex
xxdiv2pp(Sex a, Sex b)
{
    f = cons(nil, nil);
    c = f;

    while (a != nil) {
	d = diff2(power(lterm(a)), power(lterm(b)));
	if (minusp(d))
	    break;
	e = xdiv2pp(coeff(lterm(a)), coeff(lterm(b)));
	if (ZEROP(e)) {
	    a = reduct(a);
	    continue;
	}
	e = mkterm(e, d);
	c = rcons(c, e);
	e = times2tp(e, reduct(b));
	a = diff2pp(reduct(a), e);
    }
    c = cdr(f);
    return c;
}

#endif


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
Lsimplify(Sex a)
{
    if (zerop(a))
	return mkfxat(0);
    if (is_vechdr(a)) {
	switch (gnumtype_of(a)) {
	case N_POLY:
	    return simplify_poly(a);
	case N_DIPOLY:
	    return simplify_dipoly(a);
	case N_RPOLY:
	    return simplify_rpoly(a);
	case N_PREF:
	    return simplify_prefix(a);
	default:
	    break;
	}
    }
    return a;
}

Sex
simplify_poly(Sex a)
{
    Sex             x,
                    p;

    x = poly_var(a);
    p = poly_pol(a);
    a = mkfxat(0);
    stackcheck();
    push5(a, nil, nil, x, p);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define x stack[-1]
#define p stack[0]
    while (p != nil) {
	b = lterm(p);
	c = make_polynomial(x, mkfxat(1), power(b));
	b = coeff(b);
	if (!zerop(b)) {
	    c = gtimes2(c, Lsimplify(b));
	    a = gplus2(a, c);
	}
	p = reduct(p);
    }
#undef a
#undef b
#undef c
#undef x
#undef p
    a = stack[-4];
    popv(5);
    return a;
}

Sex
simplify_rpoly(Sex a)
{
    Sex             b;
    b = rpoly_num(a);
    a = rpoly_den(a);
    stackcheck();
    push(a);
    b = Lsimplify(b);
    a = stack[0];
    stack[0] = b;
    a = Lsimplify(a);
    pop(b);
    return gratio2(b, a);
}


Sex
Lreorder(Sex a)
{
    Sex             x,
                    p;
    Sex             s;

    if (zerop(a))
	return a;

    if (!is_poly(a)) {
	if (is_number(a))
	    return a;
	if (is_algnum(a))
	    return a;
	error1("reorder: ", a);
	return a;
    }

    x = poly_var(a);
    p = poly_pol(a);
    stackcheck();
    push(a);
    push5(mkfxat(0), nil, nil, x, p);
#define a stack[-4]
#define b stack[-3]
#define c stack[-2]
#define x stack[-1]
#define p stack[0]
    while (p != nil) {
	b = lterm(p);
	// printf("b = 0x%08lx\n", b);
	if (!is_cons(b)) {
	    printf("p = 0x%08llx\n", p);
	    printf("fwd = 0x%08lx -- 0x%08lx, 0x%08lx\n", (long) fwdbase,
		   (long) fwdlimit, (long) fwdpointer);
	    printf("cell = 0x%08lx -- 0x%08lx, 0x%08lx\n", (long) cellbase,
		   (long) celllimit, (long) cellpointer);
	    printf("b = 0x%08llx\n", b);
	    // lisp_printf("reorder: a=%p\n", stack[-5]);
	    p = poly_pol(stack[-5]);
	    while (p != nil) {
		b = lterm(p);
		printf("  0x%08llx : 0x%08llx\n", p, b);
		if (!is_cons(b))
		    break;
		p = reduct(p);
	    }
	    error("not term");
	}
	c = make_polynomial(x, mkfxat(1), power(b));
	s = Lreorder(coeff(b));
	c = gtimes2(c, s);
	a = gplus2(a, c);
	p = reduct(p);
    }
#undef a
#undef b
#undef c
#undef x
#undef p
    p = stack[-4];
    popv(6);
    return p;
}

Sex
Lsetorder(Sex a)
{
    Sex             c;
    c = kernel_order;
    if (c == a)
	return a;
    if (Llistp(a) == nil)
	return error("setkorder: argument should be a list");
    if (a != nil) {
	switch (car(a)) {
	case mkfxat(TERM_ORDER_LEX):
	case mkfxat(TERM_ORDER_TLEX):
	case mkfxat(TERM_ORDER_REVLEX):
	case mkfxat(TERM_ORDER_TREVLEX):
	    term_order = mkint(car(a));
	    dipoly_kernel_order = a;
	    a = cdr(a);
	    break;
	case mkfxat(TERM_ORDER_GRADINGS):
	    Lsetorder_gradings(cadr(a), cddr(a));
	    a = cddr(a);
	}
    }
    kernel_order = a;
    return c;
}


Sex
Lgetorder()
{
    return kernel_order;
}

Sex
Lgetdipolyorder()
{
    return dipoly_kernel_order;
}


Sex
Lsetkorder(Sex a)
{
    Sex             c;
    c = kernel_order;
    if (c == a)
	return a;
    if (Llistp(a) == nil)
	return error("setkorder: argument should be a list");
    kernel_order = a;
    return c;
}



long
compare_term(Sex a, Sex b)
{
    long            r;

    switch (term_order) {
    case TERM_ORDER_LEX:
	return compare_term_lex(a, b);
    case TERM_ORDER_TLEX:
	return compare_term_tlex(a, b);
    case TERM_ORDER_TREVLEX:
	r = compare_term_trevlex(a, b);
#if 0
	printf("compare_term returns %ld\n", r);
#endif
	return r;
    case TERM_ORDER_GRADINGS:
	return compare_term_gradings(a, b);
    }
    return aerror1("invalid term order", mkfxat(term_order));
}

long
compare_term_lex(Sex a, Sex b)
{
    Sex             x;
    Sex             y;
    long            w;

    while (1) {
	x = poly_var(a);
	y = poly_var(b);
	w = compare_var(x, y);
	if (w != 0)
	    return -w;
	a = poly_pol(a);
	b = poly_pol(b);
	a = lterm(a);
	b = lterm(b);
	x = power(a);
	y = power(b);
	if (x == y) {
	    a = coeff(a);
	    b = coeff(b);
	    if (is_poly(a)) {
		if (is_poly(b))
		    continue;
		return 1;
	    }
	    if (is_poly(b))
		return -1;
	    return 0;
	}
	return (long) x - (long) y;
    }
}

long
compare_term_tlex(Sex a, Sex b)
{
    Sex             x;
    Sex             y;
    long            w;
    long            n = 0,
	m = 0;


    x = a;
    while (is_poly(x)) {
	x = lterm(poly_pol(x));
	n += mkint(power(x));
	x = coeff(x);
    }
    x = b;
    while (is_poly(x)) {
	x = lterm(poly_pol(x));
	m += mkint(power(x));
	x = coeff(x);
    }
    if (n != m)
	return n - m;
    while (1) {
	x = poly_var(a);
	y = poly_var(b);
	w = compare_var(x, y);
	if (w != 0)
	    return -w;
	a = poly_pol(a);
	b = poly_pol(b);
	a = lterm(a);
	b = lterm(b);
	x = power(a);
	y = power(b);
	if (x == y) {
	    a = coeff(a);
	    b = coeff(b);
	    if (is_poly(a)) {
		if (is_poly(b))
		    continue;
		return 1;
	    }
	    if (is_poly(b))
		return -1;
	    return 0;
	}
	return (long) x - (long) y;
    }
}

long
compare_term_trevlex(Sex a, Sex b)
{
    Sex             x;
    Sex             y;
    Sex             c;
    long            w;
    long            n = 0,
	m = 0;
    static long     vec_a[32];

#if 0
    lisp_printf("compare %p", a);
    lisp_printf(" and %p\n", b);
#endif

    x = a;
    while (is_poly(x)) {
	x = lterm(poly_pol(x));
	n += mkint(power(x));
	x = coeff(x);
    }
    x = b;
    while (is_poly(x)) {
	x = lterm(poly_pol(x));
	m += mkint(power(x));
	x = coeff(x);
    }
    if (n != m)
	return n - m;
    n = 0;

    while (1) {
	x = poly_var(a);
	y = poly_var(b);
	w = compare_var(x, y);
	if (w > 0) {
	    b = poly_pol(b);
	    b = lterm(b);
	    y = power(b);
	    vec_a[n++] = mkint(y);
	    b = coeff(b);
	    if (!is_poly(b))
		return -1;
	    continue;
	} else if (w < 0) {
	    a = lterm(poly_pol(a));
	    x = power(a);
	    vec_a[n++] = -mkint(x);
	    a = coeff(a);
	    if (!is_poly(a))
		return +1;
	    continue;
	}
	a = poly_pol(a);
	b = poly_pol(b);
	a = lterm(a);
	b = lterm(b);
	x = power(a);
	y = power(b);
	vec_a[n++] = -(mkint(x) - mkint(y));
	a = coeff(a);
	b = coeff(b);
	if (is_poly(a)) {
	    if (is_poly(b))
		continue;
	    return -1;
	}
	if (is_poly(b))
	    return 1;
	break;
    }

#if 0
    printf("vec = ");
    for (m = 0; m < n; m++)
	printf(" %4d ", vec_a[m]);
    printf("\n");
#endif

    while (n > 0) {
	if (vec_a[--n] != 0)
	    return vec_a[n];
    }
    return 0;
}

long
compare_term_gradings(Sex a, Sex b)
{
    Sex             x;
    Sex             y;
    Sex             v,
                    w;
    long            n = 0,
	m = 0;
    long            i,
                    j;

    static long     vec_a[32];
    static long     vec_b[32];
#if 0
    lisp_printf("compare %p", a);
    lisp_printf(" and %p\n", b);
#endif

    for (v = kernel_order; v != nil; v = cdr(v)) {
	w = car(v);
	if (is_poly(w))
	    w = poly_var(w);
	x = a;
	y = b;
	if (is_poly(a))
	    x = poly_var(a);
	if (is_poly(b))
	    y = poly_var(b);
	if (x == w) {
	    if (y == w) {
		a = lterm(poly_pol(a));
		b = lterm(poly_pol(b));
		vec_a[n++] = mkint(power(a)) - mkint(power(b));
		a = coeff(a);
		b = coeff(b);
	    } else {
		a = lterm(poly_pol(a));
		vec_a[n++] = mkint(power(a));
		a = coeff(a);
	    }
	} else if (y == w) {
	    b = lterm(poly_pol(b));
	    vec_a[n++] = -mkint(power(b));
	    b = coeff(b);
	} else {
	    vec_a[n++] = 0;
	}
    }

#if 0
    printf("vec_a = ");
    for (i = 0; i < n; i++)
	printf(" %4d ", vec_a[i]);
    printf("\n");
#endif

    for (i = 0; i < gradings_matrix_m; i++) {
	m = 0;
	for (j = 0; j < gradings_matrix_n; j++) {
	    m += gradings_matrix[i * gradings_matrix_n + j] * vec_a[j];
	}
#if 0
	printf("i = %4d, m = %4d\n", i, m);
#endif
	if (m != 0)
	    return m;
    }
    return m;
}


long
compare_var(Sex a, Sex b)
{
    Sex             x;
    Sex             y;
    long            c;
    long            l1,
                    l2;
    if (a == b)
	return 0;
    x = kernel_order;
    while (x != nil) {
	y = car(x);
	if (is_poly(y))
	    y = poly_var(y);
	if (a == y)
	    return -1;
	if (b == y)
	    return 1;
	x = cdr(x);
    }

    l1 = symbol_length(a);
    l2 = symbol_length(b);
    c = memcmp(pname(a), pname(b), (size_t) (l1 < l2 ? l1 : l2));
    if (c == 0)
	c = (long) (l1 - l2);
    if (c == 0)
	c = (long) (a - b);
    return c;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2xp(Sex t, Sex b)
{
    Sex             p;

    stackcheck();
    push2(t, b);
    p = cons(nil, nil);
    push4(p, p, nil, nil);
#define t stack[-5]
#define b stack[-4]
#define p stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (b != nil) {
	bc = lterm(b);
	if (power(bc) == mkfxat(0)) {
	    c = gplus2(t, coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, mkfxat(0));
		p = rcons(p, c);
	    }
	    cdr(p) = reduct(b);
	    goto ax;
	} else if (plusp(power(bc))) {
	    p = rcons(p, bc);
	    b = reduct(b);
	} else {
	    c = mkterm(t, mkfxat(0));
	    p = rcons(p, c);
	    cdr(p) = b;
	    goto ax;
	}
    }
    c = mkterm(t, mkfxat(0));
    p = rcons(p, c);
  ax:
#undef t
#undef b
#undef p
#undef bc
#undef c
    p = stack[-3];
    popv(6);
    return cdr(p);
}



Sex
gplus2pn(Sex a, Sex b)
{
    Sex             x,
                    p;

    if (ZEROP(b))
	return a;
    x = poly_var(a);
    p = poly_pol(a);
    stackcheck();
    push(x);
    p = plus2xp(b, p);
    pop(x);
    return make_poly(x, p);
}


Sex
gplus2np(Sex a, Sex b)
{
    Sex             x,
                    p;

    if (ZEROP(a))
	return b;

    x = poly_var(b);
    p = poly_pol(b);
    stackcheck();
    push(x);
    p = plus2xp(a, p);
    pop(x);
    return make_poly(x, p);
}

Sex
plus2pp(Sex a, Sex b)
{
    Sex             p;

    stackcheck();
    push2(a, b);
    p = cons(nil, nil);
    push5(p, p, nil, nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-3]
#define ac stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (a != nil && b != nil) {
	ac = lterm(a);
	bc = lterm(b);
	if (numeq2(power(ac), power(bc))) {
	    c = gplus2(coeff(ac), coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, power(ac));
		p = rcons(p, c);
	    }
	    a = reduct(a);
	    b = reduct(b);
	} else if (lessp2(power(ac), power(bc))) {
	    p = rcons(p, bc);
	    b = reduct(b);
	} else {
	    p = rcons(p, ac);
	    a = reduct(a);
	}
    }
    if (a != nil)
	cdr(p) = a;
    else if (b != nil)
	cdr(p) = b;
#undef a
#undef b
#undef p
#undef ac
#undef bc
#undef c
    p = cdr(stack[-4]);
    popv(7);
    return p;
}



Sex
gplus2pp(Sex a, Sex b)
{
    long            o;

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return gplus2np(a, b);
    else if (o < 0)
	return gplus2np(b, a);
    stackcheck();
    push(a);
    b = plus2pp(poly_pol(a), poly_pol(b));
    pop(a);
    return make_poly(poly_var(a), b);
}



Sex
gplus2nr(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = gtimes2(a, rpoly_den(b));
    a = gplus2(a, rpoly_num(stack[0]));
    pop(b);
    return make_rpoly(a, rpoly_den(b));
}

Sex
gplus2rn(Sex b, Sex a)
{
    stackcheck();
    push(b);
    a = gtimes2(a, rpoly_den(b));
    a = gplus2(a, rpoly_num(stack[0]));
    pop(b);
    return make_rpoly(a, rpoly_den(b));
}


Sex
gplus2rr(Sex a, Sex b)
{
    Sex             na = rpoly_num(a),
	nb = rpoly_num(b);
    Sex             da = rpoly_den(a),
	db = rpoly_den(b);
    Sex             w = nil;

    stackcheck();
    push5(na, nb, da, db, nil);
#define g   stack[0]
#define db  stack[-1]
#define da  stack[-2]
#define nb  stack[-3]
#define na  stack[-4]
    g = Mgcd(da, db);
    db = gquot2(db, g);
    g = gquot2(da, g);
    na = gtimes2(na, db);
    nb = gtimes2(nb, g);
    na = gplus2(na, nb);
    da = gtimes2(da, db);
    g = Mgcd(na, da);
    na = gquot2(na, g);
    da = gquot2(da, g);
    if (ZEROP(na))
	w = mkfxat(0);
    else if (is_poly(da))
	w = make_rpoly(na, da);
    else
	w = gratio2(na, da);
    if (na != mkfxat(0))
	/*
	 * if(is_poly(da)) w = make_rpoly(na, da); else w = make_rpoly(na, 
	 * da); w = gratio2(na, da); else w = mkfxat(0); 
	 */
	popv(5);
    return w;
#undef na
#undef nb
#undef da
#undef db
#undef g
}

Sex
diff2xp(Sex t, Sex b)
{
    Sex             p;

    stackcheck();
    push2(t, b);
    p = cons(nil, nil);
    push4(p, p, nil, nil);
#define t stack[-5]
#define b stack[-4]
#define p stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (b != nil) {
	bc = lterm(b);
	if (power(bc) == mkfxat(0)) {
	    c = gdiff2(t, coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, mkfxat(0));
		p = rcons(p, c);
	    }
	    b = reduct(b);
	    goto ax;
	} else if (plusp(power(bc))) {
	    c = gnegate(coeff(bc));
	    c = mkterm(c, power(bc));
	    p = rcons(p, c);
	    b = reduct(b);
	} else {
	    c = mkterm(t, mkfxat(0));
	    p = rcons(p, c);
	    goto ax;
	}
    }
    c = mkterm(t, mkfxat(0));
    p = rcons(p, c);
  ax:
    while (b != nil) {
	bc = lterm(b);
	c = gnegate(coeff(bc));
	c = mkterm(c, power(bc));
	p = rcons(p, c);
	b = reduct(b);
    }
#undef t
#undef b
#undef p
#undef bc
#undef c
    p = stack[-3];
    popv(6);
    return cdr(p);
}

Sex
diff2px(Sex a, Sex t)
{
    Sex             p;

    stackcheck();
    push2(a, t);
    p = cons(nil, nil);
    push4(p, p, nil, nil);
#define a stack[-5]
#define t stack[-4]
#define p stack[-2]
#define ac stack[-1]
#define c stack[0]
    while (a != nil) {
	ac = lterm(a);
	if (power(ac) == mkfxat(0)) {
	    c = gdiff2(coeff(ac), t);
	    if (!ZEROP(c)) {
		c = mkterm(c, power(ac));
		p = rcons(p, c);
	    }
	    cdr(p) = reduct(a);
	    goto ax;
	} else if (minusp(power(ac))) {
	    c = gnegate(t);
	    c = mkterm(c, mkfxat(0));
	    p = rcons(p, c);
	    cdr(p) = a;
	    goto ax;
	} else {
	    p = rcons(p, ac);
	    a = reduct(a);
	}
    }
    c = gnegate(t);
    c = mkterm(c, mkfxat(0));
    p = rcons(p, c);
  ax:
#undef a
#undef t
#undef p
#undef ac
#undef c
    p = stack[-3];
    popv(6);
    return cdr(p);
}


Sex
gdiff2pn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    stackcheck();
    push(a);
    b = diff2px(poly_pol(a), b);
    pop(a);
    return make_poly(poly_var(a), b);
}


Sex
gdiff2np(Sex a, Sex b)
{
    if (zerop(a))
	return gnegatep(b);
    stackcheck();
    push(b);
    a = diff2xp(a, poly_pol(b));
    pop(b);
    return make_poly(poly_var(b), a);
}

Sex
diff2pp(Sex a, Sex b)
{
    Sex             p;

    stackcheck();
    push2(a, b);
    p = cons(nil, nil);
    push5(p, p, nil, nil, nil);
#define a stack[-6]
#define b stack[-5]
#define p stack[-3]
#define ac stack[-2]
#define bc stack[-1]
#define c stack[0]
    while (a != nil && b != nil) {
	ac = lterm(a);
	bc = lterm(b);
	if (numeq2(power(ac), power(bc))) {
	    c = gdiff2(coeff(ac), coeff(bc));
	    if (!ZEROP(c)) {
		c = mkterm(c, power(ac));
		p = rcons(p, c);
	    }
	    a = reduct(a);
	    b = reduct(b);
	} else if (lessp2(power(ac), power(bc))) {
	    c = gnegate(coeff(bc));
	    bc = mkterm(c, power(bc));
	    p = rcons(p, bc);
	    b = reduct(b);
	} else {
	    p = rcons(p, ac);
	    a = reduct(a);
	}
    }
    if (a != nil)
	cdr(p) = a;
    else if (b != nil) {
	while (b != nil) {
	    bc = lterm(b);
	    c = gnegate(coeff(bc));
	    c = mkterm(c, power(bc));
	    p = rcons(p, c);
	    b = reduct(b);
	}
    }
#undef a
#undef b
#undef p
#undef ac
#undef bc
#undef c
    p = cdr(stack[-4]);
    popv(7);
    return p;
}


Sex
gdiff2pp(Sex a, Sex b)
{
    long            o;

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return gdiff2np(a, b);
    else if (o < 0)
	return gdiff2pn(a, b);
    stackcheck();
    push(a);
    b = diff2pp(poly_pol(a), poly_pol(b));
    pop(a);
    return make_poly(poly_var(a), b);
}


Sex
negatep(Sex a)
{
    Sex             d;

    Sex             p;

    stackcheck();
    push(a);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = gnegate(coeff(c));
	c = mkterm(d, power(c));
	p = rcons(p, c);
	a = reduct(a);
    }
#undef a
#undef p
#undef c
    p = stack[-2];
    popv(4);
    return cdr(p);
}


Sex
gnegatep(Sex a)
{
    Sex             c;

    stackcheck();
    push(a);
    c = negatep(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}


Sex
gnegater(Sex a)
{
    Sex             b,
                    c;
    b = rpoly_den(a);
    c = rpoly_num(a);
    stackcheck();
    push(b);
    c = gnegate(c);
    pop(b);
    return make_rpoly(c, b);
}


Sex
modulusp(Sex a, Sex b)
{
    Sex             d;

    Sex             p;

    stackcheck();
    push2(a, b);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-4]
#define b stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = gmodulus(coeff(c), b);
	if (!zerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef p
#undef c
    p = stack[-2];
    popv(5);
    return cdr(p);
}

Sex
smodulusp(Sex a, Sex b)
{
    Sex             d;

    Sex             p;

    stackcheck();
    push2(a, b);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-4]
#define b stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = gsmodulus(coeff(c), b);
	if (!zerop(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef b
#undef p
#undef c
    p = stack[-2];
    popv(5);
    return cdr(p);
}




Sex
gmodulusp(Sex a, Sex b)
{
    Sex             c;

    stackcheck();
    push(a);
    c = modulusp(poly_pol(a), b);
    pop(a);
    if (c == nil)
	return mkfxat(0);
    return make_poly(poly_var(a), c);
}


Sex
gmodulusr(Sex a, Sex m)
{
    Sex             b,
                    c;
    b = rpoly_den(a);
    c = rpoly_num(a);
    stackcheck();
    push2(b, m);
    c = gmodulus(c, m);
    push(c);
    b = stack[-2];
    m = stack[-1];
    b = gmodulus(b, m);
    c = stack[0];
    popv(3);
    return make_rpoly(c, b);
}

Sex
gsmodulusp(Sex a, Sex b)
{
    Sex             c;

    stackcheck();
    push(a);
    c = smodulusp(poly_pol(a), b);
    pop(a);
    if (c == nil)
	return mkfxat(0);
    return make_poly(poly_var(a), c);
}


Sex
gsmodulusr(Sex a, Sex m)
{
    Sex             b,
                    c;
    b = rpoly_den(a);
    c = rpoly_num(a);
    stackcheck();
    push2(b, m);
    c = gsmodulus(c, m);
    push(c);
    b = stack[-2];
    m = stack[-1];
    b = gsmodulus(b, m);
    c = stack[0];
    popv(3);
    return make_rpoly(c, b);
}

Sex
gdiff2nr(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = gtimes2(a, rpoly_den(b));
    a = gdiff2(a, rpoly_num(stack[0]));
    pop(b);
    return make_rpoly(a, rpoly_den(b));
}

Sex
gdiff2rn(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = gtimes2(rpoly_den(a), b);
    b = gdiff2(rpoly_num(stack[0]), b);
    pop(a);
    return make_rpoly(b, rpoly_den(a));
}

Sex
gdiff2rr(Sex a, Sex b)
{
    Sex             na = rpoly_num(a),
	nb = rpoly_num(b);
    Sex             da = rpoly_den(a),
	db = rpoly_den(b);
    Sex             w = nil;

    stackcheck();
    push5(na, nb, da, db, nil);
#define g   stack[0]
#define db  stack[-1]
#define da  stack[-2]
#define nb  stack[-3]
#define na  stack[-4]
    g = Mgcd(da, db);
    db = gquot2(db, g);
    g = gquot2(da, g);
    na = gtimes2(na, db);
    nb = gtimes2(nb, g);
    na = gdiff2(na, nb);
    da = gtimes2(da, db);
    g = Mgcd(na, da);
    na = gquot2(na, g);
    da = gquot2(da, g);
    if (ZEROP(na))
	w = mkfxat(0);
    else if (is_poly(da))
	w = make_rpoly(na, da);
    else
	w = gratio2(na, da);
    /*
     * if (na != mkfxat(0)) // w = make_rpoly(na, da); w = gratio2(na,
     * da); else w = mkfxat(0); 
     */
    popv(5);
    return w;
#undef na
#undef nb
#undef da
#undef db
#undef g
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
times2xp(Sex t, Sex a)
{
    Sex             d,
                    p;

    stackcheck();
    push2(a, t);
    p = cons(nil, nil);
    push3(p, p, nil);
#define a stack[-4]
#define t stack[-3]
#define p stack[-1]
#define c stack[0]
    while (a != nil) {
	c = lterm(a);
	d = gtimes2(t, coeff(c));
	if (!ZEROP(d)) {
	    c = mkterm(d, power(c));
	    p = rcons(p, c);
	}
	a = reduct(a);
    }
#undef a
#undef t
#undef p
#undef c
    p = stack[-2];
    popv(5);
    return cdr(p);
}


Sex
gtimes2pn(Sex p, Sex t)
{
    if (ZEROP(t))
	return t;
    if (ONEP(t))
	return p;
    stackcheck();
    push(p);
    t = times2xp(t, poly_pol(p));
    pop(p);
    if (t == nil)
	return mkfxat(0);
    return make_poly(poly_var(p), t);
}

Sex
gtimes2np(Sex t, Sex p)
{
    if (ZEROP(t))
	return t;
    if (ONEP(t))
	return p;
    stackcheck();
    push(p);
    t = times2xp(t, poly_pol(p));
    pop(p);
    if (t == nil)
	return mkfxat(0);
    return make_poly(poly_var(p), t);
}


Sex
times2tt(Sex t1, Sex t2)
{
    Sex             a;

    stackcheck();
    push4(t1, t2, nil, nil);
#define t1 stack[-3]
#define t2 stack[-2]
#define a stack[-1]
#define b stack[0]
    a = gtimes2(coeff(t1), coeff(t2));
    b = gplus2(power(t1), power(t2));
    a = mkterm(a, b);
#undef a
#undef b
#undef t1
#undef t2
    a = stack[-1];
    popv(4);
    return a;
}

Sex
times2tp(Sex t, Sex p)
{
    Sex             a;

    if (p == nil)
	return nil;
    stackcheck();
    push5(t, p, nil, nil, nil);
#define t stack[-4]
#define p stack[-3]
#define a stack[-2]
#define b stack[-1]
#define c stack[0]
    c = lterm(p);
    a = gtimes2(coeff(t), coeff(c));
    b = gplus2(power(t), power(c));
    a = mkterm(a, b);
    p = times2tp(t, reduct(p));
    a = cons(a, p);
#undef a
#undef b
#undef c
#undef p
#undef t
    a = stack[-2];
    popv(5);
    return a;
}
#ifdef NEW_POLYMUL
#else

Sex
times2pp(Sex a, Sex b)
{
    Sex             e;

    if (a == nil || b == nil)
	return nil;
    stackcheck();
    push4(a, b, nil, nil);
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[ 0]
    c = times2pp(reduct(a), b);
    d = times2tp(lterm(a), reduct(b));
    c = plus2pp(c, d);
    a = lterm(a);
    b = lterm(b);
    d = gtimes2(coeff(a), coeff(b));
    e = gplus2(power(a), power(b));
    d = mkterm(d, e);
    e = cons(d, c);
#undef a
#undef b
#undef c
#undef d
    popv(4);
    return e;
}

#endif

Sex
gtimes2pp(Sex a, Sex b)
{
    long            o;

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return gtimes2np(a, b);
    else if (o < 0)
	return gtimes2np(b, a);
    stackcheck();
    push(a);
    b = times2pp(poly_pol(a), poly_pol(b));
    pop(a);
    return make_poly(poly_var(a), b);
}


Sex
exptip(Sex a, long n)
{
    Sex             p;
    long            i;
    Sex             u;

    if (n == 1)
	return a;
    p = lterm(a);
    a = reduct(a);

    stackcheck();
    push2(a, p);
    push5(nil, nil, nil, nil, nil);
#define a stack[-6]
#define p stack[-5]
#define c stack[-4]
#define f stack[-3]
#define q stack[-2]
#define r stack[-1]
#define s stack[ 0]

    c = coeff(p);
    if (is_number(c))
	c = expti(c, n);
    else if (is_poly(c))
	c = gexpt2pi(c, mkfxat(n));
    else
	c = gexpt(c, mkfxat(n));
    if (!ZEROP(c)) {
	q = times2(power(p), mkfxat(n));
	s = mkterm(c, q);
	s = cons(s, nil);
    } else {
	s = nil;
    }
    r = mkterm(mkfxat(1), mkfxat(0));
    r = cons(r, nil);
    if (a != nil) {
	f = mkfxat(n);
	for (i = 1; i < n; i++) {
	    c = coeff(p);
	    if (is_number(c))
		c = expti(c, n - i);
	    else if (is_poly(c))
		c = gexpt2pi(c, mkfxat(n - i));
	    else
		c = gexpt(c, mkfxat(n - i));
	    /*
	     * c = gexpt2pi(coeff(p), n-i);
	     */
	    q = gtimes2(power(p), mkfxat(n - i));
	    c = gtimes2(f, c);
	    f = gtimes2(f, mkfxat(n - i));
	    f = gquot2(f, mkfxat(i + 1));
	    r = times2pp(r, a);
	    if (!ZEROP(c)) {
		u = mkterm(c, q);
		u = times2tp(u, r);
		s = plus2pp(s, u);
	    }
	}
	r = times2pp(r, a);
	s = plus2pp(s, r);
    }
#undef a
#undef p
#undef c
#undef f
#undef q
#undef r
#undef s
    p = stack[0];
    popv(7);
    return p;
}

Sex
gexpt2pi(Sex a, Sex b)
{
    Sex             x;
    long            n = mkint(b);

    if (n == 0)
	return mkfxat(1);
    else if (n == 1)
	return a;
    else if (n < 0) {
	a = gexpt2pi(a, mkfxat(-n));
	return gratio2(mkfxat(1), a);
    }
    x = poly_var(a);
    a = poly_pol(a);
    stackcheck();
    push(x);
    a = exptip(a, n);
    pop(x);
    return make_poly(x, a);
}

Sex
gexpt2ri(Sex a, Sex b)
{
    Sex             x;

    if (b == mkfxat(0))
	return mkfxat(1);
    else if (b == mkfxat(1))
	return a;

    stackcheck();
    push(rpoly_den(a));
    a = rpoly_num(a);
    a = gexpt2pi(a, b);
    pop(x);
    push(a);
    x = gexpt2pi(x, b);
    pop(a);
    return make_rpoly(a, x);
}


Sex
gtimes2rn(Sex a, Sex b)
{
    return gtimes2nr(b, a);
}

Sex
gtimes2nr(Sex a, Sex b)
{
    Sex             w = nil;
    if (a == mkfxat(0))
	return a;
    else if (a == mkfxat(1))
	return b;
    stackcheck();
    push3(b, a, nil);
#define g   stack[0]
#define a   stack[-1]
#define b   stack[-2]
    g = Mgcd(a, rpoly_den(b));
    a = gquot2(a, g);
    g = gquot2(rpoly_den(b), g);
    a = gtimes2(a, rpoly_num(b));
    // w = make_rpoly(a, g);
    if (is_poly(g))
	w = make_rpoly(a, g);
    else
	w = gratio2(a, g);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}


Sex
gtimes2rr(Sex a, Sex b)
/*
 * multiply a pair of rational numbers
 */
{
    Sex             w = nil;
    stackcheck();
    push5(rpoly_num(a), rpoly_den(a), rpoly_num(b), rpoly_den(b), nil);
#define g   stack[0]
#define db  stack[-1]
#define nb  stack[-2]
#define da  stack[-3]
#define na  stack[-4]
    g = Mgcd(na, db);
    na = gquot2(na, g);
    db = gquot2(db, g);
    g = Mgcd(nb, da);
    nb = gquot2(nb, g);
    da = gquot2(da, g);
    na = gtimes2(na, nb);
    da = gtimes2(da, db);
    /*
     * w = make_rpoly(na, da); 
     */
    if (is_poly(da))
	w = make_rpoly(na, da);
    else
	w = gratio2(na, da);
    popv(5);
    return w;
#undef g
#undef db
#undef nb
#undef da
#undef na
}

Sex
quot2px(Sex a, Sex b)
{
    if (b == nil)
	aerror2("quotient", a, b);
    if (a == nil)
	return nil;
    stackcheck();
    push5(a, b, nil, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define e stack[-1]
#define f stack[0]

    // debug_printf("quot2px: %p", a);
    // debug_printf(" and %p\n", b);

    f = cons(nil, nil);
    c = f;
    while (a != nil) {
	d = power(lterm(a));
	e = gquot2(coeff(lterm(a)), b);
	if (!ZEROP(e)) {
	    e = mkterm(e, d);
	    c = rcons(c, e);
	}
	a = reduct(a);
    }
    c = cdr(f);

#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
    a = cdr(stack[0]);
    popv(6);
    return a;
}


Sex
gquot2pn(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push(a);
    c = quot2px(poly_pol(a), b);
    pop(a);
    return make_poly(poly_var(a), c);
}


Sex
gquot2np(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
quot2pp(Sex a, Sex b)
{
    if (b == nil)
	aerror2("quotient", a, b);
    if (a == nil)
	return nil;
    stackcheck();
    push5(a, b, nil, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define c stack[-3]
#define d stack[-2]
#define e stack[-1]
#define f stack[0]
    f = cons(nil, nil);
    c = f;
    // lisp_printf("quot2pp: a = %p", a);
    // lisp_printf(", b = %p\n", b);
    while (a != nil) {
	d = diff2(power(lterm(a)), power(lterm(b)));
	if (minusp(d))
	    break;
	e = gquot2(coeff(lterm(a)), coeff(lterm(b)));
	if (ZEROP(e)) {
	    a = reduct(a);
	    continue;
	}
	e = mkterm(e, d);
	c = rcons(c, e);
	e = times2tp(e, reduct(b));
	/*
	 * lisp_printf("quot2pp: reduct(a) = %p", reduct(a));
	 * lisp_printf(", e = %p\n", e); 
	 */
	a = diff2pp(reduct(a), e);
    }
    c = cdr(f);

    // lisp_printf("quot2pp result is %p\n", c);

#undef a
#undef b
#undef c
#undef d
#undef e
#undef f
    a = cdr(stack[0]);
    popv(6);
    return a;
}

Sex
gquot2pp(Sex a, Sex b)
{
    Sex             c;
    long            o;

    o = compare_var(poly_var(a), poly_var(b));
    if (o > 0)
	return mkfxat(0);
    stackcheck();
    // lisp_printf("gquot2pp: a = %p", a);
    // lisp_printf(", b=%p\n", b);
    push(a);
    if (o < 0)
	c = quot2px(poly_pol(a), b);
    else
	c = quot2pp(poly_pol(a), poly_pol(b));
    pop(a);
    // lisp_printf("gquot2pp: result is %p\n", a);
    return make_poly(poly_var(a), c);
}



Sex
gquot2nr(Sex a, Sex b)
{
    Sex             w;
    /*
     * if (a == mkfxat(0)) return a; 
     */
    if (ZEROP(a))
	return a;
    stackcheck();
    push3(b, a, nil);
#define g   stack[0]
#define a   stack[-1]
#define b   stack[-2]
    g = Mgcd(a, rpoly_num(b));
    a = gquot2(a, g);
    if (minusp(rpoly_num(b)))
	g = gnegate(g);
    g = gquot2(rpoly_num(b), g);
    a = gtimes2(a, rpoly_den(b));
    w = make_rpoly(a, g);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}



Sex
gquot2rn(Sex a, Sex b)
{
    Sex             w;
    /*
     * if (b == mkfxat(1)) return a; else if (b == mkfxat(0)) return
     * aerror2("bad arg for quotient", a, b); 
     */
    if (ONEP(b))
	return a;
    else if (ZEROP(b))
	return aerror2("bad arg for gquotient", a, b);
    stackcheck();
    push3(a, b, nil);
#define g   stack[0]
#define b   stack[-1]
#define a   stack[-2]
    g = Mgcd(b, rpoly_num(a));
    if (minusp(b))
	g = gnegate(g);
    b = gquot2(b, g);
    g = gquot2(rpoly_num(a), g);
    a = gtimes2(b, rpoly_den(a));
    w = make_rpoly(g, a);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}



Sex
gquot2rr(Sex a, Sex b)
{
    Sex             w;
    stackcheck();
    push5(rpoly_num(a), rpoly_den(a), rpoly_den(b), rpoly_num(b),	/* 
									 * * NB switched order  
									 */
	  nil);
#define g   stack[0]
#define db  stack[-1]
#define nb  stack[-2]
#define da  stack[-3]
#define na  stack[-4]
    g = Mgcd(na, db);
    w = minusp(db);
    if (w) {
	g = gnegate(g);
    }
    na = gquot2(na, g);
    db = gquot2(db, g);
    g = Mgcd(nb, da);
    nb = gquot2(nb, g);
    da = gquot2(da, g);
    na = gtimes2(na, nb);
    da = gtimes2(da, db);
    // w = make_rpoly(na, da);
    if (is_poly(da))
	w = make_rpoly(na, da);
    else
	w = gratio2(na, da);
    popv(5);
    return w;
#undef g
#undef db
#undef nb
#undef da
#undef na
}



/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
gplus2rp(Sex a, Sex b)
{
    return gplus2rn(a, b);
}

Sex
gplus2pr(Sex a, Sex b)
{
    return gplus2nr(a, b);
}

Sex
gdiff2rp(Sex a, Sex b)
{
    return gdiff2rn(a, b);
}

Sex
gdiff2pr(Sex a, Sex b)
{
    return gdiff2nr(a, b);
}

Sex
gtimes2rp(Sex a, Sex b)
{
    return gtimes2rn(a, b);
}

Sex
gtimes2pr(Sex a, Sex b)
{
    return gtimes2nr(a, b);
}

Sex
gquot2rp(Sex a, Sex b)
{
    Sex             n,
                    d;

    n = rpoly_num(a);
    d = rpoly_den(a);
    stackcheck();
    push4(n, d, b, nil);
#define g stack[0]
#define b stack[-1]
#define d stack[-2]
#define n stack[-3]
    g = Mgcd(n, b);
    n = gquot2(n, g);
    b = gquot2(b, g);
    d = gtimes2(b, d);
#undef g
#undef b
#undef d
#undef n
    n = stack[-3];
    d = stack[-2];
    popv(4);
    return make_rpoly(n, d);
}

Sex
gquot2pr(Sex a, Sex b)
{
    return gquot2nr(a, b);
}

Sex
ratio2px(Sex a, Sex b)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push3(a, b, c);
    d = gratio2(coeff(c), b);
    pop(c);
    c = mkterm(d, power(c));
    pop2(b, a);
    push(c);
    a = ratio2px(a, b);
    pop(c);
    return cons(c, a);
}

Sex
gratio2pn(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push(a);
    c = ratio2px(poly_pol(a), b);
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
gratio2np(Sex a, Sex b)
{
    if (ZEROP(a))
	return a;
    return make_rpoly(a, b);
}


Sex
gratio2pp(Sex a, Sex b)
{
    Sex             g;
    stackcheck();
    push3(a, b, nil);
#define g stack[0]
#define b stack[-1]
#define a stack[-2]
    g = Mgcd(a, b);
    if (is_number(g)) {
	a = gquot2pn(a, g);
	b = gquot2pn(b, g);
    } else {
	/*
	 * lisp_printf("gratio2pp: (%p", a); lisp_printf(", %p)", b);
	 * lisp_printf(", %p\n", g); 
	 */
	a = gquot2pp(a, g);
	b = gquot2pp(b, g);
    }
#undef a
#undef b
#undef g
    pop3(g, b, a);
    if (is_poly(b))
	return make_rpoly(a, b);
    return gratio2(a, b);
}

Sex
mod2xP(Sex a, Sex b)
{
    return nil;
}

Sex
mod2Px(Sex a, Sex b)
{
    return nil;
}

Sex
mod2PP(Sex a, Sex b)
{
    return nil;
}

Sex
mod2xR(Sex a, Sex b)
{
    return nil;
}

Sex
mod2Rx(Sex a, Sex b)
{
    return nil;
}

Sex
mod2RP(Sex a, Sex b)
{
    return mod2Rx(a, b);
}

Sex
mod2PR(Sex a, Sex b)
{
    return mod2xR(a, b);
}

Sex
mod2RR(Sex a, Sex b)
{
    return nil;
}

Sex
grem2np(Sex a, Sex b)
{
    return a;
}

Sex
grem2pn(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = gquot2pn(a, b);
    pop(b);
    c = gtimes2(c, b);
    pop(a);
    return gdiff2(a, c);
}

Sex
grem2pp(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = gquot2pp(a, b);
    pop(b);
    c = gtimes2(c, b);
    // lisp_printf("grem2pp: a = %p\n", a);
    // lisp_printf("grem2pp: c = %p\n", c);
    pop(a);
    return gdiff2(a, c);
}

Sex
grem2nr(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2rn(Sex a, Sex b)
{
    return mkfxat(0);
}

Sex
grem2rp(Sex a, Sex b)
{
    return grem2rn(a, b);
}

Sex
grem2pr(Sex a, Sex b)
{
    return grem2nr(a, b);
}

Sex
grem2rr(Sex a, Sex b)
{
    return mkfxat(0);
}

/***************************************************************/
/***************************************************************/
Sex
ggcd2pp(Sex a, Sex b)
{
    return gcd_sprs(a, b);
}

Sex
ggcd2pn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    return mkfxat(1);
}

Sex
ggcd2np(Sex a, Sex b)
{
    if (zerop(a))
	return b;
    return mkfxat(1);
}

Sex
ggcd2rr(Sex a, Sex b)
{
    return mkfxat(1);
}

Sex
ggcd2rn(Sex a, Sex b)
{
    if (zerop(b))
	return a;
    return mkfxat(1);
}

Sex
ggcd2nr(Sex a, Sex b)
{
    if (zerop(a))
	return b;
    return mkfxat(1);
}

Sex
ggcd2rp(Sex a, Sex b)
{
    return mkfxat(1);
}

Sex
ggcd2pr(Sex a, Sex b)
{
    return mkfxat(1);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 * #define putv(v,i,a) vector_value(v,i) = a 
 * #define getv(v,i)   vector_value(v,i) 
 */
#define putv(v,i,a)        Lputv(v,mkfxat(i),a)
#define getv(v,i)        Lgetv(v,mkfxat(i))

Sex
gpdivide2(Sex a, Sex b, Sex x)
{
    long            i,
                    j,
                    k;
    long            n,
                    m;
    if (!is_poly(b)) {
	if (!is_poly(a))
	    return cons(a, mkfxat(0));
	n = mkint(power(lterm(poly_pol(a))));
	// printf("a = %p, n = 0x%0x\n", a, n);
	push(a);
	b = Aexpti(b, n);
	pop(a);
	a = gtimes2(a, b);
	return cons(a, mkfxat(0));
    } else if (!is_poly(a)) {
	return cons(mkfxat(0), a);
    }

    stackcheck();

    push5(nil, nil, kernel_order, a, b);
    push4(x, nil, nil, nil);
    push2(nil, nil);

#define q0        stack[-10]
#define cu         stack[-9]
#define ko stack[-8]
#define a stack[-7]
#define b stack[-6]
#define x stack[-5]
#define u stack[-4]
#define v stack[-3]
#define cv stack[-2]
#define p stack[-1]
#define q stack[0]

    if (is_poly(x))
	x = poly_var(x);
    kernel_order = cons(x, kernel_order);

    a = Lreorder(a);
    b = Lreorder(b);
    if (poly_var(a) != x) {
	b = a;
	a = mkfxat(0);
    } else if (poly_var(b) != x) {
	n = mkint(power(lterm(poly_pol(a))));
	b = Aexpti(b, n);
	a = gtimes2(a, b);
	b = mkfxat(0);
    } else {
	cv = poly_pol(a);
	b = poly_pol(b);
	m = mkint(power(lterm(cv)));
	n = mkint(power(lterm(b)));
	k = m - n;
	if (k < 0) {
	    b = a;
	    a = mkfxat(0);
	} else {
	    a = cv;
	    u = Lmkvect(mkfxat(m));
	    for (i = m; i >= 0; i--) {
		if (a != nil && power(lterm(a)) == mkfxat(i)) {
		    putv(u, i, coeff(lterm(a)));
		    a = reduct(a);
		} else
		    putv(u, i, mkfxat(0));
	    }
	    v = Lmkvect(mkfxat(n));
	    for (i = n; i >= 0; i--) {
		if (b != nil && power(lterm(b)) == mkfxat(i)) {
		    putv(v, i, coeff(lterm(b)));
		    b = reduct(b);
		} else
		    putv(v, i, mkfxat(0));
	    }
	    cv = getv(v, n);
	    for (; k >= 0; k--) {
		q0 = cons(getv(u, k + n), q0);
		cu = getv(u, n + k);
		for (j = n + k - 1; j >= k; j--) {
		    /*
		     * putv(u, j, gdiff2(gtimes2(cv, getv(u, j)),
		     * gtimes2(cu, getv(v, j - k)))); 
		     */
		    p = gtimes2(cv, getv(u, j));
		    q = gtimes2(cu, getv(v, j - k));
		    p = gdiff2(p, q);
		    putv(u, j, p);
		}
		for (j = k - 1; j >= 0; j--) {
		    p = gtimes2(cv, getv(u, j));
		    putv(u, j, p);
		}
	    }
	    a = nil;
	    if (q0 != nil) {
		p = cons(car(q0), mkfxat(0));
		a = cons(p, nil);
		q0 = cdr(q0);
		cu = mkfxat(1);
		k = 1;
		while (q0 != nil) {
		    cu = gtimes2(cu, cv);
		    if (!zerop(car(q0))) {
			p = gtimes2(cu, car(q0));
			p = cons(p, mkfxat(k));
			a = cons(p, a);
		    }
		    q0 = cdr(q0);
		    k++;
		}
	    }
	    b = nil;
	    for (k = 0; k < n; k++) {
		if (!zerop(getv(u, k))) {
		    p = cons(getv(u, k), mkfxat(k));
#ifdef DEBUG
		    if (p & 1) {
			printf("p = 0x%08lx, k = %ld, n = %ld, m=%ld\n",
			       p, k, n, m);
		    }
#endif
		    b = cons(p, b);
		}
	    }
	    a = make_poly(x, a);
	    b = make_poly(x, b);
	}
    }
#undef q0
#undef cu
#undef ko
#undef x
#undef u
#undef v
#undef cv
    kernel_order = stack[-8];
#ifndef DEBUG
    a = Lreorder(a);
    b = Lreorder(b);
#endif

#undef a
#undef b
#ifdef DEBUG
    a = stack[-7];
    if (is_poly(a)) {
	p = poly_pol(stack[-7]);
	while (p != nil) {
	    b = lterm(p);
	    // printf(" 0x%08lx : 0x%08lx\n", p, b);
	    if (b & 1)
		printf("gpdivide2: a = 0x%08lx\n", b);
	    if (!is_cons(b))
		break;
	    p = reduct(p);
	}
    }
    a = stack[-6];
    if (is_poly(a)) {
	p = poly_pol(stack[-6]);
	while (p != nil) {
	    b = lterm(p);
	    if (b & 1)
		printf("gpdivide2: b = 0x%08lx\n", b);
	    // printf(" 0x%08lx : 0x%08lx\n", p, b);
	    if (!is_cons(b))
		break;
	    p = reduct(p);
	}
    }
    a = stack[-7];
    a = Lreorder(a);
    stack[-7] = a;
    b = stack[-6];
    b = Lreorder(b);
    a = stack[-7];
#else
    a = stack[-7];
    b = stack[-6];
#endif
#undef p
#undef q
    popv(11);
    return cons(a, b);
}



Sex
gpquotient2(Sex a, Sex b, Sex x)
{
    return car(gpdivide2(a, b, x));
}

Sex
gpremaind2(Sex a, Sex b, Sex x)
{
    return cdr(gpdivide2(a, b, x));
}

Sex
grdivide2(Sex a, Sex b, Sex x)
{
    long            i,
                    j,
                    k;
    long            n,
                    m;

    if (!is_poly(b)) {
	a = gratio2(a, b);
	return cons(a, mkfxat(0));
    }

    if (!is_poly(a)) {
	return cons(mkfxat(0), a);
    }

    stackcheck();

    push2(nil, nil);
    push5(nil, nil, kernel_order, a, b);
    push4(x, nil, nil, nil);

#define p	stack[-10]
#define q	stack[-9]
#define q0        stack[-8]
#define cu         stack[-7]
#define ko stack[-6]
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define u stack[-2]
#define v stack[-1]
#define cv stack[0]

    if (is_poly(x))
	x = poly_var(x);
    kernel_order = cons(x, kernel_order);

    a = Lreorder(a);
    b = Lreorder(b);
    if (poly_var(a) != x) {
	b = a;
	a = mkfxat(0);
    } else if (poly_var(b) != x) {
	a = gratio2(a, b);
	b = mkfxat(0);
    } else {
	cv = poly_pol(a);
	b = poly_pol(b);
	m = mkint(power(lterm(cv)));
	n = mkint(power(lterm(b)));
	k = m - n;
	if (k < 0) {
	    b = a;
	    a = mkfxat(0);
	} else {
	    a = cv;
	    u = Lmkvect(mkfxat(m));
	    for (i = m; i >= 0; i--) {
		if (a != nil && power(lterm(a)) == mkfxat(i)) {
		    putv(u, i, coeff(lterm(a)));
		    a = reduct(a);
		} else
		    putv(u, i, mkfxat(0));
	    }
	    v = Lmkvect(mkfxat(n));
	    for (i = n; i >= 0; i--) {
		if (b != nil && power(lterm(b)) == mkfxat(i)) {
		    putv(v, i, coeff(lterm(b)));
		    b = reduct(b);
		} else
		    putv(v, i, mkfxat(0));
	    }
	    cv = getv(v, n);
	    for (; k >= 0; k--) {
		q0 = cons(getv(u, k + n), q0);
		cu = getv(u, n + k);
		for (j = n + k - 1; j >= k; j--) {
		    p = gtimes2(cv, getv(u, j));
		    q = gtimes2(cu, getv(v, j - k));
		    p = gdiff2(p, q);
		    putv(u, j, p);
		}
		for (j = k - 1; j >= 0; j--) {
		    p = gtimes2(cv, getv(u, j));
		    putv(u, j, p);
		}
	    }
	    a = nil;
	    if (q0 != nil) {
		p = cons(car(q0), mkfxat(0));
		a = cons(p, nil);
		q0 = cdr(q0);
		cu = mkfxat(1);
		k = 1;
		while (q0 != nil) {
		    cu = gtimes2(cu, cv);
		    if (!zerop(car(q0))) {
			p = gtimes2(cu, car(q0));
			p = cons(p, mkfxat(k));
			a = cons(p, a);
		    }
		    q0 = cdr(q0);
		    k++;
		}
	    }
	    b = nil;
	    for (k = 0; k < n; k++) {
		if (!zerop(getv(u, k))) {
		    p = cons(getv(u, k), mkfxat(k));
		    b = cons(p, b);
		}
	    }
	    /*
	     * lisp_printf("grdivide2: cv = %p", cv); lisp_printf("^
	     * %d\n", m-n+1); 
	     */
	    cv = gexpt(cv, mkfxat(m - n + 1));
	    /*
	     * lisp_printf("grdivide2: cv = %p\n", cv); 
	     */
	    a = make_poly(x, a);
	    b = make_poly(x, b);
	    a = gratio2(a, cv);
	    b = gratio2(b, cv);
	}
    }
#undef p
#undef q
#undef q0
#undef cu
#undef ko
#undef x
#undef u
#undef v
#undef cv
    kernel_order = stack[-6];
    a = Lreorder(a);
    b = Lreorder(b);
#undef a
#undef b
    a = stack[-5];
    b = stack[-4];
    popv(11);
    return cons(a, b);
}

Sex
grquotient2(Sex a, Sex b, Sex x)
{
    return car(grdivide2(a, b, x));
}

Sex
grremaind2(Sex a, Sex b, Sex x)
{
    return cdr(grdivide2(a, b, x));
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

bool
lessp2PP(Sex a, Sex b)
{
    long            c;

    c = compare_var(poly_var(a), poly_var(b));
    if (c > 0)
	return YES;
    else if (c < 0)
	return NO;
    a = poly_pol(a);
    b = poly_pol(b);
    if (b == nil)
	return NO;
    if (a == nil)
	return YES;
    return lessp2(power(lterm(a)), power(lterm(b)));
}


bool
lessp2RR(Sex a, Sex b)
{
    return NO;
}


bool
numeq2PP(Sex a, Sex b)
{
    Sex             ca,
                    cb;
    if (poly_var(a) != poly_var(b))
	return NO;
    stackcheck();
    push2(poly_pol(a), poly_pol(b));
#define a stack[-1]
#define b stack[0]
    while (a != nil && b != nil) {
	ca = car(a);
	cb = car(b);
	if (!numeq2(car(ca), car(cb)))
	    break;
	if (!numeq2(cdr(ca), cdr(cb)))
	    break;
	a = cdr(a);
	b = cdr(b);
    }
#undef a
#undef b
    pop2(b, a);
    if (a != b)
	return NO;
    return YES;
}

bool
numeq2RR(Sex a, Sex b)
{
    stackcheck();
    push2(a, b);
    if (numeq2(rpoly_den(a), rpoly_den(b))) {
	pop2(b, a);
	return numeq2(rpoly_num(a), rpoly_num(b));
    }
    popv(2);
    return NO;
}


bool
appeq2PP(Sex a, Sex b)
{
    Sex             ca,
                    cb;
    if (poly_var(a) != poly_var(b))
	return NO;
    stackcheck();
    push2(poly_pol(a), poly_pol(b));
#define a stack[-1]
#define b stack[0]
    while (a != nil && b != nil) {
	ca = car(a);
	cb = car(b);
	if (!appeq2(car(ca), car(cb)))
	    break;
	if (!appeq2(cdr(ca), cdr(cb)))
	    break;
	a = cdr(a);
	b = cdr(b);
    }
#undef a
#undef b
    pop2(b, a);
    if (a != b)
	return NO;
    return YES;
}

bool
appeq2RR(Sex a, Sex b)
{
    stackcheck();
    push2(a, b);
    if (appeq2(rpoly_den(a), rpoly_den(b))) {
	pop2(b, a);
	return appeq2(rpoly_num(a), rpoly_num(b));
    }
    popv(2);
    return NO;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
fixitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lfix(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = fixitp(a);
    pop(c);
    return cons(c, a);
}

Sex
fixitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = fixitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
floatitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lfloat(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = floatitp(a);
    pop(c);
    return cons(c, a);
}

Sex
floatitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = floatitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
bfloatitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lbfloat(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = bfloatitp(a);
    pop(c);
    return cons(c, a);
}

Sex
bfloatitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = bfloatitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
efloatitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lefloat(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = efloatitp(a);
    pop(c);
    return cons(c, a);
}

Sex
efloatitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = efloatitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
ecomplexitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lecomplex(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = ecomplexitp(a);
    pop(c);
    return cons(c, a);
}

Sex
ecomplexitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = intervalitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
intervalitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Linterval(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = intervalitp(a);
    pop(c);
    return cons(c, a);
}

Sex
intervalitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = intervalitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
befloatitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lbefloat(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = befloatitp(a);
    pop(c);
    return cons(c, a);
}

Sex
befloatitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = befloatitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

Sex
becomplexitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lbecomplex(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = becomplexitp(a);
    pop(c);
    return cons(c, a);
}

Sex
becomplexitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = becomplexitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}


Sex
bintervalitp(Sex a)
{
    Sex             c,
                    d;

    if (a == nil)
	return a;
    c = lterm(a);
    a = reduct(a);
    stackcheck();
    push2(a, c);
    d = Lbinterval(coeff(c));
    pop(c);
    c = mkterm(d, power(c));
    a = stack[0];
    stack[0] = c;
    a = bintervalitp(a);
    pop(c);
    return cons(c, a);
}

Sex
bintervalitP(Sex a)
{
    Sex             c;
    stackcheck();
    push(a);
    c = bintervalitp(poly_pol(a));
    pop(a);
    return make_poly(poly_var(a), c);
}

/***************************************************************/
/***************************************************************/
/***************************************************************/

Sex
wuremainder(Sex a, Sex b)
{
    long            i,
                    j,
                    k;
    long            n,
                    m;
    long            o;

    if (!is_poly(b))
	return mkfxat(0);

    if (!is_poly(a))
	return a;

    stackcheck();
    push5(nil, nil, nil, a, b);
    push4(nil, nil, nil, nil);

#define p	stack[-8]
#define q	stack[-7]
#define cu         stack[-6]
#define a stack[-5]
#define b stack[-4]
#define x stack[-3]
#define u stack[-2]
#define v stack[-1]
#define cv stack[0]


    x = poly_var(b);
    u = poly_var(a);
    o = compare_var(x, u);
    if (o < 0) {
	b = a;
	goto owari1;
    } else if (o > 0) {
	b = mkfxat(0);
	goto owari1;
    } else {
	a = poly_pol(a);
	b = poly_pol(b);
	m = mkint(power(lterm(a)));
	n = mkint(power(lterm(b)));
	k = m - n;
	if (k < 0) {
	    b = a;
	    goto owari2;
	}
	u = Lmkvect(mkfxat(m + 1));
	for (i = m; i >= 0; i--) {
	    if (a != nil && power(lterm(a)) == mkfxat(i)) {
		putv(u, i, coeff(lterm(a)));
		a = reduct(a);
	    } else
		putv(u, i, mkfxat(0));
	}
	v = Lmkvect(mkfxat(n + 1));
	for (i = n; i >= 0; i--) {
	    if (b != nil && power(lterm(b)) == mkfxat(i)) {
		putv(v, i, coeff(lterm(b)));
		b = reduct(b);
	    } else
		putv(v, i, mkfxat(0));
	}
	cv = getv(v, n);
	for (; k >= 0; k--) {
	    cu = getv(u, n + k);
	    for (j = n + k - 1; j >= k; j--) {
		p = gtimes2(cv, getv(u, j));
		q = gtimes2(cu, getv(v, j - k));
		p = gdiff2(p, q);
		putv(u, j, p);
	    }
	    for (j = k - 1; j >= 0; j--) {
		p = gtimes2(cv, getv(u, j));
		putv(u, j, p);
	    }
	}
	b = nil;
	for (k = 0; k < n; k++) {
	    a = getv(u, k);
	    if (!zerop(a)) {
		p = cons(a, mkfxat(k));
		b = cons(p, b);
	    }
	}
    }
  owari2:
    b = make_poly(x, b);
  owari1:
#undef p
#undef q
#undef cu
#undef x
#undef u
#undef v
#undef cv
#undef a
#undef b
    b = stack[-4];
    popv(9);
    return b;
}
/***************************************************************/
/***************************************************************/
Sex
Adiff(Sex a, Sex v)
{
    if (is_cons(a))
	return list_diff(a, v);
    if (is_number(a))
	return mkfxat(0);
    if (is_vechdr(a)) {
	Tag             h = vechdr(a);
	switch (type_of_header(h)) {
	case T_ARRAY:
	    return array_diff(a, v);
	case T_SIMPLE:
	    return simple_array_diff(a, v);
	}
	switch (gnumhdr_of(h)) {
	case N_POLY:
	    return poly_diff(a, v);
	case N_DIPOLY:
	    return dipoly_diff(a, v);
	case N_RPOLY:
	    return rpoly_diff(a, v);
	case N_VECT:
	    return vect_diff(a, v);
	    /*
	     * case N_ARRAY: return array_diff(a, v); 
	     */
	case N_MATRIX:
	    return matrix_diff(a, v);
	case N_PREF:
	    return prefix_diff(a, v);
	}
    }
    return mkfxat(0);
}

Sex
list_diff(Sex a, Sex v)
{
    stackcheck();
    push4(a, v, nil, nil);
#define a stack[-3]
#define v stack[-2]
#define p stack[-1]
#define q stack[0]
    while (is_cons(a)) {
	q = Adiff(car(a), v);
	a = cdr(a);
	p = cons(q, p);
    }
#undef a
#undef v
#undef p
#undef q
    a = stack[-1];
    popv(4);
    return Lnreverse(a);
}

Sex
simple_array_diff(Sex a, Sex v)
{
    long            i,
                    siz;
    siz = veclen(a);
    stackcheck();
    push3(a, v, nil);
#define a stack[-2]
#define w stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_SIMPLE, siz);
    siz = bnum_n(siz);
    for (i = 0; i < siz; i++) {
	vector_value(p, i) = Adiff(vector_value(a, i), w);
    }
#undef a
#undef w
#undef p
    a = stack[0];
    popv(3);
    return a;
}

Sex
array_diff(Sex a, Sex v)
{
    long            siz,
                    i;
    Sex             r;

    r = array_dim(a);
    siz = 1;
    while (is_cons(r)) {
	siz = siz * mkint(car(r));
	r = cdr(r);
    }
    stackcheck();
    push3(a, v, nil);
#define a stack[-2]
#define w stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (siz - 1) * sizeof(Sex));
    r = array_dim(a);
    array_dim(p) = r;
    for (i = 0; i < siz; i++) {
	array_value(p, i) = Adiff(array_value(a, i), w);
    }
#undef a
#undef w
#undef p
    a = stack[0];
    popv(3);

    return a;
}



Sex
poly_diff(Sex a, Sex x)
{
    Sex             res;

    stackcheck();
    push5(a, x, nil, nil, nil);

#define a stack[-4]
#define x stack[-3]
#define b stack[-2]
#define c stack[-1]
#define res stack[0]

    res = mkfxat(0);
    if (is_poly(x))
	x = Mvar(x);

    while (is_poly(a)) {
	if (Mvar(a) == x) {
	    b = Minit(a);
	    c = Mdeg(a);
	    b = gtimes2(b, c);
	    if (!gzerop(b)) {
		c = gdiff2(c, mkfxat(1));
		c = make_polynomial(Mvar(a), b, c);
		res = gplus2(res, c);
	    }
	} else {
	    b = Adiff(Minit(a), x);
	    if (!gzerop(b)) {
		c = make_polynomial(Mvar(a), b, Mdeg(a));
		res = gplus2(res, c);
	    }
	}
	a = Mreduct(a);
    }

#undef res
    res = stack[0];
#undef a
#undef x
#undef b
#undef c
    popv(5);
    return res;
}

Sex
dipoly_diff(Sex a, Sex v)
{
    Sex             p,
                    o;
    o = dipoly_order(a);
    p = dipoly_pol(a);
    stackcheck();
    push2(o, v);
    p = Lpolynomial(a);
    pop(v);
    p = Adiff(p, v);
    pop(o);
    a = poly_to_dipoly(p, o);
    return a;
}

Sex
rpoly_diff(Sex a, Sex v)
{
    stackcheck();
    push3(a, v, nil);
#define a stack[-2]
#define v stack[-1]
#define p stack[0]
    p = Adiff(rpoly_num(a), v);
    a = Adiff(rpoly_den(a), v);
    a = gratio2(p, a);
#undef a
#undef v
#undef p
    a = stack[-2];
    popv(3);
    return a;
}

/***************************************************************/
/***************************************************************/
Sex
Areval(Sex a)
{
    if (is_cons(a))
	return list_reval(a);
    if (is_number(a))
	return a;
    if (is_vechdr(a)) {
	Tag             h = vechdr(a);
	switch (type_of_header(h)) {
	case T_ARRAY:
	    return array_reval(a);
	case T_SIMPLE:
	    return simple_array_reval(a);
	}
	switch (gnumhdr_of(h)) {
	case N_POLY:
	    return poly_reval(a);
	case N_DIPOLY:
	    return dipoly_reval(a);
	case N_RPOLY:
	    return rpoly_reval(a);
	case N_VECT:
	    return vect_reval(a);
	    /*
	     * case N_ARRAY: return array_reval(a);
	     */
	case N_MATRIX:
	    return matrix_reval(a);
	}
    }
    return a;
}

Sex
list_reval(Sex a)
{
    stackcheck();
    push5(a, nil, nil, nil, nil);
#define a stack[-4]
#define v stack[-3]
#define b stack[-2]
#define p stack[-1]
#define q stack[0]
    while (is_cons(a)) {
	q = Areval(car(a));
	a = cdr(a);
	p = cons(q, p);
    }
#undef a
#undef v
#undef b
#undef p
#undef q
    a = stack[-1];
    popv(5);
    return Lnreverse(a);
}

Sex
simple_array_reval(Sex a)
{
    long            i,
                    siz;
    siz = veclen(a);
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define w stack[-2]
#define b stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_SIMPLE, siz);
    siz = bnum_n(siz);
    for (i = 0; i < siz; i++) {
	vector_value(p, i) = Areval(vector_value(a, i));
    }
#undef a
#undef w
#undef b
#undef p
    a = stack[0];
    popv(4);
    return a;
}

Sex
array_reval(Sex a)
{
    long            siz,
                    i;
    Sex             r;

    r = array_dim(a);
    siz = 1;
    while (is_cons(r)) {
	siz = siz * mkint(car(r));
	r = cdr(r);
    }
    stackcheck();
    push4(a, nil, nil, nil);
#define a stack[-3]
#define w stack[-2]
#define b stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (siz - 1) * sizeof(Sex));
    r = array_dim(a);
    array_dim(p) = r;
    for (i = 0; i < siz; i++) {
	array_value(p, i) = Areval(array_value(a, i));
    }
#undef a
#undef w
#undef b
#undef p
    a = stack[0];
    popv(4);

    return a;
}


Sex
poly_reval(Sex a)
{
    Sex             w;

    ASSERT(is_poly(a));
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define u stack[-1]
#define v stack[-0]
    v = all_variables(a);
    while (is_cons(v)) {
	u = car(v);
	w = value(u);
	if (w != undefined_var)
	    a = Asub(a, u, w);
	v = cdr(v);
    }
#undef a
#undef u
#undef v
    a = stack[-2];
    popv(3);
    return a;
}

Sex
dipoly_reval(Sex a)
{
    Sex             q,
                    o;
    // o = dipoly_order(a);
    // p = dipoly_pol(a);
    ASSERT(is_dipoly(a));
    stackcheck();
    push2(a, nil);
#define a stack[-1]
#define p stack[-0]
    p = Lpolynomial(a);
    q = Areval(p);
    if (p != q) {
	o = dipoly_order(a);
	a = poly_to_dipoly(q, o);
    }
#undef a
#undef p
    a = stack[-1];
    popv(2);
    return a;
}

Sex
rpoly_reval(Sex a)
{
    int             dirty = NO;
    Sex             r;

    ASSERT(is_rpoly(a));
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define p stack[-1]
#define q stack[0]

    p = rpoly_num(a);
    r = Areval(p);
    if (p != r) {
	dirty = YES;
	p = r;
    }
    q = rpoly_den(a);
    r = Areval(q);
    if (q != r) {
	dirty = YES;
	q = r;
    }
    if (dirty)
	a = gratio2(p, q);
#undef a
#undef p
#undef q
    a = stack[-2];
    popv(3);
    return a;
}

/***************************************************************/
/***************************************************************/

Sex
Asub(Sex a, Sex v, Sex b)
{
    if (is_cons(a))
	return list_subst(a, v, b);
    if (is_number(a))
	return a;
    if (is_vechdr(a)) {
	Tag             h = vechdr(a);
	switch (type_of_header(h)) {
	case T_ARRAY:
	    return array_subst(a, v, b);
	case T_SIMPLE:
	    return simple_array_subst(a, v, b);
	}
	switch (gnumhdr_of(h)) {
	case N_POLY:
	    return poly_subst(a, v, b);
	case N_DIPOLY:
	    return dipoly_subst(a, v, b);
	case N_RPOLY:
	    return rpoly_subst(a, v, b);
	case N_VECT:
	    return vect_subst(a, v, b);
	    /*
	     * case N_ARRAY: return array_subst(a, v, b); 
	     */
	case N_MATRIX:
	    return matrix_subst(a, v, b);
	}
    }
    return a;
}

Sex
list_subst(Sex a, Sex v, Sex b)
{
    stackcheck();
    push5(a, v, b, nil, nil);
#define a stack[-4]
#define v stack[-3]
#define b stack[-2]
#define p stack[-1]
#define q stack[0]
    while (is_cons(a)) {
	q = Asub(car(a), v, b);
	a = cdr(a);
	p = cons(q, p);
    }
#undef a
#undef v
#undef b
#undef p
#undef q
    a = stack[-1];
    popv(5);
    return Lnreverse(a);
}

Sex
simple_array_subst(Sex a, Sex v, Sex b)
{
    long            i,
                    siz;
    siz = veclen(a);
    stackcheck();
    push4(a, v, b, nil);
#define a stack[-3]
#define w stack[-2]
#define b stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_SIMPLE, siz);
    siz = bnum_n(siz);
    for (i = 0; i < siz; i++) {
	vector_value(p, i) = Asub(vector_value(a, i), w, b);
    }
#undef a
#undef w
#undef b
#undef p
    a = stack[0];
    popv(4);
    return a;
}

Sex
array_subst(Sex a, Sex v, Sex b)
{
    long            siz,
                    i;
    Sex             r;

    r = array_dim(a);
    siz = 1;
    while (is_cons(r)) {
	siz = siz * mkint(car(r));
	r = cdr(r);
    }
    stackcheck();
    push4(a, v, b, nil);
#define a stack[-3]
#define w stack[-2]
#define b stack[-1]
#define p stack[-0]
    p = getvector(TAG_VECTOR, T_ARRAY,
		  sizeof(Array_Head) + (siz - 1) * sizeof(Sex));
    r = array_dim(a);
    array_dim(p) = r;
    for (i = 0; i < siz; i++) {
	array_value(p, i) = Asub(array_value(a, i), w, b);
    }
#undef a
#undef w
#undef b
#undef p
    a = stack[0];
    popv(4);

    return a;
}


Sex
poly_subst(Sex a, Sex v, Sex b)
{
    if (!is_cons(v))
	return poly_subst_sub(a, v, b);

    stackcheck();
    push3(a, v, b);
#define a stack[-2]
#define v stack[-1]
#define b stack[-0]
    while (is_cons(v) && is_cons(b)) {
	a = poly_subst_sub(a, car(v), car(b));
	v = cdr(v);
	b = cdr(b);
    }
    if (is_cons(v) || is_cons(b)) {
	popv(3);
	return error("subst: number of variables and values not equal");
    }
#undef a
#undef v
#undef b
    a = stack[-2];
    popv(3);
    return a;
}

Sex
poly_subst_sub(Sex a, Sex v, Sex b)
{

    stackcheck();
    push5(a, b, v, nil, nil);
    push(nil);
#define a stack[-5]
#define b stack[-4]
#define v stack[-3]
#define c stack[-2]
#define d stack[-1]
#define p stack[0]
    if (is_poly(v))
	v = Mvar(v);
    p = mkfxat(0);
    while (is_poly(a)) {
	c = Minit(a);
	c = Asub(c, v, b);
	d = Mdeg(a);
	// lisp_printf("::c = %p", c);
	// lisp_printf(", d=%p\n", d);
	if (Mvar(a) == v) {
	    // lisp_printf(":: b= %p", b);
	    d = gexpt(b, Mdeg(a));
	    // lisp_printf(", d=%p\n", d);
	} else {
	    d = make_polynomial(Mvar(a), mkfxat(1), Mdeg(a));
	}
	// lisp_printf("::d=%p\n", d);
	d = gtimes2(d, c);
	// lisp_printf("::d=%p\n", d);
	p = gplus2(p, d);
	// lisp_printf("::p = %p\n", p);
	a = Mreduct(a);
    }
    p = gplus2(p, a);
#undef a
#undef b
#undef v
#undef c
#undef d
#undef p
    a = stack[0];
    popv(6);
    return a;
}

Sex
dipoly_subst(Sex a, Sex v, Sex b)
{
    Sex             p,
                    o;
    o = dipoly_order(a);
    p = dipoly_pol(a);
    stackcheck();
    push3(o, v, b);
    p = Lpolynomial(a);
    pop2(b, v);
    p = Asub(p, v, b);
    pop(o);
    a = poly_to_dipoly(p, o);
    return a;
}

Sex
rpoly_subst(Sex a, Sex v, Sex b)
{
    stackcheck();
    push4(a, v, b, nil);
#define a stack[-3]
#define v stack[-2]
#define b stack[-1]
#define p stack[0]
    p = Asub(rpoly_num(a), v, b);
    a = Asub(rpoly_den(a), v, b);
    a = gratio2(p, a);
#undef a
#undef b
#undef v
#undef p
    a = stack[-3];
    popv(4);
    return a;
}

/***************************************************************/
/***************************************************************/
Sex
Lresultant(Sex a, Sex b, Sex x)
{
    return Lresultant_bezout2(a, b, x);
    // return Lresultant_poly(a, b, x);
    // return Lresultant_det(a, b, x);
}


static          Sex
Lresultant_det(Sex u, Sex v, Sex w)
{
    long            i,
                    j,
                    n,
                    m;

    Sex             A;
    Sex            *p,
                   *q;

    stackcheck();
    push3(u, v, w);
#define u stack[-2]
#define v stack[-1]
#define w stack[-0]
    n = mkint(Adeg(u, w));
    m = mkint(Adeg(v, w));
    u = Acoeff(u, w);
    v = Acoeff(v, w);
#undef u
#undef v
#undef w
    A = new_matrix(n + m, n + m);
    pop3(w, v, u);
    p = &matrix_value(A, 0);
    j = (n + m) * (n + m);
    for (i = 0; i < j; i++) {
	*p++ = mkfxat(0);
    }


    for (i = 0; i < m; i++) {
	q = &array_value(u, n);
	p = &matrix_value(A, i * (n + m) + i);
	for (j = 0; j <= n; j++)
	    *p++ = *q--;
    }
    for (i = 0; i < n; i++) {
	q = &array_value(v, m);
	p = &matrix_value(A, (i + m) * (n + m) + i);
	for (j = 0; j <= m; j++)
	    *p++ = *q--;
    }
    return Ldeterminant(A);
}

static          Sex
Lresultant_poly(Sex u, Sex v, Sex w)
{
    long            s;
    long            n,
                    m,
                    k;
    long            delta;
    Sex             c;

    s = 1;

    stackcheck();
    push3(u, v, w);
    push3(nil, nil, nil);
    push5(mkfxat(1), mkfxat(1), mkfxat(-1), mkfxat(1), nil);
#define u stack[-10]
#define v stack[-9]
#define w stack[-8]
#define t stack[-7]
#define x stack[-6]
#define z stack[-5]
#define cd stack[-4]
#define cn stack[-3]
#define gam stack[-2]
#define r stack[-1]
#define beta stack[-0]
    m = mkint(Adeg(u, w));
    n = mkint(Adeg(v, w));
    if (n == (0) || m == (0)) {
	popv(11);
	return mkfxat(1);
    } else if (m < n) {
	if (n % 2 == 1 && m % 2 == 1)
	    s = -1;
	t = u;
	u = v;
	v = t;
    }

    while (!ZEROP(v)) {
	delta = mkint(Adeg(u, w)) - mkint(Adeg(v, w));
	beta = gexpt(gam, mkfxat(delta));
	beta = gtimes2(r, beta);
	beta = gnegate(beta);
	r = Ainit(v, w);
	if (delta != 0) {
	    t = gexpt(gam, mkfxat(delta - 1));
	    gam = gexpt(gnegate(r), mkfxat(delta));
	    gam = gratio2(gam, t);
	}
	t = u;
	u = v;
	n = mkint(Adeg(t, w));
	m = mkint(Adeg(u, w));
	if (n % 2 == 1 && m % 2 == 1)
	    s = -s;
	v = gpremaind2(t, v, w);
	v = gratio2(v, beta);
	if (!ZEROP(v)) {
	    z = gexpt(beta, mkfxat(m));
	    cn = gtimes2(cn, z);
	    k = (1 + delta) * m - n + mkint(Adeg(v, w));
	    z = gexpt(r, mkfxat(k));
	    cd = gtimes2(cd, z);
	    x = grquotient2(cd, cn, w);
	    if (!ZEROP(x)) {
		cn = mkfxat(1);
		cd = x;
	    }
	}
    }

    if (Adeg(u, w) != mkfxat(0))
	u = mkfxat(0);
    else if (Adeg(t, w) != mkfxat(1)) {
	n = Adeg(t, w);
	u = gexpt(u, n);
	u = gtimes2(cn, u);
	if (s < 0)
	    u = gnegate(u);
    }
    c = u;
#undef u
#undef v
#undef w
#undef x
#undef t
#undef z
#undef cd
#undef cn
#undef r
#undef gam
#undef beta

    popv(11);
    return c;
}





static          Sex
Lresultant_bezout1(Sex u, Sex v, Sex w)
{
    long            n,
                    m,
                    nm,
                    s;
    int             i,
                    j;
    Sex             t;

    stackcheck();

    push(nil);
    push4(u, v, w, nil);
    push4(nil, nil, nil, nil);
    push2(nil, nil);

#define z stack[-10]
#define u stack[-9]
#define v stack[-8]
#define w stack[-7]
#define x stack[-6]
#define uh stack[-5]
#define ut stack[-4]
#define vh stack[-3]
#define vt stack[-2]
#define ap stack[-1]
#define ep stack[-0]
    if (is_poly(w))
	w = poly_var(w);

    kernel_order = cons(w, kernel_order);
    u = Lreorder(u);
    v = Lreorder(v);
    if (poly_var(u) == w) {
	u = poly_pol(u);
	n = mkint(power(lterm(u)));
	if (n == 0)
	    u = coeff(lterm(u));
    } else
	n = 0;
    if (poly_var(v) == w) {
	v = poly_pol(v);
	m = mkint(power(lterm(v)));
	if (m == 0)
	    v = coeff(lterm(v));
    } else
	m = 0;
    if (n == 0) {
	if (m == 0)
	    t = mkfxat(1);
	else {
	    t = gexpt(u, mkfxat(m));
	}
	kernel_order = cdr(kernel_order);
	popv(11);
	return t;
    }
    if (m == 0) {
	t = gexpt(v, mkfxat(n));
	kernel_order = cdr(kernel_order);
	popv(11);
	return t;
    }
    n = m - n;
    s = 1;
    if (n < 0) {
	if ((n % 2) == 0 && (m % 2) == 1)
	    s = -1;
	t = u;
	u = v;
	v = t;
	n = -n;
	m = n + m;
    }
    ep = mkfxat(1);
    uh = coeff(lterm(u));
    vh = coeff(lterm(v));
    ut = reduct(u);
    x = mkterm(mkfxat(1), mkfxat(n));
    ut = times2tp(x, ut);
    vt = reduct(v);
    x = times2xp(uh, vt);
    ap = times2xp(vh, ut);
    ap = diff2pp(x, ap);
    ap = sf2exb1(ap);
    ep = ext_mult(ap, ep);
    uh = mkterm(uh, mkfxat(0));
    uh = cons(uh, nil);
    vh = mkterm(vh, mkfxat(0));
    vh = cons(vh, nil);

    z = mkterm(mkfxat(1), mkfxat(1));
    for (j = m - 1; j >= n + 1; j--) {
	if (ut != nil && power(lterm(ut)) == mkfxat(j)) {
	    ap = coeff(lterm(ut));
	    x = times2tp(z, uh);
	    uh = plus2xp(ap, x);
	    ut = reduct(ut);
	} else {
	    uh = times2tp(z, uh);
	}
	if (vt != nil && power(lterm(vt)) == mkfxat(j)) {
	    ap = coeff(lterm(vt));
	    x = times2tp(z, vh);
	    vh = plus2xp(ap, x);
	    vt = reduct(vt);
	} else {
	    vh = times2tp(z, vh);
	}
	x = times2pp(uh, vt);
	ap = times2pp(vh, ut);
	ap = plus2pp(x, ap);
	ap = sf2exb1(ap);
	ep = ext_mult(ap, ep);
    }
    if (n != 0) {
	// lisp_printf("u = %p\n", u);
	// lisp_printf("ep = %p\n", ep);
	ap = sf2exb1(u);

	// lisp_printf("ap = %p\n", ap);
	ep = ext_mult(ap, ep);
	for (j = 1; j < n; j++) {
	    x = mkterm(mkfxat(1), mkfxat(j));
	    ap = times2tp(x, u);
	    ap = sf2exb1(ap);
	    ep = ext_mult(ap, ep);
	}
    }
    if (ep == nil)
	t = mkfxat(0);
    else
	t = cdar(ep);
    kernel_order = cdr(kernel_order);
    popv(11);
#undef z
#undef u
#undef v
#undef w
#undef x
#undef uh
#undef ut
#undef vh
#undef vt
#undef ap
#undef ep
    if (s < 0)
	t = gnegate(t);
    return t;
}


static          Sex
sf2exb1(Sex u)
{
    Sex             c;

    if (u == nil)
	return nil;
    stackcheck();
    push4(u, nil, nil, nil);
#define u stack[-3]
#define a stack[-2]
#define b stack[-1]
#define k stack[-0]
    k = power(lterm(u));
    if (k == mkfxat(0)) {
	a = cons(mkfxat(0), nil);
	a = cons(a, coeff(lterm(u)));
	c = cons(a, nil);
    } else {
	b = reduct(u);
	b = sf2exb1(b);
	a = cons(k, nil);
	a = cons(a, coeff(lterm(u)));
	c = cons(a, b);
    }
    popv(4);
#undef u
#undef v
#undef a
#undef b
#undef k
    return c;
}

static          Sex
Lresultant_bezout2(Sex u, Sex v, Sex w)
{
    long            n,
                    m,
                    nm,
                    s;
    int             i,
                    j;
    Sex             t;
    Sex             p;

    stackcheck();
    push4(u, v, w, nil);
    push4(nil, nil, nil, nil);
    push2(nil, nil);

#define u stack[-9]
#define v stack[-8]
#define w stack[-7]
#define x stack[-6]
#define uh stack[-5]
#define ut stack[-4]
#define vh stack[-3]
#define vt stack[-2]
#define ap stack[-1]
#define ep stack[-0]

    n = mkint(Adeg(u, w));
    m = mkint(Adeg(v, w));
    if (n == 0) {
	if (m == 0)
	    t = mkfxat(1);
	else
	    t = gexpt(u, mkfxat(m));
	popv(10);
	return t;
    }
    if (m == 0) {
	t = gexpt(v, mkfxat(n));
	popv(10);
	return t;
    }
    n = m - n;
    s = 1;
    if (n < 0) {
	if ((n % 2) == 0 && (m % 2) == 1)
	    s = -1;
	t = u;
	u = v;
	v = t;
	n = -n;
	m = n + m;
    }
    ep = mkfxat(1);
    uh = Ainit(u, w);
    vh = Ainit(v, w);
    ut = Areduct(u, w);
    x = gexpt(w, mkfxat(n));
    ut = gtimes2(x, ut);
    vt = Areduct(v, w);
    x = gtimes2(uh, vt);
    ap = gtimes2(vh, ut);
    ap = gdiff2(x, ap);
    ap = sf2exb2(ap, w);
    ep = ext_mult(ap, ep);

    for (j = m - 1; j >= n + 1; j--) {
	if (Adeg(ut, w) == mkfxat(j)) {
	    ap = Ainit(ut, w);
	    x = gtimes2(w, uh);
	    uh = gplus2(ap, x);
	    ut = Areduct(ut, w);
	} else {
	    uh = gtimes2(w, uh);
	}
	if (Adeg(vt, w) == mkfxat(j)) {
	    ap = Ainit(vt, w);
	    x = gtimes2(w, vh);
	    vh = gplus2(ap, x);
	    vt = Areduct(vt, w);
	} else {
	    vh = gtimes2(w, vh);
	}
	x = gtimes2(uh, vt);
	ap = gtimes2(vh, ut);
	ap = gdiff2(x, ap);
	ap = sf2exb2(ap, w);
	ep = ext_mult(ap, ep);
    }
    if (n != 0) {
	p = sf2exb2(u, w);
	ep = ext_mult(p, ep);
	for (j = 1; j < n; j++) {
	    x = gexpt(w, mkfxat(j));
	    ap = gtimes2(x, u);
	    ap = sf2exb2(ap, w);
	    ep = ext_mult(ap, ep);
	}
    }
    if (ep == nil)
	t = mkfxat(0);
    else
	t = cdar(ep);
    popv(10);
#undef u
#undef v
#undef w
#undef x
#undef uh
#undef ut
#undef vh
#undef vt
#undef ap
#undef ep
    if (s < 0)
	t = gnegate(t);
    return t;
}


static          Sex
sf2exb2(Sex u, Sex v)
{
    Sex             c;

    if (ZEROP(u))
	return nil;
    stackcheck();
    push5(u, v, nil, nil, nil);
#define u stack[-4]
#define v stack[-3]
#define a stack[-2]
#define b stack[-1]
#define k stack[-0]
    k = Adeg(u, v);
    if (k == mkfxat(0)) {
	a = cons(mkfxat(0), nil);
	a = cons(a, u);
	c = cons(a, nil);
    } else {
	b = Areduct(u, v);
	b = sf2exb2(b, v);
	a = cons(k, nil);
	a = cons(a, Ainit(u, v));
	c = cons(a, b);
    }
    popv(5);
#undef u
#undef v
#undef a
#undef b
#undef k
    return c;
}


static          Sex
ext_mult(Sex u, Sex v)
{
    Sex             c;
    if (u == nil || v == nil)
	return nil;
    if (v == mkfxat(1))
	return u;
    stackcheck();
    push4(u, v, nil, nil);
    push2(nil, nil);
#define u stack[-5]
#define v stack[-4]
#define x stack[-3]
#define y stack[-2]
#define a stack[-1]
#define b stack[-0]
    x = ext_ordexn(caaar(u), caar(v));
    if (x != nil) {
	y = gtimes2(cdar(u), cdar(v));
	if (car(x) != nil)
	    y = gnegate(y);
	a = cons(car(u), nil);
	a = ext_mult(a, cdr(v));
	b = ext_mult(cdr(u), v);
	a = ext_add(a, b);
	y = cons(cdr(x), y);
	c = cons(y, a);
    } else {
	a = ext_mult(cdr(u), v);
	b = cons(car(u), nil);
	b = ext_mult(b, cdr(v));
	c = ext_add(a, b);
    }
#undef u
#undef v
#undef x
#undef y
#undef a
#undef b
    popv(6);
    return c;
}

static          Sex
ext_add(Sex u, Sex v)
{
    Sex             r;

    if (u == nil)
	return v;
    else if (v == nil)
	return u;

    stackcheck();
    push4(u, v, nil, nil);
    push2(nil, nil);
#define u stack[-5]
#define v stack[-4]
#define a stack[-3]
#define b stack[-2]
#define c stack[-1]
#define d stack[-0]
    c = caar(u);
    d = caar(v);
    if (equal(c, d) != nil) {
	a = gplus2(cdar(u), cdar(v));
	b = ext_add(cdr(u), cdr(v));
	if (!ZEROP(a)) {
	    a = cons(c, a);
	    r = cons(a, b);
	} else
	    r = b;
    } else if (ext_ordexp(c, d) != nil) {
	a = ext_add(cdr(u), v);
	r = cons(car(u), a);
    } else {
	a = ext_add(u, cdr(v));
	r = cons(car(v), a);
    }
#undef u
#undef v
#undef a
#undef b
#undef c
#undef d
    popv(6);
    return r;
}


static          Sex
ext_ordexp(Sex u, Sex v)
{
    Sex             r;

    if (u == nil)
	return true;
    else if (car(u) > car(v))
	return true;
    else if (car(u) == car(v))
	return ext_ordexp(cdr(u), cdr(v));
    return nil;
}

static          Sex
ext_ordexn(Sex u, Sex v)
{
    Sex             c;

    stackcheck();
    push2(u, v);
    push3(nil, nil, nil);
#define u stack[-4]
#define v stack[-3]
#define s stack[-2]
#define x stack[-1]
#define w stack[-0]
    s = nil;
    x = nil;

    while (v != nil) {
	w = car(v);
	if (w == u) {
	    popv(5);
	    return nil;
	}
	if (u > w) {
	    u = cons(u, x);
	    u = Lreverse(u);
	    u = Lappend(u, v);
	    c = cons(s, u);
	    popv(5);
	    return c;
	}
	x = cons(w, x);
	v = cdr(v);
	s = s == nil ? true : nil;
    }
    u = Lreverse(cons(u, x));
    c = cons(s, u);
#undef u
#undef v
#undef s
#undef x
#undef w
    popv(5);
    return c;
}

/***************************************************************/
/***************************************************************/
/***************************************************************/


setup_type const polynomial_setup[] = {
    {"setkorder", (lisp_proc *) Lsetkorder, 1},
    {"setorder", (lisp_proc *) Lsetorder, 1},
    {"set_order", (lisp_proc *) Lsetorder, 1},
    {"reorder", (lisp_proc *) Lreorder, 1},
    {"simplify", (lisp_proc *) Lsimplify, 1},
    {"simp", (lisp_proc *) Lsimplify, 1},
    {"symbol_to_polynomial", (lisp_proc *) symbol_to_polynomial, 1},



#ifdef ASUKA
    {"resultant", (lisp_proc *) Lresultant, 3},
    {"resultant_bezout1", (lisp_proc *) Lresultant_bezout1, 3},
    {"resultant_bezout2", (lisp_proc *) Lresultant_bezout2, 3},
    {"resultant_det", (lisp_proc *) Lresultant_det, 3},
    {"resultant_poly", (lisp_proc *) Lresultant_poly, 3},
    {"gquotient", (lisp_proc *) gquot2, 2},
    {"gremainder", (lisp_proc *) gremaind2, 2},
    {"wuremainder", (lisp_proc *) wuremainder, 2},
    {"premainder", (lisp_proc *) gpremaind2, 3},
    {"pquotient", (lisp_proc *) gpquotient2, 3},
    {"pdivide", (lisp_proc *) gpdivide2, 3},
    {"rremainder", (lisp_proc *) grremaind2, 3},
    {"rquotient", (lisp_proc *) grquotient2, 3},
    {"rdivide", (lisp_proc *) grdivide2, 3},
#endif
    {"sub", (lisp_proc *) Asub, 3},
    {"reval", (lisp_proc *) Areval, 1},

    {NULL, 0, 0}
};
