/*
 * rational.c           (C) Copyright 1998-2000, 2007 by Fujio Kako, Nara Women's University.
 * $Header: /Users/kako/asuka/reduce/../src/RCS/rational.c,v 2.5 2015/09/10 23:49:21 kako Exp $
 */

static char const rcsid[] =
    "$Id: rational.c,v 2.5 2015/09/10 23:49:21 kako Exp $";


#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"


Sex             plus2rr_x(Sex a, Sex b);
bool            lessp2xr(Sex a, Sex b);
bool            lessp2rr(Sex a, Sex b);
bool            lessp2rx(Sex a, Sex b);
bool            numeq2rr(Sex a, Sex b);
Sex             my_exp2(long a);
Sex             rationald(Sex a);

extern long     recipim(long n, long m);
extern long     mult2mm(long a, long b, long m);
extern Sex      recipiM(Sex a, Sex m);

/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
Ldenominator(Sex a)
{
    if (!is_number(a))
	return aerror1("denominator", a);
    if (is_numhdr(a) && is_ratio(a))
	return rational_den(a);
    else
	return mkfxat(1);
}

Sex
Lnumerator(Sex a)
{
    if (!is_number(a))
	return aerror1("numerator", a);
    if (is_numhdr(a) && is_ratio(a))
	return rational_num(a);
    else
	return mkfxat(1);
}



Sex
Lrational(Sex a)
{
    Sex             r,
                    i;
    double          dr,
                    di;

    if (is_fixnum(a)) {
	return a;
    } else if (is_numhdr(a)) {
	switch (numtype_of(a)) {
	case N_LOGICAL:
	case N_BIGNUM:
	case N_RATIONAL:
	    return a;
	case N_BFLOAT:
	    return rationald(a);
	case N_EFLOAT:
	    return rationalf(efloat_value(a));
	case N_INTERVAL:
	    return rationalf((interval_lower(a) + interval_upper(a)) *
			     0.5);
	case N_FLOAT:
	    return rationalf(float_val(a));
	case N_BEFLOAT:
	    return rationald(befloat_value(a));
	case N_BINTERVAL:
	    a = plus2dd(binterval_lower(a), binterval_upper(a));
	    a = leftshiftd(a, -1);
	    return rationald(a);
	case N_COMPLEX:
	    r = complex_real(a);
	    i = complex_imag(a);
	    push(i);
	    r = Lrational(r);
	    pop(i);
	    push(r);
	    i = Lrational(i);
	    pop(r);
	    return Lmkcomplex(r, i);
	case N_ECOMPLEX:
	    dr = ecomplex_real(a);
	    di = ecomplex_imag(a);
	    r = rationalf(dr);
	    push(r);
	    i = rationalf(di);
	    pop(r);
	    return Lmkcomplex(r, i);
	case N_BECOMPLEX:
	    r = becomplex_real(a);
	    i = becomplex_imag(a);
	    push(i);
	    r = Lrational(r);
	    pop(i);
	    push(r);
	    i = Lrational(i);
	    pop(r);
	    return Lmkcomplex(r, i);
	default:
	    return mkfxat(0);
	}
    } else {
	return mkfxat(0);
    }
}


Sex
Lmkrational(Sex a, Sex b)
{
    Sex             p = nil;
    if (a == mkfxat(0))
	return mkfxat(0);
    stackcheck();
    push3(a, b, p);
#define a stack[-2]
#define b stack[-1]
#define p stack[0]
    p = gcd(a, b);
    if (p != mkfxat(1)) {
	a = quot2(a, p);
	b = quot2(b, p);
    }
    if (minusp(b)) {
	a = negate(a);
	b = negate(b);
    }
    if (b != mkfxat(1))
	a = make_ratio(a, b);
#undef a
#undef b
#undef p
    popv(2);
    pop(a);
    return a;
}

Sex
make_ratio(Sex p, Sex q)
{
    Sex             v;
    if (q == mkfxat(1))
	return p;
    if (p == mkfxat(0))
	return mkfxat(0);
    if (q == mkfxat(0))
	error("make_ratio: zero divide");
    if (minusp(q))
	error("make_ratio: negative denominator");
    stackcheck();
    push2(p, q);
    v = new_rational();
    pop2(q, p);
    rational_num(v) = p;
    rational_den(v) = q;
    return v;
}


/***************************************************************/
/***************************************************************/
/***************************************************************/


Sex
plus2ir(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = times2(a, rational_den(b));
    a = plus2(a, rational_num(stack[0]));
    pop(b);
    return make_ratio(a, rational_den(b));
}


Sex
plus2rr_x(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = plus2rr(a, b);
    pop2(b, a);
    push3(a, b, c);
    lisp_printf("%p + ", stack[-2]);
    lisp_printf("%p = ", stack[-1]);
    lisp_printf("%p\n", stack[0]);
    pop3(c, b, a);
    return c;
}

/*
 * henrici algorithm
 * 
 * r = (r1/r2); s = (s1/s2); d = gcd(r2, s2); if(d == 1) { t1 =
 * r1*s2+r2*s1; t2=r2*s2} else { r2 =r2/d; s2=s2/d; t1=r1*s2+s1*r2;
 * t2=r2*s2 if(t1 == 0) {t2 = 1} else { e = gcd(t1, d); if(e != 1) { t1 =
 * t1/e; t2 = t2/e} } } return t1/t2; 
 */

Sex
plus2rr(Sex a, Sex b)
{
    Sex             na = rational_num(a),
	nb = rational_num(b);
    Sex             da = rational_den(a),
	db = rational_den(b);
    Sex             w = nil;

    stackcheck();
    push2(a, b);
    if (da == db) {
	push(da);
	na = plus2(na, nb);
	push(na);
	w = gcd(na, stack[-1]);
	if (minusp(w))
	    w = negate(w);
	push(w);
	stack[-1] = quot2(stack[-1], stack[0]);
	stack[-2] = quot2(stack[-2], stack[0]);
	popv(1);
	pop2(na, da);
	if (da == mkfxat(0))
	    aerror3("plus2rr", stack[-1], stack[0], na);
	popv(2);
	return make_ratio(na, da);
    }
    stackcheck();
    push5(na, nb, da, db, nil);
#define g   stack[0]
#define db  stack[-1]
#define da  stack[-2]
#define nb  stack[-3]
#define na  stack[-4]
    g = gcd(da, db);
    db = quot2(db, g);
    g = quot2(da, g);
    na = times2(na, db);
    nb = times2(nb, g);
    na = plus2(na, nb);
    da = times2(da, db);
    g = gcd(na, da);
    na = quot2(na, g);
    da = quot2(da, g);
    if (minusp(da)) {
	da = negate(da);
	na = negate(na);
    }
#undef na
#undef nb
#undef da
#undef db
#undef g
    na = stack[-4];
    da = stack[-2];
    popv(5);
    if (da == mkfxat(0))
	aerror3("plus2rr", stack[-1], stack[0], na);
    popv(2);
    return make_ratio(na, da);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
diff2ir(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = times2(a, rational_den(b));
    a = diff2(a, rational_num(stack[0]));
    pop(b);
    return make_ratio(a, rational_den(b));
}

Sex
diff2ri(Sex a, Sex b)
{
    stackcheck();
    push(a);
    b = times2(rational_den(a), b);
    b = diff2(rational_num(stack[0]), b);
    pop(a);
    return make_ratio(b, rational_den(a));
}

Sex
diff2rr(Sex a, Sex b)
{
    Sex             na = rational_num(a),
	nb = rational_num(b);
    Sex             da = rational_den(a),
	db = rational_den(b);
    Sex             w = nil;

    stackcheck();
    if (da == db) {
	push(da);
	na = diff2(na, nb);
	push(na);
	w = gcd(na, stack[-1]);
	if (minusp(w))
	    w = negate(w);
	push(w);
	stack[-1] = quot2(stack[-1], w);
	stack[-2] = quot2(stack[-2], stack[0]);
	popv(1);
	pop2(na, da);
	return make_ratio(na, da);
    }
    push5(na, nb, da, db, nil);
#define g   stack[0]
#define db  stack[-1]
#define da  stack[-2]
#define nb  stack[-3]
#define na  stack[-4]
    g = gcd(da, db);
    db = quot2(db, g);
    g = quot2(da, g);
    na = times2(na, db);
    nb = times2(nb, g);
    na = diff2(na, nb);
    da = times2(da, db);
    g = gcd(na, da);
    na = quot2(na, g);
    da = quot2(da, g);
    if (minusp(da)) {
	da = negate(da);
	na = negate(na);
    }
    w = make_ratio(na, da);
    popv(5);
    return w;
#undef na
#undef nb
#undef da
#undef db
#undef g
}


/***************************************************************/
/***************************************************************/
/***************************************************************/




Sex
times2ir(Sex a, Sex b)
{
    Sex             w = nil;
    if (a == mkfxat(0))
	return a;
    else if (a == mkfxat(1))
	return b;
    stackcheck();
    push3(b, a, nil);
#define g   stack[0]
#define a   stack[-1]
#define b   stack[-2]
    g = gcd(a, rational_den(b));
    a = quot2(a, g);
    g = quot2(rational_den(b), g);
    a = times2(a, rational_num(b));
    if (minusp(g)) {
	g = negate(g);
	a = negate(a);
    }
    w = make_ratio(a, g);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}

/*
 * Henrici algorithm r = (r1/r2), s=(s1/s2) d1 = gcd(r1, s2); d2 = gcd(r2, 
 * s1); if(d1 != 1) { r1 = r1/d1; s2 = s2/d1} if(d2 != 1) { s1 = s1/d2; r2 
 * = r2/d2} t1 = r1*s1; t2 = r2*s2; return t1/t2; 
 */


Sex
times2rr(Sex a, Sex b)
/*
 * multiply a pair of rational numbers
 */
{
    Sex             w = nil;
    stackcheck();
    push5(rational_num(a), rational_den(a),
	  rational_num(b), rational_den(b), nil);
#define g   stack[0]
#define db  stack[-1]
#define nb  stack[-2]
#define da  stack[-3]
#define na  stack[-4]
    g = gcd(na, db);
    na = quot2(na, g);
    db = quot2(db, g);
    g = gcd(nb, da);
    nb = quot2(nb, g);
    da = quot2(da, g);
    na = times2(na, nb);
    da = times2(da, db);
    w = make_ratio(na, da);
    popv(5);
    return w;
#undef g
#undef db
#undef nb
#undef da
#undef na
}




/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
quot2ir(Sex a, Sex b)
{
    Sex             w;
    if (a == mkfxat(0))
	return a;
    stackcheck();
    push3(b, a, nil);
#define g   stack[0]
#define a   stack[-1]
#define b   stack[-2]
    g = gcd(a, rational_num(b));
    a = quot2(a, g);
    g = quot2(rational_num(b), g);
    a = times2(a, rational_den(b));
    if (minusp(g)) {
	g = negate(g);
	a = negate(a);
    }
    w = make_ratio(a, g);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}


Sex
quot2br(Sex a, Sex b)
{
    Sex             w;

    stackcheck();
    push3(b, a, nil);
#define g   stack[0]
#define a   stack[-1]
#define b   stack[-2]
    g = gcd(a, rational_num(b));
    a = quot2(a, g);
    g = quot2(rational_num(b), g);
    a = times2(a, rational_den(b));
    if (minusp(g)) {
	g = negate(g);
	a = negate(a);
    }
    w = make_ratio(a, g);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}

Sex
quot2ri(Sex a, Sex b)
{
    Sex             w;
    if (b == mkfxat(1))
	return a;
    else if (b == mkfxat(0))
	return aerror2("bad arg for quotient", a, b);
    stackcheck();
    push3(a, b, nil);
#define g   stack[0]
#define b   stack[-1]
#define a   stack[-2]
    g = gcd(b, rational_num(a));
    b = quot2(b, g);
    g = quot2(rational_num(a), g);
    a = times2(b, rational_den(a));
    if (minusp(a)) {
	g = negate(g);
	a = negate(a);
    }
    w = make_ratio(g, a);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}

Sex
quot2rb(Sex a, Sex b)
{
    Sex             w;

    stackcheck();
    push3(a, b, nil);
#define g   stack[0]
#define b   stack[-1]
#define a   stack[-2]
    g = gcd(b, rational_num(a));
    b = quot2(b, g);
    g = quot2(rational_num(a), g);
    a = times2(b, rational_den(a));
    if (minusp(a)) {
	g = negate(g);
	a = negate(a);
    }
    w = make_ratio(g, a);
    popv(3);
    return w;
#undef a
#undef b
#undef g
}


Sex
quot2rr(Sex a, Sex b)
{
    Sex             w;
    stackcheck();
    push5(rational_num(a), rational_den(a), rational_den(b), rational_num(b),	/* 
										 * NB switched order 
										 */
	  nil);
#define g   stack[0]
#define db  stack[-1]
#define nb  stack[-2]
#define da  stack[-3]
#define na  stack[-4]
    g = gcd(na, db);
    na = quot2(na, g);
    db = quot2(db, g);
    g = gcd(nb, da);
    nb = quot2(nb, g);
    da = quot2(da, g);
    na = times2(na, nb);
    da = times2(da, db);
    if (minusp(da)) {
	na = negate(na);
	da = negate(da);
    }
    w = make_ratio(na, da);
    popv(5);
    return w;
#undef g
#undef db
#undef nb
#undef da
#undef na
}



/***************************************************************/
/***************************************************************/
/***************************************************************/



Sex
ratio2ii(Sex a, Sex b)
{
    Sex             p = nil;
    if (a == mkfxat(0))
	return mkfxat(0);
    stackcheck();
    push3(a, b, p);
#define a stack[-2]
#define b stack[-1]
#define p stack[0]
    p = gcd(a, b);
    if (p != mkfxat(1)) {
	a = quot2(a, p);
	b = quot2(b, p);
    }
    if (minusp(b)) {
	a = negate(a);

	b = negate(b);

    }
    if (b != mkfxat(1)) {





	a = make_ratio(a, b);
    }
#undef a
#undef b
#undef p
    popv(2);
    pop(a);
    return a;
}

Sex
my_exp2(long s)
{
    Sex             a;
    long            m;
    long            i;


    m = s >> LP_LOG2;
    s = s & (LPSIZE - 1);

    a = new_binteger(T_BIGNUM, bnum_len(m + 1));

    for (i = 0; i < m; i++)
	bignum_value(a)[i] = 0;

    i = 1L << s;
    if (m == 0 && (i & fix_mask) == 0)
	return mkfxat(i);
    bignum_value(a)[m] = i;
    return a;
}


/***************************************************************/
/***************************************************************/

Sex
mod2ri(Sex a, Sex b)
{
    long            m = mkint(b);
    long            k;
    long            n;

    push(a);
    a = rational_num(a);
    a = remaind2(a, mkfxat(m));
    n = mkint(a);
    if (n < 0)
	n = n + m;
    pop(a);
    a = rational_den(a);
    a = remaind2(a, mkfxat(m));
    k = mkint(a);
    k = recipim(k, m);
    k = mult2mm(k, n, m);
    return mkfxat(k);
}

Sex
mod2rb(Sex a, Sex m)
{
    Sex             b;
    b = rational_den(a);
    a = rational_num(a);
    push2(a, m);
    b = recipiM(b, m);
    a = stack[-1];
    a = times2jj(a, b);
    m = stack[0];
    a = rem2jj(a, m);
    popv(2);
    return a;
}

Sex
smod2ri(Sex a, Sex b)
{
    long            k;
    long            m,
                    m2;
    long            n;

    m = mkint(b);
    m2 = m / 2;
    push(a);
    a = rational_num(a);
    a = remaind2(a, mkfxat(m));
    n = mkint(a);
    pop(a);
    a = rational_den(a);
    a = remaind2(a, mkfxat(m));
    k = mkint(a);
    k = recipim(k, m);
    n = mult2mm(k, n, m);
    if (n > m2)
	n = n - m;
    else if (n <= -m2)
	n = n + m;
    return mkfxat(n);
}

Sex
smod2rb(Sex a, Sex m)
{
    Sex             b,
                    m2;

    stackcheck();
    b = rational_den(a);
    a = rational_num(a);
    push2(a, m);
    b = recipiM(b, m);
    a = stack[-1];
    a = times2jj(a, b);
    m = stack[0];
    a = rem2jj(a, m);
    stack[-1] = a;
    m = stack[0];
    m2 = quot2(m, mkfxat(2));
    push(m2);
#define a stack[-2]
#define m stack[-1]
#define m2 stack[0]
    if (lessp2(m2, a))
	a = diff2(a, m);
    else {
	m2 = negate(m2);
	if (lessp2(a, m2))
	    a = plus2(a, m);
    }
#undef a
#undef m
#undef m2
    popv(3);
    a = stack[-2];
    return a;
}

/***************************************************************/
/***************************************************************/


Sex
rationalf(double d)
{
    Sex             a;
    int             sign = 0;

    long            m;
    double          er;
    double          eq0,
                    eq1,
                    eq2;


    if (d == 0.0)
	return mkfxat(0);

    if (d > 1.0e7)
	return fixitf(d + 0.5);
    else if (d < -1.0e7)
	return fixitf(d - 0.5);
    else if (fabs(d) < 1.0e-7) {
	if (d < 0) {
	    a = fixitf(-1 / d + 0.5);
	    return make_ratio(mkfxat(-1L), a);
	} else {
	    a = fixitf(1 / d + 0.5);
	    return make_ratio(mkfxat(1L), a);
	}
    }
    if (d < 0) {
	sign = 1;
	d = -d;
    }
    er = d * 1.0e-15;

    stackcheck();
    push5(nil, nil, nil, nil, nil);
#define p0 stack[-0]
#define p1 stack[-1]
#define q0 stack[-2]
#define q1 stack[-3]
#define t0 stack[-4]
    p0 = mkfxat(1);
    q0 = mkfxat(0);
    m = (long) d;
    p1 = mkfxat(m);
    q1 = mkfxat(1);
    d = d - m;
    eq0 = 0;
    eq1 = 1;
    while (d > 1.e-10 && er * eq0 * eq1 < 1) {
	d = 1 / d;
	if (d > 1.0e+7) {
	    t0 = fixitf(d + 0.5);
	    p1 = times2(p1, t0);
	    p1 = plus2(p1, p0);
	    q1 = times2(q1, t0);
	    q1 = plus2(q1, q0);
	    break;
	}
	m = (long) d;
	d = d - m;
	t0 = times2(mkfxat(m), p1);
	t0 = plus2(t0, p0);
	p0 = p1;
	p1 = t0;
	t0 = times2(mkfxat(m), q1);
	t0 = plus2(t0, q0);
	q0 = q1;
	q1 = t0;
	eq2 = eq1 * m + eq0;
	eq0 = eq1;
	eq1 = eq2;
    }
    if (sign)
	p1 = negate(p1);
    if (q1 == mkfxat(1))
	a = p1;
    else {
	a = new_rational();
	rational_num(a) = (p1);
	rational_den(a) = (q1);
    }
#undef p0
#undef p1
#undef q0
#undef q1
#undef t0
    popv(5);
    return a;
}

Sex
rationald(Sex a)
{
    double          d;
    int             sign = 0;

    Sex             m;
    double          mm;
    double          er;
    double          eq0,
                    eq1,
                    eq2;
    double          pr,
                    pri;


    d = floatitd(a);
    if (d == 0.0)
	return mkfxat(0);

    pr = ldexp(1.0, current_b_precision / 2);
    pri = 1.0 / pr;


    if (d > pr || d < -pr)
	return fixitd(a);
    else if (fabs(d) < pri) {
	if (d < 0) {
	    a = fixitd(quot2(mkfxat(-1), a));
	    return make_ratio(mkfxat(-1L), a);
	} else {
	    a = fixitd(quot2(mkfxat(1), a));
	    return make_ratio(mkfxat(1L), a);
	}
    }
    if (d < 0) {
	sign = 1;
	d = -d;
	a = negated(a);
    }
    er = d * pri * pri;

    stackcheck();
    push5(a, nil, nil, nil, nil);
    push2(nil, nil);

#define p0 stack[-0]
#define p1 stack[-1]
#define q0 stack[-2]
#define q1 stack[-3]
#define t0 stack[-4]
#define m stack[-5]
#define a stack[-6]
    p0 = mkfxat(1);
    q0 = mkfxat(0);
    m = fixitd(a);
    mm = floatit(m);
    p1 = (m);
    q1 = mkfxat(1);
    a = diff2(a, m);
    d = floatitd(a);
    eq0 = 0;
    eq1 = 1;
    while (d > pri && er * eq0 * eq1 < 1) {
	a = quot2(mkfxat(1), a);
	d = floatitd(a);
	if (d > pr) {
	    t0 = fixitd(a);
	    p1 = times2(p1, t0);
	    p1 = plus2(p1, p0);
	    q1 = times2(q1, t0);
	    q1 = plus2(q1, q0);
	    break;
	}
	m = fixitd(a);
	mm = floatit(m);
	a = diff2(a, m);
	d = floatitd(a);
	t0 = times2((m), p1);
	t0 = plus2(t0, p0);
	p0 = p1;
	p1 = t0;
	t0 = times2((m), q1);
	t0 = plus2(t0, q0);
	q0 = q1;
	q1 = t0;
	eq2 = eq1 * mm + eq0;
	eq0 = eq1;
	eq1 = eq2;
    }
    if (sign)
	p1 = negate(p1);
    if (q1 == mkfxat(1))
	a = p1;
    else {
	a = new_rational();
	rational_num(a) = (p1);
	rational_den(a) = (q1);
    }
#undef p0
#undef p1
#undef q0
#undef q1
#undef t0
#undef m
#undef a
    a = stack[-6];
    popv(7);
    return a;
}


#if 0
Sex
rationald(Sex a)
{
    Sex             b,
                    c;
    long            exp;

    exp = bfloat_exp(a);
    a = bfloat_man(a);
    if (exp > 0)
	return leftshift(a, exp);
    else if (exp == 0)
	return a;
    stackcheck();
    push(a);
    b = my_exp2(-exp);
    push(b);
    c = new_rational();
    pop2(b, a);
    rational_num(c) = a;
    rational_den(c) = b;
    return c;
}
#endif


Sex
rational(Sex a)
{
    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	return a;
    case TAG_NUMBERS:
	{
	    long            ha = type_of_header(numhdr(a));
	    switch (ha) {
	    case T_BIGNUM:
		return a;
	    case T_FLOAT:
		return rationalf(floatit(a));
	    case T_BFLOAT:
		return rationald(a);
	    default:
		return aerror1("bad arg for rational", a);
	    }
	}
    default:
	return aerror1("bad arg for rational", a);
    }
}



/***************************************************************/
/***************************************************************/
/***************************************************************/

/*
 * Arithmetic comparison: lessp2
 */


bool
lessp2xr(Sex a, Sex b)
{
    /*
     * compute a < p/q  as a*q < p
     */
    stackcheck();
    push(rational_num(b));
    a = times2(a, rational_den(b));
    pop(b);
    return lessp2(a, b);
}



bool
lessp2rr(Sex a, Sex b)
{
    Sex             c;
    stackcheck();
    push2(a, b);
    c = times2(rational_num(a), rational_den(b));
    pop2(b, a);
    push(c);
    b = times2(rational_num(b), rational_den(a));
    pop(c);
    return lessp2(c, b);
}



bool
lessp2rx(Sex a, Sex b)
{
    stackcheck();
    push(rational_num(a));
    b = times2(b, rational_den(a));
    pop(a);
    return lessp2(a, b);
}


/***************************************************************/
/***************************************************************/
/***************************************************************/



bool
numeq2rr(Sex a, Sex b)
{
    return (numeq2(rational_num(a), rational_num(b)) &&
	    numeq2(rational_den(a), rational_den(b)));
}


/***************************************************************/
/***************************************************************/
/***************************************************************/
