/*
 * tags.h      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 *      $Header: /Users/kako/asuka/src/RCS/tags.h,v 2.13 2017/11/27 00:45:12 kako Exp $
 */

#include <stdint.h>

// typedef long bool;
// typedef long Sex;
#ifdef _LP64
typedef int64_t bool;
typedef int64_t Sex;
#else
typedef int32_t bool;
typedef int32_t Sex;
#endif

#define YES                 1
#define NO                  0

#define CELL	(sizeof(Sex)*2)
#define HDR	(sizeof(Sex)*2)
#define SEX	(sizeof(Sex))


/*
 * Tag codes in every Lisp Object
 *       if LP == 32
 *     x x x | x 0 0 0      cons cell
 *     x x x | x 0 0 1      for fwd relocation
 *     x x x | G 0 1 0      vec-hdrs and others
 *     x x x | G 0 1 1      28-bit integer
 *     x x x | x 1 0 0      a symbol
 *     x x x | x 1 0 1      binary data
 *     x x x | x 1 1 0      Lisp vector or array
 *     x x x | x 1 1 1      bignum. {rational number, complex number}
 *
 *       if LP == 64
 *     x x x | 0 0 0 0      cons cell
 *     x x x | x 0 0 1      for fwd relocation
 *     x x x | G 0 1 0      vec-hdrs and others
 *     x x x | G 0 1 1      60-bit integer
 *     x x x | 0 1 0 0      a symbol
 *     x x x | 0 1 0 1      binary data 
 *     x x x | 0 1 1 0      vector or array
 *     x x x | 0 1 1 1      bignum. {rational number, complex number}
 */


#define TAG_BITS        7

#define TAG_CONS        0	/* Cons cells */
#define TAG_RELOC       1	/* For FWD Relocation */
#define TAG_HDRS        2	/* Vechdrs etc */
#define TAG_FIXNUM      3	/* 60-bit integers */
#define TAG_SYMBOL      4	/* Symbols */
#define TAG_BINARY      5	/* String vector */
#define TAG_VECTOR      6	/* Regular Lisp vectors */
#define TAG_NUMBERS     7	/* Bignum, Rational, Complex */

#define tag_type(p)  (((long)(p)) & TAG_BITS)

#define is_cons(p)   ((((long)(p)) & TAG_BITS) == TAG_CONS)
#define is_fixnum(p) ((((long)(p)) & TAG_BITS) == TAG_FIXNUM)
#define is_hdrs(p)   ((((long)(p)) & TAG_BITS) == TAG_HDRS)
#define is_relhdr(p)  ((((long)(p)) & 3) == TAG_RELOC)

#define is_symhdr(p) ((((long)(p)) & TAG_BITS) == TAG_SYMBOL)
#define is_symbol(p)	is_symhdr(p)
#define is_numhdr(p)((((long)(p)) & TAG_BITS) == TAG_NUMBERS)
#define is_vechdr(p) ((((long)(p)) & TAG_BITS) == TAG_VECTOR)
#define is_binhdr(p) ((((long)(p)) & TAG_BITS) == TAG_BINARY)

#define is_number(p) ((((long)(p)) & 3) == 3)
#define is_immed(p)  ((((long)(p)) & 6) == 2)

#define mkfxat(x)    ((Sex)(TAG_FIXNUM | (((unsigned long)x) << 4)))

#define mkint(x)    ((long)(((long)(x)) >> 4))



/*********************************************************
 *
 * Here we assume that every pointer has it top_bit clear 
 * 
 *********************************************************/

#ifdef _LP64
#define GC_BIT_I        (8L)	/* Used with FIXNUM */
#define GC_BIT_P        (0x8000000000000000L)	/* Used with pointer types 
						 */
#else
#define GC_BIT_I        (8L)	/* Used with FIXNUM */
#define GC_BIT_P        (0x80000000L)	/* Used with pointer types */
#endif

#ifdef _LP64

#define is_i_marked(w)      (((long)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      (((long)(w) & GC_BIT_P) != 0)

#define set_p_mark(w)   ((Sex)((long)(w) | GC_BIT_P))
#define reset_p_mark(w) ((Sex)((long)(w) & ~GC_BIT_P))

#define set_i_mark(h)   ((Sex)((long)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((long)(h) & ~GC_BIT_I))

#else

#if MEMORY_MODEL == MEMORY_LOW
/*
 * malloc memory address is low end 
 */

#define is_i_marked(w)      (((int)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      ((int)(w) < 0)

#define set_p_mark(w)   ((Sex)((int)(w) | GC_BIT_P))
#define reset_p_mark(w) ((Sex)((int)(w) & ~GC_BIT_P))

#define set_i_mark(h)   ((Sex)((int)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((int)(h) & ~GC_BIT_I))

#else
#if MEMORY_MODEL == MEMORY_HIGH
/*
 * malloc memory address is above 0x80000000 
 */
#define is_i_marked(w)      (((int)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      ((int)(w) > 0)

#define set_p_mark(w)   ((Sex)((int)(w) & GC_BIT_P))
#define reset_p_mark(w) ((Sex)((int)(w) | ~GC_BIT_P))

#define set_i_mark(h)   ((Sex)((int)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((int)(h) & ~GC_BIT_I))

#else
#if MEMORY_MODEL == MEMORY_MIXED
/*
 * malloc memory address is mix of low and high end 
 */
/*
 * Linux kernel 2.6 seems this case 
 */

#define is_i_marked(w)      (((int)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      mixed_is_p_marked((int)w)

#define set_p_mark(w)   mixed_set_p_mark((int)w)
#define reset_p_mark(w) mixed_reset_p_mark((int)w)

#define set_i_mark(h)   ((Sex)((int)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((int)(h) & ~GC_BIT_I))

#else
#if MEMORY_MODEL == MEMORY_VARIABLE
/*
 * malloc memory address is mix of low and high end 
 */
/*
 * Linux kernel 2.6 seems this case 
 */

#define is_i_marked(w)      (((int)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      variable_is_p_marked((int)w)

#define set_p_mark(w)   variable_set_p_mark((int)w)
#define reset_p_mark(w) variable_reset_p_mark((int)w)

#define set_i_mark(h)   ((Sex)((int)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((int)(h) & ~GC_BIT_I))

#else

#define is_i_marked(w)      (((long)(w) & GC_BIT_I) != 0)
#define is_p_marked(w)      (((long)(w) & GC_BIT_P) != 0)

#define set_p_mark(w)   ((Sex)((long)(w) | GC_BIT_P))
#define reset_p_mark(w) ((Sex)((long)(w) & ~GC_BIT_P))

#define set_i_mark(h)   ((Sex)((long)(h) | GC_BIT_I))
#define reset_i_mark(h) ((Sex)((long)(h) & ~GC_BIT_I))

#endif

#endif
#endif
#endif
#endif

/*********************************************************/
/*********************************************************/


typedef struct Cons_Cell {
    Sex             kar;
    Sex             kdr;
} Cons_Cell;


#define car(p) (((Cons_Cell *) (p))->kar)
#define cdr(p) (((Cons_Cell *) (p))->kdr)
#define cadr(p)  car(cdr(p))
#define caddr(p) car(cdr(cdr(p)))

typedef Sex     Special(Sex, Sex);

typedef Sex     lisp_proc(void);


typedef Sex     zero_args(void);
typedef Sex     one_args(Sex);
typedef Sex     two_args(Sex, Sex);
typedef Sex     three_args(Sex, Sex, Sex);
typedef Sex     n_args4(Sex, Sex, Sex, Sex);
typedef Sex     n_args5(Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args6(Sex, Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args7(Sex, Sex, Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args8(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args9(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args10(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex);
typedef Sex     n_args11(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex,
			 Sex);
typedef Sex     n_args12(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex,
			 Sex, Sex);
typedef Sex     n_args13(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex,
			 Sex, Sex, Sex);
typedef Sex     n_args14(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex,
			 Sex, Sex, Sex, Sex);
typedef Sex     n_args15(Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex, Sex,
			 Sex, Sex, Sex, Sex, Sex);



typedef unsigned long Tag;


#define header_mask          0xffff0
#define type_of_header(h)    (((unsigned long)(h)) & header_mask)

/*
 * 0 0 | * 0 1 0     string header 
 * 0 1 | * 0 1 0     symbol header 
 * 1 0 | * 0 1 0     number header 
 * 1 1 | * 0 1 0     vector header 
 */

/*
 *   P S O C | V t p m | s e b c | g f 0 1 | * 0 1 0       symbol header
 *   0 0 0 0 | 0 0 0 0 | 0 0 0 0 | g f 0 1 | * 0 1 0       undefined
 *   0 0 0 0 |     1 0 | 0 0 0 1 | g f 0 1 | * 0 1 0       cfunc
 *   0 0 0 0 |     1 0 | 0 1 0 0 | g f 0 1 | * 0 1 0       expr
 *   0 0 0 0 |     1 0 | 0 1 0 1 | g f 0 1 | * 0 1 0       subr
 *   0 0 0 0 |     1 0 | 0 1 1 0 | g f 0 1 | * 0 1 0       code
 *   0 0 0 0 |     1 0 | 1 0 0 0 | g f 0 1 | * 0 1 0       fexpr
 *   0 0 0 0 |     1 0 | 1 0 0 1 | g f 0 1 | * 0 1 0       fsubr
 *   0 0 0 0 |     1 0 | 1 0 1 0 | g f 0 1 | * 0 1 0       fcode
 *   0 0 0 0 |     1 1 | 0 0 0 0 | g f 0 1 | * 0 1 0       macro
 *   0 0 0 0 |     1 1 | 0 0 0 1 | g f 0 1 | * 0 1 0       cmacro
 *   0 0 0 0 |     1 1 | 0 0 1 0 | g f 0 1 | * 0 1 0       mcode
 *   0 0 0 0 |     1 1 | 0 1 0 0 | g f 0 1 | * 0 1 0       smacro
 *   0 0 0 0 |     1 1 | 0 1 0 1 | g f 0 1 | * 0 1 0       csmacro
 *   0 0 0 0 |     1 1 | 0 1 1 0 | g f 0 1 | * 0 1 0       scode
 *
 *   0 0 0 0 | 1 0 0 0 | 0 0 0 0 | g f 0 1 | * 0 1 0       algebraic value
 *   1 1 0 0 | 0 0 0 0 | 0 0 0 0 | g f 0 1 | * 0 1 0       algebraic subr
 *   1 0 1 0 | 0 0 0 0 | 0 0 0 0 | g f 0 1 | * 0 1 0       algebraic operator
 *   1 0 0 1 | 0 0 0 0 | 0 0 0 0 | g f 0 1 | * 0 1 0       compiled func
 *
 *         p   proc is defined
 *         t   trace flag
 *         c   compiled code (fsubr, subr)
 *         b   byte code (fsubr, subr)
 *         m   symbol is defined as a macro
 *         s   symbol names a special form (fsubr)
 *         e   symbol names a usual function (or smacro)
 *         g   global variable
 *         f   fluid variable
 */

#define  SYM_FLUID          0x00000040	/* (fluid '(xxx)) */
#define  SYM_GLOBAL         0x00000080	/* (global '(xxx)) */
#define  PROC_MASK          0x00003f00	/* mask bit */
#define  PROC_CFUNC         0x00002100	/* C function */
#define  PROC_EXPR          0x00002400	/* expr */
#define  PROC_SUBR          0x00002500	/* subr */
#define  PROC_CODE          0x00002600	/* byte code */
#define  PROC_FEXPR         0x00002800	/* fexpr */
#define  PROC_FSUBR         0x00002900	/* fsubr or special */
#define  PROC_FCODE         0x00002a00	/* bytecode (special) */
#define  PROC_MACRO         0x00003000	/* macro */
#define  PROC_CMACRO        0x00003100	/* compiled macro */
#define  PROC_MCODE         0x00003200	/* bytecode macro */
#define  PROC_SMACRO        0x00003400	/* symbolic macro */
#define  PROC_CSMACRO       0x00003500	/* compiled symbolic macro */
#define  PROC_SCODE         0x00003600	/* bytecode symbolic macro */
#define  PROC_CCODE         0x00000100	/* procedure is compiled * code */
#define  PROC_BCODE         0x00000200	/* procedure is compiled * byte
					 * code */
#define  SYM_PROC           0x00003f00
#define  SYM_TRACED         0x00004000

#define  SYM_AVALUE	    0x00008000
#define  GPROC_SUBR	    0x000c0000
#define  GPROC_CSUBR	    0x000d0000
#define  GPROC_OPER	    0x000b0000
#define  GPROC_CCODE	    0x00010000
#define  SYM_APROC	    0x000f0000


#define  CFUNC_INT		1	/* return value is integer */
#define  CFUNC_FLOAT		2	/* return value is float */
#define	 CFUNC_STRING		3	/* return value is string */
#define	 CFUNC_INTVEC		4	/* return value is integer *
					 * vector */
#define	 CFUNC_FLOATVEC		5	/* return value is float vector */

#define T_NEG         	   0x0100
#define S_NEG		   0x0008
#define T_FIX	   	   0x0200
#define T_LOG		   0x0400
#define T_APP	   	   0x1000
#define T_CMP		   0x2000
#define T_BIG		   0x0800
#define T_NUM		   0x0020
#define T_EFF		   0x4000
#define T_POL		   0x8000

#define T_VEC		   0x0030
#define T_CANO		   0x1000
#define T_FORM		   0x2000
#define T_PREF		   0x4000
#define T_EXPR		   0x8000

/*
 *                     PECA BLFN
 *                0000 0000 1100 0010 0010         logical
 *                0001 0000 1010 0010 0010         bignum
 *                0001 0000 1011 0010 0010         bignum (negative)
 *                0010 0000 0100 0010 0010         modular
 *                0011 0000 0000 0010 0010         rational
 *                0100 0001 0000 0010 0010         float
 *                0101 0101 0000 0010 0010         efloat
 *                0110 0010 0000 0010 0010         complex
 *                0111 0001 0000 0010 0010         interval
 *                1000 0001 1000 0010 0010         bfloat
 *                1001 0111 0000 0010 0010         ecomplex
 *                1010 0101 1000 0010 0010         befloat
 *                1011 0111 1000 0010 0010         becomplex
 *                1100 0001 1000 0010 0010         binterval
 *                1101 0000 1100 0010 0010         bmodular
 *                1110 0000 0100 0010 0010         qmodular
 *                1111 0000 0100 0010 0010         fmodular
 *
 *                0001 0001 0110 1111 0010         variable
 *                0010 0001 0110 1111 0010         algebraic number
 *                0011 0001 0110 1111 0010         transendential number
 *                0100 0001 0110 1111 0010         polynomial
 *                0101 0001 0110 1111 0010         distributed polynomial
 *                0110 0001 0110 1111 0010         algebraic function
 *                0111 0001 0110 1111 0010         rational function
 *                1000 0010 0110 1111 0010         vector
 *                1001 0010 0110 1111 0010         matrix
 *                1010 0010 0110 1111 0010         tensor
 *                1011 0100 0110 1111 0010         prefix expression
 *                1100 1000 0110 1111 0010         relations
 *                1101 1000 0110 1111 0010         set expression
 *                1110 1000 0110 1111 0010         logical expression
 *                1111 1000 0110 1111 0010         tropical algebra
 
 *                0000 0000 0001 1111 0010         simple vector
 *                0000 0000 0010 1111 0010         array
 *                0000 0000 0011 1111 0010         v_array (aka. list)
 *                0000 0000 0100 1111 0010         package
 *                0000 0000 0101 1111 0010         hash table
 *                0000 0000 0110 1111 0010         gal structure
 */

#define N_LOGICAL          0
#define N_BIGNUM           1
#define N_MODULAR          2
#define N_RATIONAL         3
#define N_FLOAT            4
#define N_EFLOAT	   5
#define N_COMPLEX          6
#define N_INTERVAL         7
#define N_BFLOAT           8
#define N_ECOMPLEX         9
#define N_BEFLOAT          10
#define N_BECOMPLEX        11
#define N_BINTERVAL        12
#define N_BMODULAR         13
#define N_QMODULAR         14
#define N_FMODULAR         15
#define TYPEMAX	   	   16

#define N_NON		   0	/* non g-num object */
#define N_VAR		   1
#define N_ALGNUM	   2
#define N_TRNNUM	   3
#define N_POLY             4
#define N_DIPOLY           5
#define N_ALGFUN           6
#define N_RPOLY            7
#define N_VECT             8
#define N_MATRIX           9
#define N_TENSOR           10
#define N_PREF             11
#define N_RELEXP           12
#define N_SETEXP           13
#define N_LOGEXP           14
#define N_TROPICAL	   15
#define GTYPEMAX	   16


#define T_LOGICAL          ((N_LOGICAL  <<16)|T_BIG|T_LOG|      T_NUM)
#define T_BIGNUM           ((N_BIGNUM   <<16)|T_BIG|T_FIX|      T_NUM)
#define T_BIGNEG           ((N_BIGNUM   <<16)|T_BIG|T_FIX|T_NEG|T_NUM)
#define T_RATIONAL         ((N_RATIONAL <<16)|                  T_NUM)
#define T_MODULAR          ((N_MODULAR  <<16)|      T_LOG|      T_NUM)
#define T_FLOAT            ((N_FLOAT    <<16)|T_APP|            T_NUM)
#define T_EFLOAT	   ((N_EFLOAT   <<16)|T_APP|            T_NUM|T_EFF)
#define T_COMPLEX          ((N_COMPLEX  <<16)|            T_CMP|T_NUM)
#define T_INTERVAL         ((N_INTERVAL <<16)|T_APP|            T_NUM)
#define T_BFLOAT           ((N_BFLOAT   <<16)|T_APP|T_BIG|      T_NUM)
#define T_ECOMPLEX         ((N_ECOMPLEX <<16)|T_APP|      T_CMP|T_NUM|T_EFF)
#define T_BEFLOAT          ((N_BEFLOAT  <<16)|T_APP|T_BIG|      T_NUM|T_EFF)
#define T_BECOMPLEX        ((N_BECOMPLEX<<16)|T_APP|T_BIG|T_CMP|T_NUM|T_EFF)
#define T_BINTERVAL        ((N_BINTERVAL<<16)|T_APP|T_BIG|      T_NUM)
#define T_BMODULAR         ((N_BMODULAR <<16)|T_BIG|T_LOG|      T_NUM)
#define T_QMODULAR         ((N_QMODULAR <<16)|      T_LOG|      T_NUM)
#define T_FMODULAR         ((N_FMODULAR <<16)|      T_LOG|      T_NUM)

#define T_VAR    	   ((N_VAR    << 16) | T_CANO|   T_GAL|T_VEC)
#define T_ALGNUM	   ((N_ALGNUM << 16) | T_CANO|   T_GAL|T_VEC)
#define T_TRNNUM	   ((N_TRNNUM << 16) | T_CANO|   T_GAL|T_VEC)
#define T_POLY		   ((N_POLY   << 16) | T_CANO|   T_GAL|T_VEC)
#define T_DIPOLY	   ((N_DIPOLY << 16) | T_CANO|   T_GAL|T_VEC)
#define T_ALGFUN	   ((N_ALGFUN << 16) | T_CANO|   T_GAL|T_VEC)
#define T_RPOLY	   	   ((N_RPOLY  << 16) | T_CANO|   T_GAL|T_VEC)
#define T_VECT	   	   ((N_VECT   << 16) | T_FORM|   T_GAL|T_VEC)
#define T_MATRIX	   ((N_MATRIX << 16) | T_FORM|   T_GAL|T_VEC)
#define T_TENSOR	   ((N_TENSOR << 16) | T_FORM|   T_GAL|T_VEC)
#define T_PREFIX 	   ((N_PREF   << 16) | T_FORM|   T_GAL|T_VEC)
#define T_RELEXP	   ((N_RELEXP << 16) | T_EXPR|   T_GAL|T_VEC)
#define T_SETEXP	   ((N_SETEXP << 16) | T_EXPR|   T_GAL|T_VEC)
#define T_LOGEXP	   ((N_LOGEXP << 16) | T_EXPR|   T_GAL|T_VEC)
#define T_TROPICAL	   ((N_TROPICAL << 16) | T_EXPR|   T_GAL|T_VEC)

typedef struct Number_Head {
    Tag             h;
    long            len;
    Sex             s[1];
} Number_Head;

#define length_of_header(v) (((Number_Head *)((long )(v) &~TAG_BITS))->len)
#define numhdr(v) (((Number_Head *)((long )(v) - TAG_NUMBERS ))->h)
#define numlen(v) (((Number_Head *)((long )(v) - TAG_NUMBERS ))->len)
#define numvec(v,n) (((Number_Head *)((long )(v) - TAG_NUMBERS ))->s[n])

#define numtype_of(h)        (numhdr(h)>>16)
#define numhdr_of(h)        ((h)>>16)
#define gnumtype_of(h)       (vechdr(h)>>16)
#define gnumhdr_of(h)       ((h)>>16)

#define T_NUMBER	0x020
#define T_VECTOR	0x030
#define T_BINARY	0x000
#define T_SYMBOL        0x010
#define T_MASK		0x030

#define is_numhdr_header(h) (((long)(h) & T_MASK) == T_NUMBER)
#define is_vector_header(h)  (((long)(h) & T_MASK) == T_VECTOR)
#define is_binhdr_header(h)  (((long)(h) & T_MASK) == T_BINARY)
#define is_symbol_header(h)  (((long)(h) & T_MASK) == T_SYMBOL)

/*
 * The following tests are valid provided that n is already known to have tag
 * TAG_NUMBERS, i.e. it is a bignum, ratio or complex.
 */
#define is_logical(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_LOGICAL)

#define is_modular(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_MODULAR)

#define is_bigint(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BIGNUM)

#define is_float(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_FLOAT)

#define is_rational(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_RATIONAL)

#define is_ratio(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_RATIONAL)

#define is_complex(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_COMPLEX)

#define is_bfloat(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BFLOAT)

#define is_efloat(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_EFLOAT)
#define is_interval(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_INTERVAL)
#define is_ecomplex(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_ECOMPLEX)
#define is_befloat(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BEFLOAT)
#define is_binterval(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BINTERVAL)
#define is_becomplex(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BECOMPLEX)
#define is_bmodular(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_BMODULAR)
#define is_qmodular(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_QMODULAR)
#define is_fmodular(n) \
    (is_numhdr(n) && (numtype_of(n)) == N_FMODULAR)

#define is_algnum(n) \
    (is_vechdr(n) && (gnumtype_of(n)) == N_ALGNUM)
#define is_poly(n) \
    (is_vechdr(n) && (gnumtype_of(n)) == N_POLY)
#define is_dipoly(n) \
    (is_vechdr(n) && (gnumtype_of(n)) == N_DIPOLY)
#define is_rpoly(n) \
    (is_vechdr(n) && (gnumtype_of(n)) == N_RPOLY)



#define is_bigneg(n) ((numhdr(n) & T_NEG) != 0)
#define is_integer(n) (is_fixnum(n) || is_bigint(n))
#define is_fix(n)    (is_numhdr(n) && (numhdr(n)&T_FIX) != 0)
#define is_neg(n)    (is_numhdr(n) && (numhdr(n)&T_NEG) != 0)
#define is_app(n)    (is_numhdr(n) && (numhdr(n)&T_APP) != 0)
#define is_log(n)    (is_numhdr(n) && (numhdr(n)&T_LOG) != 0)
#define is_eff(n)    (is_numhdr(n) && (numhdr(n)&T_EFF) != 0)
#define is_cmp(n)    (is_numhdr(n) && (numhdr(n)&T_CMP) != 0)
#define is_big(n)    (is_numhdr(n) && (numhdr(n)&T_BIG) != 0)

/*
 * tag for binary data 
 */

#define T_STRING         0x040	/* simple character vector */
#define T_IARRAY         0x150	/* simple integer vector */
#define T_LARRAY         0x160	/* simple long integer vector */
#define T_SARRAY         0x170	/* simple short float vector */
#define T_DARRAY         0x180	/* simple double float vector */
#define T_IARRAY2        0x250	/* simple integer array */
#define T_LARRAY2        0x260	/* simple long integer array */
#define T_SARRAY2        0x270	/* simple short float array */
#define T_DARRAY2        0x280	/* simple double float array */
#define T_IARRAY3        0x350	/* simple integer 3d array */
#define T_LARRAY3        0x360	/* simple long integer 3d array */
#define T_SARRAY3        0x370	/* simple short float 3d array */
#define T_DARRAY3        0x380	/* simple double float 3d array */
#define T_FILE           0x440	/* a file-handle */

#define T_ARRAY_MASK	0x0f00
#define T_ARRAY_1	0x0100
#define T_ARRAY_2	0x0200
#define T_ARRAY_3	0x0300

/*
 * tag for lisp vector 
 */

#define T_SIMPLE         0x1f0	/* simple vector */
#define T_ARRAY          0x2f0	/* header record for general array */
#define T_V_ARRAY        0x3f0	/* variable array (aka. list) */
#define T_PACKAGE        0x4f0	/* packages.  */
#define T_HASH           0x5f0	/* hash tables.  */

#define T_GAL		 0x6f0	/* Gal structures */

#define CHAR_EOF    vector_value(current_charvec, 0x1a)
#define CHAR_OF(x)  vector_value(current_charvec, x)
#define LISP_CHAR_OF(x)  vector_value(sugar_charvec, x)
#define ASUKA_CHAR_OF(x)  vector_value(asuka_charvec, x)


typedef struct Symbol_Head {
    Tag             header;
    long            len;
    Sex             value;	/* 
				 * Global or special value cell 
				 */
#ifdef ALGEBRAIC_PROPERTY
    Sex             gvalue;	/* 
				 * Global or special value cell 
				 */
#endif
    union {
	Sex             lambda;	/* lambda definition */
	zero_args      *proc;	/* 
				 * Executable code  
				 */
	long            code;	/* bytecode location */
    } proc;

#ifdef ALGEBRAIC_PROPERTY
    zero_args      *gproc;	/* 
				 * Executable code  
				 */
    unsigned short  gtag;	/* 
				 * argument count 
				 */
    char            nargs;	/* 
				 * argument count 
				 */
    char            gnargs;	/* 
				 * argument count 
				 */
#else
    long            nargs;	/* 
				 * argument count 
				 */
#endif
    Sex             plist;	/* 
				 * A list 
				 */
    char            pname[CELL];	/* 
					 * A string (always) 
					 */
} Symbol_Head;

#define qtag(p)       (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->header)
#define gtag(p)       (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->gtag)
#define symlen(p)     (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->len)
#define value(p)      (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->value)
#define lambda(p)     (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->proc.lambda)
#define proc(p)       (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->proc.proc)
#define bcode(p)      (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->proc.code)
#define gvalue(p)     (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->gvalue)
#define gproc(p)      (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->gproc)
#define qnargs(p)     (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->nargs)
#define gnargs(p)     (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->gnargs)
#define pname(p)      (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->pname)
#define plist(p)      (((Symbol_Head *)((char *)(p) - TAG_SYMBOL))->plist)

#define symhdr_length       ((sizeof(Symbol_Head) + (CELL-1)) & -(CELL))

#define proc0(p)        ((zero_args *)  proc(p))
#define proc1(p)        ((one_args *)   proc(p))
#define proc2(p)        ((two_args *)   proc(p))
#define proc3(p)        ((three_args *) proc(p))
#define proc4(p)        ((n_args4 *)    proc(p))
#define proc5(p)        ((n_args5 *)    proc(p))
#define proc6(p)        ((n_args6 *)    proc(p))
#define proc7(p)        ((n_args7 *)    proc(p))
#define proc8(p)        ((n_args8 *)    proc(p))
#define proc9(p)        ((n_args9 *)    proc(p))
#define proc10(p)       ((n_args10 *)   proc(p))
#define proc11(p)       ((n_args11 *)   proc(p))
#define proc12(p)       ((n_args12 *)   proc(p))
#define proc13(p)       ((n_args13 *)   proc(p))
#define proc14(p)       ((n_args14 *)   proc(p))
#define proc15(p)       ((n_args15 *)   proc(p))
#define procn(p)        ((n_args *)     proc(p))
#ifdef ASUKA
#define gproc0(p)       ((zero_args *)  gproc(p))
#define gproc1(p)       ((one_args *)   gproc(p))
#define gproc2(p)       ((two_args *)   gproc(p))
#define gproc3(p)       ((three_args *) gproc(p))
#define gproc4(p)       ((n_args4 *)    gproc(p))
#define gproc5(p)       ((n_args5 *)    gproc(p))
#define gproc6(p)       ((n_args6 *)    gproc(p))
#define gproc7(p)       ((n_args7 *)    gproc(p))
#define gproc8(p)       ((n_args8 *)    gproc(p))
#define gproc9(p)       ((n_args9 *)    gproc(p))
#define gproc10(p)      ((n_args10 *)   gproc(p))
#define gproc11(p)      ((n_args11 *)   gproc(p))
#define gproc12(p)      ((n_args12 *)   gproc(p))
#define gproc13(p)      ((n_args13 *)   gproc(p))
#define gproc14(p)      ((n_args14 *)   gproc(p))
#define gproc15(p)      ((n_args15 *)   gproc(p))
#define gprocn(p)       ((n_args *)     gproc(p))
#endif


typedef struct Logical_Number {
    Tag             h;
    long            len;
    unsigned long   d[1];
} Logical_Number;

#define logical_length(b)  (((Logical_Number *)((char *)(b)-TAG_NUMBERS))->len)
#define logical_value(b)   (((Logical_Number *)((char *)(b)-TAG_NUMBERS))->d)

typedef struct Big_Number {
    Tag             h;
    long            len;
    unsigned long   d[1];
} Big_Number;

#define bignum_length(b)  (((Big_Number *)((char *)(b)-TAG_NUMBERS))->len)
#define bignum_value(b)   (((Big_Number *)((char *)(b)-TAG_NUMBERS))->d)

#define bnum_len(n)	  (HDR + (n)*SEX)
#define bnum_n(len)	  ((len)/SEX - 2)

#define bignum_precision(a)	((bignum_length(a) - HDR) * 8)

typedef struct Modular_Number {
    Tag             h;
    long            len;
    long            d;
    long            m;
} Modular_Number;

#define modular_length(b)  (((Modular_Number *)((char *)(b)-TAG_NUMBERS))->len)
#define modular_value(b)   (((Modular_Number *)((char *)(b)-TAG_NUMBERS))->d)
#define modular_modulus(b) (((Modular_Number *)((char *)(b)-TAG_NUMBERS))->m)

typedef struct Rational_Number {
    Tag             header;
    long            len;
    Sex             num;
    Sex             den;
} Rational_Number;

#define rational_num(r)  (((Rational_Number *)((char *)(r)-TAG_NUMBERS))->num)
#define rational_den(r)  (((Rational_Number *)((char *)(r)-TAG_NUMBERS))->den)

typedef struct Complex_Number {
    Tag             header;
    long            len;
    Sex             real;
    Sex             imag;
} Complex_Number;

#define complex_real(r)    (((Complex_Number *)((char *)(r)-TAG_NUMBERS))->real)
#define complex_imag(r)    (((Complex_Number *)((char *)(r)-TAG_NUMBERS))->imag)

typedef struct EFloat_Number {
    Tag             header;
    long            len;
    double          value;
    double          error;
} EFloat_Number;

#define efloat_value(r)    (((EFloat_Number *)((char *)(r)-TAG_NUMBERS))->value)
#define efloat_error(r)    (((EFloat_Number *)((char *)(r)-TAG_NUMBERS))->error)

#ifdef BEFLOAT_ERROR_IS_DOUBLE
typedef double  bf_error_type;
#else
typedef long    bf_error_type;
#endif

typedef struct Big_EFloat_Number {
    Tag             header;
    long            len;
    Sex             value;
    bf_error_type   error;
} Big_EFloat_Number;

#define befloat_value(r)    \
	(((Big_EFloat_Number *)((char *)(r)-TAG_NUMBERS))->value)
#define befloat_error(r)    \
	(((Big_EFloat_Number *)((char *)(r)-TAG_NUMBERS))->error)

typedef struct EComplex_Number {
    Tag             header;
    long            len;
    double          real;
    double          imag;
    double          error;
} EComplex_Number;

#define ecomplex_real(r)  \
        (((EComplex_Number *)((char *)(r)-TAG_NUMBERS))->real)
#define ecomplex_imag(r)  \
        (((EComplex_Number *)((char *)(r)-TAG_NUMBERS))->imag)
#define ecomplex_error(r)  \
        (((EComplex_Number *)((char *)(r)-TAG_NUMBERS))->error)

typedef struct Big_EComplex_Number {
    Tag             header;
    long            len;
    Sex             real;
    Sex             imag;
    bf_error_type   error;
} Big_EComplex_Number;

#define becomplex_real(r)  \
        (((Big_EComplex_Number *)((char *)(r)-TAG_NUMBERS))->real)
#define becomplex_imag(r)  \
        (((Big_EComplex_Number *)((char *)(r)-TAG_NUMBERS))->imag)
#define becomplex_error(r)  \
        (((Big_EComplex_Number *)((char *)(r)-TAG_NUMBERS))->error)


typedef struct Interval_Number {
    Tag             header;
    long            len;
    double          lower;
    double          upper;
} Interval_Number;

#define interval_lower(r) \
  (((Interval_Number *)((char *)(r)-TAG_NUMBERS))->lower)
#define interval_upper(r) \
  (((Interval_Number *)((char *)(r)-TAG_NUMBERS))->upper)

typedef struct Big_Interval_Number {
    Tag             header;
    long            len;
    Sex             lower;
    Sex             upper;
} Big_Interval_Number;

#define binterval_lower(r)  \
	(((Big_Interval_Number *)((char *)(r)-TAG_NUMBERS))->lower)
#define binterval_upper(r)  \
	(((Big_Interval_Number *)((char *)(r)-TAG_NUMBERS))->upper)


typedef struct Float_Number {
    Tag             header;
    long            len;
    double          f;
} Float_Number;

#define float_val(v)     (((Float_Number *)((char *)(v) -TAG_NUMBERS))->f)


typedef struct Big_Float_Number {
    Tag             header;
    long            len;
    long            exp;
    Sex             man;
} Big_Float_Number;

#define bfloat_exp(v)   (((Big_Float_Number *)((char *)(v) -TAG_NUMBERS))->exp)
#define bfloat_man(v)   (((Big_Float_Number *)((char *)(v) -TAG_NUMBERS))->man)

typedef struct Big_Modular_Number {
    Tag             header;
    long            len;
    Sex             value;
    Sex             modulus;
} Big_Modular_Number;

#define bmodular_value(v) (((Big_Modular_Number *)((char *)(v) -TAG_NUMBERS))->value)
#define bmodular_modulus(v) (((Big_Modular_Number *)((char *)(v) -TAG_NUMBERS))->modulus)

typedef struct q_Modular_Number {
    Tag             header;
    long            len;
    long            modulus;
    long            value[1];
} q_Modular_Number;

#define qmodular_value(v) (((q_Modular_Number *)((char *)(v) -TAG_NUMBERS))->value)
#define qmodular_modulus(v) (((q_Modular_Number *)((char *)(v) -TAG_NUMBERS))->modulus)

typedef struct f_Modular_Number {
    Tag             header;
    long            len;
    long            ivalue;
    double          dvalue;
} f_Modular_Number;

#define fmodular_ivalue(v) (((f_Modular_Number *)((char *)(v) -TAG_NUMBERS))->ivalue)
#define fmodular_dvalue(v) (((f_Modular_Number *)((char *)(v) -TAG_NUMBERS))->dvalue)

typedef struct Numtype_Head {
    union {
	struct Number_Head n;
	struct Logical_Number l;
	struct Big_Number b;
	struct Modular_Number m;
	struct Rational_Number r;
	struct Complex_Number c;
	struct EFloat_Number e;
	struct Big_EFloat_Number E;
	struct EComplex_Number z;
	struct Big_EComplex_Number Z;
	struct Interval_Number v;
	struct Big_Interval_Number V;
	struct Float_Number f;
	struct Big_Float_Number d;
	struct Big_Modular_Number M;
	struct q_Modular_Number q;
	struct f_Modular_Number Y;
    } u;
} Numtype_Head;


/*
 * Gal structures 
 */

typedef struct Variable {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
    Sex             def;
} Variable;

#define var_var(v)   (((Variable *)((char *)(v) -TAG_VECTOR))->var)
#define var_pol(v)   (((Variable *)((char *)(v) -TAG_VECTOR))->pol)
#define var_def(v)   (((Variable *)((char *)(v) -TAG_VECTOR))->def)

typedef struct Algebraic_Number {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
    Sex             def;
} Algebraic_Number;

#define algnum_var(v)   (((Algebraic_Number *)((char *)(v) -TAG_VECTOR))->var)
#define algnum_pol(v)   (((Algebraic_Number *)((char *)(v) -TAG_VECTOR))->pol)
#define algnum_def(v)   (((Algebraic_Number *)((char *)(v) -TAG_VECTOR))->def)

typedef struct Transendential {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
    Sex             def;
} Transendential;

#define const_var(v)   (((Transendential *)((char *)(v) -TAG_VECTOR))->var)
#define const_pol(v)   (((Transendential *)((char *)(v) -TAG_VECTOR))->pol)
#define const_def(v)   (((Transendential *)((char *)(v) -TAG_VECTOR))->def)

typedef struct Polynomial {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
} Polynomial;

typedef struct Distributed_Polynomial {
    Tag             header;
    long            len;
    // Sex var;
    Sex             order;
    Sex             pol;
} Distributed_Polynomial;

#define poly_var(v)   (((Polynomial *)((char *)(v) -TAG_VECTOR))->var)
#define poly_pol(v)   (((Polynomial *)((char *)(v) -TAG_VECTOR))->pol)

// #define dipoly_var(v) \
// (((Distributed_Polynomial *)((char *)(v) -TAG_VECTOR))->var)
#define dipoly_order(v)   \
	(((Distributed_Polynomial *)((char *)(v) -TAG_VECTOR))->order)
#define dipoly_pol(v)   \
	(((Distributed_Polynomial *)((char *)(v) -TAG_VECTOR))->pol)

typedef struct Algebraic_Function {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
} Algebraic_Function;

typedef struct Rational_Polynomial {
    Tag             header;
    long            len;
    Sex             num;
    Sex             den;
} Rational_Polynomial;

#define rpoly_num(v)   (((Rational_Polynomial  *)((char *)(v) -TAG_VECTOR))->num)
#define rpoly_den(v)   (((Rational_Polynomial  *)((char *)(v) -TAG_VECTOR))->den)


typedef struct Vector {
    Tag             header;
    long            len;
    Sex             dimx;
    Sex             __v[1];
} Vector;

#define is_vect(s)         (is_vechdr(s) && type_of_header(vechdr(s)) == T_VECT)
#define vect_dimx(u)       (((Vector *)((char *)(u) - TAG_VECTOR))->dimx)
#define vect_value(u, k)   (((Vector *)((char *)(u) - TAG_VECTOR))->__v[k])

typedef struct Matrix {
    Tag             header;
    long            len;
    Sex             dimx;
    Sex             dimy;
    Sex             __v[1];
} Matrix;

#define is_matrix(s)         (is_vechdr(s) && type_of_header(vechdr(s)) == T_MATRIX)
#define matrix_dimx(u)       (((Matrix *)((char *)(u) - TAG_VECTOR))->dimx)
#define matrix_dimy(u)       (((Matrix *)((char *)(u) - TAG_VECTOR))->dimy)
#define matrix_value(u, k)   (((Matrix *)((char *)(u) - TAG_VECTOR))->__v[k])

typedef struct Tensor {
    Tag             header;
    long            len;
    Sex             __v[1];
} Tensor;

typedef struct Prefix_Form {
    Tag             header;
    long            len;
    Sex             __v[1];
} Prefix_Form;

#define is_prefix(s)    (is_vechdr(s) && type_of_header(vechdr(s)) == T_PREFIX)
#define prefix_form(u)	(((Prefix_Form *)((char *)(u) - TAG_VECTOR))->__v[0])

typedef struct Relation_Expression {
    Tag             header;
    long            len;
    Sex             __v[1];
} Relation_Expression;

typedef struct Set_Expression {
    Tag             header;
    long            len;
    Sex             __v[1];
} Set_Expression;

typedef struct Logical_Expression {
    Tag             header;
    long            len;
    Sex             __v[1];
} Logical_Expression;

typedef struct Tropical {
    Tag             header;
    long            len;
    Sex             var;
    Sex             pol;
} Tropical;

#define is_tropical(s)         (is_vechdr(s) && type_of_header(vechdr(s)) == T_TROPICAL)
#define tropical_var(v)   (((Tropical *)((char *)(v) -TAG_VECTOR))->var)
#define tropical_pol(v)   (((Tropical *)((char *)(v) -TAG_VECTOR))->pol)

typedef struct Galtype_Head {
    union {
	struct Variable v;
	struct Algebraic_Number a;
	struct Transendential t;
	struct Polynomial p;
	struct Distributed_Polynomial D;
	struct Algebraic_Function A;
	struct Rational_Polynomial r;
	struct Vector   V;
	struct Matrix   M;
	struct Tensor   T;
	struct Prefix_Form P;
	struct Relation_Expression R;
	struct Set_Expression S;
	struct Logical_Expression L;
	struct Tropical X;
    } u;
} Galtype_Head;

#define alignit(n) ((Sex)(((long)(n) + CELL - 1) & (-CELL)))
#define alignits(n) ((Sex)(((long)(n) + SEX - 1) & (-SEX)))

typedef struct Vector_Head {
    Tag             header;
    long            len;
    Sex             __v[1];
} Vector_Head;

typedef struct Array_Head {
    Tag             header;
    long            len;
    Sex             dim;
    Sex             __v[1];
} Array_Head;

typedef struct Binary_Head {
    Tag             header;
    long            len;
    union {
	char            s[4];
	int             i[1];
	long            l[1];
	float           f[1];
	double          d[1];
	Sex             __w[1];
    } __v;
} Binary_Head;

#define vechdr(v)  (((Vector_Head *)((char *)(v) - TAG_VECTOR))->header)
#define veclen(v)  (((Vector_Head *)((char *)(v) - TAG_VECTOR))->len)
#define binhdr(v)  (((Binary_Head *)((char *)(v) - TAG_BINARY))->header)
#define binlen(v)  (((Binary_Head *)((char *)(v) - TAG_BINARY))->len)
#define symbol_length(a)	(symlen(a) - sizeof(Symbol_Head) + CELL)
#define string_length(a)	(binlen(a) - CELL)

#define is_simple_vector(s) \
    (is_vechdr(s) && type_of_header(vechdr(s)) == T_SIMPLE)
#define is_vector(s)	is_simple_vector(s)

#define is_array(s) \
    (is_vechdr(s) && type_of_header(vechdr(s)) == T_ARRAY)

#define vector_value(u, n) \
                     (((Vector_Head *)((char *)(u) - TAG_VECTOR))->__v[n])
#define array_value(u, n) \
                     (((Array_Head *)((char *)(u) - TAG_VECTOR))->__v[n])
#define array_dim(u) \
                     (((Array_Head *)((char *)(u) - TAG_VECTOR))->dim)
#define string_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.s[n])
#define iarray_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.i[n])
#define larray_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.l[n])
#define sarray_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.f[n])
#define darray_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.d[n])
#define object_value(u, n) \
                   (((Binary_Head *)((char *)(u) - TAG_BINARY))->__v.__w[n])

#define is_string(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_STRING)
#define is_iarray(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_IARRAY)
#define is_larray(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_LARRAY)
#define is_sarray(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_SARRAY)
#define is_darray(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_DARRAY)

typedef struct Binary2_Head {
    Tag             header;
    long            len;
    long            dim_y;
    union {
	char            s[4];
	int             i[1];
	long            l[1];
	float           f[1];
	double          d[1];
	Sex             __w[1];
    } __v;
} Binary2_Head;

#define barray_dimy(v)  (((Binary2_Head *)((char *)(v) - TAG_BINARY))->dim_y)

#define iarray2_value(u, n) \
                   (((Binary2_Head *)((char *)(u) - TAG_BINARY))->__v.i[n])
#define larray2_value(u, n) \
                   (((Binary2_Head *)((char *)(u) - TAG_BINARY))->__v.l[n])
#define sarray2_value(u, n) \
                   (((Binary2_Head *)((char *)(u) - TAG_BINARY))->__v.f[n])
#define darray2_value(u, n) \
                   (((Binary2_Head *)((char *)(u) - TAG_BINARY))->__v.d[n])
#define is_iarray2(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_IARRAY2)
#define is_larray2(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_LARRAY2)
#define is_sarray2(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_SARRAY2)
#define is_darray2(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_DARRAY2)

typedef struct Binary3_Head {
    Tag             header;
    long            len;
    long            dim_y;
    long            dim_z;
    union {
	char            s[4];
	int             i[1];
	long            l[1];
	float           f[1];
	double          d[1];
	Sex             __w[1];
    } __v;
} Binary3_Head;

#define barray_dimz(v)  (((Binary3_Head *)((char *)(v) - TAG_BINARY))->dim_z)

#define iarray3_value(u, n) \
                   (((Binary3_Head *)((char *)(u) - TAG_BINARY))->__v.i[n])
#define larray3_value(u, n) \
                   (((Binary3_Head *)((char *)(u) - TAG_BINARY))->__v.l[n])
#define sarray3_value(u, n) \
                   (((Binary3_Head *)((char *)(u) - TAG_BINARY))->__v.f[n])
#define darray3_value(u, n) \
                   (((Binary3_Head *)((char *)(u) - TAG_BINARY))->__v.d[n])
#define is_iarray3(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_IARRAY3)
#define is_larray3(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_LARRAY3)
#define is_sarray3(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_SARRAY3)
#define is_darray3(s) \
    (is_binhdr(s) && type_of_header(binhdr(s)) == T_DARRAY3)

#define barray_dim(u, h) \
	((h&0x300) >> 8)
#define barray_type(u, h) \
	(h&0xf0)
/*
 * Variable Array
 */

typedef struct V_Array {
    Tag             Header;
    long            len;
    long            size;
    Sex             __v[1];
} V_Array;

#define is_v_array(s) \
    (is_vechdr(s) && type_of_header(vechdr(s)) == T_V_ARRAY)

#define v_array_len(u) \
                     (((V_Array *)((char *)(u) - TAG_VECTOR))->len)
#define v_array_value(u, n) \
                     (((V_Array *)((char *)(u) - TAG_VECTOR))->__v[n])
#define v_array_size(u) \
                     (((V_Array *)((char *)(u) - TAG_VECTOR))->size)

/*
 * Package structure            
 */

typedef struct Package {
    Tag             header;
    long            len;
    Sex             id;		/* 
				 * Package as type of this struct 
				 */
    Sex             vector;
    Sex             chrvec;
    Sex             count;
} Package;

#define package_hdr(p)   (((Package *)((char *)(p) - TAG_VECTOR))->header)
#define package_id(p)    (((Package *)((char *)(p) - TAG_VECTOR))->id)
#define package_vec(p)   (((Package *)((char *)(p) - TAG_VECTOR))->vector)
#define package_charvec(p)   (((Package *)((char *)(p) - TAG_VECTOR))->chrvec)
#define package_cnt(p)   (((Package *)((char *)(p) - TAG_VECTOR))->count)
#define is_package(s) \
    (is_vechdr(s) && type_of_header(vechdr(s)) == T_PACKAGE)

/*
 * Hash table
 */

typedef struct HashEntry {
    Sex             key;
    Sex             value;
} HashEntry;

typedef struct HashTable {
    Tag             header;
    long            len;
    Sex             size;
    Sex             count;
    Sex             entry;
} HashTable;

#define hash_hdr(p)    (((HashTable *)((char *)(p) - TAG_VECTOR))->header)
#define hash_size(p)   (((HashTable *)((char *)(p) - TAG_VECTOR))->size)
#define hash_count(p)  (((HashTable *)((char *)(p) - TAG_VECTOR))->count)
#define hash_entry(p)  (((HashTable *)((char *)(p) - TAG_VECTOR))->entry)
#define is_hashtable(s) \
    (is_vechdr(s) && type_of_header(vechdr(s)) == T_HASH)

/*
 * values to go in exit_reason 
 */

#define	EXIT_NULL	0
#define	EXIT_GO		1
#define	EXIT_RETURN	2
#define	EXIT_THROW	3

/*
 * argument to exit system call
 */

#define EXIT_SUCCESS	0
#define	EXIT_FAILURE	1


#define RUN_ASUKA       0x06
#define RUN_GAL         0x05
#define RUN_REDUCE      0x04
#define RUN_SUGAR       0x02
#define RUN_SLISP       0x01
#define RUN_GSUGAR      0x03
#define RUN_COMPILER    0x08
#define RUN_GBP         0x10
#define RUN_LBP         0x20
#define RUN_BP          (RUN_GBP | RUN_LBP)
#define RUN_MASK        0x07
