/*
 * trig.c      (C) Copyright 1998-2001, 2007 by Fujio Kako, Nara Women's University.
 * Transcendental Functions.
 *      $Header: /Users/kako/asuka/src/RCS/trig.c,v 2.12 2015/11/16 07:34:16 kako Exp $
 */

static char const rcsid[] =
    "$Id: trig.c,v 2.12 2015/11/16 07:34:16 kako Exp $";

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "config.h"
#include "tags.h"
#include "errors.h"
#include "externs.h"
#include "arith.h"
#include "mathop.h"
#include "compare.h"
#include "entries.h"
#include "trig.h"

#include "bfloat.h"

Sex             Le_x(Sex);

static double   fhypot(double u, double v);
static double   fcbrt(double a);
static double   my_log2(double x);
static double   flog10(double x);
static double   fcabs(Complex z);
static Complex  my_csin(Complex z);
static Complex  my_ccos(Complex z);
static Complex  my_ctan(Complex z);
static Complex  my_ccosh(Complex z);
static Complex  my_csinh(Complex z);
static Complex  my_ctanh(Complex z);
static Complex  my_cexp(Complex z);
static Complex  my_clog(Complex z);
static Complex  my_clog2(Complex z);
static Complex  my_clog10(Complex z);
static Complex  my_csqrt(Complex z);
static Complex  my_casinh(Complex z);
static Complex  my_cacosh(Complex z);
static Complex  my_catanh(Complex z);
static Complex  ccbrt(Complex a);
static Complex  my_cacos(Complex z);
static Complex  my_casin(Complex z);
static Complex  my_catan(Complex z);
static Sex      make_complex(double x, double y);
static Sex      dsin(Sex x);
static Sex      dcos(Sex x);
static Sex      datan2(Sex a, Sex b);
static Sex      dtan(Sex x);
static Sex      dasin(Sex x);
static Sex      dacos(Sex x);
static Sex      datan(Sex x);
static Sex      dsinh(Sex a);
static Sex      dcosh(Sex a);
static Sex      dtanh(Sex a);
static Sex      dasinh(Sex a);
static Sex      dacosh(Sex a);
static Sex      datanh(Sex a);
static Sex      dsqrt(Sex a);
static Sex      dcbrt(Sex a);
static Sex      dlog10(Sex a);
static Sex      dlog2(Sex a);
static Sex      dlog(Sex a);
static Sex      dexp(Sex a);
static Sex      dcabs(Sex zr, Sex zi);
static Sex      dcsin(Sex zr, Sex zi);
static Sex      dccos(Sex zr, Sex zi);
static Sex      dcexp(Sex zr, Sex zi);
static Sex      dcsinh(Sex zr, Sex zi);
static Sex      dccosh(Sex zr, Sex zi);
static Sex      dctanh(Sex zr, Sex zi);
static Sex      dldexp(Sex a, long n);
static Sex      dabs(Sex a);
static Sex      dclog(Sex x, Sex y);
static Sex      dclog2(Sex zr, Sex zi);
static Sex      dclog10(Sex zr, Sex zi);
static Sex      dcsqrt(Sex x, Sex y);
static Sex      dccbrt(Sex zr, Sex zi);
static Sex      dctan(Sex zr, Sex zi);
static Sex      dcasin(Sex zr, Sex zi);
static Sex      dcacos(Sex zr, Sex zi);
static Sex      dcatan(Sex zr, Sex zi);
static Sex      dcasinh(Sex zr, Sex zi);
static Sex      dcacosh(Sex zr, Sex zi);
static Sex      dcatanh(Sex zr, Sex zi);

extern long     power_mod(long n, long m, long k);
extern Sex      modular_expt(Sex a, Sex b, Sex m);

#if defined(VisualC) || defined(Linux)
#define cbrt fcbrt
extern double   log(double);
extern double   sqrt(double);

double
atanh(double x)
{
    return 0.5 * log((1 + x) / (1 - x));
}

double
asinh(double x)
{
    return log(x + sqrt(1 + x * x));
}

double
acosh(double x)
{
    return 2 * log(sqrt((x + 1) / 2) + sqrt((x - 1) / 2));
}

#endif

#define _pi       3.14159265358979323846
#define _half_pi  1.57079632679489661923
#define _log_2    0.693147180559945309417
#define _log_10   2.30258509299404568402

static double
fhypot(double u, double v)
{
    double          r;
    if (u < 0.0)
	u = -u;
    if (v < 0.0)
	v = -v;
    if (u > v) {
	r = u;
	u = v;
	v = r;
    }
    if (u + v == v)
	r = v;
    else {
	r = u / v;
	r = v * sqrt(1.0 + r * r);
    }
    return r;
}

static double
fcbrt(double a)
{
    long

                    i,
                    neg = 0;
    double          b;
    int             xx,
                    x;

    if (a == 0.0)
	return 0.0;
    else if (a < 0.0)
	a = -a, neg = 1;
    b = frexp(a, &xx);		/* 
				 * end-conditions unimportant 
				 */
    x = xx;
    b = 0.5996 + 0.4081 * b;
    while (x % 3 != 0) {
	x--;
	b *= 1.26;
    }
    b = ldexp(b, x / 3);
    for (i = 0; i < 6; i++) {
	b = (2.0 * b + a / (b * b)) / 3.0;
    }
    if (neg)
	return -b;
    else
	return b;
}

static double
my_log2(double x)
{
    return log(x) / _log_2;
}

static double
flog10(double x)
{
    return log(x) / _log_10;
}

static double
fcabs(Complex z)
{
    return sqrt(z.real * z.real + z.imag * z.imag);
}


static          Complex
my_csin(Complex z)
{
    /*
     * sin(x + iy) = sin(x)*cosh(y) + i cos(x)*sinh(y) 
     */
    double          x = z.real,
	y = z.imag;
    double          s = sin(x),
	c = cos(x);
    double          sh,
                    ch;
    if (y <= -22.0) {
	ch = exp(-y) * 0.5;
	sh = -ch;
    } else if (y >= 22.0) {
	ch = sh = exp(y) * 0.5;
    } else {
	ch = cosh(y);
	sh = sinh(y);
    }
    z.real = s * ch;
    z.imag = c * sh;
    return z;
}

static          Complex
my_ccos(Complex z)
{
    /*
     * cos(x + iy) = cos(x)*cosh(y) - i sin(x)*sinh(y) 
     */
    double          x = z.real,
	y = z.imag;
    double          s = sin(x),
	c = cos(x);
    double          sh,
                    ch;
    if (y < -22.0) {
	ch = exp(-y) / 2;
	sh = -ch;
    } else if (y > 22.0) {
	sh = ch = exp(y) / 2;
    } else {
	ch = cosh(y);
	sh = sinh(y);
    }
    z.real = c * ch;
    z.imag = -s * sh;
    return z;
}

static          Complex
my_ctan(Complex z)
{
    double          x = z.real,
	y = z.imag;
    /*
     * tan(x + iy) = (tan(x) + i tanh(y))/(1 - i tan(x)*tanh(y))
     */
    double          t = tan(x),
	th = tanh(y);
    double          t2 = t * t,
	th2 = th * th;
    double          d = 1.0 + t2 * th2;

    z.real = t * (1.0 - th2) / d;
    z.imag = th * (1.0 + t2) / d;
    return z;
}

static          Complex
my_ccosh(Complex z)
{
    /*
     * cosh(z) = cos(iz) 
     */
    double          t;
    t = z.real;
    z.real = -z.imag;
    z.imag = t;
    z = my_ccos(z);
    return z;
}

static          Complex
my_csinh(Complex z)
{
    /*
     * sinh(z) = -i sin(iz) 
     */
    double          t;
    t = z.real;
    z.real = -z.imag;
    z.imag = t;
    z = my_csin(z);
    t = z.real;
    z.real = z.imag;
    z.imag = -t;
    return z;
}

static          Complex
my_ctanh(Complex z)
{
    /*
     * tan(z) = i tanh(-i z) tanh(z) = -i tan(iz) 
     */
    /*
     * tanh(x+yi) = (sinh(2x)+isin(2y))/(cosh2x + cos2y) tanh(z) = -i
     * tan(iz) 
     */
    double          t;
    t = z.real;
    z.real = -z.imag;
    z.imag = t;
    z = my_ctan(z);
    t = z.real;
    z.real = z.imag;
    z.imag = -t;
    return z;
}


static          Complex
my_cexp(Complex z)
{
    double          x = z.real,
	y = z.imag;
    double          s = sin(y),
	c = cos(y);
    double          w = exp(x);
    z.real = w * c;
    z.imag = w * s;
    return z;
}

static          Complex
my_clog(Complex z)
{
    double          x = z.real,
	y = z.imag;
    double          scale,
                    r;
    int             n1,
                    n2;
    if (x == 0.0)
	r = log(fabs(y));
    else if (y == 0.0)
	r = log(fabs(x));
    else {
	(void) frexp(x, &n1);
	(void) frexp(y, &n2);
	if (n2 > n1)
	    n1 = n2;
	scale = ldexp(1.0, n1);
	x /= scale;
	y /= scale;
	r = log(scale) + 0.5 * log(x * x + y * y);
    }
    z.real = r;
    if (y == 0.0)
	if (x < 0.0)
	    z.imag = _pi;
	else
	    z.imag = 0.0;
    else
	z.imag = atan2(y, x);
    return z;
}


static          Complex
my_clog2(Complex z)
{
    z = my_clog(z);
    z.real = z.real / _log_2;
    z.imag = z.imag / _log_2;
    return z;
}

static          Complex
my_clog10(Complex z)
{
    z = my_clog(z);
    z.real = z.real / _log_10;
    z.imag = z.imag / _log_10;
    return z;
}




static          Complex
my_csqrt(Complex z)
{
    double          x = z.real,
	y = z.imag;
    /*
     * */
    long            j;
    int             n;
    double          scale;
    double          vx,
                    vy;
    if (y == 0.0) {
	if (x < 0.0) {
	    z.real = 0.0;
	    z.imag = sqrt(-x);
	} else {
	    z.real = sqrt(x);
	    z.imag = 0.0;
	}
	return z;		/* 
				 * Pure real arguments are easy 
				 */
    }
    (void) frexp(y, &n);
    if (x != 0.0) {
	int             n1;
	(void) frexp(x, &n1);
	if (n1 > n)
	    n = n1;
    }
    n &= ~1;			/* 
				 * ensure it is even 
				 */
    scale = ldexp(1.0, n);
    x = x / scale;
    y = y / scale;		/* 
				 * Now max(|x|, |y|) is in [0.5, 1) 
				 */
    if (y < x) {
	if (x > -y)
	    vx = 0.7071, vy = 0.0;
	else
	    vx = 0.7071, vy = -0.7071;
    } else if (x > -y)
	vx = 0.7071, vy = 0.7071;
    else if (y > 0.0)
	vx = 0.0, vy = 0.7071;
    else
	vx = 0.0, vy = -0.7071;
    for (j = 0; j < 6; j++) {
	double          t = vx * vx + vy * vy;
	double          qx = (x * vx + y * vy) / t,
	    qy = (y * vx - x * vy) / t;
	vx = (vx + qx) * 0.5;
	vy = (vy + qy) * 0.5;
    }
    n = n / 2;
    z.real = ldexp(vx, n);
    z.imag = ldexp(vy, n);
    return z;
}




static          Complex
my_casinh(Complex z)
{				/* 
				 * log(z + sqrt(1 + z^2)) 
				 */
    long            quadrant = 0;
    Complex         w;
    if (z.real < 0.0) {
	z.real = -z.real;
	z.imag = -z.imag;
	quadrant = 1;
    }
    if (z.imag < 0.0) {
	z.imag = -z.imag;
	quadrant |= 2;
    }
    w.real = z.real * z.real - z.imag * z.imag + 1.0;
    w.imag = 2 * z.real * z.imag;
    w = my_csqrt(w);
    w.real += z.real;
    w.imag += z.imag;
    w = my_clog(w);
    if (quadrant & 2)
	w.imag = -w.imag;
    if (quadrant & 1)
	w.real = -w.real, w.imag = -w.imag;
    return w;
}

static          Complex
my_cacosh(Complex z)
{				/* 
				 * 2*log(sqrt((z+1)/2) + sqrt((z-1)/2)) 
				 */
    Complex         w1,
                    w2;
    w1.real = (z.real + 1.0) / 2.0;
    w2.real = (z.real - 1.0) / 2.0;
    w1.imag = w2.imag = z.imag / 2.0;
    w1 = my_csqrt(w1);
    w2 = my_csqrt(w2);
    w1.real += w2.real;
    w1.imag += w2.imag;
    z = my_clog(w1);
    z.real *= 2.0;
    z.imag *= 2.0;
    return z;
}

static          Complex
my_catanh(Complex z)
{				/* 
				 * (log(1+z) - log(1-z))/2 
				 */
    /*
     * atanh(z) = n*pi*i + 1/2 log((1+z)/(1-z)) atanh(x+yi) = 1/2
     * atanh(2x/(1+x^2+y^2))  + i/2 atan(2y/(1-x^2-y^2)) 
     */
    Complex         w1,
                    w2;
    w1.real = 1.0 + z.real;
    w2.real = 1.0 - z.real;
    w1.imag = z.imag;
    w2.imag = -z.imag;
    w1 = my_clog(w1);
    w2 = my_clog(w2);
    w1.real -= w2.real;
    w1.imag -= w2.imag;
    w1.real /= 2.0;
    w1.imag /= 2.0;
    return w1;
}





static          Complex
ccbrt(Complex a)
{
    int             xr,
                    xi,
                    i,
                    neg = 0;
    Complex         b;
    double          c,
                    cr,
                    ci;
    if (a.real == 0.0 && a.imag == 0.0)
	return a;
    else if (a.real < 0.0) {
	a.real = -a.real;
	a.imag = -a.imag;
	neg = 1;
    }
    b.real = frexp(a.real, &xr);
    b.imag = frexp(a.imag, &xi);
    /*
     * end-conditions unimportant 
     */
    if (xr > xi) {
	b.imag = ldexp(a.imag, -xr);
	b.real = 0.5996 + 0.4081 * b.real;
	b.imag = 0.5996 + 0.4081 * b.imag;
	while ((xr % 3) != 0) {
	    xr--;
	    b.real = b.real * 1.26;
	    b.imag = b.imag * 1.26;
	}
	b.real = ldexp(b.real, xr / 3);
	b.imag = ldexp(b.imag, xr / 3);
    } else {
	b.real = ldexp(a.real, -xi);
	b.real = 0.5996 + 0.4081 * b.real;
	b.imag = 0.5996 + 0.4081 * b.imag;
	while ((xi % 3) != 0) {
	    xi--;
	    b.real = b.real * 1.26;
	    b.imag = b.imag * 1.26;
	}
	b.real = ldexp(b.real, xi / 3);
	b.imag = ldexp(b.imag, xi / 3);
    }
    for (i = 0; i < 8; i++) {
	c = b.real * b.real + b.imag * b.imag;
	c = c * c;
	cr = b.real * b.real - b.imag * b.imag;
	ci = -2.0 * b.real * b.imag;
	cr = cr / c;
	ci = ci / c;
	b.real = (2.0 * b.real + (a.real * cr - a.imag * ci)) / 3.0;
	b.imag = (2.0 * b.imag + (a.real * ci + a.imag * cr)) / 3.0;
    }
    if (neg) {
	b.real = -b.real;
	b.imag = -b.imag;
    }
    return b;
}



static          Complex
my_cacos(Complex z)
{
    /*
     * acos(z) = i acosh(z) = pi/2 - i asinh(-iz) acos(x+yi) =
     * acos1/2(sqrt((1+x)^2+y^2) - sqrt((1-x)^2+y^2)) - i
     * acosh((sqrt((1+x)^2+y^2) + sqrt((1-x)^2+y^2)) 
     */
    double          t;
    z = my_cacosh(z);
    t = z.real;
    z.real = z.imag;
    z.imag = -t;
    return z;
}

static          Complex
my_casin(Complex z)
{
    /*
     * asin(z) = i sinh(-iz) asin(x+yi) = asin1/2 (sqrt((1+x)^2+y^2) -
     * sqrt((1-x)^2+y^2)) + i acosh1/2 (sqrt((1+x)^2+y^2) +
     * sqrt((1-x)^2+y^2)) 
     */
    double          t;
    t = z.real;
    z.real = -z.imag;
    z.imag = t;
    z = my_casinh(z);
    t = z.real;
    z.real = z.imag;
    z.imag = -t;
    return z;
}

static          Complex
my_catan(Complex z)
{
    /*
     * atan(z) = i atanh(-iz) atan(x+yi) = 1/2 * (pi - atan(x/(y^2-1)) +
     * atan(x/(y^2+1)) + i/4 log (x^2+(1+y)^2)/(x^2+(1-y)^2)) = 1/2
     * atan(2x/(1-x^2-y^2)) + i/2 atanh(2y/(1+x^2+y^2)) 
     */
    double          t;
    t = z.real;
    z.real = -z.imag;
    z.imag = t;
    z = my_catanh(z);
    t = z.real;
    z.real = z.imag;
    z.imag = -t;
    return z;
}


static          Sex
make_complex(double x, double y)
{
    Sex             r,
                    i;

    r = make_float(x);
    stackcheck();
    push(r);
    i = make_float(y);
    pop(r);
    return Lmkcomplex(r, i);
}

Sex
expti(Sex a, long n)
{
    Sex             b = mkfxat(1);
    if (n == 0)
	return b;
    if (n == 1)
	return a;

    if (is_modular(a)) {
	long            m = modular_modulus(a);
	n = power_mod(modular_value(a), n, m);
	b = new_modular();
	modular_value(b) = n;
	modular_modulus(b) = m;
	return b;
    }
    if (is_bmodular(a)) {
	Sex             m = bmodular_modulus(a);
	modular_expt(bmodular_value(a), mkfxat(n), m);
	b = new_bmodular();
	bmodular_value(b) = a;
	bmodular_modulus(b) = m;
	return b;
    }

    push2(a, b);
    while (n > 1) {
	if (n & 1)
	    stack[-0] = times2(stack[-1], stack[0]);
	stack[-1] = times2(stack[-1], stack[-1]);
	n = n >> 1;
    }
    pop2(b, a);
    return times2(b, a);
}



/*
 * 
 * bigfloat routine 
 */

#ifdef USE_MPFR


Sex
dsin(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;
    long            n;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    n = bnum_n(numlen(b)) - 1;
    if (bignum_value(b)[n] == 0)
	return x;

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_sin(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}


Sex
dcos(Sex x)
{

    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_cos(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

static          Sex
datan2(Sex a, Sex b)
{
    __mpfr_struct   op1,
                    op2,
                    rop;
    Sex             c,
                    d;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a) && is_bfloat(b));
    d = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op1 = bfloat2mpfr(a);
    op2 = bfloat2mpfr(b);

    mpfr_atan2(&rop, &op1, &op2, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}


Sex
dtan(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_tan(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dasin(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_asin(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dacos(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_acos(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
datan(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_atan(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

/*
 * hyperbolic functions  
 */

Sex
dsinh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_sinh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;
}

Sex
dcosh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_cosh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dtanh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_tanh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dasinh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_asinh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;


}

Sex
dacosh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_acosh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;
}

Sex
datanh(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_atanh(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

static          Sex
dsqrt(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_sqrt(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

static          Sex
dcbrt(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_cbrt(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dlog10(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_log10(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;
}

Sex
dlog2(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_log2(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}

Sex
dlog(Sex x)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(x));
    b = bfloat_man(x);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(x);

    mpfr_log(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;

}




Sex
dexp(Sex a)
{
    __mpfr_struct   op,
                    rop;
    Sex             b,
                    c;

    mpfr_prec_t     p;
    ASSERT(is_bfloat(a));
    b = bfloat_man(a);

    // precision
    p = bignum_precision(b);

    mpfr_init2(&rop, p);
    op = bfloat2mpfr(a);

    mpfr_exp(&rop, &op, MPFR_RNDN);

    c = mpfr2bfloat(&rop);
    mpfr_clear(&rop);

    return c;


}

#else


Sex
dsin(Sex x)
{
    long            m,
                    j;
    long            sign = 1;
    Sex             s;

    ASSERT(is_bfloat(x));

    if (zerop(x))
	return x;
    stackcheck();
    push4(x, nil, nil, nil);
#define x stack[-3]
#define tm stack[-2]
#define pi4 stack[-1]
#define r stack[0]
    if (minusp(x)) {
	sign = -1;
	x = negated(x);
    }
    pi4 = leftshiftd(Lpi(), -2);
    if (lessp2dd(x, pi4)) {
	m = 0;
	r = x;
    } else {
	tm = Lfix(quot2dd(x, pi4));
	if (is_fixnum(tm)) {
	    r = times2id(tm, pi4);
	    r = diff2dd(x, r);
	    /*
	     * r = diff2dd(x, times2id(tm, pi4)); 
	     */
	    m = mkint(tm) % 8;
	} else {
	    r = times2bd(tm, pi4);
	    r = diff2dd(x, r);
	    /*
	     * r = diff2dd(x, times2bd(tm, pi4)); 
	     */
	    tm = rem2bi(tm, mkfxat(8));
	    m = mkint(tm);
	}
    }
    if (m >= 4) {
	sign = -sign;
	m = m - 4;
    }
    switch (m) {
    case 0:
	break;
    case 1:
	if (zerop(r)) {
	    s = dsqrt(Lbfloat(mkfxat(2)));
	    s = leftshiftd(s, -1);
	    if (sign < 0)
		s = negated(s);
	    popv(4);
	    return s;
	}
	r = diff2dd(pi4, r);
    case 2:
	s = dcos(r);
	if (sign < 0)
	    s = negated(s);
	popv(4);
	return s;
    case 3:
	r = diff2dd(pi4, r);
    }
    x = tm = r;
    if (!zeropd(x)) {
	r = negated(times2dd(r, r));
	j = 1;
	m = current_b_precision;
	do {
	    j = j + 2;
	    tm = times2dd(tm, r);
	    tm = quot2di(tm, mkfxat(j * (j - 1)));
	    x = plus2dd(x, tm);
	} while (bfloat_exp(x) - bfloat_exp(tm) < m);
    }
    s = x;
    if (sign < 0)
	s = negated(s);
    popv(4);
#undef x
#undef tm
#undef pi4
#undef r
    return s;
}


Sex
dcos(Sex x)
{
    long            m,
                    j;
    long            sign = 1;
    Sex             s;

    ASSERT(is_bfloat(x));

    if (zerop(x))
	return Lbfloat(mkfxat(1));
    stackcheck();
    push4(x, nil, nil, nil);
#define x stack[-3]
#define tm stack[-2]
#define pi4 stack[-1]
#define r stack[0]
    if (minusp(x)) {
	x = negated(x);
    }
    pi4 = leftshiftd(Lpi(), -2);
    if (lessp2dd(x, pi4)) {
	m = 0;
	r = x;
    } else {
	tm = Lfix(quot2dd(x, pi4));
	if (is_fixnum(tm)) {
	    r = times2id(tm, pi4);
	    r = diff2dd(x, r);
	    /*
	     * r = diff2dd(x, times2id(tm, pi4)); 
	     */
	    m = mkint(tm) % 8;
	} else {
	    r = times2bd(tm, pi4);
	    r = diff2dd(x, r);
	    /*
	     * r = diff2dd(x, times2bd(tm, pi4)); 
	     */
	    tm = rem2bi(tm, mkfxat(8));
	    m = mkint(tm);
	}
    }
    if (m >= 4) {
	sign = -sign;
	m = m - 4;
    }
    switch (m) {
    case 0:
	break;
    case 1:
	if (zerop(r)) {
	    popv(4);
	    s = dsqrt(Lbfloat(mkfxat(2)));
	    s = leftshiftd(s, -1);
	    if (sign < 0)
		s = negated(s);
	    popv(4);
	    return s;
	}
	sign = -sign;
	r = diff2dd(pi4, r);
    case 2:
	sign = -sign;
	s = dsin(r);
	if (sign < 0)
	    s = negated(s);
	popv(4);
	return s;
    case 3:
	sign = -sign;
	r = diff2dd(pi4, r);
	break;
    }
    x = tm = Lbfloat(mkfxat(1));
    if (!zeropd(r)) {
	r = negated(times2dd(r, r));
	j = 0;
	m = current_b_precision;
	do {
	    j = j + 2;
	    tm = times2dd(tm, r);
	    tm = quot2di(tm, mkfxat(j * (j - 1)));
	    x = plus2dd(x, tm);
	} while (bfloat_exp(x) - bfloat_exp(tm) < m);
    }
    s = x;
    if (sign < 0)
	s = negated(s);
    popv(4);
#undef x
#undef tm
#undef pi4
#undef r
    return s;
}

static          Sex
datan2(Sex a, Sex b)
{
    long            q = 0;

    ASSERT(is_bfloat(a) && is_bfloat(b));
    stackcheck();
    push3(a, b, nil);
#define a stack[-2]
#define b stack[-1]
#define r stack[ 0]
    if (minusp(a)) {
	a = negated(a);
	q = 1;
    }
    if (minusp(b)) {
	b = negated(b);
	q |= 2;
    }
    if (lessp2dd(a, b)) {
	Sex             x;
	x = a;
	a = b;
	b = x;
	q |= 4;
    }
    if (zerop(a) && zerop(b))
	r = Lbfloat(mkfxat(0));
    else {

	r = datan(quot2dd(b, a));
	switch (q) {
	default:
	case 0:
	    break;
	case 1:
	    a = Lpi();
	    r = diff2dd(a, r);
	    break;
	case 2:
	    r = negated(r);
	    break;
	case 3:
	    a = Lpi();
	    r = diff2dd(r, a);
	    break;
	case 4:
	    a = Lpi();
	    a = dldexp(a, -1);
	    r = diff2dd(a, r);
	    break;
	case 5:
	    a = Lpi();
	    a = dldexp(a, -1);
	    r = plus2dd(a, r);
	    break;
	case 6:
	    a = Lpi();
	    a = dldexp(a, -1);
	    r = diff2dd(r, a);
	    break;
	case 7:
	    a = Lpi();
	    a = dldexp(a, -1);
	    r = negated(plus2dd(r, a));
	    break;
	}
    }
#undef a
#undef b
#undef r
    a = stack[0];
    popv(3);
    return a;
}


Sex
dtan(Sex x)
{
    long            m;
    long            sign = 1;
    Sex             s;

    ASSERT(is_bfloat(x));

    if (zerop(x))
	return Lbfloat(mkfxat(0));
    stackcheck();
    push4(x, nil, nil, nil);
#define x stack[-3]
#define tm stack[-2]
#define pi4 stack[-1]
#define r stack[0]
    if (minusp(x)) {
	x = negated(x);
	sign = -1;
    }
    pi4 = leftshiftd(Lpi(), -2);
    if (lessp2dd(x, pi4)) {
	m = 0;
	r = x;
    } else {
	tm = Lfix(quot2dd(x, pi4));
	if (is_fixnum(tm)) {
	    s = times2id(tm, pi4);
	    r = diff2dd(x, s);
	    m = mkint(tm) % 4;
	} else {
	    s = times2bd(tm, pi4);
	    r = diff2dd(x, s);
	    tm = rem2bi(tm, mkfxat(4));
	    m = mkint(tm);
	}
    }
    switch (m) {
    case 1:
	r = diff2dd(pi4, r);
	sign = -sign;
    case 2:
	sign = -sign;
	x = dsin(r);
	s = times2dd(x, x);
	s = diff2id(mkfxat(1), s);
	s = dsqrt(s);
	s = quot2dd(s, x);
	break;
    case 3:
	sign = -sign;
	r = diff2dd(pi4, r);
    case 0:
	x = dsin(r);
	s = times2dd(x, x);
	s = diff2id(mkfxat(1), s);
	s = dsqrt(s);
	s = quot2dd(x, s);
	break;
    default:
	return error("dtan");
    }
#undef x
#undef tm
#undef pi4
#undef r
    if (sign < 0)
	s = negated(s);
    popv(4);
    return s;
}

Sex
dasin(Sex x)
{
    Sex             a;
    long            sign = 1;
    ASSERT(is_bfloat(x));
    if (zerop(x))
	return x;
    if (minusp(x)) {
	sign = -1;
	x = negated(x);
    }
    if (onep(x)) {
	return quot2di(Lpi(), mkfxat(2 * sign));
    }
    stackcheck();
    push(x);
    if (lessp2id(mkfxat(1), x))
	error("asin");
    x = stack[0];
    a = times2dd(x, x);
    a = diff2id(mkfxat(1), a);
    a = dsqrt(a);
    pop(x);
    a = quot2dd(x, a);
    a = datan(a);
    if (sign < 0)
	a = negated(a);
    return a;
}

Sex
dacos(Sex x)
{
    Sex             a;
    long            sign = 1;
    ASSERT(is_bfloat(x));
    if (minusp(x)) {
	sign = -1;
	x = negated(x);
    }
    if (onep(x)) {
	if (sign > 0)
	    return Lbfloat(mkfxat(0));
	else
	    return Lpi();
    }
    stackcheck();
    push(x);
    if (lessp2id(mkfxat(1), x))
	error("acos");
    x = stack[0];
    a = times2dd(x, x);
    a = diff2id(mkfxat(1), a);
    a = dsqrt(a);
    pop(x);
    a = quot2dd(a, x);
    a = datan(a);
    if (sign < 0) {
	push(a);
	x = Lpi();
	pop(a);
	a = diff2dd(x, a);
    }
    return a;
}

Sex
datan(Sex x)
{
    long            sign = 1;
    long            cot = 0;
    long            j,
                    m;
    Sex             s;
    ASSERT(is_bfloat(x));
    if (zerop(x))
	return x;
    stackcheck();
    push3(x, nil, nil);
#define x stack[-2]
#define y stack[-1]
#define tm stack[0]
    if (minusp(x)) {
	sign = -1;
	x = negated(x);
    }
    if (lessp2id(mkfxat(1), x)) {
	cot = 1;
	x = quot2id(mkfxat(1), x);
    }
    if (lessp2fd(make_float(0.419921875), x)) {
	s = times2dd(x, x);
	s = plus2id(mkfxat(1), s);
	s = dsqrt(s);
	s = plus2id(mkfxat(1), s);
	x = quot2dd(x, s);
	sign = sign * 2;
    }
    /*
     * taylor expansion of atan(x) atan(x) = x - x^3/3 + x^5/5 - x^7/7 + .... 
     */
    tm = x;
    if (!zeropd(x)) {
	x = negated(times2dd(x, x));
	j = 1;
	y = tm;
	m = current_b_precision;
	do {
	    j = j + 2;
	    tm = times2dd(tm, x);
	    s = quot2di(tm, mkfxat(j));
	    y = plus2dd(y, s);
	} while (bfloat_exp(y) - bfloat_exp(tm) < m);
    }

    x = leftshiftd(Lpi(), -1);
    if (cot) {
	y = diff2dd(x, y);
    }
    if (sign != 1)
	y = times2id(mkfxat(sign), y);
    if (lessp2dd(x, y))
	y = diff2dd(y, x);
    s = y;
    popv(3);
#undef x
#undef y
#undef tm
    return s;
}

/*
 * hyperbolic functions  
 */

Sex
dsinh(Sex a)
{
    Sex             e;
    ASSERT(is_bfloat(a));
    e = dexp(a);
    stackcheck();
    push(e);
    a = quot2id(mkfxat(1), e);
    pop(e);
    a = diff2dd(e, a);
    a = leftshiftd(a, -1);
    return a;
}

Sex
dcosh(Sex a)
{
    Sex             e;
    ASSERT(is_bfloat(a));
    e = dexp(a);
    stackcheck();
    push(e);
    a = quot2id(mkfxat(1), e);
    pop(e);
    a = plus2dd(e, a);
    a = leftshiftd(a, -1);
    return a;
}

Sex
dtanh(Sex a)
{
    Sex             e;
    ASSERT(is_bfloat(a));
    e = dexp(a);
    stackcheck();
    push(e);
    a = quot2id(mkfxat(1), e);
    push(a);
    push(nil);
#define e stack[-2]
#define e1 stack[-1]
#define s stack[ 0]
    s = diff2dd(e, e1);
    a = plus2dd(e, e1);
    a = quot2dd(s, a);
    popv(3);
#undef e
#undef e1
#undef s
    return a;
}

Sex
dasinh(Sex a)
{				/* 
				 * log(x+sqrt(x^2+1)) 
				 */
    long            sign = 1;
    Sex             x;

    ASSERT(is_bfloat(a));
    if (minusp(a)) {
	sign = -1;
	a = negated(a);
    }
    stackcheck();
    push(a);
    x = times2dd(a, a);
    x = plus2id(mkfxat(1), x);
    x = dsqrt(x);
    pop(a);
    x = plus2dd(x, a);
    x = dlog(x);
    if (sign < 0)
	x = negated(x);
    return x;
}

Sex
dacosh(Sex a)
{				/* 
				 * log(x+sqrt(x^2-1)) 
				 */
    long            sign = 1;
    Sex             x;

    ASSERT(is_bfloat(a));
    if (minusp(a)) {
	sign = -1;
	a = negated(a);
    }
    stackcheck();
    push(a);
    x = times2dd(a, a);
    x = diff2di(x, mkfxat(1));
    x = dsqrt(x);
    pop(a);
    x = plus2dd(x, a);
    x = dlog(x);
    if (sign < 0)
	x = negated(x);
    return x;
    /*
     * Sex             x; int             f; push(a); x = times2dd(a, a); x =
     * diff2id(mkfxat(1), x); x = dsqrt(x); pop(a); f = minusp(a); x =
     * quot2dd(x, a); x = datan(x); if (f) return plus2dd(x, Lpi()); else
     * return x; 
     */

}

Sex
datanh(Sex a)
{

    /*
     *  atanh(x) = x + x^3/3 + x^5/5+...  abs(x) < 2^-bprec
     *             x                       |x| < 2^(-bprec/2)
     *             x+x^3/3                 |x| < 2^(-bprec/4)
     *             x+x^3/3+x^5/5           |x| < 2^(-bprec/8)
     *             0.5*log((1+x)/(1-x))
     */

    Sex             x;

    ASSERT(is_bfloat(a));
    stackcheck();
    push(a);
    x = plus2id(mkfxat(1), a);
    push(x);
    a = diff2id(mkfxat(1), stack[-1]);
    pop(x);
    x = quot2dd(x, a);
    x = dlog(x);
    x = leftshiftd(x, -1);
    popv(1);
    return x;
}

static          Sex
dsqrt(Sex a)
{
    /*
     * x~ = 1/2*(x+ a/x); x -> sqrt(a); 
     * or x~ = 1/2*(3x - ax^3) = 1/2* x * (3-a x^2);  ax -> sqrt(a); 
     */
    long            k;
    ASSERT(is_bfloat(a));

    if (zerop(a))
	return a;
    if (minusp(a))
	return aerror1("sqrt", a);
    k = msd(current_w_precision) + 4;
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define s stack[-1]
#define t stack[0]
    s = new_bfloat();
    bfloat_man(s) = bfloat_man(a);
    bfloat_exp(s) = (-current_b_precision + bfloat_exp(a)) / 2;
    while (k-- > 0) {
	t = quot2dd(a, s);
	s = plus2dd(s, t);
	bfloat_exp(s) = bfloat_exp(s) - 1;
    }
#undef a
#undef s
#undef t
    a = stack[-1];
    popv(3);
    return a;
}

static          Sex
dcbrt(Sex a)
{
    /*
     * f = x^3 - a 
     * f' = 3x^2 
     * x~ = x - f/f' = x - (x^3 - a)/3x^2 
     *    = 2/3*x + 1/3*a/x^2 or
     * x~ = 1/2(x + 3x/(1+(2/a)*x^3))  
     */
    long            k;
    ASSERT(is_bfloat(a));
    if (zerop(a))
	return a;
    k = msd(current_w_precision) + 5;
    stackcheck();
    push3(a, nil, nil);
#define a stack[-2]
#define s stack[-1]
#define t stack[0]
    s = new_bfloat();
    bfloat_man(s) = bfloat_man(a);
    bfloat_exp(s) = (-2 * current_b_precision + bfloat_exp(a)) / 3;
    while (k-- > 0) {
	t = times2dd(s, s);
	s = leftshiftd(s, 1);
	t = quot2dd(a, t);
	s = plus2dd(s, t);
	s = quot2di(s, mkfxat(3));
    }
#undef a
#undef s
#undef t
    a = stack[-1];
    popv(3);
    return a;
}

Sex
dlog10(Sex a)
{
    Sex             log10;
    ASSERT(is_bfloat(a));
    a = dlog(a);
    stackcheck();
    push(a);
    log10 = dlog(bfloatiti(mkfxat(10)));
    pop(a);
    return quot2dd(a, log10);
}

Sex
dlog2(Sex a)
{
    Sex             log2;
    ASSERT(is_bfloat(a));
    a = dlog(a);
    stackcheck();
    push(a);
    log2 = dlog(bfloatiti(mkfxat(2)));
    pop(a);
    return quot2dd(a, log2);
}

Sex
dlog(Sex x)
{
    /*
     * y = (1+x)/(1-x) x = (y-1)/(y+1) log(y) = 2*(x+x^3/3+x^5/5+...) 
     */
    Sex             a;
    long            k = current_b_precision + 7;
    long            m;
    long            nn;

    ASSERT(is_bfloat(x));
    if (minusp(x))
	error("dlog with negative");

    stackcheck();
    push3(x, nil, nil);
#define x stack[-2]
#define n stack[0]
#define ans stack[-1]
    nn = bfloat_log2(x) + 1;
    if (nn > 10000000 || nn < -10000000) {
	nn = nn / 277;
	nn = nn * 192;
    } else {
	nn = nn * 192;
	nn = nn / 277;
    }
    ans = Le();
    if (nn > 0) {
	ans = expti(ans, nn);
	x = quot2dd(x, ans);
    } else if (nn < 0) {
	ans = expti(ans, -nn);
	x = times2dd(x, ans);
    }
    n = diff2id(mkfxat(1), x);
    x = plus2id(mkfxat(1), x);
    x = quot2dd(n, x);
    x = leftshift(bfloat_man(x), bfloat_exp(x) + k);
    n = leftshifti(mkfxat(1), k);
    if (lessp2(n, x))
	error("log_x");
    n = x;
    ans = x;
    x = times2(x, x);
    x = leftshift(x, -k);
    m = 3;
    while (n != mkfxat(0)) {
	n = times2(n, x);
	n = leftshift(n, -k);
	a = quot2(n, mkfxat(m));
	ans = plus2(ans, a);
	m = m + 2;
    }
    a = times2(ans, mkfxat(2));
    a = negate(a);
#undef x
#undef n
#undef ans
    popv(3);
    a = make_bfloat(a, -k);
    if (nn != 0)
	a = plus2id(mkfxat(nn), a);
    ASSERT(is_bfloat(a));
    return a;
}




Sex
dexp(Sex a)
{
    long            n;

    ASSERT(is_bfloat(a));
    /*
     * lisp_printf("dexp a = %p\n", a);
     */
    if (zerop(a))
	return Lbfloat(mkfxat(1));

    stackcheck();
    push3(a, nil, nil);
#define b stack[-2]
#define i stack[-1]
#define e stack[0]
    b = a;
    i = Lfix(a);
    if (i == mkfxat(0)) {
	popv(3);
	return Le_x(a);
    }
    b = diff2(b, i);
    if (!is_fixnum(i)) {
	if (minusp(i)) {
	    popv(3);
	    return Lbfloat(mkfxat(0));
	} else {
	    return error("dexp: argument too big");
	}
    }
    n = mkint(i);
    if (n < 0) {
	e = Le();
	e = expti(e, -n);
	b = Le_x(b);
	a = quot2dd(b, e);
    } else {
	e = Le();
	e = expti(e, n);
	b = Le_x(b);
	a = times2dd(e, b);
    }
    popv(3);
#undef b
#undef i
#undef e
    ASSERT(is_bfloat(a));
    return a;
}

#endif


/*
 * double complex
 */


Sex
dcabs(Sex zr, Sex zi)
{
    Sex             r,
                    i;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push(zi);
    r = times2dd(zr, zr);
    pop(zi);
    push(r);
    i = times2dd(zi, zi);
    pop(r);
    r = plus2dd(r, i);
    return dsqrt(r);
}

Sex
dcsin(Sex zr, Sex zi)
{
    /*
     * sin(x + iy) = sin(x)*cosh(y) + i cos(x)*sinh(y) 
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
    push(nil);
#define zr stack[-5]
#define zi stack[-4]
#define s  stack[-3]
#define c  stack[-2]
#define sh stack[-1]
#define ch stack[ 0]
    s = dsin(zr);
    c = dcos(zr);
    sh = dsinh(zi);
    ch = dcosh(zi);
    zr = times2dd(s, ch);
    zi = times2dd(c, sh);
#undef zr
#undef zi
#undef s
#undef c
#undef sh
#undef ch
    zr = stack[-5];
    zi = stack[-4];
    popv(6);
    return Lmkcomplex(zr, zi);
}

Sex
dccos(Sex zr, Sex zi)
{
    /*
     * cos(x+yi) = cos x cosh y - i sin x sinh y 
     */
    /*
     * cos(x + iy) = cos(x)*cosh(y) - i sin(x)*sinh(y) 
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
    push(nil);
#define zr stack[-5]
#define zi stack[-4]
#define s  stack[-3]
#define c  stack[-2]
#define sh stack[-1]
#define ch stack[ 0]
    s = dsin(zr);
    c = dcos(zr);
    sh = dsinh(zi);
    ch = dcosh(zi);
    zr = times2dd(c, ch);
    zi = times2dd(s, sh);
    zi = negated(zi);
#undef zr
#undef zi
#undef s
#undef c
#undef sh
#undef ch
    zr = stack[-5];
    zi = stack[-4];
    popv(6);
    return Lmkcomplex(zr, zi);
}

Sex
dcexp(Sex zr, Sex zi)
{
    /*
     * exp(x + iy) = cos(y)*exp(x) + i sin(y)*exp(x) 
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
#define zr stack[-4]
#define zi stack[-3]
#define s  stack[-2]
#define c  stack[-1]
#define e  stack[00]
    s = dsin(zi);
    c = dcos(zi);
    e = dexp(zr);
    zr = times2dd(c, e);
    zi = times2dd(s, e);
#undef zr
#undef zi
#undef s
#undef c
#undef e
    zr = stack[-4];
    zi = stack[-3];
    popv(5);
    return Lmkcomplex(zr, zi);
}

Sex
dcsinh(Sex zr, Sex zi)
{
    /*
     * sinh(z) = -i sin(iz) sinh(x+yi) = sinhx cos y + i cosh x sin y 
     */
    /*
     * sinh(x + iy) = (e(x+iy)-e(-x-iy))/2
     *              =(e^x+e^-x)*(e^iy-e^-iy)+(e^x-e^-x)(e^iy+e^-iy)/4
     *              =icosh(x)sin(y)+sinh(x)cos(y);
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
    push(nil);
#define zr stack[-5]
#define zi stack[-4]
#define s  stack[-3]
#define c  stack[-2]
#define sh stack[-1]
#define ch stack[ 0]
    s = dsin(zi);
    c = dcos(zi);
    sh = dsinh(zr);
    ch = dcosh(zr);
    zr = times2dd(c, sh);
    zi = times2dd(s, ch);
#undef zr
#undef zi
#undef s
#undef c
#undef sh
#undef ch
    zr = stack[-5];
    zi = stack[-4];
    popv(6);
    return Lmkcomplex(zr, zi);
}

Sex
dccosh(Sex zr, Sex zi)
{
    /*
     * cosh(x+yi) = cosh x coshx + i sinhx sin y 
     */
    /*
     * cosh(x + iy) = (e(x+iy)+e(-x-iy))/2
     *              =(e^x+e^-x)*(e^iy+e^-iy)+(e^x-e^-x)(e^iy-e^-iy)/4
     *              =cosh(x)cos(y)+isinh(x)sin(y);
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
    push(nil);
#define zr stack[-5]
#define zi stack[-4]
#define s  stack[-3]
#define c  stack[-2]
#define sh stack[-1]
#define ch stack[ 0]
    s = dsin(zi);
    c = dcos(zi);
    sh = dsinh(zr);
    ch = dcosh(zr);
    zr = times2dd(c, ch);
    zi = times2dd(s, sh);
#undef zr
#undef zi
#undef s
#undef c
#undef sh
#undef ch
    zr = stack[-5];
    zi = stack[-4];
    popv(6);
    return Lmkcomplex(zr, zi);
}

static          Sex
dctanh(Sex zr, Sex zi)
{
    /*
     * tanh(x+yi) = (sinh(2x)+isin(2y))/(cosh2x + cos2y) 
     * tanh(z) = -i  tan(iz) 
     */
    Sex             z,
                    x;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    push(zr);
    zi = negate(zi);
    pop(zr);
    z = dctan(zi, zr);
    if (is_complex(z)) {
	push(z);
	x = negate(complex_real(z));
	pop(z);
	return Lmkcomplex(complex_imag(z), x);
    }
    z = negate(z);
    return Lmkcomplex(mkfxat(0), z);
}


Sex
dldexp(Sex a, long n)
{
    Sex             c;

    ASSERT(is_bfloat(a));
    stackcheck();
    push(a);
    c = new_bfloat();
    pop(a);
    bfloat_man(c) = bfloat_man(a);
    bfloat_exp(c) = bfloat_exp(a) + n;
    return c;
}

Sex
dabs(Sex a)
{
    if (minusp(a))
	a = negate(a);
    return a;
}

Sex
dclog(Sex x, Sex y)
{
    stackcheck();
    ASSERT(is_bfloat(x) && is_bfloat(y));
    push4(x, y, nil, nil);
#define x stack[-3]
#define y stack[-2]
#define s stack[-1]
#define r stack[ 0]
    if (zerop(x))
	r = dlog(dabs(y));
    else if (zerop(y))
	r = dlog(dabs(x));
    else {
	s = times2dd(x, x);
	r = times2dd(y, y);
	r = dlog(plus2dd(s, r));
	r = dldexp(r, -1);
    }
    if (zerop(y))
	if (minusp(x))
	    y = Lpi();
	else
	    y = Lbfloat(mkfxat(0));
    else
	y = datan2(y, x);
#undef x
#undef y
#undef s
#undef r
    x = stack[0];
    y = stack[-2];
    popv(4);
    return Lmkcomplex(x, y);
}


Sex
dclog2(Sex zr, Sex zi)
{
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    z = dclog(zr, zi);
    stackcheck();
    push3(z, nil, nil);
#define z stack[-2]
#define log2 stack[-1]
#define t stack[0]
    log2 = dlog(bfloatiti(mkfxat(2)));
    t = quot2dd(complex_real(z), log2);
    zi = quot2dd(complex_imag(z), log2);
    zr = t;
    popv(3);
#undef z
#undef log2
#undef t
    return Lmkcomplex(zr, zi);
}

Sex
dclog10(Sex zr, Sex zi)
{
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    z = dclog(zr, zi);
    stackcheck();
    push3(z, nil, nil);
#define z stack[-2]
#define log10 stack[-1]
#define t stack[0]
    log10 = dlog(bfloatiti(mkfxat(10)));
    t = quot2dd(complex_real(z), log10);
    zi = quot2dd(complex_imag(z), log10);
    zr = t;
    popv(3);
#undef z
#undef log10
#undef t
    return Lmkcomplex(zr, zi);
}

Sex
dcsqrt(Sex x, Sex y)
{
    /*
     * f = x^3 - a f' = 3x^2 
     * x~ = x - f/f' = x - (x^3 - a)/3x^2 
     *               = (2x + a/x^2 )/3
     */

    long            k,
                    n;
    ASSERT(is_bfloat(x) && is_bfloat(y));
    stackcheck();
    if (zerop(y)) {
	if (minusp(x)) {
	    push(y);
	    y = dsqrt(negate(x));
	    pop(y);
	} else {
	    push(y);
	    x = dsqrt(x);
	    pop(y);
	}
	return Lmkcomplex(x, y);	/* 
					 * * Pure real arguments are easy  
					 */
    }
    k = msd(current_w_precision) + 4;
    push4(x, y, nil, nil);
    push4(nil, nil, nil, nil);
#define x stack[-7]
#define y stack[-6]
#define t stack[-5]
#define s stack[-4]
#define u stack[-3]
#define v stack[-2]
#define zr stack[-1]
#define zi stack[0]
    n = bfloat_exp(x);
    if (n < bfloat_exp(y))
	n = bfloat_exp(y);
    n = (-current_b_precision - n) / 2;
    zr = make_bfloat(bfloat_man(x), bfloat_exp(x) + n);
    zi = make_bfloat(bfloat_man(y), bfloat_exp(y) + n);
    while (k-- > 0) {
	t = times2dd(zr, zr);
	s = times2dd(zi, zi);
	u = plus2dd(t, s);

	t = times2dd(x, zr);
	s = times2dd(y, zi);
	v = plus2dd(t, s);

	t = times2dd(x, zi);
	s = times2dd(y, zr);
	t = diff2dd(s, t);

	t = quot2dd(t, u);
	s = quot2dd(v, u);

	zr = plus2dd(zr, s);
	zi = plus2dd(zi, t);
	zr = leftshiftd(zr, -1);
	zi = leftshiftd(zi, -1);
    }
#undef x
#undef y
#undef t
#undef s
#undef u
#undef v
#undef zr
#undef zi
    x = stack[-1];
    y = stack[0];
    popv(8);
    return Lmkcomplex(x, y);
}

Sex
dccbrt(Sex x, Sex y)
{
    /*
     * */
    long            k,
                    n;
    ASSERT(is_bfloat(x) && is_bfloat(y));
    if (zerop(y)) {
	return dcbrt(x);
    }
    k = msd(current_w_precision) + 4;
    stackcheck();
    push4(x, y, nil, nil);
    push5(nil, nil, nil, nil, nil);
#define x stack[-8]
#define y stack[-7]
#define t stack[-6]
#define s stack[-5]
#define u stack[-4]
#define v stack[-3]
#define w stack[-2]
#define zr stack[-1]
#define zi stack[0]
    n = bfloat_exp(x);
    if (n < bfloat_exp(y))
	n = bfloat_exp(y);
    n = -(current_b_precision + n) * 2 / 3;
    zr = make_bfloat(bfloat_man(x), bfloat_exp(x) + n);
    zi = make_bfloat(bfloat_man(y), bfloat_exp(y) + n);
    while (k-- > 0) {
	t = times2dd(zr, zr);
	s = times2dd(zi, zi);
	u = diff2dd(t, s);
	v = times2dd(zr, zi);
	v = leftshiftd(v, 1);

	t = times2dd(u, u);
	s = times2dd(v, v);
	w = plus2dd(t, s);

	t = times2dd(x, u);
	s = times2dd(y, u);

	u = times2dd(x, v);
	v = times2dd(y, v);

	t = plus2dd(t, v);
	s = diff2dd(s, u);


	t = quot2dd(t, w);
	s = quot2dd(s, w);

	zr = leftshiftd(zr, 1);
	zi = leftshiftd(zi, 1);
	zr = plus2dd(zr, t);
	zi = plus2dd(zi, s);
	zr = quot2di(zr, mkfxat(3));
	zi = quot2di(zi, mkfxat(3));

    }
#undef x
#undef y
#undef t
#undef s
#undef u
#undef v
#undef w
#undef zr
#undef zi
    x = stack[-1];
    y = stack[0];
    popv(9);
    return Lmkcomplex(x, y);
}


Sex
dctan(Sex zr, Sex zi)
{
    /*
     * tan(x + iy) = (tan(x) + i tanh(y))/(1 - i tan(x)*tanh(y))
     * tan(x+yi) = (sin 2x + i sinh(2y))/(cos2x + cosh2y)
     */
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push4(zr, zi, nil, nil);
    push3(nil, nil, nil);
#define zr stack[-6]
#define zi stack[-5]
#define t  stack[-4]
#define th stack[-3]
#define t2 stack[-2]
#define th2 stack[-1]
#define d  stack[ 0]
    t = dtan(zr);
    th = dtanh(zi);
    t2 = times2dd(t, t);
    th2 = times2dd(th, th);
    d = times2dd(t2, th2);
    d = plus2id(mkfxat(1), d);
    d = quot2id(mkfxat(1), d);
    zr = diff2id(mkfxat(1), th2);
    zi = plus2id(mkfxat(1), t2);
    zr = times2dd(t, zr);
    zi = times2dd(th, zi);
    zr = times2dd(zr, d);
    zi = times2dd(zi, d);
#undef zr
#undef zi
#undef t
#undef th
#undef th2
#undef t2
#undef d
    zr = stack[-6];
    zi = stack[-5];
    popv(7);
    return Lmkcomplex(zr, zi);
}

static          Sex
dcasin(Sex zr, Sex zi)
{
    /*
     * arcsin(z) = i log(-iz + sqrt(1-z^2)) 
     * arcsin(z) = pi/2 - i arccosh(z) 
     * asin(z) = i sinh(-iz) 
     */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push(zi);
    zr = negate(zr);
    pop(zi);
    z = dcasinh(zi, zr);
    zr = complex_imag(z);
    zi = complex_real(z);
    push(zr);
    zi = negate(zi);
    pop(zr);
    return Lmkcomplex(zi, zr);
}


static          Sex
dcacos(Sex zr, Sex zi)
{
    /*
     * arccos(z) = i log(z + sqrt(z^2-1)) 
     * arccos(z) = i arccosh(z) 
     */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    z = dcacosh(zr, zi);
    zr = complex_imag(z);
    zi = complex_real(z);
    stackcheck();
    push(zr);
    zi = negate(zi);
    pop(zr);
    return Lmkcomplex(zr, zi);
}

static          Sex
dcatan(Sex zr, Sex zi)
{
    /*
     * arctan z = (i/2) log(1-iz)/(1+iz) 
     * arctan iz = i arctan(z) 
     */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push(zr);
    zi = negate(zi);
    pop(zr);
    z = dcatanh(zi, zr);
    zr = complex_imag(z);
    zi = complex_real(z);
    push(zr);
    zi = negate(zi);
    pop(zr);
    return Lmkcomplex(zr, zi);
}

Sex
dcasinh(Sex zr, Sex zi)
{				/* 
				 * log(z + sqrt(1 + z^2)) 
				 */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
#define zr stack[-4]
#define zi stack[-3]
#define w1 stack[-2]
#define w2 stack[-1]
#define a  stack[ 0]
    w1 = times2dd(zr, zr);
    w2 = times2dd(zi, zi);
    w1 = diff2dd(w1, w2);
    w2 = times2id(mkfxat(2), times2dd(zr, zi));
    w1 = plus2id(mkfxat(1), w1);
    a = dcsqrt(w1, w2);
    zr = plus2dd(zr, complex_real(a));
    zi = plus2dd(zi, complex_imag(a));
    z = dclog(zr, zi);
#undef zr
#undef zi
#undef w1
#undef w2
#undef a
    popv(5);
    return z;
}


Sex
dcacosh(Sex zr, Sex zi)
{				/* 
				 * 2*log(sqrt((z+1)/2) + sqrt((z-1)/2)) 
				 */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push5(zr, zi, nil, nil, nil);
#define zr stack[-4]
#define zi stack[-3]
#define w1 stack[-2]
#define w2 stack[-1]
#define a  stack[ 0]
    a = leftshiftd(zi, -1);
    z = plus2id(mkfxat(1), zr);
    z = leftshiftd(z, -1);
    w1 = dcsqrt(z, a);
    z = diff2di(zr, mkfxat(1));
    z = leftshiftd(z, -1);
    w2 = dcsqrt(z, a);
    z = plus2(w1, w2);
    z = dclog(complex_real(z), complex_imag(z));
    z = times2(z, mkfxat(2));
#undef zr
#undef zi
#undef w1
#undef w2
#undef a
    popv(5);
    return z;
}

Sex
dcatanh(Sex zr, Sex zi)
{				/* 
				 * n*pi*i + (log(1+z) - log(1-z))/2 
				 */
    Sex             z;
    ASSERT(is_bfloat(zr) && is_bfloat(zi));
    stackcheck();
    push4(zr, zi, nil, nil);
#define zr stack[-3]
#define zi stack[-2]
#define w1 stack[-1]
#define w2 stack[ 0]
    z = plus2id(mkfxat(1), zr);
    w1 = dclog(z, zi);
    z = diff2id(mkfxat(1), zr);
    w2 = dclog(z, zi);
    z = diff2(w1, w2);
    z = quot2(z, mkfxat(2));
#undef zr
#undef zi
#undef w1
#undef w2
    popv(4);
    return z;
}


/*
 */


typedef double  double_fn(double);
typedef Sex     bfloat_fn(Sex);
typedef Complex complex_fn(Complex);
typedef Sex     dcomplex_fn(Sex, Sex);



#ifdef ASUKA
static          Sex
trig_fn(Sex a,
	double_fn * f_fn,
	bfloat_fn * b_fn, complex_fn * c_fn, dcomplex_fn * d_fn,
	Sex fn_symbol)
#else
static          Sex
trig_fn(Sex a,
	double_fn * f_fn,
	bfloat_fn * b_fn, complex_fn * c_fn, dcomplex_fn * d_fn, char *msg)
#endif
{
    double          x;
    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	a = bfloatiti(a);
	return b_fn(a);
    case TAG_NUMBERS:
	switch (numtype_of(a)) {
	case N_BIGNUM:
	    a = bfloatitb(a);
	    return b_fn(a);
	case N_RATIONAL:
	    a = bfloatitr(a);
	    return b_fn(a);
	case N_INTERVAL:
	case N_EFLOAT:
	    x = floatit(a);
	    return make_float(f_fn(x));
	case N_FLOAT:
	    x = float_val(a);
	    return make_float(f_fn(x));
	case N_COMPLEX:
	    {
		Sex             ar,
		                ai;
		ar = complex_real(a);
		ai = complex_imag(a);
#if TOWARD_LOWER_PRECISION
		if ((is_app(ar) && !is_big(ar))
		    || (is_app(ai) && !is_big(ai))) {
		    Complex         z;
		    z.real = floatit(ar);
		    z.imag = floatit(ai);
		    z = c_fn(z);
		    return make_complex(z.real, z.imag);
		} else {
		    stackcheck();
		    push(ai);
		    ar = bfloatit(ar);
		    a = stack[0];
		    stack[0] = ar;
		    ai = bfloatit(a);
		    pop(ar);
		    return d_fn(ar, ai);
		}
#else
		if (is_big(ar) || is_big(ai)) {
		    stackcheck();
		    push(ai);
		    ar = bfloatit(ar);
		    ai = stack[0];
		    stack[0] = ar;
		    ai = bfloatit(ai);
		    pop(ar);
		    return d_fn(ar, ai);
		} else {
		    Complex         z;
		    z.real = floatit(ar);
		    z.imag = floatit(ai);
		    z = c_fn(z);
		    return make_complex(z.real, z.imag);
		}
#endif
	    }
	case N_BEFLOAT:
	    a = Lbfloat(a);
	case N_BFLOAT:
	    return b_fn(a);
	}
    }
#ifdef ASUKA
    {
	Sex             p;
	a = Llist2(fn_symbol, a);
	stackcheck();
	push(a);
	p = new_prefix();
	pop(a);
	prefix_form(p) = a;
	return p;
    }
#else
    return aerror1(msg, a);
#endif
}


Sex
Lsin(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, sin, dsin, my_csin, dcsin, sin_symbol);
#else
    return trig_fn(a, sin, dsin, my_csin, dcsin, "sin");
#endif
}

Sex
Lcos(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, cos, dcos, my_ccos, dccos, cos_symbol);
#else
    return trig_fn(a, cos, dcos, my_ccos, dccos, "cos");
#endif
}

Sex
Ltan(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, tan, dtan, my_ctan, dctan, tan_symbol);
#else
    return trig_fn(a, tan, dtan, my_ctan, dctan, "tan");
#endif
}

Sex
Lasin(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, asin, dasin, my_casin, dcasin, asin_symbol);
#else
    return trig_fn(a, asin, dasin, my_casin, dcasin, "asin");
#endif
}

Sex
Lacos(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, acos, dacos, my_cacos, dcacos, acos_symbol);
#else
    return trig_fn(a, acos, dacos, my_cacos, dcacos, "acos");
#endif
}

Sex
Latan(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, atan, datan, my_catan, dcatan, atan_symbol);
#else
    return trig_fn(a, atan, datan, my_catan, dcatan, "atan");
#endif
}

Sex
Lsinh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, sinh, dsinh, my_csinh, dcsinh, sinh_symbol);
#else
    return trig_fn(a, sinh, dsinh, my_csinh, dcsinh, "sinh");
#endif
}

Sex
Lcosh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, cosh, dcosh, my_ccosh, dccosh, cosh_symbol);
#else
    return trig_fn(a, cosh, dcosh, my_ccosh, dccosh, "cosh");
#endif
}

Sex
Ltanh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, tanh, dtanh, my_ctanh, dctanh, tanh_symbol);
#else
    return trig_fn(a, tanh, dtanh, my_ctanh, dctanh, "tanh");
#endif
}

Sex
Lasinh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, asinh, dasinh, my_casinh, dcasinh, asinh_symbol);
#else
    return trig_fn(a, asinh, dasinh, my_casinh, dcasinh, "asinh");
#endif
}

Sex
Lacosh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, acosh, dacosh, my_cacosh, dcacosh, acosh_symbol);
#else
    return trig_fn(a, acosh, dacosh, my_cacosh, dcacosh, "acosh");
#endif
}

Sex
Latanh(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, atanh, datanh, my_catanh, dcatanh, atanh_symbol);
#else
    return trig_fn(a, atanh, datanh, my_catanh, dcatanh, "atanh");
#endif
}

Sex
Lexp(Sex a)
{
    if (((long) a & TAG_BITS) == TAG_NUMBERS && numtype_of(a) == N_EFLOAT) {
	double          v = efloat_value(a);
	double          e = efloat_error(a);
	double          vp = v + e;
	double          vm = v - e;
	vp = exp(vp);
	vm = exp(vm);
	v = 0.5 * (vp + vm);
	e = vp - v;
	return make_efloat(v, e);
    }
#ifdef ASUKA
    return trig_fn(a, exp, dexp, my_cexp, dcexp, exp_symbol);
#else
    return trig_fn(a, exp, dexp, my_cexp, dcexp, "exp");
#endif
}

Sex
Llog(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, log, dlog, my_clog, dclog, log_symbol);
#else
    return trig_fn(a, log, dlog, my_clog, dclog, "log");
#endif
}

Sex
Llog2(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, my_log2, dlog2, my_clog2, dclog2, log2_symbol);
#else
    return trig_fn(a, my_log2, dlog2, my_clog2, dclog2, "log2");
#endif
}

Sex
Llog10(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, flog10, dlog10, my_clog10, dclog10, log10_symbol);
#else
    return trig_fn(a, flog10, dlog10, my_clog10, dclog10, "log10");
#endif
}

Sex
Lsqrt(Sex a)
{
    if (((long) a & TAG_BITS) == TAG_NUMBERS && numtype_of(a) == N_EFLOAT) {
	double          v = efloat_value(a);
	double          e = efloat_error(a);
	double          vp = v + e;
	double          vm = v - e;
	if (vm < 0)
	    vm = 0;
	vp = sqrt(vp);
	vm = sqrt(vm);
	v = 0.5 * (vp + vm);
	e = vp - v;
	return make_efloat(v, e);
    }
#ifdef ASUKA
    return trig_fn(a, sqrt, dsqrt, my_csqrt, dcsqrt, sqrt_symbol);
#else
    return trig_fn(a, sqrt, dsqrt, my_csqrt, dcsqrt, "sqrt");
#endif
}

Sex
Lcbrt(Sex a)
{
#ifdef ASUKA
    return trig_fn(a, cbrt, dcbrt, ccbrt, dccbrt, cbrt_symbol);
#else
    return trig_fn(a, cbrt, dcbrt, ccbrt, dccbrt, "cbrt");
#endif
}


Sex
Lasec(Sex a)
{
    return Lacos(quot2(mkfxat(1), a));
}

Sex
Lasech(Sex a)
{
    return Lacosh(quot2(mkfxat(1), a));
}

Sex
Lacsc(Sex a)
{
    if (zerop(a))
	return mkfxat(0);
    else
	return Lasin(quot2(mkfxat(1), a));
}

Sex
Lacsch(Sex a)
{
    if (zerop(a))
	return mkfxat(0);
    else
	return Lasinh(quot2(mkfxat(1), a));
}

Sex
Lacot(Sex a)
{
    if (zerop(a))
	return quot2(Lpi(), 2);
    else
	return Latan(quot2(mkfxat(1), a));
}

Sex
Lacoth(Sex a)
{
    if (zerop(a))
	return mkfxat(0);
    else
	return Latanh(quot2(mkfxat(1), a));
}


Sex
Lsec(Sex a)
{
    return quot2(mkfxat(1), Lcos(a));
}

Sex
Lsech(Sex a)
{
    return quot2(mkfxat(1), Lcosh(a));
}

Sex
Lcsc(Sex a)
{
    return quot2(mkfxat(1), Lsin(a));
}

Sex
Lcsch(Sex a)
{
    return quot2(mkfxat(1), Lsinh(a));
}

Sex
Lcot(Sex a)
{
    return quot2(mkfxat(1), Ltan(a));
}

Sex
Lcoth(Sex a)
{
    return quot2(mkfxat(1), Ltanh(a));
}



Sex
Lexpt(Sex a, Sex b)
{
    if (is_fixnum(b)) {
	long            n = mkint(b);
	if (n < 0) {
	    a = expti(a, -n);
	    a = ratio2(mkfxat(1), a);
	} else
	    a = expti(a, n);
	return a;
    }
    stackcheck();
    push(b);
    a = Llog(a);
    pop(b);
    a = times2(a, b);
    return Lexp(a);
}


#ifdef ASUKA
Sex
Aexpti(Sex a, long n)
{
    if (n == 0)
	return mkfxat(1);	/* 
				 * value 1 of appropriate type 
				 */
    else if (n == 1)
	return a;

    else if ((n & 1) == 0) {
	a = Aexpti(a, n / 2);
	return gtimes2(a, a);
    } else {
	Sex             b;
	stackcheck();
	push(a);
	b = Aexpti(a, n / 2);
	b = gtimes2(b, b);
	pop(a);
	return gtimes2(a, b);
    }
}

Sex
Aexpt(Sex a, Sex b)
{
    if (is_fixnum(b)) {
	long            n = mkint(b);
	if (n < 0) {
	    a = Aexpti(a, -n);
	    a = gratio2(mkfxat(1), a);
	} else
	    a = Aexpti(a, n);
	return a;
    }
    stackcheck();
    push(b);
    a = Llog(a);
    pop(b);
    a = gtimes2(a, b);
    return Lexp(a);
}

#endif

Sex
Llog_2(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = Llog(a);
    pop(b);
    push(a);
    b = Llog(b);
    pop(a);
    return quot2(a, b);
}


Sex
Labs(Sex a)
{
    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	break;
    case TAG_NUMBERS:
	switch (numtype_of(a)) {
	case N_COMPLEX:
	    {
		double          cr,
		                ci;
		double          d;
		cr = floatit(complex_real(a));
		ci = floatit(complex_imag(a));
		d = sqrt(cr * cr + ci * ci);
		a = make_float(d);
		return a;
	    }
	    break;
	case N_ECOMPLEX:
	case N_BECOMPLEX:
	    return aerror1("bad arg for abs", a);
	}
	break;
    default:
	return aerror1("bad arg for abs", a);
    }
    if (minusp(a))
	a = negate(a);
    return a;
}

Sex
Lhypot(Sex a, Sex b)
{
    stackcheck();
    push(b);
    a = times2(a, a);
    pop(b);
    push(a);
    b = times2(b, b);
    pop(a);
    a = plus2(a, b);
    return Lsqrt(a);
}


Sex
Latan2(Sex a, Sex b)
{
    double          u,
                    v,
                    r;
    long            q = 0;
    v = floatit(a);
    u = floatit(b);
    if (u < 0.0)
	u = -u, q = 1;
    if (v < 0.0)
	v = -v, q |= 2;
    if (v > u) {
	r = u;
	u = v;
	v = r;
	q |= 4;
    }
    if (u == 0.0 && v == 0.0)
	r = 0.0;
    else {
	r = atan(v / u);
	switch (q) {
	default:
	case 0:
	    break;
	case 1:
	    r = _pi - r;
	    break;
	case 2:
	    r = -r;
	    break;
	case 3:
	    r = -_pi + r;
	    break;
	case 4:
	    r = _half_pi - r;
	    break;
	case 5:
	    r = _half_pi + r;
	    break;
	case 6:
	    r = -_half_pi + r;
	    break;
	case 7:
	    r = -_half_pi - r;
	    break;
	}
    }
    a = make_float(r);
    return a;
}





Sex
Lpi(void)
{
    Sex             a;
    long            k,
                    m,
                    x;
    k = current_b_precision + 3;
    stackcheck();
    push3(nil, nil, nil);
#define n	stack[0]
#define s	stack[-1]
#define ss	stack[-2]
    ss = quot2(expti(mkfxat(2), k), mkfxat(5));
    n = ss;
    x = -25;
    m = 1;
    while (n != mkfxat(0)) {
	n = quot2(n, mkfxat(x));
	m = m + 2;
	a = quot2(n, mkfxat(m));
	ss = plus2(ss, a);
    }
    s = quot2(expti(mkfxat(2), k), mkfxat(239));
    n = s;
    x = -239 * 239;
    m = 1;
    while (n != mkfxat(0)) {
	n = quot2(n, mkfxat(x));
	m = m + 2;
	a = quot2(n, mkfxat(m));
	s = plus2(s, a);
    }
    ss = times2(ss, mkfxat(16));
    s = times2(s, mkfxat(4));
    a = diff2(ss, s);
#undef s
#undef ss
#undef n
    popv(3);
    return make_bfloat(a, -k);
}

Sex
Le(void)
{
    Sex             a;
    long            k = current_b_precision + 7;
    long            m = 1;

    stackcheck();
    push2(nil, nil);
#define n stack[0]
#define ans stack[-1]
    n = leftshifti(mkfxat(1), k);
    ans = mkfxat(0);
    while (n != mkfxat(0)) {
	m++;
	n = quot2(n, mkfxat(m));
	ans = plus2(ans, n);
    }
    n = leftshifti(mkfxat(1), k + 1);
    a = plus2(ans, n);
#undef n
#undef ans
    popv(2);
    return make_bfloat(a, -k);
}



Sex
Le_x(Sex x)
{
    Sex             a;
    long            k = current_b_precision + 7;
    long            m;

    ASSERT(is_bfloat(x));
    stackcheck();
    push3(x, nil, nil);
#define x stack[-2]
#define n stack[0]
#define ans stack[-1]
    x = leftshift(bfloat_man(x), bfloat_exp(x) + k);
    n = leftshifti(mkfxat(1), k);
    ans = n;
    m = 1;
    while (n != mkfxat(0)) {
	n = quot2(n, mkfxat(m));
	n = times2(n, x);
	n = leftshift(n, -k);
	ans = plus2(ans, n);
	m++;
    }
    a = ans;
#undef x
#undef n
#undef ans
    popv(3);
    return make_bfloat(a, -k);
}


Sex
ifactorial_part(long m, long n)
{
    long            p;
    Sex             a,
                    b;
    if (m == n)
	return mkfxat(n);
    else if (m == n - 1)
	return times2ii(mkfxat(m), mkfxat(n));
    p = (n + m) / 2;
    a = ifactorial_part(m, p);
    stackcheck();
    push(a);
    b = ifactorial_part(p + 1, n);
    pop(a);
    return times2jj(a, b);
}


Sex
ifactorial(long n)
{
    long            s = 1;
    long            i;
    Sex             a;
    if (n < 10) {
	for (i = 2; i <= n; i++)
	    s = s * i;
	return mkfxat(s);
    }
    if (n > 20)
	return ifactorial_part(1, n);

    s = 3628800;		/* 
				 * factorial(10) 
				 */
    a = mkfxat(s);
    for (i = 11; i <= n; i++)
	a = times2ji(a, mkfxat(i));
    return a;
}

Sex
ffactorial(double x)
{
    return nil;
}

Sex
bffactorial(Sex a)
{
    return nil;
}

Sex
Lpfactorial(Sex a, Sex b)
{
    long            n,
                    m;
    n = mkint(a);
    m = mkint(b);
    return ifactorial_part(n, m);
}


Sex
Lfactorial(Sex a)
{
    long            n;
    double          x;

    switch ((long) a & TAG_BITS) {
    case TAG_FIXNUM:
	n = mkint(a);
	return ifactorial(n);
    case TAG_NUMBERS:
	switch (numtype_of(a)) {
	case N_BIGNUM:
	case N_RATIONAL:
	case N_INTERVAL:
	case N_EFLOAT:
	    x = floatit(a);
	    return ffactorial(x);
	case N_FLOAT:
	    x = float_val(a);
	    return ffactorial(x);
	case N_BINTERVAL:
	case N_BEFLOAT:
	    a = bfloatit(a);
	    return bffactorial(a);
	case N_BFLOAT:
	    return bffactorial(a);
	case N_COMPLEX:
	    break;
	}
    }
    return aerror1("factorial", a);

}

Sex
ibinomial(long n, long m)
{
    Sex             a,
                    b;

    if (n < m || m < 0)
	aerror2("binomial", mkfxat(n), mkfxat(m));
    if (m > n - m)
	m = n - m;
    stackcheck();
    a = ifactorial_part(n - m + 1, n);
    push(a);
    b = ifactorial_part(1, m);
    pop(a);
    a = quot2(a, b);
    return a;
}

Sex
Lbinomial(Sex a, Sex b)
{
    if (is_fixnum(a) && is_fixnum(b))
	return ibinomial(mkint(a), mkint(b));
    return aerror2("binomial", a, b);
}


/*
 * Sex CC_csc(Sex a) {return Lcsc(a);} Sex CC_cot(Sex a) {return Lcot(a);}
 * Sex CC_coth(Sex a) {return Lcoth(a);} Sex CC_csch(Sex a) {return
 * Lcsch(a);} Sex CC_acot(Sex a) {return Lacot(a);} Sex CC_acsc(Sex a)
 * {return Lacsc(a);}  Sex CC_acoth(Sex a) {return Lacoth(a);} Sex
 * CC_acsch(Sex a) {return Lacsch(a);} 
 */


setup_type const trig_setup[] = {
    {"abs", (lisp_proc *) Labs, 1},
    {"acos", (lisp_proc *) Lacos, 1},
    {"acosh", (lisp_proc *) Lacosh, 1},
    {"asec", (lisp_proc *) Lasec, 1},
    {"asech", (lisp_proc *) Lasech, 1},
    {"asin", (lisp_proc *) Lasin, 1},
    {"asinh", (lisp_proc *) Lasinh, 1},
    {"atan2", (lisp_proc *) Latan2, 2},
    {"atanh", (lisp_proc *) Latanh, 1},
    {"cbrt", (lisp_proc *) Lcbrt, 1},
    {"cos", (lisp_proc *) Lcos, 1},
    {"cosh", (lisp_proc *) Lcosh, 1},
    {"exp", (lisp_proc *) Lexp, 1},
#ifdef ASUKA
    {"expt", (lisp_proc *) Aexpt, 2},
#else
    {"expt", (lisp_proc *) Lexpt, 2},
#endif
    {"hypot", (lisp_proc *) Lhypot, 2},
    {"ln", (lisp_proc *) Llog, 1},
    {"log", (lisp_proc *) Llog, 1},
    {"log10", (lisp_proc *) Llog10, 1},
    {"sec", (lisp_proc *) Lsec, 1},
    {"sech", (lisp_proc *) Lsech, 1},
    {"sin", (lisp_proc *) Lsin, 1},
    {"sinh", (lisp_proc *) Lsinh, 1},
    {"sqrt", (lisp_proc *) Lsqrt, 1},
    {"tan", (lisp_proc *) Ltan, 1},
    {"tanh", (lisp_proc *) Ltanh, 1},
    {"logb", (lisp_proc *) Llog_2, 2},	/* 
					 * 2 arg version 
					 */
    {"atan", (lisp_proc *) Latan, 1},

    {"csc", (lisp_proc *) Lcsc, 1},
    {"csch", (lisp_proc *) Lcsch, 1},
    {"cot", (lisp_proc *) Lcot, 1},
    {"coth", (lisp_proc *) Lcoth, 1},
    {"acsc", (lisp_proc *) Lacsc, 1},
    {"acsch", (lisp_proc *) Lacsch, 1},
    {"acot", (lisp_proc *) Lacot, 1},
    {"acoth", (lisp_proc *) Lacoth, 1},

    {"pi", (lisp_proc *) Lpi, 0},
    {"e", (lisp_proc *) Le, 0},
    {"e_x", (lisp_proc *) Le_x, 1},
    {"factorial", (lisp_proc *) Lfactorial, 1},
    {"pfactorial", (lisp_proc *) Lpfactorial, 2},
    {"binomial", (lisp_proc *) Lbinomial, 2},
    {NULL, 0, 0}
};
